// CoCoA Application (QT, Portable graphics interaction)

//#ifdef CA_QT

#include <fstream>
#include <stdio.h>
#include <signal.h>

#include "qtmultilineedit.h"
#include <qthread.h>

#include "ca_qt.h"
#include "engine_thread.h"

// [ PUBLIC ] [ PRIVATE ] [ CDEV ]

// ---[ PUBLIC ]-----------------------------------------------

extern "C" void ca_terminate(void);

/* VECCHIO MAIN
int main_old(void) 
{
  int n;
  cterm t;

  n=ca_init();

  if (n != ck_NOERR) return n;

  n = ck_source_from_string("<< CocoaPackagePath()+'/cocoa/init.coc';");
  if (n != ck_NOERR && n != ck_ERR_interrupted) return n;
  do {
    ca_interrupted=FALSE;

    t=ca_get(cdev_stdin);
    ck_engine_top(t);

  } while (! ca_finished()); // user asked for ca to finish
  ca_terminate();
  return 0;
}
*/

bool ca_finished()
{ 
  return ck_finished();
}

void ca_continue()
{
	ck_continue();
}

void ca_monitor(int what)
{
// do nothing (for now)
}

int ca_message(cterm t)
{
  ca_put(t,cdev_out);
  return ca_NOERR;
}

int ca_warning(cterm t)
{
  ca_message_string("-- WARNING: ");
  ca_put(t,cdev_out);
  ca_put_nl(cdev_out);
  //ca_put_dashes(cdev_out);
  return ca_NOERR;
}

int ca_internal_error(cterm t)
{
  ca_message_string("-- INTERNAL ERROR ");
  ca_put(t,cdev_out);
  ca_put_nl(cdev_out);
  return ca_NOERR;
}

/* ---[ PRIVATE ]----------------------------------------------- */

void disp_ctrlc(int n) {
  signal(SIGINT,&disp_ctrlc);
  ca_message_string("[ Caught interruption ]\n");
  ca_interrupted=TRUE;
}

int ca_init_cdevs(void)
{
  cdev_stdin = cdev_init("stdin",cdev_type_ifile, cdev_prot_cocoal);
  cdev_stdout = cdev_init("stdout",cdev_type_owindow, cdev_prot_cocoal);
  cdev_stderr = cdev_init("stderr",cdev_type_owindow, cdev_prot_cocoal);
  cdev_set_ifile(cdev_stdin,stdin);
  cdev_parser = cdev_link(cdev_stdin);
  cdev_monitor = cdev_link(cdev_stdout);
  return ca_NOERR;
}

int ca_init(void)
{
  int n;
//  signal(SIGINT,&disp_ctrlc);
  ca_pi_init();
  ca_init_cdevs();
  cterm_initialize();
  n=ck_init(); // cocoa kernel init
  if (n != ck_NOERR) return n;
  return ca_NOERR;
}

void ca_terminate(void)
{
  ck_quit();
  ca_message_string("-- Bye\n");
}


// ---[ CDEV ]----------------------------------------------- 

int cdev_puts_ofile(char * s, cdev d)
{
  fputs(s,cdev_the_ofile(d));
  return ca_NOERR;
}

cdev cdev_open_ifile(char * filename, int prot) {
  cdev res_d;
  FILE * f;
  f = fopen(filename,"r");
  if (!f) return cdev_NULL;
  res_d=cdev_init(filename, cdev_type_ifile, prot);
  cdev_set_ifile(res_d,f);
  return res_d;
}

cdev cdev_open_ofile(char * filename, int prot, cdev_ofile_params p) {
  cdev res_d;
  FILE * f;
  if (strcmp(filename,"") == 0 || strcmp(filename,"stdout") == 0) {
    f = stdout;
  } else {
    char *params = (char*) p;
    if (params && !strcmp(params, "w"))
        f = fopen(filename, "w");
    else
        f = fopen(filename, "a");
  }
  if (!f) return cdev_NULL;
  res_d=cdev_init(filename, cdev_type_ofile, prot);
  cdev_set_ofile(res_d,f);
  return res_d;
}

cdev cdev_open_iwindow(char * filename, int prot) {
  return cdev_NULL;
}

int cdev_close_ofile(cdev D) {
  fclose(cdev_the_ofile(D));
  return ca_NOERR;
}

int cdev_close_ifile(cdev D) {
  fclose(cdev_the_ifile(D));
  return ca_NOERR;
}

int cdev_flush_ofile(cdev D) {
  fflush(cdev_the_ofile(D));
  return ca_NOERR;
}

char cdev_getc_ifile(cdev d) {
  return getc(cdev_the_ifile(d));
}

int cdev_putc_ofile(char c, cdev d) {
//cerr << "CDEV_PUTC_OFILE";
  putc(c,cdev_the_ofile(d));
  return ca_NOERR;
}

void cdev_free_iwindow(cdev d) {}

coc_bool cdev_eof(cdev d) {
  int res_n;
  if (d==cdev_NULL) return TRUE;
  switch (cdev_type(d)) {
  case cdev_type_ifile:
    res_n =feof(cdev_the_ifile(d));
    break;
  case cdev_type_istring:
    res_n=cdev_istring_eof(cdev_the_istring(d));
    break;
  default: return TRUE;
  }
  if (res_n) return TRUE; else return FALSE;
}

//
// owindow (QT) device handling -- Da sistemare
//
cdev cdev_open_owindow(char * filename, int prot, cdev_owindow_params p) {
	// TODO! Questo codice probabilmente e' superfluo... Controllare
  cdev res_d;
  res_d=cdev_init(filename, cdev_type_owindow, prot);
  return res_d;
}

int cdev_close_owindow(cdev D) {
  return ca_NOERR;
}

int cdev_puts_owindow(char * s, cdev d) {
	EngineThread::currentEngineThread()->addString(s);
	return ca_NOERR;
}

void cdev_free_owindow(cdev d) {}

int cdev_putc_owindow(char c, cdev d) {
	EngineThread::currentEngineThread()->addChar(c);
	return ca_NOERR;
}

int cdev_flush_owindow(cdev D) {
	EngineThread::currentEngineThread()->flush();
	return ca_NOERR;
}

int cdev_put_dashes_owindow(cdev d)
{
	ca_put(cterm_dashes,d);
	/*
	EngineThread::currentEngineThread()->addString("-------------------------------------------------------\n");
	EngineThread::currentEngineThread()->flush();
	*/
	return ca_NOERR;
}

// ---[ pretty printing (IO dep.) ]---------------------------------------- 

int pp_pow(cdev d, cterm t)
{
  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  char * exponent;
  exponent=cterm_string(params[1]);
  if (strcmp(exponent,"0")!=0) {
    pp_indet(d,params[0]);
    if (strcmp(exponent,"1")!=0) {
      cdev_puts("^",d);
      cdev_puts(exponent,d); } }
  return ca_NOERR;
}
//#endif
