;; cocoa.el --- major mode for editing and running CoCoA
(setq cocoa-mode-version "18 Dec 2003")
;; by: Giovanna Roda, Giorgio Vecchiocattivi and Anna Bigatti
;; maintained by: Anna Bigatti <bigatti@dima.unige.it>

;;==============================================================>
;;     CUSTOMIZATION moved into cocoa.emacs
;;==============================================================>

;;==============================================================>
;  \M-x cocoa
;; will open a buffer called *cocoa* with a running CoCoA

;; you can edit a CoCoA file and send commands from it to *cocoa*:
;   "\C-c\C-l" 'cocoa-send-line
;   "\C-c\C-b" 'cocoa-send-buffer
;   "\C-c\C-r" 'cocoa-send-region
;   "\C-c\C-c" 'comment-region
;   "\C-c\C-e" 'cocoa-close-block  ;; writes the appropriate "End"
;   "\C-c\C-a" 'cocoa-send-file
;   "\C-c\C-w" 'cocoa-open-wordlist-file
;   "\C-c\C-n" 'cocoa-open-file
;   "\C-c\C-q" 'cocoa-shell-quit
;   "\C-c?"    'cocoa-word-man
;   "\C-cm"    'cocoa-word-man
;   "\C-c\C-m" 'cocoa-word-man
;;==============================================================>




;;-------------------------------------------------------------;;
;;                    EDITING CoCoA files                      ;;
;;-------------------------------------------------------------;;

;;-------------------------------------------------------------->
;; abbrev-table
;;-------------------------------------------------------------->
;; to use this, in cocoa-mode, type Esc-x abbrev-mode.
;; it capitalizes the CoCoA keywords 
(defvar cocoa-mode-abbrev-table nil
  "Abbrev table in use in CoCoA-mode buffers.")

(define-abbrev-table 'cocoa-mode-abbrev-table '(
    ("alias"   		"Alias"   	nil 0)
    ("and"    		"And"   	nil 0)
    ("ciao"    		"Ciao;"   	nil 0)
    ("cond"    		"Cond"    	nil 0)
    ("define"  		"Define"  	nil 0)
    ("deglex"		"DegLex"     	nil 0)
    ("degrevlex"	"DegRevLex"     nil 0)
    ("do"      		"Do"      	nil 0)
    ("else"    		"Else"    	nil 0)
    ("elsif"   		"ElsIf"   	nil 0)
    ("end"     		"End;"    	nil 0)
    ("endalias"         "EndAlias;"   	nil 0)
    ("endblock"         "EndBlock;"   	nil 0)
    ("endcatch"         "EndCatch;"   	nil 0)
    ("endcond"          "EndCond;"   	nil 0)
    ("enddefine"        "EndDefine;"   	nil 0)
    ("endfor"           "EndFor;"   	nil 0)
    ("endforeach"       "EndForeach;"  	nil 0)
    ("endif"            "EndIf;"   	nil 0)
    ("endpackage"       "EndPackage;"  	nil 0)
    ("endrepeat"        "EndRepeat;"  	nil 0)
    ("endusing"         "EndUsing;"   	nil 0)
    ("endwhile"         "EndWhile;"   	nil 0)
    ("error"   		"Error"    	nil 0)
    ("false"     	"False"     	nil 0)
    ("for"     		"For"     	nil 0)
    ("foreach" 		"Foreach" 	nil 0)
    ("ideal"      	"Ideal"      	nil 0)
    ("if"      		"If"      	nil 0)
    ("in"      		"In"      	nil 0)
    ("isin"            	"IsIn"      	nil 0)
    ("indentation"	"Indentation;"	nil 0)
    ("lex"    		"Lex"  	        nil 0)
    ("mat"    		"Mat"  	        nil 0)
    ("module"    	"Module"        nil 0)
    ("not"	        "Not"           nil 0)
    ("on"    		"On"   	        nil 0)
    ("or"    		"Or"   	        nil 0)
    ("posto"   		"PosTo"   	nil 0)
    ("print"   		"Print"   	nil 0)
    ("println" 		"PrintLn" 	nil 0)
    ("record"  		"Record"  	nil 0)
    ("return"  		"Return"  	nil 0)
    ("set"     		"Set"   	nil 0)
    ("step"    		"Step"   	nil 0)
    ("quit"    		"Quit;"   	nil 0)
    ("then"    		"Then"    	nil 0)
    ("time"    		"Time"    	nil 0)
    ("true"    		"True"    	nil 0)
    ("to"      		"To"      	nil 0)
    ("topos"   		"ToPos"      	nil 0)
    ("unset"     	"Unset"     	nil 0)
    ("use"     		"Use"     	nil 0)
    ("using"   		"Using"   	nil 0)
    ("var"   		"Var"   	nil 0)
    ("vector"  		"Vector"   	nil 0)
    ("while"   		"While"   	nil 0)
    ("xel"    		"Xel"  	        nil 0)
    ))

(defvar cocoa-imenu-generic-expression
  '("^[ \t]*\\(Define\\)[ \t\n]+\\([a-zA-Z0-9_.:]+\\)" . (2))
  "Imenu expression for CoCoA-mode.  See `imenu-generic-expression'.")

(defconst cocoa-startkeywords-list
(concat "\\(Alias\\|Block\\|Catch\\|Cond\\|Define\\|For\\|Foreach\\|"
	"If\\|Package\\|Using\\|While\\)"
	))
(defconst cocoa-endkeywords-list
(concat "\\<End" 
	"\\(\\|Alias\\|Block\\|Catch\\|Cond\\|Define\\|For\\|Foreach\\|"
	"If\\|Package\\|Using\\|While\\)"
	"\\>"))

;;-------------------------------------------------------------->
;; Regular expressions used to calculate indent, etc.
;;-------------------------------------------------------------->

(defconst cocoa-symbol-re      "\\<[a-zA-Z_][a-zA-Z_0-9.]*\\>")
(defconst cocoa-beg-block-re
"\\<\\(Alias\\|Block\\|Catch\\|Define\\|If\\|Foreach\\|For\\|Repeat\\|Using\\|While\\|Cond\\)\\>")
(defconst cocoa-end-block-re
(concat "\\<\\("
	cocoa-endkeywords-list "\\|"
	"Until"
	"\\)\\>"))
(defconst cocoa-declaration-re "\\<\\(NIENTE_DA_FISSARE\\)\\>")
(defconst cocoa-defun-re       "\\<\\(Define\\)\\>")
(defconst cocoa-sub-block-re   
"\\<\\(Alias\\|Block\\|Catch\\|If\\|Foreach\\|For\\|Repeat\\|Using\\|While\\|Cond\\)\\>")
(defconst cocoa-noindent-re
(concat "\\<\\("
	cocoa-endkeywords-list "\\|"
	"Else\\|Elsif\\|Until"
	"\\)\\>"))
(defconst cocoa-nosemi-re      "\\<\\(Then\\|Do\\|Else\\)\\>")
(defconst cocoa-autoindent-lines-re
(concat "\\<\\("
	cocoa-endkeywords-list "\\|"
	"Define\\|While\\|Else"
	"\\)\\>"))

;;; Strings used to mark beginning and end of excluded text
(defconst cocoa-exclude-str-start "{-----\\/----- EXCLUDED -----\\/-----")
(defconst cocoa-exclude-str-end " -----/\\----- EXCLUDED -----/\\-----}")

;;-------------------------------------------------------------->
;; cocoa-mode-syntax-table
;;-------------------------------------------------------------->

(defvar cocoa-mode-syntax-table nil
  "Syntax table in use in cocoa-mode buffers.")

(if cocoa-mode-syntax-table
    ()
  (setq cocoa-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "."   cocoa-mode-syntax-table)
  (modify-syntax-entry ?( "()"  cocoa-mode-syntax-table)
  (modify-syntax-entry ?) ")("  cocoa-mode-syntax-table)
;  (modify-syntax-entry ?* "." cocoa-mode-syntax-table)
  (modify-syntax-entry ?+ "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?= "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?% "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?< "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?> "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?& "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?| "."    cocoa-mode-syntax-table)
  (modify-syntax-entry ?_ "w"    cocoa-mode-syntax-table)
  ; string
  (modify-syntax-entry ?\' "\""  cocoa-mode-syntax-table)
  ; -- comments 
  (modify-syntax-entry ?- ". 12"    cocoa-mode-syntax-table)
  ; // and /* */ comments 
  (modify-syntax-entry ?/  ". 124" cocoa-mode-syntax-table)
  (modify-syntax-entry ?*  ". 23b" cocoa-mode-syntax-table)
  (modify-syntax-entry ?\n ">"   cocoa-mode-syntax-table)
)

;;-------------------------------------------------------------->
;; font-lock-keywords
;;-------------------------------------------------------------->

(defvar cocoa-font-lock-keywords
  (list
   ; function definition
   '("^[ \t]*\\(Define\\)\\>[ \t]*\\(\\sw+\\)?"
     (1 font-lock-keyword-face) (2 font-lock-function-name-face nil t))
   ; types
   (cons    (concat "\\<\\("
	   "NULL\\|BOOL\\|STRING\\|TYPE\\|ERROR\\|RECORD\\|DEVICE\\|"
	   "INT\\|RAT\\|ZMOD\\|POLY\\|RATFUN\\|VECTOR\\|IDEAL\\|MODULE\\|"
	   "MAT\\|LIST\\|RING\\|TAGGED\\|FUNCTION"
	   "\\)\\>")
	 'font-lock-type-face)
   ; constants
   (cons   (concat "\\<\\("
		   "TRUE\\|FALSE\\|True\\|False\\|"
		   "Lex\\|Xel\\|DegLex\\|DegRevLex\\|"
		   "PosTo\\|ToPos\\|Null\\)\\>") 
	   'font-lock-constant-face)
   ; Ideal, Module...
   (cons   (concat "\\<\\("
		   "Ideal\\|Module\\|Mat\\|Record\\|Vector\\|Error"
		   "\\)\\>") 
	   'font-lock-builtin-face)
   ; keywords
   (concat "\\<\\("
	   "And\\|Or\\|Not\\|Do\\|"
	   "Els\\(e\\|If\\|if\\)\\|"
	   cocoa-startkeywords-list "\\|"
	   cocoa-endkeywords-list "\\|"
	   "In\\|IsIn\\|"
	   "On\\|Print\\(\\|Ln\\)\\|"
	   "Repeat\\|Until\\|Return\\|Set\\|Step\\|"
	   "T\\(hen\\|ime\\|o\\)\\|"
	   "Unset\\|Use\\|Var"
	   "\\)\\>")
   '("\\<\\(NULLA_DA_FISSARE\\)\\>[ \t]*\\([0-9]+\\)?"
     (1 font-lock-keyword-face) (2 font-lock-reference-face nil t)))
  "Additional expressions to highlight in CoCoA mode.")

;;-------------------------------------------------------------->

(defvar cocoa-indent-level 2
  "*Indentation of CoCoA statements with respect to containing block.")

(defvar cocoa-case-indent 2
  "*Indentation for case statements.")

(defvar cocoa-auto-newline nil
  "*Non-nil means automatically newline after semicolons and the punctuation
mark after an end.")

(defvar cocoa-tab-always-indent t
  "*Non-nil means TAB in CoCoA mode should always reindent the current line,
regardless of where in the line point is when the TAB command is used.")

(defvar cocoa-auto-endcomments t
  "*Non-nil means a comment { ... } is set after the ends which ends cases and
functions. The name of the function or case will be set between the braces.")

(defvar cocoa-auto-lineup '(all)
  "*List of contexts where auto lineup of :'s or ='s should be done.
Elements can be of type: 'paramlist', 'declaration' or 'case', which will
do auto lineup in parameterlist, declarations or case-statements
respectively. The word 'all' will do all lineups. '(case paramlist) for
instance will do lineup in case-statements and parameterlist, while '(all)
will do all lineups.")

(defvar cocoa-toggle-completions nil
  "*Non-nil means \\<cocoa-mode-map>\\[cocoa-complete-word] should try all
possible completions one by one.
Repeated use of \\[cocoa-complete-word] will show you all of them.
Normally, when there is more than one possible completion,
it displays a list of all possible completions.")

(defvar cocoa-type-keywords
  '("NULLA_DA_FISSARE_COCOA_TYPE_KEYWORDS")
  "*Keywords for types used when completing a word in a declaration or
parmlist.
\(eg. integer, real, char.)  The types defined within the CoCoA program
will be completed runtime, and should not be added to this list.")

(defvar cocoa-start-keywords
  '("Define" "End" "While")
  "*Keywords to complete when standing at the first word of a statement.
\(eg. begin, repeat, until, readln.)
The procedures and variables defined within the CoCoA program
will be completed runtime and should not be added to this list.")

(defvar cocoa-separator-keywords
  '("To" "Else" "Mod" "Div" "Then")
  "*Keywords to complete when NOT standing at the first word of a statement.
\(eg. downto, else, mod, then.)
Variables and function names defined within the
CoCoA program are completed runtime and should not be added to this list.")


;;;
;;;  Macros
;;;

(defsubst cocoa-get-beg-of-line (&optional arg)
  (save-excursion
    (beginning-of-line arg)
    (point)))

(defsubst cocoa-get-end-of-line (&optional arg)
  (save-excursion
    (end-of-line arg)
    (point)))

(defun cocoa-declaration-end ()
  (let ((nest 1))
    (while (and (> nest 0)
		(re-search-forward
		 "[:=]\\|\\(\\<record\\>\\)\\|\\(\\<end\\>\\)"
		 (save-excursion (end-of-line 2) (point)) t))
      (cond ((match-beginning 1) (setq nest (1+ nest)))
	    ((match-beginning 2) (setq nest (1- nest)))
	    ((looking-at "[^(\n]+)") (setq nest 0))))))


(defun cocoa-declaration-beg ()
  (let ((nest 1))
    (while (and (> nest 0)
		(re-search-backward
"[:=]\\|\\<\\(NULLA_DECLARATION_BEG\\)\\>\\|\\(\\<record\\>\\)\\|\\(\\<end\\>\\)
" (cocoa-get-beg-of-line 0) t))
      (cond ((match-beginning 1) (setq nest 0))
	    ((match-beginning 2) (setq nest (1- nest)))
	    ((match-beginning 3) (setq nest (1+ nest)))))
    (= nest 0)))


(defsubst cocoa-within-string ()
  (save-excursion
    (nth 3 (parse-partial-sexp (cocoa-get-beg-of-line) (point)))))




;;;
;;;  Electric functions
;;;
(defun electric-cocoa-terminate-line ()
  "Terminate line and indent next line."
  (interactive)
  ;; First, check if current line should be indented
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (if (looking-at cocoa-autoindent-lines-re)
	(cocoa-indent-line)))
  (delete-horizontal-space) ; Removes trailing whitespaces
  (newline)
  ;; Indent next line
  (cocoa-indent-line)
  ;; Maybe we should set some endcomments
  (if cocoa-auto-endcomments
      (cocoa-set-auto-comments))
  ;; Check if we shall indent inside comment
  (let ((setstar nil))
    (save-excursion
      (forward-line -1)
      (skip-chars-forward " \t")
      (cond ((looking-at "\\*[ \t]+/")
	     ;; Delete region between `*' and `/' if there is only whitespaces.
	     (forward-char 1)
	     (delete-horizontal-space))
	    ((and (looking-at "/\\*\\|\\*[^/]")
		  (not (save-excursion
			 (search-forward "\\*/" (cocoa-get-end-of-line) t))))
	     (setq setstar t))))
    ;; If last line was a star comment line then this one shall be too.
;    (if (null setstar)
;	(cocoa-indent-line)
;      (insert "*  "))
))

(defun electric-cocoa-semi-or-dot ()
  "Insert `;' or `.' character and reindent the line."
  (interactive)
  (insert last-command-char)
  (save-excursion
    (beginning-of-line)
    (cocoa-indent-line))
  (if cocoa-auto-newline
      (electric-cocoa-terminate-line)))

(defun electric-cocoa-colon ()
  "Insert `:' and do all indentions except line indent on this line."
  (interactive)
  (insert last-command-char)
  ;; Do nothing if within string.
  (if (cocoa-within-string)
      ()
    (save-excursion
      (beginning-of-line)
      (cocoa-indent-line))
    (let ((cocoa-tab-always-indent nil))
      (cocoa-indent-command))))

(defun electric-cocoa-equal ()
  "Insert `=', and do indention if within type declaration."
  (interactive)
  (insert last-command-char)
  (if (eq (car (cocoa-calculate-indent)) 'declaration)
      (let ((cocoa-tab-always-indent nil))
	(cocoa-indent-command))))

(defun electric-cocoa-hash ()
  "Insert `#', and indent to column 0 if this is a CPP directive."
  (interactive)
  (insert last-command-char)
  (if (save-excursion (beginning-of-line) (looking-at "^[ \t]*#"))
      (save-excursion (beginning-of-line)
		      (delete-horizontal-space))))

(defun electric-cocoa-tab ()
  "Function called when TAB is pressed in CoCoA mode."
  (interactive)
  ;; Do nothing if within a string or in a CPP directive.
  (if (or (cocoa-within-string)
	  (and (not (bolp))
	       (save-excursion (beginning-of-line) (eq (following-char) ?#))))
      (insert "\t")
    ;; If cocoa-tab-always-indent, indent the beginning of the line.
    (if cocoa-tab-always-indent
	(save-excursion
	  (beginning-of-line)
	  (cocoa-indent-line))
      (if (save-excursion
	    (skip-chars-backward " \t")
	    (bolp))
	  (cocoa-indent-line)
	(insert "\t")))
    (cocoa-indent-command)))



(defun cocoa-beg-of-defun ()
  "Move backward to the beginning of the current function or procedure."
  (interactive)
  (re-search-backward cocoa-defun-re nil 'move)
)

(defun cocoa-end-of-statement ()
  "Move forward to end of current statement."
  (interactive)
  (let ((parse-sexp-ignore-comments t)
	(nest 0) pos
	(regexp (concat "\\(" cocoa-beg-block-re "\\)\\|\\("
			cocoa-end-block-re "\\)")))
    (if (not (looking-at "[ \t\n]")) (forward-sexp -1))
    (or (looking-at cocoa-beg-block-re)
	;; Skip to end of statement
	(setq pos (catch 'found
		    (while t
		      (forward-sexp 1)
		      (cond ((looking-at "[ \t]*;")
			     (skip-chars-forward "^;")
			     (forward-char 1)
			     (throw 'found (point)))
			    ((save-excursion
			       (forward-sexp -1)
			       (looking-at cocoa-beg-block-re))
			     (goto-char (match-beginning 0))
			     (throw 'found nil))
			    ((eobp)
			     (throw 'found (point))))))))
    (if (not pos)
	;; Skip a whole block
	(catch 'found
	  (while t
	    (re-search-forward regexp nil 'move)
	    (setq nest (if (match-end 1)
			   (1+ nest)
			 (1- nest)))
	    (cond ((eobp)
		   (throw 'found (point)))
		  ((= 0 nest)
		   (throw 'found (cocoa-end-of-statement))))))
      pos)))

;;;
;;; Other functions
;;;
(defun cocoa-set-auto-comments ()
  "Insert `{ case }' or `{ NAME }' on this line if appropriate.
Insert `{ case }' if there is an `end' on the line which
ends a case block.  Insert `{ NAME }' if there is an `end'
on the line which ends a function or procedure named NAME."
  (save-excursion
    (forward-line -1)
    (skip-chars-forward " \t")
    (if (and (or (looking-at "\\<EndDefine;") (looking-at "\\<End;") )
	     (not (save-excursion
		    (end-of-line)
		    (search-backward "--" (cocoa-get-beg-of-line) t))))
	(let ((type (car (cocoa-calculate-indent))))
	  (if (eq type 'declaration)
	      ()
	    (if (eq type 'case)
		;; This is a case block
		(progn
		  (end-of-line)
		  (delete-horizontal-space)
		  (insert " -- case "))
	      (let ((nest 1))
		;; Check if this is the end of a function
		(save-excursion
		  (while (not (or (looking-at cocoa-defun-re) (bobp)))
		    (backward-sexp 1)
		    (cond ((looking-at cocoa-beg-block-re)
			   (setq nest (1- nest)))
			  ((looking-at cocoa-end-block-re)
			   (setq nest (1+ nest)))))
		  (if (bobp)
		      (setq nest 1)))
		(if (zerop nest)
		    (progn
		      (end-of-line)
		      (delete-horizontal-space)
		      (insert " -- ")
		      (let (b e)
			(save-excursion
			  (setq b (progn (cocoa-beg-of-defun)
					 (skip-chars-forward "^ \t")
					 (skip-chars-forward " \t")
					 (point))
				e (progn (skip-chars-forward "a-zA-Z0-9_")
					 (point))))
			(insert-buffer-substring (current-buffer) b e))
		      (insert ""))))))))))



;;;
;;; Indentation
;;;
(defconst cocoa-indent-alist
  '((block . (+ ind cocoa-indent-level))
    (case . (+ ind cocoa-case-indent))
    (caseblock . ind) (cpp . 0)
    (declaration . (+ ind cocoa-indent-level))
    (paramlist . (cocoa-indent-paramlist t))
    (comment . (cocoa-indent-comment t))
    (defun . ind) (contexp . ind)
    (unknown . 0) (string . 0)))

(defun cocoa-indent-command ()
  "Indent for special part of code."
  (let* ((indent-str (cocoa-calculate-indent))
	 (type (car indent-str))
	 (ind (car (cdr indent-str))))
    (cond ((and (eq type 'paramlist)
		(or (memq 'all cocoa-auto-lineup)
		    (memq 'paramlist cocoa-auto-lineup)))
	   (cocoa-indent-paramlist)
	   (cocoa-indent-paramlist))
	  ((and (eq type 'declaration)
		(or (memq 'all cocoa-auto-lineup)
		    (memq 'declaration  cocoa-auto-lineup)))
	   (cocoa-indent-declaration))
	  ((and (eq type 'case) (not (looking-at "^[ \t]*$"))
		(or (memq 'all cocoa-auto-lineup)
		    (memq 'case cocoa-auto-lineup)))
	   (cocoa-indent-case)))
    (if (looking-at "[ \t]+$")
	(skip-chars-forward " \t"))))

(defun cocoa-indent-line ()
  "Indent current line as a CoCoA statement."
  (let* ((indent-str (cocoa-calculate-indent))
	 (type (car indent-str))
	 (ind (car (cdr indent-str))))
    (if (looking-at "^[0-9a-zA-Z]+[ \t]*:[^:=]")
	(search-forward ":" nil t))
    (delete-horizontal-space)
    ;; Some things should not be indented
    (if (or (and (eq type 'declaration) (looking-at cocoa-declaration-re))
	    (eq type 'cpp)
	    (looking-at cocoa-defun-re))
	()
      ;; Other things should have no extra indent
      (if (looking-at cocoa-noindent-re)
	  (indent-to ind)
	;; But most lines are treated this way:
	(indent-to (eval (cdr (assoc type cocoa-indent-alist))))
	))))

(defun cocoa-calculate-indent ()
  "Calculate the indent of the current CoCoA line.
Return a list of two elements: (INDENT-TYPE INDENT-LEVEL)."
  (save-excursion
    (let* ((parse-sexp-ignore-comments t)
	   (oldpos (point))
	   (state (save-excursion (parse-partial-sexp (point-min) (point))))
	   (nest 0) (par 0) (complete (looking-at "[ \t]*End\\>"))
	   (elsed (looking-at "[ \t]*Else\\>"))
	   (type (catch 'nesting
		   ;; Check if inside a string, comment or parenthesis
		   (cond ((nth 3 state) (throw 'nesting 'string))
			 ((nth 4 state) (throw 'nesting 'comment))
			 ((> (car state) 0)
			  (goto-char (scan-lists (point) -1 (car state)))
			  (setq par (1+ (current-column))))
			 ((save-excursion (beginning-of-line)
					  (eq (following-char) ?#))
			  (throw 'nesting 'cpp)))
		   ;; Loop until correct indent is found
		   (while t
		     (if (bobp) (throw 'nesting 'unknown))
		     (backward-sexp 1)
		     (cond (;--Escape from case statements
			    (and (looking-at "[A-Za-z0-9]+[ \t]*:[^:=]")
				 (not complete)
				 (save-excursion (skip-chars-backward " \t")
						 (bolp))
				 (= (save-excursion
				      (end-of-line) (backward-sexp) (point))
				    (point))
				 (> (save-excursion (goto-char oldpos)
						    (beginning-of-line)
						    (point))
				    (point)))
			    (throw 'nesting 'caseblock))
			   (;--Nest block outwards
			    (looking-at cocoa-beg-block-re)
			    (if (= nest 0)
				(cond ((looking-at "NULLA_DA_FISSARE\\>")
				       (throw 'nesting 'case))
				      ((looking-at "NULLA_DA_FISSARE\\>")
				       (throw 'nesting 'declaration))
				      (t (throw 'nesting 'block)))
			      (setq nest (1- nest))))
			   (;--Nest block inwards
			    (looking-at cocoa-end-block-re)
			    (if (and (looking-at "End\\s ")
				     elsed (not complete))
				(throw 'nesting 'block))
			    (setq complete t
				  nest (1+ nest)))
			   (;--Defun (or parameter list)
			    (looking-at cocoa-defun-re)
			    (if (= 0 par)
				(throw 'nesting 'defun)
			      (setq par 0)
			      (let ((n 0))
				(while (re-search-forward

	"\\(\\<NULLA_DA_FISSARE\\>\\)\\|\\<NULLA_DA_FISSARE\\>"
					oldpos t)
				  (if (match-end 1)
				      (setq n (1+ n)) (setq n (1- n))))
				(if (> n 0)
				    (throw 'nesting 'declaration)
				  (throw 'nesting 'paramlist)))))
			   (;--Declaration part
			    (looking-at cocoa-declaration-re)
			    (if (save-excursion
				  (goto-char oldpos)
				  (forward-line -1)
				  (looking-at "^[ \t]*$"))
				(throw 'nesting 'unknown)
			      (throw 'nesting 'declaration)))
			   (;--If, else or while statement
			    (and (not complete)
				 (looking-at cocoa-sub-block-re))
			    (throw 'nesting 'block))
			   (;--Found complete statement
			    (save-excursion (forward-sexp 1)
					    (= (following-char) ?\;))
			    (setq complete t))
			   (;--No known statements
			    (bobp)
			    (throw 'nesting 'unknown))
			   )))))

      ;; Return type of block and indent level.
      (if (> par 0)                               ; Unclosed Parenthesis
	  (list 'contexp par)
	(list type (cocoa-indent-level))))))

(defun cocoa-indent-level ()
  "Return the indent-level the current statement has.
Do not count labels, case-statements or records."
  (save-excursion
    (beginning-of-line)
    (if (looking-at "[ \t]*[0-9a-zA-Z]+[ \t]*:[^:=]")
	(search-forward ":" nil t)
      (if (looking-at ".*=[ \t]*NULLA_DA_FISSARE\\>")
	  (search-forward "=" nil t)))
    (skip-chars-forward " \t")
    (current-column)))

(defun cocoa-indent-comment (&optional arg)
  "Indent current line as comment.
If optional arg is non-nil, just return the
column number the line should be indented to."
    (let* ((stcol (save-excursion
		    (re-search-backward "--" nil t)
		    (1+ (current-column)))))
      (if arg stcol
	(delete-horizontal-space)
	(indent-to stcol))))


;from b to e nicely. The lineup string is str."

(defun cocoa-get-lineup-indent (b e str)
  (save-excursion
    (let ((ind 0)
	  (reg (concat str "\\|\\(\\<NULLA_DA_FISSARE\\>\\)"))
	  nest)
      (goto-char b)
      ;; Get rightmost position
      (while (< (point) e)
	(setq nest 1)
	(if (re-search-forward reg (min e (cocoa-get-end-of-line 2)) 'move)
	    (progn
	      ;; Skip record blocks
	      (if (match-beginning 1)
		  (cocoa-declaration-end)
		(progn
		  (goto-char (match-beginning 0))
		  (skip-chars-backward " \t")
		  (if (> (current-column) ind)
		      (setq ind (current-column)))
		  (goto-char (match-end 0)))))))
      ;; In case no lineup was found
      (if (> ind 0)
	  (1+ ind)
	;; No lineup-string found
	(goto-char b)
	(end-of-line)
	(skip-chars-backward " \t")
	(1+ (current-column))))))

;;-------------------------------------------------------------->
;; cocoa-close-block  {like tex-close-latex-block} C-c C-e
;;-------------------------------------------------------------->

(defun cocoa-last-unended-begin ()
  "Leave point at the beginning of the last `\\begin{...}' that is unended."
;  (while (and (re-search-backward "\\(\\\\begin\\s *{\\)\\|\\(\\\\end\\s *{\\)")
;              (looking-at "\\\\end{"))
  (while (and (re-search-backward 
	       (concat 
;		"\\([^a-zA-Z0-9]\\(If\\|Define\\)[^a-zA-z0-9]\\|"
		"\\(\\<" cocoa-startkeywords-list "\\>\\|"
		cocoa-endkeywords-list "\\)")
	       )
              (looking-at cocoa-endkeywords-list))
    (cocoa-last-unended-begin)))

(defun cocoa-close-block ()
  "Creates an \\end{...} to match the last unclosed \\begin{...}."
  (interactive "*")
  (let ((new-line-needed (bolp))
	text indentation)
    (save-excursion
      (condition-case nil
          (cocoa-last-unended-begin)
        (error (error "Couldn't find unended command")))
      (setq indentation (current-column))
      (re-search-forward "[ \n]*\\(\\s *[^)\n ]*\\)")
      (setq text (buffer-substring (match-beginning 1) (match-end 1))))
    (indent-to indentation)
    (insert "End" text ";")
    (if new-line-needed (insert ?\n))))

;;-------------------------------------------------------------;;
;;                        RUNNING CoCoA                        ;;
;;-------------------------------------------------------------;;

(defun cocoa ()
  "CoCoA interface.
The function splits the current window in two parts, one containing
a CoCoA shell and the other containing a file chosen by the user
in cocoa mode.

The file in cocoa mode can be modified as any text file; in addition
to that, there is a special key binding (C-c C-l) that allows to send
the current line as input to CoCoA. This saves the effort of cutting
and pasting lines of input.

Type C-h C-f cocoa-mode to know about the other functions
available in the cocoa mode.

Customization:
- change the value of the global variable cocoa-dir to the 
  directory where your CoCoA input files are;
- change the value of cocoa-shell-launch-dir to the directory where the 
  CoCoA executable is.
"
;  (setq default-directory cocoa-shell-launch-dir)
  (interactive)
  (require 'comint)
  (require 'easymenu)  ;; MENU

;;;; uncomment the following lines if you have problems with trailing ^M
;;;; you can also add them to your .emacs file
;;(if (null system-uses-terminfo) nil
;;    (autoload 'shell-strip-ctrl-m "shell" nil t)
;;    (add-hook 'comint-output-filter-functions 'shell-strip-ctrl-m)
;;    (add-hook 'comint-input-filter-functions 'shell-ctrl-m)
;;)
  (cocoa-make-shell)
  (delete-other-windows)
  (save-excursion
    (switch-to-buffer (process-buffer (get-process "cocoa")))
;;;; opens a cocoa file from cocoa-dir
;;    (split-window-vertically)
;;    (other-window 1)
  )
)

;;----------------------------------------------------------------------
;; cocoa-mode-map
;;----------------------------------------------------------------------
(defvar cocoa-mode-map ()  "Keymap used in CoCoA mode.")
(if cocoa-mode-map
    ()
;; syntax
  (setq cocoa-mode-map (make-sparse-keymap))
  (define-key cocoa-mode-map ";"       'electric-cocoa-semi-or-dot)
  (define-key cocoa-mode-map "."       'electric-cocoa-semi-or-dot)
  (define-key cocoa-mode-map ":"       'electric-cocoa-colon)
  (define-key cocoa-mode-map "="       'electric-cocoa-equal)
  (define-key cocoa-mode-map "#"       'electric-cocoa-hash)
  (define-key cocoa-mode-map "\r"      'electric-cocoa-terminate-line)
  (define-key cocoa-mode-map "\t"      'electric-cocoa-tab)
  (define-key cocoa-mode-map "\177"    'backward-delete-char-untabify)
  (define-key cocoa-mode-map "\M-\C-a" 'cocoa-beg-of-defun)
;; to send commands to CoCoA
  (define-key cocoa-mode-map "\C-c\C-l" 'cocoa-send-line)
  (define-key cocoa-mode-map "\C-c\C-b" 'cocoa-send-buffer)
  (define-key cocoa-mode-map "\C-c\C-r" 'cocoa-send-region)
  (define-key cocoa-mode-map "\C-c\C-a" 'cocoa-send-file)
  (define-key cocoa-mode-map "\C-c?" 'cocoa-word-man)
  (define-key cocoa-mode-map "\C-c\C-m" 'cocoa-word-man)
  (define-key cocoa-mode-map "\C-cm" 'cocoa-word-man)
  (define-key cocoa-mode-map "\C-c\C-w" 'cocoa-open-wordlist-file)
  (define-key cocoa-mode-map "\C-c\C-n" 'cocoa-open-file)
  (define-key cocoa-mode-map "\C-c\C-q" 'cocoa-shell-quit)
  (define-key cocoa-mode-map "\C-c\C-s" 'cocoa-shell-restart-process)
  (define-key cocoa-mode-map "\C-c\C-e" 'cocoa-close-block)
  (define-key cocoa-mode-map "\C-c\C-c" 'comment-region)
  (define-key cocoa-mode-map "\C-c\C-u" 'uncomment-region)
)

;;-------------------------------------------------------------->
;; cocoa-menu
;;-------------------------------------------------------------->

(defvar cocoa-mode-menu nil)
(defvar cocoa-menu nil
  "Menu for `cocoa-mode'.")
(setq cocoa-menu
      '("CoCoA"
	["Send line to cocoa"             cocoa-send-line   t]
	["Send region to cocoa"           cocoa-send-region t]
	["Send buffer to cocoa"           cocoa-send-buffer t]
	"---"
	["Manual Word"                    cocoa-word-man]
	["Complete Word"                  dabbrev-expand]
	["Indent Region"                  indent-region]
	["Comment Region"                 comment-region]
	["Uncomment Region"               uncomment-region]
	["Print End***;"                  cocoa-close-block]
	"---"
	["Show cocoa-mode version"        cocoa-show-mode-version t]
	["Show cocoa version"             cocoa-show-version t]
	"---"
	["Kill cocoa"                     cocoa-shell-quit t]
	))


(defconst cocoa-xemacs-p (string-match "XEmacs" (emacs-version)))

(defun cocoa-setup-menubar ()
  "Initial setup of cocoa and insertions menus."
  (progn
    (easy-menu-define			; set up cocoa menu
      cocoa-mode-menu cocoa-mode-map "Menu used in cocoa-mode"
      cocoa-menu)
    (easy-menu-add cocoa-mode-menu cocoa-mode-map) )
)




;;-------------------------------------------------------------->
;; cocoa-open-wordlist-file
;;-------------------------------------------------------------->

(defun cocoa-open-wordlist-file ()
  "The function opens a wordlist file chosen by the user in cocoa-mode"
  (interactive)
  (setq cocoa-wordlist-file 
	(read-file-name "CoCoA word list file: " cocoa-wordlist-default-file))
  (if cocoa-wordlist-file  
      (setq cocoa-wordlist-file cocoa-wordlist-default-file))
  (find-file-noselect cocoa-wordlist-file)
)

;;-------------------------------------------------------------->
;; cocoa-open-file
;;-------------------------------------------------------------->

(setq cocoa-shell-buffer-list '())

(defun cocoa-open-file ()
  "The function opens a new file chosen by the user in cocoa-mode"
  (interactive)
  (setq cocoa-file 
	(read-file-name "Enter name of a CoCoA file: " cocoa-dir))
  (setq cocoa-file-buffer-name (find-file-noselect cocoa-file))
  (switch-to-buffer cocoa-file-buffer-name)
  (cocoa-mode)
  (setq cocoa-shell-buffer-list 
	(cons cocoa-file-buffer-name cocoa-shell-buffer-list))
)

;;-------------------------------------------------------------->
;; cocoa-make-shell
;;-------------------------------------------------------------->

(defun cocoa-make-shell ()
  (save-excursion
;;    (setq default-directory cocoa-shell-launch-dir)
    (set-buffer  (make-comint "cocoa" 
			      (concat cocoa-executable)))
;; Load wordlist
    (if cocoa-auto-load-wordlist
	(find-file-noselect cocoa-wordlist-default-file 'NOWARN))
  )
)

;;-------------------------------------------------------------->
;; cocoa-mode
;;-------------------------------------------------------------->

;;;###autoload
(defun cocoa-mode ()
  "Major mode for editing a notebook file that sends input to CoCoA. 
Like Text mode with \\[cocoa-send-line] sending the current line as input.

Special commands:
\\{cocoa-mode-map}
"
;;  (setq default-directory cocoa-shell-launch-dir)
  (interactive)
  (kill-all-local-variables)
  (use-local-map cocoa-mode-map)
  (setq major-mode 'cocoa-mode)
  (setq mode-name "CoCoA")
;;;; syntax
  (setq local-abbrev-table cocoa-mode-abbrev-table)
  (set-syntax-table cocoa-mode-syntax-table)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'cocoa-indent-line)
  (make-local-variable 'case-fold-search)
  (setq case-fold-search t)

  (make-local-variable 'comment-start)
  (setq comment-start "--")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip  "/\\*+ *\\|// *")
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'cocoa-indent-comment)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  
;; Font lock support
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(cocoa-font-lock-keywords nil nil))
;; Imenu support
  (make-local-variable 'imenu-generic-expression)
  (setq imenu-generic-expression cocoa-imenu-generic-expression)
;; Load wordlist
  (if cocoa-auto-load-wordlist
      (find-file-noselect cocoa-wordlist-default-file 'NOWARN))
;; Menu ???
  (cocoa-setup-menubar)

  (run-hooks 'cocoa-mode-hook)
)

;;-------------------------------------------------------------->
;; cocoa-send-line
;;-------------------------------------------------------------->

(defun cocoa-send-line ()
  "Send current line as input to CoCoA in the other window.
Note that the cursor can be at any position in the line;
after the input has been sent, the cursor goes to the next non-empty
line in the cocoa file.

Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (beginning-of-line)
  (setq cocoa-line-beg (point))
  (end-of-line)
  (setq cocoa-input-line (buffer-substring cocoa-line-beg (point)))
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
;    (pop-to-buffer "*cocoa*")
    (goto-char (point-max))
    (insert cocoa-input-line)
    (comint-send-input)
;    (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
    )
  (skip-chars-forward "-*\n-*")
)

;;-------------------------------------------------------------->
;; cocoa-word-man
;;-------------------------------------------------------------->

(defun cocoa-word-man1 ()
  "Send current word input to CoCoA 'Man' in the other window.
Note that the cursor can be at any position in the word;
after the input has been sent, the cursor goes to the next non-empty
line in the cocoa file.
Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (setq word-regexp "[a-zA-Z][a-zA-Z0-9_]*")
  (setq word-boundary "\\(^\\|[^a-zA-Z_]\\)")
  ;; Now mark the word and save to string.
  (forward-char)
  (re-search-backward (concat word-boundary word-regexp))
  (or (re-search-forward word-regexp (point-max) t)
      (error "No word found to check!"))
  (setq start (match-beginning 0)
	end (point)
	word (buffer-substring start end))
;  (setq cocoa-input-line (concat "Man('" word "');"))
  (setq cocoa-input-line (concat "? " word))
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
;    (pop-to-buffer "*cocoa*")
    (goto-char (point-max))
    (insert cocoa-input-line)
    (comint-send-input)
;    (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
    )
  (skip-chars-forward "-*\n-*")
)

;;-------------------------------------------------------------->
;; cocoa-show-version
;;-------------------------------------------------------------->


(defun cocoa-exec-this (arg)
  "Send arg to CoCoA
"
  (cocoa-shell-check-process)
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
;    (pop-to-buffer "*cocoa*")
    (goto-char (point-max))
    (insert arg)
    (comint-send-input)
;    (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
    )
  (skip-chars-forward "-*\n-*")
)

(defun cocoa-show-version ()
  (interactive)
  (cocoa-exec-this "Comp(CocoaInfos(), \"Version\");")
)

(defun cocoa-show-mode-version ()
  (interactive)
  (cocoa-exec-this (concat "-- " cocoa-mode-version))
)

(defun cocoa-word-man ()
  "Send current word input to CoCoA 'Man' in the other window.
Note that the cursor can be at any position in the word;
after the input has been sent, the cursor goes to the next non-empty
line in the cocoa file.
Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (setq word-regexp "[a-zA-Z][a-zA-Z0-9_]*")
  (setq word-boundary "\\(^\\|[^a-zA-Z_]\\)")
  ;; Now mark the word and save to string.
  (forward-char)
  (re-search-backward (concat word-boundary word-regexp))
  (or (re-search-forward word-regexp (point-max) t)
      (error "No word found to check!"))
  (setq start (match-beginning 0)
	end (point)
	word (buffer-substring start end))
  (cocoa-exec-this (concat "? " word))
)
;;-------------------------------------------------------------->
;; cocoa-shell-quit
;;-------------------------------------------------------------->

(defun cocoa-shell-quit ()
  "Quit CoCoA after saving all notebook files."
  (interactive)
  (message "saving all notebook files . . .")
  (save-excursion
    (while (not (null cocoa-shell-buffer-list))
      (setq cocoa-shell-buffer-tosave (car cocoa-shell-buffer-list))
      (if (buffer-name cocoa-shell-buffer-tosave) 
	  (progn 
	    (switch-to-buffer cocoa-shell-buffer-tosave)
	    (save-buffer)
	    (kill-buffer cocoa-shell-buffer-tosave)))
      (setq cocoa-shell-buffer-list (cdr cocoa-shell-buffer-list))
      ))
  (cocoa-shell-quit-cocoa-shell)
  (message "  . . . bye!  ")
)

;;-------------------------------------------------------------->
;; cocoa-shell-quit-cocoa-shell
;;-------------------------------------------------------------->

(defun cocoa-shell-quit-cocoa-shell ()
  (if (not (eq (get-buffer-process "*cocoa*") nil)) 
      (process-kill-without-query (get-process "cocoa")))
  (kill-buffer "*cocoa*")
)

;;-------------------------------------------------------------->
;; cocoa-shell-restart-process
;;-------------------------------------------------------------->

(defun cocoa-shell-restart-process ()
"Restart CoCoA process in the buffer named *cocoa*."

(interactive)
;;(process-kill-without-query (get-process "cocoa"))
(if (not (eq (get-buffer-process "*cocoa*") nil)) (delete-process "*cocoa*"))
(cocoa-make-shell)
)

;;-------------------------------------------------------------->
;; cocoa-send-buffer
;;-------------------------------------------------------------->

(defun cocoa-send-buffer ()
  "Send current buffer as input to CoCoA. See \\[cocoa-send-region] for 
more information. Does not save the buffer, so it's useful for trying 
experimental versions.
See \\[cocoa-send-file] for an alternative.

Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (setq kill-ring '())
  (copy-region-as-kill (point-min) (point-max))
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
    (if (not (eq (process-status "cocoa") "run")) (cocoa-make-shell))
    (goto-char (point-max))
    (yank)
    (comint-send-input)
;  (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
  )
)

;;-------------------------------------------------------------->
;; cocoa-send-region
;;-------------------------------------------------------------->

(defun cocoa-send-region ()
  "Send current region as input to CoCoA.

Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (setq kill-ring '())
  (copy-region-as-kill (point) (mark))
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
    (if (not (eq (process-status "cocoa") "run")) (cocoa-make-shell))
    (goto-char (point-max))
    (yank)
    (comint-send-input)
;    (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
  )
)

;;-------------------------------------------------------------->
;; cocoa-send-file
;;-------------------------------------------------------------->

(defun cocoa-send-file ()
  "Prompt to save current buffer and send current buffer's file as input to CoCoA.

Note: if the *cocoa* buffer has no process, CoCoA will be restarted
automatically when this function is called.
"
  (interactive)
  (cocoa-shell-check-process)
  (setq cocoa-shell-input-file buffer-file-name)
  (setq cocoa-shell-input-file-name (concat (concat "<< '" cocoa-shell-input-file) "';"))
  ;;    (if (cocoa-shell-running)
  ;;        (cocoa-shell-kill-job)
  ;;      (cocoa-make-shell)
  ;;     )
  (cocoa-shell-check-window)
  (save-excursion
    (switch-to-buffer-other-window "*cocoa*")
    (if (not (eq (process-status "cocoa") "run")) (cocoa-make-shell))
    (goto-char (point-max))
    (insert cocoa-shell-input-file-name)
    (comint-send-input)
;    (switch-to-buffer-other-window cocoa-file-buffer-name)
    (other-window 1)
  )
)

;;-------------------------------------------------------------->
;; cocoa-shell-check-process
;;-------------------------------------------------------------->

(defun cocoa-shell-check-process ()
  (if (eq (get-buffer-process "*cocoa*") nil) (cocoa-make-shell) )
)

(defun cocoa-shell-check-window ()
  (if (get-buffer-window "*cocoa*" 0)
      ()
      (progn
	(delete-other-windows)
	(split-window-vertically)
	(other-window 1)
      )
  )
)
