/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (03/2001)
 */

#include "ca_qt.h" /* only needed for the function ca_finished() */
#include "engine_thread.h"
#include "output_update_events.h"
#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>

#define UPDATE_TIMEOUT 200

typedef void (*signalHandler)(int);
void errorHandler(int);

static jmp_buf callingEnvironment;

EngineThread::EngineThread(QObject* _receiver, const char* _source)
	: receiver(_receiver)
{
	source = new char[strlen(_source) + 1];
	strcpy(source, _source);

	engineThread = this;
}

EngineThread* EngineThread::engineThread = 0;

EngineThread::~EngineThread()
{
	delete[] source;
}

EngineThread* EngineThread::currentEngineThread()
{
	return engineThread;
}

void EngineThread::addChar(char c)
{
	buffer += c;

	if (timeout.elapsed() > UPDATE_TIMEOUT)
		sendBufferedData();
}

void EngineThread::addString(const char *s)
{
	buffer += s;

	if (timeout.elapsed() > UPDATE_TIMEOUT)
		sendBufferedData();
}

void EngineThread::flush()
{
	sendBufferedData();
	QThread::postEvent(receiver, new FlushEvent());
}

void EngineThread::sendBufferedData()
{
	QThread::postEvent(receiver, new StringAddEvent(buffer));
	buffer = "";
	timeout.restart();
}

void EngineThread::run()
{
	timeout.start();

	signalHandler abrtHandler = signal(SIGABRT, errorHandler);
	signalHandler fpeHandler = signal(SIGFPE, errorHandler);
	signalHandler illHandler = signal(SIGILL, errorHandler);
	signalHandler segvHandler = signal(SIGSEGV, errorHandler);

	int value = setjmp(callingEnvironment);
	if (!value)
		ck_source_from_string(source);
	else
		QThread::postEvent(receiver, new EngineCrashEvent());

	flush();
	QThread::postEvent(receiver, new ExecutionStateEvent(false));
	if (ca_finished())
		QThread::postEvent(receiver, new QuitEvent());
	engineThread = 0;
	
	signal(SIGABRT, abrtHandler);
	signal(SIGFPE, fpeHandler);
	signal(SIGILL, illHandler);
	signal(SIGSEGV, segvHandler);
}

void errorHandler(int signal)
{
	longjmp(callingEnvironment, 1);
}
