#include "pixmaps.h"
#include <iostream.h>
#include "helpwindow.h"
#include <qpixmap.h>
#include <qpopupmenu.h>

#if defined(MAKE_KDE)
#include <kmenubar.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#else
#include <qmenubar.h>
#include <qstatusbar.h>
#include <qtoolbar.h>
#endif

#include <qtoolbutton.h>
#include <qiconset.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qstylesheet.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qapplication.h>
#include <qcombobox.h>
#include <qevent.h>
#include <qlineedit.h>
#include <qobjectlist.h>
#include <qfileinfo.h>
#include <qfile.h>
#include <qdatastream.h>
#include <qpainter.h>
#include <qprinter.h>
#include <qsimplerichtext.h>
#include <qpaintdevicemetrics.h>

#include <ctype.h>

HelpWindow::HelpWindow(const QString& _home, const QString& _path,
			QWidget* parent, const char *name, const int _maximumSize)
#if defined(MAKE_KDE)
	/*
	 * We have to pass the additional WFlag WType_TopLevel to KMainWindow, because
	 * KMainWindow passes WDestructiveClose to QMainWindow by default which is unwanted
	 * in this case.
	 */
	: KMainWindow(parent, name, WType_TopLevel), home(_home), pathCombo(0), maximumSize(_maximumSize)
#else
	: QMainWindow(parent, name), home(_home), pathCombo(0), maximumSize(_maximumSize)
#endif
{
	history = new PropertySection("History dummy");
	bookmarks = new PropertySection("Bookmark dummy");

	browser = new QTextBrowser(this);
	browser->mimeSourceFactory()->setFilePath(_path);
	browser->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
	connect(browser, SIGNAL(textChanged()), this, SLOT(textChanged()));

	setCentralWidget(browser);

	if (!home.isEmpty())
		browser->setSource(home);

	connect(browser, SIGNAL(highlighted(const QString&)), statusBar(), SLOT(message(const QString&)));

	QWidget *d = QApplication::desktop();
	int screen_width = d->width();
	int screen_height = d->height();
	int xpos = (int)(0.5 * 0.2 * screen_width);
	int ypos = (int)(0.5 * 0.2 * screen_height);
	move(xpos, ypos);
	resize((int)(0.8 * screen_width), (int)(0.8 * screen_height));

	QIconSet newIconSet = createIconSet(window_new_xpm);
	QIconSet openIconSet = createIconSet(fileopen_xpm);
	QIconSet printIconSet = createIconSet(fileprint_xpm);
	QIconSet closeIconSet = createIconSet(fileclose_xpm);

	QIconSet backIconSet = createIconSet(back_xpm);
	QIconSet forwardIconSet = createIconSet(forward_xpm);
	QIconSet homeIconSet = createIconSet(home_xpm);
	QIconSet copyIconSet = createIconSet(editcopy_xpm);

	QPopupMenu* file = new QPopupMenu(this);
	file->insertItem(newIconSet, "&New Window", this, SLOT(newWindow()), ALT | Key_N);
	file->insertItem(openIconSet, "&Open File", this, SLOT(openFile()), ALT | Key_O);
	file->insertItem(printIconSet, "&Print", this, SLOT(print()), ALT | Key_P);
	file->insertSeparator();
	file->insertItem(closeIconSet, "&Close", this, SLOT(close()), ALT | Key_Q);

	QPopupMenu* go = new QPopupMenu(this);
	backwardId = go->insertItem(backIconSet, "&Backward", browser, SLOT(backward()), ALT | Key_Left);
	forwardId = go->insertItem(forwardIconSet, "&Forward", browser, SLOT(forward()), ALT | Key_Right);
	homeId = go->insertItem(homeIconSet, "&Home", browser, SLOT(home()));

	hist = new QPopupMenu(this);
	connect(hist, SIGNAL(activated(int)), this, SLOT(historyEntryChosen(int)));

	bookm = new QPopupMenu( this );
	bookm->insertItem("Add Bookmark", this, SLOT(addBookmark()));
	bookm->insertSeparator();
	connect(bookm, SIGNAL(activated(int)), this, SLOT(bookmarkEntryChosen(int)));

	menuBar()->insertItem("&File", file);
	menuBar()->insertItem("&Go", go);
	menuBar()->insertItem("History", hist);
	menuBar()->insertItem("Bookmarks", bookm);

	menuBar()->setItemEnabled(forwardId, false);
	menuBar()->setItemEnabled(backwardId, false);
	connect(browser, SIGNAL(backwardAvailable(bool)), this, SLOT(setBackwardAvailable(bool)));
	connect(browser, SIGNAL(forwardAvailable(bool)), this, SLOT(setForwardAvailable(bool)));

#if defined(MAKE_KDE)
	KToolBar* toolbar = new KToolBar(this);
#else
	QToolBar* toolbar = new QToolBar(this);
#endif

	addToolBar(toolbar, "Toolbar");
	QToolButton* button;

	button = new QToolButton(backIconSet, "Backward", "", browser, SLOT(backward()), toolbar);
	connect(browser, SIGNAL(backwardAvailable(bool)), button, SLOT(setEnabled(bool)));
	button->setEnabled(false);
	button = new QToolButton(forwardIconSet, "Forward", "", browser, SLOT(forward()), toolbar);
	connect(browser, SIGNAL(forwardAvailable(bool)), button, SLOT(setEnabled(bool)));
	button->setEnabled(false);
	button = new QToolButton(homeIconSet, "Home", "", browser, SLOT(home()), toolbar);
	button = new QToolButton(copyIconSet, "Copy", "", browser, SLOT(copy()), toolbar);

	toolbar->addSeparator();

	pathCombo = new QComboBox(true, toolbar);
	connect(pathCombo, SIGNAL(activated(const QString &)), this, SLOT(pathSelected(const QString &)));
	toolbar->setStretchableWidget(pathCombo);
	setRightJustification(true);
	setDockEnabled(Left, false);
	setDockEnabled(Right, false);

	pathCombo->insertItem(home);

	browser->setFocus();
}

void HelpWindow::showAndRaise()
{
	show();
#if defined(SIS_WIN32)
	showNormal();
#endif
	raise();
}

void HelpWindow::setBackwardAvailable( bool b)
{
	menuBar()->setItemEnabled( backwardId, b);
}

void HelpWindow::setForwardAvailable( bool b)
{
	menuBar()->setItemEnabled( forwardId, b);
}


void HelpWindow::textChanged()
{
	if ( browser->documentTitle().isNull() )
		setCaption( browser->context() );
	else
		setCaption( browser->documentTitle() ) ;

	QString selectedURL = browser->context();
	if (!selectedURL.isEmpty() && pathCombo)
	{
		bool exists = false;
		int i;
		for (i = 0; i < pathCombo->count(); i++)
		{
			if (pathCombo->text(i) == selectedURL)
			{
				exists = true;
				break;
			}
		}

		if (!exists) {
			pathCombo->insertItem( selectedURL, 0 );
			pathCombo->setCurrentItem( 0 );

			/*
			 * Ensure that the size of the history allows the insertion
			 * of a new entry.
			 */
			truncateHistory();

			/* Insert the new history entry and history menu entry */
			history->append(selectedURL, QString::null);
			mHistory[hist->insertItem(selectedURL)] = selectedURL;
		} else
			pathCombo->setCurrentItem(i);
	}
}

void HelpWindow::truncateHistory()
{
	/*
	 * Delete items from the history menu and the history so that the size of the
	 * history doesn't exceed the limit.
	 */
	PropertySectionEntry *entry;
	for (entry = history->first(); history->count() >= maximumSize && entry != 0; entry = history->first())
	{
		QMap<int, QString>::Iterator it2;
		for (it2 = mHistory.begin(); it2 != mHistory.end(); ++it2)
		{
			if (entry->key() == *it2)
			{
				hist->removeItem(it2.key());
				mHistory.remove(it2);
				break;
			}
		}
		history->remove(entry->key());
	}
}

void HelpWindow::openFile()
{
#ifndef QT_NO_FILEDIALOG
	QString fn = QFileDialog::getOpenFileName(QString::null, QString::null, this);
	if (!fn.isEmpty())
		browser->setSource(fn);
#endif
}

void HelpWindow::newWindow()
{
	HelpWindow *helpWindow = new HelpWindow(home, "qbrowser", NULL);
	helpWindow->show();
	helpWindow->pathSelected(browser->source());
}

void HelpWindow::print()
{
#ifndef QT_NO_PRINTER
	QPrinter printer;
	printer.setFullPage(TRUE);
	if (printer.setup())
	{
		QPainter p(&printer);
		QPaintDeviceMetrics metrics(p.device());
		int dpix = metrics.logicalDpiX();
		int dpiy = metrics.logicalDpiY();
		const int margin = 72; // pt
		QRect body(margin * dpix / 72, margin * dpiy / 72,
			metrics.width() - margin * dpix / 72 * 2,
			metrics.height() - margin * dpiy / 72 * 2);
		QFont font("times", 10);
		QSimpleRichText richText(browser->text(), font, browser->context(), browser->styleSheet(),
					browser->mimeSourceFactory(), body.height());
		richText.setWidth(&p, body.width());
		QRect view(body);
		int page = 1;
		do {
			richText.draw(&p, body.left(), body.top(), view, colorGroup());
			view.moveBy(0, body.height());
			p.translate(0 , -body.height());
			p.setFont(font);
			p.drawText(view.right() - p.fontMetrics().width(QString::number(page)),
				view.bottom() + p.fontMetrics().ascent() + 5, QString::number(page));
			if (view.top() >= richText.height())
				break;
			printer.newPage();
			page++;
		} while (true);
	}
#endif
}

void HelpWindow::pathSelected( const QString &_path )
{
	browser->setSource( _path );

	QMap<int, QString>::Iterator it;
	bool found = false;
	for (it = mHistory.begin(); it != mHistory.end(); ++it)
	{
		if (it.data() == _path)
		{
			found = true;
			break;
		}
	}

	if (!found)
	{
		/*
		 * Ensure that the size of the history allows the insertion
		 * of a new entry.
		 */
		truncateHistory();

		/* Insert the new entry into the history and the history menu */
		history->append(_path, QString::null);
		mHistory[hist->insertItem(_path)] = _path;
	}
}

PropertySection* HelpWindow::setBookmarks(PropertySection* _bookmarks)
{
	PropertySection *oldBookmarks = bookmarks;
	bookmarks = _bookmarks;

	/* Remove the old bookmark menu entries */
	QMap<int, QString>::Iterator it1;
	for (it1 = mBookmarks.begin(); it1 != mBookmarks.end(); ++it1)
		bookm->removeItem(it1.key());

	mBookmarks.clear();

	/* Insert the new bookmark menu entries */
	PropertySectionEntry *entry;
	for (entry = bookmarks->first(); entry != 0; entry = bookmarks->next())
		mBookmarks[bookm->insertItem(entry->key())] = entry->key();

	return oldBookmarks;
}

PropertySection* HelpWindow::setHistory(PropertySection* _history)
{
	PropertySection *oldHistory = history;
	history = _history;

	/* Remove the old history menu entries */
	QMap<int, QString>::Iterator it1;
	for (it1 = mHistory.begin(); it1 != mHistory.end(); ++it1)
		hist->removeItem(it1.key());

	/*
	 * Truncate the history to the maximum size.
	 */
	PropertySectionEntry *entry;
	for (entry = history->first(); entry != 0 && history->count() > maximumSize; entry = history->first())
		history->remove(entry->key());

	/* Insert the new history menu entries */
	for (entry = history->first(); entry != 0; entry = history->next())
		mHistory[hist->insertItem(entry->key())] = entry->key();

	return oldHistory;
}

void HelpWindow::historyEntryChosen(int i)
{
	if (mHistory.contains(i))
		browser->setSource(mHistory[i]);
}

void HelpWindow::bookmarkEntryChosen(int i)
{
	if (mBookmarks.contains(i))
		browser->setSource(mBookmarks[i]);
}

void HelpWindow::addBookmark()
{
	QString url = browser->context();

	QMap<int, QString>::Iterator it;
	bool found = false;
	for (it = mBookmarks.begin(); it != mBookmarks.end(); ++it)
	{
		if (it.data() == url)
		{
			found = true;
			break;
		}
	}

	if (!found)
	{
		/* Insert the new entry into the bookmarks and the bookmark menu */
		bookmarks->append(url, QString::null);
		mBookmarks[bookm->insertItem(url)] = url;
	}
}

#include "helpwindow.moc"
