/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (02/2001)
 */

#include "preferences_dialog.h"
#include <qfiledialog.h>
#include <qlayout.h>

PreferencesDialog::PreferencesDialog(QWidget *parent, const char *name)
	: QTabDialog(parent, name, true, WDestructiveClose)
{
	setCaption("Preferences");

	QWidget *wordFileTab = new QWidget(this);
	QVBoxLayout *vertical = new QVBoxLayout(wordFileTab, 5);

	QHBoxLayout *horizontal = new QHBoxLayout();
	wordFileListView = new QListView(wordFileTab);
	wordFileListView->setMultiSelection(true);
	wordFileListView->setAllColumnsShowFocus(true);
	wordFileListView->setMinimumWidth(320);

	wordFileListView->addColumn("File");
	wordFileListView->setColumnWidthMode(0, QListView::Manual);

	wordFileListView->addColumn("Load");
	wordFileListView->setColumnWidthMode(1, QListView::Maximum);

	wordFileListView->addColumn("Load state");
	wordFileListView->setColumnWidthMode(2, QListView::Maximum);

	connect(wordFileListView, SIGNAL(selectionChanged()), this, SLOT(checkSelection()));

	horizontal->addWidget(wordFileListView, 3);

	QVBoxLayout *buttonContainer = new QVBoxLayout();
	QPushButton *addButton = new QPushButton("Add", wordFileTab);
	connect(addButton, SIGNAL(clicked()), this, SLOT(addCompletionFiles()));
	buttonContainer->addWidget(addButton);

	loadButton = new QPushButton("Load", wordFileTab);
	connect(loadButton, SIGNAL(clicked()), this, SLOT(loadCompletionFiles()));
	loadButton->setEnabled(false);
	buttonContainer->addWidget(loadButton);

	unloadButton = new QPushButton("Unload", wordFileTab);
	connect(unloadButton, SIGNAL(clicked()), this, SLOT(unloadCompletionFiles()));
	unloadButton->setEnabled(false);
	buttonContainer->addWidget(unloadButton);

	removeButton = new QPushButton("Remove", wordFileTab);
	connect(removeButton, SIGNAL(clicked()), this, SLOT(removeCompletionFiles()));
	removeButton->setEnabled(false);
	buttonContainer->addWidget(removeButton);

	QPushButton *OKButton = new QPushButton("Ok", wordFileTab);
	connect(OKButton, SIGNAL(clicked()), this, SLOT(close()));
	buttonContainer->addWidget(OKButton);

	horizontal->addLayout(buttonContainer, 1);

	vertical->addLayout(horizontal);

	vertical->activate();
	addTab(wordFileTab, "&Word completion files");

	setOkButton(NULL);
}

void PreferencesDialog::addCompletionFiles()
{
	QStringList paths = QFileDialog::getOpenFileNames(QString::null, QString::null, this);
	emit addCompletionFiles(paths);
}

void PreferencesDialog::loadCompletionFiles()
{
	QStringList paths = getSelectedItems();
	emit loadCompletionFiles(paths);
}

void PreferencesDialog::removeCompletionFiles()
{
	QStringList paths = getSelectedItems();
	emit removeCompletionFiles(paths);
}

void PreferencesDialog::unloadCompletionFiles()
{
	QStringList paths = getSelectedItems();
	emit unloadCompletionFiles(paths);
}

void PreferencesDialog::updateWordFileList(WordFileStateList wordFileStateList)
{
	wordFileListView->clear();

	QStringList keys;
	QMap<QString, WordFileState>::Iterator it;
	for (it = wordFileStateList.begin(); it != wordFileStateList.end(); ++it)
		keys.append(it.key());
	keys.sort();

	QStringList::Iterator kit;
	int count = 0;
	for (kit = keys.begin(); kit != keys.end(); ++kit)
	{
		QString fileName = *kit;
		WordFileState state = wordFileStateList[fileName];
		wordFileListView->insertItem(
				new QListViewItem(
					wordFileListView, fileName,
					loadFlagNames[state.loadFlag],
					loadedFlagNames[state.loadedFlag]));
	}
	updateColumnWidths();
}

void PreferencesDialog::paintEvent(QPaintEvent *e)
{
	updateColumnWidths();
}

void PreferencesDialog::updateColumnWidths()
{
	int width = wordFileListView->visibleWidth();
	wordFileListView->setColumnWidth(0, width
		- wordFileListView->columnWidth(1)
		- wordFileListView->columnWidth(2));
}

QStringList PreferencesDialog::getSelectedItems()
{
	QStringList list;
	QListViewItemIterator it(wordFileListView);
	for (; it.current(); ++it)
	{
		if (it.current()->isSelected())
			list.append(it.current()->text(0));
	}
	return list;
}

void PreferencesDialog::checkSelection()
{
	QStringList list = getSelectedItems();
	if (list.count() > 0)
	{
		loadButton->setEnabled(true);
		removeButton->setEnabled(true);
		unloadButton->setEnabled(true);
	} else
	{
		loadButton->setEnabled(false);
		removeButton->setEnabled(false);
		unloadButton->setEnabled(false);
	}
}

#include "preferences_dialog.moc"
