/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (02/2001)
 */

#include <qfile.h>
#include <qregexp.h>
#include <qtextstream.h>
#include "word_completions.h"

WordFileLoadedFlag WordCompletions::loadWordCompletions(QString filename)
{
	QFile file(filename);

	if (!file.exists())
		return NOT_FOUND;

	if (!file.open(IO_ReadOnly))
		return NOT_READABLE;

	QTextStream s(&file);
	while (!s.atEnd())
	{
		QString word = s.readLine();
		if (!word)
			return CORRUPTED;

		/*
		 * We only need a single copy of each word
		 */
		if (!word.isEmpty() && !wordList.contains(word))
			wordList << word;
        }
	file.close();

	wordList.sort();
	return LOADED;
}

void WordCompletions::clear()
{
	wordList.clear();
}

QStringList WordCompletions::completionsFor(QString head)
{
	/*
	 * Ifhead contains characters, that have special meaning in regular expressions, mask them.
	 */
	QString specials = "\\.^$[]*+?";
	for (unsigned int i = 0; i < specials.length(); i++)
		head.replace(QRegExp("\\" + specials.at(i)), "\\" + specials.at(i));

	/* Search for all words in wordList starting with head */
	QRegExp regExp(QString("^") + head);
	QStringList results = wordList.grep(regExp);

	/*
	 * We don't need items which are equal to head, cause then there is
	 * nothing to complete.
	 */
	results.remove(head);
	return results;
}
