<?xml version="1.0" encoding="ascii"?>

<xsl:stylesheet version="1.1"
		xmlns:xsl="http://www.w3.org/1999/XSL/Transform">

  <xsl:preserve-space elements="coc-quotes"/>
  <xsl:preserve-space elements="code"/>

<!--xslt document for creating the GUI-man directory for CoCoA-->

  <xsl:output method="text"/>
  <xsl:strip-space elements="*"/>	

  <xsl:template match="help">

<!-- ____ COMMANDS ______________________________________________ -->

  <xsl:for-each select="cocoa_commands/command">
    <xsl:sort select="title"/>
    <xsl:variable name="pos_num" select="position()" data-type="number"/>
    <xsl:variable name="file_cmd" select="concat('cmd',position(),'.html')"/>
    <xsl:document href="{$file_cmd}">

&lt;HTML&gt;
&lt;HEAD&gt;
&lt;link REL="stylesheet" TYPE="text/css" href="gui.css"&gt;
&lt;TITLE&gt;
  <xsl:apply-templates select="title"/>
&lt;/TITLE&gt;
&lt;/HEAD&gt;

&lt;BODY&gt;

&lt;DIV&gt;

  &lt;A HREF=
<!--   &quot;commands.html&quot; -->
  &quot;big_index.html#list&quot;
  &gt;up&lt;/A&gt;  

<xsl:choose>
<xsl:when test="position()!=1">
  &lt;A HREF=
  &quot;cmd<xsl:value-of select="$pos_num - 1"/>.html&quot;
  &gt;previous&lt;/A&gt;  
</xsl:when>
<xsl:otherwise>
&lt;span class="white"&gt;previous&lt;/span&gt;
</xsl:otherwise>
</xsl:choose>

<xsl:choose>
<xsl:when test="position()!=last()">
  &lt;A HREF=
  &quot;cmd<xsl:value-of select="$pos_num + 1"/>.html&quot;
  &gt;next&lt;/A&gt;  
</xsl:when>
<xsl:otherwise>
&lt;span class="white"&gt;next&lt;/span&gt;
</xsl:otherwise>
</xsl:choose>

&lt;H1&gt;
  <xsl:apply-templates select="title"/>
&lt;/H1&gt;

&lt;H2&gt;Syntax&lt;/H2&gt;
&lt;PRE&gt;<xsl:apply-templates select="syntax"/>&lt;/PRE&gt;

&lt;H2&gt;Summary&lt;/H2&gt;
<xsl:apply-templates select="short_description"/>

&lt;H2&gt;Description&lt;/H2&gt;
<xsl:apply-templates select="description"/>

<xsl:if test="see">
&lt;H2&gt;See Also&lt;/H2&gt;
  &lt;UL&gt;
  <xsl:for-each select="see">
    &lt;LI&gt;
<!--     &lt;A HREF=&quot;commands.html#<xsl:apply-templates/>&quot;&gt;<xsl:apply-templates/>&lt;/A&gt; -->
    &lt;A HREF=&quot;toc.html#<xsl:apply-templates/>&quot;&gt;<xsl:apply-templates/>&lt;/A&gt;
  </xsl:for-each>
  &lt;/UL&gt;
</xsl:if>

&lt;/DIV&gt;

&lt;/BODY&gt;
&lt;/HTML&gt;

    </xsl:document>
  </xsl:for-each>

<!-- ____ CHAPTERS ______________________________________________ -->

  <xsl:for-each select="manual_parts/part">
    <xsl:sort select="@number" data-type="number"/> 
    <xsl:variable name="part_num" select="@number"/>
    <xsl:for-each select="chapter"> 
      <xsl:sort select="@number" data-type="number"/>
      <xsl:variable name="chap_num" select="@number"/>
      <xsl:for-each select="section">
	<xsl:sort select="@number" data-type="number"/>
	<xsl:variable name="sect_num" select="@number"/>

	<xsl:variable name="file" select="concat('p',$part_num,'c',$chap_num,'s',$sect_num,'.html')"/> 

	<xsl:document href="{$file}">

&lt;HTML&gt;
&lt;HEAD&gt;
&lt;link REL="stylesheet" TYPE="text/css" href="gui.css"&gt;
&lt;TITLE&gt;
  <xsl:apply-templates select="title"/>
&lt;/TITLE&gt;
&lt;/HEAD&gt;

&lt;BODY&gt;

&lt;DIV&gt;

  &lt;A HREF=
<!--   &quot;<xsl:value-of select="concat('toc.html#p',$part_num,'c',$chap_num)"/>&quot; -->
  &quot;<xsl:value-of select="concat('toc.html#p',$part_num,'c',$chap_num)"/>&quot;
  &gt;up&lt;/A&gt;  


<xsl:if test="position()!=1">
  &lt;A HREF=
  &quot;<xsl:value-of select="concat('p',$part_num,'c',$chap_num,'s',$sect_num - 1,'.html')"/>&quot;
  &gt;previous&lt;/A&gt;  
</xsl:if>
<xsl:if test="position()=1">&lt;span class="white"&gt;previous&lt;/span&gt;</xsl:if>

<xsl:if test="position()!=last()">
  &lt;A HREF=
  &quot;<xsl:value-of select="concat('p',$part_num,'c',$chap_num,'s',$sect_num + 1,'.html')"/>&quot;
  &gt;next&lt;/A&gt;  
</xsl:if>
<xsl:if test="position()=last()">&lt;span class="white"&gt;next&lt;/span&gt;</xsl:if>


&lt;H1&gt;
  <xsl:value-of select="concat($part_num,'.',$chap_num,'.',$sect_num,' ')"/>
  <xsl:apply-templates select="title"/>
&lt;/H1&gt;

<!-- &lt;H2&gt;Description&lt;/H2&gt; -->
<xsl:apply-templates select="description"/>

&lt;/DIV&gt;

&lt;/BODY&gt;
&lt;/HTML&gt;

    </xsl:document>
      </xsl:for-each>
    </xsl:for-each>
  </xsl:for-each>
  </xsl:template>


<!-- ____  ______________________________________________ -->

  <xsl:template match="title"><xsl:apply-templates/></xsl:template>
  <xsl:template match="description"><xsl:apply-templates/></xsl:template>

  <xsl:template match="commands_and_functions_for">  
   <xsl:variable name="my_type" select="@type"/>
&lt;CENTER&gt;
&lt;TABLE BORDER=1&gt;
   <xsl:for-each select="/help/cocoa_commands/command[type=$my_type]">
     <xsl:sort select="title"/>
&lt;TR&gt;
&lt;TD&gt;
  &lt;A HREF=&quot;toc.html#<xsl:apply-templates select="title"/>&quot;
     &gt;<xsl:apply-templates select="title"/>&lt;/A&gt;
&lt;/TD&gt;
&lt;TD&gt;
 <xsl:apply-templates select="short_description"/>
&lt;/TD&gt;
&lt;/TR&gt;
   </xsl:for-each>
&lt;/TABLE&gt;
&lt;/CENTER&gt;
&lt;BR&gt;&lt;BR&gt;
  </xsl:template>

  <xsl:template match="coc_version">
    <xsl:value-of select="/help/version/cocoa_version"/>
  </xsl:template>


  <xsl:template match="em">&lt;B&gt;<xsl:value-of select="."/>&lt;/B&gt;</xsl:template>

  <xsl:template match="coc-quotes">&quot;<xsl:apply-templates/>&quot;</xsl:template>


  <xsl:template match="code">&lt;code&gt;<xsl:apply-templates/>&lt;/code&gt;</xsl:template>

  <xsl:template match="lt">&amp;lt;</xsl:template>
  <xsl:template match="less_eq">&amp;lt;=</xsl:template>
  <xsl:template match="tilde">~</xsl:template>
  <xsl:template match="dollar">$</xsl:template>
  <xsl:template match="ocb">{</xsl:template>
  <xsl:template match="ccb">}</xsl:template>
   <xsl:template match="times"> x </xsl:template>
  <xsl:template match="_">_</xsl:template>

<!--   <xsl:template match="ref">&lt;A HREF=&quot;toc.html#<xsl:apply-templates/>&quot;&gt;&quot;<xsl:apply-templates/>&quot;&lt;/A&gt;</xsl:template> -->

  <xsl:template match="ref">&lt;A HREF=&quot;toc.html#<xsl:apply-templates/>&quot;&gt;&quot;<xsl:apply-templates/>&quot;&lt;/A&gt;</xsl:template>

  <xsl:template match="coderef">&lt;A HREF=&quot;toc.html#<xsl:apply-templates/>&quot;&gt;&lt;code&gt;<xsl:apply-templates/>&lt;/code&gt;&lt;/A&gt;</xsl:template>

  <xsl:template match="par">&lt;BR&gt;&lt;BR&gt;</xsl:template>

  <xsl:template match="example">
    &lt;h2&gt;example&lt;/h2&gt;
    &lt;pre&gt;
    <xsl:apply-templates/>
    &lt;/pre&gt;
  </xsl:template>

  <xsl:template match="see">
    &lt;h2&gt;See also&lt;/h2&gt;
    &lt;pre&gt;
    <xsl:apply-templates/>
    &lt;/pre&gt;
  </xsl:template>

  <xsl:template match="coc_date">
    <xsl:value-of select="/help/version/date"/>
  </xsl:template>

  <xsl:template match="verbatim">&lt;PRE&gt;<xsl:apply-templates/>&lt;/PRE&gt;</xsl:template>


</xsl:stylesheet> 

