; needs nsis 1.4 beta or better

Name "CoCoA 4.2"
Caption "CoCoA 4.2 Installer"
Icon ..\UI\res\wincocoa.ico

; only as a fallback if no Windows version (WIN9X/WINNT) is defined (in the makefile)
!ifndef WIN9X & WINNT
  OutFile CoCoA-4.2-Installer.exe
!endif
!ifdef WIN9X
  OutFile CoCoA-4.2-Installer-Win9x.exe
!endif
!ifdef WINNT
  OutFile CoCoA-4.2-Installer-Win2k.exe
!endif

CRCCheck on
InstallColors 000000 FFFFFF

UninstallText "This will uninstall CoCoA 4.2 from your system."
UninstallExeName "uninst.exe"

ComponentText "This will install CoCoA 4.2 on your system."

EnabledBitmap "enabled.bmp"
DisabledBitmap "disabled.bmp"

DirText "Please select a location to install CoCoA (or use the default)."
InstallDir $PROGRAMFILES\CoCoA-4.2

; we want to see what has been done before quitting the installer
AutoCloseWindow false

; seems to install much faster if log is hidden
ShowInstDetails hide

; this makes building a bit slower but increases compression rate
SetDatablockOptimize on

; Typical is default
InstType "Typical"
InstType "Full"
InstType "WinCoCoA only"
InstType "Console only"
; note: there is InstType Custom automatically

; try to get $INSTDIR from HKCU first, then from HKLM, don't change it as fallback
Function .onInit
  ClearErrors
; ReadHKCU:
    ReadRegStr $0 HKCU "Software\CoCoA\4.2\Environment" "COCOA_LIBRARY"
    IfErrors ReadHKLM
    StrCmp $0 "" ReadHKLM DoReplace
  ReadHKLM:
    ReadRegStr $0 HKLM "Software\CoCoA\4.2\Environment" "COCOA_LIBRARY"
    IfErrors DontReplace
    StrCmp $0 "" DontReplace DoReplace
  DoReplace:
    StrCpy $INSTDIR $0
  DontReplace:
    ClearErrors
FunctionEnd

Section "Library (required)"
  SectionIn 12345 ; Library will always be selected, even in "Custom"
  SetOutPath "$INSTDIR"
  File "..\UI\Readme.txt"
  File "..\UI\userinit.coc"
  File /r "..\lib"
  WriteRegStr HKLM "Software\CoCoA" "CurrentVersion" "4.2"
  WriteRegStr HKLM "Software\CoCoA\4.2\Environment" "COCOA_LIBRARY" "$INSTDIR"
  WriteRegStr HKCU "Software\CoCoA" "CurrentVersion" "4.2"
  WriteRegStr HKCU "Software\CoCoA\4.2\Environment" "COCOA_LIBRARY" "$INSTDIR"
  ClearErrors
SectionEnd

Section "Graphical Interface"
  SectionIn 123
  SetOutPath "$INSTDIR"
  File "..\UI\wincocoa.exe"
  File "..\UI\wincocoa.bat"
  File "..\UI\wordlist.txt"
  ClearErrors
; I'm not happy with this, but we can't ask the user where to put the wincocoa.ini :-(
; maybe this will be possible in a future version of nsis?
; COCOA_GUI_PROPERTIES only are overwritten with defaults if they are empty or don't exist  in HKCU AND HKLM
; ReadHKCU:
    ReadRegStr $0 HKCU "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES"
    IfErrors ReadHKLM
    StrCmp $0 "" ReadHKLM DontWrite
  ReadHKLM:
    ReadRegStr $0 HKLM "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES"
    IfErrors DoWrite
    StrCmp $0 "" DoWrite DontWrite
  DoWrite:
    WriteRegStr HKCU "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES" "$INSTDIR\wincocoa.ini"
    WriteRegStr HKLM "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES" "$INSTDIR\wincocoa.ini"
  DontWrite:
    ClearErrors
SectionEnd

Section "HTML Documentation"
  SectionIn 123
  SetOutPath "$INSTDIR"
  File /r "..\UI\doc"
  ClearErrors
SectionEnd

Section "CoCoA Console"
  SectionIn 124
  SetOutPath "$INSTDIR"
  File "..\UI\cocoa.exe"
  File "..\UI\cocoa.bat"
; buggy win95 needs an extra icon for the console version
  IfFileExists "$INSTDIR\cocoa.ico" DontCopy
    File /oname=cocoa.ico "..\UI\res\wincocoa.ico"
  DontCopy:
    ClearErrors
SectionEnd

; separate the "file install" sections and the "create link" sections
SectionDivider

Section "Create CoCoA Start Menu Group"
  SectionIn 1234
  CreateDirectory "$SMPROGRAMS\CoCoA 4.2"
  CreateShortCut "$SMPROGRAMS\CoCoA 4.2\Readme.lnk" "$INSTDIR\Readme.txt"
  CreateShortCut "$SMPROGRAMS\CoCoA 4.2\Uninstall.lnk" "$INSTDIR\uninst.exe" "" "$INSTDIR\uninst.exe" 0
; CheckWinCoCoA:
    IfFileExists "$INSTDIR\wincocoa.exe" CreateWinCoCoAShortCut CheckCoCoA
  CreateWinCoCoAShortCut:
    CreateShortCut "$SMPROGRAMS\CoCoA 4.2\WinCoCoA.lnk" "$INSTDIR\wincocoa.exe" "" "$INSTDIR\wincocoa.exe" 0
  CheckCoCoA:
    IfFileExists "$INSTDIR\cocoa.exe" CreateCoCoAShortCut EndOfSection
  CreateCoCoAShortCut:
    CreateShortCut "$SMPROGRAMS\CoCoA 4.2\CoCoA Console.lnk" "$INSTDIR\cocoa.exe" "" "$INSTDIR\cocoa.ico" 0
  EndOfSection:
    ClearErrors
SectionEnd

Section "Create Desktop Shortcuts"
  SectionIn 2
; CheckWinCoCoA:
    IfFileExists "$INSTDIR\wincocoa.exe" CreateWinCoCoADesktopShortCut CheckCoCoA
  CreateWinCoCoADesktopShortCut:
    CreateShortCut "$DESKTOP\WinCoCoA.lnk" "$INSTDIR\wincocoa.exe" "" "$INSTDIR\wincocoa.exe" 0
  CheckCoCoA:
    IfFileExists "$INSTDIR\cocoa.exe" CreateCoCoADesktopShortCut EndOfSection
  CreateCoCoADesktopShortCut:
    CreateShortCut "$DESKTOP\CoCoA Console.lnk" "$INSTDIR\cocoa.exe" "" "$INSTDIR\cocoa.ico" 0
  EndOfSection:
    ClearErrors
SectionEnd

Section "Create Context Menu Entries"
  SectionIn 123
  ClearErrors

; Icon detection
; CheckForExe:
    IfFileExists "$INSTDIR\wincocoa.exe" UseExe CheckForIcon
  CheckForIcon:
    IfFileExists "$INSTDIR\cocoa.ico" UseIcon CopyIcon
  UseExe:
    StrCpy $0 "$INSTDIR\wincocoa.exe,0"
    Goto EndDetection
  CopyIcon:
    File /oname=cocoa.ico "..\UI\res\wincocoa.ico"
  UseIcon:
    StrCpy $0 "$INSTDIR\cocoa.ico"
    Goto EndDetection
  EndDetection:
  
; Try to register the file extensions .coc and .pkg for CoCoA 4.2
; Check_coc:
    ReadRegStr $1 HKCR ".coc" ""
    IfErrors Register_coc
    StrCmp $1 "CoCoAFile" Register_CoCoAFile ReplaceAsk_coc
  ReplaceAsk_coc:
    MessageBox MB_YESNO|MB_ICONQUESTION "The file extension .coc is currently registered for the file type $1. $\nDo you want to register it for CoCoA 4.2 (thus overwriting the old value)?" IDYES Register_coc
    Goto Check_pkg
  Register_coc:
    WriteRegStr HKCR ".coc" "" "CoCoAFile"
  Register_CoCoAFile:
    WriteRegStr HKCR "CoCoAFile" "" "CoCoA Source"
    WriteRegStr HKCR "CoCoAFile\DefaultIcon" "" "$0"
    IfFileExists "$INSTDIR\wincocoa.exe" Register_coc_shell Check_pkg
  Register_coc_shell:
    WriteRegStr HKCR "CoCoAFile\shell\open\command" "" "$INSTDIR\wincocoa.exe %1"
  Check_pkg:
    ReadRegStr $1 HKCR ".pkg" ""
    IfErrors Register_pkg
    StrCmp $1 "CoCoAPackage" Register_CoCoAPackage ReplaceAsk_pkg
  ReplaceAsk_pkg:
    MessageBox MB_YESNO|MB_ICONQUESTION "The file extension .pkg is currently registered for the file type $1. $\nDo you want to register it for CoCoA 4.2 (thus overwriting the old value)?" IDYES Register_pkg
    Goto ExtensionsRegistered
  Register_pkg:
    WriteRegStr HKCR ".pkg" "" "CoCoAPackage"
  Register_CoCoAPackage:
    WriteRegStr HKCR "CoCoAPackage" "" "CoCoA Package"
    WriteRegStr HKCR "CoCoAPackage\DefaultIcon" "" "$0"
    IfFileExists "$INSTDIR\wincocoa.exe" Register_pkg_shell ExtensionsRegistered
  Register_pkg_shell:
    WriteRegStr HKCR "CoCoAPackage\shell\open\command" "" "$INSTDIR\wincocoa.exe %1"
  ExtensionsRegistered:
  ClearErrors
SectionEnd

Section "-EasterEgg"
  ClearErrors
; advanced users know what to put into the registry :-)
; ReadEasterEggKey:
    ReadRegStr $0 HKCU "Software\CoCoA" "EasterEgg"
    IfErrors EndOfSection
; ReadHKCU:
    ReadRegStr $0 HKCU "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES"
    IfErrors ReadHKLM
    StrCmp $0 "" ReadHKLM CheckWinCoCoAIni
  ReadHKLM:
    ReadRegStr $0 HKLM "Software\CoCoA\4.2\Environment" "COCOA_GUI_PROPERTIES"
    IfErrors SetDefaultWinCoCoAIniPath
    StrCmp $0 "" SetDefaultWinCoCoAIniPath CheckWinCoCoAIni
  SetDefaultWinCoCoAIniPath:
    StrCpy $0 "$INSTDIR\wincocoa.ini"
  CheckWinCoCoAIni:
    IfFileExists "$0" ReadWinCoCoAIni SetDefaultValue
  ReadWinCoCoAIni:
    ReadINIStr $1 "$0" "Application properties" "autoIndentation"
    IfErrors SetDefaultValue
    StrCmp $1 "" SetDefaultValue WriteWinCoCoAIni
  SetDefaultValue:
    StrCpy $1 "1"
  WriteWinCoCoAIni:
    WriteINIStr "$0" "Application properties" "autoIndentation" "$1"
  EndOfSection:
    ClearErrors
SectionEnd

Section "-PostInstall"
; register in WindowsSoftware/Uninstall tab. this is not visible for the user
  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\CoCoA 4.2" "DisplayName" "CoCoA 4.2 (remove only)"
  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\CoCoA 4.2" "UninstallString" '"$INSTDIR\uninst.exe"'
SectionEnd

Section Uninstall
  ClearErrors
; delete registry stuff
  DeleteRegKey HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\CoCoA 4.2"
  DeleteRegKey HKLM "Software\CoCoA\4.2\Environment"
  DeleteRegKey HKLM "Software\CoCoA\4.2"
; Don't delete "HKLM\Software\CoCoA" since there might be other versions...(ok, not yet)
  DeleteRegKey HKCU "Software\CoCoA\4.2\Environment"
  DeleteRegKey HKCU "Software\CoCoA\4.2"
; Don't delete "HKCU\Software\CoCoA" since there might be other versions...(ok, not yet)

; If the file extensions .coc or .pkg are registered for CoCoA 4.2 remove them
; Check_coc:
    ReadRegStr $0 HKCR ".coc" ""
    IfErrors Check_pkg
    StrCmp $0 "CoCoAFile" Remove_coc Check_pkg
  Remove_coc:
    DeleteRegKey HKCR ".coc"
  Check_pkg:
    ReadRegStr $0 HKCR ".pkg" ""
    IfErrors ExtensionsRemoved
    StrCmp $0 "CoCoAFile" Remove_pkg ExtensionsRemoved
  Remove_pkg:
    DeleteRegKey HKCR ".pkg"
  ExtensionsRemoved:

  DeleteRegKey HKCR "CoCoAFile\DefaultIcon"
  DeleteRegKey HKCR "CoCoAFile\shell\open\command"
  DeleteRegKey HKCR "CoCoAFile\shell\open"
  DeleteRegKey HKCR "CoCoAFile\shell"
  DeleteRegKey HKCR "CoCoAFile"
  
; delete files in INSTDIR
  RMDIR /r "$INSTDIR\doc"
; if there are contrib directory created by the user, they will not be deleted
  RMDIR /r "$INSTDIR\lib\cocoa"
  RMDIR /r "$INSTDIR\lib\contrib\bigatti"
  RMDir /r "$INSTDIR\lib\contrib\bimacros"
  RMDir /r "$INSTDIR\lib\contrib\biro"
  RMDir /r "$INSTDIR\lib\contrib\caboara"
  RMDir /r "$INSTDIR\lib\contrib\dalbaz"
; if there are no contrib directories created by the user, delete the lib directory
  RMDIR "$INSTDIR\lib\contrib"
  RMDIR "$INSTDIR\lib"
  Delete "$INSTDIR\cocoa.exe"
  Delete "$INSTDIR\cocoa.bat"
  Delete "$INSTDIR\Readme.txt"
  Delete "$INSTDIR\uninst.exe"
  Delete "$INSTDIR\wincocoa.exe"
  Delete "$INSTDIR\wincocoa.bat"
  Delete "$INSTDIR\cocoa.ico"
  Delete "$INSTDIR\wordlist.txt"
; don't delete wincocoa.ini and userinit.coc as they contain user data
; INSTDIR should never be empty now, but who knows?
  RMDir "$INSTDIR"

; delete start menu group
  RMDIR /r "$SMPROGRAMS\CoCoA 4.2"

; delete desktop links (if exist)
  Delete "$DESKTOP\CoCoA Console.lnk"
  Delete "$DESKTOP\WinCoCoA.lnk"
SectionEnd