/* $Id: ca.c,v 4.7 2004/06/29 17:11:52 cocoa Exp $
[ CDEV (OMULTI) ]
*/

#include "ca.h"
#include "cterm.h"
#include "ck.h"
#include "map.h"
#include "olist.h"
#include "object.h"
#include "options.h"
#include "operator.h"
#include "ctbuf.h"

#include "machine.h"

/* ---[ PLATFORM Independent ]--------------------------------------------- */

char * cdev_prot_names[CDEV_NUM_PROTOCOLS+1] = {
  "None", "CoCoAL", "CTerm", "Multi", "UserDef" };
char * cdev_type_names[CDEV_NUM_TYPES+1] = {
  "Null", "IFile", "OFile", "IString", "OString", "IWindow", "OWindow",
  "Multiple", "Socket"
};

#ifdef DEBUG_CDEV_RC

cdev cdev_link(cdev D)
{
  return (D != cdev_NULL ? (cdev_incr_rc(D),D) : cdev_NULL);
}

void cdev_unlink(cdev D)
{
  if (D != cdev_NULL) { 
    cdev_decr_rc(D);
    if (cdev_rc(D)==0) cdev_free(D);
  }
}

#endif


coc_bool ca_interrupted=FALSE;
int ca_err=ca_NOERR;

int ca_pi_init(void)
{
#ifdef DEBUG
  initmem();
#endif
#ifdef COCOA_WITH_RUM
  rum_init_all();
#endif
  options_init();
  return ca_NOERR;
}

int ca_put_message_string(char * s, int (* ca_put_message)(cterm))
{
  cterm t;
  int n;
  t=cterm_mk_NBS(s);
  n=ca_put_message(t);
  cterm_free(t);
  return n;
}

/* ---[ CDEV ]--------------------------------------------- */

cdev cdev_stdin, cdev_stdout, cdev_stderr, cdev_parser, cdev_echo;
cdev cdev_monitor, cdev_log, cdev_out;

void cdev_free(cdev D) 
{
  switch (cdev_type(D)) {
  case cdev_type_istring: cdev_free_istring(D); break;
  case cdev_type_ostring: cdev_free_ostring(D); break;
  case cdev_type_iwindow: cdev_free_iwindow(D); break;
  case cdev_type_owindow: cdev_free_owindow(D); break;
  case cdev_type_omulti: cdev_free_omulti(D); break;
  }
  free_str1(cdev_name(D));
  free_cdev(D);
}

int cdev_unput(cdev D, char c) {
  return ca_NOERR; /* TODO: what? */
}

/* ---[ CDEV (cterm) ]--------------------------------------------- */

int ca_put_list_cterm(cdev d, int n, cterm_list l)
{ int i;
  if (n == 0) return ca_NOERR;
  ca_put_cterm(d,l[0]);
  for (i=1;i<n;i++) {
    cdev_puts(",",d);
    ca_put_cterm(d,l[i]);
    }
  return ca_NOERR;
}

int ca_put_cterm(cdev d, cterm t)
{
  int pkg_n;
  coop op=cterm_op(t);
  if (op == op_NBS) {
    cdev_puts("\"",d);
    cdev_puts(cterm_string(t),d);
    cdev_puts("\"",d); }
  else {
    pkg_n=op_package(op);
    op=op_index(op);
    cdev_puts(op_map_op_name(pkg_n,op),d);
    cdev_puts("(",d);
    ca_put_list_cterm(d,cterm_arity(t),cterm_params(t));
    cdev_puts(")",d); 
  }
  return ca_NOERR;
}

/* ---[ cdev_get_cterm ]--------------------------------------------- */

/* JAA: the variable below used to be of type char but was used for holding */
/* the value EOF -- bad programming practice, and troublesome on some       */
/* platforms.  I have changed it to type "int" but have assumed that        */
/* comparisons against char constants and char variables do not require     */
/* that cdev_cterm_la_char be cast into type char.                          */
int cdev_cterm_la_char=' '; /* look-ahead character for cdev_get_cterm */

void cdev_cterm_getc(cdev d);
cterm cdev_cterm_read_cterm(cdev d);
coop cdev_cterm_read_op(cdev d);
cterm cdev_cterm_read_NBS(cdev d);
ctbuf cdev_cterm_read_params(cdev d, char sep, char t);

cterm cdev_get_cterm(cdev d)
{
  /* cdev_cterm_getc(d); */
  return cdev_cterm_read_cterm(d);
}

void cdev_cterm_getc(cdev d)
{
  do {
    cdev_cterm_la_char=cdev_getc(d);
  } while (cdev_cterm_la_char > (char)0 && cdev_cterm_la_char <= (char)32);
}

cterm cdev_cterm_read_cterm(cdev d)
{
  ctbuf aux_b;
  coop op;
  while (cdev_cterm_la_char=='.' || cdev_cterm_la_char==' ')
    cdev_cterm_getc(d); /* skip it */
  if (cdev_cterm_la_char=='"') return cdev_cterm_read_NBS(d);
  op=cdev_cterm_read_op(d);
  if (cdev_cterm_la_char=='(') {
    cdev_cterm_getc(d); /* read '(' */
    aux_b=cdev_cterm_read_params(d,',',')');
    cdev_cterm_getc(d); /* read ')' */
    return ctbuf_mk_cterm(aux_b,op); }
  return cterm_mk_const(op);
}

coop cdev_cterm_read_op(cdev d)
{
  char op_name[100];
  int n;
  if (cdev_cterm_la_char==EOF) return op_NULL;
  op_name[0]=(char)cdev_cterm_la_char;
  n=1;
  cdev_cterm_getc(d);
  while (isalpha(cdev_cterm_la_char) || cdev_cterm_la_char == '_') {
    op_name[n++]=(char)cdev_cterm_la_char;
    cdev_cterm_getc(d); }
  op_name[n]='\0';
  return op_map_coop(pkg_builtin_n,op_name); /* TODO: just a trick */
}

cterm cdev_cterm_read_NBS(cdev d)
{
  int n;
  char str[1000]; /* it MUST suffice (TODO) */
  cdev_cterm_la_char=cdev_getc(d); /* read '"' */
  n=0;
  while(cdev_cterm_la_char!='"') {
    str[n++]=(char)cdev_cterm_la_char;
    cdev_cterm_la_char=cdev_getc(d); }
  str[n]='\0';
  cdev_cterm_getc(d); /* read '"' */
  return cterm_mk_NBS(str);
}

ctbuf cdev_cterm_read_params(cdev d, char sep, char t)
{
  ctbuf res_b;
  cterm aux_t;
  res_b=ctbuf_empty();
  while(cdev_cterm_la_char!=t) {
    aux_t=cdev_cterm_read_cterm(d);
    res_b=ctbuf_append(res_b,aux_t);
    if (cdev_cterm_la_char==sep) cdev_cterm_getc(d); } /* read sep */
  return res_b;
}

/* ---[ CDEV (CoCoAL) ]--------------------------------------------- */

int ca_put_cocoal(cdev d, cterm t)
{
  pp_cterm_main(d,t,0);
  return ca_NOERR;
}

/* ---[ CDEV (OMULTI) ]------------------------------------------ */

cdev cdev_dup_omulti(cdev D) {
  cdev res_D;
  res_D=cdev_init(cdev_name(D), cdev_type_omulti, cdev_prot_multi);
  cdev_set_omulti(res_D, olist_dup(cdev_the_omulti(D)));
  return res_D;
}

void cdev_omulti_add(cdev OMD, cdev D) {
  cdev_set_omulti(OMD,
    olist_append_this(cdev_the_omulti(OMD), cdev2object(cdev_link(D))));
}

void cdev_omulti_remove(cdev OMD, cdev D) {
  int i,n=0;
  object aux_o;
  olist ol;
  ol=cdev_the_omulti(OMD);
  for (i=1; i<=olist_len(ol); i++) {
    aux_o=olist_nth(ol,i);
    if (cdev_id(aux_o->v.cdev_value) == cdev_id(D)) n=i;
  }
  olist_remove(cdev_the_omulti(OMD), n);
}

cdev cdev_open_omulti(char * devname) {
  cdev res_d;
  res_d=cdev_init(devname, cdev_type_omulti, cdev_prot_multi);
  cdev_set_omulti(res_d, olist_init(0));
  return res_d;
}

int cdev_flush_omulti(cdev D) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=cdev_flush(aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

int cdev_close_omulti(cdev D) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=cdev_close(aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

int cdev_put_dashes_omulti(cdev D) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=ca_put_dashes(aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

int cdev_putc_omulti(char c, cdev D) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=cdev_putc(c,aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

int cdev_puts_omulti(char * s, cdev D) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=cdev_puts(s,aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

int ca_put_multi(cdev D, cterm t) {
  olist aux_ol;
  object aux_o;
  int i, res_n;
  aux_ol=cdev_the_omulti(D);
  for (i=1;i<=olist_len(aux_ol);i++) {
    aux_o=olist_nth(aux_ol,i);
    res_n=ca_put(t,aux_o->v.cdev_value);
    if (res_n != ca_NOERR) return res_n;
  }
  return ca_NOERR;
}

/* ---[ CDEV (ISTRING, OSTRING) ]------------------------------------------ */

cdev_istring cdev_istring_init(char * b, int p)
{
  cdev_istring res_is;
  res_is=malloc_cdev_istring();
  cdev_istring_set_buffer(res_is,b);
  cdev_istring_set_pos(res_is,p);
  return res_is;
}

cdev cdev_open_istring(char * devname, char * s, int prot) {
  cdev res_d;
  cdev_istring is;
  res_d=cdev_init(devname, cdev_type_istring, prot);
  is=cdev_istring_init(s,0);
  cdev_set_istring(res_d,is);
  return res_d;
}

void cdev_free_omulti(cdev D)
{
  TODO("cdev_free_omulti");
}

void cdev_free_istring(cdev D) 
{
  cdev_istring is;
  is=cdev_the_istring(D);
  free_str1(cdev_istring_buffer(is));
  free_cdev_istring(is);
}

void cdev_free_ostring(cdev D) 
{
  charbuf_free(cdev_the_ostring(D)); /* TODO: trasformare in macro */
}

cdev cdev_open_ostring(char * devname, int prot) 
{
  cdev res_d;
  charbuf os;
  res_d=cdev_init(devname, cdev_type_ostring, prot);
  os=charbuf_init(50,50);
  cdev_set_ostring(res_d,os);
  return res_d;
}

/* ---[ CDEV (type driven) ]------------------------- */

cdev cdev_init(char * name, int t, int prot)
{
  static progressivo=0;
  cdev res_d;
  res_d=malloc_cdev();
  cdev_set_curr_line(res_d, 1);
  cdev_set_id(res_d,progressivo++);
  cdev_set_rc(res_d,1);
  cdev_set_name(res_d,my_strdup(name));
  cdev_set_type(res_d,t);
  cdev_set_prot(res_d,prot);
  return res_d;
}

cdev cdev_dup(cdev D) {
  cdev res_D;
  switch (cdev_type(D)) {
    case cdev_type_omulti: res_D=cdev_dup_omulti(D); break;
    default: res_D=cdev_link(D);
  }
  return res_D;
}

int cdev_flush(cdev D) {
  int res_n=ca_NOERR;
  switch (cdev_type(D)) {
    case cdev_type_ofile: res_n=cdev_flush_ofile(D); break;
    case cdev_type_owindow: res_n=cdev_flush_owindow(D); break;
    case cdev_type_omulti: res_n=cdev_flush_omulti(D); break;
  }
  return res_n;
}

int cdev_close(cdev D) {
  int res_n=ca_NOERR;
  switch (cdev_type(D)) {
    case cdev_type_ifile: res_n=cdev_close_ifile(D); break;
    case cdev_type_ofile: res_n=cdev_close_ofile(D); break;
    case cdev_type_omulti: res_n=cdev_close_omulti(D); break;
    case cdev_type_owindow: res_n=cdev_close_owindow(D); break;
    case cdev_type_socket: res_n=cdev_close_socket(D); break;
  }
  if (res_n == ca_NOERR) cdev_set_type(D,-cdev_type(D));
  return res_n;
}

int cdev_close_socket(cdev D)
{
  cocoa_close(cdev_the_socket(D));
  free(cdev_the_socket(D));
  cdev_the_socket(D) = 0; /* set socket ptr to null */
  return ca_NOERR;
}

int cdev_getc(cdev d) {
  int res_c=EOF;
  if (cdev_is_closed(d)) { ca_err=ca_ERR_cdev_closed; return res_c; }
  switch (cdev_type(d)) {
    case cdev_type_ifile: res_c=cdev_getc_ifile(d); break;
    case cdev_type_istring:
      res_c=cdev_istring_getc(cdev_the_istring(d));
      if (res_c=='\0') {
        res_c=EOF;
        cdev_istring_decr_pos(cdev_the_istring(d)); }
      break;
    case cdev_type_socket: res_c=cdev_getc_socket(d); break;
    default: res_c=EOF;
  }
  if (res_c=='\n') cdev_incr_curr_line(d);
  return res_c;
}

int cdev_getc_socket(cdev d)
{
  int c;
  cocoa_sock_getc(cdev_the_socket(d), &c); /* TODO: return value check */
  return c;
}

int cdev_puts(char * s, cdev d)
{
  switch (cdev_type(d)) {
  case cdev_type_ofile: return cdev_puts_ofile(s, d); break;
  case cdev_type_owindow: return cdev_puts_owindow(s, d); break;
  case cdev_type_omulti: return cdev_puts_omulti(s, d); break;
  case cdev_type_ostring:
    charbuf_puts(cdev_the_ostring(d), s);
    return ca_NOERR;
    break;
  case cdev_type_socket: return cdev_puts_socket(s,d); break;
  default: return ca_ERR_unknown_cdev_type;
  }
}

int cdev_puts_socket(char * s, cdev d)
{
  cocoa_send(cdev_the_socket(d), s, strlen(s));
  return ca_NOERR;
}

int cdev_putc(char c, cdev d) {
  if (cdev_is_closed(d)) return ca_ERR_cdev_closed;
  switch (cdev_type(d)) {
  case cdev_type_ofile: return cdev_putc_ofile(c,d); break;
  case cdev_type_owindow: return cdev_putc_owindow(c, d); break;
  case cdev_type_omulti: return cdev_putc_omulti(c, d); break;
  case cdev_type_ostring: charbuf_putc(cdev_the_ostring(d), c);
    return ca_NOERR;
    break;
  case cdev_type_socket: return cdev_putc_socket(c,d); break;
  default: return ca_ERR_unknown_cdev_type;
  }
}

int cdev_putc_socket(char c, cdev d)
{
  cocoa_send(cdev_the_socket(d), &c, sizeof(char));
  return ca_NOERR;
}

cdev cdev_open(int t, int prot, char * aux_s, cdev_params p)
{
  cdev res_d;
  switch (t) {
  case cdev_type_ifile: res_d=cdev_open_ifile(aux_s, prot); break;
  case cdev_type_ofile: res_d=cdev_open_ofile(aux_s, prot, (cdev_ofile_params)p); break;
  case cdev_type_istring: 
    res_d=cdev_open_istring(aux_s, my_strdup((char *)p), prot); break;
  case cdev_type_ostring: res_d=cdev_open_ostring(aux_s, prot); break;
  case cdev_type_iwindow: res_d=cdev_open_iwindow(aux_s, prot); break;
  case cdev_type_owindow: res_d=cdev_open_owindow(aux_s, prot, (cdev_owindow_params)p); break;
  case cdev_type_omulti: res_d=cdev_open_omulti(aux_s); break;
  case cdev_type_socket: res_d=cdev_open_socket(aux_s, prot); break;
  }
  return res_d;
}

cdev cdev_open_socket(char * hostname, int prot)
{
	cdev res_d;
  	struct sock_t * sock;
	int res;
	
	sock = malloc(sizeof(struct sock_t)); /* I couldn't come up with anything better */
	init_sock(sock);
	res = cocoa_connect(sock, hostname, COCOA_PORT);
	if (res < 0)
        {
          free(sock);
          return cdev_NULL;
        }
	res_d = cdev_init(hostname, cdev_type_socket, prot);
	cdev_set_socket(res_d, sock);
	return res_d;
}

int ca_put_dashes(cdev D)
{
	switch (cdev_type(D)) {
                case cdev_type_owindow: cdev_put_dashes_owindow(D); break;
		case cdev_type_omulti: cdev_put_dashes_omulti(D); break;
		default: ca_put(cterm_dashes,D); break;
	}
	return ca_NOERR;
}


/* ---[ CDEV (protocol driven) ]------------------------- */

cterm ca_get(cdev d)
{
  switch (cdev_prot(d)) { 
  case cdev_prot_cocoal: {
    cdev_parser=d;
    yyparse(); 
    return yycterm; }
  case cdev_prot_cterm:
    return cdev_get_cterm(d);
  }
  return cterm_NULL;
}

int ca_put(cterm t, cdev d)
{
  if (d==cdev_NULL) return ca_NOERR;
  switch (cdev_prot(d)) { 
  case cdev_prot_cterm: return ca_put_cterm(d,t); break;
  case cdev_prot_cocoal: return ca_put_cocoal(d,t); break;
  case cdev_prot_multi: return ca_put_multi(d,t); break;
  }
  return ca_ERR_unknown_cdev_type;
}

cterm cdev2cterm(cdev D)
{
  ctbuf res_b;
  res_b=ctbuf_empty();
  ctbuf_append(res_b,
    cterm_mk_table_entry("Name",string2cterm(cdev_name(D))));
  ctbuf_append(res_b,
    cterm_mk_table_entry("Type",string2cterm(cdev_type_name(D))));
  ctbuf_append(res_b,
    cterm_mk_table_entry("Protocol",string2cterm(cdev_prot_name(D))));
  return ctbuf_mk_cterm(res_b,op_RECORD);
}

/* ---[ utils ]--------------------------------------------- */

void printf_string(char * s)
{
  cterm t;
  t=cterm_mk_NBS(s);
  put_cterm(t);
  cterm_free(t);
}
