/* $Id: ca.h,v 4.5 2004/06/21 12:17:46 cocoa Exp $ */

#ifndef CA_H
#define CA_H

#include "cocdefs.h"
#include "auxtypes.h"
#include "utils.h"
#include "net_lib.h"

#define ca_NOERR 0
#define ca_ERR_unknown_cdev_type 2001
#define ca_ERR_cdev_closed 2002
#define ca_ERR_socket 2003

extern coc_bool ca_interrupted;
extern int ca_err;

cdev cdev_init(char * name, int cdev_type, int prot);
int cdev_close(cdev D);
int cdev_close_ifile(cdev D);
int cdev_close_ofile(cdev D);
int cdev_close_owindow(cdev D);
int cdev_close_socket(cdev D);
cdev cdev_dup(cdev D);
int cdev_flush(cdev D);
int cdev_flush_ofile(cdev D);
int cdev_getc(cdev D);
int cdev_putc(char c, cdev d);

/* ---[ PUBLIC ]----------------------------------------------- */

#define cdev_is_closed(d) ((d)->t < 0)

cterm ca_get(cdev d);
int ca_put(cterm t, cdev d);
void ca_monitor(int what);
int ca_message(cterm t);
int ca_warning(cterm t);
int ca_error(cterm t);
int ca_internal_error(cterm t);

char* ca_cocoa_packages();
char* ca_getenv(const char* name);


extern cdev cdev_stdin, cdev_stdout, cdev_stderr, cdev_echo;
extern cdev cdev_monitor, cdev_log, cdev_out, cdev_parser;
extern cterm yycterm;

/* ---[ PLATFORM Independent ]--------------------------------------------- */

int ca_pi_init(void);
/* initialization of the common modules between ca and ck */

#define ca_open_log(D) cdev_omulti_add(cdev_log, D)
#define ca_close_log(D) cdev_omulti_remove(cdev_log, D)

#define ca_message_string(s) ca_put_message_string(s,&ca_message)
#define ca_warning_string(s) ca_put_message_string(s,&ca_warning)
#define ca_error_string(s) ca_put_message_string(s,&ca_error)
#define ca_internal_error_string(s) ca_put_message_string(s,&ca_internal_error)

int ca_put_message_string(char * s, int (* ca_put_message)(cterm));
int ca_put_cterm(cdev d, cterm t);

#define ca_put_nl(d) ca_put(cterm_newline, d)

int ca_put_dashes(cdev d);

#define ca_put_and_log(t,d) \
  do { ca_put(t,d); ca_put(t,cdev_log); } while(FALSE)

#define ca_redef_cdev_log(d) \
  do { cdev_unlink(cdev_log); cdev_log=d; } while(FALSE)

/* ---[ compatibility ]--------------------------------------------- */

#define put_cterm(t) ca_put(t,cdev_out)

/* ---[ CDEV ]--------------------------------------------- */

#define cdev_NULL NULL

#define CDEV_NUM_TYPES 8

#define cdev_type_ifile 1
#define cdev_type_ofile 2
#define cdev_type_istring 3
#define cdev_type_ostring 4
#define cdev_type_iwindow 5
#define cdev_type_owindow 6
#define cdev_type_omulti 7
#define cdev_type_socket 8

#define CDEV_NUM_PROTOCOLS 4

#define cdev_prot_NULL 0
#define cdev_prot_cocoal 1
#define cdev_prot_cterm 2
#define cdev_prot_multi 3
#define cdev_prot_userdef 4


extern char * cdev_prot_names[];
extern char * cdev_type_names[];

#define malloc_cdev_istring() \
  (cdev_istring)mymalloc(sizeof(cdev_istring_info),"cdev_istring")
#define free_cdev_istring(D) myfree(sizeof(cdev_istring_info),D,"cdev_istring")

#define cdev_id(d) ((d)->id)
#define cdev_set_id(d,n) (d)->id=n
#define cdev_curr_line(d) ((d)->curr_line)
#define cdev_set_curr_line(d,l) (d)->curr_line=l
#define cdev_incr_curr_line(d) ((d)->curr_line++)

#define cdev_istring_set_buffer(is,b) (is)->buffer=b
#define cdev_istring_set_pos(is,p) (is)->pos=p
#define cdev_istring_decr_pos(is) (is)->pos--
#define cdev_istring_getc(is) (is)->buffer[(is)->pos++]
#define cdev_istring_eof(is) ((is)->buffer[(is)->pos]=='\0')
#define cdev_istring_buffer(is) ((is)->buffer)

void cdev_free(cdev D);
int cdev_unput(cdev D, char c);

int cdev_puts(char * s, cdev d);
int cdev_puts_ofile(char * s, cdev d);
int cdev_puts_owindow(char * s, cdev d);
int cdev_puts_socket(char * s, cdev d);
int cdev_put_dashes_owindow(cdev d);

int cdev_putc_socket(char c, cdev d);
int cdev_getc_socket(cdev d);

cdev cdev_open_ifile(char * filename, int prot);
cdev cdev_open_ofile(char * filename, int prot, cdev_ofile_params p);
cdev cdev_open_istring(char * devname, char * s, int prot);
cdev cdev_open_ostring(char * devname, int prot);
cdev cdev_open_iwindow(char * devname, int prot);
cdev cdev_open_owindow(char * devname, int prot, cdev_owindow_params p);
cdev cdev_open_omulti(char * devname);
cdev cdev_open_socket(char * hostname, int prot);

cterm cdev2cterm(cdev D);

/* check if can open. If not so return cdev_null, otherwise call
   cdev_init and assign its parameters */
cdev cdev_open(int t, int prot, char * aux_s, cdev_params p);

/* ---[ CDEV (OMULTI) ]------------------------------------------ */

cdev cdev_dup_omulti(cdev D);
void cdev_omulti_add(cdev OMD, cdev D);
void cdev_omulti_remove(cdev OMD, cdev D);
void cdev_free_omulti(cdev OMD);

/* ---[ CDEV (cterm) ]--------------------------------------------- */

int cdev_put_cterm(cdev d, cterm t);
cdev_istring cdev_istring_init(char * b, int p);
void cdev_free_istring(cdev d);
void cdev_free_ostring(cdev d);
void cdev_free_iwindow(cdev d);
void cdev_free_owindow(cdev d);

cterm cdev_get_cterm(cdev d);

#define malloc_cdev() (cdev)mymalloc(sizeof(cdev_info),"cdev")
#define free_cdev(D) myfree(sizeof(cdev_info),D,"cdev")

#ifdef DEBUG_CDEV_RC

cdev cdev_link(cdev D);
void cdev_unlink(cdev D);

#else

#define cdev_link(ev) (cdev_incr_rc(ev),ev)

#define cdev_unlink(D) \
  if (D != cdev_NULL) do { cdev_decr_rc(D); \
       if (cdev_rc(D)==0) cdev_free(D);} while(FALSE)
       
#endif

#define cdev_set_ifile(D,f) (D)->d.ifile=f
#define cdev_set_ofile(D,f) (D)->d.ofile=f
#define cdev_set_istring(D,IS) (D)->d.is=IS
#define cdev_set_ostring(D,OS) (D)->d.os=OS
#define cdev_set_omulti(D,OM) (D)->d.om=OM
#define cdev_set_socket(D,s) (D)->d.socket=s

#define cdev_the_ifile(D) ((D)->d.ifile)
#define cdev_the_ofile(D) ((D)->d.ofile)
#define cdev_the_istring(D) ((D)->d.is)
#define cdev_the_ostring(D) ((D)->d.os)
#define cdev_the_omulti(D) ((D)->d.om)
#define cdev_the_socket(D) ((D)->d.socket)

#define cdev_incr_rc(D) ((D)->rc++)
#define cdev_decr_rc(D) ((D)->rc--)

#define cdev_rc(D) ((D)->rc)
#define cdev_name(D) ((D)->name)
#define cdev_type(D) (ABS((D)->t))
/* When closed, a cdev negates its type; line below preserves its type name */
#define cdev_type_name(D) cdev_type_names[abs((D)->t)]
#define cdev_prot(D) ((D)->prot)
#define cdev_prot_name(D) cdev_prot_names[(D)->prot]
#define cdev_set_rc(D,N) (D)->rc=N
#define cdev_set_name(D,N) (D)->name=N
#define cdev_set_type(D,T) (D)->t=T
#define cdev_set_prot(D,P) (D)->prot=P

/* ---[ utils ]--------------------------------------------- */

void printf_string(char * s);

#endif
