/* $Id: ck.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include <stdio.h>

#include "cocdefs.h"
#include "cterm.h"
#include "ctbuf.h"
#include "ck.h"
#include "ca.h"
#include "engine.h"
#include "memoryco.h"
#include "mathcoc.h"
#include "olist.h"
#include "operator.h"
#include "options.h"
#include "ring.h"

extern int funcall_level;

/*
[ PRIVATE headers ] [ PUBLIC ] [ PRIVATE ] [ option panels ] [ coops ]
[ ck_engine() ]
*/


/* ---[ PRIVATE headers ]------------------------------------------ */

void ck_opt_init(void);
void ck_init_coops(void);

/* ---[ PUBLIC ]----------------------------------------------- */

ring ring_Z,ring_Q,ring_ShZ;
env current_env, cocoa_env, ring_env, panels_env, memory_env, pkg_env, err_env;
env cdev_env;
cterm cterm_newline, cterm_dashes;
ctbuf comp_stack;

/* GMP_RANDSTATE is referred to in math.c for mi_seed (qv) */
gmp_randstate_t GMP_RANDSTATE;
int GMP_RANDSIZE;

void initialize_predef_panels(ring R);
void initialize_predef_rings(void);
void ck_rum_init(void);

void ck_init_cterms(void)
{
  cterm_newline=cterm_mk_const(op_NEWLINE);
  cterm_dashes=cterm_mk_const(op_DASHES);
}

int ck_init(void)
{
  cdev cdev_echo;
  GMP_RANDSIZE = 8*sizeof(long);
  gmp_randinit(GMP_RANDSTATE, GMP_RAND_ALG_LC, GMP_RANDSIZE);
  gmp_randseed_ui(GMP_RANDSTATE, 1); /* arbitrary but fixed seed */
  map_init_maps();
#ifdef COCOA_WITH_RUM
  ck_rum_init();
#endif
  ck_opt_init();
  ck_init_coops();
  ck_init_cterms();
  comp_stack = ctbuf_empty();
  pp_init();
  opt_set_value(opt_panel_INTERNAL,flag_ENGINE_EXIT,flag_OFF);
  object_init_adt();
  current_env=env_init();
  cocoa_env=env_init();  
  memory_env=env_init();  
  cdev_env=env_init();
  pkg_env=env_init();
  err_env=env_init();
  initialize_predef_rings();
  current_ring=ring_Z;
  funcall_level=0;
  initialize_predef_panels(current_ring);
  cdev_out=cdev_open_omulti("out");
  cdev_log=cdev_open_omulti("log");
  /* cdev_echo=cdev_open_ofile("echo", cdev_prot_cterm); */
  cdev_omulti_add(cdev_out, cdev_log);
  cdev_omulti_add(cdev_log, cdev_stdout);
  /* cdev_omulti_add(cdev_log, cdev_echo); */

  env_declare_and_assign(&cdev_env,"STDOUT",
    cdev2object(cdev_link(cdev_stdout)));
  env_declare_and_assign(&cdev_env,"STDIN",
    cdev2object(cdev_link(cdev_stdin)));
  env_declare_and_assign(&cdev_env,"OUT",
    cdev2object(cdev_link(cdev_out)));
  env_declare_and_assign(&cocoa_env,"DEV",
    record2object(cdev_env));
  env_declare_and_assign(&cocoa_env,"ENV",
    record2object(ring_env));
  env_declare_and_assign(&cocoa_env,"PKG",
    record2object(pkg_env));
  env_declare_and_assign(&cocoa_env,"ERR",
    record2object(err_env));
  env_declare_and_assign(&cocoa_env, "ComputationStack", object_link(object_null));

  env_declare_and_assign(&memory_env,"MEMORY",
    record2object(cocoa_env));
  mi_init();
  engine_init();
  return ck_NOERR;
}

void initialize_predef_panels(ring R)
{
  panels_env=env_init();
  panels_env_assign(opt_panel_descr(opt_panel_GROBNER),opt_panel_GROBNER,R);
  panels_env_assign(opt_panel_descr(opt_panel_ENGINE),opt_panel_ENGINE,R);
#ifdef DEBUG
  panels_env_assign(opt_panel_descr(opt_panel_DEBUG),opt_panel_DEBUG,R);
  panels_env_assign(opt_panel_descr(opt_panel_MONITOR),opt_panel_MONITOR,R);
#endif
}

void initialize_predef_rings(void)
{
  ring_env=env_init();
  ring_Z=ring_of_coeff("Z",characteristic_Li);
  ring_env_assign(ring_Z);
  ring_Q=ring_of_coeff("Q",characteristic_Lr);
  ring_env_assign(ring_Q);
  ring_ShZ=ring_of_coeff("ShZ",characteristic_Si);
}

/* ---[ ck_engine() ]----------------------------------------------- */

int ck_open_dev(cterm t);

/* SEE ALSO int ck_engine_test(cterm t) */

void ck_mem_stat(void)
{
#ifdef DEBUG
  cterm t, mem_params[2];
  int n;
  n=maxmemused;
  mem_params[1]=int2cterm(memused);
  mem_params[0]=int2cterm(n);
  t = cterm_mk_classic(op_MEM_STAT,2,mem_params);
  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_MEMORY)) put_cterm(t);
  cterm_free(t);
#endif
}

int ck_engine_top(cterm t)
{ 
  /* executes commands at top-level */

  object aux_o;
  cterm aux_t;

  ca_interrupted=FALSE;

  /* return ck_engine_test(t); */

  if (math_error != object_NULL) {
    cocoa_internal_error("unexpected math_error != object_NULL");
    object_unlink(math_error);
  }
  math_error=object_NULL;
  error_set_context(cterm_NULL);
  error_set_param(cterm_NULL);
  engine_made_output=FALSE;

  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_ECHO)) {
    ca_put(t, cdev_out);
    ca_put_nl(cdev_out);
  }

  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_COMP_STACK))
    comp_stack=ctbuf_empty();

  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_TIMER))
    aux_o=engine_and_time(t);
  else
    aux_o=engine(t);

  if (aux_o == object_BREAK) {
    object_unlink(aux_o);
    aux_o=cocoa_error("BREAK_WITHOUT_LOOP");
  } else if (aux_o == object_RETURN_NULL) {
    object_unlink(aux_o);
    aux_o=object_link(object_null);
  }

  if (object_is_error(aux_o) && flag_is_on(opt_panel_ENGINE, flag_ENGINE_COMP_STACK)) {
    object stack_thing = olist2object(this_ctbuf2olist(comp_stack));
    env_assign(cocoa_env, "ComputationStack", stack_thing);
  }

  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_TIMER) 
      || aux_o != object_null) {
    if (engine_made_output) {
      ca_put_nl(cdev_out);
      ca_put_dashes(cdev_out);
    }
    engine_print_object(aux_o,cdev_out,current_ring);
    ca_put_nl(cdev_out);
    ca_put_dashes(cdev_out);
    engine_made_output=FALSE;
    object_unlink(aux_o);
  } else if (engine_made_output) {
    ca_put_nl(cdev_out);
    ca_put_dashes(cdev_out); 
    engine_made_output=FALSE;
  }


  cterm_free(t);
  ck_mem_stat();

  return ck_NOERR;
}

int ck_engine(cterm t)
{ 
  object aux_o;

  if (math_error != object_NULL) {
    cocoa_internal_error("unexpected math_error != object_NULL");
    object_unlink(math_error);
  }
  math_error=object_NULL;
  engine_made_output=FALSE;

  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_COMP_STACK))
    comp_stack=ctbuf_empty();

  aux_o=engine(t);

  if (aux_o == object_BREAK) {
    object_unlink(aux_o);
    aux_o=cocoa_error("BREAK_WITHOUT_LOOP");
  } else if (aux_o == object_RETURN_NULL) {
    object_unlink(aux_o);
    aux_o=object_link(object_null);
  }

  if (object_is_error(aux_o)) {
    if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_COMP_STACK)) {
      object stack_thing=olist2object(this_ctbuf2olist(comp_stack));
      env_assign(cocoa_env, "ComputationStack", stack_thing);
    }
    printf_object(aux_o, cdev_out);
    math_error=NULL; /* error object ALIASED in aux_o and DEALLOCATED below */
    ca_put_nl(cdev_out);
    ca_put_dashes(cdev_out);
    engine_made_output=FALSE;
  } else if (engine_made_output) {
    ca_put_nl(cdev_out);
    ca_put_dashes(cdev_out);
    engine_made_output=FALSE;
  }
  cterm_free(t);
  ck_mem_stat();

  if (ca_interrupted) return ck_ERR_interrupted;

  object_unlink(aux_o);

  return ck_NOERR;
}

int ck_source(cdev D)
{
  cterm t;
  int status=ck_NOERR;
  env save_current_env=current_env;
  if (funcall_level>0) {
    current_env=env_init();
  }
  while (status == ck_NOERR && ! ck_finished() && ! cdev_eof(D)) {
    t=ca_get(D);
    if (cterm_op(t)==op_EOF) {
      cterm_free(t);
      break;
    }
    status=ck_engine(t);
  }
  if (cdev_type(D) != cdev_type_socket)
    cdev_close(D);
  if (funcall_level>0) {
    env_free(current_env);
    current_env=save_current_env;
  }
  return status;
}

int ck_source_top(cdev D)
{
  cterm t;
  int status=ck_NOERR;
  while (status == ck_NOERR && ! cdev_eof(D)) {
    t=ca_get(D);
    if (cterm_op(t)==op_EOF) {
      cterm_free(t);
      break;
    }
    status=ck_engine_top(t);
  }
  cdev_close(D);
  return status;
}

int ck_source_from_string(const char * s)
{
  int n;
  cdev D;
  D = cdev_open_istring("",my_strdup((char *)s), cdev_prot_cocoal);
  n = ck_source(D);
  cdev_unlink(D);
  return n;
}

void ck_continue(void)
{
  opt_set_value(opt_panel_INTERNAL,flag_ENGINE_EXIT,flag_OFF);
}

coc_bool ck_finished(void)
{
  return flag_is_on(opt_panel_INTERNAL,flag_ENGINE_EXIT);
}

void ck_quit(void)
{
  gmp_randclear(GMP_RANDSTATE); /* minimal amount of tidying up */
  /* do nothing (for now) */
}

/* ---[ PRIVATE ]----------------------------------------------- */

/* ---[ coops ]-------------------------------------------- */

coop op_POLYNOMIAL, op_MAT, op_FLOAT, op_ZPVALUE, op_ZP, op_SHORT_INT,
op_SHORT_RAT, op_SHORT_DOUBLE, op_LONG_INT, op_LONG_RAT, op_DASHES, 
op_MEM_STAT, op_OBJECTS, op_PANELS, op_MEMORY, op_HANDLER, op_ERROR, op_MKERR,
op_PAIR, op_IDEAL, op_GLOBAL_MEMORY, op_objtype_NULL, op_BREAK,
op_objtype_coeff, op_objtype_ring, op_objtype_poly, op_objtype_bool, 
op_objtype_cterm, op_objtype_int, op_objtype_ideal, op_objtype_cocfun,
op_objtype_pvector, op_objtype_vlist, op_objtype_string, op_objtype_module, 
op_objtype_pseries, op_objtype_objtype, op_objtype_mat, 
op_objtype_olist, op_objtype_omat, op_objtype_ratfun, op_objtype_panel, 
op_objtype_err, op_objtype_arbint, op_objtype_arbrat, op_objtype_record, 
op_objtype_zp, op_objtype_double, op_objtype_tagged, op_objtype_cdev,
op_CURR_RING, op_ISDEF, op_LCM, op_BUILTIN,
op_INDET_NTH, op_LEX, op_XEL, op_DEGLEX, op_DEGREVLEX, op_POS_TO, op_TO_POS, 
op_LEVEL, op_ORD, op_WEIGHTS, op_ENV, op_ENV_SLOT, op_MODULE, op_MONOMIAL, 
op_OBJECT, op_VECTOR, op_RATFUN, op_TERM, op_EEE, op_EEE_MON, op_EEE_LBL,
op_TRIPOLI, op_HOMO, op_CLASSIC, op_SCALARS, op_MON, op_AT;

coop op_ALIAS, op_ALIASES, op_ALIAS_CMD, 
op_ASCII, op_ALIAS_IN, op_ALL, op_AND, op_ADD_DEVICE,
op_ASSIGN, op_CARTESIAN, op_CATCH, op_CHANGE_SIGN, op_CLEAR, 
op_CLOSE, op_CIP,
op_CLOSELOG, op_COCOA_PACKAGES, op_COLON, op_COMMANDS, op_COMPONENT, op_COND, op_CALL,
op_DEFINE_AND_USE_RING, op_DEFINE_RING, op_DELETE_RINGS, op_DELETE_VARS, 
op_DESCRIBE, op_DOTS, op_EOF, op_EOD, op_ELIM,
op_EVAL, op_EXPORT, op_FOR, op_FOREACH, op_FUNCTION, op_EVAL_POLY,
op_FRAC, op_FUN_APPL, op_FLUSH, op_GET, op_GetErrMesg,
op_HELP, op_IF, op_IMPORT, op_INIT_CDEV, 
op_INDET, op_INFOS, op_PROGRAM2,
op_INDET_DECL, op_INDET_INFO, op_IS_EQ, op_IS_GE, op_IS_GT, op_IS_IN, op_IS_LE, 
op_IS_LT, op_MAP, op_MULT, op_NEWLINE, op_NOT, op_NULL_VALUE, op_OPENLOG, op_OR,
op_PACKAGE, op_PARENTHESIS, op_PKG_ID, op_POLY_RING, op_POWER, op_PRINT, 
op_PRINTON, op_PRINTON_AND_CR, op_PRINT_AND_CR, op_PUT, op_PACKAGES,
op_PROGRAM, op_QUIT, op_RANGE, op_REMOVE_DEVICE,
op_RECORD, op_RECORDSEL, op_REPEAT, op_RETURN, op_RING_MODIFIERS, op_RING_ZP, 
op_SET, op_SET_CURRENT_RING, op_SKIP, op_SOURCE, op_SUBTRACT, op_SUM, op_SYSTEM,
op_TERROR, op_TIME, op_UNLABEL, op_VAR, op_WHILE, op_ZPMK, op_PANEL_CMD,
op_SHIFTS;

coop op_RING, op_LIMITS,
op_ADJOINT, op_APPEND, op_REMOVE, op_CAST, op_CHAR, op_CONCAT, op_DATE, op_DEG,
op_DENSE_POLY, op_DER, op_DET, op_DIV, op_FACT, op_FACTOR, op_GCD, op_GB_ALL, op_MDEG,
op_GB_INIT, op_GB_ONE_MOVE, op_GB_STAT,
op_GENS, op_HEAD, op_HOMOG, op_INDET_IND, op_INDET_INDEX, op_INDET_NAME,
op_LABEL, op_LEADING_COEFF, 
op_LEADING_TERM, op_LEN, op_LISTCAST, op_LOG, op_MINIMALIZE, op_MINORS, op_MOD,
op_MVPOINCARE, op_NEWLIST, op_NEWMAT, op_NEWVECTOR, op_NEXTPRIME, op_MOD2RAT,
op_NUM_INDETS, op_POINCARE, op_POLY, op_RANDOMIZE,
op_RES, op_RUM_STAT, op_SHOW_ALIASES, op_SIZE, op_STARTING, op_SUBST, 
op_TAG, op_TAIL, op_TAO, op_TYPEOF, op_FIELDS, op_IS_HOMOG, op_IS_RING_INDEP,
op_COCOLON, op_USING;

coop op_SUPPORT, op_MONOMIALS, op_COEFFS,  op_ISPOSTO, op_ISTOPOS,
op_FIRSTNONZERO, op_FIRSTNONZEROPOS, op_ISTERM, op_LPOS, op_LPP, op_LM,
op_COEFFOFTERM, op_NR, op_NEWPPSTREAM, op_NEXTPP, op_AVOIDLASTPP;

ck_handler * ck_handlers[MAX_NUM_OPERATORS];

coop ck_add_builtin_op(char * id, int arity, ck_handler * h)
{
  coop op;
  op=op_map_add(pkg_builtin_n, id, arity);
  ck_handlers[op]=h;
  return op;
}

void ck_init_coops(void)
{
  int i;
  for (i=0;i<MAX_NUM_OPERATORS;i++) ck_handlers[i]=NULL;

  ck_handlers[op_HELPQUERY]=&mi_helpquery;
  ck_handlers[op_STRING]=&mi_string;
  ck_handlers[op_FALSE]=&mi_bool;
  ck_handlers[op_TRUE]=&mi_bool;
  ck_handlers[op_LIST]=&mi_list;

#ifdef COCOA_DEVELOPING
  ck_add_builtin_op("Hook",2,mi_hook);
#endif
  ck_add_builtin_op("TestSet",2,mi_anna_BuiltinTestSet);
  ck_add_builtin_op("TestSetLex",2,mi_anna_BuiltinTestSetLex);
  ck_add_builtin_op("TestSetXel",2,mi_anna_BuiltinTestSetXel);
  ck_add_builtin_op("NonNegSol",2,mi_anna_BuiltinNonNegSol);
  
  op_EEE = op_map_add(pkg_builtin_n,"EEE",NO_ARITY);
  op_EEE_MON = op_map_add(pkg_builtin_n,"EEE_MON",NO_ARITY);
  op_EEE_LBL = op_map_add(pkg_builtin_n,"EEE_LBL",NO_ARITY);
  op_ADD_DEVICE = ck_add_builtin_op("AddDevice",2,mi_add_device);
  op_ALIAS = op_map_add(pkg_builtin_n,"ALIAS",2);
  op_SHIFTS = op_map_add(pkg_builtin_n,"Shifts",1);
  op_ALIASES = op_map_add(pkg_builtin_n,"\040ALIASES",NO_ARITY);
  op_ALIAS_CMD = ck_add_builtin_op("\040ALIAS_CMD",NO_ARITY,engine_alias);
  op_ASCII = ck_add_builtin_op("Ascii",1,mi_ascii);
  op_ALIAS_IN = ck_add_builtin_op("\040ALIAS_IN",2,engine_alias_in);
  op_ALL = ck_add_builtin_op("\040ALL",4,mi_all);
  op_AND = ck_add_builtin_op("AND",2,mi_bool);
  op_ASSIGN = ck_add_builtin_op("\040ASSIGN",NO_ARITY,engine_assign);
  op_BUILTIN = op_map_add(pkg_builtin_n,"\040BUILTIN",0);
  op_BREAK = ck_add_builtin_op("\040BREAK",NO_ARITY,engine_break);
  op_CARTESIAN = ck_add_builtin_op("\040CARTESIAN",NO_ARITY,mi_cartesian);
  op_CATCH = ck_add_builtin_op("\040CATCH",NO_ARITY,engine_catch);
  op_USING = ck_add_builtin_op("\040USING",2,engine_using);
  op_CHANGE_SIGN = ck_add_builtin_op("\040CHANGE_SIGN",1,mi_change_sign);
  op_CALL = ck_add_builtin_op("Call",NO_ARITY,mi_call);
  op_CLEAR = ck_add_builtin_op("CLEAR",NO_ARITY,engine_clear_rings);
  op_CLOSE = ck_add_builtin_op("Close",1,engine_close);
  op_CLOSELOG = ck_add_builtin_op("CloseLog",1,engine_closelog);
  op_COCOA_PACKAGES = ck_add_builtin_op("CocoaPackagePath",0,mi_cocoa_packages);
  op_COMMANDS = ck_add_builtin_op("\040COMMANDS",NO_ARITY,engine_commands);
  op_COMPONENT = ck_add_builtin_op("Comp",2,mi_component);
  op_COND = ck_add_builtin_op("Cond",NO_ARITY,mi_cond);
  ck_add_builtin_op("DebugVersion",NO_ARITY,mi_debug_version);
  op_DEFINE_AND_USE_RING = 
    ck_add_builtin_op("\040DEFINE_AND_USE",3,engine_define_ring);
  op_DEFINE_RING = ck_add_builtin_op("\040DEFINE_RING",3,engine_define_ring);
  op_DELETE_RINGS = ck_add_builtin_op("DESTROY",NO_ARITY,engine_delete_rings);
  op_DELETE_VARS = 
    ck_add_builtin_op("\040DELETE_VARS",NO_ARITY,engine_delete_vars);
  op_DESCRIBE = ck_add_builtin_op("DESCRIBE",1,engine_describe);
  op_DOTS = op_map_add(pkg_builtin_n,"\040DOTS",0);
  op_ELIM = op_map_add(pkg_builtin_n,"Elim",NO_ARITY);
  op_EOD = ck_add_builtin_op("EOD",0,mi_eod);
  op_EOF = ck_add_builtin_op("\040EOF",0,engine_quit);
  op_EVAL_POLY = ck_add_builtin_op("EvalPoly",2,mi_jaa_eval);
  op_EVAL = ck_add_builtin_op("\040EVAL",NO_ARITY,engine_eval); /* 1 or 2 */
  op_EXPORT = op_map_add(pkg_builtin_n,"\040EXPORT",NO_ARITY);
  op_FLUSH = ck_add_builtin_op("Flush",1,engine_flush);
  op_FIELDS = ck_add_builtin_op("Fields",1,mi_fields);
  op_FUNCTION = ck_add_builtin_op("Function",NO_ARITY,mi_function);
  op_FOR = ck_add_builtin_op("\040FOR",NO_ARITY,engine_for);
  op_FOREACH = ck_add_builtin_op("\040FOREACH",3,engine_foreach);
  op_FRAC = ck_add_builtin_op("Fraction",2,mi_frac);
  op_FUN_APPL = ck_add_builtin_op("\040FUN_APPL",NO_ARITY,mi_fun_appl);
  op_GET = ck_add_builtin_op("Get",2,mi_get);
  op_GetErrMesg = ck_add_builtin_op("GetErrMesg",1,mi_geterrmesg);
  ck_add_builtin_op("GetEnv",1,mi_getenv);
  op_HELP = ck_add_builtin_op("Help",1,mi_help);
  op_IS_RING_INDEP = ck_add_builtin_op("IsRingIndep",1,mi_is_ring_indep);
  op_IS_HOMOG = ck_add_builtin_op("IsHomog",1,mi_is_homog);
  op_IF = ck_add_builtin_op("\040IF",NO_ARITY,engine_if);
  op_IMPORT = op_map_add(pkg_builtin_n,"\040IMPORT",NO_ARITY);
  op_INIT_CDEV = ck_add_builtin_op("OpenDevice",4,mi_init_cdev);
  op_INDET = ck_add_builtin_op("\040INDET",NO_ARITY,mi_indet);
  op_INDET_DECL = op_map_add(pkg_builtin_n,"\040INDET_DECL",NO_ARITY);
  op_INDET_INFO = op_map_add(pkg_builtin_n,"\040INDET_INFO",NO_ARITY);
  ck_add_builtin_op("InElim",2,mi_in_elim);
  op_IS_EQ = ck_add_builtin_op("IsEQ",2,mi_bool);
  op_IS_GE = ck_add_builtin_op("IsGE",2,mi_bool);
  op_IS_GT = ck_add_builtin_op("IsGT",2,mi_bool);
  op_IS_IN = ck_add_builtin_op("IsIn",2,mi_bool);
  op_IS_LE = ck_add_builtin_op("IsLE",2,mi_bool);
  op_IS_LT = ck_add_builtin_op("IsLT",2,mi_bool);
  ck_add_builtin_op("IsZero",1,mi_is_zero);
  op_COCOLON = ck_add_builtin_op("\040COCOLON",2,mi_cocolon);
  op_MAP = ck_add_builtin_op("Map",4,mi_map);
  op_MULT = ck_add_builtin_op("\040MULT",2,mi_mult);
  op_NEWLINE = op_map_add(pkg_builtin_n,"NewLine",0);
  op_NOT = ck_add_builtin_op("NOT",1,mi_bool);
  op_NULL_VALUE = ck_add_builtin_op("Null",0,mi_null);
  ck_add_builtin_op("NewId",NO_ARITY,mi_new_id);
  op_OPTION = ck_add_builtin_op("Option",1,engine_option);
  op_OPENLOG = ck_add_builtin_op("OpenLog",1,engine_openlog);
  op_OR = ck_add_builtin_op("OR",2,mi_bool);
  op_PACKAGE = ck_add_builtin_op("\040PACKAGE",NO_ARITY,engine_package);
  op_PARENTHESIS = ck_add_builtin_op("\040PARENTHESIS",1,mi_parenthesis);
  op_PKG_ID = ck_add_builtin_op("\040PKG_ID",1,mi_path);
  op_POLY_RING = op_map_add(pkg_builtin_n,"\040POLY_RING",NO_ARITY);
  op_PACKAGES = ck_add_builtin_op("Packages",0,mi_packages);
  op_POWER = ck_add_builtin_op("\040POW",2,mi_power);
  op_PRINT = ck_add_builtin_op("Print",NO_ARITY,engine_print);
  op_PRINTON = ck_add_builtin_op("\040PRINTON",2,engine_print);
  op_PRINTON_AND_CR = ck_add_builtin_op("\040PRINTON_AND_CR",2,engine_print);
  op_PRINT_AND_CR = ck_add_builtin_op("PrintLn",NO_ARITY,engine_print);
  op_PUT = ck_add_builtin_op("Put",2,mi_put);
  op_PROGRAM = ck_add_builtin_op("\040DEFINE",5,engine_define_func);
  op_QUIT = ck_add_builtin_op("\040QUIT",0,engine_quit);
  op_RANGE = ck_add_builtin_op("\040RANGE",2,mi_range);
  op_RECORD = ck_add_builtin_op("Record",NO_ARITY,mi_record);
  op_RECORDSEL = ck_add_builtin_op("\040RecordComp",2,mi_record_sel);
  op_REMOVE_DEVICE = ck_add_builtin_op("RemoveDevice",2,mi_remove_device);
  op_REMOVE = ck_add_builtin_op("Remove",2,mi_remove);
  op_REPEAT = ck_add_builtin_op("\040REPEAT",2,engine_repeat);
  op_RETURN = ck_add_builtin_op("Return",NO_ARITY,engine_return);
  op_RING_MODIFIERS = op_map_add(pkg_builtin_n,"\040RING_MODIFIERS",NO_ARITY);
  op_RING_ZP = op_map_add(pkg_builtin_n,"\040ZMod",NO_ARITY);
  ck_add_builtin_op("Seed",1,mi_seed);
  op_SET = ck_add_builtin_op("\040SET",NO_ARITY,engine_set_option);
  op_SET_CURRENT_RING = ck_add_builtin_op("\040USE",1,engine_set_curr_ring);
  op_SKIP = ck_add_builtin_op("\040SKIP",0,engine_skip);
  op_SOURCE = ck_add_builtin_op("Source",1,engine_source);
  op_SUBTRACT = ck_add_builtin_op("\040SUBTRACT",2,mi_subtract);
  op_SUM = ck_add_builtin_op("\040SUM",2,mi_sum);
  op_SYSTEM = ck_add_builtin_op("System", 1, mi_system);
  op_TERROR = ck_add_builtin_op("\040TERROR",NO_ARITY,engine_terror);
  op_TIME = ck_add_builtin_op("\040TIME",1,mi_time);
  op_UNLABEL = ck_add_builtin_op("Untagged",NO_ARITY,mi_unlabel);
  op_VAR = ck_add_builtin_op("Var",NO_ARITY,mi_var);
  op_WHILE = ck_add_builtin_op("\040WHILE",2,engine_while);
  op_ZPMK = ck_add_builtin_op("\040ZPMK",2,mi_zpmk);
  op_PANEL_CMD = ck_add_builtin_op("Panel",1,engine_panel);

  op_POLYNOMIAL = op_map_add(pkg_builtin_n,"\040POLY",NO_ARITY);
  op_MAT = ck_add_builtin_op("Mat",1,mi_mat);
  op_FLOAT = op_map_add(pkg_builtin_n,"\040FLOAT",1);
  op_ZPVALUE = op_map_add(pkg_builtin_n,"\040ZPVALUE",2);
  op_ZP = op_map_add(pkg_builtin_n,"\040ZP",2);
  op_SHORT_INT = op_map_add(pkg_builtin_n,"\040SHORT_INT",1);
  op_SHORT_RAT = op_map_add(pkg_builtin_n,"\040SHORT_RAT",2);
  op_SHORT_DOUBLE = op_map_add(pkg_builtin_n,"\040SHORT_DOUBLE",1);
  op_LONG_INT = op_map_add(pkg_builtin_n,"\040LONG_INT",1);
  op_LONG_RAT = op_map_add(pkg_builtin_n,"\040LONG_RAT",2);
  op_DASHES = op_map_add(pkg_builtin_n,"\040DASHES",NO_ARITY);
  op_MEM_STAT = op_map_add(pkg_builtin_n,"\040MEM_STAT",2);
#ifdef DEBUG
  op_OBJECTS = op_map_add(pkg_builtin_n,"Objects",NO_ARITY);
#endif
  op_PANELS = ck_add_builtin_op("Panels",0,mi_panels);
  op_MEMORY = ck_add_builtin_op("Memory",NO_ARITY,mi_memory);
  op_HANDLER = op_map_add(pkg_builtin_n,"\040HANDLER",3);
  op_ERROR = op_map_add(pkg_builtin_n,"\040ERROR",3);
  op_MKERR = ck_add_builtin_op("Error",NO_ARITY,mi_error);
  op_PAIR = op_map_add(pkg_builtin_n,"\040PAIR",2);
  op_GLOBAL_MEMORY = 
    ck_add_builtin_op("GlobalMemory",0,mi_memory);

  op_AT = op_map_add(pkg_builtin_n,"AT",2);

  op_objtype_NULL = ck_add_builtin_op("NULL",NO_ARITY,mi_objtype);

  op_objtype_ring = ck_add_builtin_op("RING",0,mi_objtype);
  op_objtype_poly = ck_add_builtin_op("POLY",0,mi_objtype);
  op_objtype_bool = ck_add_builtin_op("BOOL",0,mi_objtype);
  op_objtype_cterm = ck_add_builtin_op("CTERM",0,mi_objtype);
  op_objtype_ideal = ck_add_builtin_op("IDEAL",0,mi_objtype);
  op_objtype_pvector = ck_add_builtin_op("VECTOR",0,mi_objtype);
  op_objtype_string = ck_add_builtin_op("STRING",0,mi_objtype);
  op_objtype_module = ck_add_builtin_op("MODULE",0,mi_objtype);
  op_objtype_objtype = ck_add_builtin_op("TYPE",0,mi_objtype);
  op_objtype_olist = ck_add_builtin_op("LIST",0,mi_objtype);
  op_objtype_omat = ck_add_builtin_op("MAT",0,mi_objtype);
  op_objtype_ratfun = ck_add_builtin_op("RATFUN",0,mi_objtype);
  op_objtype_panel = ck_add_builtin_op("PANEL",0,mi_objtype);
  op_objtype_err = ck_add_builtin_op("ERROR",0,mi_objtype);
  op_objtype_arbint = ck_add_builtin_op("INT",0,mi_objtype);
  op_objtype_arbrat = ck_add_builtin_op("RAT",0,mi_objtype);
  op_objtype_record = ck_add_builtin_op("RECORD",0,mi_objtype);
  op_objtype_zp = ck_add_builtin_op("ZMOD",0,mi_objtype);
  op_objtype_double = ck_add_builtin_op("\040DOUBLE",0,mi_objtype);
  op_objtype_tagged = ck_add_builtin_op("TAGGED",1,mi_objtype);
  op_objtype_cdev = ck_add_builtin_op("DEVICE",0,mi_objtype);
  op_objtype_cocfun = ck_add_builtin_op("FUNCTION",0,mi_objtype);
  op_objtype_mat = ck_add_builtin_op("\040OLDMAT",0,mi_objtype);

  op_CURR_RING = ck_add_builtin_op("CurrentRing",0,mi_curr_ring);
  op_ISDEF = ck_add_builtin_op("Defined",1,mi_bool);
  op_INDET_NTH = ck_add_builtin_op("Indet",1,mi_indet_nth);
  op_LEX = op_map_add(pkg_builtin_n,"Lex",0);
  op_XEL = op_map_add(pkg_builtin_n,"Xel",0);
  op_DEGLEX = op_map_add(pkg_builtin_n,"DegLex",0);
  op_DEGREVLEX = op_map_add(pkg_builtin_n,"DegRevLex",0);
  op_POS_TO = op_map_add(pkg_builtin_n,"PosTo",0);
  op_TO_POS = op_map_add(pkg_builtin_n,"ToPos",0);
  op_LEVEL= op_map_add(pkg_builtin_n,"Level",1);
  op_ORD = ck_add_builtin_op("Ord",NO_ARITY,mi_ord);
  op_WEIGHTS = op_map_add(pkg_builtin_n,"Weights",NO_ARITY);
  op_ENV = ck_add_builtin_op("RingEnv",NO_ARITY,mi_env);
  ck_add_builtin_op("NoMoreThanOneRing",NO_ARITY,mi_is_ring_hom);
  op_ENV_SLOT = op_map_add(pkg_builtin_n,"\040ENV_SLOT",2);
  op_MODULE = ck_add_builtin_op("Module",NO_ARITY,mi_module);
  op_MONOMIAL = op_map_add(pkg_builtin_n,"\040MONOMIAL",NO_ARITY);
  op_OBJECT = op_map_add(pkg_builtin_n,"\040OBJECT",2);
  op_VECTOR = ck_add_builtin_op("Vector",NO_ARITY,mi_vector);
  op_RATFUN = op_map_add(pkg_builtin_n,"\040RATFUN",2); /* used for pp with parenthesis */
  op_TERM = op_map_add(pkg_builtin_n,"\040TERM",NO_ARITY);

  op_RING = ck_add_builtin_op("Ring",1,mi_ring);
  op_LIMITS = ck_add_builtin_op("CocoaLimits",0,mi_limits);
  op_LEADING_TERM = ck_add_builtin_op("LT",1,mi_leading_term);
  op_LEADING_COEFF = ck_add_builtin_op("LC",1,mi_leading_coeff);
  op_DIV = ck_add_builtin_op("Div",2,mi_div);
  op_POINCARE = ck_add_builtin_op("UVPoincareQuotient",1,mi_poincare_quotient);
  ck_add_builtin_op("IsPureLT",1,mi_ispurelt);
  op_DEG = ck_add_builtin_op("Deg",1,mi_deg);
  op_MDEG = ck_add_builtin_op("MDeg",1,mi_mdeg);
  op_SIZE = ck_add_builtin_op("Size",1,mi_size);
  op_MOD = ck_add_builtin_op("Mod",2,mi_mod);
  op_RANDOMIZE = ck_add_builtin_op("Randomized",1,mi_randomized);
  op_DENSE_POLY = ck_add_builtin_op("DensePoly",NO_ARITY,mi_dense_poly);
  op_INDET_NAME = ck_add_builtin_op("IndetName",1,mi_indet_name);
  op_INDET_IND = ck_add_builtin_op("IndetInd",1,mi_indet_ind);
  op_INDET_INDEX = ck_add_builtin_op("IndetIndex",1,mi_indet_index);
  op_SHOW_ALIASES = ck_add_builtin_op("Aliases",NO_ARITY,mi_aliases);
  op_GB_INIT = ck_add_builtin_op("GB_Init",1,mi_gb_init);
  op_GB_ONE_MOVE = ck_add_builtin_op("GB_OneMove",1,mi_gb_one_move);
  op_NOINDENT = op_map_add(pkg_builtin_n,"\040NOINDENT",1);
  op_GB_STAT = ck_add_builtin_op("GB_Stat",1,mi_gb_stat);
  op_GB_ALL = ck_add_builtin_op("GB_DoAll",1,mi_gb_all);
  op_TYPEOF = ck_add_builtin_op("Type",1,mi_typeof);
  op_MVPOINCARE = ck_add_builtin_op("MVPoincare",NO_ARITY,mi_mvpoincare);
  op_HEAD = ck_add_builtin_op("Head",1,mi_head);
  op_TAIL = ck_add_builtin_op("Tail",1,mi_tail);
  op_GENS = ck_add_builtin_op("\040Gens",1,mi_gens);
  op_LEN = ck_add_builtin_op("Len",1,mi_len);
  op_DER = ck_add_builtin_op("Der",NO_ARITY,mi_der);
  op_HOMOG = ck_add_builtin_op("Homogenized",NO_ARITY,mi_homog);
  op_DET = ck_add_builtin_op("Det",1,mi_det);
  op_ADJOINT = ck_add_builtin_op("Adjoint",1,mi_adjoint);
  op_MINORS = ck_add_builtin_op("Minors",2,mi_minors);
  op_SUBST = ck_add_builtin_op("SubstPoly",2,mi_subst);
  op_DATE = ck_add_builtin_op("Date",NO_ARITY,mi_date);
  op_CHAR = ck_add_builtin_op("Characteristic",NO_ARITY,mi_characteristic);
  op_LABEL = ck_add_builtin_op("Tagged",NO_ARITY,mi_label);
  op_POLY = ck_add_builtin_op("Poly",NO_ARITY,mi_poly);
  op_CONCAT = ck_add_builtin_op("Concat",NO_ARITY,mi_concat);
  op_APPEND = ck_add_builtin_op("Append",2,mi_append);
  op_LISTCAST = ck_add_builtin_op("List",1,mi_olistcast);
  op_NUM_INDETS = ck_add_builtin_op("NumIndets",NO_ARITY,mi_num_indets);
  op_CAST = ck_add_builtin_op("Cast",2,mi_cast);
  op_CIP = ck_add_builtin_op("ComputationInProgress",1,mi_cip);
  op_FACT = ck_add_builtin_op("Fact",1,mi_fact);
  ck_add_builtin_op("BinToric",2,mi_bin_toric);
  ck_add_builtin_op("MatToric",1,mi_mat_toric);
  ck_add_builtin_op("MatSequentialToric",1,mi_mat_seqtoric);
  ck_add_builtin_op("BinSequentialToric",2,mi_bin_seqtoric);
  ck_add_builtin_op("MatParallelToric",1,mi_mat_partoric);
  ck_add_builtin_op("BinParallelToric",2,mi_bin_partoric);
  ck_add_builtin_op("LinSol",2,mi_lin_sol);
  ck_add_builtin_op("LinKer",1,mi_lin_ker);
  op_TAO = ck_add_builtin_op("Tao",0,mi_tao);
  op_STARTING = ck_add_builtin_op("Starting",NO_ARITY,mi_starting);
  op_FACTOR = ck_add_builtin_op("Factor",1,mi_factor);
  op_GCD = ck_add_builtin_op("GCD",2,mi_gcd);
  op_NEWPPSTREAM = ck_add_builtin_op("NewPPStream",NO_ARITY,mi_NewPPStream);
  op_NEXTPP = ck_add_builtin_op("NextPP",1,mi_NextPP);
  op_AVOIDLASTPP = ck_add_builtin_op("AvoidLastPP",1,mi_AvoidLastPP);
  op_NEXTPRIME = ck_add_builtin_op("NextPrime",1,mi_NextPrime);
  op_MOD2RAT = ck_add_builtin_op("Mod2Rat",3,mi_Mod2Rat);
  ck_add_builtin_op("BuchbergerMoeller",2,mi_bm);
  ck_add_builtin_op("GBM",1,mi_gbm);
  ck_add_builtin_op("Isqrt",1,mi_isqrt);
  ck_add_builtin_op("Iroot",2,mi_iroot);
  ck_add_builtin_op("QZP",1,mi_QZP);
  ck_add_builtin_op("ZPQ",1,mi_ZPQ);
  ck_add_builtin_op("IsPPrime",1,mi_IsPPrime);
  ck_add_builtin_op("NextPPrime",1,mi_NextPPrime);
  ck_add_builtin_op("PowerMod",3,mi_PowerMod);
  op_LOG = ck_add_builtin_op("Log",1,mi_log);
  op_TAG = ck_add_builtin_op("Tag",1,mi_tag);
  op_NEWLIST = ck_add_builtin_op("NewList",NO_ARITY,mi_new_list);
  op_NEWVECTOR = ck_add_builtin_op("NewVector",NO_ARITY,mi_new_vector);
  op_NEWMAT = ck_add_builtin_op("NewMat",NO_ARITY,mi_new_mat);
#ifdef RUM_DEBUG
  op_RUM_STAT = ck_add_builtin_op("RumStat",NO_ARITY,mi_rum_stat);
#endif
  op_INFOS = ck_add_builtin_op("CocoaInfos",0,mi_infos);
  op_PROGRAM2 = op_map_add(pkg_builtin_n,"\040Program2",5);

  op_SUPPORT = ck_add_builtin_op("Support", 1, mi_support);
  op_MONOMIALS = ck_add_builtin_op("Monomials", 1, mi_monomials);
  op_COEFFS = ck_add_builtin_op("Coefficients", 1, mi_coeffs);
  
  op_ISPOSTO = ck_add_builtin_op("IsPosTo", 1, mi_isposto);
  op_ISTOPOS = ck_add_builtin_op("IsToPos", 1, mi_istopos);
  op_FIRSTNONZERO = ck_add_builtin_op("FirstNonZero", 1, mi_firstnonzero);
  op_FIRSTNONZEROPOS = ck_add_builtin_op("FirstNonZeroPos", 1, mi_firstnonzeropos);
  op_ISTERM = ck_add_builtin_op("IsTerm", 1, mi_isterm);
  op_LPOS = ck_add_builtin_op("LPos", 1, mi_lpos);
  op_LPP = ck_add_builtin_op("LPP", 1, mi_lpp);
  op_LM = ck_add_builtin_op("LM", 1, mi_lm);
  op_COEFFOFTERM = ck_add_builtin_op("CoeffOfTerm", 2, mi_coeffofterm);
  op_NR = ck_add_builtin_op("NR", 2, mi_nr);

}

/* ---[ option panels ]-------------------------------------------- */

opt_panel opt_panel_DEBUG;
opt_id flag_DEBUG_MEM,flag_DEBUG_OBJ,flag_DEBUG_MEMFILE, 
  opt_DEBUG_EEE_PP, flag_SHOW_TODO, flag_SHOW_OPTIMIZE, flag_DEBUG_GB,
  flag_DEBUG_FASTSOURCE, flag_SHOW_OBJNUM;
opt_value optv_DEBUG_EEE_PP_GRAPH1, optv_DEBUG_EEE_PP_GRAPH2, 
  optv_DEBUG_EEE_PP_NORMAL;

opt_panel opt_panel_MONITOR;
opt_id flag_MON_GB_STATUS;

#ifdef DEBUG
opt_id flag_MON_GB_SPOLY, flag_MON_GB_REDUNDANT_POLY,
  flag_MON_GB_KILLED_PAIRS,flag_MON_GB_SKIPPED_PAIRS, flag_MON_GBASIS;
#endif

opt_panel opt_panel_ENGINE;
opt_id flag_ENGINE_ECHO, 
  flag_ENGINE_TRACE, flag_ENGINE_TIMER, flag_ENGINE_MEMORY,
  flag_ENGINE_INDENT, flag_ENGINE_TRACE_SOURCES, flag_ENGINE_WARNINGS,
  flag_ENGINE_COMP_STACK, flag_ENGINE_NO_ERROR_CONTEXT;

#ifdef PRETTY_EXPONENTS
opt_id flag_ENGINE_PP_EXP;
#endif


opt_panel opt_panel_GROBNER;

opt_id flag_GB_SUGAR, flag_GB_RED_FULL, flag_GB_STEP, flag_GB_VERBOSE;

opt_panel opt_panel_HDRIVEN;
opt_id opt_HD_INS_TYPE;
opt_value optv_HD_INS_TYPE_AT_END, optv_HD_INS_TYPE_BY_LEN, 
  optv_HD_INS_TYPE_BY_ECART;

opt_id flag_HD_FINAL_MONIC;

opt_id opt_HD_RED_TYPE;
opt_value optv_HD_RED_TYPE_FULL, optv_HD_RED_TYPE_LT;

opt_id opt_HD_I_RED_TYPE;
opt_value optv_HD_I_RED_TYPE_FULL, optv_HD_I_RED_TYPE_LT;

opt_id opt_HD_F_RED_TYPE;
opt_value optv_HD_F_RED_TYPE_NONE, optv_HD_F_RED_TYPE_FULL, 
  optv_HD_F_RED_TYPE_LT;


opt_panel opt_panel_IMATH;

opt_id opt_IM_INS_TYPE;
opt_value optv_IM_INS_TYPE_AT_END, optv_IM_INS_TYPE_BY_LEN, 
  optv_IM_INS_TYPE_BY_ECART;

opt_id flag_IM_SUGAR;

opt_panel opt_panel_INTERNAL;
opt_id flag_ENGINE_EXIT;

void opt_init_debug(void)
{
  opt_panel_DEBUG=opt_panel_new("DEBUG");
  flag_DEBUG_MEM=opt_new_flag(opt_panel_DEBUG,"ShowMemoryUsage");
  flag_DEBUG_OBJ=opt_new_flag(opt_panel_DEBUG,"ShowObjects");
  flag_DEBUG_MEMFILE=opt_new_flag(opt_panel_DEBUG,"MemoryUsageOnFile");
  flag_SHOW_TODO=opt_new_flag(opt_panel_DEBUG,"ShowTodo");
  flag_set(opt_panel_DEBUG,flag_SHOW_TODO);
  flag_SHOW_OPTIMIZE=opt_new_flag(opt_panel_DEBUG,"ShowOptimize");
  flag_DEBUG_GB=opt_new_flag(opt_panel_DEBUG,"DebugGB");
  flag_DEBUG_FASTSOURCE=opt_new_flag(opt_panel_DEBUG,"FastSource");
  flag_SHOW_OBJNUM=opt_new_flag(opt_panel_DEBUG,"ShowObjectNumber");
  opt_DEBUG_EEE_PP=
    opt_new_enum(opt_panel_DEBUG,"EEE_PP");
  optv_DEBUG_EEE_PP_NORMAL=
    optv_new(opt_panel_DEBUG,opt_DEBUG_EEE_PP,"Normal");
  optv_DEBUG_EEE_PP_GRAPH1=
    optv_new(opt_panel_DEBUG,opt_DEBUG_EEE_PP,"GraphicWithLabels");
  optv_DEBUG_EEE_PP_GRAPH2=
    optv_new(opt_panel_DEBUG,opt_DEBUG_EEE_PP,"GraphicWithEmptyLabels");
  opt_set_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP,optv_DEBUG_EEE_PP_NORMAL);
}

#ifdef DEBUG
void opt_init_monitor(void)
{
  opt_panel_MONITOR=opt_panel_new("MONITOR (ONLY DEBUG MODE, OBSOLETE)");
  flag_MON_GB_STATUS=opt_new_flag(opt_panel_MONITOR,"Show GB status");

  flag_MON_GB_SPOLY=opt_new_flag(opt_panel_MONITOR,"Show S-polys");
  flag_MON_GB_REDUNDANT_POLY=
    opt_new_flag(opt_panel_MONITOR,"Show redundant polys");
  flag_MON_GB_KILLED_PAIRS=
    opt_new_flag(opt_panel_MONITOR,"Show killed pairs");
  flag_MON_GB_SKIPPED_PAIRS=
    opt_new_flag(opt_panel_MONITOR,"Show skipped pairs");
  flag_MON_GBASIS=
    opt_new_flag(opt_panel_MONITOR,"Show partially computed GBasis");
}
#endif

void opt_init_engine(void)
{
  opt_panel_ENGINE=opt_panel_new("GENERAL");
  flag_ENGINE_ECHO=opt_new_flag(opt_panel_ENGINE,"Echo");
  flag_ENGINE_TIMER=opt_new_flag(opt_panel_ENGINE,"Timer");
#ifdef DEBUG
  flag_ENGINE_MEMORY=opt_new_flag(opt_panel_ENGINE,"Memory");
#endif
  flag_ENGINE_TRACE=opt_new_flag(opt_panel_ENGINE,"Trace");
  flag_ENGINE_INDENT=opt_new_flag(opt_panel_ENGINE,"Indentation");
  flag_ENGINE_TRACE_SOURCES=opt_new_flag(opt_panel_ENGINE,"TraceSources");
  flag_ENGINE_WARNINGS=opt_new_flag(opt_panel_ENGINE,"SuppressWarnings");
  flag_ENGINE_COMP_STACK=opt_new_flag(opt_panel_ENGINE,"ComputationStack");
  flag_ENGINE_NO_ERROR_CONTEXT=opt_new_flag(opt_panel_ENGINE,"NoErrorContext");
#ifdef PRETTY_EXPONENTS
  flag_ENGINE_PP_EXP=opt_new_flag(opt_panel_ENGINE,"PrettyExponents");
#endif
}

void opt_init_gb(void)
{
  opt_panel_GROBNER=opt_panel_new("GROEBNER");

  flag_GB_SUGAR = opt_new_flag(opt_panel_GROBNER,"Sugar");
  flag_GB_RED_FULL = opt_new_flag(opt_panel_GROBNER,"FullRed");
  flag_GB_STEP = opt_new_flag(opt_panel_GROBNER,"SingleStepRed");
  flag_GB_VERBOSE = opt_new_flag(opt_panel_GROBNER,"Verbose");

  opt_set_value(opt_panel_GROBNER,flag_GB_SUGAR,flag_ON);
  opt_set_value(opt_panel_GROBNER,flag_GB_RED_FULL,flag_ON);
  opt_set_value(opt_panel_GROBNER,flag_GB_STEP,flag_OFF);
  opt_set_value(opt_panel_GROBNER,flag_GB_VERBOSE,flag_OFF);
}

void opt_init_internal(void)
{
  opt_panel_INTERNAL=opt_panel_new("Internal Flags");
  flag_ENGINE_EXIT=opt_new_flag(opt_panel_INTERNAL,"Exit Cocoa");
}

void ck_opt_init(void)
{
  options_init();
#ifdef DEBUG
  opt_init_debug();
  opt_init_monitor();
#endif
  opt_init_engine();
  opt_init_gb();
  opt_init_internal();
}

void ck_rum_init(void)
{
  int i;
  rum_init_all();
  for (i=1;i<=3;i++) rum_init(i*4,3*RUM_STD_SIZE);
  rum_init(16,6*RUM_STD_SIZE);
  for (i=5;i<=5;i++) rum_init(i*4,3*RUM_STD_SIZE);
  for (i=6;i<=10;i++) rum_init(i*4,RUM_STD_SIZE);
}
