/* ------------------------------------------------------------------------ *\

   Convention: given the macro MACRO, the definition of the
   macro _MACRO (note the underscore) is a fast way to undefine MACRO

\* ------------------------------------------------------------------------ */

#ifndef COCDEFS_H
#define COCDEFS_H

#define _WITH_SYSTEM_CMD
#define COCOA_WITH_RUM
#define _COCOA_SHOW_MINOR_VERSION
#define _COCOA_DEVELOPING
#define _DEBUG
#define _DEBUG_MALLOC
/* Note: turn off COCOA_WITH_RUM if you want to use DEBUG_MALLOC     */
/* DEBUG_MALLOC prints out a message for every memory (de)allocation */
/* It is of limited use, as CoCoA never frees memory held in globals */

/* Port number used by the CoCoA command OpenSocket */
#define COCOA_PORT 49344

#ifndef WITH_SYSTEM_CMD
#define COCOA_MAJOR_VERSION() "4.4 PRE-RELEASE"
#define COCOA_MINOR_VERSION() "a"
#else
#define COCOA_MAJOR_VERSION() "4.4 PRE-RELEASE **SPECIAL**"
#define COCOA_MINOR_VERSION() "a"
#endif

#define IS_DOMAIN
#define COCOA



/* ------[ developing ]------ */

#ifdef COCOA_DEVELOPING
#define COCOA_HOOK
#endif

/* ------[ debugging ]------ */

#define GB_STATS

#ifdef DEBUG

#define MEMORY_CLEAR_WHEN_FREE
#define OBJECT_DEBUG
#define _RUM_DEBUG
#define _RUM_DEBUG_SHOW_UNUSED

#define DEBUG_CHECK_NIL_FREE
#define DEBUG_RING_RC
#define _DEBUG_CDEV_RC
#define _DEBUG_SHIFTS_RC
#define _DEBUG_MAP
#define _DEBUG_RUM
#define _DEBUG_RESOLVING
#define _DEBUG_GB
#define _DEBUG_IVECTORS
#define _DEBUG_GB_REDUCE

#endif

#define _COCOA_MONITORING
#define _HDRIVEN_STAT

/* ------[ polynomials ]------ */

/* polynomials used for arithmetics */
#define ARPOLY_MONOPOLI
#define _ARPOLY_TRIPOLI
#define _ARPOLY_EEE

/* components of vectors */
#define PVCOMP_MONOPOLI
#define _PVCOMP_TRIPOLI
#define _PVCOMP_EEE

#define MONOPOLI

#if defined(PVCOMP_TRIPOLI)
#  define TRIPOLI
#  define PVCOMP_VERSION "TRIPOLI"
#  define PVCOMP_SHORT_VERSION "T"
#elif defined(PVCOMP_EEE)
#  define EEE
#  define PVCOMP_VERSION "EEE"
#  define PVCOMP_SHORT_VERSION "E"
#elif defined(PVCOMP_MONOPOLI)
#  define PVCOMP_VERSION "MONOPOLI"
#  define PVCOMP_SHORT_VERSION "M"
#else
#  define PVCOMP_VERSION "unknown"
#  define PVCOMP_SHORT_VERSION "?"
#endif

#if defined(ARPOLY_TRIPOLI) 
#  define TRIPOLI 
#  define ARPOLY_VERSION "TRIPOLI"
#  define ARPOLY_SHORT_VERSION "T"
#elif defined(ARPOLY_EEE)
#  define EEE
#  define ARPOLY_VERSION "EEE"
#  define ARPOLY_SHORT_VERSION "E"
#elif defined(ARPOLY_MONOPOLI)
#  define ARPOLY_VERSION "MONOPOLI"
#  define ARPOLY_SHORT_VERSION "M"
#else
#  define ARPOLY_VERSION "unknown"
#  define ARPOLY_SHORT_VERSION "?"
#endif

#define POLY_VERSION "[" ARPOLY_VERSION "," PVCOMP_VERSION "]"
#define POLY_SHORT_VERSION ARPOLY_SHORT_VERSION PVCOMP_SHORT_VERSION

#ifdef COCOA_SHOW_MINOR_VERSION
#define COCOA_VERSION() COCOA_MAJOR_VERSION() "." COCOA_MINOR_VERSION()
#else
#define COCOA_VERSION() COCOA_MAJOR_VERSION()
#endif

/* ------[ platform dependent ]------ */

/* ------[ derived macros (do not edit) ]------ */

#ifdef CA_MAC
# define PRETTY_EXPONENTS
#endif

#ifdef SIS_MAC
# define SIS_TYPE "MAC"
#elif defined(SIS_UNIX)
# define SIS_TYPE "UNIX"
#elif defined(SIS_DOS)
# define SIS_TYPE "DOS"
#elif defined(SIS_WIN32)
# define SIS_TYPE "SISWIN32"
#endif

#ifdef EEE
# define EEE_KEEP_OPT
#endif

#if (defined(EEE) || defined(TRIPOLI))
# define SCALARS
# define TRIPOLI_OR_EEE
#endif

#if (defined(PVCOMP_EEE) || defined(PVCOMP_TRIPOLI))
# define SCALARS
# define PVCOMP_TRIPOLI_OR_EEE
#endif

#if (defined(ARPOLY_EEE) || defined(ARPOLY_TRIPOLI))
# define SCALARS
# define ARPOLY_TRIPOLI_OR_EEE
#endif

/* ------[ Misc ]---------- */

#define COCOA_DEFAULT_ZMOD 32003
/* new criterions for pairs */
#define GB_NEWCRIT

#endif
