/* $Id: coctypes.h,v 4.5 2004/06/28 13:14:01 cocoa Exp $

[ coeff ] [ cocfun ] [ cord ] [ cdev ] [ eterms ] [ evector ] [ indet ] 
[ misc ] [ monomial ] [ mlist ] [ module ] [ memory ] [ mvector ]
[ monopoli ] [ object ][ ordering ] [ olist ] [ omatrix ]
[ package ] [ poly ] [ pseries ] [ pvector ] [ pair ] 
[ ring ] [ ratfun ] [ scalars ] [ shifts ] [ subst ] [ terms ] 
[ tlist ] [ tripoli ] [ unipoly ] [ vpair ] [ vlist ] [ weights ] 

------------------------------------------------------------------------------*/

#ifndef COCTYPES_H
#define COCTYPES_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#define MALLOC malloc
#define FREE free

/* ====================================== */

/* mutually RECURSIVE types */

#ifndef AUXTYPES_H
typedef struct ring_el * ring;
typedef struct object_info * object;
#endif

typedef struct cdev_info * cdev;

typedef object * loc;
typedef object * olist;
#define omat olist
typedef struct module_info *module;

/* ====================================== */


/* -----------[ cdev ]-------------- */

typedef void * cdev_params;

#if defined(CA_TEXT) || defined(CA_QT)

/* W A R N I N G !  Making CA_TEXT and CA_QT distinct could lead to      */
/*                  problems.  ../UI/makefile assumes they are the same. */
typedef FILE * cdev_ifile;
typedef FILE * cdev_ofile;
typedef void * cdev_ofile_params;
typedef void * cdev_iwindow;
typedef void * cdev_owindow;
typedef void * cdev_owindow_params;
typedef struct sock_t * cdev_socket; /* TODO: Mac and Win32 ??? */

#elif defined(CA_MAC)

#include "macHdrs.h"

typedef void * cdev_ifile;
typedef void * cdev_ofile;
typedef WindowPtr cdev_owindow_params;

typedef struct cdev_iwindow_info {
  char * buffer;
  int pos;
} cdev_iwindow_info;

typedef struct cdev_owindow_info {
  WindowPtr targetWindow;
  char * buffer;
  int pos;
} cdev_owindow_info;

typedef cdev_iwindow_info * cdev_iwindow;
typedef cdev_owindow_info * cdev_owindow;

#elif defined(CA_WIN32)

#include "cdev_w32.h"

#else

#error "Undefined Interface" 

#endif


typedef struct cdev_istring_info * cdev_istring;
typedef charbuf cdev_ostring;

typedef olist cdev_omulti;

typedef struct cdev_istring_info{
  char * buffer;
  int pos;
} cdev_istring_info;

/* typedef struct cdev_info * cdev; */ /* see coctypes.h */

typedef union {
    cdev_ifile ifile;
    cdev_ofile ofile;
    cdev_istring is;
    cdev_ostring os;
    cdev_iwindow iw;
    cdev_owindow ow;
    cdev_omulti om;
    cdev_socket socket;
  } cdevice;

typedef struct cdev_info{
  int rc;
  int id; /* unique identifier generated by cdev_init */
  char * name;
  int t;
  int prot;
  cdevice d;
  int curr_line;
} cdev_info;

/* -----------[ cocfun ]-------------- */

typedef struct cocfun_info
{  
  char * pkg_id;
  char * fun_id;
  cterm header;
  cterm body;
  cterm help;
  int arity;
  int rc;
  coc_bool is_short; /* defined in the short form F(X1,...,XN) := E */
} cocfun_info;

typedef cocfun_info * cocfun;


/* -----------[ coeff ]-------------- */

#ifndef __GMP_H__
# include "gmp.h"
#endif

typedef MP_INT * verylongint;
typedef MP_RAT * verylongrat;

typedef struct aux_rat
{  
  int num;
  int den;
} aux_rat;

typedef aux_rat * rat;

typedef struct aux_zmod
{  
  int p;
  int n; /* n modulo p */
} aux_zmod;

typedef aux_zmod * zmod;

#define zmod_NULL NULL
#define double_NULL ZERO

typedef void * coeff;

#define cal(pc,c) *pc=(coeff)(c)

#define arbint coeff
#define arbrat coeff

#define arbint_NULL NULL
#define arbrat_NULL NULL

#define coeff_NULL NULL

#define coeff_zp unsigned int

#if 0

typedef union {
  verylongint Li;  /* intero lungo */
  verylongrat Lr;  /* razionale lungo */
  unsigned int  Mi;  /* intero modulo p */
  int Si;  /* intero macchina  */
  rat * Sr;  /* razionale corto  */
  double * Sd;     /* double corto */ 
} coeff;

#endif

#define characteristic_Li 1
#define characteristic_Lr 0
#define characteristic_Si -1
#define characteristic_Sr -2
#define characteristic_Sd -3
#define characteristic_ERROR -4
/* #define characteristic_L_Zp -5 (see p in R->characteristic_L) */

/* characteristic > 1 => module p */

/* -----------[ indet ]-------------- */

typedef struct indet_info_aux * indet_info;

typedef short int unknown;
#define unknown_NULL -1

#define MAX_NUM_INDET_ENTRY 26

typedef struct indet_entry_aux * indet_entry;

typedef struct indet_entry_aux {
  unknown offset;
  unknown size;
  char * ident;
  ivector arrinfo;
  ivector from;
} indet_entry_aux;

typedef struct indet_info_aux {
  short int num_entries;
  short int num_indets;
  indet_entry indets[MAX_NUM_INDET_ENTRY];
} indet_info_aux;

#define indet_info_NULL NULL

/* -----------[ terms ]-------------- */

typedef ivector term;
typedef iv_elem expType;

#define term_NULL iv_NULL

/* -----------[ scalars ]-------------- */

#define scalar iv_elem
#define scalars ivector

/* -----------[ weights ]-------------- */

#define weight ivector
typedef weight * weilist;

#define weight_NULL iv_NULL
#define weilist_NULL NULL



/* -----------[ ints ]-------------- */

typedef int * ints;

/* -----------[ generic arrays ]-------------- */

typedef void ** flexarray;

/* -----------[ eterms ]-------------- */

typedef  int  * eterm;

/* -----------[ unipoly ]-------------- */

typedef union
{
  int i;
  mpz_t z;
} UPCoeff;

typedef UPCoeff * unipoly;

/* ====================================== */

/* -----------[ monomial ]-------------- */

typedef struct monomial_aux * monomial;

typedef struct monomial_aux {
  coeff c;
  term t; /* here exponents */
  monomial next; /* used only by monopoli */
} monomial_info;

/* -----------[ tlist ]-------------- */

#define tlist coc_pipe

/* ====================================== */

/* -----------[ mlist ]-------------- */

#define mlist coc_pipe

/* -----------[ monopoli ]-------------- */

#define monopoli monomial

/* -----------[ tripoli ]-------------- */

typedef struct mon_aux * mon;
typedef struct trnode_aux * trnode;
typedef void * tripoli;
#if 0
typedef union { trnode tp; /* or */ mon m; } tripoli;
#endif

typedef struct trnode_aux {
  scalar sp; /* scalar product of this bucket */
  tripoli p;
  trnode next; /* next bucket */
} trnode_aux;

typedef struct mon_aux {
  coeff c;
  scalars ss; /* scalar products only from j+1 */
  mon next;
} mon_aux;

/* -----------[ eee ]-------------- */

typedef expType * eee_lbl;

typedef struct eee_mon_aux * eee;

typedef struct eee_mon_aux * eee_mon;

typedef struct eee_mon_aux {
  expType hd;
  eee_lbl tl;
  eee c;
  eee_mon next;
} eee_mon_aux;

#define eee_NULL NULL
#define eee_mon_NULL NULL

/* -----------[ poly ]-------------- */

/* -----------[ evcomp ]-------------- */

#ifdef PVCOMP_MONOPOLI
# define pvcomp monopoli
#elif defined(PVCOMP_TRIPOLI)
# define pvcomp tripoli
#elif defined(PVCOMP_EEE)
# define pvcomp eee
#endif


#ifdef ARPOLY_MONOPOLI
# define arpoly monopoli
#elif defined(ARPOLY_TRIPOLI)
# define arpoly tripoli
#elif defined(ARPOLY_EEE)
# define arpoly eee
#endif


/* ====================================== */

/* -----------[ pseries ]-------------- */

typedef struct PSeries_info *PSeries;
typedef struct MVPSeries_info *MVPSeries;

#define pseries_NULL NULL
#define mvpseries_NULL NULL

typedef struct PSeries_info {
  unipoly  PolyNum;
  olist    denom_factors;
} PSeries_info;

typedef struct MVPSeries_info {
  arpoly  PolyNum;
  olist    denom_factors;
  ring MVPRing; 
} MVPSeries_info;

/* -----------[ shifts ]-------------- */

typedef term shift;

typedef struct shifts_aux {
  int r; /* if r < 0 it means that shifts are numbers */
  shift * ss;
  int rc;
} shifts_aux;

typedef shifts_aux * shifts;

#define shifts_NULL NULL

/* -----------[ pvector ]-------------- */

typedef struct pv_aux * pvector;

typedef struct pv_aux {
  int r;
  shifts ss;
  pvcomp * compo;
} pv_aux;

#define pv_NULL NULL

/* -----------[ mvector ]-------------- */

typedef struct mv_aux * mvector;

typedef struct mv_aux {
  int r;
  monomial * compo; /* array of inverted monopoli */
} mv_aux;

#define mv_NULL NULL

/* -----------[ ratfun ]-------------- */

typedef struct ratfun_info *ratfun;

typedef struct ratfun_info {
  arpoly num;
  arpoly den;
} ratfun_info;

#define ratfun_NULL NULL

/* ====================================== */

#define gb_NUM_STATS 10

#define gb_STATS_H_KILLED_PAIRS 0 /* Hilbert */
#define gb_STATS_CREATED_PAIRS 1
#define gb_STATS_USEFUL_PAIRS 2
#define gb_STATS_ZERO_PAIRS 3
#define gb_STATS_C_KILLED_PAIRS 4 /* criterions */

#define gb_STATS_H_KILLED_VECTORS 5
#define gb_STATS_CREATED_VECTORS 6
#define gb_STATS_USEFUL_VECTORS 7
#define gb_STATS_ZERO_VECTORS 8
#define gb_STATS_BACK_KILLED 9

/* -----------[ evector ]-------------- */

typedef struct ev_aux * evector;

typedef struct ev_aux {
  int rc;
  mvector irr;
  pvector red;
  pvector tail;
  int wD, tD;
  monomial lm;
  eterm et;
  int lmpos;
  int label;
  coc_bool Is_SPoly;
  pvector rel; /* used for syzygies */
} ev_info;

/* -----------[ vpair ]-------------- */

typedef struct vpairEl {
  evector VL1, VL2;
  term lcm; /* lcm(lt(VL1),lt(VL2)) */
#ifdef GB_NEWCRIT
  term A; /* lcm/lt(VL1) */
  term B; /* lcm/lt(VL2) */
#endif
  int wD, tD;
  coc_bool coprimes;
  struct vpairEl * next;
} vpairEl, *vpair;

#define vpair_NULL NULL
#define vpairs_NULL NULL

/* -----------[ vlist ]-------------- */

typedef struct vlist_header * vlist;

typedef struct vlist_atom_aux * vlist_atom;

typedef struct vlist_header {
  int len;
  vlist_atom first;
  vlist_atom last;
} vlist_header;

typedef struct vlist_atom_aux {
  evector ev;
  coc_bool marked;
  vlist_atom next;
} vlist_atom_info;

/* ====================================== */

/* -----------[ module ]-------------- */

#define gb_NUM_MODULE_KEY_FUNCTIONS 15

/*
#define gb_NUM_STATS 6

#define gb_STATS_KILLED_PAIRS 0
#define gb_STATS_CREATED_PAIRS 1
#define gb_STATS_USEFUL_PAIRS 2
#define gb_STATS_KILLED_VECTORS 3
#define gb_STATS_CREATED_VECTORS 4
#define gb_STATS_USEFUL_VECTORS 5
*/

#define module_FLAG_IS_HOMOG 1
#define module_FLAG_SYZ_OF_MINGENS 2
#define module_FLAG_JUST_LTNF 3
/* #define module_FLAG_NEXT 4 */

#define module_NULL NULL

typedef struct gb_opt {
  int goal;
  int alg;
  coc_bool initialized;
  int deg_trunc;
  int syz_trunc;
  int res_trunc;
  int reg_trunc;
  int elim_indets;
  int pair_order;
  int rules_order;
  int discrepancy;
  int stats[gb_NUM_STATS];
} gb_opt;

typedef struct module_info {
  int r;
  int rc;
  ring R;
  shifts ss;
  vlist gens;
  vlist mingens;
  vlist gbasis;
  vlist rules;
  vlist incoming_vectors;
  vpair incoming_pairs;
  evector current_vector;
  module syz_of_gens;
  module syz;
  module zys;
  PSeries pser;
  MVPSeries mvpser;
  gb_opt gb;
  int module_flags;

  int KF[gb_NUM_MODULE_KEY_FUNCTIONS];

} module_info;

/* ====================================== */

typedef struct env_slot_info * env; /* NULL if empty */

/* -----------[ ordering ]-------------- */

typedef short int toElem;

typedef toElem     * ord_row;
typedef ord_row    * ordering;

#define to_NULL NULL
#define ordering_NULL NULL

#define lex_num 1
#define xel_num 2
#define dlex_num 3
#define drev_num 4
#define elim_num 5
#define user_defined_ordering_num 6

/*	un ordinamento e' rappresentato da una matrice di interi N x N+1, 
	dove N e' il numero delle variabili. Se l'ultima componente della 
	riga i  e' j >= 0, allora i primi N elementi della riga sono tutti
	nulli tranne l'elemento j-esimo. 
*/

#define TO_POS 1
#define POS_TO 2

/* -----------[ cord ]-------------- */

#define cord_Lex 1
#define cord_Xel 2
#define cord_DegLex 3
#define cord_DegRevLex 4
#define cord_Elim 5
/* user defined */
#define cord_Ord 6

typedef struct cord_info *cord;

typedef struct cord_info {
  int t; /* type */
  unknown x, y;
  ordering o;
} cord_info;

/* -----------[ object ]-------------- */

typedef enum objtype {
  objtype_NULL,
  objtype_objtype,
  objtype_coeff,
  objtype_ring,
  objtype_poly,
  objtype_bool,
  objtype_cterm,
  objtype_int,
  objtype_pseries,
  objtype_pvector,
  objtype_vlist,
  objtype_string,
  objtype_ideal,
  objtype_module,
  objtype_panel,
  objtype_olist,
  objtype_omat,
  objtype_ratfun,
  objtype_arbint,
  objtype_arbrat,
  objtype_err,
  objtype_record,
  objtype_zmod,
  objtype_double,
  objtype_tagged,
  objtype_cdev,
  objtype_cord,
  objtype_cocfun
} objtype;

#define record env

typedef union{
  void * NULL_value;
  ring ring_value;
  coeff coeff_value;
  arpoly poly_value;
  coc_bool bool_value;
  cterm cterm_value;
  int int_value;
  pvector pvector_value;
  vlist vlist_value;
  char * string_value;
  module module_value;
  PSeries pseries_value;
  objtype objtype_value;
  opt_panel panel_value;
  olist olist_value;
  omat omat_value;
  ratfun ratfun_value;
  arbint arbint_value;
  arbrat arbrat_value;
  char * err_value;
  record record_value;
  zmod zmod_value;
  double * double_value;
  cdev cdev_value;
  cord cord_value;
  cocfun cocfun_value;  
} valueType;

#define value_NULL NULL

typedef struct object_info {
  char * id;
  objtype t;
  ring R;
  valueType v;
  int rc; /* reference count */

#ifdef OBJECT_DEBUG
  int num; /* progressive number to monitor alive objects */
#endif

} object_info;

#define object_NULL NULL

extern object object_null,object_ZERO,object_ONE;

/* -----------[ memory ]-------------- */

/* typedef struct env_slot_info * env; */ /* NULL if empty */

typedef struct env_slot_info * env_slot;

typedef struct env_slot_info {
  char * id;
  loc address;
  env_slot next;
} env_slot_info;

typedef struct env_stack_aux * env_stack; /* NULL => is empty */

typedef struct env_stack_aux {
  env top;
  env_stack pop;
} env_stack_info;

#define env_NULL NULL

/* TODO: check servono? e dove? */
#define env_ERROR_NONE 0
#define env_ERROR_ID_NOT_DECLARED 1

#define env_slot_NULL NULL


/* -----------[ subst ]-------------- */

typedef struct subst_info * subst;

typedef struct subst_info {
  ivector filter; /* indicates which indets has subst (0=false,1=true) */
  arpoly * sub;
} subst_info;

/* -----------[ ring ]-------------- */

typedef int ringtype;

#define ringtype_NULL NULL
#define ringtype_coeff 1
#define ringtype_poly 2
#define ringtype_quotient 3

#define NUM_AUX_IV 3

#define MAX_NUM_INDETS SHRT_MAX

typedef struct ring_el {

  ringtype rtype;
  
  int rc;
  coc_bool locked;
  env local_env;
  int characteristic;
  ring coeffs;
  module quid; /* quel quid in piu' che ha un anello quoziente */
  char * ring_name;
  short indetsNo; /* number of indeterminates */
  indet_info indets; /* informations about indeterminates */
  ordering to;

  ordering ot;
  int det;

  trnode * tripoli_dum;
  eee_mon * eee_dum;

  short whatTO;
  unknown first_u;
  unknown last_u; /* useful if whatTO=elim_num */

  short mord; /* module ordering (POS_TO or TO_POS) */

  weight weights;
  weilist weightlist;

  int level; /* tripoli level */

  ivector aux_iv[NUM_AUX_IV];

  void ( * InitCoeff )	( coeff * c );
  void ( * ClearCoeff )	( coeff * c );
  void ( * CopyCoeff )	( coeff * theCopy, coeff c );
  void ( * MakeCoeff )	( coeff * c, int num, int den, ring R );
  void ( * Str2Coeff ) 	( coeff * c, char * s, ring R );
  void ( * Cterm2Coeff) ( coeff * c, cterm t );

  void ( * AddCoeff ) 	( coeff * c, coeff c1, coeff c2, ring R);
  void ( * SubCoeff ) 	( coeff * c, coeff c1, coeff c2, ring R);
  void ( * MulCoeff ) 	( coeff * c, coeff c1, coeff c2, ring R);
  void ( * DivCoeff ) 	( coeff * c, coeff c1, coeff c2, ring R);
  void ( * ModCoeff ) 	( coeff * c, coeff c1, coeff c2, ring R);

  void ( * InvertCoeff ) 	( coeff * theInv, coeff c, ring R);
  void ( * NegCoeff )		( coeff * theOpp, coeff c, ring R );
  void ( * ChangeSignToCoeff)	( coeff * c, ring R );
  void ( * RandomizeCoeff )	( coeff * c, ring R );

  coc_bool  ( * AreEqualsCoeff )	( coeff c1, coeff c2 );
  coc_bool  ( * IsPositiveCoeff ) 	( coeff c1, ring R );
  coc_bool  ( * IsZeroCoeff )	( coeff c );
  coc_bool  ( * IsOneCoeff )	( coeff c );
  coc_bool  ( * IsMinusOneCoeff )	( coeff c, ring R );

  void ( * Coeff2Cterm )	( cterm * t, coeff c, ring R );
  void ( * Li2Coeff )	( coeff * c, verylongint n, ring R );

#ifdef TRIPOLI_OR_EEE

  term ( * scalars2term )       ( scalars ss, ring R );
  scalars ( * term2scalars )    ( term t, ring R );

#if defined(EEE)

  term ( * ivec_scalars2term )       ( ivec ss, ring R );
  ivec ( * ivec_term2scalars )    ( term t, ring R );

#endif

#endif

  int ( * term_which_gt)	( term t1, term t2, ordering ord, ring R );

  /* auxiliary fields */

  int halfP;

} ring_el;

#define ring_NULL NULL

/* -----------[ olist ]-------------- */

#define olist_NULL NULL
#define SIZE_OF_OLIST_BLOCK 20

/* -----------[ omatrix ]-------------- */

#define omat_NULL NULL

/* -----------[ package ]-------------- */

#define MAX_NUM_OPERATORS 1024
#define MAX_NUM_PACKAGES 1024
/* MAX_NUM_OPERATORS limits the number of operators for each package */

#define pkg_n_NULL -1

typedef char * pkg_id;

typedef struct pkg_info * package;

typedef struct pkg_info {
  op_info op_infos[MAX_NUM_OPERATORS];
  map op_map;
} pkg_info;

#endif
