/* $Id: coeffs.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include <stdio.h>

#include "gmp.h"

# include "ca.h"
# include "ck.h"
# include "coeffs.h"
# include "cterm.h"
# include "engine.h"
# include "operator.h"
# include "options.h"
# include "utils.h"
# include "mathcoc.h"

static void Normalize( rat Q );
static void MakeRat( rat Q, int N, int D );
/* static void MultRat( rat Q, rat Q1, rat Q2 ); */ /* TODO: REMOVE */
static void AddRat( rat Q, rat Q1, rat Q2 );
static void SubRat( rat Q, rat Q1, rat Q2 );
static void InvertRat( rat invQ, rat Q );
static void DivRat( rat Q, rat Q1, rat Q2 );

int coeff_size(coeff c, ring R)
{
  int res_i;
  switch(R->characteristic) {
    case characteristic_Li: return mpz_size(c)*sizeof(unsigned long int);
    case characteristic_Lr: {
      res_i=mpz_size(&((verylongrat)c)->_mp_num)*sizeof(unsigned long int)+
            mpz_size(&((verylongrat)c)->_mp_den)*sizeof(unsigned long int);
    } break;
    default:
      res_i=sizeof(void *); break; /* TODO: verif. Sugg. di Dusty */
  }
  return res_i;
}

zmod zmod_mult_arbint( zmod z, arbint n)
{
  arbint aux_n;
  int aux_n1;
  zmod aux_z,res_z;
  aux_n = malloc_verylongint();
  mpz_init ((verylongint)aux_n); 
  mpz_mmod_ui((verylongint)aux_n,n,zmod_p(z));
  aux_n1=this_arbint2int(aux_n);
  aux_z=zmod_mk(aux_n1,zmod_p(z));
  res_z=zmod_mult(z,aux_z);
  free_zmod(aux_z);
  return res_z;
}

zmod zmod_mult_int( zmod z, int n)
{
  int aux_n;
  zmod aux_z,res_z;
  aux_n=MMOD(n,zmod_p(z));
  aux_z=zmod_mk(aux_n,zmod_p(z));
  res_z=zmod_mult(z,aux_z);
  free_zmod(aux_z);
  return res_z;
}

double * double_dup(double * d)
{
  double * res_d;
  res_d=malloc_double();
  *res_d=*d;
  return res_d;
}

cterm double2cterm(double * d)
{
  char buf[100];
  cterm params[1];
  sprintf(buf,"%f",*d);
  params[0]=cterm_mk_NBS(buf);
  return cterm_mk_classic(op_FLOAT,1,params);
}

int zmod_sym_n(zmod z)
{
  if (zmod_n(z) > zmod_p(z)/2) return zmod_n(z)-zmod_p(z);
  else return zmod_n(z);
}

zmod zmod_mk(int n, int p)
{
  zmod res_z;
  res_z=malloc_zmod();
  zmod_set_p(res_z,p);
  zmod_set_n(res_z,n);
  return res_z;
}

zmod zmod_mk_norm(int n, int p)
{
  zmod res_z;
  res_z=malloc_zmod();
  zmod_set_p(res_z,p);
  zmod_set_n(res_z,MMOD(n,p));
  return res_z;
}

zmod zmod_inv(zmod z)
{
  int a0, a1, s0, s1, q, temp;
  a0 = zmod_n(z);    
  a1 = zmod_p(z);
  s0 = 1;      s1 = 0;
  while ( a1 ) {
    q = a0 / a1;
    temp = a0; a0 = a1;  a1 = temp - a1 * q;
    temp = s0; s0 = s1;  s1 = temp - s1 * q;
  }
  if (a0 != 1) { math_error=cocoa_error("CANNOT_INV"); return zmod_mk(0,zmod_p(z)); }

  return zmod_mk(MMOD(s0,zmod_p(z)),zmod_p(z));
}

zmod zmod_dup(zmod z)
{
  zmod res_z;
  res_z=malloc_zmod();
  zmod_set_p(res_z,zmod_p(z));
  zmod_set_n(res_z,zmod_n(z));
  return res_z;
}

cterm zmod2cterm(zmod z)
{
  cterm params[2];
  params[0]=int2cterm(zmod_sym_n(z));
  params[1]=int2cterm(zmod_p(z));
  return cterm_mk_classic(op_ZPVALUE,2,params);
}

arbint this_arbrat2arbint(arbrat r)
{
  arbint res_n;
  LiInit( &res_n );
  arbrat2arbint(res_n,r);
  LrClear(&r);
  return res_n;
}

arbrat arbrat_mk(arbint n1, arbint n2)
{
  arbrat res_r,aux_n1,aux_n2;
  LrInit( &res_r );
  aux_n1=arbint2arbrat(n1);
  aux_n2=arbint2arbrat(n2);
  LrDiv( &res_r, aux_n1, aux_n2, ring_NULL);
  LrClear(&aux_n1);
  LrClear(&aux_n2);
  return res_r;
}

arbrat arbint2arbrat(arbint n)
{
  arbrat res_r;
  LrInit( &res_r );
  Li2Lr(&res_r,n,ring_NULL);
  return res_r;
}

arbrat this_arbint2arbrat(arbint n)
{
  arbrat res_r;
  LrInit( &res_r );
  Li2Lr(&res_r,n,ring_NULL);
  LiClear(&n); /* TODO: optimize */
  return res_r;
}

arbint int2arbint(int n)
{
  arbint c;
  LiInit( &c );
  LiMakeCoeff(&c,n,1,ring_NULL);
  return c;
}

arbint arbint_dup(arbint n)
{
  arbint res_n;
  LiInit( &res_n );
  LiCopy( &res_n, n);
  return res_n;
}

arbrat arbrat_dup(arbrat n)
{
  arbrat res_n;
  LrInit( &res_n );
  LrCopy( &res_n, n);
  return res_n;
}

arbrat int2arbrat(int n)
{
  arbrat c;
  LrInit( &c );
  LrMakeCoeff(&c,n,1,ring_NULL);
  return c;
}

coeff coeff_opposite(coeff c,ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->CopyCoeff(&res_c,c);
  R->ChangeSignToCoeff(&res_c,R);
  return res_c;
}

coeff coeff_init(ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  return res_c;
}

coeff int2coeff(int n,ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->MakeCoeff(&res_c,n,1,R);
  return res_c;
}

coeff this_arbint2coeff(arbint n,ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->Li2Coeff(&res_c,n,R);
  LiClear(&n);
  return res_c;
}

coeff arbint2coeff(arbint n,ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->Li2Coeff(&res_c,n,R);
  return res_c;
}

coeff arbrat2coeff(arbrat q, ring R)
{
  coeff num, den, ans;
  R->InitCoeff(&num);
  R->InitCoeff(&den);
  R->InitCoeff(&ans);
  R->Li2Coeff(&num, mpq_numref((MP_RAT*)q), R);
  R->Li2Coeff(&den, mpq_denref((MP_RAT*)q), R);
  R->DivCoeff(&ans, num, den, R);
  coeff_free(num, R);
  coeff_free(den, R);
  return ans;
}

coeff coeff_x_int(coeff c, int n, ring R)
{
  coeff aux_c;
  aux_c=int2coeff(n,R);
  R->MulCoeff(&aux_c,aux_c,c,R);
  return aux_c;  
}

coc_bool IsGreaterCoeff( coeff c1, coeff c2, ring R ) {
  coeff c;
  coc_bool res;
  R ->InitCoeff( &c );
  R->SubCoeff( &c, c1, c2, R );
  res = R->IsPositiveCoeff( c, R );
  R->ClearCoeff( &c );
  return res;
}

coc_bool IsLiGreater( coeff c1, coeff c2, ring R ) {
  arbint c;
  coc_bool res;
  LiInit( &c );
  LiSub( &c, c1, c2, R );
  res = IsLiPositive( c, R );
  LiClear( &c );
  return res;
}

coc_bool IsLrGreater( coeff c1, coeff c2, ring R ) {
  arbrat c;
  coc_bool res;
  LrInit( &c );
  LrSub( &c, c1, c2, R );
  res = IsLrPositive( c, R );
  LrClear( &c );
  return res;
}

coeff coeff_power(coeff c,int exp,ring R)
{  /* exponent in {0,1,...} */
  coeff aux_c,res_c;
  if (exp == 1) return coeff_dup(c,R);
  if (exp == 0) return coeff_one(R);
  aux_c=coeff_power(c,exp/2,R);
  res_c=coeff_mult(aux_c,aux_c,R);
  coeff_free(aux_c,R);
  if (exp & 1) {
    aux_c=coeff_mult(res_c,c,R);
    coeff_free(res_c,R);
    res_c=aux_c;
  }
  return res_c;
}

coc_bool coeff_is_int(coeff c,ring R)
{
  switch(R->characteristic) {
    case characteristic_Li: return arbint_is_int((arbint)c);
    case characteristic_Lr: return arbrat_is_int((arbrat)c);
    default: return TRUE;
  }
}

int this_coeff2int(coeff c,ring R)
{
  int res_i;
  /* TODO: complete */
  switch(R->characteristic) {
    case characteristic_Li: return this_arbint2int((arbint)c);
    case characteristic_Si: return (int)c;
    case characteristic_Lr: {
      res_i=mpz_get_si(arbrat_num(c));
      LrClear(&c);
    } break;
    default:
      res_i=(coeff_zp)c; break;
  }
  return res_i;
}

int this_arbint2int(arbint r)
{
  int res_i;
  res_i=mpz_get_si(r);
  LiClear(&r);
  return res_i;
}

arbint this_coeff2arbint(coeff c,ring R)
{
  /* TODO: complete */
  verylongint res_i;
  switch(R->characteristic) {
    case characteristic_Li: return (arbint)c;
    case characteristic_Si: return int2arbint((int)c);
    case characteristic_Lr: {
      if (!arbrat_is_arbint((verylongrat)c)) return arbint_NULL;
      res_i = malloc_verylongint();
      mpz_init(res_i);
      mpq_get_num(res_i,(verylongrat)c);
      coeff_free(c,R);
    } break;
    default: 
      coeff_free(c,R);
      math_error=cocoa_error("CANNOT_CAST");
      return arbint_NULL;
  }
  return (arbint)res_i;
}

arbrat this_coeff2arbrat(coeff c,ring R)
{
  /* TODO: complete */
  switch(R->characteristic) {
    case characteristic_Lr: return (arbrat)c;
    case characteristic_Si: return int2arbrat((int)c);
    case characteristic_Li: return this_arbint2arbrat((arbint)c);
    default: 
/*    math_error=cocoa_error("CANNOT_CAST"); */
/*    coeff_free(c,R);                       */
/*    return arbrat_NULL;                    */      
    {
      long r = (int)c;
      arbrat ans;
      if (2*r > R->characteristic) r -= R->characteristic;
      ans = int2arbrat(r);
      coeff_free(c, R);
      return ans;
    }
  }
}

/* START GDD (941221) */
long coeff2long(coeff c,ring R)
{
  TODO("ELIMINARE coeff2long");
  return 0L;
}

#if 0

{
  long res_i;
  if (R->characteristic==characteristic_Li)          
    res_i=mpz_get_si(c.Li);
  else if (R->characteristic==characteristic_Si)
    res_i=c.Si;
  else if (R->characteristic==characteristic_Lr) {
    MP_INT aux_z;
    mpz_init(&aux_z);
    mpq_get_num(&aux_z,c.Lr);
    res_i=mpz_get_si(&aux_z);
    mpz_clear(&aux_z); }
  else if (R->characteristic==characteristic_Sr)
    res_i=c.Sr->num;
  else if (R->characteristic==characteristic_Sd)
    res_i=*c.Sd;
  else { 
    long p=(long)R->characteristic;
    if (c.Mi > p/2) res_i=(long)c.Mi-p;
    else res_i=c.Mi; }
  return res_i;
}
/* END GDD */

#endif

coeff cterm2coeff(cterm t,ring R)
{  
  coeff res_c;
  R->InitCoeff(&res_c);
  R->Cterm2Coeff(&res_c,t);
  return res_c;
}

cterm coeff2cterm(coeff c,ring R)
{
  cterm res_t;
  R->Coeff2Cterm(&res_t,c,R);
  return res_t;
}

coeff coeff_one(ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->Str2Coeff(&res_c, "1", R );
  return res_c;
}

coeff coeff_zero(ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->Str2Coeff(&res_c, "0", R );
  return res_c;
}

coeff coeff_dup(coeff c,ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->CopyCoeff(&res_c,c);
  return res_c;
}

coeff coeff_sum(coeff c1, coeff c2, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->AddCoeff(&res_c,c1,c2,R);
  return res_c;
}

coeff coeff_sum_these(coeff c1, coeff c2, ring R)
{
  R->AddCoeff(&c1,c1,c2,R);
  R->ClearCoeff(&c2);
  return c1;
}

coeff coeff_sum_these_free_if_zero(coeff c1, coeff c2, ring R)
{
  R->AddCoeff(&c1,c1,c2,R);
  R->ClearCoeff(&c2);
  if (coeff_is_zero(c1,R)) R->ClearCoeff(&c1);
  return c1;
}

coeff coeff_sub(coeff c1, coeff c2, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->SubCoeff(&res_c,c1,c2,R);
  return res_c;
}

coeff coeff_sub_these(coeff c1, coeff c2, ring R)
{
  R->SubCoeff(&c1,c1,c2,R);
  R->ClearCoeff(&c2);
  return c1;
}

coeff coeff_mult(coeff c1, coeff c2, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->MulCoeff(&res_c,c1,c2,R);
  return res_c;
}

coeff coeff_div(coeff c1, coeff c2, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->DivCoeff(&res_c,c1,c2,R);
  return res_c;
}

coeff coeff_mmod(coeff c1, coeff c2, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->ModCoeff(&res_c,c1,c2,R);
  return res_c;
}

coeff coeff_inverse(coeff c, ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->InvertCoeff(&res_c,c,R);
  return res_c;
}

coeff coeff_random( ring R)
{
  coeff res_c;
  R->InitCoeff(&res_c);
  R->RandomizeCoeff(&res_c,R);
  return res_c;
}

/*  --------------------------------------------------------
          integers modulo p
  --------------------------------------------------------
*/

void ZpInit ( coeff * c ) {}

void ZpClear ( coeff * c ) {}

void ZpCopy ( coeff * theCopy, coeff c ) { 
  cal(theCopy,c); 
}

void ZpMakeCoeff( coeff * c, int num, int den, ring R ) {
  int temp;
  int sign = 1;
  if (( num < 0 && den > 0 ) || ( num >= 0 && den < 0 ) ) /* TODO: check */
    sign = -1;
  temp = ABS( den );
  ZpInvert( (coeff *)&temp, (coeff)temp, R );
  temp *= ABS( num );
  temp %= R->characteristic;
  if ( sign < 0 )
    temp = - temp + R->characteristic;
  cal(c,temp);
}

void ZpStr2Coeff( coeff * c, char * s, ring R ) { 
  register int N, i;
  for ( N = 0, i = 0; s[i] >= '0' && s[i] <= '9' ; i++)
    N = (10 * N + s[i] - '0' ) % R->characteristic;
  cal(c,N);
}

void ZpCterm2Coeff( coeff * c, cterm t ) {
  /* t=zp(p,n) */
  cterm_list params=cterm_params(t);
  char * s=cterm_string(params[1]);
  register int i, N;
  int p = atoi(cterm_string(params[0]));
  for ( N = 0, i = 0; s[i] >= '0' && s[i] <= '9' ; i++)
    N = (10 * N + s[i] - '0' ) % p;
  cal(c,N);
}

void ZpInvert( coeff * theInv, coeff c, ring R ) {
  int a0, a1, s0, s1, q, temp;
  a0 = (int)c;    a1 = R->characteristic;
  s0 = 1;      s1 = 0;
  while ( a1 ) {
    q = a0 / a1;
    temp = a0;    a0 = a1;  a1 = temp % a1;
    temp = s0;    s0 = s1;   s1 = temp - s1 * q;
  }  
  if ( a0 != 1 ) math_error = cocoa_error("CANNOT_DIV");
  cal(theInv,((s0 % R->characteristic) + R->characteristic ) % R->characteristic);
}

void ZpNeg( coeff * theOpp, coeff c, ring R) {
  cal(theOpp, -(coeff_zp)c + R->characteristic);
}

void ZpChangeSign( coeff * c, ring R) {
  cal(c, -(coeff_zp)*c + R->characteristic);
}

void ZpRandomize( coeff * c, ring R) {
  cal(c,rand() % R->characteristic);
}

void ZpAdd( coeff * c, coeff c1, coeff c2, ring R ) {
  cal(c,( (coeff_zp)c1 + (coeff_zp)c2 ) % R->characteristic);
}

void ZpSub( coeff * c, coeff c1, coeff c2, ring R ) {
  cal(c, ( (coeff_zp)c1 - (coeff_zp)c2 + R->characteristic )
    % R->characteristic);
}

void ZpMul( coeff * c, coeff c1, coeff c2, ring R ) {
  cal(c,( (coeff_zp)c1 * (coeff_zp)c2 ) % R->characteristic);
}

void ZpDiv( coeff * c, coeff c1, coeff c2, ring R ) {
  coeff temp;
  ZpInvert( &temp, c2, R );
  cal(c, ( (coeff_zp)c1 * (coeff_zp)temp ) % R->characteristic);
}

void ZpMod( coeff * c, coeff c1, coeff c2, ring R ) {
  cal(c,0); /* TODO: Eh? */
}

coc_bool AreZpEquals( coeff c1, coeff c2 ) {
  return (coeff_zp)c1 == (coeff_zp)c2;
}

coc_bool IsZpPositive( coeff c1, ring R ) {
  return (coeff_zp)c1 <= (coeff_zp)(R->halfP);
}

coc_bool IsZpZero( coeff c ) {
  return (coeff_zp)c == 0;
}

coc_bool IsZpOne( coeff c ) {
  return (coeff_zp)c == 1;
}

coc_bool IsZpMinusOne( coeff c, ring R ) {
  return (coeff_zp)c == (coeff_zp)R->characteristic - 1;
}

void Zp2Cterm( cterm *t, coeff c, ring R) {
  char buf[100];
  char p[10];
  cterm params[2];
  sprintf( buf, "%d", (coeff_zp)c );
  sprintf( p, "%d", R->characteristic);
  params[0]=cterm_mk_NBS(p);
  params[1]=cterm_mk_NBS(buf);
  *t=cterm_mk_classic(op_ZP,2,params);
}

void Li2Zp( coeff * c, verylongint n, ring R)
{
  MP_INT aux_n1,aux_n2;
  int num;
  
  mpz_init(&aux_n1);
  mpz_init(&aux_n2);
  mpz_set_si( &aux_n1, R->characteristic);
  mpz_mmod ( (verylongint)(&aux_n2), n, &aux_n1 ); 
  num=mpz_get_si(&aux_n2);
  ZpMakeCoeff( c, num, 1, R );
  mpz_clear( &aux_n1 );
  mpz_clear( &aux_n2 );
}

/*  --------------------------------------------------------
          short integers
  --------------------------------------------------------
*/

int  IntGCD(register int m, register int n) {
  register int r;

  m = ABS( m );  
  n = ABS( n );  
  if ( n == 0 ) 
    return( m );
  else
    while ( 1 ) {
      r = m % n;
      if (r == 0) 
        return( n );
      else {
        m = n % r;
        if ( m == 0 )
          return( r );
        else {
          n = r % m;
          if ( n == 0 )
            return( m );
        }
      }
    }
}

int  IntLCM (int x, int y) {
  return( (x / IntGCD( x, y )) * y );
}

coc_bool IsPrime( long int N ) {
  int h, d;
  if ( N <= 1 )
    return FALSE;
  if ( N == 2 || N == 3 )
    return TRUE;
  if ( ( N % 2 == 0 ) || ( N % 3 == 0))
    return FALSE;
  d = 5; 
  h = 4;
  while ( d * d <= N ) {
    if ( N % d != 0 ) {
      h = 6 - h;
      d += h;
    } else 
      return FALSE;
  }
  return TRUE;
}



void SiInit ( coeff * c ) {}

void SiClear ( coeff * c ) {}

void SiCopy ( coeff * theCopy, coeff c ) { 
  cal(theCopy, c);
}

void SiMakeCoeff( coeff * c, int num, int den, ring R ) { 
  cal(c,num);
}

void SiStr2Coeff( coeff * c, char * s, ring R )   {  
  register int N, i;
  for ( N = 0, i = 0; s[i] >= '0' && s[i] <= '9' ; i++)
    N = (10 * N + s[i] - '0' );
  cal(c,N);
}


void SiCterm2Coeff( coeff * c, cterm t ) 
{ /* t=short_int(signed_number) */
  cterm_list params=cterm_params(t);
  cal(c,atoi(cterm_string(params[0])));
}

void SiInvert( coeff * theInv, coeff c, ring R )  {
  cal(theInv,c);        /* TODO: Errore se debug */
}

void SiNeg( coeff * theOpp, coeff c, ring R )  {
  cal(theOpp,-(int)c);
}

void SiChangeSign( coeff * c, ring R) { 
  cal(c,-(int)*c);
}

void SiRandomize( coeff * c , ring R) { 
  cal(c,rand());
}

void SiAdd( coeff * c, coeff c1, coeff c2, ring R ) { 
  cal(c,(int)c1 + (int)c2);
}

void SiSub( coeff * c, coeff c1, coeff c2, ring R ) { 
  cal(c,(int)c1 - (int)c2);
}

void SiMul( coeff * c, coeff c1, coeff c2, ring R ) { 
  cal(c,(int)c1 * (int)c2);
}

void SiDiv( coeff * c, coeff c1, coeff c2, ring R ) { 
  if ((int)c2<0) {
    c2=(coeff)-(int)c2; 
    c1=(coeff)-(int)c1; }
  cal(c,((int)c1-MMOD((int)c1,(int)c2))/(int)c2);
}

void SiMod( coeff * c, coeff c1, coeff c2, ring R ) { 
  cal(c,MMOD((int)c1,(int)c2));
}

coc_bool AreSiEquals( coeff c1, coeff c2 ) { 
  return (int)c1 == (int)c2; 
}

coc_bool IsSiPositive( coeff c1, ring R ) {
  return (int)c1 > 0; 
}

coc_bool IsSiZero( coeff c )  { 
  return (int)c == 0; 
}

coc_bool IsSiOne( coeff c ) { 
  return (int)c == 1; 
}

coc_bool IsSiMinusOne( coeff c, ring R )  { 
  return (int)c == - 1; 
}

void Si2Cterm( cterm *t, coeff c, ring R ) 
{
  char buf[20];
  cterm params[1];
  sprintf( buf, "%d", (int)c );
  params[0]=cterm_mk_NBS(buf);
  *t=cterm_mk_classic(op_SHORT_INT,1,params);
}

void Li2Si( coeff * c, verylongint n, ring R)
{
  *c=(coeff)mpz_get_si(n);
}

/*  --------------------------------------------------------
          short rationals
  --------------------------------------------------------
*/

static void Normalize( rat Q ) {
  register int m;
  if ( Q -> num == 0 )
    Q -> den = 1;
  else
  {  m = IntGCD( Q -> num, Q -> den );
    Q -> num /= m;
    Q -> den /= m;
  }
}

/* TODO: macro? */
static void MakeRat( rat Q, int N, int D ) {
  Q -> num = N;    
  Q -> den = D;
  Normalize( Q );
}

static void MulRat( rat Q, rat Q1, rat Q2 ) {
  aux_rat auxQ1, auxQ2;
  MakeRat( &auxQ1, Q1->num, Q2->den );
  MakeRat( &auxQ2, Q2->num, Q1->den );
  Q -> num = auxQ1.num * auxQ2.num;
  Q -> den = auxQ1.den * auxQ2.den;
}

static void AddRat( rat Q, rat Q1, rat Q2 ) {
  register int d1, d2, t;  
  
  d1 = IntGCD( Q1->den, Q2->den );
  if ( d1 == 1 ) {
    Q->num = Q1->num * Q2->den + Q2->num * Q1->den;
    Q->den = Q1->den * Q2->den;
  } else {
    t = Q1->num * ( Q2->den / d1 ) + Q2->num * ( Q1->den / d1 );
    d2 = IntGCD( t, d1 );
    Q->num = t / d2;
    Q->den = (Q1->den / d1) * (Q2->den / d2 );
  }
}

static void SubRat( rat Q, rat Q1, rat Q2 ) {
  register int d1, d2, t;  
  
  d1 = IntGCD( Q1->den, Q2->den );
  if ( d1 == 1 ) {
    Q->num = Q1->num * Q2->den - Q2->num * Q1->den;
    Q->den = Q1->den * Q2->den;
  } else {
    t = Q1->num * ( Q2->den / d1 ) - Q2->num * ( Q1->den / d1 );
    d2 = IntGCD( t, d1 );
    Q->num = t / d2;
    Q->den = (Q1->den / d1) * (Q2->den / d2 );
  }
}

static void InvertRat( rat invQ, rat Q ) {
  invQ -> num = Q->den;
  if ( Q->num < 0 )
    invQ -> num =  - invQ -> num;
  invQ -> den = ABS( Q->num );
}

static void DivRat( rat Q, rat Q1, rat Q2 ) {
  aux_rat temp;
  InvertRat( &temp, Q2 );
  MulRat( Q,  Q1, &temp );  
}


/*.....................................*/

void SrInit ( coeff * c ) { 
  cal(c, malloc_rat());
  MakeRat( (rat)(*c), 0, 1 );
}

void SrClear ( coeff * c ) {
  if ( *c != NULL ) {
    free_rat( *c );
    *c = NULL;
  }
}


void SrCopy ( coeff * theCopy, coeff c ) { 
  rat_set_num((rat)(*theCopy),rat_num((rat)c));
  rat_set_den((rat)(*theCopy),rat_den((rat)c));
}

void SrMakeCoeff( coeff * c, int num, int den, ring R ) {   
  MakeRat( (rat)(*c), num, den ); 
}

void SrStr2Coeff( coeff * c, char * s, ring R )  {
  register int N, i;
  for ( N = 0, i = 0; s[i] >= '0' && s[i] <= '9' ; i++)
    N = 10 * N + s[i] - '0' ;
  rat_set_num((rat)(*c),N) ;
  rat_set_den((rat)(*c),1) ;
}

void SrCterm2Coeff( coeff * c, cterm t) 
{  /* t=short_rat(signed_number,number) */
  cterm_list params=cterm_params(t);
  rat_set_num((rat)(*c),atoi(cterm_string(params[0])));
  rat_set_den((rat)(*c),atoi(cterm_string(params[1])));
}

void SrInvert( coeff * theInv, coeff c, ring R ) { 
  InvertRat( (rat)(*theInv), (rat)c ); 
}

void SrNeg( coeff * theOpp, coeff c, ring R ) { 
  rat_set_num((rat)(*theOpp), -rat_num((rat)c)) ;
  rat_set_den((rat)(*theOpp),rat_den((rat)c));
}

void SrChangeSign( coeff * c, ring R) {  
  rat_set_num((rat)(*c), -rat_num((rat)*c)) ;
}

void SrRandomize( coeff * c, ring R) { 
  int n, d;
  n = rand();
  while ((d = rand())==0);
  SrMakeCoeff( c, n, d, R );
}

void SrAdd( coeff * c, coeff c1, coeff c2, ring R ) { 
  AddRat( (rat)(*c), (rat)c1, (rat)c2 ); 
}

void SrSub( coeff * c, coeff c1, coeff c2, ring R ) { 
  SubRat( (rat)(*c), (rat)c1, (rat)c2 ); 
}

void SrMul( coeff * c, coeff c1, coeff c2, ring R ) { 
  MulRat( (rat)(*c), (rat)c1, (rat)c2 ); 
}

void SrDiv( coeff * c, coeff c1, coeff c2, ring R ) { 
  DivRat( (rat)(*c), (rat)c1, (rat)c2 ); 
}

void SrMod( coeff * c, coeff c1, coeff c2, ring R ) { 
  MakeRat( (rat)(*c), 0, 1 ); 
}

coc_bool AreSrEquals( coeff c1, coeff c2 ) {
  return ( rat_num((rat)c1) == rat_num((rat)c2) 
    && rat_den((rat)c1) == rat_den((rat)c2) ); 
}

coc_bool IsSrPositive( coeff c, ring R ) {
  return rat_num((rat)c) > 0; 
}

coc_bool IsSrZero( coeff c )  {
  return rat_num((rat)c) == 0; 
}

coc_bool IsSrOne( coeff c ) { 
  return rat_num((rat)c) == rat_den((rat)c); 
}

coc_bool IsSrMinusOne( coeff c, ring R )  { 
  return rat_num((rat)c) == -rat_den((rat)c); 
}

void Sr2Cterm( cterm * t, coeff c, ring R )
{
  cterm params[2];
  char buf[20];
  sprintf(buf,"%d",rat_num((rat)c));
  params[0]=cterm_mk_NBS(buf);
  sprintf(buf,"%d",rat_den((rat)c));
  params[1]=cterm_mk_NBS(buf);
  *t=cterm_mk_classic(op_SHORT_RAT,2,params);
}

void Li2Sr( coeff * c, verylongint n, ring R)
{
  MakeRat( (rat)(*c), mpz_get_si(n), 1 ); 
}

/*  --------------------------------------------------------
          doubles
  --------------------------------------------------------
*/

#define dass(c,D) *(double *)(*c)=(double)(D)

void SdInit ( coeff * c ) { 
  *c=(coeff)malloc_double();
  dass(c,0.0);
}

void SdClear ( coeff * c ) {
  if ( *c != NULL ) {
    free_double( *c );
    *c = NULL;
  }
}

void SdCopy ( coeff * theCopy, coeff c ) { 
  dass(theCopy,*(double *)c);
}

void SdMakeCoeff( coeff * c, int num, int den, ring R ) { 
  dass(c,(0.0 + num)/ den) ; 
}

void SdStr2Coeff( coeff * c, char * s, ring R )  {  
  dass(c, atof(s));
}

void SdCterm2Coeff( coeff * c, cterm t )
{ /* t=double(s:NBS) */
  cterm_list params=cterm_params(t);
  dass(c,atof(cterm_string(params[0])));
}

void SdInvert( coeff * theInv, coeff c, ring R ) { 
  dass(theInv,1.0/(*(double *)c)) ; 
}

void SdNeg( coeff * theOpp, coeff c, ring R ) { 
  dass(theOpp, - (*(double *)c)) ;
}

void SdChangeSign( coeff * c, ring R) {
  dass(c, - (*(double *)(*c))) ;
}

void SdRandomize( coeff * c, ring R) { 
  dass(c, (( 0.0 + rand() )/12345) * rand());
}

void SdAdd( coeff * c, coeff c1, coeff c2, ring R ) {
  dass(c, *((double *)c1) + *((double *)c2)); 
}

void SdSub( coeff * c, coeff c1, coeff c2, ring R ) {
  dass(c, *((double *)c1) - *((double *)c2)); 
}

void SdMul( coeff * c, coeff c1, coeff c2, ring R ) {
  dass(c, *((double *)c1) * *((double *)c2)); 
}

void SdDiv( coeff * c, coeff c1, coeff c2, ring R ) {
  dass(c, *((double *)c1) / *((double *)c2)); 
}

void SdMod( coeff * c, coeff c1, coeff c2, ring R ) {
  dass(c, 0.0); 
}

coc_bool AreSdEquals( coeff c1, coeff c2 ) {
  return *((double *)c1) == *((double *)c2) ; 
}

coc_bool IsSdPositive( coeff c1, ring R ){ 
  return  *((double *)c1) > 0; 
}

coc_bool IsSdZero( coeff c )  { 
  return *((double *)c) == 0 ; 
}

coc_bool IsSdOne( coeff c ) { 
  return *((double *)c) == 1; 
}

coc_bool IsSdMinusOne( coeff c, ring R )  { 
  return *((double *)c) == -1; 
}

void Sd2Cterm( cterm * t, coeff c, ring R)
{
  char buf[50];
  cterm params[1];
  sprintf( buf, "%f", *((double *)c) );
  params[0]=cterm_mk_NBS(buf);
  *t=cterm_mk_classic(op_SHORT_DOUBLE,1,params);
}

void Li2Sd( coeff * c, verylongint n, ring R)
{
  char * s;
  int sz=mpz_sizeinbase(n,10)+200;
  s=malloc_str(sz);
  mpz_get_str(s,10,n);
  dass(c, atof(s));
  free_str(sz,s);
}

/*  --------------------------------------------------------
          big integers
  --------------------------------------------------------
*/

void LiInit ( coeff * c ) { 
  *c = (coeff)malloc_verylongint();
  mpz_init ((verylongint)(*c)); 
}

void LiClear ( coeff * c ) { 
  if ( *c != NULL ) {
    mpz_clear ((verylongint)(*c)); 
    free_verylongint( (verylongint)(*c) );
    *c = NULL;
  }
}

void LiCopy ( coeff * theCopy, coeff c ) { 
  mpz_set ( (verylongint)(*theCopy), (verylongint)c ); 
}

void LiMakeCoeff( coeff * c, int num, int den, ring R ) {  
  mpz_set_si( (verylongint)(*c), num );
}

void LiStr2Coeff( coeff * c, char * s, ring R )  {
  mpz_set_str ((verylongint)(*c), s, 10);
}

void LiCterm2Coeff(coeff * c,cterm t)
{ /* t=long_int(signed_number) */
  /* c is assumed to be already allocated */
  cterm_list params=cterm_params(t);
  mpz_set_str((verylongint)(*c),cterm_string(params[0]),10);
}

void LiInvert( coeff * theInv, coeff c, ring R ) {
  mpz_set( (verylongint)(*theInv), (verylongint)c );
}

void LiNeg( coeff * theOpp, coeff c, ring R )  {  
  mpz_neg ( (verylongint)(*theOpp), (verylongint)c );
}

void LiChangeSign( coeff * c, ring R)    {   
  /* mpz_neg ( (verylongint)(*c), (verylongint)(*c) ); */
  ((verylongint)(*c))->_mp_size=-((verylongint)(*c))->_mp_size;
}

void LiRandomize( coeff * c, ring R) {
  int negate;
  mpz_urandomb((verylongint)(*c), GMP_RANDSTATE, GMP_RANDSIZE); /* only for choosing the sign */
  negate = mpz_tstbit((verylongint)(*c), 1);
  mpz_urandomb((verylongint)(*c), GMP_RANDSTATE, GMP_RANDSIZE);
  if (negate) mpz_neg((verylongint)(*c), (verylongint)(*c));
}

void LiAdd( coeff * c, coeff c1, coeff c2, ring R )  {
  mpz_add ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 ); 
}

void LiSub( coeff * c, coeff c1, coeff c2, ring R )  {
  mpz_sub ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 ); 
}

void LiMul( coeff * c, coeff c1, coeff c2, ring R )  {
  mpz_mul ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 ); 
}

void LiDiv( coeff * c, coeff c1, coeff c2, ring R ) {
  /* TODO: when GNU will be corrected, we MUST modify */
  if ( mpz_cmp_si( (verylongint)c2, 0 ) < 0 ) {
    LiChangeSign(&c2,R);
    mpz_mdiv ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 );
    LiChangeSign(&c2,R);
    LiChangeSign(c,R); }
  else
    mpz_mdiv ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 );
  /* if ( mpz_cmp_si( (verylongint)c2, 0 ) < 0 )
    mpz_add_ui ( (verylongint)(*c), (verylongint)(*c), 1 ); */
}

void LiMod( coeff * c, coeff c1, coeff c2, ring R ) {
  /* TODO: when GNU will be corrected, we MUST modify */
  if ( mpz_cmp_si( (verylongint)c2, 0 ) < 0 ) {
    LiChangeSign(&c2,R);
    mpz_mmod ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 );
    LiChangeSign(&c2,R); }
  else
    mpz_mmod ( (verylongint)(*c), (verylongint)c1, (verylongint)c2 );
  /* if ( mpz_cmp_si( (verylongint)c2, 0 ) < 0 )
    mpz_sub ( (verylongint)(*c), (verylongint)(*c), (verylongint)c2 ); */
}

coc_bool AreLiEquals( coeff c1, coeff c2 ) {
  return mpz_cmp( (verylongint)c1, (verylongint)c2 ) == 0;
}

coc_bool IsLiPositive( coeff c1, ring R ) {  
  if ( mpz_cmp_si( (verylongint)c1, 0 ) > 0 )
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLiZero( coeff c )  {  
  if ( mpz_cmp_si( (verylongint)c, 0 ) == 0 )
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLiOne( coeff c )  {  
  if ( mpz_cmp_si( (verylongint)c, 1 ) == 0 )
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLiMinusOne( coeff c, ring R )  {  
  if ( mpz_cmp_si( (verylongint)c, -1 ) == 0 )
    return TRUE;
  else
    return FALSE;
}

void Li2Cterm( cterm * t, coeff c, ring R )
{
  cterm params[1];
  char * s;
  int sz=mpz_sizeinbase((verylongint)c,10)+200;
  s=malloc_str(sz);
  mpz_get_str(s,10,(verylongint)c);
  params[0]=cterm_mk_NBS(s);
  *t=cterm_mk_classic(op_LONG_INT,1,params);
  free_str(sz,s);
}

void Li2Li( coeff * c, verylongint n, ring R)
{
  mpz_set ( (verylongint)(*c), n ); 
}

/*  --------------------------------------------------------
          big rationals
  --------------------------------------------------------
*/

void LrInit ( coeff * c ) { 
  *c = (coeff)malloc_verylongrat();
  mpq_init ((verylongrat)(*c)); 
}

void LrClear ( coeff * c ) { 
  if ( *c != NULL ) {
    mpq_clear ((verylongrat)(*c)); 
    free_verylongrat( (verylongrat)(*c) );
    *c = NULL;
  }
}

void LrCopy ( coeff * theCopy, coeff c ) { 
  mpq_set ( (verylongrat)(*theCopy), (verylongrat)c ); 
}

void LrMakeCoeff( coeff * c, int num, int den, ring R ) {  
  mpq_set_si( (verylongrat)(*c), num, den );
}

void LrStr2Coeff( coeff * c, char * s, ring R )  {
  MP_INT numer, denom;
  mpz_init(&numer);
  mpz_init(&denom);
  mpz_set_str( &numer, s, 10 );
  mpq_set_num( (verylongrat)(*c), &numer );
  mpz_set_si( &denom, 1);
  mpq_set_den( (verylongrat)(*c), &denom );
  mpz_clear( &numer );
  mpz_clear( &denom );
}

void LrCterm2Coeff(coeff * c,cterm t)
{ /* t=long_rat(signed_number,number) */
  /* c is absumed to be already allocated */
  cterm_list params=cterm_params(t);
  MP_INT num,den;
  mpz_init(&num);
  mpz_init(&den);
  mpz_set_str(&num,cterm_string(params[0]),10);
  mpz_set_str(&den,cterm_string(params[1]),10);
  mpq_set_num((verylongrat)(*c),&num);
  mpq_set_den((verylongrat)(*c),&den);
  mpz_clear(&num);
  mpz_clear(&den);
}

void LrInvert( coeff * theInv, coeff c, ring R ) { 
  mpq_inv ( (verylongrat)(*theInv), (verylongrat)c );
}

void LrNeg( coeff * theOpp, coeff c, ring R )  {  
  mpq_neg ( (verylongrat)(*theOpp), (verylongrat)c ); 
}

void LrChangeSign( coeff * c, ring R)    {   
  mpq_neg ( (verylongrat)(*c), (verylongrat)(*c) ); 
}

void LrRandomize( coeff * c, ring R) { 
  int negate;
  mpz_urandomb(mpq_numref((verylongrat)(*c)), GMP_RANDSTATE, GMP_RANDSIZE);
  negate = mpz_tstbit(mpq_numref((verylongrat)(*c)), 1);
  mpz_urandomb(mpq_numref((verylongrat)(*c)), GMP_RANDSTATE, GMP_RANDSIZE);
  if (negate) mpz_neg(mpq_numref((verylongrat)(*c)), mpq_numref((verylongrat)(*c)));
  mpz_set_ui(mpq_denref((verylongrat)(*c)), 1);
}

void LrAdd( coeff * c, coeff c1, coeff c2, ring R )  {
  mpq_add ( (verylongrat)(*c), (verylongrat)c1, (verylongrat)c2 ); 
}

void LrSub( coeff * c, coeff c1, coeff c2, ring R )  {
  mpq_sub ( (verylongrat)(*c), (verylongrat)c1, (verylongrat)c2 ); 
}

void LrMul( coeff * c, coeff c1, coeff c2, ring R )  {
  mpq_mul ( (verylongrat)(*c), (verylongrat)c1, (verylongrat)c2 ); 
}

void LrDiv( coeff * c, coeff c1, coeff c2, ring R ) {  
  mpq_div ( (verylongrat)(*c), (verylongrat)c1, (verylongrat)c2 ); 
}

void LrMod( coeff * c, coeff c1, coeff c2, ring R ) {  
  LrMakeCoeff(c,0,1,R);
}

coc_bool AreLrEquals( coeff c1, coeff c2 )    {
  return mpq_cmp( (verylongrat)c1, (verylongrat)c2 ) == 0;
}

coc_bool IsLrPositive( coeff c1, ring R )  {  
  if ( mpz_cmp_si( &(((verylongrat)c1)->_mp_num), 0 ) > 0 )
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLrZero( coeff c )    {  
  if ( mpz_cmp_si( &(((verylongrat)c)->_mp_num), 0 ) == 0 )
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLrOne( coeff c )    {  
  if (mpz_cmp( &(((verylongrat)c)->_mp_num),&(((verylongrat)c)->_mp_den)) == 0)
    return TRUE;
  else
    return FALSE;
}

coc_bool IsLrMinusOne( coeff c, ring R )    {  
  if ( (mpz_cmp_si( &(((verylongrat)c)->_mp_num), -1 ) == 0) && 
          (mpz_cmp_si( &(((verylongrat)c)->_mp_den),1 ) == 0) )
    return TRUE;
  else
    return FALSE;
}

void Lr2Cterm( cterm * t, coeff c, ring R )
{
  
  cterm params[2];
  int sz_num,sz_den;
  char *num=NULL;
  char *den=NULL;
  sz_num=mpz_sizeinbase(&(((verylongrat)c)->_mp_num),10)+200;
  sz_den=mpz_sizeinbase(&(((verylongrat)c)->_mp_den),10)+200;
  num=malloc_str(sz_num);
    /* TRICK */
  mpz_get_str(num,10,&(((verylongrat)c)->_mp_num));
  den=malloc_str(sz_den);
    /* TRICK */
  mpz_get_str(den,10,&(((verylongrat)c)->_mp_den));

  params[0]=cterm_mk_NBS(num);
  params[1]=cterm_mk_NBS(den);
  *t=cterm_mk_classic(op_LONG_RAT,2,params);
  free_str(sz_num,num);
  free_str(sz_den,den);
}

void Li2Lr( coeff * c, verylongint n, ring R)
{
  MP_INT denom;
  mpz_init(&denom);
  mpq_set_num( (verylongrat)(*c), n );
  mpz_set_si( &denom, 1);
  mpq_set_den( (verylongrat)(*c), &denom );
  mpz_clear( &denom );
}
