/* $Id: cterm.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "ca.h"
#include "cocdefs.h"
#include "cterm.h"
#include "ivectors.h"
#include "map.h"
#include "operator.h"
#include "utils.h"

cterm error_param=cterm_NULL;
cterm error_context=cterm_NULL;
cterm cterm_unknown_context;
extern coop op_FUN_APPL;

/* ---[ cterm OPERATORS ]----------------------------------------------- */

coop op_NBS;
coop op_LIST;
coop op_NOINDENT;
coop op_BIND;
coop op_STRING;
coop op_HELPQUERY;
coop op_TRUE;
coop op_FALSE;

void cterm_initialize(void)
{
  static coc_bool already_called=FALSE;
  if (already_called) return;
  already_called=TRUE;
  coop_init();
  op_NBS = op_map_add(pkg_builtin_n,"\040NBS",NO_ARITY);
  op_LIST = op_map_add(pkg_builtin_n,"\040LIST",NO_ARITY); 
  op_NOINDENT = op_map_add(pkg_builtin_n,"\040NOINDENT",NO_ARITY); 
  op_BIND = op_map_add(pkg_builtin_n,"\040BIND",2);
  op_STRING = op_map_add(pkg_builtin_n,"\040STRING",1);
  op_FALSE = op_map_add(pkg_builtin_n,"FALSE",0);
  op_TRUE = op_map_add(pkg_builtin_n,"TRUE",0);
  cterm_unknown_context=cterm_mk_NBS("Unknown");
}

cterm cterm_mk_table_entry(char * descr, cterm value)
{
  cterm params[2];
  params[0]=cterm_mk_NBS(descr);
  params[1]=value;
  return cterm_mk_classic(op_BIND,2,params);
  /* return cterm_mk_classic(op_TABLE_ENTRY,2,params); */
}

cterm cterm_noindent(cterm t)
{
  cterm params[1];
  params[0]=t;
  return cterm_mk_classic(op_NOINDENT,1,params);
}

cterm cterm_init(void)
{
  cterm p;
  p = malloc_cterm();
  p->op=op_NULL;
  p->string=NULL;
  p->params=NULL;
  return(p);
}

/* CONSTRUCTORS of cterm */

cterm cterm_mk_fun_appl2(char * op,int arity,cterm * params)
{
  cterm aux_t;
  cterm aux_params[2];
  aux_params[0]=cterm_mk_NBS(op);
  aux_params[1]=cterm_mk_classic(op_LIST,arity,params);
  return cterm_mk_classic(op_FUN_APPL,2,aux_params);
}

cterm cterm_mk_fun_appl(coop op,int arity,cterm * params)
{
  cterm aux_t;
  cterm aux_params[2];
  aux_params[0]=cterm_mk_NBS(op_map_op_name(pkg_builtin_n, op));
  aux_params[1]=cterm_mk_classic(op_LIST,arity,params);
  return cterm_mk_classic(op_FUN_APPL,2,aux_params);
}

cterm cterm_mk(coop op,int arity,cterm * params,char * str)
{
  cterm t;
  int i;

  t = cterm_init();
  t->op = op;
  t->arity = arity;
  if (op == op_NBS) {
    t->string = my_strdup(str);
    t->params = NULL;
    }
  else /* if (op != op_NBS) */ {
    t->string = NULL;
    if (params!=NULL) {
      t->params = malloc_cterm_list(arity);
      for (i=0;i<arity;i++) t->params[i]=params[i];
      }
    else t->params = NULL;
    }
  return t;
}

cterm cterm_dup(cterm t)
{
  cterm res_t;
  if (t==cterm_NULL) return res_t=cterm_NULL;
  else {
    coop op=cterm_op(t);
    int arity=cterm_arity(t);
    cterm_list params=cterm_params(t);
    cterm_list res_params=NULL;
    char * res_s=cterm_string(t);
    register int i;
    if (arity!=0) {
      res_params = malloc_cterm_list(arity);
      for (i=0;i<arity;i++) res_params[i]=cterm_dup(params[i]);
      res_t=cterm_mk(op,arity,res_params,res_s);
      free_cterm_list(arity,res_params); }
    else
      res_t=cterm_mk(op,arity,res_params,res_s);
    return res_t; }
}

void cterm_free(cterm t)
{
  register int i;
  if (t!=cterm_NULL) {
    if (t->op==op_NBS) free_str1(t->string);
    else
      if (t->arity!=0) {
        for (i=0;i<t->arity;i++) cterm_free(t->params[i]);
        free_cterm_list(t->arity,t->params); }
    free_cterm(t); }
}

/* ---[ cterm converters ]----------------------------------------------- */

cterm int2cterm(int i)
{
  char aux_s[20]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  sprintf(aux_s,"%d",i);
  return cterm_mk_NBS(aux_s);
}

cterm string2cterm(char * s)
{
  cterm params[1];
  params[0]=cterm_mk_NBS(s);
  return cterm_mk_classic(op_STRING,1,params);
}

cterm bool2cterm(coc_bool b)
{
  if (b) return cterm_mk_const(op_TRUE);
  else return cterm_mk_const(op_FALSE);
}

cterm iv2cterm(ivector v)
{
  int i,n;
  cterm_list params;
  n=iv_get_len(v);
  if (n==0) return cterm_mk_const(op_LIST);
  params=malloc_cterm_list(n);
  for (i=1;i<=n;i++) params[i-1]=int2cterm(iv_get_nth(v,i));
  return cterm_mk_classic(op_LIST,n,params);
}

void printf_iv(ivector v)
{
  cterm aux_t;
  aux_t=iv2cterm(v);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

#if 0 /* TODO: cheffare? */

cterm matrix2cterm(matrix M,int nr, int nc)
{
  int r,c;
  cterm_list cterm_mat,cterm_row;
  cterm res_t;
  cterm_mat=malloc_cterm_list(nr);
  for (r=0;r<nr;r++) {
    cterm_row=malloc_cterm_list(nc);
    for (c=0;c<nc;c++)
      cterm_row[c]=int2cterm(matrix_elem(M,r,c));
    cterm_mat[r]=cterm_mk_classic(op_ROW,nc,cterm_row); }
  res_t=cterm_mk_classic(op_MAT,nr,cterm_mat);
  return res_t;
}

void printf_matrix(matrix M,int nr, int nc)
{
  cterm aux_t;
  aux_t=matrix2cterm(M,nr,nc);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

#endif
