/* $Id: cterm.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef CTERM_H
#define CTERM_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#include <stdio.h>
#include <ctype.h>

/* ---[ cterm OPERATORS ]----------------------------------------------- */

#define cterm_ERR_NONE 0    
#define cterm_ERR_UNKN_OP 2001
#define cterm_ERR_BAD_PARAMS_NUM 2002

extern coop op_NBS;
extern coop op_LIST;
extern coop op_NOINDENT;
extern coop op_BIND;
extern coop op_STRING;
extern coop op_HELPQUERY;
extern coop op_FALSE;
extern coop op_TRUE;

#define malloc_cterm() (cterm)mymalloc(sizeof(cterm_node),"cterm")
#define free_cterm(p) myfree(sizeof(cterm_node),p,"cterm")

#define malloc_cterm_list(size) \
  (cterm *)mymalloc(sizeof(cterm)*(size),"cterm_list")
#define free_cterm_list(size,p) myfree((size)*sizeof(cterm),p,"cterm_list")


/* error handling */

extern cterm error_param;
extern cterm error_context;
extern cterm cterm_unknown_context;

#define error_set_param(P) do {\
  cterm_free(error_param); \
  error_param=cterm_dup(P); } while(FALSE)

#define error_set_this_param(P) do {\
  cterm_free(error_param); \
  error_param=P; } while(FALSE)

#define error_set_context(P) do {\
  cterm_free(error_context); \
  error_context=cterm_dup(P); } while(FALSE)

/*------------------------------- ADT cterm ------------------------------------

  A cterm looks like a classic term. The only difference is that in a cterm you
  can have a NOT BETTER SPECIFIED (SUB)TERM, represented by a string
  (classic 'quoting' procedure).
  Here is a list of examples of cterms.

  assign(var("X"),sum("3","5"))
  not(equal(var("X"),sum(var("Y"),"34")))
  zero()

  In the previous examples assign,var,sum,not,equal are PREDEFINED coops.
  zero is a constant.
  "X","3","5","34","Y" are NOT BETTER SPECIFIED (NBS) cterms.

------------------------------------------------------------------------------*/

#if 0

/* cterm is implemented as a tree (classic implementation of classic terms) */


struct cterm_node_aux{
  coop op; /* the coop number (OP_NBS if it is an NBS) */
  char * string; /* a zero-arity coop not better specified */
  int arity;
  cterm_list params;
};

typedef struct cterm_node_aux cterm_node;

/* extern cterm cterm_EOF; */ /* declared in cterm.c */

#endif

/* CONSTRUCTORS of cterm */

void cterm_initialize(void);

cterm cterm_mk_table_entry(char * descr, cterm value);

cterm cterm_noindent(cterm t);
/* t -> NOINDENT(t) */

cterm cterm_init(void);
/* cterm_init() alloc memory for a cterm and returns the pointer to it */

cterm cterm_mk_fun_appl2(char * op,int arity,cterm * params);
cterm cterm_mk_fun_appl(coop op,int arity,cterm * params);

cterm cterm_mk(coop op,int arity,cterm * params,char * str);
/* cterm_mk(op_NBS,0,NULL,str) return an NBS cterm using the string str */
/* cterm_mk(op,0,NULL,NULL) with op<>op_NBS make the cterm op() (arity is 0) */
/* cterm_mk(op,n,params,NULL) with op<>op_NBS and params=(t1,...,tn)
  make the cterm op(t1,...,tn) */


#define cterm_mk_NBS(s) cterm_mk(op_NBS,0,NULL,s)
/* used to make a NBS-cterm */

#define cterm_mk_classic(op,arity,params) cterm_mk(op,arity,params,NULL)
/* used to make a "classic" term */

#define cterm_mk_const(c) cterm_mk(c,0,NULL,NULL)
/* constant cterms are very simple */


/* MACROS for cterm */

#define cterm_op(t) ((t)->op)
#define cterm_package(t) ((t)->pkg)
#define cterm_params(t) ((t)->params)
#define cterm_string(t) ((t)->string)
#define cterm_arity(t) ((t)->arity)
#define cterm_param(t,p) ((t)->params[p])

#define cterm_set_op(t,o) (t)->op=o
#define cterm_set_arity(t,n) (t)->arity=n
#define cterm_set_string(t,s) (t)->string=s
#define cterm_set_params(t,p) (t)->params=p

cterm cterm_dup(cterm t);
/* return a duplicate of the cterm t */

void cterm_free(cterm t);
/* cterm_free(t) free the memory allocated for the cterm t */

/* ---[ cterm converters ]----------------------------------------------- */

cterm int2cterm(int i);
cterm string2cterm(char * s);
cterm bool2cterm(coc_bool b);
cterm iv2cterm(ivector v);

#endif


