/* $Id: eee.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

int eee_h_dummy;

#ifdef EEE

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "coeffs.h"
#include "ctbuf.h"
#include "eee.h"
#include "ivectors.h"
#include "monomial.h"
#include "mlist.h"
#include "ordering.h"
#include "operator.h"
#include "options.h"
#include "pipe.h"
#include "polys.h"
#include "ring.h"
#include "scalars.h"
#include "terms.h"

void eee_adt_init(ring R)
{
  /* nothing to do, for now */
}


/* -----[ EEE ]-------------- */

eee eee_remove_lm(eee p, ring R)
{
  monomial aux_m;
#ifdef DEBUG
  if (p == NULL) {
    INTERNAL_ERROR("eee_remove_lm(p,R): p is NULL");
    return p;
  }
#endif
  OPTIMIZE("eee_remove_lm");
  aux_m=eee_leading_monomial(p,R);
  monomial_set_next(aux_m,NULL);
  return
   eee_sum_this_monomial_list_to_this_eee(monomial_opposite(aux_m,R),p,R);
}

eee eee_sum_this_monomial_list_to_this_eee(monomial m, eee p, ring R)
{
  monomial next_m;
  while (m != NULL){
    next_m = monomial_next(m);
    p=eee_add_this_monomial( p, m, R );
    m=next_m; }
  return p;
}

eee eee_this_x_coeff(eee p, coeff c, ring R)
{
  eee_this_x_coeff_l(&p,c,ring_indets_num(R),R);
  return p;
}

eee eee_randomize(eee p, ring R)
{
  monopoli aux_m,aux_m2;
  eee res_p;
  aux_m=eee2monopoli(p,R);
  aux_m2=monopoli_randomize(aux_m,R);
  res_p=monopoli2eee(aux_m2,R);
  monopoli_free(aux_m,R);
  monopoli_free(aux_m2,R);
  return res_p;
}

eee eee_this_minus_term_x_eee (eee p1, term ME, eee p2, ring R)
{
  TODO("eee_this_minus_term_x_eee");
  return NULL;
}

eee eee_this_plus_term_x_eee (eee p1, term ME, eee p2, ring R)
{
  TODO("eee_this_plus_term_x_eee");
  return NULL;
}

eee eee_this_minus_term_x_eee_truncated (eee p1, term ME, 
  eee p2, iv_elem Trunc, ring R)
{
  TODO("eee_this_minus_term_x_eee_truncated");
  return NULL;
}

eee eee_this_plus_term_x_eee_truncated (eee p1, term ME,
  eee p2, iv_elem Trunc, ring R)
{
  TODO("eee_this_plus_term_x_eee_truncated");
  return NULL;
}

eee this_coeff2eee(coeff c, ring R)
{
  if (coeff_is_zero(c,R)) {
    coeff_free(c,R);
    return NULL;
  }
  return this_monomial2eee(this_coeff2monomial(c,R),R);
}

monopoli this_eee2monopoli(eee p,ring R) 
{
  monopoli res_m;
  res_m=eee2monopoli(p,R);
  eee_free(p,R);
  return res_m;
}

ivector eee_max_mdeg(eee p, ring R)
{
  monopoli aux_m;
  ivector res_iv;
  OPTIMIZE("eee_max_mdeg");
  aux_m=eee2monopoli(p,R);
  res_iv=monopoli_max_mdeg(aux_m,R);
  monopoli_free(aux_m,R);
  return res_iv;
}

eee eee_sum_these(eee p1, eee p2, ring R)
{
  if (p1 == NULL) return p2;
  if (p2 == NULL) return p1;
  return eee_sum_these_l(p1,p2,ring_indets_num(R),R);
}

eee eee_sub_these(eee p1, eee p2, ring R)
{
  eee_change_sign(p2,R);
  return eee_sum_these(p1,p2,R);
}

coc_bool eee_eq(eee p1, eee p2, ring R)
{
  eee aux_p;
  OPTIMIZE("eee_eq");
  if (p1==NULL) return (p2==NULL);
  if (p2==NULL) return (p1==NULL);
  aux_p=eee_sub(p1,p2,R);
  if (aux_p == NULL) return TRUE;
  eee_free(aux_p,R);
  return FALSE;
}

coeff eee_leading_coeff(eee p, ring R)
{
  int n,i;
  monomial res_m;
  if (p == NULL) return coeff_zero(R);
  n=ring_indets_num(R);
  i=1;
  while (i<=n) {
    i+=eee_mon_lbl_len(p);
    p=eee_mon_cff(p);
  }
  return coeff_dup(p,R);
}

eee eee_x_monomial(eee p, monomial m, ring R)
{
  eee_lbl aux_l;
  eee res_p;
  if (p == NULL) return p;                                             
  if (iv_is_zero(monomial_term(m))) {    
    if (coeff_is_one(monomial_coeff(m),R)) return eee_dup(p,R);
    return eee_x_coeff_l(p,monomial_coeff(m),ring_indets_num(R),R);
  }                             
  aux_l=ivec_term2scalars(monomial_term(m),R);
  if (coeff_is_one(monomial_coeff(m),R))
    res_p=eee_shift(p,aux_l,ring_indets_num(R),R);
  else
    res_p=eee_shift_and_coeff(p,aux_l,monomial_coeff(m),ring_indets_num(R),R);
  eee_lbl_free_ck(aux_l);
  return res_p;
}

eee eee_this_x_monomial(eee p, monomial m, ring R)
{
  eee_lbl aux_l;
  if (p == NULL) return p;
  if (iv_is_zero(monomial_term(m))) {
    if (coeff_is_one(monomial_coeff(m),R)) return p;
    eee_this_x_coeff_l(&p,monomial_coeff(m),ring_indets_num(R),R);
    return p; }
  aux_l=ivec_term2scalars(monomial_term(m),R);
  if (coeff_is_one(monomial_coeff(m),R))
    eee_shift_this(&p,aux_l,ring_indets_num(R),R);
  else
    eee_shift_this_and_coeff(&p,aux_l,monomial_coeff(m),ring_indets_num(R),R);
  eee_lbl_free_ck(aux_l);
  return p;
}

eee eee_square(eee p, ring R)
{
  OPTIMIZE("eee_square");
  if (p==NULL) return NULL;
  return eee_mult_l(p,p,ring_indets_num(R),R);
}

olist eee2sppnum(eee p, ring R)
{
  monopoli aux_m;
  olist res_ol;
  OPTIMIZE("eee2sppnum");
  aux_m=eee2monopoli(p,R);
  res_ol=monopoli2sppnum(aux_m,R);
  /* monopoli_free(aux_m,R); */
  return res_ol;
}

int eee_deg(eee p, ring R)
{
  monopoli aux_m;
  int res_n;
  OPTIMIZE("eee_deg"); /* look at order; (*) avoid coefficients */
  /* (*) if the first row of order is all > 0 then Deg(P)=Deg(LT(P)) */
  aux_m=eee2monopoli(p,R);
  res_n=monopoli_max_deg(aux_m);
  monopoli_free(aux_m,R);
  return res_n;
}

coc_bool eee_is_homog(eee p, ring R)
{
  monopoli aux_m;
  coc_bool res_b;
  OPTIMIZE("eee_is_homog");
  aux_m=eee2monopoli(p,R);
  res_b=monopoli_is_homog(aux_m,R);
  monopoli_free(aux_m,R);
  return res_b;
}

int eee_is_homog_return_deg(eee p, ring R)
{
  monopoli aux_m;
  int res_n;
  OPTIMIZE("eee_is_homog_return_deg");
  aux_m=eee2monopoli(p,R);
  res_n=monopoli_is_homog_return_deg(aux_m,R);
  monopoli_free(aux_m,R);
  return res_n;
}

coc_bool eee_is_homog_deg(eee p, int hd, ring R)
{
  monopoli aux_m;
  coc_bool res_b;
  OPTIMIZE("eee_is_homog");
  aux_m=eee2monopoli(p,R);
  res_b=monopoli_is_homog_deg(aux_m,hd,R);
  monopoli_free(aux_m,R);
  return res_b;
}

eee eee_der(eee p, unknown x, ring R)
{
  monopoli aux_p,aux_p2;
  eee res_p;
  OPTIMIZE("eee_der");
  aux_p=eee2monopoli(p,R);
  aux_p2=monopoli_der(aux_p,x,R);
  res_p=this_monopoli2eee(aux_p2,R);
  monopoli_free(aux_p,R);
  return res_p;
}

term eee_leading_term(eee p, ring R)
{
  monomial aux_m;
  term res_t;
  OPTIMIZE("eee_leading_term");
  aux_m=eee_first(p,R);
  if (aux_m == NULL) return term_one(R);
  res_t=monomial_term(aux_m);
  coeff_free(monomial_coeff(aux_m),R);
  free_monomial(aux_m);
  return res_t;  
}

eee eee_map(eee p, ring R, ring S, ivector m)
{
  monopoli aux_p,aux_p2;
  eee res_p;
  OPTIMIZE("eee_map");
  aux_p=eee2monopoli(p,R);
  aux_p2=monopoli_map(aux_p,R,S,m);
  res_p=this_monopoli2eee(aux_p2,S);
  monopoli_free(aux_p,R);
  return res_p;
}

eee eee_mult(eee p1, eee p2, ring R)
{
  if (p1 == NULL) return NULL;
  if (p2 == NULL) return NULL;
  return eee_mult_l(p1,p2,ring_indets_num(R),R);
}

monomial eee_get_first(eee *p, ring R)
{
  eee last_p, aux_p; /* last poly such that it is not a monomial */
  eee_mon aux_m;
  eee_lbl aux_ss,l;
  int n,i,j, last_l;
  monomial res_m;
  if (p == NULL) return monomial_zero(R);
  n=ring_indets_num(R);
  aux_ss=eee_lbl_init(n);
  aux_m=*p;
  last_p=eee_NULL;
  last_l=n;
  i=1;
  while (i<=n) {
    l=eee_mon_lbl_tl(aux_m);
    eee_lbl_set_nth(aux_ss,i,eee_mon_lbl_hd(aux_m));
    for (j=1;j<=eee_lbl_len_ck(l);j++) 
      eee_lbl_set_nth(aux_ss,j+i,eee_lbl_nth(l,j));
    i+=eee_mon_lbl_len(aux_m);
    if (i<=n && eee_mon_next(eee_mon_cff(aux_m)) != eee_NULL) 
      { last_p=aux_m; last_l=n-i+1; }
    aux_m=eee_mon_cff(aux_m);
  }
  res_m=monomial_mk(aux_m,ivec_scalars2term(aux_ss,R));
  eee_lbl_free_ck(aux_ss);

  if (last_p==eee_NULL) {
      aux_p=*p;
      *p=eee_mon_next(*p); }
  else {
    aux_p=eee_mon_cff(last_p);
    eee_mon_set_cff(last_p,eee_mon_next(aux_p)); }

  i=last_l;
  while (i>0) {
    aux_m=eee_mon_cff(aux_p);
    i-=eee_mon_lbl_len(aux_p);
    eee_lbl_free_ck(eee_mon_lbl_tl(aux_p));
    free_eee_mon(aux_p);
    aux_p=aux_m;
  }
  return res_m;
}

coeff eee_this_leading_coeff(eee p, ring R)
{
  int n,i;
  monomial res_m;
  OPTIMIZE("eee_this_leading_coeff"); /* use the true leading coeff */
  if (p == NULL) return coeff_zero(R);
  n=ring_indets_num(R);
  i=1;
  while (i<=n) {
    i+=eee_mon_lbl_len(p);
    p=eee_mon_cff(p);
  }
  return p;
}

eee eee_add_this_monomial(eee p, monomial m, ring R)
{
  eee aux_p;
  aux_p=this_monomial2eee(m,R);
  return eee_sum_these(aux_p,p,R);
}

eee eee_sum_this_monopoli(monopoli m, ring R)
{
  eee res_p=NULL;
  monomial next_m;
  while (m != NULL){
    next_m = monomial_next(m);
    res_p=eee_add_this_monomial( res_p, m, R );
    m=next_m; }
  return res_p;
}

eee eee_sum_monopoli(monopoli m, ring R)
{
  eee res_p=NULL;
  monomial next_m;
  OPTIMIZE("eee_sum_monopoli");
  while (m != NULL) {
    next_m = monomial_next(m);
    res_p=eee_add_this_monomial( res_p, monomial_dup(m,R), R ); /* OPT here */
    m=next_m; }
  return res_p;
}

/* -----[ EEE_L ]-------------- */

eee eee_shift_and_coeff(eee p, eee_lbl ss, coeff c, int l, ring R)
{ /* assumed p <> NULL */
  int h,i,j,ll;
  eee_mon aux_m;
  eee_lbl lbl,res_lbl;
  if (l == 0) return coeff_mult(p,c,R);
  j=ring_indets_num(R)-l+1;
  h=eee_lbl_nth(ss,j);
  aux_m=eee_dummy(R,l);
  do {
    eee_mon_set_next(aux_m,malloc_eee_mon());
    aux_m=eee_mon_next(aux_m);
    eee_mon_set_lbl_hd(aux_m,eee_mon_lbl_hd(p)+h);
    if (eee_mon_lbl_tl(p)!=eee_lbl_NULL) {
      lbl=eee_mon_lbl_tl(p);
      ll=eee_lbl_len(lbl);
      res_lbl=eee_lbl_init(ll);
      for (i=1;i<=ll;i++)
        eee_lbl_set_nth(res_lbl,i,eee_lbl_nth(lbl,i)+eee_lbl_nth(ss,j+i));
      eee_mon_set_cff(aux_m,eee_shift_and_coeff(eee_mon_cff(p),ss,c,l-ll-1,R));
      eee_mon_set_lbl_tl(aux_m,res_lbl);
    } else {
      eee_mon_set_lbl_tl(aux_m,eee_lbl_NULL);
      eee_mon_set_cff(aux_m,eee_shift_and_coeff(eee_mon_cff(p),ss,c,l-1,R));
    }
    p=eee_mon_next(p);
  } while (p != NULL);
  eee_mon_set_next(aux_m,NULL);
  return eee_mon_next(eee_dummy(R,l));
}

eee eee_shift(eee p, eee_lbl ss, int l, ring R) 
{ /* assumed p <> NULL */
  int h,i,j,ll;
  eee_mon aux_m;
  eee_lbl lbl,res_lbl;
  if (l == 0) return coeff_dup(p,R);
  j=ring_indets_num(R)-l+1;
  h=eee_lbl_nth(ss,j);
  aux_m=eee_dummy(R,l);
  do {
    eee_mon_set_next(aux_m,malloc_eee_mon());
    aux_m=eee_mon_next(aux_m);
    eee_mon_set_lbl_hd(aux_m,eee_mon_lbl_hd(p)+h);
    if (eee_mon_lbl_tl(p)!=eee_lbl_NULL) {
      lbl=eee_mon_lbl_tl(p);
      ll=eee_lbl_len(lbl);
      res_lbl=eee_lbl_init(ll);
      for (i=1;i<=ll;i++)
        eee_lbl_set_nth(res_lbl,i,eee_lbl_nth(lbl,i)+eee_lbl_nth(ss,j+i));
      eee_mon_set_cff(aux_m,eee_shift(eee_mon_cff(p),ss,l-ll-1,R));
      eee_mon_set_lbl_tl(aux_m,res_lbl);
    } else {
      eee_mon_set_lbl_tl(aux_m,eee_lbl_NULL);
      eee_mon_set_cff(aux_m,eee_shift(eee_mon_cff(p),ss,l-1,R));
    }
    p=eee_mon_next(p);
  } while (p != NULL);
  eee_mon_set_next(aux_m,NULL);
  return eee_mon_next(eee_dummy(R,l));
}

eee eee_x_coeff_l(eee p, coeff c, int l, ring R)
{ /* assumed p <> NULL */
  int i,j,ll;
  eee_lbl lbl;
  eee_mon aux_m;
  if (l==0) return coeff_mult(p,c,R);
  aux_m=eee_dummy(R,l);
  do {
    eee_mon_set_next(aux_m,eee_mon_mk(eee_mon_lbl_hd(p),
      eee_lbl_dup_ck(eee_mon_lbl_tl(p)),
      eee_x_coeff_l(eee_mon_cff(p),c,l-eee_mon_lbl_len(p),R)));
    p=eee_mon_next(p);
    aux_m=eee_mon_next(aux_m);
  } while (p != eee_NULL);
  return eee_mon_next(eee_dummy(R,l));
}

void eee_shift_this_and_coeff(eee * p, eee_lbl ss, coeff c, int l, ring R)
{
  int i,j,ll;
  eee aux_p;
  eee_lbl lbl;
  if (l==0) { coeff_mult_this(*p,c,R); return; }
  aux_p=*p;
  j=ring_indets_num(R)-l+1;
  while (aux_p != eee_NULL) {
    eee_mon_set_lbl_hd(aux_p,eee_mon_lbl_hd(aux_p)+eee_lbl_nth(ss,j));
    lbl=eee_mon_lbl_tl(aux_p);
    if (lbl!=eee_lbl_NULL) {
      ll=eee_lbl_len(lbl); /* <-- */
      for (i=1;i<=ll;i++)
        eee_lbl_set_nth(lbl,i,eee_lbl_nth(lbl,i)+eee_lbl_nth(ss,j+i));
      eee_shift_this_and_coeff(&eee_mon_cff(aux_p),ss,c,l-ll-1,R); }
    else eee_shift_this_and_coeff(&eee_mon_cff(aux_p),ss,c,l-1,R);
    aux_p=eee_mon_next(aux_p);
  }
}

void eee_shift_this(eee * p, eee_lbl ss, int l, ring R)
{
  int i,j,ll;
  eee aux_p;
  eee_lbl lbl;
  if (l==0) return;
  aux_p=*p;
  while (aux_p != eee_NULL) {
    j=ring_indets_num(R)-l+1;
    eee_mon_set_lbl_hd(aux_p,eee_mon_lbl_hd(aux_p)+eee_lbl_nth(ss,j));
    lbl=eee_mon_lbl_tl(aux_p);
    if (lbl!=eee_lbl_NULL) {
      ll=eee_lbl_len_ck(lbl);
      for (i=1;i<=ll;i++)
        eee_lbl_set_nth(lbl,i,eee_lbl_nth(lbl,i)+eee_lbl_nth(ss,j+i));
      eee_shift_this(&eee_mon_cff(aux_p),ss,l-ll-1,R); }
    else eee_shift_this(&eee_mon_cff(aux_p),ss,l-1,R);
    aux_p=eee_mon_next(aux_p);
  }
}

void eee_this_x_coeff_l(eee * p, coeff c, int l, ring R)
{
  int i,j,ll;
  eee aux_p;
  eee_lbl lbl;
  if (l==0) { coeff_mult_this(*p,c,R); return; }
  aux_p=*p;
  while (aux_p != eee_NULL) {
    eee_this_x_coeff_l(&eee_mon_cff(aux_p),c,l-eee_mon_lbl_len(aux_p),R);
    aux_p=eee_mon_next(aux_p);
  }
}

cterm eee2cterm_graph(eee p, int l, ring R)
{
  eee_mon aux_m;
  ctbuf aux_b;
  cterm aux_t;
  if (opt_get_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP)==optv_DEBUG_EEE_PP_GRAPH2
    && ring_level(R)==l) return int2cterm(eee_num_leaves(p,l,R));
  if (l==0) return coeff2cterm(p,R);
  aux_m=p;
  aux_b=ctbuf_empty();
  while (aux_m != eee_NULL) {
    aux_t=eee_mon2cterm(aux_m,l,R);
    ctbuf_append(aux_b,aux_t);
    aux_m=eee_mon_next(aux_m);
  }
  return ctbuf_mk_cterm(aux_b,op_EEE);
}

cterm eee2cterm_normal(eee p, int l, ring R)
{
  monopoli aux_m;
  cterm res_t;
  ctbuf aux_b;
  aux_m=eee2monopoli_l(p,l,R);
  aux_b=monopoli2ctbuf(aux_m,R);
  res_t=ctbuf_mk_cterm(aux_b,op_POLYNOMIAL);
  monopoli_free(aux_m,R);
  return res_t;
}

cterm eee2cterm_l(eee p, int l, ring R)
{
#ifdef DEBUG
  if (opt_get_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP)!=optv_DEBUG_EEE_PP_NORMAL)
    return eee2cterm_graph(p,l,R);
  else
#endif
    return eee2cterm_normal(p,l,R);
}

int eee_num_leaves(eee p, int l, ring R)
{
  int res_n;
  if (l==0) return 1;
  res_n=0;
  while (p!=eee_NULL) {
    res_n+=eee_num_leaves(eee_mon_cff(p),l-eee_mon_lbl_len(p),R);
    p=eee_mon_next(p); }
  return res_n;
}

int eee_size_l(eee p, int l, ring R)
{
  int res_n,n;
  if (l==0) return (coeff_size(p,R)-4);
  res_n=0;
  while (p!=eee_NULL) {
    n=eee_mon_lbl_len(p);
    if (n==1)
      res_n+=eee_size_l(eee_mon_cff(p),l-n,R)+sizeof(struct eee_mon_aux);
    else 
      res_n+=eee_size_l(eee_mon_cff(p),l-n,R)+n*sizeof(ivec_elem)
        +sizeof(struct eee_mon_aux);
    p=eee_mon_next(p); }
  return res_n;
}

#ifndef EEE_KEEP_OPT

eee eee_sum_these_l(eee p1, eee p2, int l, ring R)
{ /* NOTE: assumed p1<>0 and p2<>0 */
  eee_mon aux_m, aux_m1;
  eee res_p, aux_p1, aux_p2;
  coeff res_c;

  if (l==0) {
    res_c=coeff_sum_these(p1,p2,R);
    if (coeff_is_zero(res_c,R)) {
      coeff_free(res_c,R);
      return eee_NULL; }
    else return res_c; }

  if (l==ring_level(R)) return eee_monopoli_sum_these(p1,p2,R);

  aux_m=malloc_eee_mon();
  res_p=aux_m;

  while (p1!=eee_NULL && p2!=eee_NULL) {
    if (eee_mon_lbl_hd(p1)>eee_mon_lbl_hd(p2)) {
      eee_mon_set_next(aux_m,p1);
      p1=eee_mon_next(p1);
      aux_m=eee_mon_next(aux_m); }
    else if (eee_mon_lbl_hd(p1)<eee_mon_lbl_hd(p2)) {
      eee_mon_set_next(aux_m,p2);
      p2=eee_mon_next(p2);
      aux_m=eee_mon_next(aux_m); }
    else {
      aux_p1=eee_mon_next(p1);
      aux_p2=eee_mon_next(p2);
      aux_m1=eee_mon_sum_these(p1,p2,l,R);
      if (aux_m1!=eee_mon_NULL) {
        eee_mon_set_next(aux_m,aux_m1);
        aux_m=eee_mon_next(aux_m); }
      p1=aux_p1;
      p2=aux_p2; } }

  if (p1==eee_NULL) eee_mon_set_next(aux_m,p2);
  else eee_mon_set_next(aux_m,p1);
  aux_m=res_p;
  res_p=eee_mon_next(res_p);
  free_eee_mon(aux_m);

  return res_p;
}

#else

eee eee_sum_these_l(eee p1, eee p2, int l, ring R)
{ /* NOTE: assumed p1<>0 and p2<>0 */
  eee_mon aux_m, aux_m1;
  eee aux_p1, aux_p2;
  coeff res_c;

  if (l==ring_level(R)) return eee_monopoli_sum_these(p1,p2,R);

#if 0 /* TODO: check if it is a good idea to treat this as a special case */
  if (eee_is_mon(p1)) {
    if (eee_is_mon(p2)) {
      if (eee_mon_lbl_hd(p1) < eee_mon_lbl_hd(p2)) { 
        eee_mon_set_next(p2,p1);
        return p2;
      }
      else if (eee_mon_lbl_hd(p1) > eee_mon_lbl_hd(p2)) {
        eee_mon_set_next(p1,p2);
        return p1;
      }
      else {
        return eee_mon_sum_these(p1,p2,l,R); 
      }
    }
  }
#endif

  aux_m=eee_dummy(R,l);

  do {
    if (eee_mon_lbl_hd(p1)>eee_mon_lbl_hd(p2)) {
      eee_mon_set_next(aux_m,p1);
      p1=eee_mon_next(p1);
      aux_m=eee_mon_next(aux_m); }
    else if (eee_mon_lbl_hd(p1)<eee_mon_lbl_hd(p2)) {
      eee_mon_set_next(aux_m,p2);
      p2=eee_mon_next(p2);
      aux_m=eee_mon_next(aux_m); }
    else {
      aux_p1=eee_mon_next(p1);
      aux_p2=eee_mon_next(p2);
      aux_m1=eee_mon_sum_these(p1,p2,l,R);
      if (aux_m1!=eee_mon_NULL) {
        eee_mon_set_next(aux_m,aux_m1);
        aux_m=eee_mon_next(aux_m); }
      p1=aux_p1;
      p2=aux_p2;
    }
  } while (p1!=eee_NULL && p2!=eee_NULL);

  if (p1==eee_NULL) eee_mon_set_next(aux_m,p2);
  else eee_mon_set_next(aux_m,p1);

  return eee_mon_next(eee_dummy(R,l));
}

#endif

eee eee_mult_l(eee p1, eee p2, int l, ring R)
{ /* optimization */
  eee res_p,save_res_p,aux_p,res_p_prec;

  if (l==0) return coeff_mult(p1,p2,R);

  if (l==ring_level(R)) return eee_monopoli_mult(p1,p2,R);

  if (eee_is_mon(p2)) return eee_x_eee_mon(p1,p2,l,R);

  save_res_p=res_p=eee_mon_x_eee(p1,p2,l,R);
  p1=eee_mon_next(p1);

  while (p1 != eee_NULL) {
    aux_p=eee_mon_x_eee(p1,p2,l,R);
    while (eee_mon_next(res_p) != eee_mon_NULL && 
        eee_mon_lbl_hd_gt(res_p,aux_p)) {
      res_p_prec=res_p;
      res_p=eee_mon_next(res_p);
      /* printf("."); */ 
    }
    /* printf("\n"); */
    eee_mon_set_next(res_p_prec,eee_sum_these_l(res_p,aux_p,l,R));
    res_p=eee_mon_next(res_p_prec);
    p1=eee_mon_next(p1);
  }
  return save_res_p;
}

monopoli eee2monopoli_l(eee p, int l, ring R)
{
  monopoli res_m, aux_m, aux_m1;
  if (p==NULL) return NULL;
  p=eee_dup_l(p,l,R);
  res_m=eee_get_first(&p,R);
  aux_m=res_m;
  while (p != NULL) {
    monomial_set_next(aux_m, eee_get_first(&p,R));
    aux_m=monomial_next(aux_m);
  } while (p != NULL);
  OPTIMIZE("eee2monopoli_l");
  return res_m;
}

eee eee_dup_l(eee p, int l, ring R)
{
  eee res_p, aux_p;
  if (l==0) return coeff_dup(p,R);

  res_p=aux_p=eee_mon_dup(p,l,R);
  p=eee_mon_next(p);
  while (p != eee_NULL) {
    eee_mon_set_next(aux_p,eee_mon_dup(p,l,R));
    p=eee_mon_next(p);
    aux_p=eee_mon_next(aux_p); }
  return res_p;
}

coc_bool eee_eq_l(eee p1, eee p2, int l, ring R)
{ /* TODO: WARNING NON UNIQUE REPRESENTATION */ /* first optimize */
  if (l==0) return coeff_eq(p1,p2,R);
  while (p1 != eee_NULL && p2 != eee_NULL) {
    if (!eee_mon_eq(p1,p2,l,R)) return FALSE;
    p1=eee_mon_next(p1);
    p2=eee_mon_next(p2);
  }
  return p1==eee_NULL && p2==eee_NULL;
}

void eee_free_l(eee p, int l, ring R)
{
  if (l==0) 
    (R)->ClearCoeff((coeff *)&p);
    /* coeff_free((coeff)p,R); */
  else {
    eee_mon aux_m;
    while (p != eee_NULL) {
      aux_m=p;
      p=eee_mon_next(p);
      /* eee_mon_free(aux_m,l,R); */ /* now is the following 3 lines (TODO) */
      eee_free_l(eee_mon_cff(aux_m),l-eee_mon_lbl_len(aux_m),R);
      eee_lbl_free_ck(eee_mon_lbl_tl(aux_m));
      free_eee_mon(aux_m); } }
}

void eee_change_sign_l(eee * p, int l, ring R)
{
  if (l==0) coeff_change_sign((coeff *)p,R);
  else {
    eee_mon aux_m;
    aux_m=*p;
    while(aux_m!=eee_mon_NULL) {
      eee_change_sign_l(&eee_mon_cff(aux_m),l-eee_mon_lbl_len(aux_m),R);
      aux_m=eee_mon_next(aux_m); } }
}

/* -----[ EEE_MON ]-------------- */

cterm eee_mon2cterm(eee_mon m, int l, ring R)
{
  cterm params[2];
#ifdef DEBUG
  if (m==eee_mon_NULL) ca_message_string("INTERNAL ERROR: eee_mon2cterm");
#endif
  params[0]=eee_mon_lbl2cterm(m);
  params[1]=eee2cterm_l(eee_mon_cff(m),l-eee_mon_lbl_len(m),R);
  return cterm_mk_classic(op_EEE_MON,2,params);
}

eee_mon eee_mon_mk(expType hd, eee_lbl tl, eee p)
{
  eee_mon res_m;
  res_m=malloc_eee_mon();
  eee_mon_set_lbl_hd(res_m,hd);
  eee_mon_set_lbl_tl(res_m,tl);
  eee_mon_set_cff(res_m,p);
  eee_mon_set_next(res_m,eee_mon_NULL);
  return res_m;
}

eee eee_mon_x_eee(eee_mon m, eee p, int l, ring R)
{
  eee save_res_p,res_p;

  save_res_p=res_p=eee_mon_mult(m,p,l,R);
  p=eee_mon_next(p);

  while (p != eee_NULL) {
    eee_mon_set_next(res_p,eee_mon_mult(m,p,l,R));
    res_p=eee_mon_next(res_p);
    p=eee_mon_next(p);
  }

  return save_res_p;
}

eee eee_x_eee_mon(eee p, eee_mon m, int l, ring R)
{
  eee save_res_p,res_p;

  save_res_p=res_p=eee_mon_mult(p,m,l,R);
  p=eee_mon_next(p);

  while (p != eee_NULL) {
    eee_mon_set_next(res_p,eee_mon_mult(p,m,l,R));
    res_p=eee_mon_next(res_p);
    p=eee_mon_next(p);
  }

  return save_res_p;
}

/* -----[ EEE_LBL ]-------------- */

cterm eee_mon_lbl2cterm(eee_mon m)
{
  cterm_list params;
  cterm res_t;
  eee_lbl aux_l;
  int i,n;
  n=eee_mon_lbl_len(m);
  params=malloc_cterm_list(n);
#ifdef DEBUG
    if (opt_get_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP) ==
      optv_DEBUG_EEE_PP_GRAPH2)
        params[0]=int2cterm(0);
    else
#endif
      params[0]=int2cterm(eee_mon_lbl_hd(m));
  aux_l=eee_mon_lbl_tl(m);
  for (i=1;i<=eee_lbl_len_ck(aux_l);i++)
#ifdef DEBUG
    if (opt_get_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP) ==
      optv_DEBUG_EEE_PP_GRAPH2)
        params[i]=int2cterm(0); 
    else
#endif
      params[i]=int2cterm(eee_lbl_nth(aux_l,i));
  res_t=cterm_mk_classic(op_EEE_LBL,n,params);
  free_cterm_list(n,params);
  return res_t;
}

/* -----[ EEE_MONOPOLI ]-------------- */

eee eee_monopoli_mult(eee p1, eee p2, ring R)
{ /* optimization */
  eee res_p,save_res_p,aux_p,res_p_prec;

  if (eee_is_mon(p2)) return eee_monopoli_mon_x_monopoli(p2,p1,R);

  save_res_p=res_p=eee_monopoli_mon_x_monopoli(p1,p2,R);
  p1=eee_mon_next(p1);

  while (p1 != eee_NULL) {
    aux_p=eee_monopoli_mon_x_monopoli(p1,p2,R);
    while (eee_mon_next(res_p) != eee_mon_NULL && 
        eee_mon_lbl_gt(res_p,aux_p)) {
      res_p_prec=res_p;
      res_p=eee_mon_next(res_p);
    } 
    eee_mon_set_next(res_p_prec,eee_monopoli_sum_these(res_p,aux_p,R));
    res_p=eee_mon_next(res_p_prec);
    p1=eee_mon_next(p1);
  }
  return save_res_p;
}

eee eee_monopoli_mon_x_monopoli(eee_mon m, eee p, ring R)
{
  eee res_p,save_res_p;

  save_res_p=res_p=eee_monopoli_mon_mult(m,p,R);
  p=eee_mon_next(p);

  while (p != eee_NULL) {
    eee_mon_set_next(res_p,eee_monopoli_mon_mult(m,p,R));
    res_p=eee_mon_next(res_p);
    p=eee_mon_next(p);
  }

  return save_res_p;
}

eee_mon eee_monopoli_mon_mult(eee_mon m1, eee_mon m2, ring R)
{
  return eee_mon_mk(eee_mon_lbl_hd(m1)+eee_mon_lbl_hd(m2),
    eee_lbl_sum_ck(eee_mon_lbl_tl(m1),eee_mon_lbl_tl(m2)),
    coeff_mult(eee_mon_cff(m1),eee_mon_cff(m2),R));
}

eee eee_monopoli_sum_these(eee p1, eee p2, ring R)
{ /* assumed p1<>0 and p2<>0 */
  eee_mon aux_m, aux_m1;
  eee aux_p1, aux_p2;

  aux_m=eee_dummy(R,0);

  do {
    switch (eee_mon_lbl_which_gt(p1,p2)) {
      case ANSWER_BOTH:
        aux_p1=eee_mon_next(p1);
        aux_p2=eee_mon_next(p2);
        aux_m1=eee_monopoli_mon_sum_these(p1,p2,R);
        if (aux_m1!=eee_mon_NULL) {
          eee_mon_set_next(aux_m,aux_m1);
          aux_m=eee_mon_next(aux_m); }
        p1=aux_p1;
        p2=aux_p2;
        break;
      case ANSWER_FIRST:
        eee_mon_set_next(aux_m,p1);
        p1=eee_mon_next(p1);
        aux_m=eee_mon_next(aux_m);
        break;
      case ANSWER_SECOND:
        eee_mon_set_next(aux_m,p2);
        p2=eee_mon_next(p2);
        aux_m=eee_mon_next(aux_m);
        break;
    }
  } while (p1!=eee_NULL && p2!=eee_NULL);
  
  if (p1==eee_NULL) eee_mon_set_next(aux_m,p2);
  else eee_mon_set_next(aux_m,p1);
  return eee_mon_next(eee_dummy(R,0));
}

eee_mon eee_monopoli_mon_sum_these(eee_mon m1, eee_mon m2, ring R)
{
  coeff aux_c;
  aux_c=coeff_sum_these(eee_mon_cff(m1),eee_mon_cff(m2),R);

  eee_lbl_free_ck(eee_mon_lbl_tl(m2));
  free_eee_mon(m2);

  if (coeff_is_zero(aux_c,R)) {
    coeff_free(aux_c,R);
    eee_lbl_free_ck(eee_mon_lbl_tl(m1));
    free_eee_mon(m1);
    return eee_mon_NULL;
  }

  eee_mon_set_cff(m1,aux_c);
  return m1;
}

monomial eee_first(eee p, ring R)
{
  eee_lbl aux_ss,l;
  int n,i,j;
  monomial res_m;
  if (p == NULL) return NULL;
  n=ring_indets_num(R);
  aux_ss=eee_lbl_init(n);
  i=1;
  while (i<=n) {
    l=eee_mon_lbl_tl(p);
    eee_lbl_set_nth(aux_ss,i,eee_mon_lbl_hd(p));
    for (j=1;j<=eee_lbl_len_ck(l);j++) 
      eee_lbl_set_nth(aux_ss,j+i,eee_lbl_nth(l,j));
    i+=eee_mon_lbl_len(p);
    p=eee_mon_cff(p);
  }
  res_m=monomial_mk(coeff_dup(p,R),ivec_scalars2term(aux_ss,R));
  eee_lbl_free_ck(aux_ss);
  return res_m;
}

eee_mon this_monomial2eee_mon(monomial m, ring R)
{ 
  eee_lbl aux_ss;
  eee_mon res_m;
  /* OPTIMIZE("this_monomial2eee_mon"); */ /* almost ok */
  aux_ss=ivec_term2scalars(monomial_term(m),R);
  res_m=eee_mon_mk_special(aux_ss,monomial_coeff(m),R);
  term_free(monomial_term(m));
  free_monomial(m);
  return res_m;
}

eee_mon eee_mon_mk_special(eee_lbl l, eee p, ring R)
{
  eee_mon aux_m,res_m;
  eee_lbl aux_l, aux_l1, aux_l2;
  int i,lev,n;

  aux_l=eee_lbl_dup_from(l,1);
  lev=ring_level(R);
  n=ring_indets_num(R);
  if (lev!=0 && lev!=n) {
    aux_l1=eee_lbl_dup_to_ck(aux_l,n-lev-1);
    aux_l2=eee_lbl_dup_from(aux_l,n-lev);
    aux_m=eee_mon_mk(eee_lbl_nth(aux_l,n-lev),aux_l2,p);
    res_m=eee_mon_mk(eee_lbl_nth(l,1),aux_l1,aux_m);
    eee_lbl_free_ck(aux_l);
  }
  else 
    res_m=eee_mon_mk(eee_lbl_nth(l,1),aux_l,p);
  eee_lbl_free_ck(l);
  return res_m;
}

eee_mon eee_mon_mult(eee_mon m1, eee_mon m2, int l, ring R)
{
  eee_mon res_m,aux_m;
  eee_lbl l1, l2, aux_l;
  int n1,n2;
  l1=eee_mon_lbl_tl(m1);
  l2=eee_mon_lbl_tl(m2);
  n1=eee_lbl_len_ck(l1);
  n2=eee_lbl_len_ck(l2);

  if (n1==n2) 
    res_m=eee_mon_mk(eee_mon_lbl_hd(m1)+eee_mon_lbl_hd(m2),eee_lbl_sum_ck(l1,l2),
      eee_mult_l(eee_mon_cff(m1),eee_mon_cff(m2),l-n1-1,R));
  else if (n1>n2) {
    aux_l=eee_lbl_dup_from(l1,n2+1);
    aux_m=eee_mon_mk(eee_lbl_nth(l1,n2+1),aux_l,eee_mon_cff(m1));
    l1=eee_lbl_dup_to_ck(l1,n2);
    res_m=eee_mon_mk(eee_mon_lbl_hd(m1)+eee_mon_lbl_hd(m2),
      eee_lbl_sum_ck(l1,l2), eee_mon_x_eee(aux_m,eee_mon_cff(m2),l-n2-1,R));
    eee_lbl_free_ck(l1);
    eee_lbl_free_ck(aux_l);
    free_eee_mon(aux_m); } /* OPTIMIZE (sum this to this) */
  else { /* n1<n2 */
    aux_l=eee_lbl_dup_from(l2,n1+1);
    aux_m=eee_mon_mk(eee_lbl_nth(l2,n1+1),aux_l,eee_mon_cff(m2));
    l2=eee_lbl_dup_to_ck(l2,n1);
    res_m=eee_mon_mk(eee_mon_lbl_hd(m1)+eee_mon_lbl_hd(m2),
      eee_lbl_sum_ck(l1,l2), eee_x_eee_mon(eee_mon_cff(m1),aux_m,l-n1-1,R));
    eee_lbl_free_ck(l2);
    eee_lbl_free_ck(aux_l);
    free_eee_mon(aux_m); } /* OPTIMIZE (sum this to this) */
  return res_m;
}

eee_mon eee_mon_sum_these(eee_mon m1, eee_mon m2, int l, ring R)
{
  eee_lbl l1, l2, a,b,c; /* m1 = [p]h.l1, m2 = [q]h.l2 */
  int i,n1,n2,n;
  l1=eee_mon_lbl_tl(m1); /* l1=a.b, l2=a.c, a=MaxCommonFirst(l1,l2) */
  l2=eee_mon_lbl_tl(m2);

  if (l1 == NULL) {
    l--;
    if (l2 == NULL) {
      if (l==0) {
        eee_mon_set_cff(m1,
          coeff_sum_these_free_if_zero(eee_mon_cff(m1),eee_mon_cff(m2),R));
        if (eee_mon_cff(m1) == coeff_NULL) {
          free_eee_mon(m1);
          m1=eee_mon_NULL;
        }
      } else { /* l > 0 */
        eee_mon_set_cff(m1,
          eee_sum_these_l(eee_mon_cff(m1),eee_mon_cff(m2),l,R));
        if (eee_mon_cff(m1)==eee_NULL) {
          free_eee_mon(m1);
          m1=eee_mon_NULL;
        } else if (eee_is_mon(eee_mon_cff(m1)) && l != ring_level(R)) {
          eee_mon_lift_this(m1);                  
        }
      }
      free_eee_mon(m2); 
    } else { /* l1 = NULL, l2 <> NULL */
      eee_mon_set_lbl_hd(m2,eee_lbl_nth(l2,1));
      eee_mon_set_lbl_tl(m2,eee_lbl_dup_from(l2,1));
      eee_lbl_free(l2);

      eee_mon_set_next(m2,eee_mon_NULL); /* m2 is now a poly */
      eee_mon_set_cff(m1,eee_sum_these_l(eee_mon_cff(m1),m2,l,R));
      if (eee_mon_cff(m1)==eee_NULL) {
        free_eee_mon(m1);
        return eee_mon_NULL;
      } else if (eee_is_mon(eee_mon_cff(m1)) && l != ring_level(R)) {
        eee_mon_lift_this(m1);                  
      }  
    }
    return m1;
  }
  if (l2 == NULL) {
    l--;
    eee_mon_set_lbl_hd(m1,eee_lbl_nth(l1,1));
    eee_mon_set_lbl_tl(m1,eee_lbl_dup_from(l1,1));
    eee_lbl_free(l1);

    eee_mon_set_next(m1,eee_mon_NULL); /* m1 is now a poly */
    eee_mon_set_cff(m2,eee_sum_these_l(m1,eee_mon_cff(m2),l,R));
    if (eee_mon_cff(m2)==eee_NULL) {
      free_eee_mon(m2);
      m2=eee_mon_NULL;
    } else if (eee_is_mon(eee_mon_cff(m2)) && l != ring_level(R)) {
      eee_mon_lift_this(m2);                  
    }
    return m2;
  }
  
  n1=eee_lbl_len(l1);
  n2=eee_lbl_len(l2);

  if (n1 < n2) {
    for (i=1;i<=n1;i++) {
      if (eee_lbl_nth(l1,i)!=eee_lbl_nth(l2,i)) break;
    }
  } else {
    for (i=1;i<=n2;i++) {
      if (eee_lbl_nth(l1,i)!=eee_lbl_nth(l2,i)) break;
    }
  } 
 
  /* i is the index of the first discrepancy (>=2) */

  if (i-1==n1) { /* l1 = a, l2 = a.? */

    if (n1==n2) { /* l1 = l2 = a => m=[p+q]h.a */
      if (l==i) {
        eee_mon_set_cff(m1,
          coeff_sum_these_free_if_zero(eee_mon_cff(m1),eee_mon_cff(m2),R));
        if (eee_mon_cff(m1) == coeff_NULL) {
          eee_lbl_free(l1);
          free_eee_mon(m1);
          m1=eee_mon_NULL;
        }
      } else { /* l <> i */
        eee_mon_set_cff(m1,
          eee_sum_these_l(eee_mon_cff(m1),eee_mon_cff(m2), l-i,R));
        if (eee_mon_cff(m1)==eee_NULL) {
          eee_lbl_free(l1);
          free_eee_mon(m1);
          m1=eee_mon_NULL;
        } else if (eee_is_mon(eee_mon_cff(m1)) && l-i != ring_level(R)) {
          eee_mon_lift_this(m1);                  
        }
      }
      eee_lbl_free(l2);
      free_eee_mon(m2); 
  
    } else { /* l1 = a, l2 = a.c => m=[p+[q]c]h.a */

      c=eee_lbl_dup_from(l2,i);
      eee_mon_set_lbl_hd(m2,eee_lbl_nth(l2,i));
      eee_mon_set_lbl_tl(m2,c);

      eee_lbl_free(l2);
      eee_mon_set_next(m2,eee_mon_NULL); /* m2 is now a poly */
      eee_mon_set_cff(m1,eee_sum_these_l(eee_mon_cff(m1),m2,l-i,R));
      if (eee_mon_cff(m1)==eee_NULL) {
        free_eee_mon(m1);
        eee_lbl_free(l1);
        m1=eee_mon_NULL;
      } else if (eee_is_mon(eee_mon_cff(m1)) && l-i != ring_level(R)) {
        eee_mon_lift_this(m1);                  
      }
    }
    return m1;

  } else if (i-1==n2) {  /* l1 = a.b, l2 = a => m=[[p]b+q]h.a */

      b=eee_lbl_dup_from(l1,i);
      eee_mon_set_lbl_hd(m1,eee_lbl_nth(l1,i));
      eee_mon_set_lbl_tl(m1,b);
      eee_lbl_free(l1);

      eee_mon_set_next(m1,eee_mon_NULL); /* m1 is now a poly */
      eee_mon_set_cff(m2,eee_sum_these_l(m1,eee_mon_cff(m2),l-i,R));
      if (eee_mon_cff(m2)==eee_NULL) {
        free_eee_mon(m2);
        eee_lbl_free(l2);
        return eee_mon_NULL;
      } else if (eee_is_mon(eee_mon_cff(m2)) && l-i != ring_level(R)) {
        eee_mon_lift_this(m2);                  
      }
      return m2;
    }

  else {  /* l1 = a.b, l2 = a.c => m=[[p]b+[q]c]h.a */
    /* NOTE: in this case the sum is made "in place".
     * moreover we know that the result will be a eee-binomial
     * so we have not to check if it is optimized 
     */

    expType h;

    /* printf("S"); */
    a=eee_lbl_dup_to_ck(l1,i-1);
    b=eee_lbl_dup_from(l1,i);
    c=eee_lbl_dup_from(l2,i);
    h=eee_mon_lbl_hd(m1);
    eee_mon_set_lbl_hd(m1,eee_lbl_nth(l1,i));
    eee_mon_set_lbl_hd(m2,eee_lbl_nth(l2,i));
    eee_lbl_free(l1);
    eee_lbl_free(l2);

    eee_mon_set_lbl_tl(m1,b);
    eee_mon_set_lbl_tl(m2,c);
    if (eee_mon_lbl_hd_gt(m1,m2)) {
      eee_mon_set_next(m1,m2);
      eee_mon_set_next(m2,eee_mon_NULL);
      m1=eee_mon_mk(h,a,m1); } /* in m1 we put the result */
    else {
      eee_mon_set_next(m2,m1);
      eee_mon_set_next(m1,eee_mon_NULL);
      m1=eee_mon_mk(h,a,m2); } /* in m1 we put the result */

  }

  return m1;
}

coc_bool eee_mon_eq(eee_mon m1, eee_mon m2, int l, ring R)
{
  TODO("eee_mon_eq");
  /* ATTENZIONE! la rappresentazione non e' unica (se non sono ottimizzati) */
}

coc_bool eee_mon_is_const(eee_mon m, int l, ring R)
{
  while (l != 0) {
    if (eee_mon_next(m) != eee_NULL) return FALSE;
    if (eee_mon_lbl_hd(m) != 0) return FALSE;
    if (!eee_lbl_is_zero_ck(eee_mon_lbl_tl(m))) return FALSE;
    l-=eee_mon_lbl_len(m);
    m=eee_mon_cff(m); }
  return TRUE;
}

void eee_change_order(eee * p, ring R1, ring R2) 
{
  monomial aux_m,next_m;
  if ((*p) != NULL) {
    eee res_p;
    aux_m=eee2monopoli(*p,R1);
    eee_free(*p,R1);
    res_p=NULL;
    while (aux_m != NULL) {
      next_m=monomial_next(aux_m);
      eee_add_this_monomial_unique(aux_m,&res_p,R2);
      aux_m=next_m; }
    *p=res_p; }
}

void eee_mon_lift_this(eee_mon m)
{
  eee_lbl res_l,aux_l1,aux_l2;
  eee_mon aux_m;
  int n,i;
  aux_m=eee_mon_cff(m);
  aux_l1=eee_mon_lbl_tl(m);
  aux_l2=eee_mon_lbl_tl(aux_m);
  n=eee_lbl_len_ck(aux_l1);
  res_l=eee_lbl_init(n+eee_lbl_len_ck(aux_l2)+1);
  for (i=n;i>0;i--)
    eee_lbl_set_nth(res_l,i,eee_lbl_nth(aux_l1,i));
  eee_lbl_set_nth(res_l,n+1,eee_mon_lbl_hd(aux_m));
  n++;
  for (i=eee_lbl_len_ck(aux_l2);i>0;i--)
    eee_lbl_set_nth(res_l,i+n,eee_lbl_nth(aux_l2,i));
  eee_lbl_free_ck(aux_l1);
  eee_lbl_free_ck(aux_l2);
  eee_mon_set_cff(m,eee_mon_cff(aux_m));
  eee_mon_set_lbl_tl(m,res_l);
  free_eee_mon(aux_m);
}

void eee_add_this_monomial_unique(monomial m, eee *p, ring R)
{
  /* TODO: Optimize (perhaps is already ok) */
  eee aux_p;
  OPTIMIZE("eee_add_this_monomial_unique");
  aux_p=this_monomial2eee(m,R);
  *p=eee_sum_these(aux_p,*p,R);
}

eee_lbl eee_first_lbl(eee p, ring R)
{
  eee_lbl res_ss,l;
  int n,i,j;
  monomial res_m;
  if (p == NULL) return NULL;
  n=ring_indets_num(R);
  res_ss=eee_lbl_init(n);
  i=1;
  while (i<=n) {
    l=eee_mon_lbl_tl(p);
    eee_lbl_set_nth(res_ss,i,eee_mon_lbl_hd(p));
    for (j=1;j<=eee_lbl_len_ck(l);j++) 
      eee_lbl_set_nth(res_ss,j+i,eee_lbl_nth(l,j));
    i+=eee_mon_lbl_len(p);
    p=eee_mon_cff(p);
  }
  return res_ss;
}

int eee_first_which_gt(eee p1, eee p2, int l, ring R)
{
  coc_bool res_b;
  eee_lbl aux_l1,aux_l2;
  aux_l1=eee_first_lbl(p1,R);
  aux_l2=eee_first_lbl(p2,R);
  res_b=eee_lbl_which_gt(aux_l1,aux_l2);
  OPTIMIZE("eee_first_which_gt");
  eee_lbl_free(aux_l1);
  eee_lbl_free(aux_l2);
  return res_b;
}


/****************************************************************/




#if 0 /* TODO_PUBLIC */

/* -----[ EEE ]-------------- */

/* ENTRY FUNCTIONS */

int eee_size(polynomial p, ring R)
{
  return eee_size_l(p,ring_indets_num(R),R)+sizeof(eee);
}

term poly_this_leading_term(polynomial p, ring R)
{
  TODO("poly_this_leading_term (note that is impossible)");
  return term_NULL;
}

/* -----[ EEE_MON ]-------------- */



void eee_optimize_this(eee p, int l, ring R)
{
  if (l==0 || p==eee_NULL) return;
  while (eee_is_mon(p) &&
         eee_mon_lbl_len(p) < l &&
         eee_is_mon(eee_mon_cff(p)) &&
         l-eee_mon_lbl_len(p) != ring_level(R))
    eee_mon_lift_this(p);
  while (p != eee_NULL) {
    eee_optimize_this(eee_mon_cff(p),l-eee_mon_lbl_len(p),R);
    p=eee_mon_next(p);
  }
}

eee eee_this_monopoli2eee(monomial m, int l, ring R)
{
  /* NOTE: monomials have scalars instead of terms */

  eee res_p, aux_p;
  eee_mon aux_m, aux_m1;
  monomial first_m,prev_m;
  scalars first_ss,aux_ss;
  eee_lbl aux_l;
  int first_hd;
  int n, r, nl, i;

  if (l==0) {
    res_p=monomial_coeff(m);
    scalars_free((scalars)monomial_term(m));
    free_monomial(m);
    return res_p; }

  if (l==ring_level(R)) return monopoli_this_monopoli2eee(m,R);
  
  n = ring_indets_num(R);
  nl = n-l;
  res_p=aux_m=malloc_eee_mon();
  while (m != NULL) {
    first_m=m;
    r=n-ring_level(R);
    first_ss=(scalars)monomial_term(m);
    first_hd=scalars_nth(first_ss,nl+1);
    prev_m=m;
    m=monomial_next(m);
    while (m != NULL &&
        scalars_nth((scalars)monomial_term(m),nl+1) == first_hd) {
      aux_ss=(scalars)monomial_term(m);
      for (i=nl+2;i<=r;i++) 
        if (scalars_nth(aux_ss,i) != scalars_nth(first_ss,i)) {r=i-1; break;}
      prev_m=m;
      m=monomial_next(m);
    }
    if (m != NULL) monomial_set_next(prev_m,NULL);

    /* lbl=PI(n-l+2,r,first_m); */
    if (r > nl+1) {
      aux_l=eee_lbl_init(r-nl-1);
      for (i=r-nl-1; i>=1; i--) 
        eee_lbl_set_nth(aux_l,i,scalars_nth(first_ss,i+nl+1));
    } else aux_l=eee_lbl_NULL;

    aux_p = eee_this_monopoli2eee(first_m, n-r, R);
    aux_m1 = eee_mon_mk_NONEXT(first_hd, aux_l, aux_p);
    eee_mon_set_next(aux_m,aux_m1);
    aux_m=aux_m1;
  }

  eee_mon_set_next(aux_m,eee_mon_NULL);
  aux_m=res_p;
  res_p=eee_mon_next(res_p);
  free_eee_mon(aux_m);

  return res_p;
}

eee monopoli_this_monopoli2eee(monomial m, ring R)
{
  eee res_p;
  eee_mon aux_m;
  scalars aux_ss;
  eee_lbl aux_l;
  monomial prev_m;
  int i, l, nl;

  /* NOTE: monomials have scalars instead of terms */

  res_p=aux_m=malloc_eee_mon();
  l = ring_level(R);
  nl = ring_indets_num(R)-l;

  if (l == 1)
    while(m != NULL) {
      eee_mon_set_next(aux_m,
        eee_mon_mk_NONEXT(scalars_nth((scalars)monomial_term(m),nl+1),
          eee_lbl_NULL,monomial_coeff(m)));
      scalars_free((scalars)monomial_term(m));
      prev_m=m;
      m=monomial_next(m);
      free_monomial(prev_m);      
      aux_m=eee_mon_next(aux_m); }

  else

    while(m != NULL) {
      aux_ss=(scalars)monomial_term(m);
      aux_l=eee_lbl_init(l-1);
      for (i=l-1; i>=1; i--) 
        eee_lbl_set_nth(aux_l,i,scalars_nth(aux_ss,i+nl+1));
      eee_mon_set_next(aux_m,
        eee_mon_mk_NONEXT(scalars_nth((scalars)monomial_term(m),nl+1),
          aux_l,monomial_coeff(m)));
      m=monomial_next(m);
      aux_m=eee_mon_next(aux_m); }

  eee_mon_set_next(aux_m,eee_mon_NULL);
  aux_m=res_p;
  res_p=eee_mon_next(res_p);
  free_eee_mon(aux_m);
  return res_p;

}

/* --------------------- *\
         EEE MON
\* --------------------- */

eee_mon eee_mon_mk_NONEXT(expType hd, eee_lbl tl, eee p)
{
  eee_mon res_m;
  res_m=malloc_eee_mon();
  eee_mon_set_lbl_hd(res_m,hd);
  eee_mon_set_lbl_tl(res_m,tl);
  eee_mon_set_cff(res_m,p);
  return res_m;
}

eee_mon eee_mon_mk_special_NOLEVEL(eee_lbl l, eee p)
{
  eee_mon res_m;
  eee_lbl aux_l;
  int i;
  res_m=malloc_eee_mon();
  eee_mon_set_lbl_hd(res_m,eee_lbl_nth(l,1));
  aux_l=eee_lbl_dup_from(l,1);
  eee_lbl_free_ck(l);
  eee_mon_set_lbl_tl(res_m,aux_l);
  eee_mon_set_cff(res_m,p);
  eee_mon_set_next(res_m,eee_mon_NULL);
  return res_m;
}


eee_mon eee_mon_free(eee_mon m, int l, ring R)
{
  eee_free(eee_mon_cff(m),l-eee_mon_lbl_len(m),R);
  eee_lbl_free_ck(eee_mon_lbl_tl(m));
  free_eee_mon(m);
}

#endif /* TODO_PUBLIC */

#endif /* EEE */
