/* $Id: engine.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */
/* [ auxiliary ] [ environments ] */

#include "cocdefs.h"
#include "engine.h"

#include "ca.h"
#include "ck.h"
#include "ctbuf.h"
#include "cterm.h"
#include "evector.h"
#include "gb.h"
#include "map.h"
#include "machine.h"
#include "mathcoc.h"
#include "memoryco.h"
#include "module.h"
#include "object.h"
#include "olist.h"
#include "operator.h"
#include "options.h"
#include "pseries.h"
#include "pvector.h"
#include "record.h"
#include "ring.h"
#include "utils.h"
#include "vlist.h"

/* Keywords:  HANDLERS, FROZEN */

#define TIMER_FOR_OP(op) (op != op_EOF && op != op_SKIP)

ring current_ring;
coc_bool engine_made_output;
map curr_al_map;
charstars curr_al_pkgs;

void engine_init(void)
{
  curr_al_map=map_new();
  curr_al_pkgs=malloc_string_list(MAX_NUM_PACKAGES); /* BIGGER? */
}

object engine_and_time(cterm t)
{ /* it calls engine() */
  long int t1,t2,c1,c2;
  record res_r;
  object res_o;
  get_time(&t1);
  get_cputime(&c1);
  res_o=engine(t);
  if (res_o != object_null) return res_o;
  get_time(&t2);
  get_cputime(&c2);
  res_r=record_init();
  record_update_field(&res_r,"User",int2object(t2-t1));
  record_update_field(&res_r,"Cpu",int2object(c2-c1));
  res_o=record2object(res_r);
  object_set_label(res_o,my_strdup("Time"));
  return res_o;
}

object engine(cterm t)
{
  coop op;
  char * aux_s;
  object res_o;

  ca_monitor(0);
  if (ca_interrupted) return cocoa_error("INTERRUPTED");
  error_set_context(t);
  error_set_param(cterm_NULL);

  if (cterm_op(t) != op_ALIAS_IN && cterm_op(t) != op_PROGRAM) {

    aux_s=cterm_expand_aliases(t,curr_al_map,curr_al_pkgs);
    if (aux_s != NULL) {
      error_set_context(t);
      error_set_this_param(cterm_mk_NBS(aux_s));
      free_str1(aux_s);
      return cocoa_error("CANNOT_USE_ALIAS");
    }

  }

  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_TRACE)) {
    coop aux_op;
    opt_set_value_check(opt_panel_ENGINE,flag_ENGINE_TRACE,flag_OFF);
    put_cterm(t);
    aux_op=op_map_coop(pkg_user_n,"Debugger");
    if (aux_op != op_NULL) {
      object aux_o;
      cterm aux_t;
      aux_t=cterm_mk_const(aux_op);
      aux_o=eval_caller(aux_t,current_ring);
      cterm_free(aux_t);
      object_unlink(aux_o); } /* TODO: what about the errors of the monitor? */
    else put_cterm(cterm_newline);
    opt_set_value_check(opt_panel_ENGINE,flag_ENGINE_TRACE,flag_ON);
  }

  op=cterm_op(t);
  if (ck_handlers[op]==NULL) return cocoa_error("UNKN_CMD");
  else if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_COMP_STACK)) {
    ctbuf_push(&comp_stack, t);
    res_o = (* ck_handlers[op])(t,current_ring);
    if (!object_is_error(res_o)) ctbuf_pop(comp_stack);
    return res_o; 
  } else {
    return (* ck_handlers[op])(t,current_ring);
  }
}

object engine_in_ring(cterm t, ring R)
{
  object res_o;
  ring aux_R;
  aux_R=current_ring;
  current_ring=R;
  res_o=engine(t);
  current_ring=aux_R;
  return res_o;
}

#if 0
--------------------------------------------------------------------
   engine HANDLERS
--------------------------------------------------------------------
object engine_alias(cterm t, ring R);
object engine_alias_in(cterm t, ring R);
object engine_assign(cterm t, ring R);
object engine_break(cterm t, ring R);
object engine_catch(cterm t, ring R);
object engine_clear_rings(cterm t,ring R);
object engine_close(cterm t, ring R);
object engine_closelog(cterm t, ring R);
object engine_commands(cterm t, ring R);
object engine_define_func(cterm t, ring R);
object engine_define_ring(cterm t, ring R); 
  /* R ::= Q[xy] *//* Use R ::= Q[xy] *//* Use Q[xy] */
object engine_delete_rings(cterm t,ring R);
object engine_delete_vars(cterm t,ring R);
object engine_describe(cterm t, ring R);
object engine_eval(cterm t, ring R);
object engine_for(cterm t, ring R);
object engine_flush(cterm t, ring R);
object engine_foreach(cterm t, ring R);
object engine_if(cterm t, ring R);
object engine_need(cterm t, ring R);
object engine_openlog(cterm t, ring R);
object engine_option(cterm t, ring R);
object engine_package(cterm t, ring R);
object engine_panel(cterm t, ring R);
object engine_print(cterm t, ring R);
object engine_quit(cterm t, ring R);
object engine_repeat(cterm t, ring R);
object engine_return(cterm t, ring R);
object engine_set_curr_ring(cterm t, ring R); /* Use R */
object engine_set_option(cterm t, ring R);
object engine_skip(cterm t, ring R);
object engine_source(cterm t, ring R);
object engine_terror(cterm t, ring R);
object engine_time(cterm t, ring R);
object engine_unset_option(cterm t, ring R);
object engine_while(cterm t, ring R);
--------------------------------------------------------------------
#endif /* 0 */

/* This function is a bit "naughty"; it uses the error handling mechanism */
/* to propagate a Quit request to top level, and then fiddles the printing*/
/* of a "QUITTING" error so that it looks reasonable.  Previously this    */
/* function merely returned object_link(object_NULL), and this meant that */
/* a Quit inside a loop was not acted upon until the loop had finished.   */
object engine_quit(cterm t, ring R)
{
  opt_set_value(opt_panel_INTERNAL,flag_ENGINE_EXIT,flag_ON);
  return cocoa_error("QUITTING");
}

object engine_break(cterm t, ring R)
{
  return object_link(object_BREAK);
}

object engine_panel(cterm t, ring R)
{ /* PANEL(STRING) */
  opt_panel aux_p;
  object aux_o;
  char * aux_s;
  aux_p=opt_panel_NULL;
  if (cterm_op(cterm_param(t,0)) == op_NBS) {
    aux_p=panels_env_get(cterm_string(cterm_param(t,0)));
  } else {
    aux_o=eval(cterm_param(t,0),R);
    aux_s=object_cast_string(aux_o);
    if (math_error != object_NULL) return cocoa_error("UNKN_OPT_OR_PANEL");
    aux_p=panels_env_get(aux_s);
    free_str1(aux_s);
  }
  if (aux_p == opt_panel_NULL) return cocoa_error("UNKN_OPT_OR_PANEL");
  aux_o=panel2object(aux_p);
  engine_print_object(aux_o,cdev_out,R);
  object_unlink(aux_o);
  ca_put_dashes(cdev_out);
  return object_link(object_null);
}

object engine_skip(cterm t, ring R)
{
  /* do nothing */
  return object_link(object_null);
}

object engine_get_panel_and_option(cterm t, opt_panel * pan, opt_id * oid, 
  ring R)
{
  char * aux_s;
  aux_s=eval_var_name(t,R);
  cocoa_panels_search(aux_s,pan,oid);
  free_str1(aux_s);
  if (*oid != opt_id_NULL) return object_NULL;
  else return cocoa_error("UNKN_OPT_OR_PANEL");
}

object engine_set_option_NUM(opt_panel pan, opt_id oid, object o)
{
  TODO("engine_set_option_NUM");
  return object_link(object_null);
}

object engine_set_option_FLAG(opt_panel pan, opt_id oid, object o)
{
  if (object_type(o) == objtype_bool) {
    if (o->v.bool_value) flag_set(pan,oid);
    else flag_unset(pan,oid);
  } else {
    return cocoa_error("WRONG_OPTV");
  }
  return object_link(object_null);
}

object engine_set_option_ENUM(opt_panel pan, opt_id oid, object o)
{
  TODO("engine_set_option_ENUM");
  return object_link(object_null);
}

object engine_set_option(cterm t, ring R)
{
  /* Set(O,V) */
  /* Panel may be "", it means look for the unique option O in all the panels */
  object res_o, aux_o;
  opt_panel pan;
  opt_id oid;
  res_o=engine_get_panel_and_option(cterm_param(t,0), &pan, &oid, R);
  if (res_o != object_NULL) return res_o;
  aux_o=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o)) return aux_o;
  switch (opt_get_type(pan,oid)) {
    case opt_type_NUM : res_o=engine_set_option_NUM(pan,oid,aux_o); break;
    case opt_type_FLAG : res_o=engine_set_option_FLAG(pan,oid,aux_o); break;
    case opt_type_ENUM : res_o=engine_set_option_ENUM(pan,oid,aux_o); break;
    default : return object_link(object_null);
  }
#ifdef DEBUG
  utils_MEMORY_FLAG=flag_is_on(opt_panel_DEBUG,flag_DEBUG_MEM);
  if (utils_MEMORY_FLAG) {
    if (flag_is_on(opt_panel_DEBUG,flag_DEBUG_MEMFILE) && debuglog == stdout)
      debuglog=fopen("debug.log","a");
    else if(debuglog!=stdout&& !flag_is_on(opt_panel_DEBUG,flag_DEBUG_MEMFILE)) 
      { fclose(debuglog); debuglog=stdout; } }
#endif
  object_unlink(aux_o);
  return res_o;
}

object engine_define_func(cterm t, ring R)
/* PROGRAM(program_name,args,body,expr,help) */
{
  char * aux_s;
  object aux_o;
  aux_s=cterm_expand_aliases(t,map_empty,NULL); /* forbid aliases */
  if (aux_s != NULL) {
    error_set_context(t);
    error_set_this_param(cterm_mk_NBS(aux_s));
    free_str1(aux_s);
    return cocoa_error("CANNOT_USE_ALIAS");
  }
  aux_o=engine_define_pkg_func(pkg_user_n,t,R);
  if (aux_o != object_NULL) return aux_o;
  else return object_link(object_null);
}

object engine_catch(cterm t, ring R)
{
  /* CATCH(Command) */
  /* CATCH(Command,Var) */
  object aux_o,res_o;
  loc l;
  aux_o=engine(cterm_param(t,0));

  /* Check whether we exited via a Return; if so just pass control upwards */
  if (object_type(aux_o) != objtype_NULL &&
      !object_is_error(aux_o))
    return aux_o;

  if (cterm_arity(t)==2) {
    res_o=engine_assign_this(cterm_param(t,1), aux_o, R);
    if (!object_is_error(res_o) || res_o==aux_o) {
      object_unlink(res_o);
      res_o=object_link(object_null); }
    }
  else {
    object_unlink(aux_o);
    res_o=object_link(object_null); }

  if (object_is_error(res_o)) error_set_context(t);
  return res_o;
}

object engine_commands(cterm t, ring R)
{
  /* COMMANDS(C1,...,Cn) */
  int i;
  object res_o;
  for (i=0;i<cterm_arity(t);i++) {
    res_o=engine_in_ring(cterm_param(t,i),R);
    if (res_o != object_null) return res_o;
    object_unlink(res_o);  }
  return object_link(object_null);
}

object engine_if(cterm t, ring R)
{ 
  /* IF(B1,C1,B2,C2,...,Bn,Cn) */
  /* IF(B1,C1,B2,C2,...,Bn,Cn,C_else) */
  coc_bool aux_b;
  int arity,i;
  arity=cterm_arity(t);
  for (i=0;i<=arity/2-1;i++) {
    aux_b=eval_bool(cterm_param(t,i*2),R);
    if (math_error!=object_NULL) return math_error;
    if (aux_b) return engine_in_ring(cterm_param(t,i*2+1),R); }
  if (arity & 1) return engine_in_ring(cterm_param(t,arity-1),R); /* else */
  return object_link(object_null); /* do nothing */
}

object engine_while(cterm t, ring R)
{
  /* WHILE(B,C) */
  coc_bool aux_b;
  object res_o;
  aux_b=eval_bool(cterm_param(t,0),R);
  if (math_error!=object_NULL) { object err=math_error; math_error=object_NULL; return err; }
  while (aux_b) {
    res_o=engine_in_ring(cterm_param(t,1),R);
    if (res_o!=object_null) {
      if (res_o == object_BREAK) {
        object_unlink(res_o);
        return object_link(object_null);
      } else {
        return res_o; 
      }
    }
    object_unlink(res_o);
    aux_b=eval_bool(cterm_param(t,0),R);
    if (math_error!=object_NULL) return math_error; }
  return object_link(object_null);
}

object engine_repeat(cterm t, ring R)
{
  /* REPEAT(C,B) */
  coc_bool aux_b;
  object res_o;
  do {
    res_o=engine_in_ring(cterm_param(t,0),R);
    if (res_o!=object_null) {
      if (res_o == object_BREAK) {
        object_unlink(res_o);
        return object_link(object_null);
      } else {
        return res_o; 
      }
    }
    object_unlink(res_o);
    aux_b=eval_bool(cterm_param(t,1),R);
    if (math_error!=object_NULL) return math_error; }
  while (!aux_b);
  return object_link(object_null);
}

object engine_package(cterm t, ring R)
{ /* PACKAGE(PKG_ID,[EXPORT,...,EXPORT] */
  cterm t_exports;
  int pkg_n;
  object aux_o;
  char * pkg_name;
  cterm params[5],aux_t;
  int i;
  t_exports=cterm_param(t,1);
  pkg_name=cterm_string(cterm_param(cterm_param(t,0),0));
  cterm_expand_dollar(t_exports,pkg_name);
  pkg_n=pkg_init(pkg_name);
  for (i=0;i<cterm_arity(t_exports);i++) {
    aux_o=engine_define_pkg_func(pkg_n,cterm_param(t_exports,i),R);
    if (aux_o != object_NULL) { /* if error occurred */
      pkg_clear(pkg_n);
      return aux_o; 
    }
  }
  /* definition of the function PkgName() */
  params[0]=cterm_mk_NBS(pkg_name);
  params[0]=cterm_mk_classic(op_STRING,1,params);
  params[2]=cterm_mk_classic(op_RETURN,1,params);
  params[0]=cterm_mk_NBS("PkgName");
  params[1]=cterm_mk_const(op_LIST);
  params[3]=cterm_mk_const(op_NULL_VALUE);
  params[4]=cterm_mk_const(op_NULL_VALUE);
  aux_t=cterm_mk_classic(op_PROGRAM,5,params);
  aux_o=engine_define_pkg_func(pkg_n,aux_t,R);
  cterm_free(aux_t);
  /* end of definition of the function PkgName() */
  if (math_error != object_NULL) return math_error;
  if (op_map_coop(pkg_n,"Initialize") != op_NULL) {
    params[0]=cterm_mk_NBS(pkg_name);
    params[1]=cterm_mk_NBS("Initialize");
    params[2]=cterm_mk_const(op_LIST);
    aux_t=cterm_mk_classic(op_FUN_APPL,3,params);
    aux_o=eval(aux_t,R);
    if (object_is_error(aux_o)) {
      error_set_context(aux_t);
      return aux_o;
    } else {
      object_unlink(aux_o);
      cterm_free(aux_t);
    }
  }
  return object_link(object_null);
}

object engine_print(cterm t, ring R)
{
  /* PRINT(E1,...,En) */
  /* PRINT_AND_CR(E1,...,En) */
  /* PRINTON(E1,...,En) */
  /* PRINTON_AND_CR(E1,...,En) */
  char * aux_filename;
  cdev aux_d = cdev_out;
  opt_panel pan;
  cterm_list aux_params;
  object res_o,aux_o;
  int i;
  int arity;
  coop op=cterm_op(t);
  coc_bool aux_b_do_cr = (op==op_PRINT_AND_CR || op==op_PRINTON_AND_CR);
  coc_bool aux_b_on_file = (op==op_PRINTON || op==op_PRINTON_AND_CR);

  engine_made_output=TRUE;
  
  if (aux_b_on_file) {
    aux_o=eval_caller(cterm_param(t,0),R);
    if (object_is_error(aux_o)) return aux_o;
    else {
      if (object_type(aux_o) != objtype_cdev) 
        return cocoa_error("EXPECTED_DEV");
      aux_d=object2cdev(aux_o);
      t=cterm_param(t,1);
      cdev_unlink(aux_d);
    } 
  }

  arity=cterm_arity(t);

  if (arity==0) {
    if (aux_b_do_cr) ca_put_nl(aux_d);
    else engine_made_output=FALSE;
    res_o=object_link(object_null); }
#ifdef DEBUG
  else if (cterm_op(cterm_param(t,0))==op_OBJECTS) printf_objects(cdev_out);
#endif
  else {
    for (i=0;i<arity;i++) {
      if (cterm_op(cterm_param(t,i))==op_NEWLINE) { ca_put_nl(aux_d); res_o=object_link(object_null); }
      else {
        aux_o = eval_caller(cterm_param(t,i),R);
        if (!object_is_error(aux_o)) {
          res_o=engine_print_object(aux_o,aux_d,R);
          if (object_is_error(res_o)) {
            object_unlink(aux_o);
            break; }
          object_unlink(aux_o); }
        else {
          res_o=aux_o;
          break; } } }
    if (aux_b_do_cr) ca_put_nl(aux_d); }
    engine_made_output=(aux_d==cdev_out);
  return res_o;
}

object engine_for(cterm t, ring R)
{
  /* FOR(VAR,C,E1,E2,E3) or FOR(VAR,C,E1,E2) */

  unsigned int nsteps;
  int i_from,i_to,i_step,arity,i;
  char * aux_v, * aux_v1;
  object aux_o,aux_o1,res_o;
  cterm aux_t,cmd_t;
  cterm_list aux_params;
  env aux_env;
  loc aux_l;
  cterm_list params=cterm_params(t);
  arity=cterm_arity(t);

  res_o=object_NULL;

  i_from = eval_int(params[2],R);
  if (math_error!=object_NULL) return math_error;
  i_to = eval_int(params[3],R);
  if (math_error!=object_NULL) return math_error;
  if (arity==5) {
    i_step = eval_int(params[4],R);
    if (math_error!=object_NULL) return math_error; }
  else i_step=1;
  if (i_step == 0) return cocoa_error("FOR_STEP_0");
  /* Check whether we should do no iterations. */
  if (i_step < 0 && i_to > i_from ||
      i_step > 0 && i_to < i_from)
    return object_link(object_null);
  
  /* These two horrible lines compute the number of steps to do */
  if (i_step >0) nsteps = ((unsigned)i_to - (unsigned)i_from)/(unsigned)i_step;
  else nsteps = ((unsigned)i_from - (unsigned)i_to)/(unsigned)(-i_step);
  nsteps++;

  /* This block creates a "local scope" for the variable of a For loop */
  /* inside the body of same.                                          */
  aux_v=eval_var_name(params[0],R);
  aux_v1=env_newvar(aux_v);
  cmd_t=params[1];
  cterm_subst_vars(&cmd_t,aux_v,aux_v1);
  cterm_subst_vars(&(params[0]),aux_v,aux_v1);

  aux_env = current_env;
  aux_l=env_declare(&aux_env,aux_v1);
  aux_o1=int2object(0);
  mem_assign(aux_l,aux_o1);
  object_unlink(aux_o1);
  /* The odd form of the loop below is to handle the extreme case of a */
  /* CoCoA "For" loop which is performed MAX_UNSIGNED times.           */
  for (i=i_from;; i+=i_step)
  {
    aux_o1=int2object(i);
    mem_assign(aux_l,aux_o1);
    /* (*aux_l)->v.int_value=i; */ /* PORCATA, NON FUNZIONA (causa ref.count) */
    aux_o=engine(cmd_t);
    object_unlink(aux_o1);
    if (aux_o!=object_null) {
      if (aux_o == object_BREAK) {
        object_unlink(aux_o);
        res_o=object_link(object_null);
      } else {
        res_o=aux_o; 
      }
      break; 
    }
    object_unlink(aux_o);
    if (--nsteps == 0) break;
  }
  cterm_subst_vars(&cmd_t,aux_v1,aux_v);
  cterm_subst_vars(&(params[0]),aux_v1,aux_v);
  env_delete_var(&aux_env,aux_v1);
  free_str(strlen(aux_v1),aux_v1);
  free_str(strlen(aux_v),aux_v);
  if (res_o==object_NULL) return object_link(object_null);
  return res_o; 
}

object engine_foreach(cterm t, ring R)
{
  /* foreach("X",L,C) */
  /* FOREACH X IN L DO C END */
  char * aux_v, * aux_v1;
  object aux_o,aux_o2,res_o;
  olist aux_ol;
  env aux_env;
  cterm aux_t,cmd_t;
  cterm_list params;
  loc aux_l;
  int n,arity,i;

  res_o=object_NULL;

  params=cterm_params(t);
  arity=cterm_arity(t);
  aux_v=eval_var_name(params[0],R);
  aux_v1=env_newvar(aux_v);
  /* NOTE: this solves the problem of collisions between X and 
     ring-local vars */
  cmd_t=params[arity-1];
  cmd_t=cterm_dup(cmd_t);
  cterm_subst_vars(&cmd_t,aux_v,aux_v1);
  cterm_subst_vars(&(params[0]),aux_v1,aux_v);
  aux_env=current_env; /* TODO: local or global? */

  aux_l=env_declare(&aux_env,aux_v1);
  aux_o=eval(params[1],R); /* the list */
  if (object_is_error(aux_o)) {
    cterm_free(cmd_t);
    env_delete_var(&aux_env,aux_v1);
    free_str1(aux_v1);
    free_str1(aux_v);
    return aux_o;
  }
  if (object_type(aux_o) != objtype_olist) {
    aux_o=object_cast(aux_o,objtype_olist,R);
    if (object_is_error(aux_o)) {
      cterm_free(cmd_t);
      env_delete_var(&aux_env,aux_v1);
      free_str1(aux_v1);
      free_str1(aux_v);
      return aux_o; 
    }
  }
  aux_ol=object2olist(aux_o);
  n=olist_len(aux_ol);
  for (i=1;i<=n;i++) {
    mem_assign(aux_l,olist_nth(aux_ol,i));
    aux_o=engine(cmd_t);
    if (aux_o!=object_null) {
      if (aux_o == object_BREAK) {
        object_unlink(aux_o);
        res_o=object_link(object_null);
      } else {
        res_o=aux_o; 
      }
      break; 
    }
    object_unlink(aux_o); }
  env_delete_var(&aux_env,aux_v1);
  olist_free(aux_ol);
  cterm_subst_vars(&cmd_t,aux_v1,aux_v); /* TODO: necessario? */
  cterm_subst_vars(&(params[0]),aux_v1,aux_v); /* TODO: necessario? */
  cterm_free(cmd_t);
  free_str1(aux_v1);
  free_str1(aux_v);
  if (res_o==object_NULL) return object_link(object_null);
  return res_o;
}

object engine_eval(cterm t, ring R)
{
  object aux_o,res_o;
  char * id;

  if (R==ring_NULL) return cocoa_error("UNDEF_RING");

  aux_o = eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  if (aux_o == object_null) return aux_o;
  if (aux_o == object_RETURN_NULL) {
    object_unlink(aux_o);
    return object_link(object_null);
  }

  env_declare_and_assign(&current_env,"It",aux_o);
  if (math_error!=object_NULL) return math_error;
  res_o=engine_print_object(aux_o,cdev_out,R);
  object_unlink(aux_o);
  engine_made_output=TRUE;
  return res_o;
}

object engine_return(cterm t, ring R)
{
  if (cterm_arity(t)==0) return object_link(object_RETURN_NULL);
  return eval_caller(cterm_param(t,0),R);
}

object engine_alias(cterm t, ring R)
{ /* ALIAS_CMD(ALIAS(A1,P1),ALIAS(A2,P2),...) */
  int i,n;
  char * aux_s;
  cterm t_alias;
  /* TODO: check overflow, check already defined aliases */
  for (i=0;i<cterm_arity(t);i++) {
    t_alias=cterm_param(t,i);
    aux_s=eval_var_name(cterm_param(t_alias,0),R);
    n=map_get_name_id(curr_al_map,aux_s);
    if (n==name_id_NULL)
      n=map_add(&curr_al_map,aux_s);
    else
      free_str1(curr_al_pkgs[n]);
    curr_al_pkgs[n]=
      my_strdup(cterm_string(cterm_param(cterm_param(t_alias,1),0)));
    free_str1(aux_s);
  }
  return object_link(object_null);
}


object engine_alias_in(cterm t, ring R)
{
  cterm t_alias,t_aliases,t_command;
  int n,i,al_pkgs_size;
  map save_curr_al_map;
  charstars save_curr_al_pkgs;
  char * aux_s;
  object res_o;
  save_curr_al_map=curr_al_map;
  save_curr_al_pkgs=curr_al_pkgs;
  curr_al_map=map_new();
  t_aliases=cterm_param(t,0);
  al_pkgs_size=cterm_arity(t_aliases)+20;
  curr_al_pkgs=malloc_string_list(al_pkgs_size); /* TODO: TRICK ? */
  /* 20 is a TRICK (just to add to the list some alias, like $ for self) */
  for (i=0;i<cterm_arity(t_aliases);i++) { /* TODO: CHECK t_imports size<100 */
    t_alias=cterm_param(t_aliases,i);
    aux_s=eval_var_name(cterm_param(t_alias,0),R);
    n=map_add(&curr_al_map,aux_s);
    curr_al_pkgs[n]=
      my_strdup(cterm_string(cterm_param(cterm_param(t_alias,1),0)));
    free_str1(aux_s);
  }
  t_command=cterm_param(t,1);
  aux_s=cterm_expand_aliases(t_command,curr_al_map,curr_al_pkgs);
  if (aux_s != NULL) {
    error_set_this_param(cterm_mk_NBS(aux_s));
    free_str1(aux_s);
    return cocoa_error("UNDEF_ALIAS"); }
  res_o=engine(t_command);
  for (i=first_name_id; i<=map_last_id(curr_al_map); i++) {
    free_str1(curr_al_pkgs[i]);
  }
  free_string_list(al_pkgs_size,curr_al_pkgs);
  map_free(&curr_al_map);
  curr_al_map=save_curr_al_map;
  curr_al_pkgs=save_curr_al_pkgs;
  return res_o;
}

object engine_assign(cterm t, ring R)
{
  /* t=ASSIGN(L:LEXPR,E:EXPR) (L := E) */
  /* t=ASSIGN(L:LEXPR,E:EXPR,R:RINGENV) (R :: L := E) */

  cterm t_left, t_right;
  cterm_list params=cterm_params(t);
  ring aux_R;
  object aux_o;
  char * id;

  if (cterm_arity(t)==3) {
    id=eval_var_name(params[2],R);
    aux_R = ring_env_get(id);
    free_str1(id);
  }
  else aux_R = R;
  if (aux_R==ring_NULL) return cocoa_error("UNDEF_RING");

  t_left=params[0];
  t_right=params[1];

  aux_o=eval_caller(t_right,aux_R);

  /* the following must be called even if aux_o is an error ... */
  /* ... because we could want to assign an error to a variable */

  /* NO! I prefer to explicitly use Catch to assign errors */
  /* Catch Error('hello') In ErrHello End; */

  if (object_is_error(aux_o)) return aux_o;

  return engine_assign_this(t_left, aux_o, aux_R);
}

object engine_terror(cterm t, ring R)
{
  ca_put(t,cdev_out);
  engine_made_output=TRUE;
  return object_link(object_null);
}

object engine_define_ring(cterm t, ring R)
{
    /* t=define_ring("name",ring:expr,annotations) */
    /* t=define_and_set("name",ring:expr,annotations) */

  char * id;
  object aux_o,aux_o1;
  ring aux_R;

  id = eval_var_name(cterm_param(t,0),R);
  if (math_error != object_NULL)
    return math_error;

  /* R ::= Q[xy] */
  if ((strcmp(id,ring_id(R))==0) && (cterm_op(t)==op_DEFINE_RING)) {
    error_set_this_param(cterm_mk_NBS(id));
    free_str1(id);
    return cocoa_error("CANNOT_RED_CURR_RING"); 
  }

  aux_o = eval_ringenv(cterm_param(t,1),cterm_param(t,2),R);
  if (object_is_error(aux_o)) {
    free_str1(id);
    return aux_o; 
  }
  
  aux_R=ring_env_get(id);
  if (aux_R != ring_NULL) {
    ring_unlink(aux_R);
    aux_o1=ring_env_delete(aux_R);
    if (object_is_error(aux_o1)) {
      free_str1(id);
      return aux_o1;
    } else {
      object_unlink(aux_o1);
    }
  }

  ring_set_id(aux_o->v.ring_value,id);
  env_declare_and_assign(&ring_env,id,aux_o);
  R = aux_o->v.ring_value;
  /* env_declare_and_assign(&current_env,id,aux_o); */
  object_unlink(aux_o);
  if (cterm_op(t)==op_DEFINE_AND_USE_RING) current_ring=R;
  return object_link(object_null);
}

object engine_set_curr_ring(cterm t, ring R)
{
  char * id;
  id = eval_var_name(cterm_param(t,0),R);
  if (math_error != object_NULL) return math_error;
  R=ring_env_get(id);
  if (R==ring_NULL) {
    error_set_this_param(cterm_mk_NBS(id));
    free_str1(id);
    return cocoa_error("UNDEF_RING"); }
  if (ring_indets_num(R) == 0) /* must test here as Q and Z already exist */
  {
    free_str1(id);
    return cocoa_error("RING_NO_INDETS");
  }
  current_ring=R;
  free_str1(id);
  return object_link(object_null);
}

object engine_using(cterm t, ring R)
{ /* USING(RINGENV,COMMANDS) */
  char * id;
  object res_o;
  ring S;
  id = eval_var_name(cterm_param(t,0),R);
  S=ring_env_get(id);
  ring_unlink(S);
  if (S==ring_NULL) {
    error_set_this_param(cterm_mk_NBS(id));
    free_str1(id);
    return cocoa_error("UNDEF_RING"); }
  free_str1(id);
  res_o=engine_in_ring(cterm_param(t,1),S);
  return res_o;
}

object engine_source(cterm t, ring R)
{ /* SOURCE(STRING) */
  /* SOURCE(DEVICE) */
  object aux_o;
  cdev aux_d;
  char * filename, *aux_s;

  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_TRACE_SOURCES)) {
    ca_put(t, cdev_out);
  }

  aux_o = eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  switch (object_type(aux_o))
  {
  case objtype_string:
    aux_s = object2string(aux_o);
    if (*aux_s == '$') {
      filename = pkg_path(aux_s);
    } else {
      filename = build_path(aux_s);
    }
    free_str1(aux_s);
    aux_d=cdev_open_ifile(filename, cdev_prot_cocoal);
    if (aux_d == cdev_NULL) {
      error_set_this_param(cterm_mk_NBS(filename));
      free_str1(filename);
      return cocoa_error("FILE_NOT_FOUND");
    }
    free_str1(filename);
    ck_source(aux_d);
    cdev_unlink(aux_d);
    break;
  case objtype_cdev:
    aux_d=object2cdev(aux_o);
    ck_source(aux_d);
    cdev_unlink(aux_d);
    break;
  default:
    return cocoa_error("SOURCE_BAD_FILE");
  }

  return object_link(object_null);
}

object engine_describe(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o = eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=engine_describe_object(aux_o,cdev_out,R);
  object_unlink(aux_o);
  engine_made_output=TRUE;
  return res_o;
}

object engine_close(cterm t, ring R)
{
  cdev D;
  object aux_o;
    
  aux_o=eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  D=object2cdev(aux_o);
  cdev_close(D);
  cdev_unlink(D);
  return object_link(object_null);
}

object engine_flush(cterm t, ring R)
{
  cdev D;
  object aux_o;
    
  aux_o=eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  D=object2cdev(aux_o);
  cdev_flush(D);
  cdev_unlink(D);
  return object_link(object_null);
}

object engine_need(cterm t, ring R)
{
  return string2object(my_strdup("NEED not yet implemented"));
}

object engine_get_option_NUM(opt_panel pan, opt_id oid)
{
  TODO("engine_get_option_NUM");
  return object_link(object_null);
}

object engine_get_option_ENUM(opt_panel pan, opt_id oid)
{
  TODO("engine_get_option_ENUM");
  return object_link(object_null);
}

object engine_get_option_FLAG(opt_panel pan, opt_id oid)
{
  if (opt_get_value(pan,oid) == 0) 
    return bool2object(FALSE);
  else return bool2object(TRUE);
}


object engine_option(cterm t, ring R)
{ /* OPTION(OID) */
  /* Panel may be "", it means look for the unique option O in all the panels */
  object res_o, aux_o;
  opt_panel pan;
  opt_id oid;
  res_o=engine_get_panel_and_option(cterm_param(t,0), &pan, &oid, R);
  if (res_o != object_NULL) return res_o;
  switch (opt_get_type(pan,oid)) {
    case opt_type_NUM : res_o=engine_get_option_NUM(pan,oid); break;
    case opt_type_FLAG : res_o=engine_get_option_FLAG(pan,oid); break;
    case opt_type_ENUM : res_o=engine_get_option_ENUM(pan,oid); break;
    default : return object_link(object_null);
  }
  return res_o;
}

object engine_openlog(cterm t, ring R)
{ /* OPENLOG(DEVICE) */
  cdev D;
  object aux_o;

  aux_o=eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  D=object2cdev(aux_o);
  ca_open_log(D);
  cdev_unlink(D);
  return object_link(object_null);
}

object engine_closelog(cterm t, ring R)
{ /* CLOSELOG(DEVICE) */
  cdev D;
  object aux_o;

  aux_o=eval_caller(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  D=object2cdev(aux_o);
  ca_close_log(D);
  cdev_unlink(D);
  if (math_error != object_NULL) return math_error;
  return object_link(object_null);
}

object engine_clear_ring(cterm t)
{
  ring R;
  int arity;
  char * id;
  cterm_list params;
  id=eval_var_name(t,current_ring);
  R = ring_env_get(id);
  ring_unlink(R);
  free_str1(id);
  if (R==ring_NULL) return cocoa_error("UNDEF_RING");
  ring_clear(R);
  return object_link(object_null);
}

object engine_delete_ring(cterm t)
{
  ring R;
  char * id;
  id=eval_var_name(t,current_ring);
  if (strcmp(id,"Z") == 0 || strcmp(id,"Q") == 0) {
    error_set_this_param(cterm_mk_NBS(id));
    return cocoa_error("CANNOT_DEL");
  }
  R = ring_env_get(id);
  ring_unlink(R);
  if (R==ring_NULL) {
    error_set_this_param(cterm_mk_NBS(id));
    free_str1(id);
    return cocoa_error("UNDEF_RING");
  } else if (R==current_ring) {
    free_str1(id);
    error_set_this_param(cterm_mk_NBS("the current ring"));
    return cocoa_error("CANNOT_DEL"); 
  }
  free_str1(id);
  return ring_env_delete(R);
}

object engine_delete_rings(cterm t,ring R)
{
  int arity,i;
  object aux_o;
  cterm_list params;
  arity=cterm_arity(t);
  params=cterm_params(t);
  for (i=0;i<arity;i++) {
    aux_o=engine_delete_ring(params[i]);
    if (object_is_error(aux_o)) return aux_o;
  }
  return object_link(object_null);
}

object engine_clear_rings(cterm t, ring R)
{
  int arity,i;
  object aux_o;
  cterm_list params;
  arity=cterm_arity(t);
  if (arity==0) {
    env_free(current_env);
    current_env=env_init();
    /* ring_clear(current_ring); */
    return object_link(object_null); }
  params=cterm_params(t);
  for (i=0;i<arity;i++) {
    aux_o=engine_clear_ring(params[i]);
    if (object_is_error(aux_o)) return aux_o;
  }
  return object_link(object_null);
}

object engine_delete_var(cterm t,ring R)
{ /* V or X.V */
  char * id;
  env aux_env;
  aux_env=current_env;
  id=eval_var_name(t,R);
  if (math_error != object_NULL) return math_error;
  if (!env_is_declared(aux_env,id))
    return cocoa_error("UNDEF_VAR");
  env_delete_var(&aux_env,id);
  free_str1(id);
  return object_link(object_null);
}

object engine_delete_vars(cterm t,ring R)
{ /* DELETE(var_id, ..., var_id) */
  int arity,i;
  object aux_o;
  cterm_list params;
  arity=cterm_arity(t);
  params=cterm_params(t);
  for (i=0;i<arity;i++) {
    aux_o=engine_delete_var(params[i],R);
    if (object_is_error(aux_o)) return aux_o;
  }
  return object_link(object_null);
}

object engine_print_object_aux(object o, cdev D, char * prefix, ring R)
{
  object res_o;
  char * label, * pkg_id, * op_id;
  coop aux_op;
  int pkg_n,n;
  char * aux_v;
  cterm aux_params[1],aux_t;

  aux_op=op_NULL;
  if (object_is_labelled(o)) {
    char opname[200];
    if ((object_type(o) == objtype_objtype) && (o->v.objtype_value == objtype_tagged))
      return object_NULL;
    label=my_strdup(object_label(o));
    n=strlen(label);
    str_split(label,".",&pkg_id,&op_id);
    if (op_id == NULL) {
      aux_op=op_NULL;
    } else {
      sprintf(opname,"%s_%s\0",prefix,op_id);
      if (pkg_id == NULL) {
        coop_pkg_std_op(opname,&pkg_n,&aux_op);
      } else {
        pkg_n=pkg_index(pkg_id);
        if (pkg_n != name_id_NULL) {
          aux_op=op_map_coop(pkg_n,opname);
        } else aux_op=op_NULL;
      }
    }
    free_str(n,label);
    if (aux_op != op_NULL) aux_op=op_mk(pkg_n,aux_op);
  }
  if (aux_op==op_NULL) return object_NULL;
  else {
    aux_v=env_newvar("PP");
    env_declare_and_assign(&current_env,aux_v,o);
    aux_params[0]=cterm_mk_NBS(aux_v);
    aux_params[0]=cterm_mk_classic(op_VAR,1,aux_params);
    aux_t=cterm_mk_classic(aux_op,1,aux_params);
    res_o=eval_caller(aux_t,R);
    env_delete_var(&current_env,aux_v);
    free_str1(aux_v);
    cterm_free(aux_t);
  }
  return res_o;
}

object engine_print_object(object o, cdev D, ring R)
{
  object res_o;
  res_o=engine_print_object_aux(o,D,"Print",R);
  if (res_o == object_NULL) {
    printf_object(o, D);
    return object_link(object_null);
  }
  return res_o;
}

object engine_describe_object(object o, cdev D, ring R)
{
  object res_o;
  res_o=engine_print_object_aux(o,D,"Describe",R);
  if (res_o == object_NULL) {
    object_describe(o, D);
    return object_link(object_null);
  }
  return res_o;
}

cterm cocoa_panels(void)
{
  ctbuf aux_b;
  env_slot aux_e;
  object aux_o;
  loc aux_l;
  opt_panel aux_p;
  cterm aux_t;
  aux_b=ctbuf_empty();
  aux_e=(env_slot)panels_env;
  while (aux_e!=env_slot_NULL) {
    aux_l=env_slot_loc(aux_e);
    aux_o=object_link(*aux_l);
    aux_p=object2panel(aux_o);
    if (aux_p!=opt_panel_NULL) {
      aux_t=cterm_mk_NBS(opt_panel_descr(aux_p));
      ctbuf_append(aux_b,aux_t); }
    aux_e=env_slot_next(aux_e);
  }
  return ctbuf_mk_cterm(aux_b,op_PANELS);
}

int cocoa_panels_search(char * s, opt_panel *pan, opt_id * oid)
{
  env_slot aux_e;
  object aux_o;
  loc aux_l;
  opt_panel aux_p;
  cterm aux_t;
  aux_e=(env_slot)panels_env;
  *oid=opt_id_NULL;
  *pan=opt_panel_NULL;
  while (aux_e!=env_slot_NULL) {
    aux_l=env_slot_loc(aux_e);
    aux_o=object_link(*aux_l);
    aux_p=object2panel(aux_o);
    if (aux_p!=opt_panel_NULL) {
      *oid=panels_env_get_oid(aux_p,s);
      if (*oid != opt_id_NULL) {
        *pan=aux_p;
        return 1; }
    }
    aux_e=env_slot_next(aux_e);
  }
  return 0;
}

/* ---[ auxiliary ]-------------------------------------------- */

void cterm_expand_dollar(cterm t, char * p_id)
{
  int i,arity;
  char * aux_s;
  arity=cterm_arity(t);
  if (cterm_op(t)==op_FUN_APPL && arity==3) {
    aux_s=cterm_string(cterm_param(t,0));
    if (aux_s[0]=='$' && aux_s[1]=='\0') {
      free_str1(aux_s);
      cterm_set_string(cterm_param(t,0),my_strdup(p_id));
    }
  }
  for (i=0;i<arity;i++) cterm_expand_dollar(cterm_param(t,i),p_id);
}

void cterm_subst_vars(cterm * t, char * v1, char * v2)
{
  coop op=cterm_op(*t);
  int arity=cterm_arity(*t);
  register int i;
  cterm_list params=cterm_params(*t);
  if (arity==0) return;
  if (op == op_VAR && cterm_op(params[0])==op_NBS
      && strcmp(cterm_string(params[0]),v1) == 0) {
    cterm_free(params[0]);
    params[0]=cterm_mk_NBS(v2);
    (*t)->params=params; }
/* DON'T TRASH RECORD SELECTOR IF IT IS A CONSTANT STRING */
  if (op==op_RECORDSEL &&
      cterm_op(params[1]) == op_VAR &&
      cterm_op(cterm_params(params[1])[0]) == op_NBS)
    arity=1;
  
  for (i=0;i<arity;i++)
    cterm_subst_vars(&params[i],v1,v2);
  (*t)->params=params;
}

char * cterm_expand_aliases(cterm t, map al_map, charstars al_pkgs)
{ /* expands FUN_APPL("F",LIST) in F(LIST) */
  /* futhermore applies the aliases in al_map */
  int n,i,arity;
  cterm aux_t;
  char * res_s, *aux_s;
  res_s=NULL;
  arity=cterm_arity(t);
  if (cterm_op(t)==op_FUN_APPL && arity==3) {
    aux_t=cterm_param(t,0);
    aux_s=cterm_string(aux_t);
    if (aux_s[0] != '$') {
      n=map_get_name_id(al_map,aux_s);
      if (n!=name_id_NULL) {
        free_str1(aux_s);
        cterm_set_string(aux_t,my_strdup(al_pkgs[n]));
      } else {
        res_s=my_strdup(aux_s);
      } 
    }
  }
  if (res_s != NULL) return res_s;
  for (i=0;i<arity;i++) {
    aux_s=cterm_expand_aliases(cterm_param(t,i),al_map,al_pkgs);
    if (aux_s != NULL) return aux_s; }
  return res_s;
}

void engine_package_expand_aliases(int pkg_n, map al_map, charstars al_pkgs)
{
  int n,i;
  cterm h;
  char * aux_s;
  n=pkg_num_ops(pkg_n);
  for (i=2;i<=n;i++) {
    h=(cterm)op_map_op_handler(pkg_n,i);
    printf("%s:",op_map_op_name(pkg_n,i));
    put_cterm((cterm)op_map_op_handler(pkg_n,i));
    printf("\n");
    aux_s=cterm_expand_aliases(cterm_param(h,1),al_map,al_pkgs);
    if (aux_s != NULL) {
      math_error=cocoa_error("UNDEF_ALIAS");
      error_set_this_param(cterm_mk_NBS(aux_s));
      free_str1(aux_s); } }
}

object engine_define_pkg_func(int pkg_n, cterm t, ring R)
/* PROGRAM(program_name,args,body,expr,help) */
{
  cterm params[3];
  coop op,aux_op;
  cterm handler;
  int arity;
  char * op_name,* aux_s;
  params[0]=cterm_dup(cterm_param(t,1));
  params[1]=cterm_dup(cterm_param(t,2));
  params[2]=cterm_dup(cterm_param(t,3));
  handler=cterm_mk_classic(op_HANDLER,3,params);
  aux_s=NULL;
  if (cterm_op(cterm_param(t,4))!=op_NULL_VALUE)
    aux_s=cterm_string(cterm_param(cterm_param(t,4),0));
  op_name=cterm_string(cterm_param(t,0));
  aux_op=(coop)op_map_coop(pkg_n,op_name);
  if (cterm_op(cterm_param(t,1))==op_DOTS) arity=NO_ARITY;
  else arity=cterm_arity(cterm_param(t,1));
  
  if (aux_op==op_NULL)
    op=op_map_add_user_function(pkg_n,op_name,arity,handler,aux_s);
  else
    op=op_map_mod_user_function(pkg_n,aux_op,arity,handler,aux_s);
  if (op==op_NULL) return cocoa_error("CANNOT_DEF");
  return object_NULL;
}

object engine_assign_this(cterm t, object o, ring R)
{ /* t:LEXPR */ 
  /* LEXPR=VAR(S:NBS|STRING)  (V := E) */
  /* LEXPR=COMP(LEXPR,N:INT)  (L[N] := E) */
  /* LEXPR=RECORDSEL(LEXPR,X:ID)  (L.X := E) */

  object res_o;
  coop aux_op;
  loc aux_l;
  char * id, * aux_label;
  env aux_e;
  int done_memory=FALSE;

  if (object_type(o) == objtype_ring) return cocoa_error("CANNOT_ASSIGN_RINGS");
  aux_op=cterm_op(t);
  res_o=object_NULL;

  if (object_type(o) == objtype_record && o->v.record_value == cocoa_env) {
    res_o=cocoa_error("CANNOT_ASSIGN");
    error_set_this_param(cterm_mk_NBS("MEMORY"));

  } else if (aux_op==op_VAR) { /* X := E */
    id=eval_var_name(t,R);
    if (math_error == object_NULL) {
      if (strcmp(id,"MEMORY") == 0) {
        res_o=cocoa_error("CANNOT_ASSIGN");
        error_set_this_param(cterm_mk_NBS("to MEMORY"));
      } else {
        env_declare_and_assign(&current_env,id,o);
      }
    }
    free_str1(id); }

  else if (aux_op==op_COMPONENT) { /* L[N] := E */
    cterm_list params=cterm_params(t);
    object aux_o;
    aux_o=eval_caller(params[0],R); /* params[0] read as right expr */
    if (object_is_error(aux_o))
      res_o=aux_o;
    else {
      objtype ot;
      ot=object_type(aux_o);
      if (ot != objtype_olist && ot != objtype_omat) {
        res_o=cocoa_error("CANNOT_ASSIGN");
        error_set_this_param(cterm_mk_NBS("(expected list or matrix)")); }
      else {
        int aux_i;
        olist aux_ol;
        aux_i=eval_int(params[1],R);
        if (math_error==object_NULL) {
          aux_label=my_strdup(object_label(aux_o));
          aux_ol=object2olist(aux_o);
          if (aux_i<1) {
            error_set_this_param(int2cterm(aux_i));
            math_error=cocoa_error("INDEX_NEG"); }
          else if (aux_i>olist_len(aux_ol)) {
            error_set_this_param(int2cterm(aux_i));
            math_error=cocoa_error("INDEX_TOO_BIG"); }
          else {
            /* aux_ol=olist_dup(aux_ol); */ /* TODO: remove this */
	    object_unlink(olist_nth(aux_ol,aux_i));
            olist_set_nth(aux_ol,aux_i,object_link(o)); 
            aux_o=olist2object(aux_ol);
            object_set_label(aux_o,aux_label);
            object_set_type(aux_o,ot);
            res_o=engine_assign_this(params[0],aux_o,R); } } } } }

  else if (aux_op==op_RECORDSEL) { /* R.X := E */
    object aux_o;
    if (cterm_op(cterm_param(t,0)) == op_VAR) {
      id=eval_var_name(cterm_param(t,0),R);
      if (strcmp(id,"MEMORY") == 0) {
	free_str1(id);
        id=eval_var_name(cterm_param(t,1),R);
        env_declare_and_assign(&cocoa_env,id,o);
	done_memory=TRUE;
      } else if (strcmp(id,"PKG") == 0) {
	free_str1(id);
        id=eval_var_name(cterm_param(t,1),R);
        env_declare_and_assign(&pkg_env,id,o);
	done_memory=TRUE;
      } else if (strcmp(id,"ERR") == 0) {
	free_str1(id);
        id=eval_var_name(cterm_param(t,1),R);
        env_declare_and_assign(&err_env,id,o);
	done_memory=TRUE;
      } else if (strcmp(id,"DEV") == 0) {
	free_str1(id);
        id=eval_var_name(cterm_param(t,1),R);
        env_declare_and_assign(&cdev_env,id,o);
	done_memory=TRUE;
      } else if (strcmp(id,"ENV") == 0) {
	free_str1(id);
        res_o=cocoa_error("CANNOT_ASSIGN");
        error_set_this_param(cterm_mk_NBS("MEMORY.ENV"));
	done_memory=TRUE;
/*      } else if (strcmp(id,"CURR") == 0) {
	free_str1(id);
        id=eval_var_name(cterm_param(t,1),R);
        env_declare_and_assign(&current_env,id,o);
	done_memory=TRUE;
*/
      }
      free_str1(id);
    }
    if (!done_memory) {
      aux_o=eval_caller(cterm_param(t,0),R); /* params[0] read as right expr */
      aux_label=my_strdup(object_label(aux_o));
      if (object_is_error(aux_o))
        res_o=aux_o;
      else {
        objtype ot;
        ot=object_type(aux_o);
        if (ot == objtype_module) {
          engine_assign_module_field(object2module(aux_o,R),t,o,aux_label,
	    object_ring(aux_o));
        } else if (ot == objtype_ideal) {
          engine_assign_ideal_field(object2module(aux_o,R),t,o,aux_label,
	    object_ring(aux_o));
        } else if (ot == objtype_ring) {
          id=eval_var_name(cterm_param(t,1),R);
          if (math_error==object_NULL)
            env_declare_and_assign(&ring_local_env(aux_o->v.ring_value),id,o);
	  object_unlink(aux_o);
	  free_str1(id);
        } else if (ot != objtype_record) {
          res_o=cocoa_error("CANNOT_ASSIGN");
          error_set_this_param(cterm_mk_NBS("(expected record)"));
        } else {
          record aux_r;
          char * id;
	  if (aux_o->v.record_value == ring_env) {
            res_o=cocoa_error("CANNOT_ASSIGN");
            error_set_this_param(cterm_mk_NBS("MEMORY.ENV"));
	    object_unlink(aux_o);
	  } else {
            aux_r=object2record(aux_o);
            id=eval_var_name(cterm_param(t,1),R);
            if (math_error==object_NULL) {
              record_update_field(&aux_r,id,o);
              aux_o=record2object(aux_r);
              object_set_label(aux_o,aux_label);
              res_o=engine_assign_this(cterm_param(t,0),aux_o,R); 
              free_str1(id);
	    }
          }
        } 
      } 
    }
  } else {
    error_set_context(t);
    res_o=cocoa_error("BAD_LEFT_EXPR"); }

  if (res_o!=object_NULL) { /* if LEXPR is wrong the error is returned */
    object_unlink(o);
    return res_o; }
  else if (math_error!=object_NULL) return math_error;
  else if (object_is_error(o)) /* LEXPR is ok, but o is an error (returned) */
    return o;
  else { /* LEXPR and o are ok */
    object_unlink(o);
    return object_link(object_null); }
}

object engine_assign_ideal_field(module M, cterm t, object o, char * l, ring R)
{
  char * id;
  module aux_M;
  object aux_o,res_o;
  id=eval_var_name(cterm_param(t,1),R);
  if (math_error!=object_NULL) {
    free_str1(id);
    return math_error; }
  if (module_rc(M) > 1) {
    aux_M=M;
    M=module_dup(aux_M,R);
    module_unlink(aux_M,R);
  }
  module_update_field(&M,id,o,R);
  free_str1(id);
  if (math_error != object_NULL) {
/*    module_free(M, R);*/
    return math_error; }
  aux_o=ideal2object(M,R);
  object_set_label(aux_o,l);
  res_o=engine_assign_this(cterm_param(t,0),aux_o,R);
  return res_o;
}

object engine_assign_module_field(module M, cterm t, object o, char * l, ring R)
{
  char * id;
  module aux_M;
  object aux_o,res_o;

  id=eval_var_name(cterm_param(t,1),R);
  if (math_error!=object_NULL) {
    free_str1(id);
    return math_error; }
  if (module_rc(M) > 1) {
    aux_M=M;
    module_unlink(M,R);
    M=module_dup(aux_M,R);
  }
  module_update_field(&M,id,o,R);
  free_str1(id);
  if (math_error != object_NULL) {
/*    module_free(M, R);*/
    return math_error; }
  aux_o=module2object(M,R);
  object_set_label(aux_o,l);
  res_o=engine_assign_this(cterm_param(t,0),aux_o,R);
  return res_o;
}

void res_update_trunc(module M)
{
  module aux_M;
  aux_M=M;
  while (module_syz(aux_M) != module_NULL) {
    module_set_ResTrunc(module_syz(aux_M),module_ResTrunc(aux_M)-1);
    module_set_RegTrunc(module_syz(aux_M),module_RegTrunc(aux_M)+1);
    module_set_DegTrunc(module_syz(aux_M),module_DegTrunc(aux_M));
    aux_M=module_syz(aux_M); }
}

int module_update_field(module * M, char * id, object o, ring R)
{
  int n;
  module aux_M;
  olist aux_ol;
  pvector aux_v;
  vlist aux_vl;
  if (strcmp(id,"DegTrunc")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_DegTrunc(*M,n);
    res_update_trunc(*M); }
  else if (strcmp(id,"RegTrunc")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_RegTrunc(*M,n);
    res_update_trunc(*M); }
  else if (strcmp(id,"ResTrunc")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_ResTrunc(*M,n);
    res_update_trunc(*M); }
  else if (strcmp(id,"Discrepancy")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_discrepancy(*M,n); }
  else if (strcmp(id,"ElimIndets")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_ElimIndets(*M,n); }
  else if (strcmp(id,"SyzCompsNum")==0) {
    n = object_cast_int(object_link(o)); /* TODO: error checking */
    module_set_SyzCompsNum(*M,n); }
  else if (strcmp(id,"KFL")==0) {
    module_gb_set_initialized(*M,FALSE); /* TODO: ci sono controindicazioni? */
    aux_ol = object_cast_olist(object_link(o)); /* TODO: error checking */
    module_link_kf(*M,aux_ol,R);
    olist_free(aux_ol); }
  else if (strcmp(id, "SyzOfGens")==0)
  {
    if ((*M)->syz_of_gens != module_NULL)
    {
      math_error = cocoa_error("CANNOT_ASSIGN");
      return 1;
    }
    (*M)->syz_of_gens = object_cast_module(object_link(o), R);
    if (math_error != object_NULL)
    {
      (*M)->syz_of_gens = module_NULL;
      return 1;
    }
  }
  else if (strcmp(id,"GBasis")==0) {
    aux_ol = object_cast_olist(object_link(o)); /* TODO: error checking */
    aux_vl = this_olist2vlist_bis(aux_ol,R);
    vlist_free(&module_gbasis(*M), R);
    module_set_gbasis(*M,aux_vl);
    if (gb_Rules(*M) == vlist_NULL) {
      vlist_atom aux_atom;
      int n;
      gb_set_Rules(*M,vlist_init());
      n=vlist_len(aux_vl);
      aux_atom=vlist_first(aux_vl);
      while (n-- > 0) {
        if (!ev_is_zero(vlist_atom_ev(aux_atom)))
          gb_Add_To_Rules_Append(*M,vlist_atom_ev(aux_atom)); /* TODO: TRICK */
        aux_atom=vlist_atom_next(aux_atom); }
      }
  } /* TODO: error checking */
  else if (strcmp(id,"MinGens")==0) {
    aux_ol = object_cast_olist(object_link(o)); /* TODO: error checking */
    aux_vl = this_olist2vlist_bis(aux_ol,R);
    vlist_free(&module_mingens(*M),R);
    module_set_mingens(*M,aux_vl); } /* TODO: error checking */
  else if (strcmp(id,"Rules")==0) {
    aux_ol = object_cast_olist(object_link(o)); /* TODO: error checking */
    aux_vl = this_olist2vlist_bis(aux_ol,R);
    vlist_to_rules(aux_vl,R);
    vlist_free(&gb_Rules(*M),R);
    gb_set_Rules(*M,aux_vl); } /* TODO: error checking */
  else if (strcmp(id,"SyzAreOfMinGens")==0) {
    coc_bool aux_b;
    aux_b = object_cast_bool(object_link(o));
    if (aux_b) module_set_flag_syz_of_mingens(*M);
    else module_reset_flag_syz_of_mingens(*M); }
  else if (strcmp(id,"CV")==0) {
    aux_v = object_cast_pvector(object_link(o),R);
    ev_unlink(gb_CV(*M),R);
    /* gb_set_CV(*M,ev_init(aux_v,R)); */
    {
      evector tmp = ev_init(aux_v,R);
      ev_apply_shifts(tmp,module_shifts(*M),R);
      gb_set_CV(*M,tmp);
    }
  }
  else if (strcmp(id,"PSeries")==0) {
    aux_ol = object_cast_olist(object_link(o));
    if (module_pseries(*M) != pseries_NULL) pseries_free(module_pseries(*M));
    module_set_pseries(*M,olist2pseries(aux_ol)); 
    olist_free(aux_ol);
  } /* TODO: error checking */
  else {
    math_error=cocoa_error("UNKN_FIELD");
    error_set_this_param(string2cterm(id));
    return 0; /* TODO: return error */
  }
  return 0; /* no errors */
}

void cocoa_declare_var(cterm t,env * e,ring R)
{
  cterm_list params=cterm_params(t);
  int arity=cterm_arity(t)-1;
  char * id;
  id = cterm_string(t);
  env_declare(e,id);
}


/* error management */

object cocoa_error(char * id)
{
  object aux_o, aux_o1;
  aux_o=env_get_value(cocoa_env,"ERR");
  if (aux_o == object_NULL)
    return error2object(my_strdup("Error Unknown"));
  if (object_type(aux_o) != objtype_record) {
    object_unlink(aux_o);
    return error2object(my_strdup("Error Unknown")); }
  aux_o1=record_sel(aux_o->v.record_value,id);
  object_unlink(aux_o);
  if (aux_o1==object_NULL)
    return error2object(my_strdup("Error Unknown"));
  object_link(aux_o1);
  return aux_o1;
}

object cocoa_error_not_impl(char * s)
{
  error_set_this_param(cterm_mk_NBS(s));
  return cocoa_error("NOT_YET_IMPL");
}

object cocoa_internal_error(char * msg)
{
  char buffer[100];
  object res_o;
  res_o=cocoa_error("INTERNAL");
  sprintf(buffer,"[%s]",msg);
  error_set_this_param(cterm_mk_NBS(buffer));
  return res_o;
}

object error_expected_found(objtype t1, objtype t2)
{
  object res_o;
  char buffer[100];
  res_o=cocoa_error("EXPECTED");
  sprintf(buffer,"%s but found %s",objtype2str(t1),objtype2str(t2));
  error_set_this_param(cterm_mk_NBS(buffer));
  return res_o;
}

object error_cannot_apply(char * op, objtype t)
{
  object res_o;
  char buffer[100];
  res_o=cocoa_error("CANNOT_APPLY");
  sprintf(buffer,"%s to %s",op,objtype2str(t));
  error_set_this_param(cterm_mk_NBS(buffer));
  return res_o;
}

/* to be moved to cterm.h (?) */
cterm err2cterm(char * s)
{
  cterm params[2];
  params[0]=cterm_mk_NBS(s);
  return cterm_mk_classic(op_ERROR,1,params);
}

/* ---[ environments ]-------------------------------------------- */

void ring_env_assign(ring R)
{
  env_declare_and_assign(&ring_env,R->ring_name,ring2object(R));
}

ring ring_env_get(char * id)
{
  object aux_o;
  aux_o=env_get_value(ring_env,id);
  if (aux_o==object_NULL) return ring_NULL;
  return object2ring(aux_o);
}

object ring_env_delete(ring R)
{ /* R != ring_NULL */
  char buffer[50];
  char * id, * new_id;
  id = ring_id(R);
  if (ring_locked(R)) {
    sprintf(buffer,"ring %s (is locked)",id);
    error_set_this_param(cterm_mk_NBS(buffer));
    return cocoa_error("CANNOT_DEL");
  }
  ring_clear(R);
  if (ring_rc(R) > 1) { /* some objects still survive, rename the ring */
    new_id=env_newvar(id);
    ring_set_id(R,new_id);
    ring_unlink(R); /* trick */
    env_rename_var(ring_env,id,new_id);
    /* env_delete_var(&ring_env,id); */
    free_str1(id);
  } else {
    env_delete_var(&ring_env,id);
  }
  return object_link(object_null);
}

void panels_env_assign(char * id, opt_panel p,ring R)
{
  env_declare_and_assign(&panels_env,id,panel2object(p));
}

opt_panel panels_env_get(char * id)
{
  object aux_o;
  aux_o=env_get_value(panels_env,id);
  if (aux_o==object_NULL) return opt_panel_NULL;
  return object2panel(aux_o);
}
