/* $Id: evector.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ : extended vectors */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "coeffs.h"
#include "eterms.h"
#include "evector.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "options.h"
#include "polys.h"
#include "pvector.h"
#include "vlist.h"
#include "terms.h"

pvector ev_vector(evector E)
{
  if (ev_irr(E) == mv_NULL && ev_tail(E) == pv_NULL) return ev_red(E);
  INTERNAL_ERROR("[ev_vector()] : evector is split\n");
  return pv_NULL;
}

pvector ev_whole_vector(evector E, ring R)
{
  pvector res_pv;
  if (ev_irr(E) == mv_NULL) {
    if (ev_tail(E) == pv_NULL) {
      res_pv=pv_dup(ev_red(E),R);
    } else {
      res_pv=pv_dup(ev_tail(E),R);
      pv_set_nth(res_pv,ev_lmpos(E),
        pvcomp_sum_this_monomial_list_to_this_pvcomp(
          monomial_dup(ev_lm(E),R),pv_nth(res_pv,ev_lmpos(E)),R));
    }
    return res_pv;
  }
  res_pv=this_pv_plus_this_mv(pv_dup(ev_red(E),R),mv_dup(ev_irr(E),R),R);
  return this_pv_plus_this_monomial(res_pv,
    monomial_dup(ev_lm(E),R),ev_lmpos(E),R);
}

void ev_extract_lmred(evector V, monomial * lmred, int * lmpos, ring R)
{
  if (ev_tail(V) != pv_NULL) {
    *lmred=monomial_dup(ev_lm(V),R);
    *lmpos=ev_lmpos(V);
    ev_set_red(V,ev_tail(V));
    ev_set_tail(V,pv_NULL);
  } else {
    pv_extract_lm(ev_red(V),lmred,lmpos,R);
  }
}

void ev_reinit(evector V, ring R)
{
  /* the vector is put in ev_red(V) */
  if (ev_irr(V) != mv_NULL) {
    ev_irr_plus_red_to_tail(V,R); /* [REDUCING] -> [REDUCED] */
  }
  if (ev_tail(V) != ev_NULL) { /* [REDUCED] -> [INIT] */
    ev_set_red(V,ev_tail(V));
    ev_set_tail(V,ev_NULL);
    pv_set_nth(ev_red(V),ev_lmpos(V),
      pvcomp_add_this_monomial(pv_nth(ev_red(V),ev_lmpos(V)),
        monomial_dup(ev_lm(V),R),R));
  }
}

void ev_to_rule(evector V, ring R)
{
  if (ev_tail(V) == pv_NULL) {
    ev_irr_plus_red_to_tail(V,R);
    if (ev_eterm(V) == NULL) ev_set_eterm(V,term2eterm(ev_lt(V)));
  }
}

void ev_irr_plus_red_to_tail(evector V, ring R)
{
  if (ev_irr(V) == mv_NULL) {
    pv_set_nth(ev_red(V),ev_lmpos(V),
      pvcomp_remove_lm(pv_nth(ev_red(V),ev_lmpos(V)),R));
    ev_set_tail(V,ev_red(V));
  } else {
    ev_set_tail(V,this_pv_plus_this_mv(ev_red(V),ev_irr(V),R));
    ev_set_irr(V,mv_NULL);
  }
  ev_set_red(V,pv_NULL);
}

mvector mv_x_monomial(mvector mv, monomial m, ring R)
{
  mvector res_mv;
  int i,r;
  if (mv == mv_NULL) return mv_NULL;
  r = mv_rank(mv);
  res_mv=mv_init(r);
  for (i=0;i<r;i++) {
    mv_set_nth(res_mv,i,monomial_list_x_monomial(mv_nth(mv,i),m,R));
  }
  return res_mv;
}

pvector this_pv_plus_this_mv(pvector pv, mvector mv, ring R)
{
  int i,n;
  if (mv == mv_NULL) return pv;
  n=MIN(pv_rank(pv),mv_rank(mv));
  for (i=0;i<n;i++) {
    pv_set_nth(pv,i,
     pvcomp_sum_this_monomial_list_to_this_pvcomp(mv_nth(mv,i),pv_nth(pv,i),R));
    mv_set_nth(mv,i,NULL);
  }
  mv_free(mv,R);
  return pv;
}

cterm mv2cterm(mvector v, shifts ss, ring R)
{
  cterm_list params;
  cterm res_t;
  register int i;
  params=malloc_cterm_list(mv_rank(v));
  if (ss != shifts_NULL) {
    mv_Apply_Anti_Shifts(v,ss,R);
    for (i=0;i<pv_rank(v);i++)
      params[i]=monopoli2cterm(mv_nth(v,i),R);
    mv_Apply_Shifts(v,ss,R); }  /* TAPPUL! (TODO) */
  else
    for (i=0;i<pv_rank(v);i++)
      params[i]=monopoli2cterm(mv_nth(v,i),R);
  res_t=cterm_mk_classic(op_VECTOR,mv_rank(v),params);
  free_cterm_list(mv_rank(v),params);
  return res_t;
}

void mv_Apply_Anti_Shifts(mvector v,shifts ss,ring R)
{
  TODO("mv_Apply_Anti_Shifts");
}

void mv_Apply_Shifts(mvector v,shifts ss,ring R)
{
  if (v != mv_NULL) {
    TODO("mv_Apply_Shifts");
  }
}

void mv_force_rank(mvector * v, int r)
{
  int i;
  if (r <= mv_rank(*v)) return;
  if (mv_comp(*v) == NULL) mv_set_comp(*v,malloc_mv_comp(r));
  else mv_set_comp(*v,realloc_mv_comp(mv_comp(*v),mv_rank(*v),r));
  for (i=mv_rank(*v);i<r;i++) mv_set_nth(*v,i,NULL);
  mv_set_rank(*v,r);
}

mvector mv_init(int r)
{
  mvector res_v;
  register int i;
  res_v=malloc_mv();
  mv_set_rank(res_v,r);
  if (r != 0) {
    mv_set_comp(res_v,malloc_mv_comp(r));
  } else {
    mv_set_comp(res_v,NULL);
  }
  for (i=0;i<r;i++) mv_set_nth(res_v,i,NULL);
  return res_v;
}

mvector mv_dup(mvector v, ring R)
{
  mvector res_v;
  register int i;
  if (v==mv_NULL) return mv_NULL;
  res_v=mv_init(mv_rank(v));
  for (i=0;i<mv_rank(v);i++)
    mv_set_nth(res_v,i,monomial_list_dup(mv_nth(v,i),R));
  return res_v;
}

void mv_free(mvector v, ring R)
{
  int i;
  if (v != mv_NULL) {
    for (i=0;i<mv_rank(v);i++) monopoli_free(mv_nth(v,i),R);
    if (mv_rank(v)>0) free_mv_comp(mv_rank(v),mv_comp(v));
    free_mv(v); }
}

evector ev_dup(evector v,ring R)
{
  evector res_v;
  res_v=malloc_ev();
  ev_set_red(res_v,pv_dup(ev_red(v),R));
  ev_set_irr(res_v,mv_dup(ev_irr(v),R));
  ev_set_tail(res_v,pv_dup(ev_tail(v),R));
  ev_set_td(res_v,ev_td(v));
  ev_set_wd(res_v,ev_wd(v));
  ev_set_lm(res_v,monomial_dup(ev_lm(v),R));
  ev_set_lmpos(res_v,ev_lmpos(v));
  ev_set_rel(res_v,pv_dup(ev_rel(v),R));
  ev_set_rc(res_v,1);
  ev_set_label(res_v,ev_NOLABEL);
  ev_set_Is_SPoly(res_v,ev_Is_SPoly(v));
  if (ev_eterm(v) != NULL) ev_set_eterm(res_v,eterm_dup(ev_eterm(v)));
  else ev_set_eterm(res_v,NULL);
  return res_v;
}

evector ev_init(pvector v, ring R)
{
  evector res_v;
  monomial m;
  int pos;
  res_v=malloc_ev();
  pv_leading_monomial(v,&m,&pos,R);
  ev_set_red(res_v,v);
  ev_set_irr(res_v,pv_NULL);
  ev_set_tail(res_v,pv_NULL);
  ev_set_td(res_v,0);
  ev_set_wd(res_v,0);
  ev_set_lm(res_v,m);
  ev_set_lmpos(res_v,pos);
  ev_set_rel(res_v,pv_NULL);
  ev_set_label(res_v,ev_NOLABEL);
  ev_set_rc(res_v,1);
  ev_set_Is_SPoly(res_v,FALSE);
  ev_set_eterm(res_v,NULL);
  return res_v;
}

evector ev_init2(monomial lm, int lmpos, pvector tail)
{
  evector res_v;
  monomial m;
  int pos;
  res_v=malloc_ev();
  ev_set_red(res_v,pv_NULL);
  ev_set_irr(res_v,pv_NULL);
  ev_set_tail(res_v,tail);
  ev_set_td(res_v,0);
  ev_set_wd(res_v,0);
  ev_set_lm(res_v,lm);
  ev_set_lmpos(res_v,lmpos);
  ev_set_rel(res_v,pv_NULL);
  ev_set_label(res_v,ev_NOLABEL);
  ev_set_rc(res_v,1);
  ev_set_Is_SPoly(res_v,FALSE);
  ev_set_eterm(res_v,NULL);
  return res_v;
}


void ev_update_lm(evector v, monomial lm, int lmpos, ring R)
{
  if (ev_lm(v) != NULL) monomial_free(ev_lm(v),R);
  ev_set_lm(v,lm);
  ev_set_lmpos(v,lmpos);
  if (ev_eterm(v) != NULL) {
    eterm_free(ev_eterm(v));
    if (lm != NULL) ev_set_eterm(v,term2eterm(ev_lt(v)));
  }
}

void ev_update(evector v, ring R)
{
/* Update the extended vector v by means of its vector */
/* NOTE: if it was marked it remain marked and viceversa */
  int lmpos;
  monomial lm;
  pv_leading_monomial(ev_vector(v),&lm,&lmpos,R);
  ev_update_lm(v,lm,lmpos,R);
}

/* End Of Todo */

void ev_free(evector * v, ring R)
{
  if (ev_rc(*v)!=ZERO) {
    my_fputs("INTERNAL ERROR: attempt to free a referenced evector\n",stdout);
    return;
  }
  mv_free(ev_irr(*v),R);
  pv_free(ev_rel(*v),R);
  pv_free(ev_red(*v),R);
  pv_free(ev_tail(*v),R);
  if (ev_lm(*v) != NULL) monomial_free(ev_lm(*v),R);
  if (ev_eterm(*v) != NULL) eterm_free(ev_eterm(*v));
  free_ev(*v);
  *v=ev_NULL;
}

void ev_make_monic(evector ev, ring R)
{
  monomial lm;
  pv_make_monic(ev_vector(ev),R);
  /* pv_make_monic(ev_rel(ev),R); */ /* TODO QUESTION: serve? */
  lm=ev_lm(ev);
  R->ClearCoeff(&monomial_coeff(lm));
  monomial_set_coeff(lm,coeff_one(R));
}

void ev_add_monomial_to_irr(evector ev, monomial m, int pos, ring R)
{
  mvector mv=ev_irr(ev);
  if (mv == mv_NULL) {
    mv=mv_init(0);
    ev_update_lm(ev,m,pos,R);
  }
  else {
    if (mv_rank(mv) <= pos) mv_force_rank(&mv,pos+1);
    monomial_set_next(m,mv_nth(mv,pos));
    mv_set_nth(mv,pos,m);
  }
  ev_set_irr(ev,mv);
}

void ev_force_rc1(evector * ev, ring R)
{
  if (ev_rc(*ev)==1) return;
  ev_unlink(*ev,R);
  *ev=ev_dup(*ev,R);    
}

void printf_ev(evector ev, ring R)
{
  cterm aux_t;
  aux_t=ev2cterm(ev,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

cterm ev2cterm_normal(evector ev, ring R)
{
  pvector aux_v;
  cterm res_t;
  if (ev_is_split(ev)) {
    aux_v=ev_whole_vector(ev,R);
    res_t=pv2cterm2(aux_v,R);
    pv_free(aux_v,R);
  } else
    res_t=pv2cterm2(ev_vector(ev),R);
  return res_t;
}

cterm ev2cterm(evector ev, ring R)
{
  cterm aux_params[2],aux_t;
  if (ev_is_split(ev)) {
    aux_params[0]=mv2cterm(ev_irr(ev),pv_shifts(ev_red(ev)),R);
    aux_params[1]=pv2cterm(ev_red(ev),R);
    aux_t=cterm_mk_classic(op_SUM,2,aux_params);
  } else 
    aux_t=pv2cterm(ev_vector(ev),R);
  if (ev_is_labelled(ev)) {
    aux_params[0]=int2cterm(ev_label(ev));
    aux_params[1]=aux_t;
    aux_t=cterm_mk_classic(op_COLON,2,aux_params);
  }
  if (ev_rel(ev) != pv_NULL && flag_is_on(opt_panel_GROBNER,flag_GB_VERBOSE)) {
    aux_params[0]=aux_t;
    aux_params[1]=pv2cterm(ev_rel(ev),R);
    aux_t=cterm_mk_classic(op_COLON,2,aux_params);
  }
  return aux_t;
}

void ev_apply_shifts(evector ev, shifts ss, ring R)
{
  if (ss==shifts_NULL) return;
  pv_Apply_Shifts(ev_red(ev),ss,R);
  mv_Apply_Shifts(ev_irr(ev),ss,R);
}
