/* $Id: gb.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $

[ ENTRY POINTS ] [ ADT ] [ Init() ] [ Do_1() ] [ Do_Case_0() ] [
Do_SPoly() ]
[ Do_Get_Rule() ] [ Do_Apply_Rule() ] [ Do_Reduce() ] [ Do_Update() ]
[ Do_Make_Pair() ] [ Do_Add_To_Rules() ] [ Do_Add_To_IVs() ][
Do_Pair_Prec() ]
[ Do_Is_In_NF() ] [ Do_1() Aux ] [ Do_Case_Not_NF() ]

---------------------------------------------------------------- */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "cterm.h"
#include "evector.h"
#include "eterms.h"
#include "gb.h"
#include "ivectors.h"
#include "monomial.h"
#include "modupoin.h"
#include "module.h"
#include "object.h"
#include "olist.h"
#include "operator.h"
#include "pair.h"
#include "pvcomp.h"
#include "pseries.h"
#include "pvector.h"
#include "polys.h"
#include "ring.h"
#include "unipoly.h"
#include "terms.h"
#include "vlist.h"

coc_bool gb_ADD_LAST, gb_FLAG_JUST_LTNF;
void * gb_KF[gb_MAX_NUM_KEY_FUNCTIONS+1];
int gb_NUM_KEY_FUNCTIONS=0;
gb_errors gb_err=gb_err_NONE;
vpair vpair_dummy;


/* ---[ ENTRY POINTS ]----------------------------------------------- */

void gb_Init(module M)
{
  gb_Do_Init(M);
  module_gb_set_initialized(M,TRUE);
}

coc_bool gb_one_move(module M)
{
  if (!module_gb_initialized(M)) gb_Init(M);
  gb_FLAG_JUST_LTNF=module_flag_get(M,module_FLAG_JUST_LTNF);
  return gb_Do_1(M);
}

int gb_Do_All(module M)
{
  int n=0;
  if (!module_gb_initialized(M)) gb_Init(M);
  gb_FLAG_JUST_LTNF=module_flag_get(M,module_FLAG_JUST_LTNF);
  while (gb_Do_1(M))
    n++;
  return n;
}

void gb_free(module M)
{
  ring R;
  R = module_ring(M);
  vlist_free(&gb_Rules(M),R);
  vlist_free(&gb_IVs(M),R);
  FreeVPairs(&gb_IPs(M),R);
  if (gb_CV(M) != ev_NULL) ev_unlink(gb_CV(M),R);
  gb_set_Rules(M,vlist_NULL);
  gb_set_IVs(M,vlist_NULL);
  gb_set_IPs(M,vpairs_NULL);
}

void gb_dup(module res_M, module M)
{
  ring R;
  /* duplicate gb-stuff of M into res_M*/
  gb_kf_dup(res_M,M);
  R = module_ring(M);
  gb_set_Rules(res_M,vlist_dup(gb_Rules(M),R));
  gb_set_IVs(res_M,vlist_dup(gb_IVs(M),R));
  gb_set_IPs(res_M,vpairs_dup(gb_IPs(M),R));
  if (gb_CV(M) != ev_NULL) gb_set_CV(res_M,ev_link(gb_CV(M)));
}

void gb_kf_dup(module res_M, module M)
{
  int i;
  for (i=0;i<gb_NUM_MODULE_KEY_FUNCTIONS;i++)
    gb_set_kf(res_M,i,gb_kf(M,i));
}

void gb_kf_init(module res_M)
{
  int i;
  module_gb_set_initialized(res_M,FALSE);
  for (i=0;i<gb_NUM_MODULE_KEY_FUNCTIONS;i++)
    gb_set_kf(res_M,i,gb_kf_NONE);
}

cterm gb_kf2cterm(module M)
{
  int i;
  cterm params[gb_MAX_NUM_KEY_FUNCTIONS+1];
  for (i=0;i<gb_NUM_MODULE_KEY_FUNCTIONS;i++) {
    params[i]=int2cterm(gb_kf(M,i));
  }
  return cterm_noindent(
    cterm_mk_classic(op_LIST,gb_NUM_MODULE_KEY_FUNCTIONS,params));
}

olist gb_kf2olist(module M)
{
  int i;
  olist res_ol;
  res_ol=olist_init(gb_NUM_MODULE_KEY_FUNCTIONS);
  for (i=0;i<gb_NUM_MODULE_KEY_FUNCTIONS;i++) {
    olist_set_nth(res_ol, i+1, int2object(gb_kf(M,i)));
  }
  return res_ol;
}

olist gb_stats2olist(module M)
{
  int i;
  olist res_ol;
  res_ol=olist_init(gb_NUM_STATS);
  for (i=0;i<gb_NUM_STATS;i++) {
    olist_set_nth(res_ol, i+1, int2object(module_stat(M,i)));
  }
  return res_ol;
}

void gb_init()
{
  int i;
  gb_KF[0]=NULL;
  gb_KF[1]=(void *)gb_Do_1_GB;
  gb_KF[2]=(void *)gb_Do_1_HGB;
  gb_KF[3]=(void *)gb_Do_1_Res_Row_Trunc;
  gb_KF[4]=(void *)gb_Do_Case_0_Nothing;
  gb_KF[5]=(void *)gb_Do_Case_0_Syz_MGS;
  gb_KF[6]=NULL; /* FREE */
  gb_KF[7]=(void *)gb_SPoly_GB;
  gb_KF[8]=(void *)gb_SPoly_Syz;
  gb_KF[9]=(void *)gb_Get_Rule_First;
  gb_KF[10]=(void *)gb_Apply_Rule_GB;
  gb_KF[11]=(void *)gb_Apply_Rule_Syz;
  gb_KF[12]=(void *)gb_Reduce_LT_1;
  gb_KF[13]=(void *)gb_Reduce_Full_1;
  gb_KF[14]=(void *)gb_Reduce_LT;
  gb_KF[15]=(void *)gb_Reduce_Full;
  gb_KF[16]=(void *)gb_Update_Basic;
  gb_KF[17]=(void *)gb_Update_MinGens;
  gb_KF[18]=(void *)gb_Make_Pair_NoSugar;
  gb_KF[19]=(void *)gb_Make_Pair_Sugar;
  gb_KF[20]=(void *)gb_Add_To_Rules_Append;
  gb_KF[21]=(void *)gb_Add_To_IVs_Append;
  gb_KF[22]=(void *)gb_Add_To_IVs_AntiDeg;
  gb_KF[23]=(void *)gb_Pair_Prec_AntiDegRevLex;
  gb_KF[24]=(void *)gb_Pair_Prec_AntiDeg;
  gb_KF[25]=(void *)gb_Pair_Prec_SugarAntiLex;
  gb_KF[26]=(void *)gb_Pair_Prec_Sugar;
  gb_KF[27]=(void *)gb_Is_In_NF_True;
  gb_KF[28]=(void *)gb_Is_In_NF_LT;
  gb_KF[29]=(void *)gb_Is_In_NF_Full;
  gb_KF[30]=(void *)gb_Do_Case_0_Syz_MG;
  gb_KF[31]=(void *)gb_Update_Syz_MGS;
  gb_KF[32]=(void *)gb_Update_Syz_MG;
  gb_KF[33]=(void *)gb_Do_1_HGB_Trunc;
  gb_KF[34]=(void *)gb_Do_1_Res_Row_Hilb_Trunc;
  gb_KF[35]=(void *)gb_Update_Syz_MGS_Hilb;
  gb_KF[36]=(void *)gb_Do_1_MinGens_Trunc;
  gb_KF[37]=(void *)gb_Do_1_MinSyzMinGens_Trunc;
  gb_KF[38]=(void *)gb_Do_Case_Not_NF_Add_IVs;
  gb_KF[39]=(void *)gb_Do_Case_Not_NF_Keep;
  gb_KF[40]=(void *)gb_Do_1_Res_Row;
  gb_KF[41]=(void *)gb_Do_1_Res_Row_Hilb;
  gb_KF[42]=(void *)gb_Init_Basic;
  gb_KF[43]=(void *)gb_Init_Syz;
  gb_KF[44]=(void *)gb_Do_Case_0_Syz_G;
  gb_KF[45]=(void *)gb_Do_1_Syz;
  gb_KF[46]=(void *)gb_Do_1_HGB_Hilb;
  gb_KF[47]=(void *)gb_Update_Discrepancy;
  gb_KF[48]=(void *)gb_Init_HSyz;
  gb_KF[49]=(void *)gb_Get_Rule_Product;
  gb_KF[50]=(void *)gb_Do_1_NF;
  gb_KF[51]=(void *)gb_Init_Nothing;
  gb_KF[52]=(void *)gb_Do_1_Interreduce;
  gb_KF[53]=(void *)gb_Init_Interreduce;
  gb_KF[54]=(void *)gb_Init_NF;
  gb_KF[55]=(void *)gb_Do_1_HGB_Hilb_Trunc;
  gb_KF[56]=(void *)gb_Init_BasicAndRels;
  gb_KF[57]=(void *)gb_Init_NFAndRel;

  /* FREE KF available */

  gb_NUM_KEY_FUNCTIONS=57;
  for (i = gb_NUM_KEY_FUNCTIONS+1; i <= gb_MAX_NUM_KEY_FUNCTIONS; i++)
    gb_KF[i]=NULL;

  vpair_dummy = malloc_vpair();
}

/* ---[ ADT ]----------------------------------------------- */

void syz_SPoly_Den( evector * SP, vpair myPair,ring R )
{ /* of monic */
  monomial M1, M2;
  pvector v1,v2,aux_v,r1,r2,aux_r;
  coeff aux_c;
  term aux_t;

  v1 = vpair_whole_first_vector(myPair,R);
  v2 = vpair_whole_second_vector(myPair,R);

  r1 = vpair_first_rel(myPair);
  r2 = vpair_second_rel(myPair);

  aux_c=coeff_inverse(vpair_first_lc(myPair),R);
  aux_t=term_div(vpair_lcm(myPair),vpair_first_lt(myPair));
  monomial_set_mk(M1,aux_c,aux_t);

  aux_c=coeff_inverse(vpair_second_lc(myPair),R);
  coeff_change_sign(&aux_c,R);
  aux_t=term_div(vpair_lcm(myPair),vpair_second_lt(myPair));
  monomial_set_mk(M2,aux_c,aux_t);

  /* SP = M1 * v1 + M2 * v2; */

  aux_v =

pv_sum_these(pv_this_x_monomial(v1,M1,R),pv_this_x_monomial(v2,M2,R),R);

  aux_r = pv_sum_these(pv_x_monomial(r1,M1,R),pv_x_monomial(r2,M2,R),R);

  *SP = ev_init(aux_v,R);
  ev_set_rel(*SP,aux_r);

  monomial_free(M1,R);
  monomial_free(M2,R);
}

void md_SPoly_Den( pvector * SP, vpair myPair,ring R )
{ /* of monic */
  monomial M1, M2;
  pvector v1,v2;
  coeff aux_c;
  term aux_t;

  v1 = vpair_whole_first_vector(myPair,R);
  v2 = vpair_whole_second_vector(myPair,R);

  aux_c=coeff_inverse(vpair_first_lc(myPair),R);
  aux_t=term_div(vpair_lcm(myPair),vpair_first_lt(myPair));
  monomial_set_mk(M1,aux_c,aux_t);

  aux_c=coeff_inverse(vpair_second_lc(myPair),R);
  coeff_change_sign(&aux_c,R);
  aux_t=term_div(vpair_lcm(myPair),vpair_second_lt(myPair));
  monomial_set_mk(M2,aux_c,aux_t);

  /* M1= LCM/LM(p1) */
  /* M2= - LCM/LM(p2) */
  /* SP = M1 * p1 + M2 * p2; */

  *SP =
pv_sum_these(pv_this_x_monomial(v1,M1,R),pv_this_x_monomial(v2,M2,R),R);

  monomial_free(M1,R);
  monomial_free(M2,R);
}

/* -----------------[ vlist ]------------------------------ */

vlist vlist_linear_comb(vlist vl, vlist coeffs, ring R)
{
  vlist res_vl;
  vlist_atom aux_atom;
  int n;
  if (coeffs==vlist_NULL) return vlist_NULL;
  n=vlist_len(coeffs);
  aux_atom=vlist_first(coeffs);
  res_vl = vlist_init();
  while (n-- > 0) {
    vlist_append_this_vector(res_vl,pv_linear_comb(vl,
      vlist_atom_vector(aux_atom),R),R);
    aux_atom=vlist_atom_next(aux_atom); }
  return res_vl;
}

pvector pv_linear_comb(vlist vl,pvector v, ring R)
{
  pvector res_v,aux_v;
  register int i;
  int n,r;
  vlist_atom aux_atom;
  aux_atom = vlist_first(vl);
  r = pv_rank(vlist_atom_vector(aux_atom));
  res_v=pv_zero(r,R);
  n = vlist_len(vl);
  for (i=0;i<n;i++) {
    aux_v = vlist_atom_vector(aux_atom);
    res_v=pv_sum_these(res_v,pv_mult(pv_nth(v,i),aux_v,R),R);
    aux_atom=vlist_atom_next(aux_atom); }
  return res_v;
}

int gb_Compute_Discrepancy_And_Degree(module M, int D, int * DD) {
  Unipoly aux_Pn, PnM;
  int mD, d, res_n;
  PSeries aux_S;

  aux_S = module_PS_of_quotient(M); /* della attuale base di Groebner */
  aux_Pn = pseries_num(aux_S);

  PnM = pseries_num(module_pseries(M));

  mD = MAX(UnipolyDegree(PnM),UnipolyDegree(aux_Pn));
  res_n = 0;
  for ( d=D ; d<=mD ; d++ )
  {
    res_n = UnipolySubCoeffsOfXd(aux_Pn, PnM, d);
    if ( res_n!=0 ) { *DD = d; break; }
  }
  pseries_free(aux_S);
  /* FreeUnipolyCheck(aux_Pn,R); */
  return res_n;
}  

int gb_Compute_Discrepancy(module M, int D) {
  Unipoly PnFM, PnM, PnS;
  coeff aux_c, PnFMc,PnMc, PnSc;
  module S;
  int res_n;

  S=module_syz(M);

  module_compute_PS_of_quotient_trunc(M,D-1);
  module_compute_PS_of_quotient_trunc(S,D);

  /* assumo sia il Pnumeratore della Ps del Modulo non del quoziente */
  PnM = pseries_num(module_pseries(M));
  PnS = pseries_num(module_pseries(S));
  /* PnFM = PnumFreemodule(M); */

#ifdef DEBUG_GB
  if (flag_is_on(opt_panel_DEBUG, flag_DEBUG_GB)) {
    printf_string("\n");
    printf_string("Poincare(M)      = ");
    printf_unipoly(PnM,R);
    printf_string("\n");
    printf_string("Poincare(Syz(M)) = ");
    printf_unipoly(PnS,R);
    printf_string("\n");
/*
    printf_string("Poincare(F(M))   = ");
    printf_unipoly(PnFM,R);
    printf_string("\n");
*/
  }
#endif

  res_n = UnipolyAddCoeffsOfXd(PnM, PnS, D);

#ifdef DEBUG_GB
  if (flag_is_on(opt_panel_DEBUG, flag_DEBUG_GB)) {
    printf("So Discrepancy(M,%d) = %d\n",D,res_n);
  }
#endif

/* FreeUnipoly(PnFM, R); */

  return res_n;
}

int gb_Kill_IP(module M, evector v1, evector v2)
{ /* returns the number of killed pairs */
  vpair aux_p, aux_p_next;
  int d;

  if (gb_IPs(M) == vpairs_NULL) return 0;

  d=term_deglcm(ev_lt(v1),ev_lt(v2));

  aux_p=vpairs_first(gb_IPs(M));
  if (vpair_first(aux_p) == v1 && vpair_second(aux_p) == v2) {
    gb_set_IPs(M,vpair_next(aux_p));
    FreeVPair(&aux_p,module_ring(M));
    return 1;
  }
  aux_p_next = vpair_next(aux_p);
  while (aux_p_next != vpairs_NULL && vpair_deg(aux_p_next) <= d) {
    if (vpair_first(aux_p_next) == v1 &&
        vpair_second(aux_p_next) == v2) {
      vpair_set_next(aux_p, vpair_next(aux_p_next));
      FreeVPair(&aux_p_next,module_ring(M));
      return 1;
    }
    aux_p = aux_p_next;
    aux_p_next = vpair_next(aux_p_next);
  }
  return 0;
}

int gb_Kill_IPs_Of_Degree_LE(module M, int D)
{ /* returns the number of killed pairs */
  vpair aux_p;
  int n=0;
  while (gb_Min_Deg_IPs(M)<=D) {
    aux_p=gb_Get_FIP(M);
    FreeVPair(&aux_p,module_ring(M));
    n++;
  }
  return n;
}

int gb_Kill_IVs_Of_Degree_LE(module M, int D)
{ /* return number of killed vectors */
  evector V;
  int n=0;
  while (gb_Min_Deg_IVs(M)<=D) {
    V=gb_Get_FIV(M);
    ev_unlink(V, module_ring(M));
    n++;
  }
  return n;
}

int gb_Kill_All_IPs(module M)
{ /* returns the number of killed pairs */
  vpair aux_p;
  int n=0;
  while (gb_Some_IP(M)) {
    aux_p=gb_Get_FIP(M);
    FreeVPair(&aux_p,module_ring(M));
    n++;
  }
  return n;
}

int gb_Kill_All_IVs(module M)
{ /* return number of killed vectors */
  evector V;
  int n=0;
  while (gb_Some_IV(M)) {
    V=gb_Get_FIV(M);
    ev_unlink(V, module_ring(M));
    n++;
  }
  return n;
}

void gb_Init_IVs(module M)
{ /* copy the generators of M into the incoming vectors */
  vlist_atom aux_atom;
  int n;
  gb_set_IVs(M,vlist_init());
  n=vlist_len(module_gens(M));
  aux_atom=vlist_first(module_gens(M));
  while (n-- > 0) {
    gb_Do_Add_To_IVs(M,vlist_atom_ev(aux_atom));
    aux_atom=vlist_atom_next(aux_atom); }
}

void gb_Init_IVs_Syz(module M)
{ /* copy the generators of M into the incoming vectors */
  vlist_atom aux_atom;
  shifts ss;
  evector aux_ev;
  pvector rel;
  int i,n,cn;
  ring R;
  R = module_ring(M);
  gb_set_IVs(M,vlist_init());
  n=vlist_len(module_gens(M));
  aux_atom=vlist_first(module_gens(M));
  ss = shifts_NULL;
  cn = MIN(n,module_SyzCompsNum(M));
  if (module_syz(M) != module_NULL) ss=module_shifts(module_syz(M));
  for (i=0;i<n;i++) {
    aux_ev=ev_dup(vlist_atom_ev(aux_atom),R);
    if (cn < i+1)
      ev_set_rel(aux_ev,pv_zero(cn,R));
    else
      ev_set_rel(aux_ev,pv_gen(cn,i,R));
    gb_Do_Add_To_IVs(M,aux_ev);
    ev_unlink(aux_ev,R); /* MARK-3 */
    aux_atom=vlist_atom_next(aux_atom); }
}

void gb_Process_Pair(module M, vpair P)
{
  evector aux_ev;
  aux_ev=gb_Do_SPoly(M,P);
  gb_Reduce_Update(M,aux_ev);
  FreeVPair(&P,module_ring(M));
}

void gb_Process_FIP(module M)
{
  vpair aux_p;
  aux_p=gb_Get_FIP(M);
  gb_Process_Pair(M,aux_p);
}

void gb_Reduce_Update(module M, evector V)
{
  int answer;
  answer=gb_Do_Reduce(M,&V);
  if (ev_is_zero(V)) {
    gb_set_CV(M,ev_NULL);
#ifdef GB_STATS
    if (ev_Is_SPoly(V)) {
      module_stat_incr(M,gb_STATS_ZERO_PAIRS);
    } else {
      module_stat_incr(M,gb_STATS_ZERO_VECTORS);
    }
#endif
    gb_Do_Case_Zero(M,V);
  }
  else if (gb_Do_Is_In_NF(M,V)) {
    gb_set_CV(M,ev_NULL);
    gb_Do_Update(M,V);
/*
 * These lines would stop the computation as soon as a constant is found
 * but we cannot do it because it may have (undesirable) side-effects
 * when the Groebner framework is used to compute syzygies, etc...
 *    if (ev_deg_lt(V) == 0)
 *    {
 *      vlist_free(&M->incoming_vectors, module_ring(M));
 *      FreeVPairs(&M->incoming_pairs, module_ring(M));
 *    }
 */
#ifdef GB_STATS
    if (ev_Is_SPoly(V)) {
      module_stat_incr(M,gb_STATS_USEFUL_PAIRS);
    } else {
      module_stat_incr(M,gb_STATS_USEFUL_VECTORS);
    }
#endif
  }
  else {
    gb_Do_Case_Not_NF(M,V);
    ev_unlink(V,module_ring(M)); }
}

void gb_Update_FIV(module M)
{
  evector aux_ev;
  aux_ev=gb_Get_FIV(M);
  if (ev_is_zero(aux_ev)) {
    ev_unlink(aux_ev,module_ring(M));
  } else {
    gb_Do_Update(M,aux_ev);
  }
}

void gb_Reduce_Update_FIV(module M)
{
  evector aux_ev;
  aux_ev=gb_Get_FIV(M);
  ev_force_rc1(&aux_ev,module_ring(M));
  gb_Reduce_Update(M,aux_ev);
}

void gb_Reduce_If_SPoly_And_Update_FIV(module M)
{
  evector aux_ev;
  aux_ev=gb_Get_FIV(M);
  ev_force_rc1(&aux_ev,module_ring(M));
  if (ev_Is_SPoly(aux_ev)) {
    gb_Reduce_Update(M,aux_ev);
  } else if (ev_is_zero(aux_ev)) {
    gb_Do_Case_Zero(M,aux_ev);
  } else {
    gb_Do_Update(M,aux_ev);
  }
}

void gb_Process_FIPV(module M)
{
  if (gb_Must_Process_Pair(M)) gb_Process_FIP(M);
  else gb_Reduce_Update_FIV(M);
}

int gb_Min_Deg_IVs(module M)
{
  if (vlist_is_empty(gb_IVs(M))) return deg_NULL;
  return ev_deg_lt(gb_FIV(M));
}

int gb_Min_Deg_IPs(module M)
{
  if (vpairs_is_empty(gb_IPs(M))) return deg_NULL;
  return vpair_deg(gb_FIP(M));
}

int gb_Min_Deg_Incomings(module M)
{
  int res_n;
  res_n=gb_Min_Deg_Incomings_NoTruncCheck(M);
  if (res_n <= MIN(module_DegTrunc(M),module_RegTrunc(M)))
    return res_n;
  else return deg_NULL;
}

int gb_Min_Deg_Incomings_NoTruncCheck(module M)
{
  if (!gb_Something_Incoming(M)) return deg_NULL;
  if (vpairs_is_empty(gb_IPs(M))) return ev_deg_lt(gb_FIV(M));
  if (vlist_is_empty(gb_IVs(M))) return vpair_deg(gb_FIP(M));
  return MIN(ev_deg_lt(gb_FIV(M)), vpair_deg(gb_FIP(M)));
}

coc_bool gb_Must_Process_Pair(module M)
{ /* PRIORITA` ALLE COPPIE ! */
  if (vpairs_is_empty(gb_IPs(M))) return FALSE;
  if (vlist_is_empty(gb_IVs(M))) return TRUE;
  return (vpair_deg(gb_FIP(M)) <= ev_deg_lt(gb_FIV(M)));
}


/* ---[ Do_Case_Not_NF() ]--------------------------------------- */

void gb_Do_Case_Not_NF_Add_IVs(module M, evector V)
{
  gb_Do_Add_To_IVs(M,V);
}

void gb_Do_Case_Not_NF_Keep(module M, evector V)
{
  gb_set_CV(M,ev_link(V));
}

/* ---[ Init() ]--------------------------------------- */

void gb_Init_NF(module M)
{
  /* nothing */
}

void gb_Init_Interreduce(module M)
{ /* copy Gens to Rules and set flag JUST_LTNF */
  vlist_atom aux_atom;
  int n;
  gb_set_Rules(M,vlist_init());
  n=vlist_len(module_gens(M));
  aux_atom=vlist_first(module_gens(M));
  while (n-- > 0) {
    if (!ev_is_zero(vlist_atom_ev(aux_atom)))
      gb_Do_Add_To_Rules(M,vlist_atom_ev(aux_atom));
    aux_atom=vlist_atom_next(aux_atom); }
  module_flag_set(M, module_FLAG_JUST_LTNF);
}

void gb_Init_Nothing(module M)
{
  /* DO NOTHING */
}

void gb_Init_Basic(module M)
{
  /* assumed that key functions are initialized */
  gb_set_CV(M,ev_NULL);
  gb_Init_IVs(M);
  module_set_gbasis(M,vlist_init());
  gb_set_Rules(M,vlist_init());
  gb_set_MinGens(M,vlist_init());
}

void gb_Init_NFAndRel(module M)
{
  evector v;
  v=gb_CV(M);
  ev_set_rel(v,pv_init(vlist_len(module_gens(M))));
  ev_set_Is_SPoly(v,TRUE); /* TODO: TRICK: force update of rules */
}

void gb_Init_BasicAndRels(module M)
{
  /* assumed that key functions are initialized */
  gb_set_CV(M,ev_NULL);
  module_set_gbasis(M,vlist_init());
  gb_set_Rules(M,vlist_init());
  gb_set_MinGens(M,vlist_init());
  gb_Init_IVs_Syz(M); /* puts also relation with Gens(M) */
}

void gb_Init_Syz_Aux(module M)
{
  /* assumed that key functions are initialized */
  gb_set_CV(M,ev_NULL);
  module_set_gbasis(M,vlist_init());
  gb_set_Rules(M,vlist_init());
  gb_set_MinGens(M,vlist_init());
  if (module_syz(M) != module_NULL)
module_unlink(module_syz(M),module_ring(M));

module_set_syz(M,module_init(vlist_len(module_gens(M)),module_ring(M)));

  module_set_gens(module_syz(M),vlist_init());
}

void gb_Init_Syz(module M)
{
  gb_Init_Syz_Aux(M);
  gb_Init_IVs_Syz(M); /* puts also relation with Gens(M) */
}

void gb_Init_HSyz(module M)
{
  shifts aux_ss;
  gb_Init_Syz_Aux(M);
  aux_ss=vlist_leading_terms_to_shifts(module_gens(M),module_ring(M));
  module_set_shifts(module_syz(M),aux_ss);
  gb_Init_IVs_Syz(M); /* puts also relation with Gens(M) and their
shifts */
}

/* ---[ Do_1() ]--------------------------------------- */

coc_bool gb_Do_1_Interreduce(module M) /* Interreduce Rules */
{
  int n;
  if (gb_CV(M) != ev_NULL) {
    if (gb_Do_1_NF(M)) return TRUE;
    else {
      if (!ev_is_zero(gb_CV(M))) gb_Do_Add_To_Rules(M,gb_CV(M));
      ev_unlink(gb_CV(M),module_ring(M));
      gb_set_CV(M,ev_NULL);
      return TRUE;
    }
  }
  n = vlist_len(gb_Rules(M));
  while (n--) {
    gb_set_CV(M,vlist_get_first_ev(gb_Rules(M)));
    if (gb_Do_1_NF(M)) return TRUE;
    gb_Do_Add_To_Rules(M,gb_CV(M)); /* Must be of type Append */
    ev_unlink(gb_CV(M),module_ring(M));
    gb_set_CV(M,ev_NULL);
  }
  /* if we are here it means that no Rule is reducible against the
others */
  /* now we check if we are in the first phase or in the second one */
  /* first phase is Interreduce LT, second phase is complete
interreduction */
  if (gb_FLAG_JUST_LTNF) {
    /* just finished the LT Interreduction, now complete it */
    module_flag_reset(M,module_FLAG_JUST_LTNF);
    gb_FLAG_JUST_LTNF=FALSE;
    return TRUE;
  } else {
    return FALSE;
  }
}

coc_bool gb_Do_1_NF(module M) /* normalize CV */
{
  int answer;
  if (gb_CV(M) == ev_NULL) return FALSE;
  answer=gb_Do_Reduce(M,&gb_CV(M));
  if (answer != gb_ANSWER_REDUCED) return FALSE;
  return TRUE;
}

coc_bool gb_Do_1_MinGens_Trunc(module M) /* computation of MinGens of M
*/
{
  int D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (vlist_is_empty(gb_IVs(M))) return FALSE;
  D = gb_Min_Deg_Incomings(M);
  gb_Process_FIPV(M);
  return TRUE;
}

coc_bool gb_Do_1_MinSyzMinGens_Trunc(module M)
/* computation of Minimal generators of Syz of MinGens(M) deg by deg */
{
  int D, SD;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (module_syz(M)==module_NULL) return gb_Do_1_HGB_Trunc(M);
  D = gb_Min_Deg_Incomings(M);
  if (vlist_is_empty(gb_IVs(module_syz(M)))) {
    if (D==deg_NULL) return FALSE; }
  else {
    SD = gb_Min_Deg_Incomings(module_syz(M));
    if (SD <= D) return gb_Do_1_MinGens_Trunc(module_syz(M)); }
  gb_Process_FIPV(M);
  return TRUE;
}

coc_bool gb_Do_1_HGB_Trunc(module M) /* computation of GBasis degree by
degree */
{
  int D,n;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  D = gb_Min_Deg_Incomings(M);
  if (D==deg_NULL) return FALSE;
  if (module_discrepancy(M) < 0) { /* no Hilb trunc */
    gb_Process_FIPV(M);
    return TRUE;
  }
  if (D < module_DegTrunc(M)) {
    n=gb_Kill_IPs_Of_Degree_LE(M,D);
    module_stat_add(M,gb_STATS_H_KILLED_PAIRS,n);
    n=gb_Kill_IVs_Of_Degree_LE(M,D);
    module_stat_add(M,gb_STATS_H_KILLED_VECTORS,n);
    return TRUE;
  }
  if (module_discrepancy(M) == 0) {
    return FALSE;
  }
  gb_Process_FIPV(M);
  return TRUE;
}

coc_bool gb_Do_1_HGB_Hilb_Trunc(module M)
{
  int DD,D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  D = gb_Min_Deg_Incomings(M);
  if (D==deg_NULL) return FALSE;
  if (gb_Something_Incoming(M)) {
    if (module_discrepancy(M) <= 0) { /* DD = discrepancy degree */
      module_set_discrepancy(M,
gb_Compute_Discrepancy_And_Degree(M,D,&DD));
      if (module_discrepancy(M) == 0) {
        gb_Kill_All_IPs(M);
        gb_Kill_All_IVs(M);
        module_set_discrepancy(M,-1);
        return FALSE;
      }
      /* I know that module_discrepancy(M) > 0 */
      gb_Kill_IPs_Of_Degree_LE(M,DD-1);
      gb_Kill_IVs_Of_Degree_LE(M,DD-1);
/*
      module_set_discrepancy(M,-1);
*/
    }
    gb_Process_FIPV(M);
    return TRUE;
  }
  return FALSE;
}

coc_bool gb_Do_1_Res_Row_Trunc(module M)
{
  int D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  D = gb_Min_Deg_Incomings_Res(M);
  if (D==deg_NULL) return FALSE;
  return gb_Do_1_In_Row(M,D);
}

coc_bool gb_Do_1_Res_Row_Hilb_Trunc(module M)
{
  int D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  do {
    D = gb_Min_Deg_Incomings_Res(M);
    if (D==deg_NULL) return FALSE;
  } while (!gb_Do_1_In_Row_Hilbert_Trunc(M,D));
  return TRUE;
}

/* ---[ Do_1() NoTruncCheck ]-----------------------------------------
*/

coc_bool gb_Do_1_Syz(module M)
{
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (!vlist_is_empty(gb_IVs(M)))
    { gb_Reduce_If_SPoly_And_Update_FIV(M); return TRUE; }
  if (!vpairs_is_empty(gb_IPs(M))) { gb_Process_FIP(M); return TRUE; }
  return FALSE;
}

coc_bool gb_Do_1_GB(module M)
{
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (!vlist_is_empty(gb_IVs(M)))
    { gb_Update_FIV(M); return TRUE; }
  if (!vpairs_is_empty(gb_IPs(M))) { gb_Process_FIP(M); return TRUE; }
  return FALSE;
}

coc_bool gb_Do_1_HGB(module M)
{
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (gb_Something_Incoming(M)) { gb_Process_FIPV(M); return TRUE; }
  return FALSE;
}

coc_bool gb_Do_1_HGB_Hilb(module M)
{
  int DD,D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  if (gb_Something_Incoming(M)) {
    if (module_discrepancy(M) <= 0) { /* DD = discrepancy degree */
      D = gb_Min_Deg_Incomings_Res_NoTruncCheck(M);
      module_set_discrepancy(M,
gb_Compute_Discrepancy_And_Degree(M,D,&DD));
      if (module_discrepancy(M) == 0) {
        gb_Kill_All_IPs(M);
        gb_Kill_All_IVs(M);
        module_set_discrepancy(M,-1);
        return FALSE;
      }
      /* I know that module_discrepancy(M) > 0 */
      gb_Kill_IPs_Of_Degree_LE(M,DD-1);
      gb_Kill_IVs_Of_Degree_LE(M,DD-1);
  /*
      module_set_discrepancy(M,-1);
  */
  }
    gb_Process_FIPV(M);
    return TRUE;
  }
  return FALSE;
}

coc_bool gb_Do_1_Res_Row(module M)
{
  int D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  D = gb_Min_Deg_Incomings_Res_NoTruncCheck(M);
  if (D==deg_NULL) return FALSE;
  return gb_Do_1_In_Row(M,D);
}

coc_bool gb_Do_1_Res_Row_Hilb(module M)
{
  int D;
  if (gb_CV(M) != ev_NULL) { gb_Reduce_Update(M,gb_CV(M)); return TRUE;
}
  do {
    D = gb_Min_Deg_Incomings_Res_NoTruncCheck(M);
    if (D==deg_NULL) return FALSE;
  } while (!gb_Do_1_In_Row_Hilbert(M,D));
  return TRUE;
}

/* ---[ Do_1() Aux ]----------------------------------------- */

int gb_Min_Deg_Incomings_Res(module M)
{
  int res_n,aux_n;
  module aux_M;
  res_n=gb_Min_Deg_Incomings(M);
  if (module_ResTrunc(M) > 0 ) aux_M=module_syz(M); else
aux_M=module_NULL;
  while (aux_M != module_NULL) {
    aux_n=gb_Min_Deg_Incomings(aux_M);
    res_n=gb_Min_Special(res_n,aux_n);
    if (module_ResTrunc(aux_M) > 0 )
      aux_M=module_syz(aux_M); else aux_M=module_NULL;
  }
  return res_n;
}

int gb_Min_Deg_Incomings_Res_NoTruncCheck(module M)
{
  int res_n,aux_n;
  module aux_M;
  res_n=gb_Min_Deg_Incomings_NoTruncCheck(M);
  aux_M=module_syz(M);
  while (aux_M != module_NULL) {
    aux_n=gb_Min_Deg_Incomings_NoTruncCheck(aux_M);
    res_n=gb_Min_Special(res_n,aux_n);
    aux_M=module_syz(aux_M);
  }
  return res_n;
}

coc_bool gb_Do_1_In_Row(module M, int D)
{
  coc_bool DoneSomething;
  if (M==module_NULL) return FALSE;
  DoneSomething = gb_Do_1_In_Row(module_syz(M),D);
  if (!DoneSomething) return gb_Do_1_In_Case(M,D);
  return TRUE;
}

coc_bool gb_Do_1_In_Row_Hilbert(module M, int D)
{
  coc_bool DoneSomething;
  if (M==module_NULL) return FALSE;
  DoneSomething = gb_Do_1_In_Row_Hilbert(module_syz(M),D);
  if (!DoneSomething) return gb_Do_1_In_Case_Hilbert(M,D);
  return TRUE;
}

coc_bool gb_Do_1_In_Row_Hilbert_Trunc(module M, int D)
{
  coc_bool DoneSomething;
  if (M==module_NULL) return FALSE; /* TODO consider trunc */
  DoneSomething = gb_Do_1_In_Row_Hilbert_Trunc(module_syz(M),D);
  if (!DoneSomething) return gb_Do_1_In_Case_Hilbert(M,D);
  return TRUE;
}

coc_bool gb_Do_1_In_Case(module M, int D)
{
  if (D==gb_Min_Deg_Incomings_NoTruncCheck(M)) {
    gb_Process_FIPV(M);
    return TRUE; }
  return FALSE; /* case is empty */
}

coc_bool gb_Do_1_In_Case_Hilbert(module M, int D)
{
  int n;
  if (D==gb_Min_Deg_Incomings_NoTruncCheck(M)) {
    if (gb_Min_Deg_IPs(M)>D) {
#ifdef DEBUG_GB
      if (module_discrepancy(M) >= 0)
        printf("ERROR in gb_Do_1_In_Case_Hilbert\n");
#endif
      gb_Reduce_Update_FIV(M);
      return TRUE;
    }
    if (module_discrepancy(M) < 0) {
      module_set_discrepancy(M,gb_Compute_Discrepancy(M,D));
      if (module_discrepancy(M) == 0) {
        n=gb_Kill_IPs_Of_Degree_LE(M,D);
#ifdef GB_STATS
        module_stat_add(M,gb_STATS_H_KILLED_PAIRS,n);
#endif
        n=gb_Kill_IVs_Of_Degree_LE(module_syz(M),D);
#ifdef GB_STATS
        module_stat_add(module_syz(M),gb_STATS_H_KILLED_VECTORS,n);
#endif
        module_set_discrepancy(M,-1);
        return FALSE; }
    } /* I know that module_discrepancy(M) > 0 */
    gb_Process_FIP(M);
    return TRUE;
  }
  return FALSE; /* case is empty */
}

/* ---[ Do_Case_0() ]--------------------------------------- */

void gb_Do_Case_0_Nothing(module M, evector V)
{
  ev_unlink(V,module_ring(M));
}

void gb_Do_Case_0_Syz_MG(module M, evector V)
{
  ring R;
  evector aux_ev;
  R = module_ring(M);
  if (!ev_Is_SPoly(V)) { ev_unlink(V,R); return; }
  else if (pv_is_zero(ev_rel(V))) { ev_unlink(V,R); return; }
  if (ev_rc(V)!=1) aux_ev=ev_init(pv_dup(ev_rel(V),R),R);
  else {
    aux_ev=ev_init(ev_rel(V),R);
    ev_set_rel(V,pv_NULL);
  }
  vlist_append_ev(module_gens(module_syz(M)),aux_ev);
  gb_Do_Add_To_IVs(module_syz(M),aux_ev);
  ev_unlink(aux_ev,R);
  ev_unlink(V,R);
}

void gb_Do_Case_0_Syz_G(module M, evector V)
{
  ring R;
  evector aux_ev;
  R = module_ring(M);
  if (pv_is_zero(ev_rel(V))) {
    ev_unlink(V,R);
    return;
  }
  if (ev_rc(V)!=1) aux_ev=ev_init(pv_dup(ev_rel(V),R),R);
  else {
    aux_ev=ev_init(ev_rel(V),R);
    ev_set_rel(V,pv_NULL);
  }
  vlist_append_ev(module_gens(module_syz(M)),aux_ev);
  ev_unlink(aux_ev,R);
  ev_unlink(V,R);
}

void gb_Do_Case_0_Syz_MGS(module M, evector V)
{
  shifts aux_ss;
  ring R;
  evector aux_ev;
  pvector aux_pv;
  R = module_ring(M);
  if (!ev_Is_SPoly(V)) { ev_unlink(V,R); return; }
  else if (pv_is_zero(ev_rel(V))) { ev_unlink(V,R); return; }
  if (ev_rc(V)!=1)
    aux_pv=pv_dup(ev_rel(V),R);
  else {
    aux_pv=ev_rel(V);
    ev_set_rel(V,pv_NULL);
  }
  pv_Apply_Shifts(aux_pv, module_shifts(module_syz(M)), R);
  aux_ev=ev_init(aux_pv,R);
  gb_Do_Add_To_IVs(module_syz(M),aux_ev);
  ev_unlink(aux_ev,R);
  ev_unlink(V,R);
}

/* ---[ Do_SPoly() ]--------------------------------------- */

evector gb_SPoly_GB(module M, vpair P)
{
  pvector aux_v;
  evector res_ev;
  ring R;
  R = module_ring(M);
  if (vpair_are_coprimes(P)) return
ev_init(pv_zero(module_rank(M),R),R);
  md_SPoly_Den(&aux_v,P,R);
  res_ev=ev_init(aux_v,R);
  ev_set_Is_SPoly(res_ev,TRUE);
  return res_ev;
}

evector gb_SPoly_Syz(module M, vpair P)
{
  evector res_ev;
  ring R;
  R = module_ring(M);
  syz_SPoly_Den(&res_ev,P,R);
  ev_set_Is_SPoly(res_ev,TRUE);
  return res_ev;
}

/* ---[ Do_Get_Rule() ]--------------------------------------- */

evector gb_Get_Rule_First(module M, term t, int n)
/* t=term, n=position */
{
  int aux_n;
  vlist rules;
  eterm aux_et;
  evector aux_ev;
  vlist_atom aux_atom;
  ring R;
  R = module_ring(M);
  rules = gb_Rules(M);
  aux_atom = vlist_first(rules);
  aux_et = term2eterm(t);
  while (aux_atom != vlist_atom_NULL) {
    aux_ev = vlist_atom_ev(aux_atom);
    aux_n=ev_lmpos(aux_ev);
    if (n == aux_n && eterm_divides(ev_eterm(aux_ev),aux_et)) {
      eterm_free(aux_et);
      return aux_ev; }
    aux_atom = vlist_atom_next(aux_atom);
  }
  eterm_free(aux_et);
  return ev_NULL;
}

evector gb_Get_Rule_Product(module M, term t, int n)
/* t=term, n=position */
{
  static evector local_ev=NULL;
  int aux_n,c;
  vlist rules;
  eterm aux_et;
  term aux_t;
  evector aux_ev,res_ev;
  vlist_atom aux_atom;
  ring R;
  monomial res_lm;
  pvcomp res_p,aux_p;
  R = module_ring(M);
  rules = gb_Rules(M);
  aux_atom = vlist_first(rules);
  aux_et = term2eterm(t);
  res_ev=ev_NULL;
  c=0;
  while (aux_atom != vlist_atom_NULL) {
    aux_ev = vlist_atom_ev(aux_atom);
    aux_n=ev_lmpos(aux_ev);
    if (n == aux_n && eterm_divides(ev_eterm(aux_ev),aux_et)) {
      c++;
      switch(c) {
        case 1 :
          res_ev=aux_ev;
          aux_t=term_div(t,ev_lt(aux_ev));
          eterm_free(aux_et);
          aux_et = term2eterm(aux_t);
          res_p = pv_nth(ev_tail(aux_ev),n);
          res_lm = ev_lm(aux_ev);
          break;
        case 2 :
          term_div_and_assign(aux_t,ev_lt(aux_ev));
          eterm_free(aux_et);
          aux_et = term2eterm(aux_t);
          res_p = pvcomp_mult(res_p,pv_nth(ev_tail(aux_ev),n),R);
          res_lm=monomial_dup(res_lm,R);
          monomial_mult_and_assign(res_lm,ev_lm(aux_ev),R);
          break;
        default : /* >=3 */
          term_div_and_assign(aux_t,ev_lt(aux_ev));
          eterm_free(aux_et);
          aux_et = term2eterm(aux_t);
          aux_p = pvcomp_mult(res_p,pv_nth(ev_tail(aux_ev),n),R);
          pvcomp_free(res_p,R);
          res_p=aux_p;
          monomial_mult_and_assign(res_lm,ev_lm(aux_ev),R);
      }
    } else {
      aux_atom = vlist_atom_next(aux_atom);
    }
  }
  if (c > 1) {
    if (module_rank(M) != 1) {
      INTERNAL_ERROR("gb_Get_Rule_Product: available only for rank = 1");
      return ev_NULL; /* force termination of GBasis Computation */
    }
    /* printf("[c=%d]",c); */
    if (local_ev != ev_NULL) ev_unlink(local_ev,R);
    if (c % 2 == 0) { /* TODO: OPTIMIZE */
      monomial_change_sign(res_lm,R);
    }
#if 0
    printf("RULE IS: ");
    printf_monomial(res_lm,R);
    printf(" --> ");
    printf_pvcomp(res_p,R);
    printf("\n");
#endif
    res_ev=local_ev=ev_init2(res_lm,n,pv_this_pvcomp2pv(res_p,0,1,R));
  }
  eterm_free(aux_et);
  return res_ev;
}

/* ---[ Do_Apply_Rule() ]--------------------------------------- */

/* NOTE for gb_Apply_Rule_GB, gb_Apply_Rule_Syz is assumed that the Rule

is
   such that: ev_irr(Rule) is always NULL and ev_tail(Rule) != ev_NULL.
*/

void gb_Apply_Rule_GB(module M, evector * V, monomial lmred, evector
Rule)
{
  ring R;
  pvector aux_v;
  R=module_ring(M);
  monomial_divide_this(lmred,ev_lm(Rule),R);
  monomial_change_sign(lmred,R);
  aux_v=ev_red(*V);
  aux_v=this_pv_plus_mxpv(aux_v,lmred,ev_tail(Rule),R);
  ev_set_red(*V,aux_v);
}

void gb_Apply_Rule_Syz(module M, evector * V, monomial lmred, evector
Rule)
{ /* NOTE: like gb_Apply_Rule_GB, but we also update rel */
  ring R;
  pvector aux_v;
  R=module_ring(M);
  monomial_divide_this(lmred,ev_lm(Rule),R);
  monomial_change_sign(lmred,R);
  aux_v=ev_red(*V);
  aux_v=this_pv_plus_mxpv(aux_v,lmred,ev_tail(Rule),R);
  ev_set_red(*V,aux_v);
  if (ev_Is_SPoly(*V))
    ev_set_rel(*V, this_pv_plus_mxpv(ev_rel(*V),lmred,ev_rel(Rule),R));
}

/* ---[ Do_Reduce() ]--------------------------------------- */

/* gb_ANSWER_REDUCED : at least one rule was applied */
/* gb_ANSWER_NOT_REDUCED : no reductions were made, don't know if is NF
*/
/* gb_ANSWER_NOT_REDUCED_NF : no reductions were made, we know it is NF
*/

int gb_Reduce_LT_1(module M, evector * V)
{
  evector Rule;
  monomial aux_m,
    lmred; /* leading monomial of (potentially) reducible part of V */
  int lmpos;
  ring R;
  R = module_ring(M);
  if (gb_FLAG_JUST_LTNF) {
    if (ev_irr(*V) != mv_NULL) return gb_ANSWER_NOT_REDUCED_NF;
  }
  ev_extract_lmred(*V,&lmred,&lmpos,R);
  if (lmred == NULL) return gb_ANSWER_NOT_REDUCED_NF;
  Rule = gb_Do_Get_Rule(M,monomial_term(lmred),lmpos);
  if (Rule == ev_NULL) {
    ev_add_monomial_to_irr(*V,lmred,lmpos,R);
    return gb_ANSWER_NOT_REDUCED;
  } else {
    gb_Do_Apply_Rule(M,V,lmred,Rule);
    monomial_free(lmred,R);
    return gb_ANSWER_REDUCED;
  }
}

int gb_Reduce_Full_1(module M, evector *V)
{ /* do one reduction if it is possible */
  int answer;
  do {
    answer = gb_Reduce_LT_1(M,V);
  } while (answer == gb_ANSWER_NOT_REDUCED);
  return answer; /* REDUCED or NOT_REDUCED_NF */
}

int gb_Reduce_LT(module M, evector *V)
{
  int answer, res_answer;
  res_answer = gb_ANSWER_NOT_REDUCED;
  do {
    answer = gb_Reduce_LT_1(M,V);
    if (answer == gb_ANSWER_REDUCED) res_answer = answer;
    else if (answer == gb_ANSWER_NOT_REDUCED_NF &&
      res_answer == gb_ANSWER_NOT_REDUCED) res_answer = answer;
  } while (answer == gb_ANSWER_REDUCED);
  return res_answer;
}

coc_bool gb_Reduce_Full(module M, evector *V)
{
  int answer, res_answer;
  res_answer = gb_ANSWER_NOT_REDUCED;
  do {
    answer = gb_Reduce_Full_1(M,V);
    if (answer == gb_ANSWER_REDUCED) res_answer = answer;
    else if (answer == gb_ANSWER_NOT_REDUCED_NF &&
      res_answer == gb_ANSWER_NOT_REDUCED) res_answer = answer;
  } while (answer == gb_ANSWER_REDUCED);
  return res_answer;
}

/* ---[ Do_Update() ]--------------------------------------- */

void gb_Update_Basic(module M, evector V)
{
  vpair NewPs;
#ifndef GB_NEWCRIT
  gb_Apply_BCriterion(M,V);
#endif
  ev_irr_plus_red_to_tail(V,module_ring(M));
  NewPs = gb_Build_New_Pairs_GM(M,V);
  gb_set_IPs(M,gb_Merge_Pair_Lists(M,gb_IPs(M),NewPs));
  ev_set_label(V,vlist_len(gb_Rules(M))+1);
  gb_Do_Add_To_Rules(M,V);
  if (gb_ADD_LAST) vlist_append_ev(module_gbasis(M),V);
  ev_unlink(V,module_ring(M));
}

void gb_Update_Discrepancy(module M, evector V)
{
  if (module_discrepancy(M) > 0 &&

term_first_nonzero_indet(ev_lt(V),module_ring(M))>module_ElimIndets(M))
{
    module_decrement_discrepancy(M);
  }
  if (!ev_Is_SPoly(V)) {
    vlist_append_ev(module_mingens(M),V);
  }
  gb_Update_Basic(M,V);
}

void gb_Update_MinGens(module M, evector V)
{
  if (!ev_Is_SPoly(V)) {
    vlist_append_ev(module_mingens(M),V);
  }
  gb_Update_Basic(M,V);
}

void gb_Update_Syz_MG(module M, evector V)
{
  int n;
  if (!ev_Is_SPoly(V)) {
    if (module_syz(M) == module_NULL) module_set_syz(M,module_Clone(M));

    module_set_rank(module_syz(M),module_rank(module_syz(M))+1);
    n=vlist_len(module_mingens(M));

ev_set_rel(V,pv_gen(MIN(n+1,module_SyzCompsNum(M)),n,module_ring(M)));
    vlist_append_ev(module_mingens(M),V);
  }
  gb_Update_Basic(M,V);
}

void gb_Update_Syz_MGS(module M, evector V)
{
  int n;
  if (!ev_Is_SPoly(V)) {
    if (module_syz(M) == module_NULL) module_set_syz(M,module_Clone(M));

    module_set_rank(module_syz(M),module_rank(module_syz(M))+1);
    shifts_append(&module_shifts(module_syz(M)),ev_lt(V));
    n=vlist_len(module_mingens(M));

ev_set_rel(V,pv_gen(MIN(n+1,module_SyzCompsNum(M)),n,module_ring(M)));
    vlist_append_ev(module_mingens(M),V);
  }
  gb_Update_Basic(M,V);
}

void gb_Update_Syz_MGS_Hilb(module M, evector V)
{
  int n;
  if (!ev_Is_SPoly(V)) { /* was IV */
    if (module_syz(M) == module_NULL) module_set_syz(M,module_Clone(M));

    module_set_rank(module_syz(M),module_rank(module_syz(M))+1);
    shifts_append(&module_shifts(module_syz(M)),ev_lt(V));
    n=vlist_len(module_mingens(M));

ev_set_rel(V,pv_gen(MIN(n+1,module_SyzCompsNum(M)),n,module_ring(M)));
    vlist_append_ev(module_mingens(M),V);
    /* HILBERT */
    if (module_zys(M) != module_NULL &&
module_discrepancy(module_zys(M)) > 0) {
      module_decrement_discrepancy(module_zys(M));
      if (module_discrepancy(module_zys(M)) == 0) {
        n=gb_Kill_IVs_Of_Degree_LE(M,ev_deg_lt(V));
#ifdef GB_STATS
        module_stat_add(M,gb_STATS_H_KILLED_VECTORS,n);
#endif
        n=gb_Kill_IPs_Of_Degree_LE(module_zys(M),ev_deg_lt(V));
#ifdef GB_STATS
        module_stat_add(module_zys(M),gb_STATS_H_KILLED_PAIRS,n);
#endif
        module_set_discrepancy(module_zys(M),-1); } }
  } else { /* was IP */
    /* HILBERT */
    if (module_discrepancy(M) > 0) {
      module_decrement_discrepancy(M);
      if (module_discrepancy(M) == 0) {
        n=gb_Kill_IVs_Of_Degree_LE(module_syz(M),ev_deg_lt(V));
#ifdef GB_STATS
        module_stat_add(module_syz(M),gb_STATS_H_KILLED_VECTORS,n);
#endif
        n=gb_Kill_IPs_Of_Degree_LE(M,ev_deg_lt(V));
#ifdef GB_STATS
        module_stat_add(M,gb_STATS_H_KILLED_PAIRS,n);
#endif
        module_set_discrepancy(M,-1); } }
  }
  gb_Update_Basic(M,V);
}

/* ---[ Do_Update() aux ]--------------------------------------- */

void gb_Apply_BCriterion(module M, evector V)
{
  vpair auxPair, myPair;
  term LF;
  ring R;
  int pos;
  LF = ev_lt(V);
  pos = ev_lmpos(V);
  myPair = gb_IPs(M);
  R = module_ring(M);
  while ( myPair != NULL ) {
    if ( vpair_lmpos(myPair)==pos && term_divides( LF, vpair_lcm(myPair)

) ) {
      term_take_lcm(R->aux_iv[0],LF, vpair_first_lt(myPair));
      if ( term_eq( R->aux_iv[0], vpair_lcm(myPair) ) ) {
        auxPair = myPair;
        myPair = myPair -> next;
      } else {
        term_take_lcm(R->aux_iv[0],LF, vpair_second_lt(myPair) );
        if ( term_eq( R->aux_iv[0], vpair_lcm(myPair) ) ) {
          auxPair = myPair;
          myPair = myPair -> next;
        } else {
#ifdef GB_STATS
          module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
          if ( myPair == gb_IPs(M) ) {
            gb_set_IPs(M, gb_IPs(M)->next);
            FreeVPair( & myPair,R );
            myPair = gb_IPs(M);
          } else {
            auxPair -> next = myPair -> next;
            FreeVPair( & myPair,R );
            myPair = auxPair -> next;
          }
        }
      }
    } else {
      auxPair = myPair;
      myPair = myPair -> next;
    }
  }
}

vpair gb_Build_New_Pairs_GM(module M, evector V)
{ /* Crea nuove coppie con GBasis e V */
  /* Inoltre cancella vecchi elementi ridondanti */
  vpair newPairs,myPair;
  int n;
  ring R;
  vlist NewGB;
  vlist_atom aux_atom;
  evector aux_ev;
  gb_ADD_LAST=TRUE;
  newPairs = vpairs_NULL;
  NewGB = vlist_init();
  n=vlist_len(module_gbasis(M));
  R=module_ring(M);
  while (n-- > 0) {
    aux_atom=vlist_get_first(module_gbasis(M));
    aux_ev=vlist_atom_ev(aux_atom);
    if (ev_lmpos(aux_ev) == ev_lmpos(V)) {
      myPair = gb_Do_Make_Pair(M,aux_ev,V);
      /* MARK-2: TODO: tener conto che nell'alg. omogeneo il test e`
sempre FALSE */
      if (term_eq(vpair_lcm(myPair),ev_lt(aux_ev))) {
        ev_unlink(aux_ev,R);
        free_vlist_atom(aux_atom);
      } else {
        if (gb_ADD_LAST && term_eq(vpair_lcm(myPair),ev_lt(V)))
          gb_ADD_LAST = FALSE;
        vlist_append(NewGB,aux_atom);
      }
      gb_Insert_Pair_GM(M, &newPairs, myPair);
    } else {
      vlist_append(NewGB,aux_atom);
    }
  } /* while */
  vlist_free(&module_gbasis(M),module_ring(M));
  module_set_gbasis(M,NewGB);
  return newPairs;
}

void gb_Insert_Pair(module M, vpair * newPairs, vpair myPair)
{
  *newPairs = gb_Merge_Pair_Lists(M,*newPairs,myPair);
}

#ifdef GB_NEWCRIT
void gb_Insert_Pair_GM(module M, vpair * newPairs, vpair myPair)
{ /* Apply GM CRITERION */
  coc_bool notInserted;
  vpair auxPair, prevPair;
  int which, n;
  ring R;

  R = module_ring(M);

      auxPair = *newPairs;
      prevPair = NULL;
      notInserted = TRUE;
      while ( (auxPair != NULL ) && notInserted ){
        which = term_which_divides( vpair_B(myPair), vpair_B(auxPair) );

        switch ( which ) {
          case ANSWER_BOTH :  /* are equals */
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            FreeVPair( &myPair,R );
            notInserted = FALSE;
            break;
          case ANSWER_SECOND :
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            notInserted = FALSE;
            FreeVPair( &myPair,R );
            break;
          case ANSWER_FIRST :
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            if ( prevPair == NULL ) {
              *newPairs = (*newPairs) -> next;
              FreeVPair( &auxPair,R );
              auxPair = *newPairs;
            } else {
              prevPair -> next = auxPair -> next;
              FreeVPair( &auxPair,R );
              auxPair = prevPair;
            }
            break;
          default:  /* NONE */
            if (term_coprime(vpair_A(myPair), vpair_A(auxPair))) {
              n=gb_Kill_IP(M,vpair_first(auxPair),vpair_first(myPair));
#ifdef GB_STATS
              module_stat_add(M,gb_STATS_BACK_KILLED,n);
              module_stat_add(M,gb_STATS_C_KILLED_PAIRS,n);
#endif
            }
            prevPair = auxPair;
            auxPair = auxPair-> next;
            break;
        }  /* switch */
      } /* while ( auxPair != NULL ) */

      if ( notInserted ) gb_Insert_Pair( M, newPairs, myPair);
}
#else
void gb_Insert_Pair_GM(module M, vpair * newPairs, vpair myPair)
{ /* Apply GM CRITERION */
  coc_bool notInserted;
  vpair auxPair, prevPair;
  int which;
  ring R;

  R = module_ring(M);

      auxPair = *newPairs;
      prevPair = NULL;
      notInserted = TRUE;
      while ( (auxPair != NULL ) && notInserted ){
        which = term_which_divides( vpair_lcm(myPair),
vpair_lcm(auxPair) );
        switch ( which ) {
          case ANSWER_BOTH :  /* are equals */
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            if ( vpair_are_coprimes(auxPair) || !
vpair_are_coprimes(myPair) ) {
              FreeVPair( &myPair,R ); }
            else {
              vpair_set_first(auxPair,vpair_first(myPair));
              vpair_set_coprimes(auxPair,vpair_are_coprimes(myPair));
              vpair_set_wd(auxPair,vpair_wd(myPair));
              vpair_set_td(auxPair,vpair_td(myPair));
              FreeVPair( &myPair,R );
            }
            notInserted = FALSE;
            break;
          case ANSWER_SECOND :
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            notInserted = FALSE;
            FreeVPair( &myPair,R );
            break;
          case ANSWER_FIRST :
#ifdef GB_STATS
            module_stat_incr(M,gb_STATS_C_KILLED_PAIRS);
#endif
            if ( prevPair == NULL ) {
              *newPairs = (*newPairs) -> next;
              FreeVPair( &auxPair,R );
              auxPair = *newPairs;
            } else {
              prevPair -> next = auxPair -> next;
              FreeVPair( &auxPair,R );
              auxPair = prevPair;
            }
            break;
          default:  /* NONE */
            prevPair = auxPair;
            auxPair = auxPair-> next;
            break;
        }  /* switch */
      } /* while ( auxPair != NULL ) */

      if ( notInserted ) gb_Insert_Pair( M, newPairs, myPair);
}
#endif

vpair gb_Merge_Pair_Lists(module M, vpair Ps1, vpair Ps2)
{ /* TODO: apply here Coprimality Criterion? */
  vpair Ps, res_Ps;
  Ps = vpair_dummy;
  res_Ps = Ps;
  while (Ps1 != vpairs_NULL && Ps2 != vpairs_NULL) {
    if (gb_Do_Pair_Prec(M,Ps2,Ps1)) {
      vpair_set_next(Ps,Ps2);
      Ps = vpair_next(Ps);
      Ps2 = vpair_next(Ps2);
    } else {
      vpair_set_next(Ps,Ps1);
      Ps = vpair_next(Ps);
      Ps1 = vpair_next(Ps1);
    }
  }
  if (Ps2 == vpairs_NULL) vpair_set_next(Ps, Ps1);
  else vpair_set_next(Ps,Ps2);
  return vpair_next(res_Ps);
}

/* ---[ Do_Is_In_NF() ]--------------------------------------- */

coc_bool gb_Is_In_NF_True(module M, evector V)
{
  return TRUE;
}

coc_bool gb_Is_In_NF_LT(module M, evector V)
{
  OPTIMIZE("gb_Is_In_NF_LT");
  return ev_is_zero(V) || ev_irr(V) != mv_NULL;
}

coc_bool gb_Is_In_NF_Full(module M, evector V)
{
  OPTIMIZE("gb_Is_In_NF_Full");
  return ev_is_zero(V) || pv_is_zero(ev_red(V));
}

/* ---[ Do_Make_Pair() ]--------------------------------------- */

vpair gb_Make_Pair_NoSugar(module M, evector V, evector W)
{
  vpair res_p;
  ring R;
  R = module_ring(M);
  res_p=NewVPair(R);
  vpair_set_first(res_p,ev_link(V));
  vpair_set_second(res_p,ev_link(W));

vpair_set_lcm(res_p,term_lcm(vpair_first_lt(res_p),vpair_second_lt(res_p)));

#ifdef GB_NEWCRIT
  vpair_set_A(res_p,term_div(vpair_lcm(res_p),vpair_first_lt(res_p)));
  vpair_set_B(res_p,term_div(vpair_lcm(res_p),vpair_second_lt(res_p)));
#endif
  if (module_rank(M)==1) {
    vpair_set_coprimes(res_p,
      term_coprime(vpair_first_lt(res_p),vpair_second_lt(res_p)) );
  } else {
    vpair_set_coprimes(res_p,FALSE);
  }
  res_p->next = vpair_NULL;
#ifdef GB_STATS
  module_stat_incr(M,gb_STATS_CREATED_PAIRS);
#endif
  return res_p;
}

vpair gb_Make_Pair_Sugar(module M, evector V, evector W)
{
  register int sDeg;
  vpair res_p;
  res_p=gb_Make_Pair_NoSugar(M,V,W);
  sDeg = term_degree(vpair_lcm(res_p));
  if ( ev_td(V) > ev_td(W) ) {
    sDeg += ev_td(V);
    vpair_set_td(res_p,ev_td(V)); }
  else {
    sDeg += ev_td(W);
    vpair_set_td(res_p,ev_td(W)); }
  vpair_set_wd(res_p,sDeg);
#ifdef GB_STATS
  module_stat_incr(M,gb_STATS_CREATED_PAIRS);
#endif
  return res_p;
}

/* ---[ Do_Add_To_Rules() ]--------------------------------------- */

vpair gb_Add_To_Rules_Append(module M, evector V)
{
  if (ev_tail(V) == pv_NULL) {
    ev_irr_plus_red_to_tail(V,module_ring(M));
  }
  if (ev_eterm(V) == NULL) ev_set_eterm(V,term2eterm(ev_lt(V)));
  vlist_append_ev(gb_Rules(M),V);
  return vpair_NULL;
}

/* ---[ Do_Add_To_IVs() ]--------------------------------------- */

vpair gb_Add_To_IVs_Append(module M, evector V)
{
  vlist_append_ev(gb_IVs(M),V);
#ifdef GB_STATS
  module_stat_incr(M,gb_STATS_CREATED_VECTORS);
#endif
  return vpair_NULL;
}

vpair gb_Add_To_IVs_AntiDeg(module M, evector V)
{
  vlist_Insert_AntiDeg(gb_IVs(M),V);
#ifdef GB_STATS
  module_stat_incr(M,gb_STATS_CREATED_VECTORS);
#endif
  return vpair_NULL;
}

/* ---[ Do_Pair_Prec() ]--------------------------------------- */

coc_bool gb_Pair_Prec_AntiDegRevLex(module M, vpair P1, vpair P2)
{
  register int k;
  ring R;
  if ( term_degree( vpair_lcm(P1)) < term_degree( vpair_lcm(P2)) )
    return TRUE;
  if ( term_degree( vpair_lcm(P1)) > term_degree( vpair_lcm(P2)) )
    return FALSE;
  R = module_ring(M);

  for ( k = R->indetsNo; k >= 1; k-- )
    if ( iv_get_nth( vpair_lcm(P1), k ) < iv_get_nth( vpair_lcm(P2), k
))
      return FALSE;
    else if ( iv_get_nth( vpair_lcm(P1), k ) > iv_get_nth(
vpair_lcm(P2), k ))
      return TRUE;
  return FALSE; /* this line should not be reached */
}

coc_bool gb_Pair_Prec_AntiDeg(module M, vpair P1, vpair P2)
{
  return ( term_degree( vpair_lcm(P1)) < term_degree( vpair_lcm(P2)) );
}

coc_bool gb_Pair_Prec_SugarAntiLex(module M, vpair P1, vpair P2)
{
  register int k;
  ring R;

  if ( vpair_wd(P1) < vpair_wd(P2) ) return TRUE;
  if ( vpair_wd(P1) > vpair_wd(P2) ) return FALSE;
  if ( vpair_td(P1) < vpair_td(P2) ) return FALSE;
  if ( vpair_td(P1) > vpair_td(P2) ) return TRUE;
  R = module_ring(M);
  for ( k = R->indetsNo; k >= 1; k-- )
    if ( iv_get_nth( vpair_lcm(P1), k ) < iv_get_nth( vpair_lcm(P2), k
))
      return FALSE;
    else if ( iv_get_nth( vpair_lcm(P1), k ) > iv_get_nth(
vpair_lcm(P2), k ))
      return TRUE;
  return FALSE; /* this line should not be reached */
}

coc_bool gb_Pair_Prec_Sugar(module M, vpair P1, vpair P2)
{
  if ( vpair_wd(P1) < vpair_wd(P2) ) return TRUE;
  if ( vpair_wd(P1) > vpair_wd(P2) ) return FALSE;
  return ( vpair_td(P1) > vpair_td(P2) );
}



