/* $Id: gb.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $

[ ENTRY POINTS ] [ ADT ] [ Do_1() ] [ Do_Case_Zero() ] [ Do_SPoly() ]
[ Do_Get_Rule() ] [ Do_Apply_Rule() ] [ Do_Reduce() ] [ Do_Update() ]
[ Do_Make_Pair() ] [ Do_Add_To_Rules() ] [ Do_Add_To_IVs() ][ Do_Pair_Prec() ]
[ Do_Is_In_NF() ] [ Do_1() Aux ] [ Do_Case_Not_NF() ]

------------------------------------------------------------- */
 
#ifndef GB_H
#define GB_H

#include "utils.h"
#include "options.h"

#define gb_ANSWER_NOT_REDUCED_NF 0
#define gb_ANSWER_REDUCED 1
#define gb_ANSWER_NOT_REDUCED 2

typedef enum {
  gb_err_NONE,
  gb_err_set_Do_1 
} gb_errors;

typedef enum {
  gb_alg_CLASSIC, gb_alg_HOMOG, gb_alg_RES_ROW, gb_alg_HDRIVEN
} gb_algorithms;

#define gb_MAX_NUM_KEY_FUNCTIONS 100
/* #define gb_NUM_MODULE_KEY_FUNCTIONS ?? */ /* see coctypes.h */

extern void * gb_KF[gb_MAX_NUM_KEY_FUNCTIONS+1];
extern int gb_NUM_KEY_FUNCTIONS;
extern gb_errors gb_err;
extern vpair vpair_dummy;

#define gb_kf_NONE 0

#define gb_alg(M) ((M)->gb.alg)
#define gb_CV(M) ((M)->current_vector)
#define gb_IVs(M) ((M)->incoming_vectors)
#define gb_IPs(M) ((M)->incoming_pairs)
#define gb_FIP(M) vpairs_first(gb_IPs(M))
#define gb_FIV(M) vlist_first_ev(gb_IVs(M))
#define gb_MinGens(M) ((M)->mingens)
#define gb_Rules(M) ((M)->rules)
#define gb_kf(M,n) ((M)->KF[n])

#define gb_set_MinGens(M,Vs) (M)->mingens=Vs
#define gb_set_Rules(M,Vs) (M)->rules=Vs
#define gb_set_CV(M,V) (M)->current_vector=V
#define gb_set_IVs(M,Vs) (M)->incoming_vectors=Vs
#define gb_set_IPs(M,Ps) (M)->incoming_pairs=Ps
#define gb_set_kf(M,n,f) (M)->KF[n]=(f)

#define gb_Get_FIP(M) vpairs_get_first(&gb_IPs(M))
#define gb_Get_FIV(M) vlist_get_first_ev(gb_IVs(M))

#define gb_Some_IP(M) (!vpairs_is_empty(gb_IPs(M)))
#define gb_Some_IV(M) (!vlist_is_empty(gb_IVs(M)))

#define gb_Something_Incoming(M) (gb_Some_IV(M) || gb_Some_IP(M))


/* the following works because deg_NULL is the biggest integer */
#define gb_Min_Special(n,m) MIN(n,m)

/* -------------------------------------------------------------------- */
/* KEY MACROS */

#define gb_kf_Do_1 0
#define gb_kf_Do_SPoly 1
#define gb_kf_Do_Reduce 2
#define gb_kf_Do_Case_Zero 3
#define gb_kf_Do_Update 4
#define gb_kf_Do_Get_Rule 5
#define gb_kf_Do_Apply_Rule 6
#define gb_kf_Do_Add_To_Rules 7
#define gb_kf_Do_Add_To_IVs 8
#define gb_kf_Do_Pair_Prec 9
#define gb_kf_Do_IsIn_NF 10
#define gb_kf_Do_Make_Pair 11
#define gb_kf_Do_Better_Rule 12
#define gb_kf_Do_Case_Not_NF 13
#define gb_kf_Do_Init 14

#define gb_Do_1(M) ((coc_bool (*)(module))(gb_KF[(M)->KF[gb_kf_Do_1]]))(M)
#define gb_Do_SPoly(M,P) ((evector (*)())(gb_KF[(M)->KF[gb_kf_Do_SPoly]]))(M,P)
#define gb_Do_Reduce(M,V) ((int (*)())(gb_KF[(M)->KF[gb_kf_Do_Reduce]]))(M,V)
#define gb_Do_Case_Zero(M,V) ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Case_Zero]]))(M,V)
#define gb_Do_Update(M,V) ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Update]]))(M,V)
#define gb_Do_Get_Rule(M,T,N) ((evector (*)())(gb_KF[(M)->KF[gb_kf_Do_Get_Rule]]))(M,T,N)
#define gb_Do_Apply_Rule(M,V,m,R) \
  ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Apply_Rule]]))(M,V,m,R)
#define gb_Do_Add_To_Rules(M,V) \
  ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Add_To_Rules]]))(M,V)
#define gb_Do_Add_To_IVs(M,V) \
  ((void (*)(module, evector))(gb_KF[(M)->KF[gb_kf_Do_Add_To_IVs]]))(M,V)
#define gb_Do_Pair_Prec(M,P1,P2) \
  ((coc_bool (*)())(gb_KF[(M)->KF[gb_kf_Do_Pair_Prec]]))(M,P1,P2)
#define gb_Do_Is_In_NF(M,V) \
  ((coc_bool (*)())(gb_KF[(M)->KF[gb_kf_Do_IsIn_NF]]))(M,V)
#define gb_Do_Make_Pair(M,V1,V2) \
  ((vpair (*)())(gb_KF[(M)->KF[gb_kf_Do_Make_Pair]]))(M,V1,V2)
#define gb_Do_Better_Rule(M,R1,R2) \
  ((vpair (*)())(gb_KF[(M)->KF[gb_kf_Do_Better_Rule]]))(M,R1,R2)
#define gb_Do_Case_Not_NF(M,V) \
  ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Case_Not_NF]]))(M,V)
#define gb_Do_Init(M) \
  ((void (*)())(gb_KF[(M)->KF[gb_kf_Do_Init]]))(M)

/* ---[ ENTRY POINTS ]----------------------------------------------- */

void gb_free(module M);
void gb_dup(module res_M, module M);
void gb_kf_dup(module res_M, module M);
void gb_kf_init(module res_M);
void gb_init();
int gb_Do_All(module M);
coc_bool gb_one_move(module M);

cterm gb_kf2cterm(module M);
olist gb_kf2olist(module M);
olist gb_stats2olist(module M);

/* ---[ ADT ]----------------------------------------------- */

void syz_SPoly_Den( evector * SP, vpair myPair,ring R );
void md_SPoly_Den( pvector * SP, vpair myPair,ring R );

pvector pv_linear_comb(vlist vl,pvector v, ring R);
/* assumption: card(vl) = rank(v) */

vlist vlist_linear_comb(vlist vl, vlist coeffs, ring R);

evector gb_FirstReducibleRule(module M);
int gb_Compute_Discrepancy(module M, int D);
int gb_Kill_IP(module M, evector v1, evector v2);
int gb_Kill_IPs_Of_Degree_LE(module M, int D);
int gb_Kill_IVs_Of_Degree_LE(module M, int D);
int gb_Kill_All_IPs(module M);
int gb_Kill_All_IVs(module M);
void gb_Init_IVs(module M);
void gb_Init_IVs_Syz(module M);
void gb_Process_Pair(module M, vpair P);
void gb_Process_FIP(module M);
void gb_Reduce_Update(module M, evector V);
void gb_Update_FIV(module M);
void gb_Reduce_Update_FIV(module M);
void gb_Reduce_If_SPoly_And_Update_FIV(module M);
void gb_Process_FIPV(module M);
int gb_Min_Deg_Incomings(module M);
int gb_Min_Deg_Incomings_NoTruncCheck(module M);
coc_bool gb_Must_Process_Pair(module M);

/* ---[ Do_Case_Not_NF() ]--------------------------------------- */

void gb_Do_Case_Not_NF_Add_IVs(module M, evector V);
void gb_Do_Case_Not_NF_Keep(module M, evector V);

/* ---[ Init() ]--------------------------------------- */

void gb_Init_NF(module M);
void gb_Init_NFAndRel(module M);
void gb_Init_Interreduce(module M);
void gb_Init_Nothing(module M);
void gb_Init_Basic(module M);
void gb_Init_BasicAndRels(module M);
void gb_Init_Syz(module M);
void gb_Init_HSyz(module M);

/* ---[ Do_1() ]--------------------------------------- */

coc_bool gb_Do_1_Interreduce(module M); /* normalize CV */
coc_bool gb_Do_1_NF(module M); /* normalize CV */
coc_bool gb_Do_1_GB(module M);
coc_bool gb_Do_1_Syz(module M);
coc_bool gb_Do_1_MinGens_Trunc(module M); /* computation of MinGens of M */
coc_bool gb_Do_1_HGB_Trunc(module M);
coc_bool gb_Do_1_HGB_Hilb_Trunc(module M);
coc_bool gb_Do_1_MinSyzMinGens_Trunc(module M); 
/* computation of MinSyz deg by deg */
coc_bool gb_Do_1_HGB_Hilb(module M);
coc_bool gb_Do_1_Res_Row_Trunc(module M);
coc_bool gb_Do_1_Res_Row_Hilb_Trunc(module M);

coc_bool gb_Do_1_HGB(module M);
coc_bool gb_Do_1_Res_Row(module M);
coc_bool gb_Do_1_Res_Row_Hilb(module M);

/* ---[ Do_1() Aux ]----------------------------------------- */

int gb_Min_Deg_Incomings_Res(module M);
coc_bool gb_Do_1_In_Row(module M, int D);
coc_bool gb_Do_1_In_Case(module M, int D);
coc_bool gb_Do_1_In_Row_Hilbert(module M, int D);
coc_bool gb_Do_1_In_Row_Hilbert_Trunc(module M, int D);
coc_bool gb_Do_1_In_Case_Hilbert(module M, int D);

/* ---[ Do_Case_Zero() ]--------------------------------------- */

void gb_Do_Case_0_Nothing(module M, evector V);
void gb_Do_Case_0_Syz_G(module M, evector V);
void gb_Do_Case_0_Syz_MG(module M, evector V);
void gb_Do_Case_0_Syz_MGS(module M, evector V);

/* ---[ Do_SPoly() ]--------------------------------------- */

evector gb_SPoly_GB(module M, vpair P);
evector gb_SPoly_Syz(module M, vpair P);

/* ---[ Do_Get_Rule() ]--------------------------------------- */

evector gb_Get_Rule_First(module M, term t, int n);
evector gb_Get_Rule_Product(module M, term t, int n);

/* ---[ Do_Apply_Rule() ]--------------------------------------- */

void gb_Apply_Rule_GB(module M, evector * V, monomial lmred, evector Rule);
void gb_Apply_Rule_Syz(module M, evector * V, monomial lmred, evector Rule);

/* ---[ Do_Reduce() ]--------------------------------------- */

int gb_Reduce_LT_1(module M, evector * V);
int gb_Reduce_Full_1(module M, evector *V);
int gb_Reduce_LT(module M, evector *V);
int gb_Reduce_Full(module M, evector *V);

/* ---[ Do_Update() ]--------------------------------------- */

void gb_Apply_BCriterion(module M, evector V);
vpair gb_Build_New_Pairs_GM(module M, evector V);
void gb_Insert_Pair(module M, vpair * newPairs, vpair myPair);
void gb_Insert_Pair_GM(module M, vpair * newPairs, vpair myPair);
vpair gb_Merge_Pair_Lists(module M, vpair Ps1, vpair Ps2);
void gb_Update_Basic(module M, evector V);
void gb_Update_Discrepancy(module M, evector V);
void gb_Update_MinGens(module M, evector V);
void gb_Update_Syz_MG(module M, evector V);
void gb_Update_Syz_MGS(module M, evector V);
void gb_Update_Syz_MGS_Hilb(module M, evector V);

/* ---[ Do_Is_In_NF() ]--------------------------------------- */

coc_bool gb_Is_In_NF_True(module M, evector V);
coc_bool gb_Is_In_NF_LT(module M, evector V);
coc_bool gb_Is_In_NF_Full(module M, evector V);

/* ---[ Do_Make_Pair() ]--------------------------------------- */

vpair gb_Make_Pair_NoSugar(module M, evector V, evector W);
vpair gb_Make_Pair_Sugar(module M, evector V, evector W);

/* ---[ Do_Add_To_Rules() ]--------------------------------------- */

vpair gb_Add_To_Rules_Append(module M, evector V);

/* ---[ Do_Add_To_IVs() ]--------------------------------------- */

vpair gb_Add_To_IVs_Append(module M, evector V);
vpair gb_Add_To_IVs_AntiDeg(module M, evector V);

/* ---[ Do_Pair_Prec() ]--------------------------------------- */

coc_bool gb_Pair_Prec_AntiDegRevLex(module M, vpair P1, vpair P2);
coc_bool gb_Pair_Prec_AntiDeg(module M, vpair P1, vpair P2);
coc_bool gb_Pair_Prec_SugarAntiLex(module M, vpair P1, vpair P2);
coc_bool gb_Pair_Prec_Sugar(module M, vpair P1, vpair P2);

#endif /*  GB_H  */
