/* $Id: hilop.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $

	module HILOP

	"This module contains algorithm based on Grobner Computation,
	like Intersection of Ideals or Modules, Elimination of variables,
	Comparizon (?) between Ideals or Modules, etc."

[ Polynomials ] [ Ideals ] [ HDriven Ideals ] [ Modules ]

--------------------------------------------------------- */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "evector.h"
#include "gb.h"
#include "gb_kf.h"
#include "hilop.h"
#include "ivectors.h"
#include "jintf.h"
#include "module.h"
#include "monomial.h"
#include "options.h"
#include "ordering.h"
#include "polys.h"
#include "pvector.h"
#include "pseries.h"
#include "pair.h"
#include "ring.h"
#include "terms.h"
#include "vlist.h"
#include "weights.h"

extern object math_error;
extern object cocoa_error(char * id);

/* ---[ Polynomials ]----------------------------------------------- */

arpoly arpoly_lcm(arpoly p, arpoly q, ring R)
{
  /* TODO: optimize ? */
  module I,J,IJ;
  arpoly res_p;
  pvcomp aux_p;

  if (p == NULL) return NULL;
  if (q == NULL) return NULL;
  if (ring_characteristic(R) < 2) /* we can handle this case faster */
  {
    arpoly lcm, zero;
    arpoly pq = arpoly_mult(p, q, R);
    arpoly gcd = arpoly_gcd_JAA(p, q, R);
    arpoly_div(pq, gcd, &lcm, &zero, R);
    arpoly_free(zero, R);
    arpoly_free(gcd, R);
    arpoly_free(pq, R);
    return lcm;
  }

  I = module_generated_by_arpoly( p,R );
  J = module_generated_by_arpoly( q,R );

  IJ = module_intersection(I,J,R);

  /* BOSSI: optimize : avoid duplication */
  aux_p=vlist_first_poly(module_gens(IJ));
  res_p=pvcomp2arpoly(aux_p,R);

  module_unlink(IJ,R);
  module_unlink(I,R);
  module_unlink(J,R);

  return res_p;
}

arpoly arpoly_gcd(arpoly p, arpoly q, ring R)
{
  /* BOSSI : optimize */

  arpoly pq,lcm,res_p,r;

  if (p == NULL) return arpoly_dup(q,R);
  if (q == NULL) return arpoly_dup(p,R);
  if (arpoly_is_nonzero_const(p,R)) return arpoly_one(R);
  if (arpoly_is_nonzero_const(q,R)) return arpoly_one(R);
  if (ring_characteristic(R) < 2) return arpoly_gcd_JAA(p, q, R);

  pq=arpoly_mult(p,q,R);
  lcm=arpoly_lcm(p,q,R);
  arpoly_div(pq,lcm,&res_p,&r,R);

  arpoly_free(pq,R);
  arpoly_free(r,R);
  arpoly_free(lcm,R);

  return res_p;
}

/* ---[ Ideals ]----------------------------------------------- */


/* ---[ Modules ]----------------------------------------------- */

void GBasisOfModule(module M)
{
  if (module_wip(M)) {
    math_error=cocoa_error("CIP");
    return;
  }
  if (module_gbasis(M) != vlist_NULL) return;
  module_gb_set_initialized(M,FALSE);
  gb_set_kf(M, gb_kf_Do_1, gb_kf_Do_1_GB);
  gb_set_kf(M, gb_kf_Do_SPoly, gb_kf_SPoly_GB);
  gb_set_kf(M, gb_kf_Do_Reduce, gb_kf_Reduce_Full);
  gb_set_kf(M, gb_kf_Do_Case_Zero, gb_kf_Do_Case_0_Nothing);
  gb_set_kf(M, gb_kf_Do_Update, gb_kf_Update_Basic);
  gb_set_kf(M, gb_kf_Do_Get_Rule, gb_kf_Get_Rule_First);
  gb_set_kf(M, gb_kf_Do_Apply_Rule, gb_kf_Apply_Rule_GB);
  gb_set_kf(M, gb_kf_Do_Add_To_Rules, gb_kf_Add_To_Rules_Append);
  gb_set_kf(M, gb_kf_Do_Add_To_IVs, gb_kf_Add_To_IVs_AntiDeg);
  gb_set_kf(M, gb_kf_Do_Pair_Prec, gb_kf_Pair_Prec_SugarAntiRevLex);
  gb_set_kf(M, gb_kf_Do_IsIn_NF, gb_kf_Is_In_NF_True);
  gb_set_kf(M, gb_kf_Do_Make_Pair, gb_kf_Make_Pair_Sugar);
  gb_set_kf(M, gb_kf_Do_Better_Rule, gb_kf_Better_Rule_Undef);
  gb_set_kf(M, gb_kf_Do_Case_Not_NF, gb_kf_Do_Case_Not_NF_Keep);
  gb_set_kf(M, gb_kf_Do_Init, gb_kf_Init_Basic);
  gb_Do_All(M);
}

evector gb_NF_aux(evector V, module M, int kf_Reduce)
{
  evector res_ev;
  if (module_wip(M)) {
    math_error=cocoa_error("CIP");
    return ev_NULL;
  }
  if (gb_Rules(M) == vlist_NULL) GBasisOfModule(M);
  module_gb_set_initialized(M,FALSE);
  gb_set_kf(M, gb_kf_Do_1, gb_kf_Do_1_NF);
  gb_set_kf(M, gb_kf_Do_SPoly, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Reduce, kf_Reduce);
  gb_set_kf(M, gb_kf_Do_Case_Zero, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Update, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Get_Rule, gb_kf_Get_Rule_First);
  gb_set_kf(M, gb_kf_Do_Apply_Rule, gb_kf_Apply_Rule_GB);
  gb_set_kf(M, gb_kf_Do_Add_To_Rules, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Add_To_IVs, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Pair_Prec, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_IsIn_NF, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Make_Pair, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Better_Rule, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Case_Not_NF, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Init, gb_kf_Init_NF);
  ev_unlink(gb_CV(M),module_ring(M));
  gb_set_CV(M, ev_dup(V,module_ring(M)));
  gb_Do_All(M);
  res_ev=gb_CV(M);
  gb_set_CV(M, ev_NULL);
  return res_ev;
}

void Interreduce_this_vlist(vlist * vl, ring R )
{
  module M;
  if (vlist_is_empty(*vl)) return;
  M = this_vlist2module(*vl,vlist_rank(*vl),R);
  module_gb_set_initialized(M,FALSE);
  gb_set_kf(M, gb_kf_Do_1, gb_kf_Do_1_Interreduce);
  gb_set_kf(M, gb_kf_Do_SPoly, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Reduce, gb_kf_Reduce_Full);
  gb_set_kf(M, gb_kf_Do_Case_Zero, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Update, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Get_Rule, gb_kf_Get_Rule_First);
  gb_set_kf(M, gb_kf_Do_Apply_Rule, gb_kf_Apply_Rule_GB);
  gb_set_kf(M, gb_kf_Do_Add_To_Rules, gb_kf_Add_To_Rules_Append);
  gb_set_kf(M, gb_kf_Do_Add_To_IVs, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Pair_Prec, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_IsIn_NF, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Make_Pair, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Better_Rule, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Case_Not_NF, gb_kf_Undef);
  gb_set_kf(M, gb_kf_Do_Init, gb_kf_Init_Interreduce);
  gb_Do_All(M);
  *vl=gb_Rules(M);
  vlist_reinit_vectors(*vl,R);
  gb_set_Rules(M,vlist_NULL);
  module_unlink(M,R);
}

void SyzygyModule( module M, int comps_num )
{
  if (module_wip(M)) {
    math_error=cocoa_error("CIP");
    return;
  }
  if (module_syz(M) != vlist_NULL) {
    if (module_SyzCompsNum(M) == comps_num) return;
    else gb_free(M); /* clear computation */
  }
  module_set_SyzCompsNum(M,comps_num);
  module_gb_set_initialized(M,FALSE);
  gb_set_kf(M, gb_kf_Do_1, gb_kf_Do_1_Syz);
  gb_set_kf(M, gb_kf_Do_SPoly, gb_kf_SPoly_Syz);
  gb_set_kf(M, gb_kf_Do_Reduce, gb_kf_Reduce_Full);
  gb_set_kf(M, gb_kf_Do_Case_Zero, gb_kf_Do_Case_0_Syz_G);
  gb_set_kf(M, gb_kf_Do_Update, gb_kf_Update_Basic);
  gb_set_kf(M, gb_kf_Do_Get_Rule, gb_kf_Get_Rule_First);
  gb_set_kf(M, gb_kf_Do_Apply_Rule, gb_kf_Apply_Rule_Syz);
  gb_set_kf(M, gb_kf_Do_Add_To_Rules, gb_kf_Add_To_Rules_Append);
  gb_set_kf(M, gb_kf_Do_Add_To_IVs, gb_kf_Add_To_IVs_Append);
  gb_set_kf(M, gb_kf_Do_Pair_Prec, gb_kf_Pair_Prec_SugarAntiRevLex);
  gb_set_kf(M, gb_kf_Do_IsIn_NF, gb_kf_Is_In_NF_True);
  gb_set_kf(M, gb_kf_Do_Make_Pair, gb_kf_Make_Pair_Sugar);
  gb_set_kf(M, gb_kf_Do_Better_Rule, gb_kf_Better_Rule_Undef);
  gb_set_kf(M, gb_kf_Do_Case_Not_NF, gb_kf_Do_Case_Not_NF_Keep);
  gb_set_kf(M, gb_kf_Do_Init, gb_kf_Init_Syz);
  gb_Do_All(M);
}

coc_bool gb_NFsAreZero(vlist vl, module M)
{
  ring R;
  int n;
  evector aux_v;
  vlist_atom aux_atom;
  R = module_ring(M);

  n=vlist_len(vl);
  aux_atom=vlist_first(vl);
  while ( n-- > 0 ) {
    aux_v = gb_LTNF(vlist_atom_ev(aux_atom),M);
    if (ev_is_zero_ck(aux_v)) {
      aux_atom=vlist_atom_next(aux_atom);
      ev_unlink(aux_v,R);
    } else {
      ev_unlink(aux_v,R);
      return FALSE; 
    }
  }
  return TRUE;
}

module module_leading_terms( module M, ring R ) 
{
  module res_M = module_init(module_rank(M),R);   
  GBasisOfModule(M);
  module_set_gens(res_M,vlist_only_leading_terms(module_gbasis(M), R));
  module_set_gbasis(res_M, vlist_dup(module_gens(res_M),R));
  return res_M;
}

module module_intersection(module M, module N, ring R)
{
  int compM, compN, comp_num;
  vlist aux_vl,res_vl;
  module aux_M,res_M;
  if (module_not_compatible(M,N)) return module_NULL;
  compM = vlist_len(module_gens(M));
  compN = vlist_len(module_gens(N));
  comp_num = MIN(compM,compN);
  if (comp_num==compM) {
    aux_M = this_vlist2module(vlist_concat(module_gens(M),module_gens(N),R),
      module_rank(M),R);
    aux_vl = module_gens(M); }
  else {
    aux_M = this_vlist2module(vlist_concat(module_gens(N),module_gens(M),R),
      module_rank(M),R);
    aux_vl = module_gens(N); }
  SyzygyModule(aux_M, comp_num);
  res_vl = vlist_linear_comb(aux_vl,module_gens(module_syz(aux_M)),R);
  Interreduce_this_vlist(&res_vl,R);
  res_M = this_vlist2module(res_vl,module_rank(M),R);
  module_unlink(aux_M,R);
  return res_M;
}
