/* $Id: indet.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include <limits.h>

#include "cocdefs.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "indet.h"
#include "ivectors.h"
#include "operator.h"
#include "utils.h"

indet_entry indet_entry_mk(char * ident,ivector arrinfo,ivector from)
{
  indet_entry res_e;
  res_e=malloc_indet_entry();
  res_e->size=0;
  res_e->offset=0;
  res_e->ident=ident;
  res_e->arrinfo=arrinfo;
  res_e->from=from;
  return res_e;
}

indet_info indet_info_init(void)
{
  indet_info res_ii;
  res_ii=malloc_indet_info();
  res_ii->num_entries=0;
  res_ii->num_indets=0;
  return res_ii;
}

void indet_entry_free(indet_entry e)
{
  free_str(strlen(e->ident),e->ident);
  iv_free_check(e->arrinfo);
  iv_free_check(e->from);
  free_indet_entry(e);
}

void indet_info_free(indet_info ii)
{
  register int i;
  if (ii==indet_info_NULL) return;
  for (i=0;i<ii->num_entries;i++) indet_entry_free(ii->indets[i]);
  free_indet_info(ii);
}

indet_info indet_info_dup(indet_info ii)
{
  register int i;
  indet_info res_ii;
  if (ii==indet_info_NULL) return indet_info_NULL;
  res_ii=malloc_indet_info();
  res_ii->num_entries=ii->num_entries;
  res_ii->num_indets=ii->num_indets;
  for (i=0;i<ii->num_entries;i++) 
    res_ii->indets[i]=indet_entry_dup(ii->indets[i]);
  return res_ii;
}

indet_entry indet_entry_dup(indet_entry e)
{
  indet_entry res_e;
  res_e=malloc_indet_entry();
  res_e->offset=e->offset;
  res_e->size=e->size;
  res_e->ident=my_strdup(e->ident);
  res_e->arrinfo=iv_dup(e->arrinfo);
  res_e->from=iv_dup(e->from);
  return res_e;
}

void indet_info_add(indet_info ii, indet_entry e)
{
  register int i;
  e->offset=ii->num_indets;
  indet_info_set_entry(ii,ii->num_entries++,e);
  e->size=1;

  for (i=1;i<=iv_get_len(e->arrinfo);i++)
  {
    int temp = iv_get_nth(e->arrinfo,i);
    double check = temp;
    check *= e->size;
    if (check > 0 && check < SHRT_MAX)
      e->size = e->size * temp;
    else e->size = 0; /* set to zero to indicate ERROR; checked in eval_indet_info (math_aux.c) */
  }

  ii->num_indets+=e->size;
  /* Check for overflow... */
  if (ii->num_indets < e->size) e->size = 0; /* set to zero to indicate ERROR */
}

unknown count_indets(indet_info ii)
{
  if (ii==indet_info_NULL) return 0;
  return ii->num_indets;
}

int indet_info_search_entry_index(indet_info ii,char * ident)
{
  int res_i;
  register int i;
  for (i=0;i<ii->num_entries;i++)
    if (strcmp(ii->indets[i]->ident,ident)==0) return i;
  return -1;
}

unknown indet_info_get_unknown(indet_info ii,char * ident,ivector index)
{
  int i,n;
  indet_entry aux_e;
  unknown res_u;
  i = indet_info_search_entry_index(ii,ident);
  if (i == -1) return unknown_NULL;
  aux_e=ii->indets[i];
  n=iv_get_len(aux_e->arrinfo);
  if (iv_get_len(index) != n) return unknown_NULL;
  res_u=indet_info_index2rel_unknown(aux_e,index);
  if (res_u != -1) res_u += aux_e->offset;
  return res_u;
}

unknown indet_info_index2rel_unknown(indet_entry e,ivector index)
{
  register int i;
  int l;
  unknown res_u,s,sz;
  res_u=0;
  sz=e->size;
  l=iv_get_len(e->arrinfo);
  if (l==0) return 0; /* no indices */
  for (i=1;i<=l-1;i++) {
    sz/=iv_get_nth(e->arrinfo,i);
    s=iv_get_nth(index,i)-iv_get_nth(e->from,i);
    if (s>=iv_get_nth(e->arrinfo,i) || s<0) return unknown_NULL;
    res_u+=s*sz; }
  s=iv_get_nth(index,l)-iv_get_nth(e->from,l);
  if (s>=iv_get_nth(e->arrinfo,l) || s<0) return unknown_NULL;
  return res_u+s;
}

ivector indet_info_rel_unknown2index(indet_entry e,unknown u)
{
  register int i;
  ivector res_iv;
  unknown sz;
  res_iv=iv_init(iv_get_len(e->arrinfo));
  sz=e->size;
  if (sz > 1) {
    for (i=1;i<=iv_get_len(e->arrinfo)-1;i++) {
      sz/=iv_get_nth(e->arrinfo,i);
      iv_put_nth(res_iv,i,u/sz);
      u=u % sz; }
    iv_put_nth(res_iv,iv_get_len(e->arrinfo),u);
  }
  return res_iv;
}

ivector indet_indices(indet_info ii, unknown u)
{
  ivector res_v;
  register int i;
  int n;
  indet_entry aux_e;
  i=ii->num_entries-1;
  while (ii->indets[i--]->offset>u);
  aux_e=ii->indets[++i];
  n=iv_get_len(aux_e->arrinfo);

  res_v=indet_info_rel_unknown2index(aux_e,(unknown)(u-aux_e->offset));
  
  for (i=1;i<=n;i++)
    iv_set_nth(res_v,i,iv_get_nth(res_v,i)+iv_get_nth(aux_e->from,i));

  return res_v;
}

char * indet_name(indet_info ii, unknown u)
{
  register int i;
  indet_entry aux_e;
  i=ii->num_entries-1;
  while (ii->indets[i--]->offset>u);
  aux_e=ii->indets[++i];
  return aux_e->ident;
}

cterm indet_info_get_cterm(indet_info ii,unknown u)
{
  /* TODO: REDO using indet_info_get_indices */
  /* TODO: extend to x_i */
  cterm_list params;
  cterm res_t;
  ivector index;
  register int i;
  int n;
  indet_entry aux_e;
  i=ii->num_entries-1;
  while (ii->indets[i--]->offset>u);
  aux_e=ii->indets[++i];
  n=iv_get_len(aux_e->arrinfo);
  params=malloc_cterm_list(n+1);
  params[0]=cterm_mk_NBS(aux_e->ident);

  index=indet_info_rel_unknown2index(aux_e,(unknown)(u-aux_e->offset));
  
  for (i=1;i<=n;i++)
    params[i]=int2cterm(iv_get_nth(index,i)+iv_get_nth(aux_e->from,i));

  iv_free(index);

  res_t=cterm_mk_classic(op_INDET,n+1,params);
  free_cterm_list(n+1,params);
  return res_t;
}

void indet_info_first_and_last(indet_info ii,char * ident,
                               unknown *first,unknown *last)
{
  int i;
  i=indet_info_search_entry_index(ii,ident);
  if (i==-1) *first=*last=unknown_NULL;
  else {
    *first=ii->indets[i]->offset;
    *last=ii->indets[i]->offset+ii->indets[i]->size-1; }
}

cterm indet_entry2cterm(indet_entry e)
{
  ctbuf aux_b;
  cterm params[2],aux_t;
  int i,n,from,to;
  aux_b=ctbuf_new(cterm_mk_NBS(e->ident));
  n=iv_get_len(e->arrinfo);
  for (i=1;i<=n;i++) {
    from=iv_get_nth(e->from,i);
    to=iv_get_nth(e->arrinfo,i)+from-1;
    params[0]=int2cterm(from);
    params[1]=int2cterm(to);
    aux_t=cterm_mk_classic(op_RANGE,2,params);
    ctbuf_append(aux_b,aux_t);
  }
  return ctbuf_mk_cterm(aux_b,op_INDET_DECL);
}

cterm indet_info2cterm(indet_info ii)
{
  ctbuf aux_b;
  int i;
  aux_b=ctbuf_empty();
  for (i=0;i<ii->num_entries;i++)
    ctbuf_append(aux_b,indet_entry2cterm(ii->indets[i]));
  return ctbuf_mk_cterm(aux_b,op_INDET_INFO);
}


ctbuf ctbuf_indets2indetinfo(char * indets)
{
  register int i;
  char aux_s[2];
  ctbuf aux_b;
  cterm subparams[1];
  int len=strlen(indets);
  aux_b=ctbuf_empty();
  for (i=0;i<len;i++) {
    aux_s[0]=indets[i];
    aux_s[1]='\0';
    subparams[0]=cterm_mk_NBS(aux_s);
    aux_b=ctbuf_append(aux_b,cterm_mk_classic(op_INDET_DECL,1,subparams)); }
  return aux_b;
}



/*

[x[1..3,2..5]y[5]z]

LIST(
_INDET_DECL("x",_RANGE("1","3"),_RANGE("2","5")),
_INDET_DECL("y","5")),
_INDET_DECL("z")
)

*/
