/* $Id: ivectors.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"
#include "ca.h"

#ifndef IVECTORS_H
# include "ivectors.h"
#endif

#define iv_sum_nth(v,n,e) { v[n] += e; }
#define iv_sub_nth(v,n,e) { v[n] -= e; }

#ifdef DEBUG_IVECTORS

iv_elem iv_get_nth(ivector v, int n)
{
  if (n <= 0 || n > iv_get_len(v)) {
    char buffer[100];
    sprintf(buffer,"iv_get_nth: index %d out of range 1..%d", n, iv_get_len(v));
    INTERNAL_ERROR(buffer);
  }
  return M_iv_get_nth(v,n);
}

void iv_set_nth(ivector v, int n, iv_elem e)
{
  if (n <= 0 || n > iv_get_len(v)) {
    char buffer[100];
    sprintf(buffer,"iv_set_nth: index %d out of range 1..%d", n, iv_get_len(v));
    INTERNAL_ERROR(buffer);
  }
  M_iv_set_nth(v,n,e);
}

void iv_put_nth(ivector v, int n, iv_elem e)
{
  if (n <= 0 || n > iv_get_len(v)) {
    char buffer[100];
    sprintf(buffer,"iv_put_nth: index %d out of range 1..%d", n, iv_get_len(v));
    INTERNAL_ERROR(buffer);
  }
  M_iv_put_nth(v,n,e);
}

void iv_put0_nth(ivector v, int n)
{
  if (n <= 0 || n > iv_get_len(v)) {
    char buffer[100];
    sprintf(buffer,"iv_put0_nth: index %d out of range 1..%d", n, iv_get_len(v));
    INTERNAL_ERROR(buffer);
  }
  M_iv_put0_nth(v,n);
}

#endif

ivector iv_map(ivector v, ivector w1, ivector w2, ivector m)
{ /* m is the map, w1 = weights(R1), w2 = weights(R2) */
  ivector res_v;
  register int i;
  res_v=iv_init(iv_get_len(m));
  for (i=iv_get_len(m);i>0;i--) {
    if (iv_get_nth(m,i) != 0)
      iv_put_nth(res_v,i,
        iv_get_nth(v,iv_get_nth(m,i))
        /iv_get_nth(w1,iv_get_nth(m,i))
        *iv_get_nth(w2,i));
  }
  return res_v;
}

ivector iv_init(iv_elem n)
{
  ivector aux_v;
  register int i;
  aux_v = malloc_iv(n);
  iv_set_len(aux_v,n);
  iv_set_total(aux_v,0);
  for (i=n;i>0;i--) iv_set_nth(aux_v,i,0);
  return aux_v;
}

int iv_which_gt_lex(ivector v1, ivector v2)
{
  register int i;
  for (i=1;i<=iv_get_len(v1);i++) {
    if (iv_get_nth(v1,i) > iv_get_nth(v2,i)) return ANSWER_FIRST;
    if (iv_get_nth(v1,i) < iv_get_nth(v2,i)) return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

void iv_assign(ivector v1, ivector v2)
{
  register int i;
  for (i=iv_get_len(v2);i>0;i--) 
    iv_set_nth(v1,i,iv_get_nth(v2,i));
  iv_set_total(v1,iv_get_sum(v2));
}

ivector iv_create_and_fill(iv_elem n,iv_elem e)
{
  register int i;
  ivector aux_v;
  aux_v=malloc_iv(n);
  iv_set_len(aux_v,n);
  iv_set_total(aux_v,n*e);
  for (i=n;i>0;i--) iv_set_nth(aux_v,i,e);
  return aux_v;
}

void iv_fill(ivector v,iv_elem e)
{
  register int i;

  iv_set_total(v,iv_get_len(v)*e);
  for (i=iv_get_len(v);i>0;i--) iv_set_nth(v,i,e);
}

void iv_change_sign(ivector v)
{
  register int i;

  iv_set_total(v,-iv_get_sum(v));
  for (i=iv_get_len(v);i>0;i--) iv_set_nth(v,i,-iv_get_nth(v,i));
}

ivector iv_dup(ivector v)
{
  ivector aux_v;
  register int i;
  if (v==iv_NULL) return iv_NULL;
  aux_v=malloc_iv(iv_get_len(v));
  iv_set_len(aux_v,iv_get_len(v));
  for (i=iv_get_len(v);i>0;i--) iv_set_nth(aux_v,i,iv_get_nth(v,i));
  iv_set_total(aux_v,iv_get_sum(v));
  return aux_v;
}

ivector iv_dup_no_sum(ivector v)
{
  ivector aux_v;
  register int i;
  if (v==iv_NULL) return iv_NULL;
  aux_v=malloc_iv(iv_get_len(v));
  for (i=iv_get_len(v);i>0;i--) iv_set_nth(aux_v,i,iv_get_nth(v,i));
  iv_set_len(aux_v,iv_get_len(v));
  return aux_v;
}

coc_bool iv_eq(ivector v1, ivector v2)
{
  register int i;
  if (iv_get_len(v1)!=iv_get_len(v2)) return FALSE;
  for (i=iv_get_len(v1);i>0;i--)
    if (iv_get_nth(v1,i)!=iv_get_nth(v2,i)) return FALSE;
  return TRUE;
}

ivector iv_sum(ivector v1,ivector v2)
{
  ivector aux_v;
  register int i;
  aux_v=malloc_iv(iv_get_len(v1));
  iv_set_len(aux_v,iv_get_len(v1));
  for (i=iv_get_len(v1);i>0;i--)
    iv_set_nth(aux_v,i,iv_get_nth(v1,i)+iv_get_nth(v2,i));
  iv_set_total(aux_v,iv_get_sum(v1)+iv_get_sum(v2));
  return aux_v;
}

ivector iv_sum_NOTOTAL(ivector v1,ivector v2)
{
  ivector aux_v;
  register int i;
  aux_v=malloc_iv(iv_get_len(v1));
  for (i=iv_get_len(v1);i>0;i--)
    iv_set_nth(aux_v,i,iv_get_nth(v1,i)+iv_get_nth(v2,i));
  iv_set_len(aux_v,iv_get_len(v1));
  return aux_v;
}

void iv_sum_and_assign(ivector v1,ivector v2)
{
  register int i;
  iv_set_total(v1,iv_get_sum(v1)+iv_get_sum(v2));
  for (i=iv_get_len(v2);i>0;i--)
    iv_sum_nth(v1,i,iv_get_nth(v2,i));
}

void iv_subtract_and_assign(ivector v1,ivector v2)
{
  register int i;
  iv_set_total(v1,iv_get_sum(v1)-iv_get_sum(v2));
  for (i=iv_get_len(v2);i>0;i--)
    iv_sub_nth(v1,i,iv_get_nth(v2,i));
}

ivector iv_subtract(ivector v1,ivector v2)
{
  ivector aux_v;
  register int i;
  aux_v=malloc_iv(iv_get_len(v1));
  iv_set_len(aux_v,iv_get_len(v1));
  for (i=iv_get_len(v1);i>0;i--)
    iv_set_nth(aux_v,i,iv_get_nth(v1,i)-iv_get_nth(v2,i));
  iv_set_total(aux_v,iv_get_sum(v1)-iv_get_sum(v2));
  return aux_v;
}

ivector iv_maxs(ivector v1,ivector v2)
{
  ivector aux_v;
  register int i;
  register int s;
  aux_v=malloc_iv(iv_get_len(v1));
  iv_set_len(aux_v,iv_get_len(v1));
  s = 0;
  for (i=iv_get_len(v1);i>0;i--) {
    iv_set_nth(aux_v,i,MAX(iv_get_nth(v1,i),iv_get_nth(v2,i)));
    s += iv_get_nth(aux_v,i); }
  iv_set_total(aux_v,s);
  return aux_v;
}

int iv_sum_maxs(ivector v1,ivector v2)
{
  register int i,s;
  s = 0;
  for (i=iv_get_len(v1);i>0;i--)
    s += MAX(iv_get_nth(v1,i),iv_get_nth(v2,i));
  return s;
}

ivector iv_mins(ivector v1,ivector v2)
{
  ivector aux_v;
  register int i;
  register int s;
  aux_v=malloc_iv(iv_get_len(v1));
  iv_set_len(aux_v,iv_get_len(v1));
  s = 0;
  for (i=iv_get_len(v1);i>0;i--) { 
    iv_set_nth(aux_v,i,MIN(iv_get_nth(v1,i),iv_get_nth(v2,i)));
    s += iv_get_nth(aux_v,i); }
  iv_set_total(aux_v,s);
  return aux_v;
}

coc_bool iv_all_le(ivector v1,ivector v2)
{
  register int i;
  for (i=iv_get_len(v1);i>0;i--)
    if (iv_get_nth(v1,i)>iv_get_nth(v2,i)) return FALSE;
  return TRUE;
}

coc_bool iv_all_pos(ivector v)
{
  register int i;
  for (i=iv_get_len(v);i>0;i--)
    if (iv_get_nth(v,i)<0) return FALSE;
  return TRUE;
}

coc_bool iv_coprime(ivector v1, ivector v2)
{
  register int i;
  for (i=iv_get_len(v1);i>0;i--)
    if (iv_get_nth(v1,i)!=0 && iv_get_nth(v2,i)!=0) return FALSE;
  return TRUE;
}


int iv_which_all_le(ivector v1, ivector v2)
/* return ANSWER_BOTH iff v1==v2 */
/* return ANSWER_FIRST iff v1[i]<=v2[i] for all i */
/* return ANSWER_SECOND iff v2[i]<=v1[i] for all i */
/* return ANSWER_NC otherwise */
{
  register int i=iv_get_len(v1);
  while (i>0 && iv_get_nth(v1,i) == iv_get_nth(v2,i)) i--;
  if (i==0) return ANSWER_BOTH;
  if (iv_get_nth(v1,i)<iv_get_nth(v2,i)) {
    while (i>0 && iv_get_nth(v1,i) <= iv_get_nth(v2,i)) i--;
    if (i==0) return ANSWER_FIRST; else return ANSWER_NC;
  } else {
    while (i>0 && iv_get_nth(v1,i) >= iv_get_nth(v2,i)) i--;
    if (i==0) return ANSWER_SECOND; else return ANSWER_NC;
  }
}

coc_bool iv_is_zero(ivector v)
{
  register int i;
  if (v==iv_NULL) return TRUE;
  for (i=iv_get_len(v); i>0; i--)
    if (iv_get_nth(v,i) != ZERO) return FALSE;
  return TRUE;
}

iv_elem iv_scalar(ivector v1, ivector v2)
{
  register int i;
  register iv_elem res=0;
  for (i=iv_get_len(v1); i>0; i--)
    res+=iv_get_nth(v1,i)*iv_get_nth(v2,i);
  return res;
}

void iv_fill_maxs(ivector v, ivector v1, ivector v2)
{
  register int i, s=0;
  for (i=iv_get_len(v1);i>0;i--) {
    iv_set_nth(v,i,MAX(iv_get_nth(v1,i),iv_get_nth(v2,i)));
    s += iv_get_nth(v,i); }
  iv_set_total(v,s);
}

void iv_fill_subtract(ivector v, ivector v1, ivector v2)
{
  register int i;
  for (i=iv_get_len(v1);i>0;i--) 
    iv_set_nth(v,i,iv_get_nth(v1,i)-iv_get_nth(v2,i));
  iv_set_total(v,iv_get_sum(v1)-iv_get_sum(v2));
}

ivector iv_elim_nth(ivector v, iv_elem n)
{
  register int i;
  ivector res_v;
  res_v=malloc_iv(iv_get_len(v)-1);
  iv_set_len(res_v,iv_get_len(v)-1);
  for (i=1;i<n;i++) iv_set_nth(res_v,i,iv_get_nth(v,i));
  for (i=n;i<iv_get_len(v);i++) iv_set_nth(res_v,i,iv_get_nth(v,i+1));
  iv_set_total(res_v,0);
  return res_v;
}

coc_bool iv_almost_null (ivector v, unknown *index, iv_elem *non_null_elem)
{
  register int i = iv_get_len(v);

  while ( iv_get_nth(v,i--) == 0 );
  return (*non_null_elem = iv_get_nth(v,(*index=i+1))) == iv_get_sum(v);
}

/* Combinatorial stuff */

coc_bool iv_next_comb(ivector v,int n)
{
  int p,k,i;
  p=iv_get_len(v);
  k=0;
  for (i=p;i>=1;i--)
    if (iv_get_nth(v,i)<n-p+i) {
      k=i;
      break; }
  if (k==0) return FALSE;
  iv_set_nth(v,k,iv_get_nth(v,k)+1);
  for (i=k+1;i<=p;i++)
    iv_set_nth(v,i,iv_get_nth(v,i-1)+1);
  return TRUE;
}

/* End Combinatorial stuff */

ivector iv_first_m_elems ( ivector v, int m)
{
  ivector result;
  register int i;

  result=iv_init(m);
  for ( i=1; i<=m ; i++) 
    iv_put_nth(result,i,iv_get_nth(v,i));
  return result;
}

/* -----[ IVEC ]------------------------------ */

ivec ivec_init(int n)
{
  ivec aux_v;
  aux_v = malloc_ivec(n);
  ivec_set_len(aux_v,n);
  return aux_v;
}

int ivec_which_gt_lex(ivec v1, ivec v2)
{
  register int i;
  int n=ivec_len(v1);
  for (i=1;i<=n;i++) {
    if (ivec_nth(v1,i) > ivec_nth(v2,i)) return ANSWER_FIRST;
    if (ivec_nth(v1,i) < ivec_nth(v2,i)) return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

ivec ivec_sum(ivec v1, ivec v2)
{
  ivec res_v;
  register int i;
  res_v=malloc_ivec(ivec_len(v1));
  for (i=ivec_len(v1);i>0;i--)
    ivec_set_nth(res_v,i,ivec_nth(v1,i)+ivec_nth(v2,i));
  ivec_set_len(res_v,ivec_len(v1));
  return res_v;
}

ivec ivec_sub(ivec v1, ivec v2)
{
  ivec res_v;
  register int i;
  res_v=malloc_ivec(ivec_len(v1));
  for (i=ivec_len(v1);i>0;i--)
    ivec_set_nth(res_v,i,ivec_nth(v1,i)-ivec_nth(v2,i));
  ivec_set_len(res_v,ivec_len(v1));
  return res_v;
}

ivec ivec_dup(ivec v)
{
  ivec res_v;
  register int i;
  int n=ivec_len(v);
  res_v=malloc_ivec(ivec_len(v));
  for (i=n;i>0;i--) ivec_set_nth(res_v,i,ivec_nth(v,i));
  ivec_set_len(res_v,n);
  return res_v;
}

ivec ivec_dup_from(ivec v, ivec_elem n)
{
  static c=0;
  ivec res_v;
  register int i,l;
  l=ivec_len(v)-n;
  if (l==0) {
    return ivec_NULL;
  }
  res_v=malloc_ivec(l);
  for (i=l;i>0;i--) ivec_set_nth(res_v,i,ivec_nth(v,i+n));
  ivec_set_len(res_v,l);
  return res_v;
}

ivec ivec_dup_to(ivec v, ivec_elem n)
{
  ivec res_v;
  register int i,l;
  if (n==0) return ivec_NULL;
  res_v=malloc_ivec(n);
  for (i=n;i>0;i--) ivec_set_nth(res_v,i,ivec_nth(v,i));
  ivec_set_len(res_v,n);
  return res_v;
}

coc_bool ivec_is_zero(ivec v)
{
  register int i;
  for (i=ivec_len(v); i>0; i--)
    if (ivec_nth(v,i) != ZERO) return FALSE;
  return TRUE;
}

ivec ivec_this_inv(ivec v)
{
  register int i;
  for (i=ivec_len(v);i>0;i--) ivec_set_nth(v,i,-ivec_nth(v,i));
  return v;
}

coc_bool ivec_neg_coprime(ivec v1, ivec v2)
{
  register int i;
  for (i=ivec_len(v1) ; i>0 ; i-- ) 
    if ( ivec_nth(v1,i)<0 && ivec_nth(v2,i)<0 ) return FALSE;
  return TRUE;
}
