/* $Id: ivectors.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef IVECTORS_H
#define IVECTORS_H

#ifndef UTILS_H
# include "utils.h"
#endif

#define malloc_iv(n) ((ivector)(mymalloc(sizeof(iv_elem)*(n+2),"ivector"))+1)
#define calloc_iv(n) ((ivector)(mycalloc(n+2,sizeof(iv_elem),"ivector"))+1)
#define free_iv(n,p) myfree(sizeof(iv_elem)*(n+2),(p)-1,"ivector")

#define malloc_ivec(n) (ivec)mymalloc(sizeof(ivec_elem)*(n+1),"ivec")
#define free_ivec(n,p) myfree(sizeof(ivec_elem)*(n+1),p,"ivec")

/* v[0] = len(v) */
/* v[1] = sum of components(v) */

#define iv_get_len(v) (v)[-1]
#define iv_get_sum(v) (v)[0]
#define iv_set_total(v,s) (v)[0]=s
#define iv_set_len(v,l) (v)[-1]=l

/* ----------------------------------------------- */
#define M_iv_get_nth(v,n) ((v)[n])
#define M_iv_set_nth(v,n,e) (v)[n] = e

#define M_iv_put_nth(v,n,e) \
  do { (v)[0]-= (v)[n]; (v)[0]+= ((v)[n] = e); } while(FALSE)
#define M_iv_put0_nth(v,n) \
  do { (v)[0]-= (v)[n]; (v)[n] = 0; } while(FALSE)

#ifdef DEBUG_IVECTORS

iv_elem iv_get_nth(ivector v, int n);
void iv_set_nth(ivector v, int n, iv_elem e);
void iv_put_nth(ivector v, int n, iv_elem e);
void iv_put0_nth(ivector v, int n);

#else

#define iv_put_nth M_iv_put_nth
#define iv_get_nth M_iv_get_nth
#define iv_set_nth M_iv_set_nth
#define iv_put0_nth M_iv_put0_nth
  
#endif

/* ----------------------------------------------- */

#define iv_incr_nth(v,n,e) iv_put_nth(v,n,iv_get_nth(v,n)+e)

/* assumed v!=iv_NULL */
#define iv_free(v) free_iv(iv_get_len(v),v);
#define iv_free_check(v) if (v!=iv_NULL) free_iv(iv_get_len(v),v)
#define iv_fill_zero(v) iv_fill(v,ZERO)

ivector iv_map(ivector v, ivector w1, ivector w2, ivector m);
coc_bool iv_is_zero(ivector v);

/* in the following we assume that v is built with iv_put_nth */
/* i.e.: the field v[1] is updated */

/* EVERYWHERE assumed v1 initialized, with the same length of v2 */

ivector iv_init(int n);
int iv_which_gt_lex(ivector v1, ivector v2);

void iv_assign(ivector v1, ivector v2);
/* v1 := v2 */

ivector iv_create_and_fill(iv_elem n,iv_elem e);
/* create and fill an ivector with all e */

void iv_fill(ivector v,iv_elem e);
/* fill an ivector with all e */

void iv_change_sign(ivector v);

ivector iv_dup(ivector v);
/* duplicate v */

ivector iv_dup_no_sum(ivector v);
/* duplicate v but not update sum */

coc_bool iv_eq(ivector v1, ivector v2);
/* TRUE iff v1==v2 */

#define iv_set_sum(v,v1,v2) do {\
  register int i;\
  v=malloc_iv(iv_get_len(v1));\
  iv_set_len(v,iv_get_len(v1));\
  for (i=iv_get_len(v1);i>0;i--)\
    iv_set_nth(v,i,iv_get_nth(v1,i)+iv_get_nth(v2,i));\
  iv_set_total(v,iv_get_sum(v1)+iv_get_sum(v2)); } while (FALSE)

ivector iv_sum(ivector v1,ivector v2);
/* returns v1+v2 */

ivector iv_sum_NOTOTAL(ivector v1,ivector v2);

void iv_sum_and_assign(ivector v1,ivector v2);
/* v1+=v2 */

ivector iv_subtract(ivector v1,ivector v2);
/* returns v1-v2 */

void iv_subtract_and_assign(ivector v1,ivector v2);
/* returns v1-=v2 */

ivector iv_maxs(ivector v1,ivector v2);
/* returns v[i]=max(v1[i],v2[i]) */

int iv_sum_maxs(ivector v1,ivector v2);

ivector iv_mins(ivector v1,ivector v2);
/* returns v[i]=min(v1[i],v2[i]) */

coc_bool iv_all_le(ivector v1,ivector v2);
/* TRUE iff v1[i]<=v2[i] for all i */

coc_bool iv_all_pos(ivector v);
/* TRUE iff v[i]>=0 for all i */

coc_bool iv_coprime(ivector v1, ivector v2);
/* TRUE iff v1[i]*v2[i]=0 for all i */

int iv_which_all_le(ivector v1, ivector v2);
/* return ANSWER_BOTH iff v1==v2 */
/* return ANSWER_FIRST iff v1[i]<=v2[i] for all i */
/* return ANSWER_SECOND iff v2[i]<=v1[i] for all i */
/* return ANSWER_NC otherwise */

iv_elem iv_scalar(ivector v1, ivector v2);
/* return v1*v2 */

void iv_fill_maxs(ivector v, ivector v1, ivector v2);
/* v[i]=max(v1[i],v2[i]) */

void iv_fill_subtract(ivector v, ivector v1, ivector v2);
/* v[i]=v1[i]-v2[i] */

ivector iv_elim_nth(ivector v, iv_elem n);

void printf_iv(ivector v);

coc_bool iv_almost_null (ivector v, unknown *index, iv_elem *non_null_elem);
/* assumed v!=0 */


/* Combinatorial stuff */

coc_bool iv_next_comb(ivector v,int n);
/* next combination: example: (n=5)
  1234 -> 1235 -> 1245... */

/* End Combinatorial stuff */

ivector iv_first_m_elems ( ivector v, int m); /* from HDRIVEN */



/* -----[ IVEC ]------------------------------ */

#define ivec_len(v) (v)[0]
#define ivec_set_len(v,n) (v)[0]=n
#define ivec_len_ck(v) (v==ivec_NULL ? ZERO : (v)[0])
#define ivec_nth(v,n) ((v)[n])
#define ivec_set_nth(v,n,x) (v)[n]=x
#define ivec_free(v) free_ivec(ivec_len(v),v)
#define ivec_free_ck(v) if (v!=ivec_NULL) free_ivec(ivec_len(v),v)
#define ivec_sum_ck(v1,v2) (v1 == ivec_NULL ? v1 : ivec_sum(v1,v2))
#define ivec_which_gt_lex_ck(v1,v2) \
  (v1 == ivec_NULL ? ANSWER_BOTH : ivec_which_gt_lex(v1,v2))

#define ivec_dup_ck(v) (v==ivec_NULL ? ivec_NULL : ivec_dup(v))
#define ivec_dup_to_ck(v,n) (n==ZERO ? ivec_NULL : ivec_dup_to(v,n))

#define ivec_is_zero_ck(v) (v==ivec_NULL ? TRUE : ivec_is_zero(v))

ivec ivec_init(int n);
ivec ivec_dup(ivec v);
ivec ivec_dup_from(ivec v, ivec_elem n);
ivec ivec_dup_to(ivec v, ivec_elem n);
ivec ivec_sum(ivec v1, ivec v2);
ivec ivec_sub(ivec v1, ivec v2);
int ivec_which_gt_lex(ivec v1, ivec v2);
coc_bool ivec_is_zero(ivec v);

#endif  /* IVECTOR_H */
