/* map.c : 0.5 : 21 Dec 94 : map library */

/*-----------------------------------------------------------------------------

	Part of project "map".
	
-----------------------------------------------------------------------------*/

#include "cocdefs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

# include "map.h"
# include "options.h"
/* need operator.h for op_NULL  (used once only in map_append) */
# include "operator.h"

/* CONSTRUCTORS of map */

map map_empty;

void map_init_maps(void)
{
  /* nothing to do */
  map_empty=map_new();
}

map map_new(void)
{ 
  map m;
  m=malloc_map(1);
  if (m!=NULL) {
    map_set_len(m,SIZE_OF_MAP_BLOCK+2); /* initially a map is big a block */
    map_set_last_id(m,first_name_id-1); }
  return m;
}

name_id map_add(map * m,nome n)
{
  if (map_is_full(*m)) {
    map old_m=*m;
    *m=realloc_map(*m,1); /* adds a block */
    map_set_len(*m,map_len(*m)+SIZE_OF_MAP_BLOCK);
#ifdef DEBUG_MAP
    { char buf[80];
    sprintf(buf,"--[ DEBUG: map_add: new block (total %d) %d-->%d  ]--\n", 
      map_len(*m),old_m,*m);
    cocoa_debug_message(buf); }
#endif
  }

  return map_append(*m,n);
}

name_id map_append(map m,nome n)
{ 
  name_id i;
  i=map_last_id(m)+1;
  if (i >= MAX_NUM_OPERATORS) return op_NULL;
  map_set_last_id(m,i);
  map_set_name(m,i,my_strdup(n));
#ifdef DEBUG_MAP
  { char buf[80];
    sprintf(buf,"--[ DEBUG: map_add: new operator %d.%d=%s ]--\n",m,i,n);
    cocoa_debug_message(buf); }
#endif
  return i;
}

void map_free(map * m)
{
  name_id i;
  for (i=first_name_id;i<=map_last_id(*m);i++)
    free_name(map_get_name(*m,i));
  free_map(*m);
  m=NULL;
}

name_id map_get_name_id(map m, nome n)
{ 
  name_id i;
  for (i=first_name_id;i<=map_last_id(m);i++)
    if (strcmp(map_get_name(m,i),n)==0) return i;
  return name_id_NULL;
}

int map_num_starting(map m, nome n)
{ 
  int i,res_n=0;
  for (i=first_name_id;i<=map_last_id(m);i++)
    if (strncmp(map_get_name(m,i),n,strlen(n))==0) res_n++;
  return res_n;
}

name_id map_next_starting(map m, nome n, name_id from)
{ 
  name_id i;
  for (i=from+1;i<=map_last_id(m);i++)
    if (strncmp(map_get_name(m,i),n,strlen(n))==0) return i;
  return name_id_NULL;
}
