/* $Id: math.c,v 4.5 2004/07/27 10:57:25 cocoa Exp $
  [ HANDLERS ]
------------------------------------------------------------ */

#include "cocdefs.h"
#include <stdlib.h>
#include <time.h>

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "coeffs.h"
#include "cocfun.h"
#include "engine.h"
#include "eterms.h"
#include "evector.h"
#include "gb.h"
#include "hilop.h"
#include "ivectors.h"
#include "indet.h"
#include "jintf.h"
#include "memoryco.h"
#include "machine.h"
#include "map.h"
#include "mathcoc.h"
#include "math_lib.h"
#include "monomial.h"
#include "module.h"
#include "olist.h"
#include "object.h"
#include "operator.h"
#include "ordering.h"
#include "options.h"
#include "omat.h"
#include "pair.h"
#include "polys.h"
#include "poincare.h"
#include "pseries.h"
#include "pvector.h"
#include "ratfun.h"
#include "record.h"
#include "ring.h"
#include "subst.h"
#include "terms.h"
#include "toric.h"
#include "utils.h"
#include "unipoly.h"
#include "vlist.h"
#include "max.h"

#ifdef COCOA_HOOK
#include "hook.h"
#endif

object math_error;
int funcall_level;
extern env panels_env;
extern map curr_al_map;
extern charstars curr_al_pkgs;

object mi_params(cterm t, ring R);
object mi_params1(cterm t, ring R);

extern object engine_in_ring(cterm t, ring R);
object eval_also_tagged(cterm t,ring R);

void mi_init(void)
{
  polynomials_init();
  gb_init();
  poincare_init(1);

}

/* ---[ math MAIN ]------------------------------------------ */

object eval(cterm t,ring R) /* evaluation of a right expression */
{
  static int recursion_level = 0;
  object res_o;
  int incr_depth = is_user_defined_coop(cterm_op(t));
  /* Maybe built-in functions should add depth 1 and user fns 8, say? */
  recursion_level += incr_depth; /* only count recursion level of user fns */
  if (recursion_level < 1024)    /* arbitrarily limit recursion level to 1024 */
    res_o=eval_also_tagged(t,R);
  else
    res_o = cocoa_error("TOO_DEEP");
  recursion_level -= incr_depth; /* restore previous recursion level */
  return res_o;
}

object eval_also_tagged(cterm t, ring R)
{
  coop op;
  object res_o;
  op=cterm_op(t);
  if (is_user_defined_coop(op)) res_o=mi_user_function(t,R);
  else if (op == op_NBS) res_o=mi_imm(t,R);
  else {
    op = op_index(op);
    if (ck_handlers[op]==NULL) res_o=cocoa_error("UNKN_OP");
    else res_o=(* ck_handlers[op])(t,R); }
  return object_clean(res_o);
}

object eval_caller(cterm t,ring R)
{
  object o;
  math_error=object_NULL;
  o=eval(t,R);
  if (math_error != object_NULL) {
    if (o==math_error) math_error=object_NULL;
  }
  if (object_is_error(o) && (error_context==cterm_NULL)) {
    error_set_context(t);
  }
  return o;
}

int eval_int(cterm t, ring R)
{
  object res_o;
  res_o=eval(t,R);
  if (object_is_error(res_o)) {
    math_error=res_o;
    return ZERO; }
  else return object_cast_int(res_o);
}

#if 0
coc_bool eval_bool_NEW(cterm t, ring R)
{
  object res_o;
  res_o=eval(t,R);
  if (object_is_error(res_o)) {
    math_error=res_o;
    return bool_NULL; }
  else return object_cast_bool(res_o);
}
#endif

#if 0
/* -----[ HANDLERS ]-------------------------------------------- */

object mi_adjoint(cterm t, ring R);
object mi_aliases(cterm t, ring R);
object mi_ascii(cterm t, ring R);
object mi_all(cterm t, ring R);
object mi_append(cterm t, ring R);
object mi_add_device(cterm t, ring R);
object mi_base_ring(cterm t, ring R);
object mi_bin_toric(cterm t, ring R);
object mi_bool(cterm t, ring R);
object mi_bm(cterm t, ring R);
object mi_call(cterm t, ring R);
object mi_cartesian(cterm t, ring R);
object mi_cast(cterm t, ring R);
object mi_change_sign(cterm t,ring R);
object mi_cip(cterm t,ring R);
object mi_characteristic(cterm t,ring R);
object mi_cocoa_packages(cterm t, ring R);
object mi_component(cterm t, ring R);
object mi_concat(cterm t, ring R);
object mi_cocolon(cterm t, ring R);
object mi_cond(cterm t, ring R);
object mi_curr_ring(cterm t, ring R);
object mi_debug_version(cterm t, ring R);
object mi_date(cterm t, ring R);
object mi_deg(cterm t,ring R);
object mi_dense_poly(cterm t, ring R);
object mi_der(cterm t, ring R);
object mi_det(cterm t, ring R);
object mi_div(cterm t,ring R);
object mi_eod(cterm t,ring R);
object mi_elim(cterm t,ring R);
object mi_env(cterm t, ring R);
object mi_error(cterm t, ring R);
object mi_fact(cterm t, ring R);
object mi_factor(cterm t, ring R);
object mi_frac(cterm t,ring R);
object mi_fields(cterm t,ring R);
object mi_function(cterm t,ring R);
object mi_fun_appl(cterm t,ring R);
object mi_gbm(cterm t, ring R);
object mi_isqrt(cterm t, ring R);
object mi_iroot(cterm t, ring R);
object mi_get(cterm t, ring R);
object mi_getenv(cterm t, ring R);
object mi_gb_all(cterm t, ring R);
object mi_gb_init(cterm t, ring R);
object mi_gb_one_move(cterm t, ring R);
object mi_gb_stat(cterm t, ring R);
object mi_gbasis(cterm t, ring R);
object mi_gens(cterm t, ring R);
object mi_head(cterm t, ring R);
object mi_help(cterm t, ring R);
object mi_hook(cterm t, ring R);
object mi_homog(cterm t,ring R);
object mi_imm(cterm t,ring R);
object mi_infos(cterm t,ring R);
object mi_init_cdev(cterm t, ring R);
object mi_indet(cterm t,ring R);
object mi_indet_ind(cterm t, ring R);
object mi_indet_index(cterm t, ring R);
object mi_indet_name(cterm t, ring R);
object mi_indet_nth(cterm t, ring R);
object mi_is_homog(cterm t, ring R);
object mi_ispurelt(cterm t, ring R);
object mi_is_ring_indep(cterm t, ring R);
object mi_is_ring_hom(cterm t, ring R);
object mi_is_zero(cterm t,ring R);
object mi_lin_sol(cterm t, ring R);
object mi_lin_ker(cterm t, ring R);
object mi_label(cterm t, ring R);
object mi_lcm(cterm t, ring R);
object mi_leading_coeff(cterm t, ring R);
object mi_leading_term(cterm t, ring R);
object mi_len(cterm t,ring R);
object mi_list(cterm t,ring R);
object mi_log(cterm t, ring R);
object mi_map(cterm t, ring R);
object mi_mat(cterm t, ring R);
object mi_mdeg(cterm t,ring R);
object mi_memory(cterm t,ring R);
object mi_minors(cterm t,ring R);
object mi_mod(cterm t,ring R);
object mi_module(cterm t,ring R);
object mi_mult(cterm t, ring R);
object mi_mvpoincare(cterm t, ring R);
object mi_new_list(cterm t, ring R);
object mi_new_mat(cterm t, ring R);
object mi_new_id(cterm t, ring R);
object mi_new_vector(cterm t, ring R);
object mi_NewPPStream(cterm t,ring R);
object mi_NextPP(cterm t,ring R);
object mi_AvoidLastPP(cterm t,ring R);
object mi_NextPrime(cterm t, ring R);
object mi_Mod2Rat(cterm t, ring R);
object mi_null(cterm t, ring R);
object mi_num_indets(cterm t,ring R);
object mi_objtype(cterm t, ring R);
object mi_olistcast(cterm t, ring R);
object mi_ord(cterm t, ring R);
object mi_packages(cterm t, ring R);
object mi_path(cterm t, ring R);
object mi_panels(cterm t, ring R);
object mi_parenthesis(cterm t, ring R);
object mi_poincare_quotient(cterm t,ring R);
object mi_poly(cterm t,ring R);
object mi_power(cterm t, ring R);
object mi_put(cterm t, ring R);
object mi_QZP(cterm t, ring R);
object mi_ZPQ(cterm t, ring R);
object mi_randomized(cterm t, ring R);
object mi_range(cterm t, ring R);
object mi_record(cterm t, ring R);
object mi_remove(cterm t, ring R);
object mi_remove_device(cterm t, ring R);
object mi_record_sel(cterm t, ring R);
object mi_res(cterm t, ring R);
object mi_ring(cterm t, ring R);
#ifdef RUM_DEBUG
object mi_rum_stat(cterm t, ring R);
#endif
object mi_seed(cterm t, ring R);
object mi_size(cterm t,ring R);
object mi_starting(cterm t,ring R);
object mi_string(cterm t, ring R);
object mi_helpquery(cterm t, ring R);
object mi_subst(cterm t, ring R);
object mi_subtract(cterm t, ring R);
object mi_sum(cterm t, ring R);
object mi_shmodule(cterm t, ring R);
object mi_syz(cterm t,ring R);
object mi_tag(cterm t, ring R);
object mi_tail(cterm t, ring R);
object mi_tao(cterm t, ring R);
object mi_time(cterm t, ring R);
object mi_typeof(cterm t, ring R);
object mi_unlabel(cterm t, ring R);
object mi_user_function(cterm t,ring R);
object mi_var(cterm t,ring R);
object mi_vector(cterm t, ring R);
object mi_zpmk(cterm t, ring R);

object mi_anna_BuiltinTestSet(cterm t, ring R);
object mi_anna_BuiltinTestSetLex(cterm t, ring R);
object mi_anna_BuiltinTestSetXel(cterm t, ring R);
object mi_anna_BuiltinNonNegSol(cterm t, ring R);

object mi_jaa_eval(cterm o, ring R);

object mi_support(cterm t, ring R);
object mi_monomials(cterm t, ring R);
object mi_coeffs(cterm t, ring R);

/* ----------------------------------------------------------------------- */

#endif

object mi_anna_BuiltinTestSet(cterm t, ring R)
{
  /* o list of polynomials */
  object res_o, o1, o2;
  olist ol;
  BitermList BL;
  ints Cost;
  int trunc;

  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  ol = object_cast_olist(o1);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    object_unlink(o2);
    return res_o;
  }
  trunc = object_cast_int(o2);
  if (math_error != object_NULL) {
    res_o=math_error;
    olist_free(ol);
    math_error=object_NULL;
    return res_o;
  }
  Cost = ring_2ordrow2Ints(R);
  StartTestSet(Cost, R->indetsNo);
  BL = olist2BList(ol, R);
  olist_free(ol);
  if (math_error != object_NULL) {
    res_o=math_error;
    ints_free(Cost);
    math_error=object_NULL;
    return res_o;
  }
  BL = TestSet(BL, trunc);
  ints_free(Cost);
  return olist2object(this_BList2olist(BL, NULL, R));
}

object mi_anna_BuiltinTestSetLex(cterm t, ring R)
{
  /* o list of polynomials */
  object res_o, o1, o2;
  olist ol;
  BitermList BL;
  int trunc;

  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  ol = object_cast_olist(o1);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    object_unlink(o2);
    return res_o;
  }
  trunc = object_cast_int(o2);
  if (math_error != object_NULL) {
    res_o=math_error;
    olist_free(ol);
    math_error=object_NULL;
    return res_o;
  }
  
  StartTestSetLex(R->indetsNo);
  BL = olist2BList(ol, R);
  olist_free(ol);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    return res_o;
  }
  BL = TestSet(BL, trunc);
  return olist2object(this_BList2olist(BL, NULL, R));
}

object mi_anna_BuiltinTestSetXel(cterm t, ring R)
{
  /* o list of polynomials */
  object res_o, o1, o2;
  olist ol;
  BitermList BL;
  int trunc;

  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  ol = object_cast_olist(o1);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    object_unlink(o2);
    return res_o;
  }
  trunc = object_cast_int(o2);
  if (math_error != object_NULL) {
    res_o=math_error;
    olist_free(ol);
    math_error=object_NULL;
    return res_o;
  }
  StartTestSetXel(R->indetsNo);
  BL = olist2BList(ol, R);
  olist_free(ol);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    return res_o;
  }
  BL = TestSet(BL, trunc);
  return olist2object(this_BList2olist(BL, NULL, R));
}

object mi_anna_BuiltinNonNegSol(cterm t, ring R)
{
  omat M;
  olist L;
  object o1, o2, res_o;
  
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  M = object_cast_omat(o1);
  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    object_unlink(o1);
    return res_o;
  }
  L = object_cast_olist(o2);
  if (math_error != object_NULL)
  {    
    res_o=math_error;
    math_error=object_NULL;
    omat_free(M);    
    return res_o;  
  }
  res_o = ip_solve(M,L);
  omat_free(M);
  olist_free(L);
  return res_o;
}

object mi_jaa_eval(cterm t, ring R)
{
  int i, nvars;
  olist point;
  object aux_o1,aux_o2,res_o;
  arpoly f;
  
  /**********************************************************************/
  /* this should do full type checking on args...                       */
  /* args should be a polynomial, and a list of rationals (or integers) */
  /* one for each indeterminate in the current ring.  Should also check */
  /* that the ring of the polynomial is the same as the current one.    */
  /**********************************************************************/
    
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  f = object_cast_arpoly(aux_o1, R); /* this destroys aux_o1 */
  if (math_error != object_NULL) return math_error;
  point = object_cast_olist(aux_o2);
  if (math_error != object_NULL) {
    arpoly_free(f,R);
    return math_error;
  }
  nvars = ring_indets_num(R);
  if (olist_len(point) != nvars) {
    olist_free(point); 
    return cocoa_error("BAD_PARAMS");
  }
  res_o=jaa_eval(f, point, R);
  arpoly_free(f,R);
  olist_free(point);
  return res_o;
}

object mi_cip(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  if (object_type(aux_o) != objtype_ideal && 
      object_type(aux_o) != objtype_module) {
    object_unlink(aux_o);
    return cocoa_error("BAD_PARAMS");
  }
  res_o=bool2object(module_wip(aux_o->v.module_value));
  object_unlink(aux_o);
  return res_o;
}

object mi_cocolon(cterm t, ring R)
{ /* COCOLON(RINGENV,EXPR) */
  ring S;
  char * id;
  id = eval_var_name(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  S = ring_env_get(id);
  if (S==ring_NULL) {
    error_set_this_param(string2cterm(id));
    free_str1(id);
    return cocoa_error("UNDEF_RING");
  }  
  free_str1(id);
  ring_unlink(S);
  return eval(cterm_param(t,1),S);
}

object mi_map(cterm t, ring R)
{
  /* MAP(EXPR, RING1, RING2, L:LIST(int)) */
  /* map EXPR from RING1 to RING2 using parameters L */
  object aux_o,aux_o1,aux_o2,res_o;
  char * aux_id;
  ring R1,R2;
  aux_o = eval(cterm_param(t,1),R);
  aux_id=object_cast_string(aux_o);
  if (math_error!=object_NULL) return math_error;
  R1 = ring_env_get(aux_id);
  free_str1(aux_id);
  ring_unlink(R1);
  if (math_error!=object_NULL) return math_error;

  aux_o = eval(cterm_param(t,2),R);
  aux_id=object_cast_string(aux_o);
  if (math_error!=object_NULL) return math_error;
  R2 = ring_env_get(aux_id);
  free_str1(aux_id);
  ring_unlink(R2);
  if (math_error!=object_NULL) return math_error;
  if (ring_characteristic(R1) != ring_characteristic(R2))
  {
    math_error = cocoa_error("CANNOT_CAST");
    return math_error;
  }
  aux_o1 = eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;

  if (object_type(aux_o1) != objtype_poly) {
    object_unlink(aux_o1);
    res_o=cocoa_error("BAD_PARAMS");
  } else if (object_ring(aux_o1) != R1) {
    res_o = object_link(aux_o1);
  } else {
    aux_o2 = eval(cterm_param(t,3),R);
    if (object_is_error(aux_o2)) {
      object_unlink(aux_o1);
      return aux_o2;
    }
    res_o = do_map(aux_o1, R2, aux_o2);
  }
  return res_o;
}

object mi_packages(cterm t,ring R)
{ /* t = PACKAGES() */
  int aux_n,i,j,id;
  char * aux_s;
  olist res_ol;
  coop aux_op;
  char * s;
  s = "";
  aux_n=map_num_starting(pkg_map,s);
  if (s[0]=='\0') res_ol=olist_init(aux_n-map_num_starting(pkg_map,"_"));
  else res_ol=olist_init(aux_n);
  id=1; j=1;
  for (i=1;i<=aux_n;i++) {
    id=map_next_starting(pkg_map,s,id);
    aux_s=pkg_name(id);
    if (s[0]!='\0' || aux_s[0] != '_')
      olist_set_nth(res_ol,j++,string2object(my_strdup(aux_s)));
  }
  return olist2object(res_ol);
}

object mi_call(cterm t,ring R)
{ /* t=Call(F,X1,...,Xn), Xi expressions */
  int pkg_n;
  coop op;
  object aux_o, res_o;
  cocfun f;
  olist aux_L;
  cterm handler,vars,body;
  cterm_list params;

  if (cterm_arity(t) < 1) return cocoa_error("BAD_PARAMS_NUM");
  aux_o = eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;

  params=cterm_params(t);
  params++;
  t=cterm_mk_classic(op_NULL, cterm_arity(t)-1, params);
  if (math_error != object_NULL) {
    free_cterm(t);
    object_unlink(aux_o);
    return math_error;
  } else if (object_type(aux_o) != objtype_cocfun) {
    free_cterm(t);
    object_unlink(aux_o);
    return cocoa_error("BAD_PARAMS");
  }
  f=object2cocfun(aux_o);
  if (strcmp(f->pkg_id, "$builtin") == 0)
    return cocoa_error("CALL_BUILTIN");
  vars=cocfun_header(f);
  body=cocfun_body(f);
  if (cterm_arity(t) == 2 && cterm_op(cterm_param(t,1)) == op_DOTS) 
    aux_L = olist_NULL;
  else aux_L=eval_params(t,vars,R);
  if (math_error != object_NULL){  /* memory leak? */
    free_cterm(t);
    return math_error;
  }
  res_o = mi_user_function_aux(vars,body,aux_L,t,R);
  free_cterm(t);
  return res_o;
}

object mi_append(cterm t, ring R)
{ /* APPEND(VAR L:LIST, EXPR) */
  object aux_o,res_o,aux_o1,aux_o2;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2; }
  if (object_rc(aux_o1)==2 && cterm_op(cterm_param(t,0)) == op_VAR)
    res_o=do_append(aux_o1,aux_o2);
  else {
    aux_o=object_dup(aux_o1);
    res_o=do_append(aux_o,aux_o2);
    engine_assign_this(cterm_param(t,0), aux_o,R);
  }
  object_unlink(aux_o1);
  object_unlink(aux_o2);
  return res_o;
}

object mi_add_device(cterm t, ring R)
{ /* ADD_DEVICE(VAR OMD:DEVICE, D:DEVICE) */
  /* OMD must be a multidevice */
  TODO("mi_add_device");
  return object_link(object_null);
}

object mi_base_ring(cterm t, ring R)
{
  object aux_o,res_o;
  ring res_R;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_R=object_ring(aux_o);
  if (res_R == ring_NULL) res_o=string2object(my_strdup(""));
  else res_o=string2object(my_strdup(ring_id(res_R)));
  object_unlink(aux_o);
  return res_o;
}

object mi_remove_device(cterm t, ring R)
{ /* REMOVE_DEVICE(VAR OMD:DEVICE, D:DEVICE) */
  /* OMD must be a multidevice */
  TODO("mi_remove_device");
  return object_link(object_null);
}

object mi_remove(cterm t, ring R)
{ /* REMOVE(VAR L:LIST,INT) */
  object aux_o,res_o,aux_o1, aux_o2;
  int aux_n;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  aux_n=object_cast_int(aux_o2);
  if (math_error != object_NULL) {
    object_unlink(aux_o1);
    return math_error; }
  if (object_rc(aux_o1)==2)
    res_o=do_remove(aux_o1,aux_n);
  else {
    aux_o=object_dup(aux_o1);
    res_o=do_remove(aux_o,aux_n);
    engine_assign_this(cterm_param(t,0), aux_o,R);
  }
  object_unlink(aux_o1);
  return res_o;
}

object mi_in_elim(cterm t,ring R)
{ /* InElim(P:POLY,N:INT) */
  /* TRUE iff LT(P) does not depend on the first N indets (useful for elim) */
  object aux_o1,aux_o2;
  monomial aux_m;
  arpoly aux_p;
  coc_bool aux_b;
  term aux_t;
  pvector aux_v;
  int aux_n,aux_e;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  aux_e=object_cast_int(aux_o2);
  if (math_error != object_NULL) {
    object_unlink(aux_o1);
    return math_error;
  }
  if (object_type(aux_o1) == objtype_pvector) {
    aux_v=aux_o1->v.pvector_value;
    pv_leading_monomial(aux_v,&aux_m,&aux_n,R);
    aux_b=term_first_nonzero_indet(monomial_term(aux_m),R)>aux_e;
    monomial_free(aux_m,R);
  } else {
    aux_p=object_cast_arpoly(aux_o1,R);
    if (math_error != object_NULL) return math_error;
    aux_t=arpoly_leading_term(aux_p,R);
    aux_b=term_first_nonzero_indet(aux_t,R)>aux_e;
    term_free(aux_t);
    arpoly_free(aux_p,R);
  }
  return bool2object(aux_b);
}

object mi_infos(cterm t, ring R)
{
  record res_r;
  res_r=record_init();
  record_update_field2(&res_r,"MaxExp",int2object(INT_MAX));
  record_update_field2(&res_r,"MaxChar",int2object(SHRT_MAX));
  record_update_field2(&res_r,"System",string2object(my_strdup(SIS_TYPE)));
  record_update_field2(&res_r,"MajorVersion",string2object(my_strdup(COCOA_MAJOR_VERSION())));
  record_update_field2(&res_r,"MinorVersion",string2object(my_strdup(COCOA_MINOR_VERSION())));
  record_update_field2(&res_r,"PolyVersion",string2object(my_strdup(POLY_VERSION)));
  record_update_field2(&res_r,"ShortPolyVersion",string2object(my_strdup(POLY_SHORT_VERSION)));
  record_update_field2(&res_r,"Version",string2object(my_strdup(COCOA_VERSION())));
  return record2object(res_r);
}

object mi_ascii(cterm t, ring R)
{ /* ASCII([N1,...,Nn]:INTEGER) ASCII(S:STRING) */
  object o1,res_o;
  int aux_n,i;
  olist aux_ol;
  char * aux_s;
  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  if (object_type(o1) == objtype_string) {
    aux_s = object2string(o1);
    aux_n = strlen(aux_s);
    aux_ol = olist_init(aux_n);
    for (i=0;i<aux_n;i++) {
      olist_set_nth(aux_ol,i+1,int2object((int)(unsigned char)aux_s[i]));
    }
    res_o = olist2object(aux_ol);
    free_str1(aux_s);
  } else {
    int charpos;
    aux_ol = object_cast_olist(o1);
    if (math_error != object_NULL) return math_error;
    aux_n = olist_len(aux_ol);
    aux_s = malloc_str(aux_n); /* might be too big but does not leak */
    charpos = 0;
    for (i=1;i<=aux_n;i++)
    {
      aux_s[charpos] = (unsigned char)object_cast_int(olist_nth(aux_ol,i));
      if (aux_s[charpos] != '\0') ++charpos;
    }
    aux_s[charpos]='\0';
    free_olist(aux_n,aux_ol);
    res_o = string2object(aux_s);
  }
  return res_o;
}

object mi_eod(cterm t, ring R)
{
  object o1;
  cdev aux_d;
  coc_bool aux_b;
  if (cterm_arity(t) != 1) return cocoa_error("BAD_PARAMS_NUM");
  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  aux_d = object_cast_cdev(o1);
  if (math_error != object_NULL) return math_error;
  aux_b=cdev_eof(aux_d);
  cdev_unlink(aux_d);
  return bool2object(aux_b);
}

object mi_cocoa_packages(cterm t, ring R)
{
  object res_o;
  char * aux_s;
  aux_s=ca_cocoa_packages();
  res_o=string2object(my_strdup(aux_s));
  return res_o;
}

object mi_getenv(cterm t, ring R)
{
  object aux_o;
  object res_o;
  char * aux_s, * aux_s1;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_s=object_cast_string(aux_o);
  if (math_error != object_NULL) return math_error;
  aux_s1=ca_getenv(aux_s);
  if (aux_s1 == NULL) aux_s1="";
  res_o=string2object(my_strdup(aux_s1));
  free_str1(aux_s);
  return res_o;
}

object mi_geterrmesg(cterm t, ring R)
{ /* GetErrMesg(E) */
  object aux_o;
  object res_o;

  aux_o = eval(cterm_param(t,0),R);
  if (object_is_error(aux_o))
  {
    /* Sometimes both aux_o and math_error are set, in which case they    */
    /* should be equal (i.e. pointers to the same identical object).      */
    /* I don't check they're equal: o/w what to do if they're different?  */
    if (math_error != object_NULL) math_error = object_NULL;
    res_o = string2object(my_strdup(aux_o->v.err_value));
    object_unlink(aux_o);
    return res_o;
  }
  object_unlink(aux_o);
  if (math_error != object_NULL)
  {
    if (aux_o == math_error) math_error = object_NULL;
    res_o = string2object(my_strdup(math_error->v.err_value));
    object_unlink(math_error);
    math_error = object_NULL;
    return res_o;
  }
  return string2object(my_strdup(""));
}

object mi_get(cterm t, ring R)
{ /* GET(D,N) */
  object o1, o2;
  int aux_n,i,j;
  olist aux_ol, aux_ol1;
  cdev aux_d;
  int aux_c;
  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  aux_d = object_cast_cdev(o1);
  if (math_error != object_NULL) return math_error;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    cdev_unlink(aux_d);
    return o2; }
  aux_n=object_cast_int(o2);
  if (math_error != object_NULL) {
    cdev_unlink(aux_d);
    return math_error; }
  aux_ol = olist_init(aux_n);
  for (i=1; i<=aux_n ; i++) {
    aux_c = cdev_getc(aux_d);
/*    if (cdev_eof(aux_d)) break;*/
    if (aux_c == EOF) break;
    o1=int2object((int)aux_c);
    olist_set_nth(aux_ol,i,o1);
  }
  if (cdev_eof(aux_d)) {
    aux_ol1 = olist_init(i-1);
    for (j=1;j<i;j++) olist_set_nth(aux_ol1,j,olist_nth(aux_ol,j));
    free_olist(aux_n,aux_ol);
    aux_ol=aux_ol1;
  }
  cdev_unlink(aux_d);
  return olist2object(aux_ol);
}

object mi_put(cterm t, ring R)
{ /* PUT(D,[ N1, ..., Nn ]) */
  object o1, o2;
  cdev aux_d;
  olist aux_ol;
  int i;
  unsigned char aux_c;
  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  aux_d = object_cast_cdev(o1);
  if (math_error != object_NULL) return math_error;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  aux_ol = object_cast_olist(o2);
  if (math_error != object_NULL) {
    cdev_unlink(aux_d);
    return math_error; }
  for (i = 1; i<= olist_len(aux_ol); i++) {
    aux_c = (unsigned char)object_cast_int(object_link(olist_nth(aux_ol,i)));
    cdev_putc(aux_c,aux_d);
  }
  olist_free(aux_ol);
  cdev_unlink(aux_d);
  return object_link(object_null);
}

object mi_init_cdev(cterm t, ring R)
{
  cdev res_D;
  int D_type, D_prot;
  char * D_name;
  object o1,o2,o3,o4,res_o;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  o3=eval(cterm_param(t,2),R);
  if (object_is_error(o3)) {
    object_unlink(o1);
    object_unlink(o2);
    return o3; }
  o4=eval(cterm_param(t,3),R);
  if (object_is_error(o4)) {
    object_unlink(o1);
    object_unlink(o2);
    object_unlink(o3);
    return o4; }
  D_type = object_cast_int(o1);
  if (math_error != object_NULL) {
    object_unlink(o2);
    object_unlink(o3);
    return math_error; }
  D_prot = object_cast_int(o2);
  if (math_error != object_NULL) {
    object_unlink(o3);
    return math_error; }
  D_name = object_cast_string(o3);
  if (math_error != object_NULL) return math_error;
  if (object_type(o4) == objtype_string) {
    res_D=cdev_open(D_type, D_prot, D_name, 
      (cdev_params)(o4->v.string_value));
  } else {
    res_D=cdev_open(D_type, D_prot, D_name,(cdev_params)NULL);
  }
  if (res_D == cdev_NULL)
    res_o=cocoa_error("CANNOT_OPEN");
  else
    res_o=cdev2object(res_D);
  object_unlink(o4);
  free_str1(D_name);
  return res_o;
}

object mi_time(cterm t, ring R)
{
  /* TIME(COMMAND) */
  return mi_time_aux(cterm_param(t,0),R);
}

object mi_aliases(cterm t, ring R)
{ /* t=ALIASES() */
  object res_o;
  int i,aux_n;
  olist aux_ol,res_ol;

  aux_n=map_last_id(curr_al_map);
  res_ol=olist_init(aux_n-1);

  for (i=2;i<=map_last_id(curr_al_map);i++) {
    aux_ol=olist_init(2);
    olist_set_nth(aux_ol,1,
      string2object(my_strdup(map_get_name(curr_al_map,i))));
    olist_set_nth(aux_ol,2,string2object(my_strdup(curr_al_pkgs[i])));
    olist_set_nth(res_ol,i-1,olist2object(aux_ol));
  }
  res_o=olist2object(res_ol);
  object_set_label(res_o,my_strdup("Aliases"));
  return res_o;
}

object mi_cond(cterm t, ring R)
{ 
  /* COND(B1,E1,B2,E2,...,Bn,En) */
  /* COND(B1,E1,B2,E2,...,Bn,En,E_else) */
  coc_bool aux_b;
  int arity,i;
  arity=cterm_arity(t);
  for (i=0;i<=arity/2-1;i++) {
    aux_b=eval_bool(cterm_param(t,i*2),R);
    if (math_error!=object_NULL) {
      object res_o;
      res_o=math_error;
      math_error=object_NULL;
      return res_o;
    }
    if (aux_b) return eval(cterm_param(t,i*2+1),R); }
  if (arity & 1) return eval(cterm_param(t,arity-1),R); /* else */
  return object_link(object_null); /* do nothing */
}

object mi_rum_stat(cterm t, ring R)
{
  int n;
  /* TODO: build a record with the statistics and print it */
  if (cterm_arity(t)==0) rum_show_statistics(0);
  else {
    n=eval_int(cterm_param(t,0),R);
    if (math_error != object_NULL) return math_error;
    rum_show_statistics(n);
  }
  return object_link(object_null);
}

object mi_imm(cterm t,ring R) /* t="0134304" string of digits (WITHOUT SIGN) */
{
  object res_o;
  res_o=object_init(objtype_arbint,NULL,R);
  LiInit(&res_o->v.arbint_value);
  LiStr2Coeff(&res_o->v.arbint_value,cterm_string(t),R);
  return res_o;
}

object mi_ring(cterm t, ring R)
{ /* RING(VAR) */
  object aux_o,res_o;
  char * aux_id;
  aux_id=eval_var_name(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  aux_o=env_get_value(ring_env,aux_id);
  if (aux_o==object_NULL) {
    error_set_this_param(cterm_mk_NBS(aux_id));
    res_o = cocoa_error("UNDEF_RING");
  } else {
    res_o=object_dup(aux_o);
    object_unlink(aux_o);
  }
  free_str1(aux_id);
  return res_o;
}

object mi_debug_version(cterm t, ring R)
{
#ifdef DEBUG
  return bool2object(TRUE);
#else
  return bool2object(FALSE);
#endif 
}

object mi_date(cterm t, ring R)
{
  object aux_o,res_o;
  char * fmt;
  if (cterm_arity(t) != 0) {
    aux_o=eval(cterm_param(t,0),R);
    if (object_type(aux_o) != objtype_string) {
      res_o=error_expected_found(objtype_string,object_type(aux_o));
      error_set_context(t);
      object_unlink(aux_o);
      return res_o; }
    fmt=object2string(aux_o);
  } else fmt=""; /* DEFAULT */
  return string2object(get_date(fmt));
}

object mi_ord_aux(ring R) 
{
  if (ring_ordering(R) != ordering_NULL) {
    return omat2object(ord2omat(ring_ordering(R),R));
  } else {
    object res_o;
    ordering to;
    to=ring_build_ordering(R);
    res_o=omat2object(ord2omat(to,R));
    ClearOrder(&to,R);
    return res_o;
  }
}

object mi_ord(cterm t, ring R)
{
  ring aux_R;
  if (cterm_arity(t)==0) return mi_ord_aux(R);
  aux_R=eval_ring(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  return mi_ord_aux(aux_R);
}

object mi_zpmk(cterm t, ring R)
{
  object res_o, o1, o2;
  cterm t1,t2;
  t1=cterm_param(t,0);
  t2=cterm_param(t,1);
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    error_set_context(t1);
    return o1; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    error_set_context(t2);
    return o2; }
  res_o=do_zpmk(o1,o2,R);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_log(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    error_set_context(t);
    return aux_o; }
  if (object_ring(aux_o) == ring_NULL) res_o=do_log(aux_o, R);
  else res_o=do_log(aux_o, object_ring(aux_o));
  object_unlink(aux_o);
  return res_o;
}

object mi_factor(cterm t, ring R)
{ /* FACTOR(POLY) */
  object res_o,aux_o;
 
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_o=object_cast(aux_o,objtype_poly,R);
  if (object_is_error(aux_o)) return aux_o;
  if (math_error != object_NULL) {
    object_unlink(aux_o);
    return math_error; }
  res_o=do_factor(aux_o);
  object_unlink(aux_o);
  return res_o;
}
 
object mi_gcd(cterm t, ring R)
{ /* GCD(POLY,POLY) */
  object res_o,aux_o, f, g;
 
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_o=object_cast(aux_o,objtype_poly,R);
  if (object_is_error(aux_o)) return aux_o;
  if (math_error != object_NULL)
  {
    object_unlink(aux_o);
    return math_error;
  }
  f = aux_o;
  aux_o=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_o=object_cast(aux_o,objtype_poly,R);
  if (object_is_error(aux_o)) return aux_o;
  if (math_error != object_NULL)
  {
    object_unlink(aux_o);
    return math_error;
  }
  g = aux_o;
  res_o = do_gcd(f, g);
  object_unlink(f);
  object_unlink(g);
  return res_o;
}
 

object mi_curr_ring(cterm t, ring R)
{
  return ring2object(ring_link(current_ring));
}

object mi_fact(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    error_set_context(t);
    return aux_o; }
  res_o=do_fact(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_cast(cterm t, ring R)
{
  object res_o, o1, o2;
  cterm t1, t2;
  t1=cterm_param(t,0);
  t2=cterm_param(t,1);
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    error_set_context(t1);
    return o1; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    error_set_context(t2);
    return o2; }
  if (object_type(o2) != objtype_objtype) {
    object_unlink(o1);
    error_set_context(t2);
    return error_expected_found(objtype_objtype,object_type(o2)); }
  res_o= object_cast(object_link(o1),o2->v.objtype_value,R);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_env(cterm t, ring R)
{ /* RING_ENV() */
  if (cterm_arity(t) == 0) return string2object(my_strdup(ring_id(R)));
  return mi_base_ring(t,R);
}

object mi_path(cterm t, ring R)
{ /* PKG_ID(STRING) */
  return string2object(pkg_path(cterm_string(cterm_param(t,0))));
}

object mi_panels(cterm t, ring R)
{
  olist res_ol;
  object res_o;
  res_ol=env2olist(panels_env);
  res_o=olist2object(res_ol);
  return res_o;
}

object mi_concat(cterm t, ring R)
{
  object res_o,aux_o,aux_o2;
  int i,arity;
  cterm_list params;
  arity=cterm_arity(t);
  if (arity == 0) return olist2object(olist_new(0,object_NULL));
  params=cterm_params(t);
  res_o=eval(params[0],R);

  for (i=1;i<arity;i++) {
    aux_o=eval(params[i],R);
    if (object_is_error(aux_o)) {
      object_unlink(res_o);
      return aux_o;
    }
    aux_o2=do_concat(res_o,aux_o);
    object_unlink(aux_o);
    object_unlink(res_o);
    res_o=aux_o2; }
  return res_o;
}

object mi_memory(cterm t,ring R)
{
  olist res_ol;
  object aux_o,res_o;
  char * id;
  ring aux_R;
  env e;
  if (cterm_op(t)==op_GLOBAL_MEMORY) e=cocoa_env;
  else if (cterm_arity(t)==0) e=current_env;
  else if (cterm_arity(t)>1) return cocoa_error("ERR.BAD_PARAMS_NUM");
  else {
    id=eval_var_name(cterm_param(t,0),R);
    if (math_error != object_NULL) {
      res_o=math_error;
      math_error=object_NULL;
      return res_o;
    }
    aux_R = ring_env_get(id);
    if (aux_R==ring_NULL) {
      error_set_this_param(string2cterm(id));
      free_str1(id);
      return cocoa_error("UNDEF_RING");
    }  
    e=ring_local_env(aux_R);
    ring_unlink(aux_R);
  }
  res_ol=env2olist_pairs(e);
  res_o=olist2object(res_ol);
  object_set_label(res_o,my_strdup("Memory"));
  return res_o;
}

object mi_tag(cterm t, ring R)
{
  object aux_o,res_o;
  char * aux_s;

  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_s=object_label(aux_o);
  if (aux_s == NULL) res_o=string2object(my_strdup(""));
  else res_o=string2object(my_strdup(aux_s));
  object_unlink(aux_o);
  return res_o;
}

object mi_lin_sol(cterm t, ring R)
{
  object aux_o1,aux_o2,res_o;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  res_o=do_lin_sol(aux_o1,aux_o2);
  object_unlink(aux_o1);
  object_unlink(aux_o2);
  return res_o;
}

object mi_lin_ker(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_lin_ker(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_label(cterm t, ring R)
{ /* Tagged(Expr) */
  /* Tagged(Expr, Tag) */
  /* Tagged(Expr, Tag, Package-name) */ /* TODO */
  object aux_o1,aux_o2,res_o;
  char * aux_s;
  cterm_list params;
  int arity;

  params=cterm_params(t);
  arity=cterm_arity(t);
  if (arity != 2) return cocoa_error("BAD_PARAMS_NUM");
  aux_o2=eval(params[1],R);
  if (object_is_error(aux_o2)) return aux_o2;
  aux_s=object2string(aux_o2);
  if (strlen(aux_s) == 0) { free_str1(aux_s); return cocoa_error("NULL_TAG"); }
  aux_o1=eval(params[0],R);
  if (object_is_error(aux_o1)) { free_str1(aux_s); return aux_o1; }
  if (object_type(aux_o1) == objtype_objtype) { free_str1(aux_s); object_unlink(aux_o1); return cocoa_error("CANNOT_TAG_TYPE"); }
  /* If the object already has the right label, just return it */
  if (object_label(aux_o1) != NULL &&
      strcmp(aux_s, object_label(aux_o1)) == 0)
  {
    free_str1(aux_s);
    return aux_o1;
  }
  res_o=object_dup(aux_o1);
  free_str1(object_label(res_o));
  object_set_label(res_o,aux_s);
  object_unlink(aux_o1);
  return res_o;
}

object mi_unlabel(cterm t, ring R)
{
  object aux_o,res_o;
  cterm_list params;
  if (cterm_arity(t) != 1) {
    error_set_context(t);
    return cocoa_error("BAD_PARAMS_NUM"); }
  params=cterm_params(t);
  aux_o=eval(params[0],R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=object_dup(aux_o);
  /* Set the tag/label to NULL unless the object is a type, e.g. TAGGED(...) */
  if (object_type(res_o) != objtype_objtype)
  {
    free_str1(object_label(res_o));
    object_set_label(res_o,NULL);
  }
  object_unlink(aux_o);
  return res_o;
}

object mi_range(cterm t, ring R)
{
  int i,j,n,from,to;
  short sign;
  object res_o,o1,o2;
  olist res_ol;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  if (object_type(o1) == objtype_poly && object_type(o2) == objtype_poly) {
    unknown u1,u2;
    if (object_ring(o1) != object_ring(o2))
    {
      object_unlink(o1);
      object_unlink(o2);
      return cocoa_error("OBJS_OF_DIFF_RINGS");
    }
    R = object_ring(o1); /* need ring later; this is last chance to grab it */
    u1 = object2indet(o1);
    u2 = object2indet(o2);
    if (u1 == unknown_NULL || u2 == unknown_NULL) 
    {
      object_unlink(o1);
      object_unlink(o2);
      return cocoa_error("EXPECTED_INDET");
    }
    if (u2 < u1) res_ol = olist_init(0);
    else {
      n=u2-u1+1;
      res_ol=olist_init(n);
      for (i=1;i<=n;i++) 
        olist_set_nth(res_ol,i,
          arpoly2object(arpoly_unknown2arpoly((unknown)(u1+i-1),R),R));
    }
    return olist2object(res_ol);
  }
  from=object_cast_int(o1);
  if (math_error!=object_NULL) {
    object_unlink(o2);
    return math_error;
  }
  to=object_cast_int(o2);
  if (math_error!=object_NULL) return math_error;
  if (from > to) res_ol=olist_init(0);
  else {
    n=to-from+1;
    res_ol=olist_init(n);
    for (i=1;i<=n;i++) olist_set_nth(res_ol,i,
      arbint2object(int2arbint(i+from-1))); 
  }
  return olist2object(res_ol);
}

object mi_subst(cterm t, ring R)
{
  unknown u;
  subst aux_s;   
  olist aux_ol,aux_ol1;
  int i,arity;
  object aux_o1,aux_o2,aux_o3,res_o;
  object aux_o;

  arity=cterm_arity(t);
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_s=subst_init(R);

    if (arity==2) {
    aux_o2=eval(cterm_param(t,1),R);
    if (object_is_error(aux_o2)) {
      object_unlink(aux_o1);
      return aux_o2; }
    if (object_type(aux_o2) != objtype_olist) {
      object_unlink(aux_o1);
      object_unlink(aux_o2);
      return cocoa_error("EXPECTED_LIST"); }
    aux_ol=object2olist(aux_o2);
    for (i=1;i<=olist_len(aux_ol);i++) {
      if (object_type(olist_nth(aux_ol,i)) != objtype_olist) {
        olist_free(aux_ol);
        object_unlink(aux_o1);
	error_set_param(cterm_param(t,1));
        return cocoa_error("NOT_A_SUBST"); }
      aux_ol1=object2olist(object_link(olist_nth(aux_ol,i)));
      if (olist_len(aux_ol1) != 2) {
        object_unlink(aux_o1);
        olist_free(aux_ol);
        olist_free(aux_ol1);
	error_set_param(cterm_param(t,1));
        return cocoa_error("NOT_A_SUBST"); }
      u=object2indet(object_link(olist_nth(aux_ol1,1)));
      if (math_error!=object_NULL) {
        olist_free(aux_ol);
        olist_free(aux_ol1);
        object_unlink(aux_o1);
        return math_error; }
      aux_o = object_link(olist_nth(aux_ol1,2)); /* 2 object_link: why? TODO */
      aux_o3=object_cast(object_link(aux_o),objtype_poly,R);
      object_unlink(aux_o);
      olist_free(aux_ol1);
      if (object_is_error(aux_o3)) {
        object_unlink(aux_o1);
        olist_free(aux_ol);
        return aux_o3;
      }
      subst_set(aux_s,u,object2arpoly(aux_o3,R)); }
    olist_free(aux_ol);
    }

  res_o=do_subst(aux_o1,aux_s);
  object_unlink(aux_o1);
  subst_free(aux_s,R);
  return res_o;
}

#if 0 /* DELETEME */

object mi_res(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_res(aux_o);
  return res_o;
}

#endif

object mi_minors(cterm t,ring R)
{ /* MINORS(MAT,INT) */
  int p;
  object aux_o,res_o;
  p=eval_int(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  aux_o=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_minors(aux_o,p);
  object_unlink(aux_o);
  return res_o;
}


object mi_homog(cterm t,ring R)
{ /* t=homog(x:indet,p1:expr,...,pn:expr) */
  object res_o,aux_o,aux_o2;
  olist aux_ol;
  int arity=cterm_arity(t);
  unknown aux_u;
  cterm_list params=cterm_params(t);
  register int i;

  if (arity<=1)
    return cocoa_error("BAD_PARAMS_NUM");

  aux_u=eval_indet(params[0],R);
  if (math_error!=object_NULL) return math_error;

  aux_o=eval(params[1],R);
  if (object_is_error(aux_o)) return aux_o;

  if (arity==2) {
    res_o=do_homog(aux_o,aux_u,R);
    object_unlink(aux_o);
    return res_o;
  }

  /* else (arity>2) */ 

  aux_ol=olist_init(arity);
  olist_set_nth(aux_ol,1,aux_o);
  for (i=2;i<arity;i++) {
    aux_o=eval(params[i],R);
    if (object_is_error(aux_o)) {
      olist_free(aux_ol);
      return aux_o; }
    aux_o2=do_homog(aux_o,aux_u,R);
    olist_set_nth(aux_ol,i,aux_o2); }
  return olist2object(aux_ol);
}

object mi_all(cterm t, ring R)
{ /* TODO: check errors (and free) */
  char * aux_v, * aux_v1;
  loc aux_l;
  coc_bool aux_b;
  int i, j, n;
  cterm t_var, t_list, t_cond, t_return;
  object aux_o,aux_o1,res_o;
  olist aux_ol, aux_ol1, res_ol;

  t_var=cterm_param(t,0);
  t_list=cterm_param(t,1);
  t_cond=cterm_param(t,2);
  t_return=cterm_param(t,3);
  aux_v=eval_var_name(t_var,R);
  if (math_error != object_NULL) return math_error;
  aux_v1=env_newvar(aux_v);
  cterm_subst_vars(&t_var,aux_v,aux_v1);
  cterm_subst_vars(&t_return,aux_v,aux_v1);
  cterm_subst_vars(&t_cond,aux_v,aux_v1);
  aux_o=eval(t_list,R);
  if (object_is_error(aux_o)) return aux_o; /* TODO: MEMORY free temps */
  aux_l=env_declare(&current_env,aux_v1);
  aux_ol=object_cast_olist(aux_o);
  if (math_error!=object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    return res_o; }
  n=olist_len(aux_ol);
  aux_ol1=olist_init(n);
  j=0;
  for (i=1;i<=n;i++) {
    mem_assign(aux_l,olist_nth(aux_ol,i));
    aux_b=eval_bool(t_cond,R);
    if (math_error!=object_NULL) {
      res_o=math_error;
      math_error=object_NULL;
      return res_o; }
    if (aux_b) { /* TODO: check memory !!!!! */
      aux_o1=eval(t_return,R);
      if (object_is_error(aux_o1)) {
        return aux_o1;
      }
      j++;
      olist_set_nth(aux_ol1,j,aux_o1); } } /* TODO: check memory */
  res_ol=olist_init(j);
  for (i=1;i<=j;i++) olist_set_nth(res_ol,i,olist_nth(aux_ol1,i));
  free_olist(n,aux_ol1);
  olist_free(aux_ol);
  env_delete_var(&current_env,aux_v1);
  cterm_subst_vars(&t_var,aux_v1,aux_v);
  cterm_subst_vars(&t_return,aux_v1,aux_v);
  cterm_subst_vars(&t_cond,aux_v1,aux_v);
  free_str(strlen(aux_v1),aux_v1);
  free_str(strlen(aux_v),aux_v);
  return olist2object(res_ol);
}

object mi_list(cterm t,ring R)
{ /* t=list(o1:expr,...,on:expr) */
  /* t=op(o1:expr,...,on:expr) */
  object res_o,aux_o;
  olist aux_ol;
  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  register int i;
  aux_ol=olist_init(arity);
  for (i=0;i<arity;i++) {
    aux_o=eval(params[i],R);
    if (object_is_error(aux_o)) { 
      olist_free(aux_ol);
      return aux_o; }
    olist_set_nth(aux_ol,i+1,aux_o); }
  return olist2object(aux_ol);
}

object mi_olistcast(cterm t, ring R)
{
  object aux_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  return object_cast(aux_o,objtype_olist,R);
}

object mi_mat_toric(cterm t,ring R)
{ /* MatToric(M) */
  /* where M is a non-negative integer matrix */
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  ToricAlg = EATI;
  res_o=do_mat_toric(aux_o,R);
  object_unlink(aux_o);
  return res_o;
}

object mi_mat_seqtoric(cterm t,ring R)
{ /* MatToric(M) */
  /* where M is a non-negative integer matrix */
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  ToricAlg = SATI;
  res_o=do_mat_toric(aux_o,R);
  object_unlink(aux_o);
  return res_o;
}

object mi_mat_partoric(cterm t,ring R)
{ /* MatToric(M) */
  /* where M is a non-negative integer matrix */
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  ToricAlg = PATI;
  res_o=do_mat_toric(aux_o,R);
  object_unlink(aux_o);
  return res_o;
}


object mi_mat(cterm t,ring R)
{ /* t=omat(m:list(list)) */
  object aux_o;
  if (cterm_arity(t)==0) return cocoa_error("BAD_PARAMS_NUM");
  aux_o=mi_params1(t,R);
  if (object_is_error(aux_o)) return aux_o;
  return object_cast(aux_o,objtype_omat,R);
}

object mi_der(cterm t,ring R)
{ /* t=der(p:poly_expr,x:indet) */
  /* t=der(p:poly_expr,x1:indet,n1:int,...,xm:indet,nm:int) */
  /* t=der(p:poly_expr,l:list({x:indet,n:int})) */

  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  unknown x;
  object aux_o1,aux_o2,res_o;
  int det_list[3]; /* TODO: levare la PORCATA */

  /* Really only the second of these two lines is right, but read on... */
  if (arity != 2) return cocoa_error("BAD_PARAMS_NUM");
  if (arity < 1) return cocoa_error("BAD_PARAMS_NUM");
  aux_o1=eval(params[0],R);
  if (object_is_error(aux_o1)) return aux_o1;

  if (arity==2) {

    if (cterm_op(params[1])!=op_LIST) {
/*
      x=mi_cterm2unknown(params[1],R);
      if (math_error!=object_NULL) {
        object_unlink(aux_o1);
        res_o=math_error;
        math_error=object_NULL;
        return res_o; }
*/
      x=eval_indet(params[1],R);
      if (math_error!=object_NULL) {
        object_unlink(aux_o1);
        return math_error; }
      det_list[0]=1;
      det_list[1]=x;
      det_list[2]=1; }
    else {
      aux_o2=eval(params[1],R);
      if (object_is_error(aux_o2)) {
        object_unlink(aux_o1);
        return aux_o2; }
      if (object_type(aux_o2)==objtype_olist) {
        TODO("derivata con lista");
        object_unlink(aux_o2);
        return object_link(object_null);
      }
      object_unlink(aux_o2);
      x=eval_indet(params[1],R);
      if (x==unknown_NULL) return cocoa_error("PLEASE_CONTACT");
      det_list[0]=1;
      det_list[1]=x;
      det_list[2]=1;
      if (math_error!=object_NULL) {
        object_unlink(aux_o1);
        return cocoa_error("BAD_PARAMS"); }
   }
  }
  else {
    TODO("derivative with many parameters");
    object_unlink(aux_o1);
    return cocoa_error("BAD_PARAMS_NUM"); }

  res_o = do_der(aux_o1,det_list);
  object_unlink(aux_o1);
  return res_o;
}

object mi_cartesian(cterm t, ring R)
{ /* t=cartesian(l1:list,...,ln:list) */
  object res_o,aux_o;
  olist aux_ol,res_ol;
  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  register int i;
  aux_ol=olist_init(arity);
  for (i=0;i<arity;i++) {
    aux_o=eval(params[i],R);
    if (object_is_error(aux_o)) {
      olist_free(aux_ol);
      return aux_o; }
    if (object_type(aux_o) != objtype_olist) {
                              /* for now >< only applied to olists */
      olist_free(aux_ol);
      return error_expected_found(objtype_olist,object_type(aux_o)); }
    /* if (object_in_memory(aux_o)) aux_o=object_dup(aux_o); */ /* TODO: ? */
    olist_set_nth(aux_ol,i+1,aux_o); 
  }
  res_ol=olist_cartesian(aux_ol);
  olist_free(aux_ol);
  if (math_error != object_NULL) return math_error;
  return olist2object(res_ol);
}
object do_record_sel(object o, char * id, ring R);

object mi_component(cterm t, ring R)
{
  /* t1:olist, t2:int -> object */
  /* t1:pvector, t2:int -> polynomial */
  /* t1:record, t2:string -> R.S */
  
  object o1,o2,res_o;
  record aux_r;
  int i,l;
  char * id;
  o1 = eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2 = eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2;
  }
  if (objtype_le(object_type(o2),objtype_string)) {
    id=object_cast_string(o2);
    res_o=do_record_sel(o1,id,R);
    object_unlink(o1);
    free_str1(id);
    return res_o;
  }
  i = object_cast_int(o2);
  if (math_error!=object_NULL) {
    object_unlink(o1);
    return math_error;
  }
  if (i<=0) { 
    object_unlink(o1); 
    error_set_param(cterm_param(t,1));
    return cocoa_error("INDEX_NEG"); }
  else {
    l=do_len(o1);
    if (math_error != object_NULL) {
      object_unlink(o1);
      return math_error; }
    if (i > l) {
      object_unlink(o1);
      error_set_param(cterm_param(t,1));
      return cocoa_error("INDEX_TOO_BIG"); } }
  res_o=do_component(o1,i);
  object_unlink(o1);
  return res_o;
}

object mi_new_list(cterm t, ring R)
{
  object aux_o,res_o;
  int aux_n,arity;
  arity=cterm_arity(t);
  if (arity<1 || arity>2) {
    error_set_context(t);
    return cocoa_error("BAD_PARAMS_NUM"); }
  aux_n=eval_int(cterm_param(t,0),R);
  if (math_error != object_NULL) return math_error;
  if (aux_n < 0) {
    error_set_this_param(string2cterm("expected non negative integer"));
    return cocoa_error("BAD_PARAMS");
  }
  if (cterm_arity(t)==2) {
    aux_o=eval(cterm_param(t,1),R);
    if (object_is_error(aux_o)) return aux_o; }
  else aux_o=object_dup(object_null);
  res_o=olist2object(olist_new(aux_n,aux_o));
  object_unlink(aux_o);
  return res_o;  
}

object mi_new_vector(cterm t, ring R)
{
  object aux_o,res_o;
  int aux_n,arity;
  pvcomp my_comp;
  arity=cterm_arity(t);
  if (arity<1 || arity>2) {
    error_set_context(t);
    return cocoa_error("BAD_PARAMS_NUM"); }
  aux_n=eval_int(cterm_param(t,0),R);
  if (math_error != object_NULL) return math_error;
  if (aux_n <= 0) {
    error_set_this_param(string2cterm("expected positive integer"));
    return cocoa_error("BAD_PARAMS");
  }
  if (cterm_arity(t)==2) {
    aux_o=eval(cterm_param(t,1),R);
    if (object_is_error(aux_o)) return aux_o;
    my_comp=object_cast_pvcomp(aux_o,R);
    if (math_error != object_NULL) return math_error; }
  else my_comp=NULL;
  res_o=pv2object(pv_fill(my_comp,aux_n,R),R);
  return res_o;  
}

object mi_new_id(cterm t, ring R)
{
  object o;
  char * prefix;
  if (cterm_arity(t)==0) 
    return string2object(env_newvar("V"));
  else if (cterm_arity(t)==1) {
    o=eval(cterm_param(t,0),R);
    if (object_is_error(o)) return o;
    prefix=object_cast_string(o);
    if (math_error != object_NULL) {
      o=math_error;
      math_error=object_NULL;
      return o;
    } else {
      o=string2object(env_newvar(prefix));
      free_str1(prefix);
      return o;
    }
  } else {
    return cocoa_error("BAD_PARAMS_NUM");
  } 
}

object mi_new_mat(cterm t, ring R)
{
  object aux_o,res_o;
  int aux_n1,aux_n2,arity;
  arity=cterm_arity(t);
  if (arity<=1 || arity>3) {
    error_set_context(t);
    return cocoa_error("BAD_PARAMS_NUM"); }
  aux_n1=eval_int(cterm_param(t,0),R);
  if (math_error != object_NULL) return math_error;
  if (aux_n1 <= 0) {
    error_set_this_param(string2cterm("expected positive integer"));
    return cocoa_error("BAD_PARAMS");
  }
  aux_n2=eval_int(cterm_param(t,1),R);
  if (math_error != object_NULL) return math_error;
  if (aux_n2 <= 0) {
    error_set_this_param(string2cterm("expected positive integer"));
    return cocoa_error("BAD_PARAMS");
  }
  if (arity==3) {
    aux_o=eval(cterm_param(t,2),R);
    if (object_is_error(aux_o)) return aux_o; }
  else aux_o=object_dup(object_null);
  res_o=omat2object(omat_new(aux_n1,aux_n2,aux_o));
  object_unlink(aux_o);
  return res_o;  
}

object mi_head(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_head(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_tail(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_tail(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_gens(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_gens(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_typeof(cterm t, ring R)
{
  object aux_o,res_o;
  t=cterm_param(t,0);
  aux_o=object_NULL;
  if (cterm_op(t)==op_VAR) {
    loc l;
    l = eval_left_expr_not_declare(t,R);
    if (l!=NULL) /* CHECK math_error */
      aux_o = mem_value(l);
    else {
      error_set_param(t);
      return cocoa_error("UNDEF_VAR"); } }
  else {
    aux_o=eval(t,R); }
  if (object_type(aux_o) == objtype_objtype)
    res_o=objtype2object(objtype_objtype,NULL);
  else
    res_o=objtype2object(object_type(aux_o),object_label(aux_o));
  object_unlink(aux_o);
  return res_o;
}

object mi_randomized(cterm t,ring R)
{ /* t=Randomized(E:expr) */
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_randomized(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_error(cterm t,ring R)
{
  /* ERROR/1(S:STRING) */
  /* ERROR/2(S:STRING,Param:EXPR) */
  /* ERROR/2(E:ERROR,Param:EXPR) */
  object res_o, aux_o; 
  char * aux_s;
  int arity;
  arity=cterm_arity(t);
  if (arity<1 || arity>2) return cocoa_error("BAD_PARAMS_NUM");
  aux_o=eval(cterm_param(t,0),R);
  if (object_type(aux_o) == objtype_err) {
    aux_s=object2error(aux_o);
  } else if (object_type(aux_o) != objtype_string) {
    res_o=error_expected_found(objtype_string,object_type(aux_o));
    error_set_param(cterm_param(t,0));
    object_unlink(aux_o);
    return res_o;
  } else {
    aux_s=object2string(aux_o);
  }
  if (arity==2) {
    aux_o=eval(cterm_param(t,1),R);
    if (object_is_error(aux_o)) {
      free_str1(aux_s);
      return aux_o; }
    error_set_this_param(object2cterm(aux_o));
    object_unlink(aux_o); }

  /* error_set_context(cterm_unknown_context); */

  return error2object(aux_s);
}

olist mi_starting_aux(int pkg, char * s)
{
  int aux_n,i,j;
  char * aux_s;
  olist res_ol;
  coop aux_op;
  aux_n=op_map_num_starting(pkg,s);
  if (s[0]=='\0') res_ol=olist_init(aux_n-op_map_num_starting(pkg,"\040")-op_map_num_starting(pkg,"_"));
  else res_ol=olist_init(aux_n);
  aux_op=1; j=1;
  for (i=1;i<=aux_n;i++) {
    aux_op=op_map_next_starting(pkg,s,aux_op);
    aux_s=op_map_op_name(pkg,aux_op);
    if (s[0]!='\0' || (aux_s[0] != '\040' && aux_s[0] != '_'))
      olist_set_nth(res_ol,j++,string2object(my_strdup(aux_s)));
  }
  return res_ol;
}

object mi_starting(cterm t,ring R)
{ /* t=STARTING(S:STRING) */ 
  /* t=STARTING(S:STRING,PKG_NAME:STRING) */
  object o1, o2,res_o;
  int pkg_n;
  char * aux_s1, * aux_s2;
  olist aux_ol1,aux_ol2,res_ol;
  if (cterm_arity(t) == 0) return cocoa_error("BAD_PARAMS_NUM");
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  aux_s1=object_cast_string(o1);
  if (math_error != object_NULL) return math_error;
  if (cterm_arity(t) == 1) {
    aux_ol1=mi_starting_aux(pkg_builtin_n,aux_s1);
    aux_ol2=mi_starting_aux(pkg_std_n,aux_s1);
    res_ol=olist_concat_these(aux_ol1,aux_ol2);
    aux_ol2=mi_starting_aux(pkg_user_n,aux_s1);
    res_ol=olist_concat_these(res_ol,aux_ol2);
  } else {
    o2=eval(cterm_param(t,1),R);
    if (object_is_error(o2)) {
      free_str1(aux_s1);
      return o2;
    }
    aux_s2=object_cast_string(o2);
    if (math_error != object_NULL) {
      free_str1(aux_s1);
      return math_error; }
    pkg_n=pkg_index(aux_s2);
    /* WHY THE BLOODY HELL COULDN'T ANTONIO BE SENSIBLE???????????? */
    /* I'D MAKE THE FOLLOWING A FUNCTION IF ONLY CAP HADN'T SCREWED */
    /* ME ROYALLY WITH SUCH A GODAWFUL MESS.  HIGHLY UNAMUSED!!!!   */
    if (pkg_n == name_id_NULL)
    {
      object aux_o;
      cterm aux_t;
      cterm aux_params2[1];
      aux_params2[0]=string2cterm(aux_s2);
      aux_t=cterm_mk_classic(op_SOURCE,1,aux_params2);
      aux_o=engine_source(aux_t,R);
      cterm_free(aux_t);
      if (object_is_error(aux_o)) {
        object_unlink(aux_o);
        error_set_this_param(cterm_mk_NBS(aux_s2));
        math_error=cocoa_error("UNKN_PKG");
        return math_error; 
      } else {
        error_set_context(cterm_NULL); /* clear error context */
        object_unlink(aux_o);
        pkg_n=pkg_index(aux_s2);
      }
    }
    if (pkg_n != name_id_NULL)
      res_ol=mi_starting_aux(pkg_n,aux_s1);
    else {
      error_set_this_param(cterm_mk_NBS(aux_s2));
      return cocoa_error("UNKN_PKG"); 
    } 
    free_str1(aux_s2);
  }
  free_str1(aux_s1);
  return olist2object(res_ol);
}

object mi_string(cterm t, ring R)
{ /* STRING(NBS) */
  return string2object(my_strdup(cterm_string(cterm_param(t,0))));
}

object mi_helpquery(cterm t, ring R)
{ /* ?string and ??string */
  cterm help_param[2];
  cterm help_call[3];
  cterm help_call_as_a_term;
  object result;
  char *helpline = cterm_string(cterm_param(t,0));
  int iskeyword = 0;
  int nargs = 2;

  // If the line after '?' is empty, call Man with no args...
  {
    int first_non_space = 0;
    while (helpline[first_non_space] == ' ') ++first_non_space;
    if (helpline[first_non_space] == '\0') nargs = 0;
  }

  /* Strip any trailing semicolons if nargs > 0 */
  if (nargs > 0)
  {
    int last_non_space = 0;
    while (helpline[last_non_space] != '\0') ++last_non_space; /* move to end */
    --last_non_space;
    /* Loop below scans backward for a char difft from ' ' or ';' */
    /* It replaces any semicolons by spaces on the way.           */
    while (last_non_space >= 0 && (helpline[last_non_space] == ';' || helpline[last_non_space] == ' '))
    {
      helpline[last_non_space] = ' ';
      --last_non_space;
    }
    if (last_non_space < 0) nargs = 0; /* Help line was effectively empty */
  }

  /* Line below assumes that the high level Man command strips leading spaces */
  if (helpline[0] == '?') { iskeyword = 1; helpline[0] = ' '; }

  if (nargs > 0)
  {
    help_param[0] = string2cterm(helpline); /* this is the string to handed to the help call */
    help_param[1] = int2cterm(1-iskeyword);
  }
  help_call[0] = cterm_mk_NBS("$help");
  help_call[1] = cterm_mk_NBS("Man");
  help_call[2] = cterm_mk_classic(op_LIST, nargs, help_param);
  help_call_as_a_term = cterm_mk_classic(op_FUN_APPL, 3, help_call);
  result = eval(help_call_as_a_term, R);
  cterm_free(help_call_as_a_term);
  return object_link(result);
}

object mi_var(cterm t, ring R)
{ loc l;
  object res_o;
  object aux_o;
  l = eval_left_expr_not_declare(t,R);
  if (math_error!=object_NULL) return math_error;
  if (l == NULL) {
    error_set_param(t);
    return cocoa_error("UNDEF_VAR");
  }
  res_o = mem_value(l);
  return res_o;
}

object mi_sum(cterm t, ring R)
{
  object o1,o2,res_o;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  res_o=do_sum(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_mult(cterm t, ring R)
{
  object o1,o2,res_o;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  res_o=do_mult(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_is_ring_indep(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=bool2object(object_is_ring_indep(aux_o));
  object_unlink(aux_o);
  return res_o;
}

object mi_is_zero(cterm t,ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=bool2object(object_is_zero(aux_o));
  object_unlink(aux_o);
  return res_o;
}

object mi_is_homog(cterm t,ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_is_homog(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_div(cterm t,ring R)
{
  object o1,o2,res_o;
  coc_bool aux_b;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  if (object_type(o2) != objtype_ideal &&
      object_type(o2) != objtype_module &&
      object_type(o2) != objtype_olist) {
    aux_b=object_is_zero(o2);
    if (math_error!=object_NULL) {
      object_unlink(o1);
      object_unlink(o2);
      return math_error; }
    if (aux_b) {
      object_unlink(o1);
      object_unlink(o2);
      error_set_context(t);
      return cocoa_error("DIV_BY_ZERO"); } }
  res_o=do_div(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_mod(cterm t,ring R)
{
  object o1,o2,res_o;
  coc_bool aux_b;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  if (object_type(o2) != objtype_ideal &&
      object_type(o2) != objtype_module &&
      object_type(o2) != objtype_olist) {
    aux_b=object_is_zero(o2);
    if (math_error!=object_NULL) {
      object_unlink(o1);
      object_unlink(o2);
      return math_error; }
    if (aux_b) {
      object_unlink(o1);
      object_unlink(o2);
      error_set_context(t);
      return cocoa_error("DIV_BY_ZERO"); } }
  res_o=do_mod(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_fields(cterm t, ring R)
{ /* FIELDS(RECORD) */
  object aux_o;
  olist aux_ol;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  if (object_type(aux_o) != objtype_record)
  {
    object_unlink(aux_o);
    return cocoa_error("FIELDS_NOT_RECORD");
  }
  aux_ol=record_fields(aux_o->v.record_value);
  object_unlink(aux_o);
  return olist2object(aux_ol);
}

object mi_function(cterm t, ring R)
{ /* FUNCTION("", "FUN_ID") */
  /* FUNCTION("PKG_ID","FUN_ID") */
  /* FUNCTION("FUN_ID") */
  /* FUNCTION("PKG_ID.FUN_ID") */
  char * pkg_id, * op_id, * aux_s;
  cocfun aux_f;
  object o1,o2;
  coop aux_op;
  int pkg_n;

  if (cterm_arity(t) == 0 || cterm_arity(t) > 2) return cocoa_error("BAD_PARAMS_NUM");
  if (cterm_arity(t) == 1) {
    o1 = eval(cterm_param(t,0),R);
    aux_s=object_cast_string(o1);
    if (math_error != object_NULL) return math_error;
    str_split(aux_s,".",&pkg_id,&op_id);
/***** THE MEMORY IN AUX_S IS LEAKED; HARD TO FIX ******/
  } else {
    o1 = eval(cterm_param(t,0),R);
    if (object_is_error(o1)) return o1;
    pkg_id=object_cast_string(o1);
    if (math_error != object_NULL) return math_error;
    o2 = eval(cterm_param(t,1),R);
    if (object_is_error(o2)) {
      object_unlink(o1);
      return o2; }
    op_id=object_cast_string(o2);
    if (math_error != object_NULL) {
      free_str1(pkg_id);
      return math_error; }
  }
  if (pkg_id == NULL || strlen(pkg_id)==0) {
    coop_pkg_std_op(op_id,&pkg_n,&aux_op);
    if (pkg_n == pkg_n_NULL) {
      error_set_this_param(string2cterm(op_id));
      free_str1(op_id);
      if (pkg_id != NULL) free_str1(pkg_id);
      return cocoa_error("UNKN_OP"); 
    }
    pkg_id=pkg_name(pkg_n);
  }
  if (pkg_id[0] != '$') {
    error_set_this_param(string2cterm(pkg_id));
    return cocoa_error("BAD_PKG_NAME");
  }
/* this section below I ripped off from mi_expand_fun_appl */
/* MA CHE MACELLO! */
{
  int pkg_n = pkg_index(pkg_id);
  if (pkg_n == name_id_NULL)
  {
    object aux_o;
    cterm aux_t;
    cterm aux_params2[1];
    aux_params2[0]=string2cterm(pkg_id);
    aux_t=cterm_mk_classic(op_SOURCE,1,aux_params2);
    aux_o=engine_source(aux_t,R);
    cterm_free(aux_t);
    if (object_is_error(aux_o)) {
      object_unlink(aux_o);
      error_set_this_param(cterm_mk_NBS(pkg_id));
      math_error=cocoa_error("UNKN_PKG");
      return math_error; 
    } else {
      error_set_context(cterm_NULL); /* Clear error context */
      object_unlink(aux_o);
    }
  }
}
  aux_f=op_map_cocfun(pkg_id,op_id);
  if (aux_f == cocfun_NULL) {
    char * aux_s;
    aux_s=malloc_str(strlen(op_id)+strlen(pkg_id)+1);
    sprintf(aux_s,"%s.%s",pkg_id,op_id);
    error_set_this_param(string2cterm(aux_s));
    free_str1(aux_s);
    return cocoa_error("UNKN_OP"); 
  }
  return cocfun2object(aux_f);
}

object mi_fun_appl(cterm t,ring R)
{ /* FUN_APPL/2(ID,LIST(E1,...,En)) */
  /* FUN_APPL/3(PKG,ID,LIST(E1,...,En)) */
  /* FUN_APPL/2(ID,DOTS) */
  /* FUN_APPL/3(PKG,ID,DOTS) */
  cterm aux_t;
  aux_t = mi_expand_fun_appl(t,R);
  if (math_error != object_NULL)
  {
    object error_ptr = math_error;
    math_error = object_NULL;
    return error_ptr;
  }
  return eval(aux_t,R);
}

cterm mi_expand_fun_appl(cterm t,ring R)
{ /* FUN_APPL/2(ID,LIST(E1,...,En)) */
  /* FUN_APPL/3(PKG,ID,LIST(E1,...,En)) */
  /* FUN_APPL/2(ID,DOTS) */
  /* FUN_APPL/3(PKG,ID,DOTS) */
  char * pkg_id, * op_id;
  char buffer[200];
  cterm t1,args;
  cterm * aux_params;
  int pkg_n, oparity;
  int aux_arity;
  coop aux_op;
  if (cterm_arity(t)==2) {
    pkg_id="";
    op_id=cterm_string(cterm_param(t,0));
    args=cterm_param(t,1);
    coop_pkg_std_op(op_id,&pkg_n,&aux_op);
  } else { /* arity == 3 */
    pkg_id=cterm_string(cterm_param(t,0));
    op_id=cterm_string(cterm_param(t,1));
    pkg_n=pkg_index(pkg_id);
    if (pkg_n != name_id_NULL) {
      args=cterm_param(t,2);
      aux_op=op_map_coop(pkg_n,op_id);
    } else {
      object aux_o;
      cterm aux_t;
      cterm aux_params2[1];
      aux_params2[0]=string2cterm(pkg_id);
      aux_t=cterm_mk_classic(op_SOURCE,1,aux_params2);
      aux_o=engine_source(aux_t,R);
      cterm_free(aux_t);
      if (object_is_error(aux_o)) {
        object_unlink(aux_o);
        error_set_this_param(cterm_mk_NBS(pkg_id));
	math_error=cocoa_error("UNKN_PKG");
        return cterm_NULL; 
      } else {
        error_set_context(cterm_NULL); /* clear error context */
        object_unlink(aux_o);
        pkg_n=pkg_index(pkg_id);
        args=cterm_param(t,2);
        aux_op=op_map_coop(pkg_n,op_id);
      }
    }
  }
  if (aux_op==op_NULL) {
    if (cterm_arity(t)==2) {
      error_set_this_param(cterm_mk_NBS(op_id));
    } else {
      char * aux_s;
      aux_s=malloc_str(strlen(op_id)+strlen(pkg_id)+1);
      sprintf(aux_s,"%s.%s",pkg_id,op_id);
      error_set_this_param(cterm_mk_NBS(aux_s));
      free_str1(aux_s);
    }
    math_error = cocoa_error("UNKN_OP");
    error_set_context(t);
    return cterm_NULL;
  }
#ifdef DEBUG_RESOLVING
  sprintf(buffer,"--> Resolving %s.%s\n",pkg_id,op_id);
  cocoa_debug_message(buffer);
#endif

  oparity=op_map_op_arity(pkg_n,op_index(aux_op));
  if ((oparity!=NO_ARITY) && (oparity!=cterm_arity(args))) {
    math_error = cocoa_error("BAD_PARAMS_NUM");
    error_set_context(t);
    return cterm_NULL;
  }

  aux_op=op_mk(pkg_n,aux_op);
  cterm_free(cterm_param(t,0));
  if (cterm_arity(t)==3) cterm_free(cterm_param(t,1));
  cterm_set_op(t,aux_op);
  aux_params=cterm_params(t);
  aux_arity=cterm_arity(t);
  if (cterm_op(args) == op_DOTS) {
    cterm params[1];
    params[0]=args;
    cterm_set_params(t,params);
    args=cterm_mk_classic(op_LIST,1,params);
  }
  cterm_set_params(t,cterm_params(args));
  cterm_set_arity(t,cterm_arity(args));
  cterm_set_params(args,NULL);
  cterm_set_arity(args,0);
  cterm_free(args);
  free_cterm_list(aux_arity,aux_params);
  return t;
}

object mi_frac(cterm t,ring R)
{
  object o1,o2,res_o;
  coc_bool aux_b;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  if (object_type(o2) != objtype_ideal &&
      object_type(o2) != objtype_module &&
      object_type(o2) != objtype_olist) {
    aux_b=object_is_zero(o2);
    if (math_error!=object_NULL) {
      object_unlink(o1);
      object_unlink(o2);
      return math_error; }
    if (aux_b) {
      object_unlink(o1);
      object_unlink(o2);
      error_set_context(t);
      return cocoa_error("DIV_BY_ZERO"); } }
  res_o=do_frac(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_poly(cterm t,ring R)
{ 
  object aux_o;
  aux_o=mi_params1(t,R);
  if (object_is_error(aux_o)) return aux_o;
  return object_cast(aux_o,objtype_poly,R);
}

object mi_module(cterm t,ring R)
{ /* MODULE(SHIFTS,OBJECT)*/
  /* MODULE(OBJECT) */
  /* OBJECT castable to MODULE */
  /* GENS = LIST(LIST(POLY)) */
  /* SHIFTS = LIST(INT) */
  /* SHIFTS = LIST(POLY) */
  object aux_o, res_o;
  shifts ss, old_ss;
  vlist aux_vl;
  module aux_M;
  olist aux_L;

  if (cterm_arity(t) == 1) {
    aux_o=eval(cterm_param(t,0),R);
    return object_cast(aux_o,objtype_module,R);
  } else if (cterm_arity(t) == 2) {
    aux_o=eval(cterm_param(t,1),R);
    aux_M=object_cast_module(aux_o,R);
    if (math_error != object_NULL) return math_error;
    aux_o=eval(cterm_param(t,0),R);
    if (strcmp(object_label(aux_o),"shifts") != 0)
      return cocoa_error("BAD_PARAMS");
    return cocoa_error("NO_SHIFTS");
    aux_L = object_cast_olist(aux_o);
    if (math_error != object_NULL) {
      module_unlink(aux_M,R);
      return math_error;
    }
    ss = this_olist2shifts(aux_L,R);
    if (math_error != object_NULL) return math_error;
    if (shifts_rank(ss) != module_rank(aux_M)) {
      return cocoa_error("BAD_SHIFTS");
    }
    vlist_apply_shifts(module_gens(aux_M),ss,R);
    module_set_shifts(aux_M,ss);
    return module2object(aux_M,R);
  } else {
    return cocoa_error("BAD_PARAMS_NUM");
  }
}

object mi_indet(cterm t,ring R)
{ /* t = indet("x","i1",...,"in") */
  /* t = indet("x") */
  arpoly aux_p;
  object res_o;
  olist aux_ol;
  char * aux_s;
  unknown u;
  u = mi_cterm2unknown_bis(t,R);
  if (math_error!=object_NULL) return math_error;
  if (u == unknown_NULL) {
    if (cterm_arity(t) != 1) {
      res_o=cocoa_error("UNDEF_INDET");
      error_set_param(t);
      return res_o;
    }
    aux_s=cterm_string(cterm_param(t,0));
    aux_ol=indet_ident2olist(aux_s,R);
    if (aux_ol == olist_NULL) {
      math_error=cocoa_error("UNDEF_INDET");
      error_set_param(t);
      return math_error;
    } else {
      return olist2object(aux_ol);
    }
  }
  aux_p=arpoly_unknown2arpoly(u,R);
  return arpoly2object(aux_p,R);
}

/* The following global is defined in ck.c */
extern gmp_randstate_t GMP_RANDSTATE;

object mi_seed(cterm t, ring R)
{ /* Seed(INT) */
  object aux_o;
  int aux_n;
  aux_n=eval_int(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  gmp_randseed_ui(GMP_RANDSTATE, aux_n);
  return object_link(object_null);
}

object mi_indet_nth(cterm t, ring R)
{
  unknown aux_u;
  object res_o;
  arpoly res_p;
  aux_u=eval_indet_nth(t,R);
  if (math_error!=object_NULL) return math_error;
  res_p=arpoly_unknown2arpoly(aux_u,R);
  return arpoly2object(res_p,R);
}

object mi_indet_name(cterm t, ring R)
{
  ivector aux_v;
  unknown x;
  x=eval_indet(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  return string2object(my_strdup(indet_name(R->indets,x)));
}

object mi_indet_ind(cterm t, ring R)
{
  ivector aux_v;
  unknown x;
  x=eval_indet(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  aux_v=indet_indices(R->indets,x);
  return olist2object(this_iv2olist(aux_v));
}

object mi_indet_index(cterm t, ring R)
{
  unknown x;
  x=eval_indet(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  return int2object(x+1);
}

object mi_gb_init_aux(object o, ring R)
{
  module aux_M;
  olist aux_ol;
  if (object_type(o) != objtype_module && object_type(o) != objtype_ideal) 
    return error_expected_found(objtype_module,object_type(o));
  aux_M = o->v.module_value;
  gb_Do_Init(aux_M);
  return object_link(object_null);
}

object mi_gb_init(cterm t, ring R)
{
  object aux_o,aux_o1,res_o;
  aux_o1=eval(cterm_param(t,0),R);
  res_o=mi_gb_init_aux(aux_o1, R);
  object_unlink(aux_o1);
  return res_o;
}


object mi_gb_all_aux(object o, ring R)
{
  module aux_M;
  int aux_n;
  olist aux_ol;
  if (object_type(o) != objtype_module && object_type(o) != objtype_ideal) 
    return error_expected_found(objtype_module,object_type(o));
  aux_M=o->v.module_value;
  aux_n=gb_Do_All(aux_M);
  return int2object(aux_n);
}

object mi_gb_all(cterm t, ring R)
{
  object aux_o,aux_o1,res_o;
  aux_o1=eval(cterm_param(t,0),R);
  res_o=mi_gb_all_aux(aux_o1, R);
  object_unlink(aux_o1);
  return res_o;
}

object mi_gb_stat(cterm t, ring R)
{
  object aux_o,res_o;
  record res_r;
  module aux_M;
  int d;
  aux_o=eval(cterm_param(t,0),R);
  if (object_type(aux_o) != objtype_module && 
      object_type(aux_o) != objtype_ideal) {
    res_o=error_expected_found(objtype_module,object_type(aux_o));
    object_unlink(aux_o);
    return res_o; }
  aux_M=aux_o->v.module_value;
  res_r=record_init();
  record_update_field(&res_r,"Gens",
    int2object(vlist_len_ck(module_gens_ck(aux_M))));
  record_update_field(&res_r,"MinGens",
    int2object(vlist_len_ck(module_mingens(aux_M))));
  record_update_field(&res_r,"GB",
    int2object(vlist_len_ck(module_gbasis(aux_M))));
  record_update_field(&res_r,"Rules",
    int2object(vlist_len_ck(gb_Rules(aux_M))));
  record_update_field(&res_r,"IVs",
    int2object(vlist_len_ck(gb_IVs(aux_M))));
  record_update_field(&res_r,"IPs",
    int2object(vpairs_len(gb_IPs(aux_M))));
  d = gb_Min_Deg_Incomings_NoTruncCheck(aux_M);
  if (d != deg_NULL) 
    record_update_field(&res_r,"Deg", int2object(d));
  else 
    record_update_field(&res_r,"Deg", int2object(-1));
  if (module_discrepancy(aux_M) != -1) 
    record_update_field(&res_r,"Discr", 
      int2object(module_discrepancy(aux_M)));
  object_unlink(aux_o);
  return record2object(res_r);
}

object mi_gb_one_move_aux(object o, ring R)
{
  module aux_M;
  coc_bool aux_b;
  olist aux_ol;
  if (object_type(o) != objtype_module && 
      object_type(o) != objtype_ideal) {
    return error_expected_found(objtype_module,object_type(o));
  }
  aux_M=o->v.module_value;
  aux_b=gb_one_move(aux_M);
  return bool2object(aux_b);
}

object mi_gb_one_move(cterm t, ring R)
{
  object aux_o,aux_o1,res_o;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  res_o=mi_gb_one_move_aux(aux_o1, R);
  object_unlink(aux_o1);
  return res_o;
}

object mi_is_ring_hom(cterm t, ring R)
{
  object aux_o;
  coc_bool aux_b;
  if (cterm_arity(t) == 0) return cocoa_error("BAD_PARAMS_NUM");
  aux_o=eval(cterm_param(t,0),R);
  aux_b=object_is_ring_homogeneous(aux_o);
  object_unlink(aux_o);
  return bool2object(aux_b);
}

object mi_vector(cterm t, ring R)
{
  object aux_o;
  ring aux_R;
  aux_o=mi_params1(t,R);
  if (object_is_error(aux_o)) return aux_o;
  if (!object_extract_ring_ck(aux_o, &aux_R)) {
    error_set_context(t);
    return cocoa_error("BAD_PARAMS");
  }
  if (aux_R == ring_NULL) aux_R=R;
  if (!is_polynomial_ring(aux_R)) {
    error_set_context(t);
    return cocoa_error("ONLY_POLY_VECTORS"); }
  return object_cast(aux_o,objtype_pvector,aux_R);
}

object mi_poincare_quotient(cterm t,ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_poincare_quotient(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_mvpoincare(cterm t, ring R)
{ /* TODO: riattivare? */
  ring MVP;
  cterm_list params;
  object res_o,o1,o2;
  MVP = ring_env_get(MVP_RING_NAME);
  ring_unlink(MVP);
  if (MVP==ring_NULL) {
    error_set_this_param(cterm_mk_NBS("MVP"));
    return cocoa_error("UNDEF_RING"); }
  o1 = eval(cterm_param(t,0),R); /* ideal */
  if (object_is_error(o1)) return o1;
  o2 = eval(cterm_param(t,1),R); /* weights */
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  res_o=do_mvpoincare_quotient(o1,o2,MVP);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_record(cterm t, ring R)
{
  /* t=RECORD(B1,...,Bn), Bi=BIND(V,E) */
  int arity,i;
  record res_r;
  char * id;
  object aux_o;
  cterm_list params;
  arity=cterm_arity(t);
  res_r=record_init();
  params=cterm_params(t);
  for (i=0; i<arity; i++) {
    id=eval_var_name(cterm_param(params[i],0),R);
    if (math_error != object_NULL) {
      record_free(res_r);
      return math_error; }
    if (cterm_arity(params[i])==2) {
      aux_o=eval(cterm_param(params[i],1),R);
      if (object_is_error(aux_o)) {
        record_free(res_r);
        return aux_o; } }
    else aux_o=object_link(object_null);
    record_update_field(&res_r,id,aux_o);
    object_unlink(aux_o);
    free_str1(id); }
  return record2object(res_r);
}

object mi_special_sel_ring(object o, char * id, ring R)
{
  object res_o;
  ring S;
  S = o->v.ring_value;
  res_o=env_get_value(ring_local_env(S),id);
  if (res_o==object_NULL) {
    error_set_this_param(cterm_mk_NBS(id));
    res_o = cocoa_error("UNDEF_VAR");
  }
  return res_o;
}

object mi_special_sel_pvector(object o, char * id, ring R)
{
  object res_o;
  pvector v;
  R = object_ring(o);
  v = o->v.pvector_value;
  if (strcmp(id,"Shifts")==0) {
    if (pv_shifts(v)==shifts_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(shifts2olist(pv_shifts(v),R)); }
  else {
    error_set_this_param(string2cterm(id));
    res_o=cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_special_sel_ideal(object o, char * id, ring R)
{
  object res_o;
  /* selection from Ideal (I.Gens, I.Syz, ...) */
  module aux_M;
  aux_M = o->v.module_value;
  R = object_ring(o);
  if (strcmp(id, "SyzOfGens")==0)
  {
    if (aux_M->syz_of_gens == module_NULL) res_o = object_dup(object_null);
    else res_o = module2object(module_link(aux_M->syz_of_gens), R);
  }
  else if (strcmp(id,"Syz")==0) {
    if (module_syz(aux_M)==module_NULL) res_o=object_dup(object_null);
    else res_o=module2object(module_link(module_syz(aux_M)),R); }
  else if (strcmp(id,"Gens")==0) {
    if (module_gens(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(this_vlist2olist_ter(
      vlist_dup(module_gens_ck(aux_M),R),R)); }
  else if (strcmp(id,"MinGens")==0) {
    if (module_mingens(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(this_vlist2olist_ter(
      vlist_dup(module_mingens(aux_M),R),R)); }
  else if (strcmp(id,"Rels")==0) {
    if (module_gbasis(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else {
      vlist res_v=vlist_get_rels(module_gbasis(aux_M),R);
      if (res_v == vlist_NULL) {
        res_o=object_dup(object_null);
      } else {
        res_o=vlist2object(res_v,R);
      }
     }
    }
  else if (strcmp(id,"GBasis")==0) {
    if (module_gbasis(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(this_vlist2olist_ter(
      vlist_dup(module_gbasis(aux_M),R),R)); }
  else if (strcmp(id,"Rules")==0) {
    if (gb_Rules(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(this_vlist2olist_ter(
      vlist_dup(gb_Rules(aux_M),R),R)); }
  else if (strcmp(id,"IVs")==0) {
    if (gb_IVs(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(this_vlist2olist_ter(
      vlist_dup(gb_IVs(aux_M),R),R)); }
  else if (strcmp(id,"CV")==0) {
    if (gb_CV(aux_M)==ev_NULL) res_o=object_dup(object_null);
    else res_o=object_cast(pv2object(
      ev_whole_vector(gb_CV(aux_M),R),R),objtype_poly,R); }
  else if (strcmp(id,"CVRel")==0) {
    if (gb_CV(aux_M)==ev_NULL) res_o=object_dup(object_null);
    else res_o=pv2object(pv_dup(ev_rel(gb_CV(aux_M)),R),R); }
  else if (strcmp(id,"DegTrunc")==0)
    res_o=int2object(module_DegTrunc(aux_M));
  else if (strcmp(id,"SyzCompsNum")==0)
    res_o=int2object(module_SyzCompsNum(aux_M));
  else if (strcmp(id,"ResTrunc")==0)
    res_o=int2object(module_ResTrunc(aux_M));
  else if (strcmp(id,"IsHomog")==0)
    res_o=bool2object(module_flag_is_homog(aux_M));
  else if (strcmp(id,"SyzAreOfMinGens")==0)
    res_o=bool2object(module_flag_syz_of_mingens(aux_M));
  else if (strcmp(id,"Discrepancy")==0)
    res_o=int2object(module_discrepancy(aux_M));
  else if (strcmp(id,"KFL")==0)
    res_o=olist2object(gb_kf2olist(aux_M));
  else if (strcmp(id,"Stats")==0)
    res_o=olist2object(gb_stats2olist(aux_M));
  else if (strcmp(id,"NumComps")==0)
    res_o=int2object(module_rank(aux_M));
  else if (strcmp(id,"PSeries")==0) {
    if (module_pseries(aux_M)==pseries_NULL) res_o=object_dup(object_null);
    else {
      res_o=olist2object(pseries2olist(module_pseries(aux_M)));
      object_set_label(res_o,my_strdup("$hp.PSeries"));
    }
  } else {
    error_set_this_param(string2cterm(id));
    res_o=cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_special_sel_module(object o, char * id, ring R)
{
  module aux_M;
  object res_o;
  aux_M = o->v.module_value;
  R = object_ring(o);
  if (strcmp(id, "SyzOfGens")==0)
  {
    if (aux_M->syz_of_gens == module_NULL) res_o = object_dup(object_null);
    else res_o = module2object(module_link(aux_M->syz_of_gens), R);
  }
  else if (strcmp(id,"Syz")==0) {
    if (module_syz(aux_M)==module_NULL) res_o=object_dup(object_null);
    else res_o=module2object(module_link(module_syz(aux_M)),R); }
  else if (strcmp(id,"MinGens")==0) {
    if (module_mingens(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_dup(module_mingens(aux_M),R),R); }
  else if (strcmp(id,"Gens")==0) {
    if (module_gens(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_dup(module_gens_ck(aux_M),R),R); }
  else if (strcmp(id,"GBasis")==0) {
    if (module_gbasis(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_dup(module_gbasis(aux_M),R),R); }
  else if (strcmp(id,"Rels")==0) {
    if (module_gbasis(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_get_rels(module_gbasis(aux_M),R),R); }
  else if (strcmp(id,"Rules")==0) {
    if (gb_Rules(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_dup(gb_Rules(aux_M),R),R); }
  else if (strcmp(id,"IVs")==0) {
    if (gb_IVs(aux_M)==vlist_NULL) res_o=object_dup(object_null);
    else res_o=vlist2object(vlist_dup(gb_IVs(aux_M),R),R); }
  else if (strcmp(id,"CV")==0) {
    if (gb_CV(aux_M)==ev_NULL) res_o=object_dup(object_null);
    else res_o=pv2object(ev_whole_vector(gb_CV(aux_M),R),R); }
  else if (strcmp(id,"CVRel")==0) {
    if (gb_CV(aux_M)==ev_NULL) res_o=object_dup(object_null);
    else res_o=pv2object(pv_dup(ev_rel(gb_CV(aux_M)),R),R); }
  else if (strcmp(id,"Shifts")==0) {
    if (module_shifts(aux_M)==shifts_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(shifts2olist(module_shifts(aux_M),R)); }
  else if (strcmp(id,"DegTrunc")==0)
    res_o=int2object(module_DegTrunc(aux_M));
  else if (strcmp(id,"SyzCompsNum")==0)
    res_o=int2object(module_SyzCompsNum(aux_M));
  else if (strcmp(id,"ResTrunc")==0)
    res_o=int2object(module_ResTrunc(aux_M));
  else if (strcmp(id,"Discrepancy")==0)
    res_o=int2object(module_discrepancy(aux_M));
  else if (strcmp(id,"IsHomog")==0)
    res_o=bool2object(module_flag_is_homog(aux_M));
  else if (strcmp(id,"SyzAreOfMinGens")==0)
    res_o=bool2object(module_flag_syz_of_mingens(aux_M));
  else if (strcmp(id,"KFL")==0)
    res_o=olist2object(gb_kf2olist(aux_M));
  else if (strcmp(id,"Stats")==0)
    res_o=olist2object(gb_stats2olist(aux_M));
  else if (strcmp(id,"NumComps")==0)
    res_o=int2object(module_rank(aux_M));
  else if (strcmp(id,"PSeries")==0) {
    if (module_pseries(aux_M)==pseries_NULL) res_o=object_dup(object_null);
    else res_o=olist2object(pseries2olist(module_pseries(aux_M))); }
  else {
    error_set_this_param(string2cterm(id));
    res_o=cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_special_sel_arbrat(object o, char * id, ring R)
{
  arbrat aux_r;
  object res_o;
  aux_r = o->v.arbrat_value;
  if (strcmp(id,"Num")==0)
    res_o=arbint2object(arbint_dup(arbrat_num(aux_r)));
  else if (strcmp(id,"Den")==0)
    res_o=arbint2object(arbint_dup(arbrat_den(aux_r)));
  else {
    error_set_this_param(string2cterm(id));
    res_o=cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_special_sel_ratfun(object o, char * id, ring R)
{
  ratfun aux_r;
  object res_o;
  aux_r = o->v.ratfun_value;
  R = object_ring(o);
  if (strcmp(id,"Num")==0)
    res_o=arpoly2object(arpoly_dup(ratfun_num(aux_r),R),R);
  else if (strcmp(id,"Den")==0)
    res_o=arpoly2object(arpoly_dup(ratfun_den(aux_r),R),R);
  else {
    error_set_this_param(string2cterm(id));
    res_o=cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_special_sel(object o, char * id, ring R)
{
  objtype ot = object_type(o);
  switch (ot) {
  case objtype_module: return mi_special_sel_module(o,id,R);
  case objtype_ideal: return mi_special_sel_ideal(o,id,R);
  case objtype_ring: return mi_special_sel_ring(o,id,R);
  case objtype_pvector: return mi_special_sel_pvector(o,id,R);
  case objtype_arbrat: return mi_special_sel_arbrat(o,id,R);
  case objtype_ratfun: return mi_special_sel_ratfun(o,id,R);
  default: error_set_this_param(objtype2cterm(ot,NULL));
           return cocoa_error("CANNOT_SEL");
  }
}

object do_record_sel(object o, char * id, ring R)
{
  object res_o;
  if (object_type(o)!=objtype_record) return mi_special_sel(o,id,R);
  res_o=record_sel(o->v.record_value,id);
  if (res_o==object_NULL) {
    error_set_this_param(string2cterm(id));
    return cocoa_error("UNKN_FIELD"); }
  return res_o;
}

object mi_record_sel(cterm t, ring R)
{ /* t=RECORD_SEL(R,Id) */
  int i;
  record res_r;
  char * id;
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  id=eval_var_name(cterm_param(t,1),R);
  if (math_error != object_NULL) return math_error;
  res_o=do_record_sel(aux_o,id,R);
  object_unlink(aux_o);
  free_str1(id);
  return res_o;
}

object mi_change_sign(cterm t, ring R)
{ 
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_change_sign(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_power(cterm t, ring R)
{
  object aux_o,res_o;
  expType aux_i;
  unknown aux_u;
  cterm t1, t2;
  coc_bool tobeinverted=FALSE;
  t1=cterm_param(t,0);
  t2=cterm_param(t,1);
  aux_i=eval_int(t2,R);
  if (aux_i<0) { aux_i=-aux_i; tobeinverted=TRUE; }
  if (math_error!=object_NULL) return math_error;
  if (cterm_op(t1)==op_INDET) {
    arpoly aux_p;
    aux_u=mi_cterm2unknown(t1,R);
    if (math_error!=object_NULL) return math_error;
    aux_p=arpoly_unknown_pow2arpoly(aux_u,aux_i,R);
    res_o=arpoly2object(aux_p,R); }
  else {
    aux_o=eval(t1,R);
    if (object_is_error(aux_o)) return aux_o;
    res_o=do_power(aux_o,aux_i);
    object_unlink(aux_o); }
  if (object_is_error(res_o)) return res_o;
  if (tobeinverted) do_inv(&res_o);
  return res_o;
}

object mi_subtract(cterm t, ring R)
{
  object o1,o2,res_o;
  o1=eval(cterm_param(t,0),R);
  if (object_is_error(o1)) return o1;
  o2=eval(cterm_param(t,1),R);
  if (object_is_error(o2)) {
    object_unlink(o1);
    return o2; }
  res_o=do_subtract(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_o;
}

object mi_det(cterm t, ring R)
{
  object aux_o,res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_det(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_adjoint(cterm t, ring R)
{
  object aux_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  return do_adj(aux_o);
}

object mi_dense_poly(cterm t, ring R)
{
  int deg;
  if (cterm_arity(t) != 1) return cocoa_error("BAD_PARAMS_NUM");
  deg = eval_int(cterm_param(t,0),R);
  if (math_error!=object_NULL) {
    object res_o;
    res_o=math_error;
    math_error=object_NULL;
    return math_error;
  }
  return do_dense_poly(deg,R);
}

object mi_leading_term(cterm t,ring R)
{ 
  object aux_o;
  object res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_eval_leading_term(aux_o,R);
  object_unlink(aux_o);
  return res_o;
}

object mi_leading_coeff(cterm t,ring R)
{ 
  object aux_o;
  object res_o;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_eval_leading_coeff(aux_o,R);
  object_unlink(aux_o);
  return res_o;
}

object mi_bm(cterm t, ring R)
{
  object args = mi_params(t,R);
  object ans;
  ans = BM_interface(args, R);
  object_unlink(args);
  return ans;
}

object mi_gbm(cterm t, ring R)
{
  object args = mi_params(t,R);
  object ans;
  if (ring_characteristic(R) < 2)
    ans = GBM_interface(args, R);
  else ans = GBM_interface_FF(args, R);
  object_unlink(args);
  return ans;
}

object mi_isqrt(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    error_set_context(t);
    return aux_o; }
  res_o = do_isqrt(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_iroot(cterm t, ring R)
{
  object res_o,aux_o1,aux_o2;
  cterm t1, t2;
  t1 = cterm_param(t,0);
  aux_o1 = eval(t1, R);
  if (object_is_error(aux_o1))
  {
    error_set_context(t1);
    return aux_o1;
  }
  t2 = cterm_param(t,1);
  aux_o2 = eval(t2, R);
  if (object_is_error(aux_o2))
  {
    object_unlink(aux_o1);
    error_set_context(t2);
    return aux_o2;
  }
  res_o = do_iroot(aux_o1, aux_o2);
  object_unlink(aux_o2);
  object_unlink(aux_o1);
  return res_o;
}

object mi_QZP(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    error_set_context(t);
    return aux_o; }
  res_o = do_QZP(aux_o, R);
  object_unlink(aux_o);
  return res_o;
}

object mi_ZPQ(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    error_set_context(t);
    return aux_o; }
  res_o = do_ZPQ(aux_o, R);
  object_unlink(aux_o);
  return res_o;
}

object mi_IsPPrime(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o))
  {
    error_set_context(t);
    return aux_o;
  }
  res_o = do_IsPPrime(aux_o);
  object_unlink(aux_o);
  return res_o;
  
}

object mi_NextPPrime(cterm t, ring R)
{
  object res_o,aux_o;
  t=cterm_param(t,0);
  aux_o=eval(t,R);
  if (object_is_error(aux_o))
  {
    error_set_context(t);
    return aux_o;
  }
  res_o = do_NextPPrime(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_PowerMod(cterm t, ring R)
{
  object res_o,aux_o1,aux_o2,aux_o3;
  cterm t1, t2, t3;
  t1 = cterm_param(t,0);
  aux_o1 = eval(t1, R);
  if (object_is_error(aux_o1))
  {
    error_set_context(t1);
    return aux_o1;
  }
  t2 = cterm_param(t,1);
  aux_o2 = eval(t2, R);
  if (object_is_error(aux_o2))
  {
    object_unlink(aux_o1);
    error_set_context(t2);
    return aux_o2;
  }
  t3 = cterm_param(t,2);
  aux_o3 = eval(t3, R);
  if (object_is_error(aux_o3))
  {
    object_unlink(aux_o2);
    object_unlink(aux_o1);
    error_set_context(t3);
    return aux_o3;
  }
  res_o = do_PowerMod(aux_o1, aux_o2, aux_o3);
  object_unlink(aux_o3);
  object_unlink(aux_o2);
  object_unlink(aux_o1);
  return res_o;
}


#ifdef COCOA_HOOK

object mi_hook(cterm t, ring R)
{ 
  int n;
  object aux_o1,aux_o2,res_o;
  aux_o1=eval(cterm_param(t,0),R);
  n=object_cast_int(aux_o1);
  if (math_error != object_NULL) return math_error;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    return aux_o2;
  }
  switch (n) {
    case 1:res_o=hook1(aux_o2,R); break;
    case 2:res_o=hook2(aux_o2,R); break;
    case 3:res_o=hook3(aux_o2,R); break;
    case 4:res_o=hook4(aux_o2,R); break;
    case 5:res_o=hook5(aux_o2,R); break;
    case 6:res_o=hook6(aux_o2,R); break;
    case 7:res_o=hook7(aux_o2,R); break;
    case 8:res_o=hook8(aux_o2,R); break;
    case 9:res_o=hook9(aux_o2,R); break;
    case 10:res_o=hook10(aux_o2,R); break;
  }
  object_unlink(aux_o2);
  return res_o;
}

#endif

object mi_help(cterm t,ring R)
{ /* TODO: check memory */
  object aux_o,res_o;
  char * aux_s, * aux_s1, * pkg_id, * op_id;
  int pkg_n,n;
  cterm aux_t;
  char buffer[50];
  coop aux_op;

  aux_o=eval(cterm_param(t,0),R);

  if (object_is_error(aux_o)) return aux_o;
  if (object_type(aux_o) != objtype_string) {
    res_o=error_expected_found(objtype_string,object_type(aux_o));
    object_unlink(aux_o);
    return res_o;
  }
  aux_s=object2string(aux_o);
  aux_op=op_map_coop(pkg_user_n,aux_s);

  if (aux_op!=op_NULL) {
    aux_s1=op_map_op_help(pkg_user_n,op_index(aux_op));
    if (aux_s1!=NULL) {
      free_str1(aux_s);
      return string2object(my_strdup(aux_s1));
    }
  }
  aux_s1=my_strdup(aux_s);
  n=strlen(aux_s1);
  str_split(aux_s1,".",&pkg_id,&op_id);
  if (op_id == NULL) {
    aux_op=op_NULL;
  } else {
    sprintf(buffer,"Help_%s\0",op_id);
    if (pkg_id == NULL) {
      coop_pkg_std_op(buffer,&pkg_n,&aux_op);
    } else {
      pkg_n=pkg_index(pkg_id);
      if (pkg_n != name_id_NULL) {
        aux_op=op_map_coop(pkg_n,buffer);
      } else aux_op=op_NULL;
    }
  }
  free_str(n,aux_s1);
  if (aux_op != op_NULL) aux_op=op_mk(pkg_n,aux_op);

  if (aux_op!=op_NULL) {
    aux_op=op_mk(pkg_n,aux_op);
    aux_t=cterm_mk_const(aux_op);
    res_o=eval_caller(aux_t,R);
    cterm_free(aux_t);
  } else {
    sprintf(buffer,"No help available for %s",aux_s);
    res_o=string2object(my_strdup(buffer));
  }
  free_str1(aux_s);
  return res_o;
}

object mi_parenthesis(cterm t, ring R)
{
  return eval(cterm_param(t,0),R);
}

object mi_null(cterm t, ring R)
{
  return object_dup(object_null); /* NOT LINK */
}

object mi_tao(cterm t, ring R)
{
  return string2object(my_strdup("Tao ;-)"));
}

object mi_bin_toric(cterm t, ring R)
{ /* BinToric(LB,LI) */
  /* where LB is a nonempty list of binomials */
  /* and LI is a list of indeterminates */
  object aux_o1,aux_o2,res_o;
  ring aux_R;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  aux_R=object_extract_ring(aux_o1);
  if (aux_R == ring_NULL) {
    object_unlink(aux_o1);
    object_unlink(aux_o2);
    return cocoa_error("BAD_PARAMS");
  }
  ToricAlg = EATI;
  res_o=do_bin_toric(aux_o1,aux_o2,aux_R);
  object_unlink(aux_o1);
  object_unlink(aux_o2);
  return res_o;
}

object mi_bin_seqtoric(cterm t, ring R)
{ /* BinToric(LB,LI) */
  /* where LB is a nonempty list of binomials */
  /* and LI is a list of indeterminates */
  object aux_o1,aux_o2,res_o;
  ring aux_R;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  aux_R=object_extract_ring(aux_o1);
  if (aux_R == ring_NULL) {
    object_unlink(aux_o1);
    object_unlink(aux_o2);
    return cocoa_error("BAD_PARAMS");
  }
  ToricAlg = SATI;
  res_o=do_bin_toric(aux_o1,aux_o2,aux_R);
  object_unlink(aux_o1);
  object_unlink(aux_o2);
  return res_o;
}

object mi_bin_partoric(cterm t, ring R)
{ /* BinToric(LB,LI) */
  /* where LB is a nonempty list of binomials */
  /* and LI is a list of indeterminates */
  object aux_o1,aux_o2,res_o;
  ring aux_R;
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2)) {
    object_unlink(aux_o1);
    return aux_o2;
  }
  aux_R=object_extract_ring(aux_o1);
  if (aux_R == ring_NULL) {
    object_unlink(aux_o1);
    object_unlink(aux_o2);
    return cocoa_error("BAD_PARAMS");
  }
  ToricAlg = PATI;
  res_o=do_bin_toric(aux_o1,aux_o2,aux_R);
  object_unlink(aux_o1);
  object_unlink(aux_o2);
  return res_o;
}



object mi_bool(cterm t, ring R)
{
  coc_bool aux_b;
  aux_b=eval_bool(t,R);
  if (math_error != object_NULL) return math_error;
  return bool2object(aux_b); 
}

object mi_objtype(cterm t, ring R)
{
  /* TODO: IMPROVE (a lot) using a table of (linkable) objects */
  object res_o;
  coop op=cterm_op(t);

  if ((cterm_arity(t) > 0) && (op != op_objtype_tagged))
    return cocoa_error("BAD_PARAMS_NUM");
  else if ((cterm_arity(t) > 1) && (op == op_objtype_tagged))
    return cocoa_error("BAD_PARAMS_NUM");

  if (op==op_objtype_NULL) return objtype2object(objtype_NULL,NULL);
  else if (op==op_objtype_NULL) return objtype2object(objtype_NULL,NULL);
  else if (op==op_objtype_coeff) return objtype2object(objtype_coeff,NULL);
  else if (op==op_objtype_ring) return objtype2object(objtype_ring,NULL);
  else if (op==op_objtype_poly) return objtype2object(objtype_poly,NULL);
  else if (op==op_objtype_bool) return objtype2object(objtype_bool,NULL);
  else if (op==op_objtype_cterm) return objtype2object(objtype_cterm,NULL);
  else if (op==op_objtype_int) return objtype2object(objtype_int,NULL);
  else if (op==op_objtype_ideal) return objtype2object(objtype_ideal,NULL);
  else if (op==op_objtype_pvector) 
    return objtype2object(objtype_pvector,NULL);
  else if (op==op_objtype_vlist) return objtype2object(objtype_vlist,NULL);
  else if (op==op_objtype_string) return objtype2object(objtype_string,NULL);
  else if (op==op_objtype_module) return objtype2object(objtype_module,NULL);
  else if (op==op_objtype_pseries) 
    return objtype2object(objtype_pseries,NULL);
  else if (op==op_objtype_tagged) {
    char * aux_s;
    object aux_o;
    if (cterm_arity(t) == 0) aux_s=my_strdup("");
    else {
      aux_o=eval(cterm_param(t,0),R);
      if (object_is_error(aux_o)) return aux_o;
      aux_s=object_cast_string(aux_o);
      if (math_error != object_NULL) return math_error;
    }
    res_o=objtype2object(objtype_objtype,aux_s);
    free_str1(aux_s);
    return res_o;
  }
  else if (op==op_objtype_olist) return objtype2object(objtype_olist,NULL);
  else if (op==op_objtype_omat) return objtype2object(objtype_omat,NULL);
  else if (op==op_objtype_ratfun) return objtype2object(objtype_ratfun,NULL);
  else if (op==op_objtype_err) return objtype2object(objtype_err,NULL);
  else if (op==op_objtype_arbint) return objtype2object(objtype_arbint,NULL);
  else if (op==op_objtype_arbrat) return objtype2object(objtype_arbrat,NULL);
  else if (op==op_objtype_zp) return objtype2object(objtype_zmod,NULL);
  else if (op==op_objtype_double) return objtype2object(objtype_double,NULL);
  else if (op==op_objtype_objtype) 
    return objtype2object(objtype_objtype,NULL);
  else if (op==op_objtype_cdev) return objtype2object(objtype_cdev,NULL);
  else if (op==op_objtype_record) return objtype2object(objtype_record,NULL);
  else if (op==op_objtype_cocfun) return objtype2object(objtype_cocfun,NULL);
  else return objtype2object(objtype_NULL,NULL);
}

object mi_ispurelt(cterm t,ring R)
{
  object aux_o;
  arpoly f;
  object ans;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  f = object_cast_arpoly(aux_o, R); /* this destroys aux_o */
  if (math_error != object_NULL)
  {
    ans = math_error;
    math_error = object_NULL;
    return ans;
  }
  if (arpoly_is_zero(f, R)) ans = cocoa_error("LT_OF_ZERO");
  else ans = do_ispurelt(f);
  arpoly_free(f,R);
  return ans; /* either a boolean or an error */
}

object mi_deg(cterm t,ring R)
{
  object aux_o;
  int res_n;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_n=do_deg(aux_o);
  object_unlink(aux_o);
  if (math_error != object_NULL) return math_error;
  return int2object(res_n);
}

object mi_mdeg(cterm t,ring R)
{
  object res_o, aux_o;
  int res_n;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_o=do_mdeg(aux_o);
  object_unlink(aux_o);
  return res_o;
}

object mi_size(cterm t,ring R)
{
  object aux_o;
  int res_n;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_n=do_size(aux_o);
  object_unlink(aux_o);
  if (math_error != object_NULL) return math_error;
  return int2object(res_n);
}

object mi_len(cterm t,ring R)
{
  object aux_o;
  int res_n;
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  res_n=do_len(aux_o);
  object_unlink(aux_o);
  if (math_error != object_NULL) return math_error;
  return int2object(res_n);
}

object mi_num_indets(cterm t,ring R)
{
  ring aux_R;
  int res_n;
  if (cterm_arity(t)==0) res_n=ring_indets_num(R);
  else {
    aux_R=eval_ring(cterm_param(t,0),R);
    if (math_error != object_NULL) return math_error;
    res_n=ring_indets_num(aux_R); }
  return int2object(res_n);
}

object mi_characteristic(cterm t,ring R)
{
  ring aux_R;
  int res_n;
  if (cterm_arity(t)==0) res_n=ring_characteristic(R);
  else {
    aux_R=eval_ring(cterm_param(t,0),R);
    if (math_error!=object_NULL) return math_error;
    res_n=ring_characteristic(aux_R); }
  if (res_n<2) res_n=0;
  return int2object(res_n);
}

/*List of the Terms/Module terms of a poly/pvector*/
object mi_support(cterm t, ring R)
{
  object aux_o;
  olist ol;
  ring aux_R;
  
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_R=object_ring(aux_o);
  switch (object_type(aux_o))
  {
      case objtype_pvector:
        ol=support_of_this_pv(object2pv(aux_o,aux_R), aux_R);
        break;

      case objtype_poly:
       ol=support_of_this_arpoly(object2arpoly(aux_o,aux_R),aux_R);
       break;

      default:
        object_unlink(aux_o);
        return error2object(my_strdup("Poly/Vector expected"));
  }
  return olist2object(ol);
}

/*List of the Monomials/Module monomials of a poly/pvector*/
object mi_monomials(cterm t, ring R)
{
  object aux_o;
  olist ol;
  ring aux_R;
  
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_R=object_ring(aux_o);
  switch (object_type(aux_o))
  {
      case objtype_pvector:
        ol=monomials_of_this_pv(object2pv(aux_o,aux_R), aux_R);
        break;

      case objtype_poly:
        ol=monomials_of_this_arpoly(object2arpoly(aux_o,aux_R),aux_R);
       break;

      default:
        object_unlink(aux_o);
        return error2object(my_strdup("Poly/Vector expected"));
  }
  return olist2object(ol);
}

/*Coeff List of a poly/pvector*/
object mi_coeffs(cterm t, ring R)
{
  object aux_o;
  olist ol;
  ring aux_R;
  pvector pv;
  
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) return aux_o;
  aux_R=object_ring(aux_o);
  switch (object_type(aux_o))
  {
      case objtype_pvector:
        pv=object2pv(aux_o,aux_R);
        ol=coeffs_of_this_pv(pv, aux_R);
        break;

      case objtype_poly:
        ol=coeffs_of_this_arpoly(object2arpoly(aux_o,aux_R),aux_R);
       break;

      default:
        object_unlink(aux_o);
        return error2object(my_strdup("Poly/Vector expected"));
  }
  return olist2object(ol);
}

/*Checks if a ring has PosTo ordering*/
object mi_isposto(cterm t, ring R)
{
  object res_o;
  ring R1;
  short module_ord;
  
  R1=eval_ring(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  module_ord=ring_mordering(R1);
  res_o=bool2object(module_ord==POS_TO);
  return res_o;
}

/*Checks if a ring has ToPos ordering*/
object mi_istopos(cterm t, ring R)
{
  object res_o;
  ring R1;
  short module_ord;
  
  R1=eval_ring(cterm_param(t,0),R);
  if (math_error!=object_NULL) return math_error;
  module_ord=ring_mordering(R1);
  res_o=bool2object(module_ord==TO_POS);
  return res_o;
}

/*Returns the first non zero power product of a pvector*/
object mi_firstnonzero(cterm t, ring R)
{
  object aux_o,res_o;
  ring aux_R;
  pvector pv;
  
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  if ((object_type(aux_o)!=objtype_pvector) && 
      (object_type(aux_o)!=objtype_poly)) 
     return error2object(my_strdup("FirstNonZero: Vector expected"));
  if (object_type(aux_o)==objtype_poly)
      return object_dup(aux_o);
  pv=object2pv(aux_o,aux_R);
  if (pv_is_zero(pv)) { pv_free(pv, aux_R); return error2object(my_strdup("FirstNonZero: zero vector")); }

  res_o=arpoly2object(pv_FirstNonZero(pv,aux_R),aux_R);
  pv_free(pv,aux_R);
  return res_o;
}

/*Returns the first non zero position of a pvector*/
object mi_firstnonzeropos(cterm t, ring R)
{
  object aux_o,res_o;
  ring aux_R;
  pvector pv;
  arpoly p;
 
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  if ((object_type(aux_o)!=objtype_pvector) && 
      (object_type(aux_o)!=objtype_poly)) 
     return error2object(my_strdup("Vector expected"));
   if (object_type(aux_o)==objtype_poly){
      p=object2arpoly(aux_o, aux_R);
      if (arpoly_is_zero(p, aux_R)) res_o=int2object(0);
      else res_o=int2object(1);
      arpoly_free(p,aux_R);
      return res_o;
      };
  pv=object2pv(aux_o,aux_R);
  if (pv_is_zero(pv)) { pv_free(pv, aux_R); return error2object(my_strdup("FirstNonZeroPos: zero vector")); }
  res_o=int2object(pv_FirstNonZeroPos(pv,aux_R));
  pv_free(pv,aux_R);
  return res_o;
}

/*Checks a poly/pvector : is it a term?*/
object mi_isterm(cterm t, ring R)
{
  object aux_o;
  ring aux_R;
  coc_bool b;
  
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  if (object_type(aux_o)==objtype_pvector) { 
      pvector pv;
      pv=object2pv(aux_o,aux_R);
      b=pvector_is_term(pv,aux_R);
      pv_free(pv,aux_R);
      return bool2object(b);}
  if (object_type(aux_o)==objtype_poly) {
      arpoly p; 
      p=object2arpoly(aux_o,aux_R);
      b=arpoly_is_term(p,aux_R);
      arpoly_free(p,aux_R);
      return bool2object(b);
  }
   
  object_unlink(aux_o);
  return error2object(my_strdup("IsTerm: Poly/Vector expected"));
}

/*Leading position of pvector*/
object mi_lpos(cterm t, ring R)
{
  object aux_o,res_o;
  ring aux_R;
  int pos;
  pvector pv;
  arpoly p;
  
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  if ((object_type(aux_o)!=objtype_pvector))
     return error2object(my_strdup("LPos: Vector expected"));
/*    if (object_type(aux_o)==objtype_poly){ */
/*        p=object2arpoly(aux_o, aux_R); */
/*        if (arpoly_is_zero(p, aux_R)) res_o=int2object(0); */
/*        else res_o=int2object(1); */
/*        arpoly_free(p,aux_R); */
/*        return res_o; */
/*        }; */
  pv=object2pv(aux_o,aux_R);
  if (pv_is_zero(pv)) pos = -1;
  else pv_leading_pos(pv,&pos,aux_R);
  pv_free(pv,aux_R);
  res_o=int2object(pos+1);
  return res_o;
}

/*Leading power product of poly or pvector*/
object mi_lpp(cterm t, ring R)
{
  object aux_o,res_o;
  ring aux_R;
  int pos;
  monomial m;
  
  
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  switch (object_type(aux_o)) {
    case objtype_poly:{
      arpoly p;
      p=object2arpoly(aux_o,aux_R);
      if (arpoly_is_zero(p, R)) { arpoly_free(p, R); return error2object(my_strdup("LPP: zero poly")); }
      m=arpoly_leading_monomial(p,aux_R);
      res_o=arpoly2object(arpoly_this_term2arpoly(term_dup(monomial_term(m)),aux_R),aux_R);
      monomial_free(m,R);
      arpoly_free(p,R);
      break;
    }
    case objtype_pvector: {
      pvector pv;
      pv=object2pv(aux_o,aux_R);
      if (pv_is_zero(pv)) { pv_free(pv, R); return error2object(my_strdup("LPP: zero vector")); }
      pv_leading_monomial(pv,&m,&pos,aux_R);
      res_o=arpoly2object(arpoly_this_term2arpoly(term_dup(monomial_term(m)),aux_R),aux_R);
      monomial_free(m,aux_R);
      pv_free(pv,aux_R);
      break;
    }
    default :return error2object(my_strdup("LPP: Poly/Vector expected"));
  }
 
   return res_o;
}


/*Leading Monomial of poly/pvector*/
object mi_lm(cterm t, ring R)
{
  object aux_o,res_o;
  ring aux_R;
  int pos;
  monomial m;
  
  aux_o=eval(cterm_param(t,0),R);
  aux_R=object_ring(aux_o);
  if (object_is_error(aux_o)) return aux_o;
  switch (object_type(aux_o)) {
    case objtype_pvector: {
      pvector pv;
      pv=object2pv(aux_o,aux_R);
      if (pv_is_NULL_or_zero(pv)) return error2object(my_strdup("LM: zero vector"));
      pv_leading_monomial(pv,&m,&pos,aux_R);
      res_o=pv2object(this_monomial2pv(m,pos,pv_rank(pv),aux_R),aux_R);
      pv_free(pv,R);
      break;}
    case objtype_poly: {
      arpoly p;
      p=object2arpoly(aux_o,aux_R);
      if (p==NULL) return error2object(my_strdup("LM: zero poly"));
      m=arpoly_leading_monomial(p,aux_R);
      res_o=arpoly2object(arpoly_this_monomial2arpoly(m,aux_R),
                          aux_R);
      arpoly_free(p,aux_R);
      break;}
    default:    
      return error2object(my_strdup("LM: Poly/Vector expected"));
  }
  /*monomial_free(m,R);*/
  return res_o;
}


/*Poly-Term / Vector-ModuleTerm : Coeff
  Type checking is done in the misc:CoeffOfTerm function */
object mi_coeffofterm(cterm t, ring R) 
{  
  object aux_o1,aux_o2,res_o;
  ring aux_R;
  coeff res_coeff;
    
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2) ) {
    object_unlink(aux_o1);
    return aux_o2;}
  if (!object_rings_compatible(aux_o1,aux_o2)) {
    object_unlink(aux_o1);
    object_unlink(aux_o2);
    return error2object(my_strdup("CoeffOfTerm: Rings are incompatible"));}
    
  aux_R=object_ring(aux_o1);
  switch (object_type(aux_o2)) {
    case objtype_poly:{
       arpoly p;
       term myterm;
       
       p=object2arpoly(aux_o2,aux_R);
       myterm=arpoly_leading_term(p,aux_R);
       res_coeff=this_arpoly_coeff_of_term(object2arpoly(aux_o1,aux_R),
                                           myterm,aux_R);
       res_o=coeff2object(res_coeff,R);
       arpoly_free(p,R);
       term_free(myterm);
       break;}
    case objtype_pvector:{
       pvector pvt;
       pvt=object2pv(aux_o2,aux_R);
       
       res_coeff=this_pvector_coeff_of_term(object2pv(aux_o1,aux_R),
                                            pvt,aux_R);
       res_o=coeff2object(res_coeff,R);
       pv_free(pvt,R);
       break;}
   default :      
      return error2object(my_strdup("CoeffOfTerm: Poly/Vector expected"));
  }
  return res_o;
}


/*Normal reduction of a Poly/Vector w.r.t. a list of Poly/Vector s
  Result : poly
  Type checking is done in the misc:NR function */
object mi_nr(cterm t, ring R) 
{  
  object aux_o1,aux_o2,res_o;
  ring aux_R;
  pvector pv, nr_res_pv;
  module m;
  coc_bool ispoly;
      
  aux_o1=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o1)) return aux_o1;
  aux_o2=eval(cterm_param(t,1),R);
  if (object_is_error(aux_o2) ) {
    object_unlink(aux_o1);
    return aux_o2;}
  
  aux_R=object_ring(aux_o1);
  ispoly=object_type(aux_o1)==objtype_poly;
       
  pv=object_cast_pvector(aux_o1,aux_R);
  if (ispoly) m=this_vlist2module(object_cast_vlist(aux_o2,aux_R),1,aux_R);
  else m=object_cast_module(aux_o2,aux_R);
       
  nr_res_pv=nr(pv,m,aux_R);
	 
  pv_free(pv,aux_R);
  module_unlink(m,aux_R);
  
  if (ispoly) {
     arpoly p;
     p=this_pvcomp2arpoly(pvcomp_dup(pv_nth(nr_res_pv,0),aux_R),aux_R);
     pv_free(nr_res_pv,aux_R);
     return arpoly2object(p,aux_R);}
  else 
     return pv2object(nr_res_pv,aux_R);
}


object mi_NewPPStream(cterm t,ring R)
{
  ring S;

  if (cterm_arity(t) != 0 && cterm_arity(t) != 1)
    return cocoa_error("BAD_PARAMS");
  S = R;
  if (cterm_arity(t) == 1)
  {
    S = eval_ring(cterm_param(t, 0), R);
    if (math_error != object_NULL) return math_error;
  }
  return do_NewPPStream(S);
}

object mi_NextPP(cterm t, ring R)
{
  object o, ans;

  o = eval(cterm_param(t,0), R);
  if (object_is_error(o)) return o;
  if (object_type(o) != objtype_arbint) return cocoa_error("BAD_PARAMS");
  ans = do_NextPP(o, R);
  object_unlink(o);
  return ans;
}


object mi_AvoidLastPP(cterm t, ring R)
{
  object o, ans;

  o = eval(cterm_param(t,0), R);
  if (object_is_error(o)) return o;
  if (object_type(o) != objtype_arbint) return cocoa_error("BAD_PARAMS");
  do_AvoidLastPP(o, R);
  ans = object_link(object_null);
  object_unlink(o);
  return ans;
}


object mi_NextPrime(cterm t, ring R)
{
  object o, ans;

  o = eval(cterm_param(t,0), R);
  if (object_is_error(o)) return o;
  if (object_type(o) != objtype_arbint && object_type(o) != objtype_int) return cocoa_error("BAD_PARAMS");
  if (object_type(o) == objtype_arbint) ans = do_NextPrime(o->v.arbint_value);
  else ans = do_NextPrime2(o->v.int_value);
  object_unlink(o);
  return ans;
}


/* ASSUMES the argument types are already checked (at high level?) */
object mi_Mod2Rat(cterm t, ring R)
{
  object o1, o2, o3, ans;
  arbint res, mod, Q;

  o1 = eval(cterm_param(t,0), R);
  o2 = eval(cterm_param(t,1), R);
  o3 = eval(cterm_param(t,2), R);
  res = o1->v.arbint_value;
  mod = o2->v.arbint_value;
  Q = o3->v.arbint_value;

  ans = do_Mod2Rat(res, mod, Q);
  object_unlink(o3);
  object_unlink(o2);
  object_unlink(o1);
  return ans;
}

object mi_system(cterm t, ring R)
{
#ifndef WITH_SYSTEM_CMD
  return cocoa_error("SYSTEM_CALL_PROHIBITED");
#else
  object o;
  int ans;
  o = eval(cterm_param(t,0), R);
  if (object_is_error(o)) return o;
  if (object_type(o) != objtype_string) return cocoa_error("BAD_PARAMS");
  if (system(NULL) == 0) return cocoa_error("NO_SHELL");
  ans = system(o->v.string_value);
  if (ans == -1) return cocoa_error("SYSTEM_CALL_FAILED");
  object_unlink(o);
  return arbint2object(int2arbint(ans));
#endif
}
