/* $Id: math_aux.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include <stdlib.h>
#include <time.h>

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "coeffs.h"
#include "engine.h"
#include "gb.h"
#include "ivectors.h"
#include "indet.h"
#include "memoryco.h"
#include "machine.h"
#include "map.h"
#include "mathcoc.h"
#include "math_lib.h"
#include "monomial.h"
#include "module.h"
#include "olist.h"
#include "object.h"
#include "operator.h"
#include "options.h"
#include "omat.h"
#include "pseries.h"
#include "polys.h"
#include "pvector.h"
#include "record.h"
#include "ring.h"
#include "subst.h"
#include "terms.h"
#include "utils.h"
#include "unipoly.h"
#include "vlist.h"
#include "weights.h"

extern map curr_al_map;
extern charstars curr_al_pkgs;
extern int funcall_level;

object mi_params(cterm t, ring R);
object mi_params1(cterm t, ring R);

extern object engine_in_ring(cterm t, ring R);
object eval_also_tagged(cterm t,ring R);

/* ---[ auxiliary ]------------------------------------------ */

object mi_limits(cterm t, ring R)
{
  record res_r;
  res_r=record_init();
  record_update_field(&res_r,"MaxExp",int2object(INT_MAX));
  record_update_field(&res_r,"MaxChar",int2object(SHRT_MAX));
  return record2object(res_r);
}

object mi_time_aux(cterm t, ring R)
{ /* NOTE: it calls engine_in_ring() */
  clock_t t1,t2,c1,c2;
  record res_r;
  object res_o,aux_o;
  get_time(&t1);
  get_cputime(&c1);
  res_o=engine_in_ring(t,R);
  if (res_o != object_null) return res_o;
  get_time(&t2);
  get_cputime(&c2);
  res_r=record_init();
  aux_o=int2object(t2-t1);
  record_update_field(&res_r,"User",aux_o);
  object_unlink(aux_o);
  aux_o=int2object(cputime_diff(c2,c1));
  record_update_field(&res_r,"Cpu",aux_o);
  object_unlink(aux_o);
  res_o=record2object(res_r);
  object_set_label(res_o,my_strdup("Time"));
  return res_o;
}

object eval_ringenv(cterm t_ring, cterm t_mod, ring R)
{
  object o,res_o;
  ring res_R;
  /* The syntax allows a weight/ordering/other modifier even on already */
  /* specified rings: e.g. R::=Q[xy],Lex; Use R,DegLex;                 */
  /* Here we check for such incorrect use and trigger an error if found.*/
  /* For some mysterious reason this test must be done before calling   */
  /* eval_ring... yet another of Antonio's little surprises (sigh).     */
  if (cterm_op(t_ring)==op_VAR && cterm_arity(t_mod) > 0)
    return cocoa_error("RING_ILLEGAL_MODIFIER");

  res_R=eval_ring(t_ring,R);
  if (math_error!=object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    return res_o;
  }
  if (ring_indets_num(res_R) == 0) return cocoa_error("RING_NO_INDETS");

  apply_ring_modifiers(t_mod,res_R,R);

  if (math_error!=object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    ring_unlink(res_R);
    return res_o;
  }
  if (!ring_coeff_in_field(res_R) && 
      !flag_is_on(opt_panel_ENGINE,flag_ENGINE_WARNINGS)) {
    ca_warning_string("Coeffs are not in a field\n-- GBasis-related computations could fail to terminate or be wrong");
    engine_made_output = TRUE; }
  return ring2object(res_R);
}

ring eval_ring(cterm t,ring R) 
/* evaluation of a ring expression */
{
  ring res_R=ring_NULL;
  coop op=cterm_op(t);
  if (op==op_VAR) res_R=eval_ring_var(t,R);
  else if (op==op_RING_ZP) res_R=eval_ring_zp(t,R);
  else if (op==op_POLY_RING) res_R=eval_poly_ring(t,R);
  /* else if (op==op_CURR_RING) res_R=current_ring; */
  else math_error=cocoa_error("BAD_RING");
  return res_R;
}

unknown eval_indet_nth(cterm t, ring R)
{ /* Indet(N) */
  unknown res_u;
  cterm_list params;

  if (cterm_arity(t)!=1) {
    math_error=cocoa_error("BAD_PARAMS"); 
    return unknown_NULL; }

  params=cterm_params(t);
  res_u=eval_int(params[0],R);
  if (math_error!=object_NULL) return unknown_NULL;
  if (res_u<=0 || res_u>R->indetsNo) {
    char buffer[10];
    sprintf(buffer,"#%d",res_u);
    math_error=cocoa_error("UNDEF_INDET");
    error_set_this_param(cterm_mk_NBS(buffer));
    return unknown_NULL;
  }
  return res_u-1;
}

ring eval_ring_var(cterm t,ring R)
{ 
  ring aux_R,res_R;
  char * id;
  cterm_list params;
  id=eval_var_name(t,R);
  if (math_error != object_NULL) return NULL;
  aux_R=ring_env_get(id);
  ring_unlink(aux_R);
  free_str1(id);
  if (aux_R==ring_NULL) {
    math_error=cocoa_error("UNDEF_RING");
    return ring_NULL; }
  res_R=aux_R;
  return res_R;
}

coc_bool eval_bool(cterm t,ring R)
{
  coop op=cterm_op(t);
  cterm_list params=cterm_params(t);
  coc_bool res_b=FALSE; /* a default */
  if (op==op_TRUE) res_b=TRUE;
  else if (op==op_FALSE) res_b=FALSE;
  else if (op==op_AND) res_b=mi_and(params[0],params[1],R);
  else if (op==op_OR) res_b=mi_or(params[0],params[1],R);
  else if (op==op_NOT) res_b=mi_not(params[0],R);
  else if (op==op_IS_EQ) res_b=mi_is_eq(params[0],params[1],R);
  else if (op==op_IS_LT) res_b=mi_is_lt(params[0],params[1],R);
  else if (op==op_IS_GT) res_b=mi_is_lt(params[1],params[0],R);
  else if (op==op_IS_LE) res_b=mi_is_le(params[0],params[1],R);
  else if (op==op_IS_GE) res_b=mi_is_le(params[1],params[0],R);
  else if (op==op_IS_IN) res_b=mi_is_in(params[0],params[1],R);
  else if (op==op_ISDEF) res_b=mi_is_defined(params[0],R);
  else if (op==op_PARENTHESIS) res_b=eval_bool(params[0],R);
  else if (op==op_VAR) res_b=eval_bool_var(t,R);
  else if (op==op_FUN_APPL) {
    object aux_o;
    aux_o = mi_fun_appl(t,R);
    if (object_is_error(aux_o))
      math_error=aux_o;
    else
      res_b = object2bool(aux_o); }
  else if (op==op_COMPONENT)  
    res_b=object2bool(mi_component(t,R));
  else if (op==op_RECORDSEL)  
    res_b=object2bool(mi_record_sel(t,R));
  else if (is_user_defined_coop(op)) {
    object aux_o;
    aux_o = mi_user_function(t,R);
    if (object_is_error(aux_o))
      math_error=aux_o;
    else
      res_b = object2bool(aux_o); }
  else {
    object aux_o;
    aux_o = eval(t,R);
    if (object_is_error(aux_o))
      math_error=aux_o;
    else
      res_b = object_cast_bool(aux_o); }
  return res_b;
}

coc_bool mi_is_defined(cterm t, ring R)
{ /* true if the expression t is correct */
  /* useful to check if a variable is assigned */
  object aux_o;
  coc_bool res_b;
  aux_o=eval(t,R);
  res_b=!object_is_error(aux_o);
  object_unlink(aux_o);
  return res_b;
}

coc_bool mi_and(cterm t1, cterm t2, ring R)
{
  coc_bool b1, b2;
  b1=eval_bool(t1,R);
  if (math_error!=object_NULL) return FALSE;
  if (b1==FALSE) return FALSE;
  b2=eval_bool(t2,R);
  if (math_error!=object_NULL) return FALSE;
  return b2;
}

coc_bool mi_not(cterm t, ring R)
{
  coc_bool b;
  b=eval_bool(t,R);
  if (math_error!=object_NULL) return FALSE;
  return !b;
}

coc_bool mi_or(cterm t1, cterm t2, ring R)
{
  coc_bool b1, b2;
  b1=eval_bool(t1,R);
  if (math_error!=object_NULL) return FALSE;
  if (b1==TRUE) return TRUE;
  b2=eval_bool(t2,R);
  if (math_error!=object_NULL) return FALSE;
  return b2;
}

unknown eval_indet(cterm t,ring R)
{ 
  object aux_o;
  aux_o=eval(t,R);
  if (object_is_error(aux_o)) {
    math_error=aux_o;
    return unknown_NULL; }
  else return object2indet(aux_o);
}

object mi_user_function_aux(cterm vars, cterm body, olist args, 
                            cterm actual_par, ring R)
{ /* if actual_par == cterm_NULL then it is CALL BY VALUE */
  object res_o;
  env new_env,save_current_env;
  
  funcall_level++;
  new_env=env_init();  
  if (args != olist_NULL) {
    assign_parameter_variables(vars,args,new_env,R);
  } else {
    env_copy_var(current_env,&new_env,"ARGV","ARGV");
  }

  if (math_error != object_NULL) {
    res_o=math_error;
    math_error=object_NULL;
    funcall_level--;
    env_free(new_env);
    return res_o;
  } /* TODO: CHECK MEMORY */
  
  save_current_env=current_env;
  current_env=new_env;

  if (ring_locked(R)) {
    res_o=engine_in_ring(body,R);
  } else {
    ring_set_locked(R,TRUE);
    res_o=engine_in_ring(body,R);
    ring_set_locked(R,FALSE);
  }
  
  current_env=save_current_env;
  if (actual_par != cterm_NULL) 
    update_out_variables(vars,actual_par,args,new_env,R);
  olist_free(args);
  /* riprendere parametri formali non VAR */
  if (!object_is_error(res_o)) {
    if (math_error != object_NULL) res_o=math_error; }
  if (res_o == object_BREAK) {
    object_unlink(res_o);
    res_o=cocoa_error("BREAK_WITHOUT_LOOP");
  }
  env_free(new_env);
  funcall_level--;
  return res_o;
}

object mi_user_function(cterm t,ring R)
{ /* t=F(X1,...,Xn), Xi expressions */
  /* t=Call(F,X1,...,Xn), Xi expressions */
  int pkg_n;
  coop op;
  olist aux_L;
  cterm handler,vars,body;
  cterm_list params;

  op=cterm_op(t);
  pkg_n=op_package(op);
  op=op_index(op);
  handler=(cterm)op_map_op_handler(pkg_n,op);
  /* handler=handler(var_list,body) */
  params=cterm_params(handler);
  vars=params[0];
  body=params[1];
  params=cterm_params(t);
  if (cterm_arity(t) == 1 && cterm_op(params[0]) == op_DOTS) aux_L = olist_NULL;
  else aux_L=eval_params(t,vars,R);
  if (math_error != object_NULL) return math_error;
  return mi_user_function_aux(vars,body,aux_L,t,R);
}

olist eval_params(cterm t, cterm vars, ring R)
{
  int i,arity;
  olist res_ol;
  object aux_o;
  coop aux_op;
  arity=cterm_arity(t);
  res_ol=olist_init(arity);
  for (i=0;i<arity;i++) {
    aux_op=cterm_op(cterm_param(t,i));
    aux_o=eval(cterm_param(t,i),R);
    if (object_type(aux_o) == objtype_ring) { math_error = cocoa_error("ARG_IS_RING"); return res_ol; }
    if (object_is_error(aux_o)) {
#if 0 /* TODO: decidere cosa e` meglio */
      if (cterm_op(vars) != op_DOTS && cterm_op(cterm_param(vars,i)) != op_VAR) {
        math_error=aux_o;
        return res_ol;
      }
    /* NOTE: if aux_o is an error and the corresponding param is VAR
             it will be notified only if the corresponding
             variable will be read-before-writing
             during the evaluation of the body */
#else
      math_error=aux_o;
      return res_ol;
#endif
    }
    olist_set_nth(res_ol,i+1,aux_o); }
  return res_ol;
}

void update_out_variables(cterm vars, cterm args, olist ivalues, env proc_env, ring R)
{
  int i;
  char * id;
  object aux_o,aux_o1;
  cterm_list params_vars,params_args;

  params_args=cterm_params(args);
  params_vars=cterm_params(vars);

  if (cterm_op(vars)==op_DOTS) return;

  for (i=0;i<cterm_arity(vars);i++)
    if (cterm_op(params_vars[i])==op_VAR) {
      id=eval_var_name(params_vars[i],R);
      aux_o1 = env_get_value(proc_env, id);
      free_str1(id);
      if (olist_nth(ivalues,i+1) != aux_o1) {
        aux_o=engine_assign_this(params_args[i], aux_o1, R);
        if (aux_o != object_null) { 
          if (object_type(aux_o) != objtype_err || 
              strcmp(aux_o->v.err_value,"Bad left expression") != 0) {
            math_error=aux_o; 
            return; 
          }
        }
      } 
      else object_unlink(aux_o1);
    }
}

coc_bool assign_parameter_variables(cterm vars,olist ol,env new_env,ring R)
/* evaluates the arguments args and assign to the variables vars of R */
{ 
  char message[200];
  char * id;
  int arity=cterm_arity(vars);
  coc_bool status=TRUE; /* default is ok */

  if (cterm_op(vars)==op_DOTS) {
    object aux_o;
    aux_o=olist2object(olist_dup(ol));
    env_declare_and_assign(&new_env,"ARGV",aux_o);
    object_unlink(aux_o);
    return TRUE; }

  if (arity!=olist_len(ol)) {
    math_error=cocoa_error("BAD_PARAMS_NUM");
    return FALSE;}
  else {
    int i;
    for (i=0;i<arity;i++) {
      id=eval_param_name(cterm_param(vars,i),R);
      env_declare_and_assign(&new_env,id,olist_nth(ol,i+1));
      free_str1(id); }
    }
  return status;
}

unknown mi_cterm2unknown_bis(cterm t, ring R)
{ /* t = Indet(S:NBS) */
  /* t = Indet(S:NBS, Index1, ..., IndexN) */
  /* t = Indet_Nth(n) */

  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  unknown res_u;
  char * aux_s;
  coop op=cterm_op(t);
  ivector index;
  register int i;

  if (arity==0) {
    math_error=cocoa_error("EXPECTED_INDET"); 
    error_set_context(t);
    return unknown_NULL; }

  if (!is_polynomial_ring(R)) {
    math_error=cocoa_error("NON_POLY_RING");
    return unknown_NULL; }

  if (op==op_INDET_NTH) 
    return eval_indet_nth(t,R);

  if (op!=op_INDET) {
    math_error=cocoa_error("EXPECTED_INDET");
    error_set_context(t);
    return unknown_NULL; }

  aux_s=cterm_string(params[0]);
  index=iv_init(arity-1);
  for (i=1;i<arity;i++) {
    iv_set_nth(index,i,eval_int(params[i],R));
    if (math_error!=object_NULL) {
      iv_free(index);
      return unknown_NULL; } }
  res_u=indet_info_get_unknown(R->indets,aux_s,index);
  iv_free(index);
  return res_u;
}

unknown mi_cterm2unknown(cterm t, ring R)
{
  unknown res_u;
  res_u = mi_cterm2unknown_bis(t,R);
  if (res_u == unknown_NULL) {
    math_error=cocoa_error("UNDEF_INDET");
    error_set_param(t);
    return unknown_NULL;
  }
  return res_u;
}

int eval_int_var(cterm t,ring R)
{ 
  loc l;
  object aux_o;
  l = eval_left_expr_not_declare(t,R);
  if (l!=NULL) {
    aux_o = mem_value(l);
    if (aux_o->t==objtype_cterm)
      return eval_int(aux_o->v.cterm_value,R);
    else
      return object2int(aux_o); }
   
  else {
    math_error=cocoa_error("UNDEF_VAR");
    error_set_param(t);
    return ZERO; }
}

coc_bool eval_bool_var(cterm t,ring R)
{ 
  loc l;
  object aux_o;
  l = eval_left_expr_not_declare(t,R);
  if (l!=NULL) {
    aux_o = mem_value(l);
    if (aux_o->t==objtype_cterm)
      return eval_bool(aux_o->v.cterm_value,R);
    else {
      if (object_type(aux_o) != objtype_bool) {
        math_error=cocoa_error("EXPECTED_BOOL");
	object_unlink(aux_o);
	return FALSE;
      } else {
        return object2bool(aux_o);
      }
    }
  } else {
    math_error=cocoa_error("UNDEF_VAR");
    return FALSE; 
  }
}

#if 0

object mi_gauss(cterm t,ring R)
{ 
  object res_o;
  cterm_list params=cterm_params(t);
  res_o=eval(params[0],R);
  do_gauss(res_o);
  return res_o;
}

object mi_float(cterm t,ring R)
{
  object res_o;
  res_o=object_init(objtype_double,NULL,R);
  SdStr2Coeff((coeff *)&res_o->v.double_value,cterm_string(t),R);
  return res_o;
}

#endif

ring eval_ring_zp(cterm t,ring R)
{ 
  /* t=zp(e:expr) */
  int n;
  cterm_list params=cterm_params(t);
  if (cterm_arity(t)==0) return eval_coeff_ring(COCOA_DEFAULT_ZMOD);
  n=eval_int(params[0],R);
  if (n<2 || n>ZP_MAX) { math_error=cocoa_error("BAD_CHAR"); return ring_NULL; }
  if (math_error!=object_NULL) return ring_NULL;
  return eval_coeff_ring((short int)n);
}

ring eval_coeff_ring(short int c)
{ 
  ring res_R;
  res_R=ring_of_coeff(NULL /* no name (set by engine) */,c);
  return res_R;
}

indet_info eval_indet_info(cterm t, ring R)
{ 
  /* for now t=list(i1:indet_decl,..,in:indet_decl) */
  register int i;
  indet_entry aux_e;
  short int num_indet;
  indet_info res_ii;
  cterm * params = cterm_params(t);
  res_ii=indet_info_init();
  num_indet=cterm_arity(t);
  for (i=0;i<num_indet;i++) {
    aux_e=eval_indet_entry(params[i],R);
    if (math_error!=object_NULL) {
      indet_info_free(res_ii);
      return NULL; 
    }
    if (indet_info_search_entry_index(res_ii,aux_e->ident) != -1) {
      error_set_this_param(string2cterm(aux_e->ident));
      math_error=cocoa_error("ALREADY_DEF_INDET");
      indet_info_free(res_ii);
      return NULL;
    }
    indet_info_add(res_ii,aux_e);
    if (aux_e->size == 0) { math_error=cocoa_error("TOO_MANY_INDETS"); indet_info_free(res_ii); return NULL; }
 }
  return res_ii;
}

void eval_range(cterm t,int * x,int * y,ring R)
{ /* t=range(i1:integer,i2:integer)) */
  cterm_list params=cterm_params(t);
  *x = eval_int(params[0],R);
  *y = eval_int(params[1],R);
}

indet_entry eval_indet_entry(cterm t,ring R)
{
  /* t = indet_decl("x",r1:range,...,rn:range) */
  cterm_list params = cterm_params(t);
  int arity = cterm_arity(t);
  int i,x,y;
  ivector arrinfo,from;
  arrinfo=iv_init(arity-1);
  from=iv_init(arity-1);
  for (i=1;i<arity;i++) {
    if (cterm_op(params[i]) == op_RANGE) {
      eval_range(params[i],&x,&y,R);
      if (x>y) {
        error_set_param(params[i]);
        math_error=cocoa_error("BAD_RANGE");
      }
    } else {
      math_error = cocoa_error("SYNTAX"); 
      /* x=1; y=eval_int(params[i],R); */
    }
    if (math_error!=object_NULL) {
      iv_free(from);
      iv_free(arrinfo);
      return NULL; }
    iv_put_nth(from,i,x);
    iv_put_nth(arrinfo,i,y-x+1); 
  }
  return indet_entry_mk(my_strdup(cterm_string(params[0])),arrinfo,from);
}

ring eval_ring_constr(cterm t, ring R)
/* return the 'constrained' ring described in the cterm t */
/* return NULL if the ring is a coefficient ring */
/* return NULL even in case of error */
/* the term t must contain the name of a ring already existent (for now) */
{ 
  ring res_R;
  coop op=cterm_op(t);
  char * id;
  if (op==op_VAR) {
    char * aux_s;
    aux_s=cterm_string(cterm_param(t,0));
    if (strcmp(aux_s,"CoeffRing")==0) {
      return ring_link(ring_coeffs(R));
    } else {
      res_R=ring_env_get(aux_s);
      ring_unlink(res_R);
    }
  }
  else if (op==op_NBS) {
    char * aux_s;
    aux_s=cterm_string(t);
    if (strcmp(aux_s,"CoeffRing")==0) {
      res_R = ring_link(ring_coeffs(R));
    } else {
      res_R=ring_env_get(aux_s);
    }
  }
  else if (op==op_RING_ZP) return ring_NULL;

  if (res_R==ring_NULL) {
    error_set_param(t);
    math_error=cocoa_error("UNDEF_RING");
    return ring_NULL; }
  if (is_polynomial_ring(res_R)) {
    error_set_param(t);
    math_error=cocoa_error("BAD_BASERING");
    return ring_NULL; }
  return res_R;
}

short int eval_characteristic(cterm t,ring R)
/* return the characteristic of the 'constrained' ring described in t */
{ 
  ring aux_R;
  coop op=cterm_op(t);
  int characteristic;
  
  if (op==op_NBS) {
    char * aux_s;
    aux_s=cterm_string(t);
    if (strcmp(aux_s,"CoeffRing") == 0) return ring_characteristic(R);
    else {
      aux_R=ring_env_get(aux_s);
      ring_unlink(aux_R);
    }
    if (aux_R==ring_NULL) characteristic=characteristic_ERROR;
    else characteristic=aux_R->characteristic; }
  if (op==op_VAR) {
    char * aux_s;
    aux_s=cterm_string(cterm_param(t,0));
    if (strcmp(aux_s,"CoeffRing") == 0) aux_R=R;
    else {
      aux_R=ring_env_get(aux_s);
      ring_unlink(aux_R);
    }
    if (aux_R==ring_NULL) characteristic=characteristic_ERROR;
    else characteristic=aux_R->characteristic; }
  else if (op==op_RING_ZP) {
    cterm_list params=cterm_params(t);
    if (cterm_arity(t)==0) characteristic=COCOA_DEFAULT_ZMOD;
    else characteristic=eval_int(params[0],R);
    if (math_error!=object_NULL) return characteristic;
    if (characteristic > ZP_MAX) {
      error_set_this_param(int2cterm(characteristic));
      math_error=cocoa_error("BAD_CHAR");
    }
}
  return characteristic;
}

#if 0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
13 April 1999 "oldstyle" has been eliminated...
void set_ordering_user_def_oldstyle(ordering * to, cterm t, ring poly_R, ring R)
{
  int counter, index;
  int n,aux_i;
  register int i,j;
  cterm_list params;
  n=ring_indets_num(poly_R);
  params=cterm_params(t);
  if (cterm_arity(t)!=n*n) {
    math_error=cocoa_error("BAD_PARAMS_NUM");
    return; }
  for (i=0;i<=n-1;i++) {
    counter = 0;
    for (j=0;j<=n-1;j++) {
      aux_i = eval_int(params[i*n+j],R);
      if (math_error!=object_NULL) return;
      if (aux_i!=0) {
        counter++; 
        index=j;
        (*to)[i][j]=(toElem)aux_i; } }
      if (counter == 1) (*to)[i][n]=(toElem)index; }
}
#endif

void set_ordering_user_def(ordering * to, cterm t, ring poly_R, ring R)
{
  object aux_o;
  omat aux_M;
  int n;
  if (cterm_arity(t) != 1) {
/*    set_ordering_user_def_oldstyle(to,t,poly_R,R);*/
    math_error=cocoa_error("BAD_PARAMS_NUM");
    return; }
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) {
    math_error=aux_o;
    return;
  }
  aux_M=object_cast_omat(aux_o);
  if (math_error!=object_NULL) return;
  n=ring_indets_num(poly_R);
  if (n!=omat_nrow(aux_M) || n!=omat_ncol(aux_M)) {
    char buffer[100];
    math_error=cocoa_error( "EXPECTED" );
    sprintf(buffer,"a %d x %d matrix (found %d x %d)",n,n,omat_nrow(aux_M),
      omat_ncol(aux_M));
    error_set_this_param(cterm_mk_NBS(buffer));
    return; }
  omat2ord(to,aux_M);
  omat_free(aux_M);
}

void set_ordering(cterm t, ring poly_R, ring R)
/* return the ordering described in t for the polynomial ring poly_R */
{ 
  ordering to=ordering_NULL;
  short int whatTO;
  unknown aux_u1=unknown_NULL,aux_u2=unknown_NULL;
  coop op=cterm_op(t);
  cterm_list params=cterm_params(t);
  char * op_name;
  int aux_i;
  int arity=cterm_arity(t);
  int n = poly_R->indetsNo;
#ifdef TRIPOLI_OR_EEE
  InitZeroOrder(&to,poly_R);
#endif
  /* TODO: CHECK int errors */
  op_name=(char *)op_map_op_name(op_package(op), op_index(op));
  if (op==op_LEX) {
#ifdef TRIPOLI_OR_EEE
    SetLexOrder(to,poly_R);
#endif
    whatTO=lex_num; }
  else if (op==op_XEL) {
#ifdef TRIPOLI_OR_EEE
    SetXelOrder(to,poly_R);
#endif
    whatTO=xel_num; }
  else if (op==op_DEGLEX) {
#ifdef TRIPOLI_OR_EEE
    SetDegLexOrder(to,poly_R->weights,poly_R);
#endif
    whatTO=dlex_num; }
  else if (op==op_DEGREVLEX) {
#ifdef TRIPOLI_OR_EEE
    SetDegRevLexOrder(to,poly_R->weights,poly_R);
#endif
    whatTO=drev_num; }
  else if (strcmp(op_name,"Elim")==0) {
    if (arity != 1) {
      math_error=cocoa_error("BAD_PARAMS_NUM");
      error_set_this_param(cterm_mk_NBS("for operator Elim"));
      return; }
    mi_indets_range(params[0],&aux_u1,&aux_u2,poly_R);
    if (math_error != object_NULL) return;
    InitZeroOrder(&to,poly_R);
    SetElimOrder(aux_u1,aux_u2,to,poly_R->weights,poly_R);
    whatTO=elim_num; }
  else if (op==op_ORD) { 
    InitZeroOrder(&to,poly_R);
    set_ordering_user_def(&to,t,poly_R,R);
    whatTO = user_defined_ordering_num;
  }
  else math_error=cocoa_error("UNKN_ORDERING");
  if (math_error!=object_NULL) return;
#ifdef TRIPOLI_OR_EEE
  ring_free_ordering(poly_R);
#endif
  ring_set_ordering(poly_R,to,whatTO,aux_u1,aux_u2);
}

void set_weights_oldstyle(cterm t, ring poly_R, ring R)
{ /* t = weights(w1,...,wn) or t = weights() */
  int arity, n, i,aux_i;
  weilist wl;
  cterm_list params;
  weight w;
  arity = cterm_arity(t);
  params = cterm_params(t);
  n = ring_indets_num(poly_R);
  if (n!=arity) {
    math_error=cocoa_error( "BAD_PARAMS_NUM" );
    error_set_context(t);
    return; }
  w = iv_init(n);
  for (i=0;i<n;i++)
  {
    aux_i=eval_int(params[i],R);
    if (math_error!=object_NULL)
    {
      iv_free(w);
      return;
    }
    if (aux_i < 1)
    {
      math_error = cocoa_error("BAD_WEIGHTS");
      iv_free(w);
      return;
    }
    iv_set_nth(w,i+1,aux_i);
  }
  weight_free(ring_weights(poly_R));
  ring_set_weights(poly_R,w);
  wl = weilist_init(1);
  weilist_set_nth(wl,1,weight_dup(w));
  weilist_free(ring_weilist(poly_R));
  ring_set_weilist(poly_R,wl);
}

void set_weights_from_omat(omat aux_M, ring poly_R) {
  object aux_o;
  olist aux_L;
  aux_o=olist_nth(aux_M,1);
  aux_L=aux_o->v.olist_value;
  weight_free(ring_weights(poly_R));
  weilist_free(ring_weilist(poly_R));
  ring_set_weights(poly_R, olist2weight(aux_L, poly_R));
  if (math_error != object_NULL) { ring_set_weilist(poly_R, weilist_NULL); return; }
  ring_set_weilist(poly_R, olist2weilist(aux_M, poly_R));
}

void set_weights(cterm t, ring poly_R, ring R)
{ /* WEIGHTS(LIST) */
  /* WEIGHTS(W1:INT,...,Wn:INT) */
  /* WEIGHTS(MAT) (multigraded) */
  object aux_o;
  olist aux_L;
  omat aux_M;
  int i,aux_i;
  weight w;
  if (cterm_arity(t) != 1) {
    set_weights_oldstyle(t,poly_R,R);
    return; }
  aux_o=eval(cterm_param(t,0),R);
  if (object_is_error(aux_o)) {
    math_error=aux_o;
    return;
  }
  if (object_type(aux_o) == objtype_olist) {
    aux_L=object2olist(aux_o);
    aux_o=olist2object(aux_L);
    aux_L=olist_init(1);
    olist_set_nth(aux_L,1,aux_o);
    set_weights_from_omat(aux_L,poly_R);
    if (math_error != object_NULL) error_set_context(t);
    olist_free(aux_L);
  } else if (object_type(aux_o) == objtype_omat) {
    aux_M=object2omat(aux_o);
    set_weights_from_omat(aux_M,poly_R);
    if (math_error != object_NULL) error_set_context(t);
    omat_free(aux_M);
  } else {
    object_unlink(aux_o);
    set_weights_oldstyle(t,poly_R,R);
  }
}

ring eval_poly_ring(cterm t,ring R)
{ /* t=poly_ring(i:indet_list, coeff:ring_expr_constr) */
  ring res_R, C;
  weight w;
  weilist wl;
  short int characteristic;
  ordering to=ordering_NULL;
  indet_info ii;
  cterm * params=cterm_params(t);
  ii=eval_indet_info(params[0],R);
  if (math_error!=object_NULL) return ring_NULL;
  C=eval_ring_constr(params[1],R);
  if (math_error!=object_NULL) { indet_info_free(ii); return ring_NULL; }
  characteristic=eval_characteristic(params[1],R);
  if (math_error!=object_NULL) { indet_info_free(ii); ring_free(C); return ring_NULL; }
  res_R=ring_init(NULL /* no name (set by engine) */,
    ringtype_poly,characteristic,C,ii,to_NULL,ZERO,TO_POS);
#ifdef TRIPOLY_OR_EEE
  InitZeroOrder(&to,res_R);
  SetDegRevLexOrder(to,res_R->weights,res_R);
#endif
  ring_set_ordering(res_R,to,drev_num,unknown_NULL,unknown_NULL);
  w = iv_create_and_fill(res_R->indetsNo,1);
  ring_set_weights(res_R,w);
  wl = weilist_init(1);
  weilist_set_nth(wl,1,weight_dup(w));
  ring_set_weilist(res_R,wl);
  return res_R;
}

void apply_ring_modifiers(cterm t, ring New_R, ring R) 
{ /* Ring_Modifiers(A1,...,An:PolyRingAnn) */
  int arity=cterm_arity(t);
  int i;
  cterm_list params=cterm_params(t);
  for (i=0;i<arity;i++) {
    apply_ring_modifier(params[i],New_R,R);
    if (math_error != object_NULL) return; }
}

void apply_ring_modifier(cterm t, ring New_R, ring R)
{
  coop op;
  coc_bool free_t;
  free_t=FALSE;
  if (cterm_op(t)==op_FUN_APPL) {
    if (cterm_arity(t)==2 && strcmp(cterm_string(cterm_param(t,0)),"Elim")==0) {
      t=cterm_mk_classic(op_ELIM,cterm_arity(cterm_param(t,1)),
          cterm_params(cterm_param(t,1))); /* MARK-10 */
      free_t=TRUE;
    } else {
      t = mi_expand_fun_appl(t,R);
    }
  }
  if (t==cterm_NULL) { math_error=cocoa_error("BAD_PARAMS"); return; }
  op = cterm_op(t);

  if (op==op_WEIGHTS)
    set_weights(t,New_R,R);
  else if (op==op_ORD || op==op_LEX || op==op_DEGLEX || op==op_DEGREVLEX ||
           op==op_XEL || op==op_ELIM) {
    set_ordering(t,New_R,R);
    if (math_error != object_NULL) return;
  }
  else if (op==op_POS_TO)
    ring_set_mordering(New_R,POS_TO);
  else if (op==op_TO_POS)
    ring_set_mordering(New_R,TO_POS);
  else if (op==op_LEVEL) {
    int aux_n;
    aux_n=eval_int(cterm_param(t,0),R);
    if ((math_error == object_NULL) && 
        (aux_n < 0 || aux_n > ring_indets_num(New_R))) {
      error_set_context(t);
      error_set_this_param(int2cterm(aux_n));
      math_error = cocoa_error("WRONG_LEVEL"); }
    if (math_error == object_NULL) 
#ifdef EEE
      ring_set_level(New_R,ring_indets_num(New_R)-aux_n); }
#else /* TRIPOLI */
      ring_set_level(New_R,aux_n); }
#endif
  else math_error=cocoa_error("UNKN_RING_ANN");
  if (free_t) {
    free_cterm_list(cterm_arity(t),cterm_params(t));
    free_cterm(t);
  }
}

/* TODO: rimuovere questa e chiamare solo mi_whereis_var */
loc eval_left_expr(cterm t, coc_bool declare_if_not, ring R)
{ /* t=var("X"), t=Var('X') */
  loc res_l;
  env aux_e;
  char * id;
  id=eval_var_name(t,R);
  if (math_error != object_NULL) return NULL;
  mi_whereis_var(cterm_op(t), id, &aux_e, &res_l, R);
  if (math_error != object_NULL) return NULL;
  if (res_l==NULL && declare_if_not) {
    aux_e=current_env;
    res_l=env_declare(&aux_e,id);
  }
  free_str1(id);
  return res_l;
}

void mi_whereis_var(coop op, char * id, env * e, loc * l, ring R)
{
  /* op = VAR */
  /* if var is found *l is the location and *e is the env */

  if (strcmp(id,"MEMORY") == 0) {
    *e=memory_env;
    *l=env_get_loc(*e,id);
    return; }

  if (op != op_VAR) {
    math_error=cocoa_error("BAD_LEFT_EXPR");
    return; }

  *l=env_get_loc(current_env,id);
  if (*l != NULL) return;

  if (is_special_var(id)) {
    *e=cocoa_env;
    *l=env_get_loc(*e,id);
  }
  if (*l != NULL) return;
  *l=env_get_loc(ring_env,id);
}

char * eval_var_name(cterm t, ring R)
{
  /* VAR("V") or VAR(S:STRING) or "V" */
  char * res_id;
  object aux_o;
  cterm_list params;
  coop op=cterm_op(t);
  params=cterm_params(t);
  if (op != op_VAR) {
    math_error=cocoa_error("EXPECTED_VAR");
    return NULL; }
  else if (cterm_op(params[0])==op_NBS) 
    res_id = my_strdup(cterm_string(params[0]));
  else { /* t=VAR(S:STRING) espressione di tipo STRING */
    aux_o=eval(params[0],R);
    if (object_is_error(aux_o)) {
      math_error=aux_o;
      return NULL; }
    res_id = object2string(aux_o);
  }
  return res_id;
}

char * eval_param_name(cterm t, ring R)
{
  /* "V" or VAR("V") */
  if (cterm_op(t)==op_NBS) return my_strdup(cterm_string(t));
  else if (cterm_op(t)==op_DOTS) return my_strdup("ARGV");
  else return my_strdup(cterm_string(cterm_param(t,0)));
}

object mi_params(cterm t, ring R)
/* return [ eval(params[0]), ..., eval(params[arity-1])] */
{
  int arity=cterm_arity(t);
  object aux_o;
/*  if (arity==1) {
    cterm_list params;
    params=cterm_params(t);
    return eval(params[0],R); }
  else */
    return mi_list(t,R);
}

object mi_params1(cterm t, ring R)
/* if arity(t)==1 return eval(params[0]) else return the list
   [ eval(params[0]), ..., eval(params[arity-1])] */
{
  int arity=cterm_arity(t);
  object aux_o;
  if (arity==1) {
    cterm_list params;
    params=cterm_params(t);
    return eval(params[0],R); }
  else
    return mi_list(t,R);
}

int cterm2int(cterm t)
{
  /* t=number */
  return atoi(cterm_string(t));
}


/* boolean evaluations */

coc_bool mi_is_eq(cterm t1,cterm t2,ring R)
{ 
  object o1,o2;
  coc_bool res_b;
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    math_error=o1;
    return FALSE; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    math_error=o2;
    object_unlink(o1);
    return FALSE; }
  res_b=do_is_eq(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_b;
}

coc_bool mi_is_lt(cterm t1,cterm t2,ring R)
{ 
  object o1,o2;
  coc_bool res_b;
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    math_error=o1;
    return FALSE; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    math_error=o2;
    object_unlink(o1);
    return FALSE; }
  res_b=do_is_lt(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_b;
}

coc_bool mi_is_in(cterm t1,cterm t2,ring R)
{ 
  object o1,o2;
  coc_bool res_b;
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    math_error=o1;
    return FALSE; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    math_error=o2;
    object_unlink(o1);
    return FALSE; }
  res_b=do_is_in(o1,o2);
  object_unlink(o1);
  object_unlink(o2);
  return res_b;
}

coc_bool mi_is_le(cterm t1,cterm t2,ring R)
{ 
  object o1,o2;
  coc_bool b1, b2;
  o1=eval(t1,R);
  if (object_is_error(o1)) {
    math_error=o1;
    return FALSE; }
  o2=eval(t2,R);
  if (object_is_error(o2)) {
    math_error=o2;
    object_unlink(o1);
    return FALSE; }
  b1=do_is_lt(o1,o2);
  if (math_error!=object_NULL) {
    object_unlink(o1);
    object_unlink(o2);
    return FALSE; }
  if (object_type(o1) == objtype_module) return b1;
  b2=do_is_eq(o1,o2);
  if (math_error!=object_NULL) {
    object_unlink(o1);
    object_unlink(o2);
    return FALSE; }
  object_unlink(o1);
  object_unlink(o2);
  return b1 || b2;
}

void mi_indets_range(cterm t,unknown * u1,unknown * u2,ring R)
{
  /* t=indet(x:indet) (with x not indexed)
     t=range(x1:indet,x2:indet)
     t=indet(x:indet_array,i1:int,...,in:int);
     t=indet(x:indet_array); (with x indexed, meaning: all x)
     t=Indet(n:number) */
  coop op;
  cterm_list params;
  int arity;
  char * aux_s;
  op=cterm_op(t);
  arity=cterm_arity(t);
  params=cterm_params(t);
  if (op==op_RANGE) {
    *u1=eval_indet(params[0],R);
    if (math_error!=object_NULL) return;
    *u2=eval_indet(params[1],R);
    if (math_error!=object_NULL) return;
    if (*u1>*u2) {
      error_set_param(t);
      math_error=cocoa_error("BAD_RANGE");
      return;
    }
    return; }
  if (arity>1 || op!=op_INDET)
    *u1=*u2=eval_indet(t,R);
  else {
    params=cterm_params(t);
    aux_s=cterm_string(params[0]);
    indet_info_first_and_last(R->indets,aux_s,u1,u2); }
}
