/* $Id: math_lib.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"
#include <limits.h>
#include <string.h>

#include "ca.h"
#include "coeffs.h"
#include "cterm.h"
#include "ck.h"
#include "evector.h"
#include "engine.h"
#include "hilop.h"
#include "ivectors.h"
#include "module.h"
#include "memoryco.h"
#include "math_lib.h"
#include "max.h"
#include "monomial.h"
#include "module.h"
#include "modupoin.h"
#include "object.h"
#include "olist.h"
#include "omat.h"
#include "options.h"
#include "operator.h"
#include "polys.h"
#include "pvector.h"
#include "pseries.h"
#include "poincare.h"
#include "record.h"
#include "ring.h"
#include "subst.h"
#include "ratfun.h"
#include "terms.h"
#include "toric.h"
#include "unipoly.h"
#include "vlist.h"
#include "weights.h"
#include "jintf.h"

extern object math_error;

#define object_true_ring(o1,o2) \
  ((object_ring(o1) == ring_NULL) ? object_ring(o2) : object_ring(o1))

/* rimuovere tappul: TODO */

object do_map(object o1, ring R2, object o2)
{ /* o1 to be moved, o2 parameters of the map */
  /* o1 and o2 are returned unlinked */
  arpoly aux_p;
  object res_o;
  olist aux_ol;
  ivector m;
  int n,aux_i,i;
  ring R1;

  if (object_is_ring_indep(o1)) return o1;

  R1=object_ring(o1);
  aux_p=object_cast_arpoly(o1,R1);

  aux_ol=object_cast_olist(o2);
  if (math_error != object_NULL) {
    return math_error;
  }
  n = olist_len(aux_ol);
  m = iv_init(n);
  for (i=1;i<=n;i++) {
    aux_i=object_cast_int(object_link(olist_nth(aux_ol,i)));
    if (math_error!=object_NULL) {
      iv_free(m);
      return math_error; }
    iv_set_nth(m,i,aux_i); }
  res_o = arpoly2object(arpoly_map(aux_p,R1,R2,m),R2);
  arpoly_free(aux_p,R1);
  olist_free(aux_ol);
  iv_free(m);
  return res_o;
}

object do_mat_toric(object o, ring R)
{
  /* o non-negative integer matrix */
  omat M;
  BitermList BL;
  ints Indices, Weights;

  M=object_cast_omat(object_link(o));
  if (math_error != object_NULL)  return math_error;
  StartToric(omat_ncol(M));
  omat_ker2BList(M, &BL, &Indices, &Weights);
  omat_free(M);
  if ( ToricAlg == EATI )
    BL = ElimToric(BL, Indices, Weights);
  if ( ToricAlg == SATI )
    BL = SequentialToric(BL, Indices);
  if ( ToricAlg == PATI )
    BL = HDToric(BL, Indices);
  return olist2object(this_WBList2olist(BL, Weights, R));
}

object do_bin_toric(object o1, object o2, ring R)
{
  /* o1 list of polynomials */
  /* o2 list of indeterminates */
  olist aux_ol;
  BitermList BL;
  ints IL,W;

  aux_ol = object_cast_olist(object_link(o1));
  if (math_error == object_NULL)
  {
    StartToric(R->indetsNo);
    BL = olist2BList(aux_ol,R);
  }
  olist_free(aux_ol);
  if (math_error != object_NULL) return math_error;

  aux_ol = object_cast_olist(object_link(o2));
  if (math_error == object_NULL)
    IL = olist2Ints(aux_ol, R);
  olist_free(aux_ol);
  if (math_error != object_NULL) { EraseAndFreeBList(BL); return math_error; }

  W = ring_weilist2Ints(R);
  if ( ToricAlg == EATI )
    BL = ElimToric(BL, IL, W);
  if ( ToricAlg == SATI )
    BL = SequentialToric(BL, IL);
  if ( ToricAlg == PATI )
    BL = HDToric(BL, IL);
  return olist2object(this_BList2olist(BL, W, R));
}


/* Cannot easily use arbrat2coeff as it destroys its argument... sigh */
static coeff arbrat2zmodp(arbrat q, ring R)
{
  coeff num, den, ans;
  R->InitCoeff(&num);
  R->InitCoeff(&den);
  R->InitCoeff(&ans);
  R->Li2Coeff(&num, mpq_numref((MP_RAT*)q), R);
  R->Li2Coeff(&den, mpq_denref((MP_RAT*)q), R);
  R->DivCoeff(&ans, num, den, R);
  return ans;
}

/* ASSUMES that ring of poly o is Q[xyz...], and R is Z/(P)[xyz...]  */
/* ASSUMES that ring R has at least as many indets as ring of poly o */
object do_QZP(object o, ring R)
{
  monomial monom, iter;
  term PP;
  coeff c;
  int var, *exps;
  arpoly ans, f;
  ring fring = object_ring(o);
  int fvars = fring->indetsNo;  /* We're working in two rings, so be  */
  int NVARS = R->indetsNo;      /* careful with the number of indets. */
  
  exps = (int*)malloc(NVARS*sizeof(int));
  for (var=fvars; var < NVARS; var++) exps[var] = 0;
  ans = NULL;
  f = object_cast_arpoly(object_dup(o), fring);
  for (iter = f; iter != NULL; iter = monomial_next(iter))
  {
    for (var = 0; var < fvars; var++)
      exps[var] = term_get_true_nth(monomial_term(iter), var+1, fring);
    c = arbrat2zmodp(monomial_coeff(iter), R);
    if (!coeff_is_zero(c, R))
    {
      PP = term_init(NVARS);
      for (var = 0; var < NVARS; var++)
        term_put_nth(PP, (var+1), exps[var] * iv_get_nth(R->weights, var+1));
      monom = monomial_mk(c, PP);
      ans = arpoly_sum_these(ans, arpoly_this_monomial2arpoly(monom, R), R);
    }
  }
  free(exps);
  arpoly_free(f, fring);
  return arpoly2object(ans, R);
}


/* ASSUMES that ring of poly o is Z/(P)[xyz...], and R is Q[xyz...] */
object do_ZPQ(object o, ring R)
{
  monomial monom, iter;
  term PP;
  coeff c;
  int var, *exps, NVARS;
  arpoly ans, f;
  ring fring = object_ring(o);

  NVARS = R->indetsNo;
  exps = (int*)malloc(NVARS*sizeof(int));
  ans = NULL;
  f = object_cast_arpoly(object_dup(o), object_ring(o));
  for (iter = f; iter != NULL; iter = monomial_next(iter))
  {
    for (var = 0; var < NVARS; var++)
      exps[var] = term_get_true_nth(monomial_term(iter), var+1, fring);
    c = int2arbrat(coeff2int(monomial_coeff(iter), fring));
    PP = term_init(NVARS);
    for (var = 0; var < NVARS; var++)
      term_put_nth(PP, (var+1), exps[var] * iv_get_nth(R->weights, var+1));
    monom = monomial_mk(c, PP);
    ans = arpoly_sum_these(ans, arpoly_this_monomial2arpoly(monom, R), R);
  }
  free(exps);
  arpoly_free(f, fring);
  return arpoly2object(ans, R);
}


object do_fact(object o)
{
  arbint res_n;
  int aux_n;
  aux_n=object_cast_int(object_link(o));
  if (math_error != object_NULL) return math_error;
  if (aux_n<0) {
    error_set_this_param(cterm_mk_NBS("non-negative INT"));
    return cocoa_error("EXPECTED"); }
  LiInit(&res_n);
  mpz_fac_ui(res_n,aux_n);
  return arbint2object(res_n);
}


object do_isqrt(object o)
{
  arbint res_n;
  arbint aux_n;
  aux_n=object_cast_arbint(object_link(o));
  if (math_error != object_NULL) return math_error;
  if (mpz_sgn((verylongint)aux_n) < 0)
  {
    LiClear(&aux_n);
    error_set_this_param(cterm_mk_NBS("non-negative INT"));
    return cocoa_error("EXPECTED");
  }
  LiInit(&res_n);
  mpz_sqrt(res_n,aux_n);
  LiClear(&aux_n);
  return arbint2object(res_n);
}


object do_iroot(object o1, object o2)
{
  arbint res_n;
  arbint aux_n;
  int aux_i;
  aux_i = object_cast_int(object_link(o2));
  if (math_error != object_NULL) return math_error;
  if (aux_i < 1)
  {
    error_set_this_param(cterm_mk_NBS("positive INT"));
    return cocoa_error("EXPECTED");
  }
  aux_n = object_cast_arbint(object_link(o1));
  if (math_error != object_NULL) return math_error;
  if ((aux_i & 1) == 0 && mpz_sgn((verylongint)aux_n) < 0)
  {
    error_set_this_param(cterm_mk_NBS("non-negative INT"));
    return cocoa_error("EXPECTED");
  }
  LiInit(&res_n);
  mpz_root(res_n, aux_n, aux_i);
  LiClear(&aux_n);
  return arbint2object(res_n);
}

object do_IsPPrime(object o)
{
  arbint aux_n;
  int b;
  aux_n=object_cast_arbint(object_link(o));
  if (math_error != object_NULL) return math_error;
  mpz_abs((verylongint)aux_n, (verylongint)aux_n);
  b = mpz_probab_prime_p(aux_n, 32);
  LiClear(&aux_n);
  if (b) return bool2object(TRUE);
  return bool2object(FALSE);
}

object do_NextPPrime(object o)
{
  arbint res_n;
  arbint aux_n;
  aux_n=object_cast_arbint(object_link(o));
  if (math_error != object_NULL) return math_error;
  if (mpz_sgn((verylongint)aux_n) < 0)
  {
    LiClear(&aux_n);
    error_set_this_param(cterm_mk_NBS("non-negative INT"));
    return cocoa_error("EXPECTED");
  }
  LiInit(&res_n);
  mpz_nextprime(res_n,aux_n);
  LiClear(&aux_n);
  return arbint2object(res_n);
}


object do_PowerMod(object o1, object o2, object o3)
{
  arbint res_n;
  arbint aux_1, aux_2, aux_3;

  aux_1 = object_cast_arbint(object_link(o1));
  if (math_error != object_NULL) return math_error;
  aux_2 = object_cast_arbint(object_link(o2));
  if (math_error != object_NULL) { LiClear(aux_1); return math_error; }
  aux_3 = object_cast_arbint(object_link(o3));
  if (math_error != object_NULL) { LiClear(aux_2); LiClear(aux_1); return math_error; }
  if (mpz_sgn((verylongint)aux_3) == 0)
  {
    LiClear(&aux_3);
    LiClear(&aux_2);
    LiClear(&aux_1);
    error_set_this_param(cterm_mk_NBS("Non-zero modulus"));
    return cocoa_error("EXPECTED");
  }
  LiInit(&res_n);
  mpz_powm(res_n, aux_1, aux_2, aux_3);
  LiClear(&aux_3);
  LiClear(&aux_2);
  LiClear(&aux_1);
  return arbint2object(res_n);
}


object do_lin_ker(object o)
{
  omat M;
  olist res_L;
  M=object_cast_omat(object_link(o));
  if (math_error != object_NULL) return math_error;
  res_L=lin_kernel(M);
  omat_free(M);
  return olist2object(res_L);
}

object do_lin_sol(object o1, object o2)
{
  omat M;
  olist L,res_L;
  M=object_cast_omat(object_link(o1));
  if (math_error != object_NULL) return math_error;
  L=object_cast_olist(object_link(o2));
  if (math_error != object_NULL) {
    omat_free(M);
    return math_error;
  }
  res_L=lin_solve(M,L);
  omat_free(M);
  olist_free(L);
  return olist2object(res_L);
}

object do_factor(object o)
{
  if (object_type(o) != objtype_poly) return cocoa_error("BAD_PARAMS");
      /* TODO: distinguere a seconda dei coefficienti */
  return cocoa_DMPZfactor(o->v.poly_value, o->R);
}
 
object do_gcd(object f, object g)
{
  if (object_type(f) != objtype_poly) return cocoa_error("BAD_PARAMS");
  if (object_type(g) != objtype_poly) return cocoa_error("BAD_PARAMS");
  if (object_ring(f) != object_ring(g)) return cocoa_error("OBJS_OF_DIFF_RINGS");
  return arpoly2object(arpoly_gcd_JAA(f->v.poly_value, g->v.poly_value, f->R), f->R);
}
 
object do_zpmk(object o1, object o2, ring R)
{
  objtype t1, t2;
  object res_o;
  int p;
  t1=object_type(o1);
  t2=object_type(o2);

  if (t2==objtype_int) p=o2->v.int_value;
  else p=object_cast_int(object_link(o2));

  if (math_error != object_NULL) return math_error;

  if (p>ZP_MAX) {
    error_set_this_param(cterm_mk_NBS("short integer"));
    return cocoa_error("EXPECTED"); }

  switch(t1) {
    case objtype_int:
      res_o=zmod2object(zmod_mk(MMOD(o1->v.int_value,p),p)); break;
    case objtype_arbint:
      res_o=object_init(objtype_arbint,NULL,R);
      LiInit(&res_o->v.arbint_value);
      mpz_mmod_ui(res_o->v.arbint_value,o1->v.arbint_value,p);
      res_o=zmod2object(zmod_mk(object_cast_int(res_o),p)); break;
    default: res_o=cocoa_error("BAD_PARAMS"); }
  return res_o;  
}

object do_log(object o, ring R)
{
  if (objtype_lt(object_type(o),objtype_poly))
    o=object_cast(object_link(o),objtype_poly, R);
    
  if (object_type(o) != objtype_poly) return cocoa_error("BAD_PARAMS");
  if (o->v.poly_value == NULL) return cocoa_error("LOG_ZERO");

  {
    object res_o;
    term t;

    res_o=object_init(objtype_olist,NULL,R);
    t=arpoly_leading_term(o->v.poly_value,R);
    res_o->v.olist_value=this_iv2olist(term_log(t,R));
    term_free(t);
    return res_o;
  }
}

object do_subst(object o, subst s)
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    res_o=object_init(object_type(o),NULL,R);
    res_o->v.poly_value=subst_in_arpoly(o->v.poly_value,s,R);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;  
}

coc_bool object_is_zero(object o)
{
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    return o->v.poly_value == NULL;
  case objtype_ratfun:
    return ratfun_is_zero(o->v.ratfun_value);
  case objtype_coeff:
    return coeff_is_zero(o->v.coeff_value,R);
  case objtype_ideal:
  case objtype_module:
    return module_is_zero(o->v.module_value);
  case objtype_int:
    return (o->v.int_value==0);
  case objtype_arbint:
    return IsLiZero(o->v.arbint_value);
  case objtype_arbrat:
    return IsLrZero(o->v.arbrat_value);
  case objtype_double:
    return (*(o->v.double_value)==0);
  case objtype_zmod:
    return (zmod_n(o->v.zmod_value)==0);
  case objtype_omat:
    return omat_is_zero(o->v.omat_value);
  case objtype_pvector:
    return pv_is_zero(o->v.pvector_value);
  default:
    math_error=cocoa_error("BAD_PARAMS");
    return FALSE;
  }
}

object object_zero(objtype t,ring R)
{
  switch(t) {
  case objtype_poly:
  case objtype_ratfun:
    return arpoly2object(NULL,R);
  case objtype_coeff:
    return coeff2object(coeff_zero(R),R);
  case objtype_int:
    return int2object(0);
  case objtype_arbint: {
    arbint n;
    LiInit(&n);
    LiMakeCoeff(&n,0,1,R);
    return arbint2object(n); }
  default:
    error_set_this_param(cterm_mk_NBS("in INTERNAL function object_zero"));
    return cocoa_error("BAD_PARAMS");
  }
}

object do_homog(object o, unknown x, ring R)
{
  object res_o;

  switch(object_type(o)) {
  case objtype_poly:
    res_o=object_init(object_type(o),NULL,R);
    res_o->v.poly_value=arpoly_homog(o->v.poly_value,x,R);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;  
}

object do_der(object o, int * der_list)
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    res_o=object_init(object_type(o),NULL,R);
    res_o->v.poly_value=arpoly_der_nn(o->v.poly_value,der_list,R);
    break;
  case objtype_arbint:
  case objtype_arbrat:
    res_o=object_init(objtype_arbint,NULL,R);
    LiInit(&res_o->v.arbint_value);
    LiMakeCoeff(&res_o->v.arbint_value,0,1,R);
    break;
  case objtype_ratfun:
    res_o=object_init(object_type(o),NULL,R);
    res_o->v.ratfun_value=ratfun_der_nn(o->v.ratfun_value,der_list,R);
    object_simplify(res_o);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;  
}

object do_component(object o,int index)
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_pvector: {
    pvcomp aux_p;
    res_o=object_init(objtype_poly,NULL,R);
    aux_p=pvcomp_dup(pv_nth(o->v.pvector_value,index-1),R);
    pvcomp_Apply_Anti_Shift(&aux_p,o->v.pvector_value,index,R);
    res_o->v.poly_value=this_pvcomp2arpoly(aux_p,R);
    } break;
  case objtype_zmod:
    if (index==1) {
      res_o=int2object(zmod_sym_n(o->v.zmod_value));
    } else if (index==2) res_o=int2object(zmod_p(o->v.zmod_value)); 
    break;
  case objtype_panel:
    res_o=int2object(opt_get_value(o->v.panel_value,index));
    break;
  case objtype_arbrat:
    if (index==1) 
      res_o=arbint2object(arbint_dup(arbrat_num(o->v.arbrat_value)));
    else if (index==2) 
      res_o=arbint2object(arbint_dup(arbrat_den(o->v.arbrat_value)));
    break;
  case objtype_olist:
    res_o=object_link(olist_nth(o->v.olist_value,index));
    break;
  case objtype_omat:
    res_o=object_link(olist_nth(o->v.olist_value,index));
    break;
  case objtype_string:
    res_o=object_init(objtype_string,NULL,R);
    res_o->v.string_value=malloc_str(1);
    sprintf(res_o->v.string_value,"%c",o->v.string_value[index-1]);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_head(object o) /* TODO: check ring problem */
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_olist:
    if (olist_len(o->v.olist_value) == 0)
      res_o=cocoa_error("CANNOT_APPLY_TO_EMPTY");
    else
      res_o=object_link(olist_head(o->v.olist_value));
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_tail(object o)
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_olist:
    if (olist_len(o->v.olist_value) == 0)
      res_o=cocoa_error("CANNOT_APPLY_TO_EMPTY");
    else {
      res_o=object_init(objtype_olist,NULL,R);
      res_o->v.olist_value=olist_tail(o->v.olist_value); }
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}


object do_gens(object o)
{
  object res_o;
  ring R;

  R=o->R;
  switch(object_type(o)) {
  case objtype_ideal:
  case objtype_module:
    res_o=object_init(objtype_vlist,NULL,R);
    res_o->v.vlist_value=vlist_dup(module_gens_ck(o->v.module_value),R);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_dense_poly(int deg,ring R)
{
  object res_o;
  res_o=object_init(objtype_poly,NULL,R);
  res_o->v.poly_value=arpoly_full_support_of_degree(deg,R);
  return res_o;
}

object do_randomized(object o)
{
  ring R;
  object res_o;
  R=o->R;

  switch(object_type(o)) {
  case objtype_poly:
    res_o=object_init(object_type(o),NULL,R);
    res_o->v.poly_value=arpoly_randomize(o->v.poly_value,R);
    break;
  case objtype_arbint: {
    arbint aux_n;
    res_o=object_init(object_type(o),NULL,R);
    LiInit(&aux_n);
    LiRandomize(&aux_n,R);
    res_o->v.arbint_value=aux_n; }
    break;
/*
 *  case objtype_arbrat: {
 *    arbrat aux_n;
 *    res_o=object_init(object_type(o),NULL,R);
 *    LrInit(&aux_n);
 *    LrRandomize(&aux_n,R);
 *    res_o->v.arbrat_value=aux_n; }
 *    break;
 */
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_mvpoincare_quotient(object o, object o1, ring MVP)
{ /* NOTE: the contents of the object o will be modified */
  object res_o;
  weilist wl;
  module M;
  ring R;
  M = o->v.module_value;
  R = module_ring(M);

  if ( module_mvpseries(M) == pseries_NULL)
  {
    wl=convert_object2weilist(o1,R);
    module_set_mvpseries(M,
      MVPoincareSeriesOfVList(module_gbasis(M), module_shifts(M),
                              module_rank(M), wl, R, MVP));
    weilist_free(wl);
  }
  res_o=olist2object(MVPSeries2olist(module_mvpseries(M)));
  return res_o;
}

object do_poincare_quotient(object o)
{
  module M;
  object res_o;
  olist aux_ol;
  vlist aux_vl;
  shifts aux_ss;
  PSeries aux_PS;
  int r;
  ring R;

  if (object_type(o) == objtype_olist) {
    aux_ol=object2olist(object_link(o));
    R=olist_extract_ring(aux_ol);
    aux_vl=this_olist2vlist_bis(aux_ol,R);
    vlist_rank_and_shifts(aux_vl,&r,&aux_ss);
    aux_PS = PoincareSeriesOfVList(aux_vl, aux_ss, r, R);
    return olist2object(pseries2olist(aux_PS));
  }

  M = o->v.module_value;
  if (module_pseries(M) == pseries_NULL) {
    if (module_gbasis(M) == vlist_NULL) {
      error_set_param(cterm_mk_NBS("(missing GBasis)"));
      return cocoa_error("CANNOT_COMPUTE");
    } else {
      module_set_pseries(M, module_PS_of_quotient(M));
    }
  }
  res_o=olist2object(pseries2olist(module_pseries(M)));
  return res_o;
}

int do_size(object o)
{
  ring R;
  int res_i;
  R=object_ring(o);
  switch(object_type(o)) {
  case objtype_coeff:
    res_i=1; /* TODO: for now */
    break;
  case objtype_poly:
    res_i=arpoly_size(o->v.poly_value,R);
    break;
  case objtype_ratfun:
    res_i=arpoly_size(ratfun_num(o->v.ratfun_value),R)+
          arpoly_size(ratfun_den(o->v.ratfun_value),R);
    break;
  case objtype_int:
    res_i=1;
    break;
  case objtype_arbint:
    res_i=mpz_size(o->v.arbint_value);
    break;
  case objtype_arbrat:
    res_i=mpz_size(&((verylongrat)(o->v.arbrat_value))->_mp_num)+
          mpz_size(&((verylongrat)(o->v.arbrat_value))->_mp_den);
    break;
  case objtype_pvector:
    res_i=pv_size(o->v.pvector_value,R);
    break;
  case objtype_olist:
    res_i=olist_do_size(o->v.olist_value);
    break;
  case objtype_omat:
    res_i=olist_do_size(o->v.omat_value);
    break;
  case objtype_string:
    res_i=strlen(o->v.string_value)+1;
    break;
  case objtype_record:
    res_i = record_do_size(o->v.record_value);
    break;
  default: {
    res_i = 0;
    math_error=error_cannot_apply("Size",object_type(o)); }
  }
  return res_i;
}

int do_len(object o)
{
  ring R;
  int res_i;
  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    res_i=arpoly_len(o->v.poly_value,R);
    break;
  case objtype_zmod:
    res_i=2;
    break;
  case objtype_pvector:
    res_i=pv_rank(o->v.pvector_value);
    break;
  case objtype_ratfun:
    res_i=arpoly_len(ratfun_num(o->v.ratfun_value),R)+
          arpoly_len(ratfun_den(o->v.ratfun_value),R);
    break;
  case objtype_panel:
    res_i=opt_panel_dim(o->v.panel_value);
    break;
  case objtype_ideal:
  case objtype_module:
    res_i=vlist_len(module_gens(o->v.module_value));
    break;
  case objtype_olist:
    res_i=olist_len(o->v.olist_value);
    break;
  case objtype_omat:
    res_i=olist_len(o->v.omat_value);
    break;
  case objtype_int:
    res_i=1; /* TODO: for now */
    break;
  case objtype_arbint:
    res_i=1; /* TODO: for now */
    break;
  case objtype_arbrat:
    res_i=2; /* TODO: for now */
    break;
  case objtype_string:
    res_i=strlen(o->v.string_value);
    break;
  default: {
    res_i = 0;
    math_error=error_cannot_apply("Len",object_type(o)); }
  }
  return res_i;
}

int do_deg(object o)
{
  ring R;
  int res_i;
  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    res_i=arpoly_deg(o->v.poly_value, R);
    break;
  case objtype_pvector:
    res_i=pv_deg(o->v.pvector_value, R);
    break;
  default:
    res_i = 0;
  }
  return res_i;
}

object do_ispurelt(arpoly f)
{
  return bool2object(arpoly_ispurelt(f));
}

object do_mdeg(object o)
{
  ring R;
  ivector res_v;
  R=o->R;
  switch(object_type(o)) {
  case objtype_poly:
    res_v=arpoly_mdeg(o->v.poly_value, R);
    break;
  case objtype_pvector:
    res_v=pv_mdeg(o->v.pvector_value, R);
    break;
  default:
    res_v=iv_init(weilist_len(ring_weilist(R)));
  }
  return olist2object(this_iv2olist(res_v));
}

object do_det(object o)
{
  object res_o;
  switch(object_type(o)) {
  case objtype_omat:
    if (! omat_is_square(o->v.omat_value)) {
      error_set_this_param(cterm_mk_NBS("a square matrix"));
      res_o=cocoa_error("EXPECTED");
    } else res_o=omat_det(o->v.omat_value);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_adj(object o)
{
  object res_o;
  switch(object_type(o)) {
  case objtype_omat:
    if (! omat_is_square(o->v.omat_value)) {
      error_set_this_param(cterm_mk_NBS("a square matrix"));
      res_o=cocoa_error("EXPECTED");
    } else res_o=omat2object(omat_adjoint(o->v.omat_value));
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_minors(object o,int p)
{
  ring R;
  object res_o;
  R=o->R;
  switch(object_type(o)) {
  case objtype_omat: {
    olist aux_ol;
    aux_ol=omat_minors(o->v.omat_value,p);
    if (math_error != object_NULL) {
      res_o=math_error;
      math_error=object_NULL;
    } else
      res_o=olist2object(aux_ol);
    } break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

void do_inv(object * o)
{ 
  ring R;

  if (object_rc_is_not_one(*o)) {
    object_unlink(*o);
    *o=object_dup(*o);
  }

  R=object_ring(*o);
  switch(object_type(*o)) {
  case objtype_zmod:
    (*o)->v.zmod_value=zmod_inv((*o)->v.zmod_value);
    break;
  case objtype_coeff: 
    R->InvertCoeff(&((*o)->v.coeff_value),(*o)->v.coeff_value,R); 
    break;
  case objtype_omat:
    if (! omat_is_square((*o)->v.omat_value)) {
      error_set_this_param(cterm_mk_NBS("a square matrix"));
      object_unlink((*o));
      *o=cocoa_error("EXPECTED");
    } else {
      omat aux_M;
      aux_M=omat_inv((*o)->v.omat_value);
      if (math_error != object_NULL) {
        (*o)=math_error;
	math_error=object_NULL;
      } else {
        omat_free((*o)->v.omat_value);
        (*o)->v.omat_value=aux_M;
      }
    }
    break;
 case objtype_int: {
    arbrat aux_r;
    aux_r=int2arbrat((*o)->v.int_value);
    object_empty(*o);
    object_unlink(*o);
    *o=object_init(objtype_arbrat,NULL,R);
    (*o)->v.arbrat_value=aux_r;
    LrInvert(&(*o)->v.arbrat_value,(*o)->v.arbrat_value,R); }
    break;
 case objtype_arbint: {
    arbrat aux_r;
    aux_r=this_arbint2arbrat((*o)->v.arbint_value);
    object_empty(*o);
    object_unlink(*o);
    *o=object_init(objtype_arbrat,NULL,R);
    (*o)->v.arbrat_value=aux_r;
    LrInvert(&(*o)->v.arbrat_value,(*o)->v.arbrat_value,R); }
    break;
  case objtype_arbrat:
    LrInvert(&(*o)->v.arbrat_value,(*o)->v.arbrat_value,R);
    break;
  case objtype_poly:
    (*o)->v.ratfun_value=ratfun_inv_this(
      this_arpoly2ratfun((*o)->v.poly_value,R),R);
    object_set_type(*o,objtype_ratfun);
    break;
  case objtype_ratfun:
    (*o)->v.ratfun_value=ratfun_inv_this((*o)->v.ratfun_value,R);
    break;
  default: {
    char * aux_s;
    aux_s=objtype2str(object_type(*o));
    error_set_this_param(cterm_mk_NBS(aux_s));
    free_str1(aux_s);
    object_unlink(*o);
    *o=cocoa_error("CANNOT_INV"); }
  }
}

object do_power(object o,int exponent)
{
  object aux_o,res_o;
  ring R;

  R=o->R;
  if (object_type(o)==objtype_int) {
    aux_o=object_cast(object_link(o),objtype_arbint,ring_NULL);
    res_o=do_power(aux_o,exponent);
    object_unlink(aux_o);
    return res_o; }
  res_o=object_init(object_type(o),NULL,R);
  switch(object_type(o)) {
  case objtype_coeff:
    R->ClearCoeff(&res_o->v.coeff_value); 
    /* TODO: eventualmente cambiare object_init in modo che non allochi
       lo spazio */
    res_o->v.coeff_value=coeff_power(o->v.coeff_value,exponent,R);
    break;
  case objtype_zmod: {
    register int i;
    res_o->v.zmod_value=zmod_mk(1,zmod_p(o->v.zmod_value));
    for (i=1;i<=exponent;i++)
      res_o->v.zmod_value=zmod_mult(res_o->v.zmod_value,o->v.zmod_value); }
    break;
  case objtype_arbrat:
    LrClear(&res_o->v.arbrat_value);
    res_o->v.arbrat_value=coeff_power(o->v.arbrat_value,exponent,ring_Q);
    break;
  case objtype_omat:
    if (! omat_is_square(o->v.omat_value)) {
      error_set_this_param(cterm_mk_NBS("a square matrix"));
      res_o=cocoa_error("EXPECTED"); }
    else
      res_o->v.omat_value=omat_pow(o->v.omat_value,exponent);
    break;
  case objtype_arbint:
    LiClear(&res_o->v.arbint_value);
    res_o->v.arbint_value=coeff_power(o->v.arbint_value,exponent,ring_Z);
    break;
  case objtype_poly:
    res_o->v.poly_value=arpoly_power(o->v.poly_value,exponent,R);
    break;
  case objtype_ideal:
    res_o->v.module_value=ideal_pow(o->v.module_value,exponent,R);
    break;
  case objtype_ratfun:
    res_o->v.ratfun_value=ratfun_power(o->v.ratfun_value,exponent,R);
    break;
  case objtype_ring:
    object_set_type(res_o,objtype_module);
    object_set_ring(res_o,ring_link(o->v.ring_value));
    res_o->v.module_value=module_freemod(exponent,o->v.ring_value);
    break;
  default:
    object_unlink(res_o);
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;
}

object do_subtract(object o1,object o2)
/* assumed homogeneous difference */
{
  object res_o,aux_o;
  objtype t1,t2,aux_t;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    return cocoa_error("OBJS_OF_DIFF_RINGS");

  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1!=t2) {

    if (!objtype_comparable(t1,t2))
      return object_error_types("BAD_PARAMS",t1,"-",t2); /* <--- */

    /* if (cond) "efficient algorithms for particolar cases" ... */

    /* ... else "general solution: cast o1 to the type of o2 */

    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) return o1;
      res_o=do_subtract(o1,o2);
      object_unlink(o1); }
    else {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) return o2;
      res_o=do_subtract(o1,o2);
      object_unlink(o2); }

    return res_o;
  }

  R=object_true_ring(o1,o2);

  res_o=object_init(object_type(o1),NULL,R);
  switch(object_type(o1)) {
  case objtype_coeff:
    res_o->v.coeff_value = coeff_sub(o1->v.coeff_value,o2->v.coeff_value,R);
    break;
  case objtype_poly:
    res_o->v.poly_value=arpoly_sub_these(arpoly_dup(o1->v.poly_value,R),
      arpoly_dup(o2->v.poly_value,R),R);
    break;
  case objtype_ratfun:
    res_o->v.ratfun_value=
      ratfun_sub(o1->v.ratfun_value,o2->v.ratfun_value,R);
    object_simplify(res_o);
    break;
  case objtype_olist: {
    olist res_ol;
    res_ol=olist_sub(o1->v.olist_value,o2->v.olist_value);
    if (math_error != object_NULL) {
      object_unlink(res_o);
      return object_link(math_error); } /* TODO: check */
    res_o->v.olist_value=res_ol; }
    break;
  case objtype_omat: {
    omat res_m;
    res_m=omat_sub(o1->v.omat_value,o2->v.omat_value);
    if (math_error != object_NULL) {
      object_unlink(res_o);
      return object_link(math_error); } /* TODO: check */
    res_o->v.omat_value=res_m; }
    break;
  case objtype_int:
    res_o->v.int_value=o1->v.int_value-o2->v.int_value;
    break;
  case objtype_zmod:
    res_o=object_init(object_type(o1),NULL,R);
    res_o->v.zmod_value=zmod_sub(o1->v.zmod_value,o2->v.zmod_value);
    break;
  case objtype_pvector:
    if (pv_rank(o1->v.pvector_value)==pv_rank(o2->v.pvector_value)) {
      res_o->v.pvector_value=
        pv_subtract(o1->v.pvector_value,o2->v.pvector_value,R); }
    else {
      object_unlink(res_o);
      return cocoa_error("NOT_RANK_HOM"); }
    break;
  case objtype_arbint:
    res_o->v.arbint_value = coeff_sub(o1->v.arbint_value,o2->v.arbint_value,ring_Z);
    break;
  case objtype_arbrat:
    res_o->v.arbrat_value = coeff_sub(o1->v.arbrat_value,o2->v.arbrat_value,ring_Q);
    object_simplify(res_o);
    break;
  default:
    object_unlink(res_o);
    res_o=object_error_types("BAD_PARAMS",t1,"-",t2); } /* <--- */
  return res_o;
}

object do_sum(object o1,object o2)
{
  object res_o,aux_o;
  objtype t1,t2,aux_t;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    return cocoa_error("OBJS_OF_DIFF_RINGS");

  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (o1 == object_ZERO) return object_link(o2);
  if (o2 == object_ZERO) return object_link(o1);

  if (t1!=t2) {

    if (!objtype_comparable(t1,t2))
      return object_error_types("BAD_PARAMS",t1,"+",t2); /* <--- */

    if (objtype_lt(t2,t1)) {
      aux_t=t1; t1=t2; t2=aux_t;
      aux_o=o1; o1=o2; o2=aux_o;
    }

    /* if (cond) "efficient algorithms for particolar cases" ... */

    /* ... else "general solution: cast o1 to the type of o2 */
    
    aux_o=object_cast(object_link(o1),t2,R);
    if (object_is_error(aux_o)) return aux_o;
    res_o=do_sum(aux_o,o2);
    object_unlink(aux_o);
    return res_o;
  }

  R=object_ring(o1);

  {
    res_o=object_init(object_type(o1),NULL,R);
    switch(t1) {
    case objtype_coeff:
      res_o->v.coeff_value=coeff_sum(o1->v.coeff_value,o2->v.coeff_value,R);
      break;
    case objtype_int:
      res_o->v.int_value=o1->v.int_value+o2->v.int_value;
      break;
    case objtype_zmod:
      if (zmod_p(o1->v.zmod_value) != zmod_p(o2->v.zmod_value)) {
        object_unlink(res_o);
        res_o=cocoa_error("BAD_PARAMS");
      } else
      res_o->v.zmod_value=zmod_sum(o1->v.zmod_value,o2->v.zmod_value);
      break;
    case objtype_arbint:
      res_o->v.arbint_value=coeff_sum(o1->v.arbint_value,o2->v.arbint_value,ring_Z);
      break;
    case objtype_arbrat:
      res_o->v.arbrat_value=coeff_sum(o1->v.arbrat_value,o2->v.arbrat_value,ring_Q);
      object_simplify(res_o);
      break;
    case objtype_poly:
      res_o->v.poly_value=arpoly_sum(o1->v.poly_value,o2->v.poly_value,R);
      break;
    case objtype_ratfun:
      res_o->v.ratfun_value=
        ratfun_sum(o1->v.ratfun_value,o2->v.ratfun_value,R);
      object_simplify(res_o);
      break;
    case objtype_ideal:
      res_o->v.module_value=module_sum(o1->v.module_value,o2->v.module_value,R);
      break;
    case objtype_module:
      if (module_not_compatible(o1->v.module_value, o2->v.module_value)) {
        object_unlink(res_o);
        res_o=cocoa_error("NOT_RANK_HOM"); }
      else 
      res_o->v.module_value=module_sum(o1->v.module_value,o2->v.module_value,R);
      break;
    case objtype_pvector:
      if (pv_rank(o1->v.pvector_value)==pv_rank(o2->v.pvector_value)) {
        res_o->v.pvector_value=
          pv_sum(o1->v.pvector_value,o2->v.pvector_value,R); }
      else {
        object_unlink(res_o);
        return cocoa_error("NOT_RANK_HOM"); }
      break;

    case objtype_olist: {
      olist res_ol;
      res_ol=olist_sum(o1->v.olist_value,o2->v.olist_value);
      if (math_error != object_NULL) {
        object_unlink(res_o);
        return object_link(math_error); } /* TODO: check */
      res_o->v.olist_value=res_ol; }
      break;
    case objtype_omat: {
      omat res_m;
      res_m=omat_sum(o1->v.omat_value,o2->v.omat_value);
      if (math_error != object_NULL) {
        object_unlink(res_o);
        return object_link(math_error); } /* TODO: check */
      res_o->v.omat_value=res_m; }
      break;
    case objtype_string:
      res_o->v.string_value=my_strsum(o1->v.string_value,o2->v.string_value);
      break;
#ifdef NOTDEFINED
    case objtype_mat:
      res_o->v.mat_value=mat_sum(o1->v.mat_value,o2->v.mat_value,R);
      break;
#endif
    default:
      object_unlink(res_o);
      res_o=object_error_types("BAD_PARAMS",t1,"+",t2); } /* <--- */
    }
  return res_o;
}

object do_concat(object o1,object o2)
{
  object res_o;
  ring R;
  olist aux_ol1,aux_ol2;
  object aux_o;

  R=object_true_ring(o1,o2);

  if (object_type(o1)==objtype_string)
    if (object_type(o2)==objtype_string)
      return string2object(my_strsum(o1->v.string_value,o2->v.string_value));
    else return cocoa_error("BAD_PARAMS");

  aux_o=object_dup(o1);
  aux_ol1=object_cast_olist(aux_o);
  if (math_error != object_NULL) {
    object_unlink(aux_o);
    res_o=math_error;
    math_error=object_NULL;
    return res_o; }
  aux_o=object_dup(o2);
  aux_ol2=object_cast_olist(aux_o);
  if (math_error != object_NULL) {
    object_unlink(aux_o);
    res_o=math_error;
    math_error=object_NULL;
    return res_o; }
  return olist2object(olist_concat_these(aux_ol1,aux_ol2));
}

object do_append(object o1,object o2)
{
  object res_o;
  ring R;

  R = object_ring(o1);

  switch object_type(o1) {

  case objtype_olist : {
    olist aux_ol;
    aux_ol=o1->v.olist_value;
    o1->v.olist_value=olist_append_this(aux_ol,object_link(o2)); }
    break;
  default :
    return cocoa_error("EXPECTED_LIST");
  }
  return object_link(object_null);
}

object do_remove(object o1, int n)
{
  object res_o;
  ring R;

  R = object_ring(o1);

  switch object_type(o1) {

  case objtype_olist : {
    olist aux_ol;
    aux_ol=o1->v.olist_value;
    olist_remove(aux_ol,n); }
    break;
  default :
    return cocoa_error("EXPECTED_LIST");
  }
  if (math_error != object_NULL) return math_error;
  return object_link(object_null);
}

object do_div(object o1,object o2)
{
  object res_o,aux_o;
  objtype t1,t2;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    res_o = cocoa_error("OBJS_OF_DIFF_RINGS");
  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1!=t2) {

    /* if (cond) "special cases: non homog operation" ... */

    /* ... else "homog operation" */

    if (!objtype_comparable(t1,t2))
      return object_error_types("CANNOT_DIV",t1," by ",t2); /* <--- */

    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) return o1;
      res_o=do_div(o1,o2);
      object_unlink(o1); }
    else {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) return o2;
      res_o=do_div(o1,o2);
      object_unlink(o2); }

    return res_o;
  }

  res_o=object_init(t1,NULL,R);
  switch(t1) {
    case objtype_coeff:
      res_o->v.coeff_value=coeff_div(o1->v.coeff_value,o2->v.coeff_value,R);
      break;

    case objtype_poly:
      if (ring_coeff_in_field(R)) {
        arpoly aux_pr;
        arpoly_div(o1->v.poly_value,o2->v.poly_value,
          &res_o->v.poly_value,&aux_pr,R);
        arpoly_free(aux_pr,R);
      }
      else res_o=object_error_types("CANNOT_DIV",t1," by ",t2);
      break;

    case objtype_int:
      res_o->v.int_value=o1->v.int_value/o2->v.int_value;
      break;

    case objtype_arbint:
      res_o->v.arbint_value=coeff_div(o1->v.arbint_value,o2->v.arbint_value,ring_Z);
      break;

    case objtype_arbrat:
      res_o->v.arbrat_value=coeff_div(o1->v.arbrat_value,o2->v.arbrat_value,ring_Q);
      object_simplify(res_o);
      break;

    default:
      object_unlink(res_o);
      return object_error_types("CANNOT_DIV",t1," by ",t2); /* <--- */
    }
  return res_o;
}

object do_mod(object o1,object o2)
{
  object res_o,aux_o;
  objtype t1,t2;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    res_o = cocoa_error("OBJS_OF_DIFF_RINGS");
  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1!=t2) {
    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) return o1;
      res_o=do_mod(o1,o2);
      object_unlink(o1); }
    else if (objtype_lt(t2,t1)) {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) return o2;
      res_o=do_mod(o1,o2);
      object_unlink(o2); }
    else return object_error_types("BAD_PARAMS",t1," mod ",t2);
    return res_o;
  }

  res_o=object_init(t1,NULL,R);

  switch (t1) {
    case objtype_coeff:
      coeff_free(res_o->v.coeff_value,R);
      res_o->v.coeff_value=coeff_mmod(o1->v.coeff_value,o2->v.coeff_value,R);
      break;

    case objtype_poly:
      if (ring_coeff_in_field(R)) {
        if (o2->v.poly_value == NULL)
          cocoa_error("Division by zero");
        else {
          arpoly aux_pq;
          arpoly_div(o1->v.poly_value,o2->v.poly_value,&aux_pq,
            &res_o->v.poly_value,R);
          arpoly_free(aux_pq,R); }
      }
      else res_o=cocoa_error("CANNOT_DIV");
      break;

    case objtype_arbint:
      LiInit(&res_o->v.arbint_value);
      LiMod(&res_o->v.arbint_value,o1->v.arbint_value,o2->v.arbint_value,R);
      break;

    case objtype_int:
      res_o->v.int_value= MMOD(o1->v.int_value,o2->v.int_value);
      break;
    default: {
      object_unlink(res_o);
      res_o=object_error_types("BAD_PARAMS",t1,",",t2); } /* <--- */
  }
  return res_o;
}
  
object do_frac(object o1,object o2)
{
  object res_o,aux_o;
  objtype t1,t2;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    return cocoa_error("OBJS_OF_DIFF_RINGS");

  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1!=t2) {

    /* if (cond) "special cases: non homog operation" ... */

    res_o=object_NULL;

    if (t1==objtype_ring) {
      if (t2==objtype_ideal) {
        if (!rings_compatible(o1->v.ring_value,object_ring(o2))) return cocoa_error("OBJS_OF_DIFF_RINGS");
        res_o=object_dup(o2);
        object_set_label(res_o,my_strdup("Quotient")); }
      else 
        object_error_types("BAD_PARAMS",t1,"/",t2);
    } else if (t1==objtype_pvector) {
      if (objtype_le(t2,objtype_arbrat) || t2==objtype_zmod) {
        aux_o=object_link(o2);
        do_inv(&aux_o);
        if (object_is_error(aux_o)) res_o=aux_o;
        else {
         res_o=do_mult(o1,aux_o);
         object_unlink(aux_o);
        }
      }
      else 
        object_error_types("BAD_PARAMS",t1,"/",t2);
    }

    if (res_o != object_NULL) return res_o;

    /* ... else "homog operation" */

    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) return o1;
      res_o=do_frac(o1,o2);
      object_unlink(o1); }
    else if (objtype_lt(t2,t1)) {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) return o2;
      res_o=do_frac(o1,o2);
      object_unlink(o2); }
    else return object_error_types("BAD_PARAMS",t1,"/",t2);

    return res_o;
  }

  /* caso t1=t2 */

  if (object_is_zero(o2)) return cocoa_error("DIV_BY_ZERO");

  switch(t1) {
    case objtype_module:
      if (!module_is_freemod(o1->v.module_value) ||
          module_not_compatible(o1->v.module_value, o2->v.module_value))
      {
        res_o = cocoa_error("MODULE_QUOTIENT");
        break;
      }
      res_o=object_dup(o2);
      object_set_label(res_o,my_strdup("Quotient"));
      break;
    case objtype_coeff:
      if (ring_coeff_in_field(R)) {
        res_o=object_init(object_type(o1),NULL,R);
        coeff_free(res_o->v.coeff_value,R);
        res_o->v.coeff_value=coeff_div(o1->v.coeff_value,o2->v.coeff_value,R); }
      else res_o=cocoa_error("CANNOT_DIV");
      break;
    case objtype_ratfun:
      res_o=object_init(objtype_ratfun,NULL,R);
      res_o->v.ratfun_value=ratfun_div(o1->v.ratfun_value,o2->v.ratfun_value,R);
      object_simplify(res_o);
      break;
    case objtype_int:
    {
      /* Just a quick fix -- god knows what Antonio was up to... */
      arbint N, D;
      N = int2arbint(o1->v.int_value);
      D = int2arbint(o2->v.int_value);
      res_o=object_init(objtype_arbrat,NULL,R);
      res_o->v.arbrat_value=arbrat_mk(N, D);
      LiClear(&N);
      LiClear(&D);
      object_simplify(res_o);
      break;
    }
    case objtype_arbint:
      res_o=object_init(objtype_arbrat,NULL,R);
      res_o->v.arbrat_value=arbrat_mk(o1->v.arbint_value,o2->v.arbint_value);
      object_simplify(res_o);
      break;
    case objtype_zmod:
      if (!zmod_invertible(o2->v.zmod_value))
        res_o=cocoa_error("CANNOT_DIV");
      else {
        res_o=object_init(object_type(o1),NULL,R);
        res_o->v.zmod_value=zmod_frac(o1->v.zmod_value,o2->v.zmod_value); }
      break;
    case objtype_arbrat:
      res_o=object_init(objtype_arbrat,NULL,R);
      res_o->v.arbrat_value=coeff_div(o1->v.arbrat_value,o2->v.arbrat_value,ring_Q);
      object_simplify(res_o);
      break;
    case objtype_poly:
      if (ring_coeff_in_field(R)) {
        res_o=object_init(objtype_ratfun,NULL,R);
        res_o->v.ratfun_value=
          ratfun_create(o1->v.poly_value,o2->v.poly_value,R);
        object_simplify(res_o); }
      else res_o=cocoa_error("CANNOT_DIV");
      break;
    default:
      return object_error_types("BAD_PARAMS",t1,"/",t2); } /* <--- */
  return res_o;
}



object do_mult(object o1,object o2)
{
  object res_o,aux_o;
  objtype t1,t2,aux_t;
  ring R;

  if (!object_rings_compatible(o1,o2)) 
    return cocoa_error("OBJS_OF_DIFF_RINGS");

  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  res_o=object_NULL;

  if (t1!=t2) {

    if (t1==objtype_zmod && t2==objtype_arbint) {
      res_o=object_init(objtype_zmod,NULL,R);
      res_o->v.zmod_value=
        zmod_mult_arbint(o1->v.zmod_value, o2->v.arbint_value);
      return res_o;
    }

    if (t2==objtype_zmod && t1==objtype_arbint) {
      res_o=object_init(objtype_zmod,NULL,R);
      res_o->v.zmod_value=
        zmod_mult_arbint(o2->v.zmod_value, o1->v.arbint_value);
      return res_o;
    }

    if (t1==objtype_zmod && t2==objtype_int) {
      res_o=object_init(objtype_zmod,NULL,R);
      res_o->v.zmod_value=
        zmod_mult_int(o1->v.zmod_value, o2->v.int_value);
      return res_o;
    }

    if (t2==objtype_zmod && t1==objtype_int) {
      res_o=object_init(objtype_zmod,NULL,R);
      res_o->v.zmod_value=
        zmod_mult_int(o2->v.zmod_value, o1->v.int_value);
      return res_o;
    }

    /* if (cond) "efficient algorithms for particular cases" ... */

    switch(t1) {

    case objtype_double:
    case objtype_zmod:
    case objtype_arbint:
    case objtype_int:
    case objtype_arbrat:
      switch(t2) {
      case objtype_pvector: {
        pvcomp aux_p;
        aux_p=object_cast_pvcomp(object_link(o1),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_pvector,NULL,R);
        res_o->v.pvector_value = 
          pv_mult(aux_p,o2->v.pvector_value,R);
        pvcomp_free(aux_p,R); }
        break;
      case objtype_poly: {
        arpoly aux_p;
        aux_p=object_cast_arpoly(object_link(o1),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_poly,NULL,R);
        res_o->v.poly_value = 
          arpoly_mult(aux_p,o2->v.poly_value,R);
        arpoly_free(aux_p,R); }
        break;
      case objtype_ideal:
      case objtype_module: {
        pvcomp aux_p;
        aux_p=object_cast_pvcomp(object_link(o1),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(t2,NULL,R);
        res_o->v.module_value = 
          module_mult_by_pvcomp(o2->v.module_value,aux_p,R); }
        break;
      case objtype_omat:
        res_o=object_init(objtype_omat,NULL,R);
        res_o->v.omat_value = 
          olist_extprod(o2->v.omat_value,o1);
        break;
      case objtype_olist:
        res_o=object_init(objtype_olist,NULL,R);
        res_o->v.olist_value = 
          olist_extprod(o2->v.olist_value,o1);
        break; }

    break;

    case objtype_ratfun:

      switch(t2) {
      case objtype_omat:
        res_o=object_init(objtype_omat,NULL,R);
        res_o->v.omat_value = 
          olist_extprod(o2->v.omat_value,o1);
        break;
      case objtype_olist:
        res_o=object_init(objtype_olist,NULL,R);
        res_o->v.olist_value = 
          olist_extprod(o2->v.olist_value,o1);
        break;
    } 

    break;    

    case objtype_poly:

      switch(t2) {
      case objtype_double:
      case objtype_zmod:
      case objtype_arbint:
      case objtype_arbrat:
      case objtype_int: {
        arpoly aux_p;
        aux_p=object_cast_arpoly(object_link(o2),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_poly,NULL,R);
        res_o->v.poly_value = 
          arpoly_mult(aux_p,o1->v.poly_value,R);
        arpoly_free(aux_p,R);
        } break;
      case objtype_pvector: {
        pvcomp aux_p;
        res_o=object_init(objtype_pvector,NULL,R);
        aux_p=arpoly2pvcomp(o1->v.poly_value,R);
        res_o->v.pvector_value = 
          pv_mult(aux_p,o2->v.pvector_value,R);
        pvcomp_free(aux_p,R);
       }  break;
      case objtype_omat:
        res_o=object_init(objtype_omat,NULL,R);
        res_o->v.omat_value = 
          olist_extprod(o2->v.omat_value,o1);
        break;
      case objtype_olist:
        res_o=object_init(objtype_olist,NULL,R);
        res_o->v.olist_value = 
          olist_extprod(o2->v.olist_value,o1);
        break;
      case objtype_ideal:
      case objtype_module:
        res_o=object_init(t2,NULL,R);
        res_o->v.module_value = 
          module_mult_by_arpoly(o2->v.module_value,o1->v.poly_value,R);
        break;
      }

    break;

    case objtype_pvector:

      switch(t2) {
      case objtype_double:
      case objtype_zmod:
      case objtype_arbint:
      case objtype_arbrat:
      case objtype_int: {
        pvcomp aux_p;
        aux_p=object_cast_pvcomp(object_link(o2),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_pvector,NULL,R);
        res_o->v.pvector_value = 
          pv_mult(aux_p,o1->v.pvector_value,R);
        pvcomp_free(aux_p,R); }
        break;
      case objtype_poly: {
        pvcomp aux_p;
        res_o=object_init(objtype_pvector,NULL,R);
        aux_p=arpoly2pvcomp(o2->v.poly_value,R);
        res_o->v.pvector_value = 
          pv_mult(aux_p,o1->v.pvector_value,R);
        pvcomp_free(aux_p,R);
        } break;
      default:
        return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      }

    break;

    case objtype_olist:

      switch(t2) {
      case objtype_arbint:
      case objtype_int:
      case objtype_arbrat:
      case objtype_double:
      case objtype_zmod:
      case objtype_poly:
      case objtype_ratfun:
        res_o=object_init(objtype_olist,NULL,R);
        res_o->v.olist_value = 
          olist_extprod(o1->v.olist_value,o2);
        break;
      default:
        return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      }

    break;

    case objtype_omat:

      switch(t2) {
      case objtype_arbint:
      case objtype_int:
      case objtype_arbrat:
      case objtype_double:
      case objtype_zmod:
      case objtype_poly:
      case objtype_ratfun:
        res_o=object_init(objtype_omat,NULL,R);
        res_o->v.omat_value = 
          olist_extprod(o1->v.omat_value,o2);
        break;
      default:
        return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      }

    break;

    case objtype_ideal:

      switch(t2) {
      case objtype_poly:
        res_o=object_init(objtype_ideal,NULL,R);
        res_o->v.module_value =
          module_mult_by_arpoly(o1->v.module_value,o2->v.poly_value,R);
        break;
      case objtype_module:
        res_o=object_init(t2,NULL,R);
        res_o->v.module_value =
          module_mult_by_ideal(o2->v.module_value,o1->v.module_value,R);
        break;
      case objtype_zmod:
      case objtype_arbint:
      case objtype_int:
      case objtype_arbrat: {
        pvcomp aux_p;
        aux_p=object_cast_pvcomp(object_link(o2),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_ideal,NULL,R);
        res_o->v.module_value = 
          module_mult_by_pvcomp(o1->v.module_value,aux_p,R); }
        break;
      default:
        return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      }

    break;

    case objtype_module:

      switch(t2) {
      case objtype_poly:
        res_o=object_init(objtype_module,NULL,R);
        res_o->v.module_value =
          module_mult_by_arpoly(o1->v.module_value,o2->v.poly_value,R);
        break;
      case objtype_ideal:
        res_o=object_init(objtype_module,NULL,R);
        res_o->v.module_value =
          module_mult_by_ideal(o1->v.module_value,o2->v.module_value,R);
        break;
      case objtype_zmod:
      case objtype_arbint:
      case objtype_int:
      case objtype_arbrat: {
        pvcomp aux_p;
        aux_p=object_cast_pvcomp(object_link(o2),R);
        if (math_error != object_NULL) return math_error;
        res_o=object_init(objtype_module,NULL,R);
        res_o->v.module_value = 
          module_mult_by_pvcomp(o1->v.module_value,aux_p,R); }
        break;
      default:
        return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      }

    break;

    }

    if (res_o != object_NULL) return res_o;

    /* ... else "general solution: cast o1 to the type of o2 */
    
    if (!objtype_comparable(t1,t2))
      return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */

    if (objtype_lt(t2,t1)) {
      aux_t=t1; t1=t2; t2=aux_t;
      aux_o=o1; o1=o2; o2=aux_o;
    }
    
    aux_o=object_cast(object_link(o1),t2,R);
    if (object_is_error(aux_o)) return aux_o;
    res_o=do_mult(aux_o,o2);
    object_unlink(aux_o);
    return res_o;
  }

  switch(t1) {
    case objtype_coeff:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.coeff_value=coeff_mult(o1->v.coeff_value,o2->v.coeff_value,R);
       break;
    case objtype_arbint:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.arbint_value=coeff_mult(o1->v.arbint_value,o2->v.arbint_value,ring_Z);
      break;
    case objtype_arbrat:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.arbrat_value=coeff_mult(o1->v.arbrat_value,o2->v.arbrat_value,ring_Q);
      object_simplify(res_o);
      break;
    case objtype_zmod:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.zmod_value=zmod_mult(o1->v.zmod_value,o2->v.zmod_value);
      break;
    case objtype_poly:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.poly_value=arpoly_mult(o1->v.poly_value,o2->v.poly_value,R);
      break;
    case objtype_ideal:
      res_o=object_init(objtype_ideal,NULL,R);
      res_o->v.module_value =
        module_mult_by_ideal(o1->v.module_value,o2->v.module_value,R);
      break;
    case objtype_ratfun:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.ratfun_value=ratfun_mult(o1->v.ratfun_value,
        o2->v.ratfun_value,R);
      object_simplify(res_o);
      break;
    case objtype_int:
      res_o=object_init(object_type(o1),NULL,R);
      res_o->v.int_value=o1->v.int_value*o2->v.int_value;
      break;
    case objtype_omat: {
      omat res_M;
      res_M = omat_mult(o1->v.omat_value,o2->v.omat_value);
      if (res_M != omat_NULL)
        res_o=omat2object(res_M);
      else
        res_o=cocoa_error("BAD_PARAMS"); }
      break;
    default:
      return object_error_types("BAD_PARAMS",t1,"*",t2); /* <--- */
      break;
    }
  return res_o;
}

object do_change_sign(object o)
{
  ring R;
  object res_o;

  R=o->R;
  res_o=object_init(object_type(o),NULL,R);
  switch(object_type(o)) {
    case objtype_coeff:
      res_o->v.coeff_value=coeff_dup(o->v.coeff_value,R);
      (*R->ChangeSignToCoeff)(&res_o->v.coeff_value,R);
      break;
    case objtype_arbint:
      LiInit(&res_o->v.arbint_value);
      LiCopy(&res_o->v.arbint_value,o->v.arbint_value);
      LiChangeSign(&res_o->v.arbint_value,R);
      break;
    case objtype_arbrat:
      LrInit(&res_o->v.arbrat_value);
      LrCopy(&res_o->v.arbrat_value,o->v.arbrat_value);
      LrChangeSign(&res_o->v.arbrat_value,R);
      break;
    case objtype_int:
      res_o->v.int_value=-o->v.int_value;
      break;
    case objtype_poly:
      res_o->v.poly_value=arpoly_dup(o->v.poly_value,R);
      arpoly_change_sign(res_o->v.poly_value,R);
      break;
    case objtype_ratfun:
      res_o->v.ratfun_value=ratfun_dup(o->v.ratfun_value,R);
      ratfun_change_sign(res_o->v.ratfun_value,R);
      break;
    case objtype_pvector:
      res_o->v.pvector_value=pv_dup(o->v.pvector_value,R);
      pv_change_sign(res_o->v.pvector_value,R);
      break;
    case objtype_zmod:
      res_o->v.zmod_value=zmod_change_sign(o->v.zmod_value);
      break;
    case objtype_olist:
      res_o->v.olist_value=olist_change_sign(o->v.olist_value);
      break;
    case objtype_omat:
      res_o->v.omat_value=omat_change_sign(o->v.omat_value);
      break;
    default:
      object_unlink(res_o);
      res_o=cocoa_error("BAD_PARAMS");
      break;
  }
  return res_o;
}

object do_is_homog(object o)
{
  ring R;
  object res_o;

  R=o->R;
  res_o=object_init(objtype_bool,NULL,R);
  switch(object_type(o)) {
    case objtype_coeff:
    case objtype_arbint:
    case objtype_arbrat:
    case objtype_zmod:
    case objtype_int:
      res_o->v.bool_value = TRUE;
      break;
    case objtype_poly:
      res_o->v.bool_value=arpoly_is_homog(o->v.poly_value,R);
      break;
    case objtype_ratfun:
      res_o->v.bool_value=
        (arpoly_is_homog(ratfun_num(o->v.ratfun_value),R) &&
         arpoly_is_homog(ratfun_den(o->v.ratfun_value),R));
      break;
    case objtype_pvector:
      res_o->v.bool_value=pv_is_homog(o->v.pvector_value,R);
      break;
    default:
      object_unlink(res_o);
      res_o=cocoa_error("BAD_PARAMS");
      break;
  }
  return res_o;
}

coc_bool do_is_in(object o1,object o2)
{
  coc_bool res_b;
  ring R;
  object aux_o;

  switch(object_type(o2)) {
  case objtype_string:
    res_b = (strstr(o2->v.string_value,o1->v.string_value)!=NULL);
    break;
  case objtype_olist:
    res_b = olist_isin(o2->v.olist_value,o1);
    break;
  default: {
    math_error=cocoa_error("BAD_PARAMS");
    res_b = FALSE; }
  }
  return res_b;
}

coc_bool do_is_eq(object o1,object o2)
/* assumed homogeneous equality */
{
  coc_bool res_b;
  ring R;
  objtype t1,t2,aux_t;
  object aux_o;

  if (!object_rings_compatible(o1,o2)) return FALSE;
  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1==objtype_NULL) return t2==objtype_NULL;
  if (t2==objtype_NULL) return FALSE;
  if ((t1 != objtype_objtype || t2 != objtype_objtype) &&
      !objects_same_label(o1, o2))
  {
/*  Objects with different tags are different; should comparison be an error? */
/*    math_error=cocoa_error("DIFF_TAGS"); */
    return FALSE;
  }

  if (t1!=t2) {

    if (!objects_eq_comparable(o1,o2)) {
      /* math_error=object_error_types("CANNOT_COMPARE",t1," with ",t2); */
      return FALSE; }

    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) {math_error=o1;return FALSE;}
      res_b=do_is_eq(o1,o2);
      object_unlink(o1); }
    else {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) {math_error=o2;return FALSE;}
      res_b=do_is_eq(o1,o2);
      object_unlink(o2); }

    return res_b;
  }

  switch(t1) {
  case objtype_coeff:
    res_b=(*R->AreEqualsCoeff)(o1->v.coeff_value,o2->v.coeff_value);
    break;
  case objtype_arbint:
    res_b=AreLiEquals(o1->v.arbint_value,o2->v.arbint_value);
    break;
  case objtype_arbrat:
    res_b=AreLrEquals(o1->v.arbrat_value,o2->v.arbrat_value);
    break;
  case objtype_int:
    res_b=o1->v.int_value==o2->v.int_value;
    break;
  case objtype_bool:
    res_b=o1->v.bool_value==o2->v.bool_value;
    break;
  case objtype_zmod:
    res_b = zmod_eq(o1->v.zmod_value,o2->v.zmod_value);
    break;
  case objtype_ratfun:
    res_b = ratfun_eq(o1->v.ratfun_value,o2->v.ratfun_value,R);
    break;
  case objtype_poly:
    res_b = arpoly_eq(o1->v.poly_value,o2->v.poly_value,R);
    break;
  case objtype_pvector:
    res_b = pv_eq(o1->v.pvector_value,o2->v.pvector_value,R);
    break;
  case objtype_ideal:
  case objtype_module:
    res_b = EqModules(o1->v.module_value,o2->v.module_value);
    break;
  case objtype_string:
    res_b = (strcmp(o1->v.string_value,o2->v.string_value)==0);
    break;
  case objtype_objtype:
    res_b = o1->v.objtype_value==o2->v.objtype_value;
    if (res_b && (o1->v.objtype_value==objtype_tagged)) {
      res_b = (strcmp(object_label(o1),object_label(o2)) == 0);
    }
    break;
  case objtype_olist:
    res_b = olist_eq(o1->v.olist_value,o2->v.olist_value);
    break;
  case objtype_omat:
    res_b = omat_eq(o1->v.omat_value,o2->v.omat_value);
    break;
  case objtype_cocfun:
  {
    cocfun f1, f2;
    int pkg_eq, fname_eq;
    f1 = (cocfun)o1->v.cocfun_value;
    f2 = (cocfun)o2->v.cocfun_value;
    pkg_eq = 0;
    if (f1->pkg_id && f2->pkg_id) pkg_eq = (strcmp(f1->pkg_id, f2->pkg_id)==0);
    if (f1->pkg_id == NULL && f2->pkg_id == NULL) pkg_eq = 1;
    fname_eq = (strcmp(f1->fun_id, f2->fun_id) == 0);
    res_b = (pkg_eq && fname_eq);
    break;
  }
  case objtype_record:
    res_b = record_eq(o1->v.record_value, o2->v.record_value);
    break;
  default: {
    math_error=object_error_types("CANNOT_COMPARE",
      object_type(o1)," with ",object_type(o2));
    res_b = FALSE; }
  }
  return res_b;
}

coc_bool do_is_lt(object o1,object o2)
{
  coc_bool res_b;
  objtype t1,t2;
  object aux_o;
  ring R;

  if (!object_rings_compatible(o1,o2)) return FALSE;
  R=object_true_ring(o1,o2);

  t1=object_type(o1);
  t2=object_type(o2);

  if (t1!=t2) {

    /* if (cond) "special cases: non homog operation" ... */

    /* ... else "homog operation" */

    if (!objtype_comparable(t1,t2)) {
      math_error=object_error_types("CANNOT_COMPARE",t1,"<",t2); /* <--- */
      return FALSE; }

    if (objtype_lt(t1,t2)) {
      o1=object_cast(object_link(o1),t2,R);
      if (object_is_error(o1)) {math_error=o1;return FALSE;}
      res_b=do_is_lt(o1,o2);
      object_unlink(o1); }
    else {
      o2=object_cast(object_link(o2),t1,R);
      if (object_is_error(o2)) {math_error=o2;return FALSE;}
      res_b=do_is_lt(o1,o2);
      object_unlink(o2); }

    return res_b;
  }

  
  switch(t1) {
    case objtype_coeff:
      res_b=IsGreaterCoeff(o2->v.coeff_value,o1->v.coeff_value,R) &&
        !(R->AreEqualsCoeff(o2->v.coeff_value,o1->v.coeff_value));
      break;
    case objtype_arbint:
      res_b=IsLiGreater(o2->v.arbint_value,o1->v.arbint_value,R) &&
        !AreLiEquals(o2->v.arbint_value,o1->v.arbint_value);
      break;
    case objtype_arbrat:
      res_b=IsLrGreater(o2->v.arbrat_value,o1->v.arbrat_value,R) &&
        !AreLrEquals(o2->v.arbrat_value,o1->v.arbrat_value);
      break;
    case objtype_int:
      res_b=o1->v.int_value<o2->v.int_value;
      break;
      /*MAX*/
    case objtype_poly: {
      arpoly aux_p1,aux_p2;
      monomial aux_m1, aux_m2;
      int answer;
      aux_p1=o1->v.poly_value;
      aux_p2=o2->v.poly_value;
      if (arpoly_is_term(aux_p1,R) && arpoly_is_term(aux_p2,R)) {
        aux_m1=arpoly_leading_monomial(aux_p1,R);
        aux_m2=arpoly_leading_monomial(aux_p2,R);
        answer=monomial_which_gt(aux_m1,aux_m2,R);
        switch(answer) {
        case ANSWER_BOTH: 
        case ANSWER_FIRST: res_b=FALSE; break;
        case ANSWER_SECOND: res_b=TRUE; break;
          break; }
        monomial_free(aux_m1,R);
        monomial_free(aux_m2,R); }
      else {
        math_error=object_error_types("CANNOT_COMPARE",t1,"<",t2);
        error_set_this_param(string2cterm("(expected terms)"));
        res_b=FALSE; } }
      break;
      /*MAX*/
    case objtype_ideal:
    case objtype_module:
      res_b=ModuleIsContained(o1->v.module_value,o2->v.module_value);
      break;
    case objtype_string:
      res_b = (strcmp(o1->v.string_value,o2->v.string_value)<0);
      break;
    case objtype_objtype:
      res_b = objtype_lt(o1->v.objtype_value,o2->v.objtype_value);
      break;
 
   /*MAX*/
   
    case objtype_pvector:
    {
      pvector aux_pv1, aux_pv2;
      monomial aux_m1, aux_m2;
      int answer, aux_pos1, aux_pos2;

      aux_pv1=o1->v.pvector_value;
      aux_pv2=o2->v.pvector_value;
      if (pvector_is_term(aux_pv1,R) && pvector_is_term(aux_pv2,R))
      {
        pv_leading_monomial(aux_pv1,&aux_m1,&aux_pos1,R);
        pv_leading_monomial(aux_pv2,&aux_m2,&aux_pos2,R);
        if (ring_mordering(R)==TO_POS)
        {
          answer=monomial_which_gt(aux_m1,aux_m2,R);
          monomial_free(aux_m1,R);
          monomial_free(aux_m2,R);     
          if (answer == ANSWER_BOTH) res_b = aux_pos1>aux_pos2;
          else res_b = (answer == ANSWER_SECOND);
        }
        else /* ring_mordering(R) == POS_TO */
        {
          if (aux_pos1<aux_pos2) {res_b=FALSE; break;}
          if (aux_pos1>aux_pos2) {res_b=TRUE; break;}
          pv_leading_monomial(aux_pv1,&aux_m1,&aux_pos1,R);
          pv_leading_monomial(aux_pv2,&aux_m2,&aux_pos2,R);
          answer=monomial_which_gt(aux_m1,aux_m2,R);
          monomial_free(aux_m1,R);
          monomial_free(aux_m2,R);     
          res_b = (answer == ANSWER_SECOND);
        }
      }
      else
      {
        math_error=object_error_types("CANNOT_COMPARE",t1,"<",t2);
        error_set_this_param(string2cterm("(expected terms)"));
        res_b=FALSE;
      }
    }
    break;
   /*MAX*/
  default: {
    math_error=object_error_types("CANNOT_COMPARE",t1,"<",t2);
    res_b = FALSE; }
    }
  return res_b;
}

object do_eval_leading_term(object o, ring R)
{
  object res_o;

  if (object_ring(o) != ring_NULL) R = object_ring(o);
  switch (object_type(o)) {
  case objtype_int:
  case objtype_zmod:
  case objtype_arbrat:
  case objtype_arbint:
    res_o=object_link(object_ONE);
    break;
  case objtype_poly: {
    arpoly aux_p;
    aux_p=arpoly_only_leading_term(o->v.poly_value,R);
    res_o=arpoly2object(aux_p,R); }
    break;
  case objtype_pvector: {
    pvector aux_v;
    aux_v=pv_only_leading_term(o->v.pvector_value,R);
    res_o=pv2object(aux_v,R); }
    break;
  case objtype_ideal:
    res_o=object_cast(
      module2object(module_leading_terms(o->v.module_value, R),R),
      objtype_ideal,R);
    break;
  case objtype_module:
    res_o=module2object(module_leading_terms(o->v.module_value, R),R);
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;  
}

object do_eval_leading_coeff(object o, ring R)
{
  object res_o;

  if (object_ring(o) != ring_NULL) R = object_ring(o);
  switch (object_type(o)) {
  case objtype_int:
  case objtype_zmod:
  case objtype_arbrat:
  case objtype_arbint: {
      arpoly aux_p;
      coeff aux_c;
      aux_p=object_cast_arpoly(object_link(o),R);
      if (math_error != object_NULL) res_o=math_error;
      else {
	aux_c=arpoly_leading_coeff(aux_p,R);
	arpoly_free(aux_p,R);
	res_o=coeff2object(aux_c,R); 
      }
    }
    break;
  case objtype_poly: {
    coeff aux_c;
    aux_c=arpoly_leading_coeff(o->v.poly_value,R);
    res_o=coeff2object(aux_c,R); }
    break;
  case objtype_pvector: {
    coeff aux_c;
    aux_c=pv_leading_coeff(o->v.pvector_value, R);
    res_o=coeff2object(aux_c,R); }
    break;
  default:
    res_o=cocoa_error("BAD_PARAMS");
  }
  return res_o;  
}
