/* matrix.c : 0.5 : 13 Oct 93 : ADT matrices (and submatrices) */

/*------------------------------------------------------------------------------ 
        Part of project CoCoa3

------------------------------------------------------------------------------*/

#include "cocdefs.h"

#ifndef MATRIX_H
# include "matrix.h"
#endif

#ifndef IVECTORS_H
# include "ivectors.h"
#endif

matrix matrix_init(int m, int n)
{
  matrix res_M;
  register int i;
  res_M=malloc_matrix(m);
  for (i=0;i<=m-1;i++) matrix_set_row(res_M,i,malloc_matrix_row(n));
  return res_M;
}

void matrix_free(matrix M, int m, int n)
{
  register int i;
  for (i=0;i<m;i++) free_matrix_row(n,matrix_row(M,i));
  free_matrix(m,M);
}

matrix_el matrix_det(matrix M,int n)
{
  submat aux_S;
  matrix_el res_el;
  aux_S=matrix2submat(M,n,n);
  res_el=submat_det(aux_S);
  submat_free(aux_S);
  return res_el;
}

matrix matrix_adjoint(matrix M,int n)
{
  matrix res_M;
  matrix_el el;
  int i,j;
  submat aux_S,S;
  res_M=matrix_init(n,n);
  S=matrix2submat(M,n,n);
  for (i=0;i<n;i++)
    for (j=0;j<n;j++) {
      aux_S=submat_elim_rc(S,j,i);
      el=submat_det(aux_S);
      submat_free(aux_S);
      if ((i+j) & 1) el=-el;
      matrix_set_elem(res_M,i,j,el); }
  submat_free(S);
  return res_M;
}

submat submat_mk(matrix M, int m, int n, ivector rr, ivector cc)
{
  submat res_S;
  res_S=malloc_submat();
  submat_set_elems(res_S,M);
  submat_set_num_rows(res_S,m);
  submat_set_num_cols(res_S,n);
  submat_set_rows(res_S,rr);
  submat_set_cols(res_S,cc);
  return res_S;  
}

submat matrix2submat(matrix M,int m,int n)
{
  ivector rr,cc;
  register int i;
  rr=iv_init(m);
  cc=iv_init(n);
  for (i=1;i<=m;i++) iv_set_nth(rr,i,i-1);
  for (i=1;i<=n;i++) iv_set_nth(cc,i,i-1);
  return submat_mk(M,m,n,rr,cc);
}

void submat_free(submat S)
{
  iv_free(submat_rows(S));
  iv_free(submat_cols(S));
  free_submat(S);
}

matrix submat2matrix(submat S)
{
  matrix res_M;
  int m,n,i,j;
  m=submat_num_rows(S);
  n=submat_num_cols(S);
  res_M=matrix_init(m,n);
  for (i=0;i<m;i++)
    for (j=0;j<n;j++)
      matrix_set_elem(res_M,i,j,submat_elem(S,i,j));
  return res_M;
}

submat submat_elim_row(submat S, int row)
{
  ivector rr,cc;
  rr=iv_elim_nth(submat_rows(S),row+1);
  cc=iv_dup(submat_cols(S));
  return 
    submat_mk(submat_elems(S),submat_num_rows(S)-1,submat_num_cols(S),rr,cc);
}

submat submat_elim_col(submat S, int col)
{
  ivector rr,cc;
  cc=iv_elim_nth(submat_cols(S),col+1);
  rr=iv_dup(submat_rows(S));
  return 
    submat_mk(submat_elems(S),submat_num_rows(S),submat_num_cols(S)-1,rr,cc);
}

submat submat_elim_rc(submat S,int row,int col)
{
  submat aux_S,res_S;
  aux_S=submat_elim_col(S,col);
  res_S=submat_elim_row(aux_S,row);
  submat_free(aux_S);
  return res_S;
}

matrix_el submat_det(submat S)
{
  matrix_el res_el;
  int n,j;
  submat aux_S;
  n=submat_num_cols(S);
  if (n==1) return submat_elem(S,0,0);
  res_el=0;
  n--;
  for (j=0;j<=n;j++) {
    aux_S=submat_elim_rc(S,n,j);
    if (submat_elem(S,n,j) != 0) {
      matrix_el det;
      det=submat_det(aux_S);
      if (n+j & 1) det=-det;
      res_el+=det*submat_elem(S,n,j); }
    submat_free(aux_S); }
  return res_el;
}
