/*Here there are some procedure waiting to be put in the appropriate files.
The linking of the procedure to cocoa is described in the file
note/addfunction.*/

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "ctbuf.h"
#include "cterm.h"
#include "coeffs.h"
#include "evector.h"
#include "ivectors.h"
#include "gb.h"
#include "gb_kf.h"
#include "module.h"
#include "monomial.h"
#include "mathcoc.h"
#include "object.h"
#include "olist.h"
#include "operator.h"
#include "options.h"
#include "pair.h"
#include "pvector.h"
#include "polys.h"
#include "pseries.h"
#include "ring.h"
#include "terms.h"
#include "vlist.h"
#include "max.h"
#include "hook.h"

/* --------------------  Procedures for Monomials, Support, Coefficients------*/

/*returns the list of the monomials of the arpoly p.
For a full description of the extremal cases, see caboara.ts */

olist monomials_of_this_arpoly(arpoly p, ring R)

{ monomial m,m1;
  olist ol; 
  int i,n;
  arpoly p1;
  
    n=arpoly_len(p,R);
    ol=olist_init(n);
    for (i=1; i<=n; i++){
      m1=arpoly_leading_monomial(p,R);
      p1=arpoly_this_monomial2arpoly(m1,R);
      olist_set_nth(ol,i,arpoly2object(p1,R));
      m = arpoly_leading_monomial(p,R);
      monomial_change_sign(m,R);
      p=arpoly_add_this_monomial(p,m,R);};
    return ol;
}


/*returns the list of the power products of the arpoly p.
For a full description of the extremal cases, see caboara.ts */

olist support_of_this_arpoly(arpoly p, ring R)

{ monomial m;
  olist ol; 
  int i,n;
  
    n=arpoly_len(p,R);
    ol=olist_init(n);
    for (i=1; i<=n; i++){
      olist_set_nth(ol,i,arpoly2object(arpoly_this_term2arpoly(arpoly_leading_term(p,R),R),R));
      m = arpoly_leading_monomial(p,R);
      monomial_change_sign(m,R);
      p=arpoly_add_this_monomial(p,m,R);};
    return ol;
}


/*Returns the list of the coefficients of the arpoly p.
For a full description of the extremal cases, see caboara.ts */

olist coeffs_of_this_arpoly(arpoly p, ring R)

{ monomial m;
  olist ol; 
  int i,n;
  
    n=arpoly_len(p,R);
    ol=olist_init(n);
    for (i=1; i<=n; i++){
      olist_set_nth(ol,i,coeff2object(arpoly_leading_coeff(p,R),R));
      m = arpoly_leading_monomial(p,R);
      monomial_change_sign(m,R);
      p=arpoly_add_this_monomial(p,m,R);};
    return ol;
}

/*Returns the olist of the mmonomials of the pvector pv1*/
olist monomials_of_this_pv(pvector pv, ring R)

{ olist ol; 
  int r,pos;
  pvector aux_pv;
  object o;
  monomial m;
  pvcomp pvc;
    
    r=pv_rank(pv)-1;
    ol=olist_init(0);
    while (!pv_is_NULL_or_zero(pv)) {
        pv_extract_lm(pv,&m,&pos,R);
	pvc=this_monomial2pvcomp(m,R);
        aux_pv=pv_this_pvcomp2pv(pvc,pos,r+1,R);
        o=pv2object(aux_pv,R);
        ol=olist_append_this(ol,o);
    };
    pv_free(pv,R);
    return ol;
}

/*Returns the olist of the mterms of the pvector pv*/
olist support_of_this_pv(pvector pv, ring R)

{ olist ol; 
  int r,n,pos;
  pvector aux_pv;
  object o;
  monomial m;
  term t;
  
    r=pv_rank(pv)-1;
    ol=olist_init(0);
    while (!pv_is_NULL_or_zero(pv)) {
        pv_extract_lm(pv,&m,&pos,R);
	t=monomial_term(m);
        aux_pv=this_term2pv(term_dup(t),pos,r+1,R);
	monomial_free(m,R);
        o=pv2object(aux_pv,R);
        ol=olist_append_this(ol,o);
    };
    pv_free(pv,R);
    return ol;
}


/*Returns the olist of the coefficients of the pvector pv1*/
olist coeffs_of_this_pv(pvector pv, ring R)

{ olist ol; 
  int r,pos;
  object o;
  monomial m;
  coeff c;
    
    r=pv_rank(pv)-1;
    ol=olist_init(0);
    while (!pv_is_NULL_or_zero(pv)) {
        pv_extract_lm(pv,&m,&pos,R);
	c=coeff_dup(monomial_coeff(m),R);
        monomial_free(m,R);
        o=coeff2object(c,R);
        ol=olist_append_this(ol,o);
    };
    pv_free(pv,R);
    return ol;
}



/*Returns the poly which is in the first non null position of v*/
/*If the vector is completely zero then the value returned is 0 */
arpoly pv_FirstNonZero(pvector v, ring R)
{ 
  int last,j;
  arpoly p;
  
  last=pv_rank(v)-1;
  j=0;
  while (j<last && (pv_nth(v,j)==NULL)){j++;}
  p=pvcomp2arpoly(pv_nth(v,j),R);
  return p;
}


/*Returns the first non null position of v*/
/*If the vector is completely zero then the position returned is 0 */
int pv_FirstNonZeroPos(pvector v, ring R)
{ 
  int last,j;
  
  last=pv_rank(v)-1;
  j=0;
  while (j<=last && (pv_nth(v,j)==NULL)){j++;}
  if (j > last) return 0;
  return j+1;
}

/*Checks is the arpoly pvcomp p is a term */
coc_bool arpoly_is_term(arpoly p, ring R)
{ coc_bool b;
  coeff c,c1;
  
  if (  arpoly_is_zero(p,R) || (arpoly_len(p,R)!=1)) return FALSE;
  c=arpoly_leading_coeff(p,R);
  c1=coeff_one(R);
  b=coeff_eq(c1,c,R);
  coeff_free(c,R);  
  coeff_free(c1,R);  
  return b;
}

/*Checks if the pvector pvcomp p is a term */
coc_bool pvcomp_is_term(arpoly p, ring R)
{ coc_bool b;
  coeff c,c1;
  
  if (pvcomp_is_zero(p,R) || pvcomp_len(p,R)!=1) return FALSE;
  c=pvcomp_leading_coeff(p,R);
  c1=coeff_one(R);	
  b=coeff_eq(c1,c,R);
  coeff_free(c,R);  
  coeff_free(c1,R);  
  return b;
}


coc_bool pvcomp_is_zero(pvcomp p, ring R) {
  return p==NULL;
}

/*Checks if is the pvector pv is a mterm : i.e. (0,0,term,0,0)*/
coc_bool pvector_is_term(pvector pv, ring R)
{ int i, nfound;
  
  i=0;
  nfound=0;
  
  while (i<pv_rank(pv) && nfound<2) {
    if (!pvcomp_is_term(pv_nth(pv,i),R) && !pvcomp_is_zero(pv_nth(pv,i),R)) 
       {return FALSE;}
    if (pv_nth(pv,i)!=NULL) {nfound++;}
    i++;
  }
  return nfound==1;
}

/*Returns the coeff of the term t of the arpoly p*/
coeff this_arpoly_coeff_of_term(arpoly p, term t, ring R) 

{  monomial m;
   coeff res_coeff;

    while (!arpoly_is_zero(p,R)) {
      m = arpoly_leading_monomial(p,R);
      if (term_eq(monomial_term(m),t)) {
        res_coeff=monomial_coeff(m);
        monomial_set_coeff(m,coeff_one(R));
        monomial_free(m,R);
        arpoly_free(p,R);
        return res_coeff;}
      monomial_change_sign(m,R);
      p=arpoly_add_this_monomial(p,m,R);        
    };
    return coeff_zero(R);
}

/*Returns the coeff of the mterm pvt of the pvector pv*/
coeff this_pvector_coeff_of_term(pvector pv, pvector pvt, ring R) 
{
   monomial aux_m, pvt_mon;
   coeff res_coeff;
   int aux_pos, pvt_pos;
   
    pv_leading_monomial(pvt,&pvt_mon,&pvt_pos,R);

    while (!pv_is_NULL_or_zero(pv)) {
      pv_extract_lm(pv,&aux_m,&aux_pos,R);
      if (term_eq(monomial_term(aux_m),monomial_term(pvt_mon))
          && aux_pos==pvt_pos) {
        res_coeff=monomial_coeff(aux_m);
        monomial_set_coeff(aux_m,coeff_one(R));
        monomial_free(aux_m,R);
        monomial_free(pvt_mon,R);
        pv_free(pv,R);
        return res_coeff;}
      else monomial_free(aux_m,R);
    };
    monomial_free(pvt_mon,R);
    return coeff_zero(R);
}




pvector nr(pvector pvf, module m, ring R)

{ evector ev;
  pvector pv;
  int b;
  
  ev=ev_init(pv_dup(pvf,R),R);
  gb_set_kf(m,gb_kf_Do_Get_Rule,gb_kf_Get_Rule_First);
  gb_set_kf(m,gb_kf_Do_Apply_Rule,gb_kf_Apply_Rule_GB);
  gb_set_kf(m,gb_kf_Do_Add_To_Rules,gb_kf_Add_To_Rules_Append);  
  gb_Init_NF(m);
  if (gb_Rules(m)==vlist_NULL) {
    gb_Init_Interreduce(m);
  }
  
  b=gb_Reduce_Full(m,&ev);
  pv=ev_whole_vector(ev,R);
  ev_unlink(ev,R);
  return pv;
}
