/* $Id: memory.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "memoryco.h"
#include "object.h"
#include "olist.h"
#include "operator.h"
#include "utils.h"

/* ADT mem */

void mem_assign(loc l, object o)
{
  object_unlink(*l);
  *l=object_link(o);
}

/* --------------------------------------------- */

int env_error=0;
unsigned int env_incremental=0; /* used to generate new variable names */

char * env_newvar(char * id)
{
  char buffer[50];
  char * res_id;
  sprintf(buffer,"%d",env_incremental++);
  res_id=malloc_str(strlen(id)+strlen(buffer)+1);
  sprintf(res_id,"%s#%s",id,buffer);
  return res_id;
}

env env_init(void)
{
  return (env)env_mk_slot("#"); /* dummy identifier */
}

int env_len(env e)
{
  env_slot aux_slot;
  int res_n=0;
  if (e==NULL) return -1;
  aux_slot=(env_slot)e;
  while (aux_slot->next!=NULL) {
    aux_slot=aux_slot->next;
    res_n++; }
  return res_n;  
}

loc env_declare(env * e,char * id)
{
  env_slot aux_slot;
  aux_slot=env_mk_slot(id);
  env_insert_slot(e,id,aux_slot);
  return aux_slot->address;
}

void env_assign(env e,char * id,object o)
{
  loc aux_l;
  aux_l=env_get_loc(e,id);
  if (aux_l!=NULL) mem_assign(aux_l,o);
}

void env_declare_and_assign(env * e,char * id,object o)
{
  loc aux_l;
  if ((aux_l=env_get_loc(*e,id))==NULL) aux_l=env_declare(e,id);
  mem_assign(aux_l,o);
}


void env_delete_var(env *e,char * id)
{
  /* e != NULL */
  env_slot aux_slot,aux_slot_previous=NULL;
  int cmp;
  if (id == NULL) return;
  aux_slot = (env_slot)*e;
  while (aux_slot!=NULL) {
    cmp=strcmp(id,aux_slot->id);
    if (cmp>0) {
      aux_slot_previous = aux_slot;
      aux_slot = aux_slot->next; }
    else if (cmp==0) {
      if (aux_slot_previous!=NULL) 
        aux_slot_previous->next = aux_slot->next;
      else
        *e = (env)aux_slot->next;
      env_free_slot(aux_slot);
      break; }
    else break;
  } /* while */
}

void env_insert_slot(env *e,char * id,env_slot slot)
{
  /* e != NULL */
  env_slot aux_slot,aux_slot_previous=NULL;
  int cmp;
  coc_bool not_inserted;
  not_inserted=TRUE;
  aux_slot = (env_slot)*e;
  while (aux_slot!=NULL && not_inserted) {
    cmp=strcmp(id,aux_slot->id);
    if (cmp>0) {
      aux_slot_previous = aux_slot;
      aux_slot = aux_slot->next; }
    else if (cmp==0) {
      if (aux_slot_previous!=NULL) 
        aux_slot_previous->next = slot;
      else
        *e = (env)slot;
      slot->next = aux_slot->next;
      env_free_slot(aux_slot);
      not_inserted=FALSE; }
    else { /* cmp<0 */
      if (aux_slot_previous!=NULL) 
        aux_slot_previous->next = slot;
      else
        *e = (env)slot;
      slot->next = aux_slot;
      not_inserted=FALSE; }
  } /* while */
  if (not_inserted) /* i.e. slot will be the last */
    aux_slot_previous->next=slot;
}

env_slot env_search_slot(env e,char * id)
{
  env_slot aux_slot,res_slot=NULL;
  int cmp;
  coc_bool not_found=TRUE;
  aux_slot = (env_slot)e;
  while (aux_slot!=NULL && not_found) {
    cmp=strcmp(id,aux_slot->id);
    if (cmp>0)
      aux_slot = aux_slot->next;
    else if (cmp==0) {
      not_found = FALSE;
      res_slot = aux_slot; }
    else /* not existent: force exit */
      aux_slot=NULL;
  } /* while */
  return res_slot;
}

int env_var_rc(env e, char * id)
{
  loc aux_l;
  aux_l=env_get_loc(e,id);
  if (aux_l==NULL) return 0;
  return object_rc(mem_this_value(aux_l));
}

void env_free_slot(env_slot slot)
{
  loc l;
  register int i;
  l=slot->address;
  object_unlink(*l);
  free_loc(l);
  free_str(strlen(slot->id),slot->id);
  free_env_slot(slot);
}

env_slot env_mk_slot(char * id)
{
  env_slot res_slot;
  loc l;
  res_slot=malloc_env_slot();
  res_slot->address=malloc_loc();
  *(res_slot->address)=object_link(object_null);
  res_slot->id=my_strdup(id);
  res_slot->next=NULL;
  return res_slot;
}

env_slot env_slot_dup(env_slot slot)
{
  env_slot res_slot;
  res_slot=malloc_env_slot();
  res_slot->address=malloc_loc();
  *(res_slot->address)=mem_value(slot->address);
  /* *(res_slot->address)=object_dup(mem_value(slot->address)); */ /* TODO ? */
  res_slot->id=my_strdup(slot->id);
  res_slot->next=NULL;
  return res_slot;
}

void env_free(env e)
{
  env_slot aux_slot,aux_slot2;
  aux_slot2=(env_slot)e;
  while (aux_slot2!=NULL) {
    aux_slot=aux_slot2;
    aux_slot2=aux_slot2->next;
    env_free_slot(aux_slot); }
}

object env_get_value(env e,char * id)
{
  int * arrinfo;
  loc aux_l;
  if (id==NULL) return object_NULL;
  aux_l=env_get_loc(e,id);
  if (aux_l==NULL) return object_NULL;
  return mem_value(aux_l);
}

loc env_get_loc(env e,char * id)
{
  int * arrinfo;
  int sz;
  env_slot aux_slot;
  register int i;
  loc res_l;
  env_error = env_ERROR_NONE;
  aux_slot = env_search_slot(e,id);
  if (aux_slot==NULL) {
    env_error = env_ERROR_ID_NOT_DECLARED;
    return NULL; }
  res_l=aux_slot->address;
  return res_l;
}

env env_dup(env e)
{
  env res_e;
  env_slot aux_slot,aux_slot_e;
  if (e==NULL) return NULL;
  aux_slot_e=(env_slot)e;
  aux_slot=env_slot_dup(aux_slot_e);
  res_e=(env)aux_slot;
  while (aux_slot_e->next!=NULL) {
    aux_slot_e=aux_slot_e->next;
    aux_slot->next=env_slot_dup(aux_slot_e);
    aux_slot=aux_slot->next; }
  return res_e;
}

void env_copy_var(env e1, env * e2, char * id1, char * id2)
{
  char * save_id1;
  env_slot aux_slot1, aux_slot2;
  aux_slot1 = env_search_slot(e1,id1);
  if (aux_slot1 == NULL) { return; }

  save_id1 = aux_slot1->id;
  aux_slot1->id=NULL;
  aux_slot2 = env_slot_dup(aux_slot1);
  aux_slot1->id=save_id1;

  aux_slot2->id=my_strdup(id2);
  env_insert_slot(e2,id2,aux_slot2);
}

void env_rename_var(env e1, char * id1, char * id2)
{
  char * save_id;
  env_slot aux_slot;
  aux_slot = env_search_slot(e1,id1);
  if (aux_slot == NULL) { return; }
  save_id = aux_slot->id;
  aux_slot->id=my_strdup(id2);
  free_str1(save_id);
}

olist env2olist_aux(env e, int * n, coc_bool pairs)
{
  olist aux_ol,pair;
  object aux_o;
  env_slot aux_slot;
  char * id;
  aux_ol=olist_init(env_len(e));
  aux_slot=(env_slot)e;
  aux_slot=aux_slot->next; /* skip dummy var "#" */
  *n=0;
  while (aux_slot != env_slot_NULL) {
    if (*(env_slot_id(aux_slot)) != '_') {
      id=my_strdup(env_slot_id(aux_slot));
      aux_o=string2object(id);
      if (pairs) {
        pair=olist_init(2);
	olist_set_nth(pair,1,aux_o);
	olist_set_nth(pair,2,mem_value(aux_slot->address));
	olist_set_nth(aux_ol,++(*n),olist2object(pair));
      } else {  
        olist_set_nth(aux_ol,++(*n),aux_o); 
      }
    }
    aux_slot=aux_slot->next;
  }
  return aux_ol;
}

olist env2olist_pairs(env e) {
  olist aux_ol,res_ol;
  int n,i;
  aux_ol=env2olist_aux(e,&n,TRUE);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) {
    olist_set_nth(res_ol,i,olist_nth(aux_ol,i));
  }
  free_olist(olist_len(aux_ol),aux_ol);
  return res_ol;
}

olist env2olist(env e) {
  olist aux_ol,res_ol;
  int n,i;
  aux_ol=env2olist_aux(e,&n,FALSE);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) olist_set_nth(res_ol,i,olist_nth(aux_ol,i));
  free_olist(olist_len(aux_ol),aux_ol);
  return res_ol;
}

void printf_env(env e)
{
  cterm t;
  t=env2cterm(e);
  put_cterm(t);
  cterm_free(t);
}

void printf_env_slot(env_slot slot)
{
  cterm t;
  t=env_slot2cterm(slot);
  put_cterm(t);
  cterm_free(t);
}

cterm env2cterm(env e)
{
  env_slot aux_slot;
  ctbuf aux_b;
  aux_slot=(env_slot)e;
  aux_slot=aux_slot->next; /* skip dummy var "#" */
  aux_b=ctbuf_empty();
  while (aux_slot!=NULL) {
    ctbuf_append(aux_b,env_slot2cterm(aux_slot));
    aux_slot=aux_slot->next; }
  return ctbuf_mk_cterm(aux_b,op_ENV);
}

cterm env_slot2cterm(env_slot slot)
{
  cterm params[2];
  object aux_o;
#ifdef OBJECT_DEBUG
  char buffer[100];
#endif

  aux_o=mem_value(slot->address);

#ifdef OBJECT_DEBUG
  sprintf(buffer,"%s [%d]",slot->id,aux_o->num);
  params[0]=cterm_mk_NBS(buffer);
#else
  params[0]=cterm_mk_NBS(slot->id);
#endif

  params[1]=object2cterm(aux_o);
  object_unlink(aux_o);
  return cterm_mk_classic(op_ENV_SLOT,2,params);
}

/* ADT env stack */

env_stack env_the_stack=NULL;

void env_push(env e)
{
  env_stack aux_stack;
  aux_stack=malloc_env_stack_elem();
  aux_stack->top=e;
  aux_stack->pop=env_the_stack;
  env_the_stack=aux_stack;
}

env env_pop(void)
{
  env res_e;
  env_stack aux_stack;
  if (env_the_stack==NULL) return NULL;
  res_e=env_the_stack->top;
  aux_stack=env_the_stack;
  env_the_stack=env_the_stack->pop;
  free_env_stack_elem(aux_stack);
  return res_e;
}
