/* Id */

#ifndef MEMORY_H
#define MEMORY_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#ifndef COCTYPES_H
# include "coctypes.h"
#endif

#define malloc_loc() (loc)mymalloc(sizeof(object),"loc");
#define free_loc(p) myfree(sizeof(object),p,"loc");

#define malloc_env_slot() (env_slot)mymalloc(sizeof(env_slot_info),"slot")
#define free_env_slot(p) myfree(sizeof(env_slot_info),p,"slot")

#define malloc_env_stack_elem() \
  (env_stack)mymalloc(sizeof(env_stack_info),"env_stack_elem")
#define free_env_stack_elem(p) \
  myfree(sizeof(env_stack_info),p,"env_stack_elem")

char * env_newvar(char * id);
void mem_assign(loc l, object o);

#define mem_value(l) object_link(*(l))
#define mem_this_value(l) (*(l))

/*------------------------------------------------------------------------------

   An environment is a map from names to locations.

------------------------------------------------------------------------------*/

#define env_is_declared(e,id) (env_search_slot(e,id)!=NULL)

#define env_slot_id(e) ((e)->id)
#define env_slot_next(e) ((e)->next)
#define env_slot_set_next(e,s) (e)->next=(s)

#define env_slot_loc(e) ((e)->address)

env env_init(void);
/* initialize and returns a new environment */

int env_len(env e);

loc env_declare(env *e,char * id);
/* allocates the space for the variable id of type t */
/* NOTE: if id is already declared then it is replaced by the new declaration */

void env_assign(env e,char * id,object o);
/* e.id[ndx]=o. If id is not defined in e then the variable env_error is set. */
/* env_ERROR_NONE: all right */
/* env_ERROR_ID_NOT_DECLARED: id was not declared */
/* env_ERROR_WRONG_NUMBER_OF_INDICES: with respect to the declaration */
/* env_ERROR_WRONG_INDEX: index out of declared range */

/* env_{declare,assign,declare_and_assign}_var are for the special case of
   array without indices */

loc env_get_loc(env e,char * id);

env_slot env_mk_slot(char * id);

void env_declare_and_assign(env * e,char * id,object o);

void env_delete_var(env *e,char * id);

void env_insert_slot(env *e,char * id,env_slot slot);
/* insert slot in the environment e */
/* NOTE: e must be non empty */

int env_var_rc(env e, char * id);
/* returns the reference count of id in e */
/* if id is not declared returns 0 */

void env_free_slot(env_slot slot);

env_slot env_search_slot(env e,char * id);
/* search the slot of e with ident = id */
/* return NULL id is not declared */

void env_free(env e);
/* free the space used for the environment e */

env_slot env_slot_dup(env_slot slot);

object env_get_value(env e,char * id);
/* return e[id] */

void env_rename_var(env e1, char * id1, char * id2);

void env_copy_var(env e1, env * e2, char * id1, char * id2);
/* e2[id2] := e1[id1] */

olist env2olist(env e);
olist env2olist_pairs(env e);

/* TODO: remove? forse servono per salvare un anello */
void printf_env(env e);
void printf_env_slot(env_slot slot);
cterm env2cterm(env e);
cterm env_slot2cterm(env_slot slot);
/* END TODO */

void env_push(env e);
env env_pop(void);
env env_dup(env e);
/* return a duplicate of the env e if possible, return env_NULL otherwise */

#endif
