/* $Id: module.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "ctbuf.h"
#include "cterm.h"
#include "evector.h"
#include "ivectors.h"
#include "gb.h"
#include "memoryco.h"
#include "module.h"
#include "monomial.h"
#include "object.h"
#include "olist.h"
#include "operator.h"
#include "options.h"
#include "pair.h"
#include "pvector.h"
#include "polys.h"
#include "pseries.h"
#include "ring.h"
#include "terms.h"
#include "vlist.h"

extern object math_error;

/* TODO: optimize building directly the result */
module ideal_pow( module J, int theExp, ring R) {
  module aux_J,res_J;
  module auxId;

  if (theExp < 1) {
    res_J = module_freemod(1,R);
    return res_J;
  }
  
  if (theExp == 2) return module_mult_by_ideal(J,J,R);

  res_J = module_dup( J, R );
  while (theExp-->1) {
    aux_J = module_mult_by_ideal(res_J,J,R);
    module_unlink(res_J,R);
    res_J = aux_J;
  }
  return res_J;
}

object module_link_kf(module M, olist ol, ring R)
{
  int i,n,aux_n;
  n=MIN(olist_len(ol),gb_NUM_MODULE_KEY_FUNCTIONS);
  for (i=1;i<=n;i++) {
    aux_n=object_cast_int(object_link(olist_nth(ol,i)));
    if (math_error != object_NULL) return math_error;
    gb_set_kf(M,i-1,aux_n);
  }
  return object_link(object_null);
}

module module_Clone(module M)
{
  module res_M;
  ring R;
  vlist aux_vl;
  R = module_ring(M);
  if (M == module_NULL) return module_NULL;
  res_M=module_init(module_rank(M),R);
  module_set_flags(res_M,module_flags(M));
  res_M->syz_of_gens = module_NULL; /*JAA*/
  module_set_zys(res_M,M);
  module_set_shifts(res_M,shifts_NULL);
  module_set_rank(res_M,0);
  module_set_gens(res_M,vlist_init());
  module_set_mingens(res_M,vlist_init());
  module_set_gbasis(res_M,vlist_init());
  ring_unlink(module_ring(res_M));
  module_set_ring(res_M,ring_link(module_ring(M)));
  module_set_DegTrunc(res_M,module_DegTrunc(M));
  module_set_SyzCompsNum(res_M,module_SyzCompsNum(M));
  if (module_ResTrunc(M) != trunc_NONE)
    module_set_ResTrunc(res_M,module_ResTrunc(M)-1);
  if (module_RegTrunc(M) != trunc_NONE)
    module_set_RegTrunc(res_M,module_RegTrunc(M)+1);
  module_gb_set_initialized(res_M,TRUE);
  gb_set_Rules(res_M,vlist_init());
  gb_set_IVs(res_M,vlist_init());
  gb_set_IPs(res_M,vpairs_NULL);
  gb_kf_dup(res_M,M);
  return res_M;
}

int module_res_length(module M)
{
  int i;
  i=0;
  while (!module_is_zero(module_syz(M))) {
    i++;
    M=module_syz(M);
  }
  return i;
}

module module_init(int rk, ring R)
{
  module res_M;
  int i;
  res_M=malloc_module();
  module_set_flags(res_M,0);
  module_set_rc(res_M,1);
  module_set_ring(res_M,ring_link(R));
  module_set_rank(res_M,rk);
  module_set_shifts(res_M,shifts_NULL);
  module_set_gens(res_M,vlist_NULL);
  module_set_gbasis(res_M,vlist_NULL);
  module_set_pseries(res_M,pseries_NULL);
  module_set_mvpseries(res_M,mvpseries_NULL);
  res_M->syz_of_gens = module_NULL; /*JAA*/
  module_set_syz(res_M,module_NULL);
  module_set_zys(res_M,module_NULL);
  module_set_DegTrunc(res_M,trunc_NONE);
  module_set_SyzCompsNum(res_M,trunc_NONE);
  module_set_ResTrunc(res_M,trunc_NONE);
  module_set_RegTrunc(res_M,trunc_NONE);
  module_set_ElimIndets(res_M,0);
  module_set_mingens(res_M,vlist_NULL);
  module_set_discrepancy(res_M,-1);
  module_gb_set_initialized(res_M,FALSE);
  gb_set_Rules(res_M,vlist_NULL);
  gb_set_CV(res_M,ev_NULL);
  gb_set_IVs(res_M,vlist_NULL);
  gb_set_IPs(res_M,vpairs_NULL);
  gb_kf_init(res_M);
  for (i=0;i<gb_NUM_STATS;i++) module_set_stat(res_M,i,0);
  return res_M;
}

void module_compute_shifts_of_syz(module M,ring R)
{
  module Syz;
  int n,p;
  shift s;
  shifts aux_ss,aux_ss1;
  register int i;
  evector aux_ev;
  vlist_atom aux_atom;
  vlist aux_vl;

  Syz=module_syz(M);
  if (Syz==module_NULL) return;
  if (module_shifts(M) != shifts_NULL) shifts_unlink(module_shifts(M));
  n=module_rank(Syz);
  aux_ss=shifts_init(n);

  aux_vl=module_gens(M);
  aux_atom=vlist_first(aux_vl);
  aux_ss1=module_shifts(M);
  for (i=1;i<=n;i++) {
    aux_ev=vlist_atom_ev(aux_atom);
    s=ev_lt(aux_ev);
    if (aux_ss1 != shifts_NULL) 
      s = term_mult(s,shifts_nth(aux_ss1,ev_lmpos(aux_ev)+1));
    else s=term_dup(s);
    shifts_set_nth(aux_ss,i,s);
    aux_atom=vlist_atom_next(aux_atom);
  }
  module_set_shifts(Syz,aux_ss);
}

void module_free( module M, ring R)
{
  if (M!=module_NULL) {
    if (module_rc(M)!=ZERO) {
      my_fputs("INTERNAL ERROR: attempt to free a referenced module\n",stdout);
      return;
    }
    ring_unlink(module_ring(M));
    shifts_unlink(module_shifts(M));
    vlist_free( & module_gens(M),R);
    vlist_free( & module_mingens(M),R);
    vlist_free( & module_gbasis(M),R);
    gb_free(M);
    pseries_free( module_pseries(M) );
    module_unlink( module_syz(M), R);
    module_unlink( M->syz_of_gens, R);
    free_module(M); }
}

module module_dup(module M, ring R)
{
  module res_M;
  int i;
  if (M==module_NULL) return module_NULL;
  res_M=module_init(module_rank(M),R);
  module_set_flags(res_M,module_flags(M));
  module_set_shifts(res_M,shifts_link(module_shifts(M)));
  module_set_gens(res_M,vlist_dup(module_gens(M),R));
  module_set_gbasis(res_M,vlist_dup(module_gbasis(M),R));
  module_set_pseries(res_M,pseries_dup(module_pseries(M)));
  module_set_mvpseries(res_M,MVPSeries_dup(module_mvpseries(M)));
  res_M->syz_of_gens = module_link(M->syz_of_gens); /*JAA*/
  module_set_syz(res_M,module_link(module_syz(M)));
  module_set_zys(res_M,module_zys(M));
  module_set_DegTrunc(res_M,module_DegTrunc(M));
  module_set_SyzCompsNum(res_M,module_SyzCompsNum(M));
  module_set_ResTrunc(res_M,module_ResTrunc(M));
  module_set_RegTrunc(res_M,module_RegTrunc(M));
  module_set_ElimIndets(res_M,module_ElimIndets(M));
  module_gb_set_initialized(res_M,module_gb_initialized(M));
  module_set_discrepancy(res_M,module_discrepancy(M));
  if (module_mingens(M) != vlist_NULL) {
    if (module_gens(M)!=module_mingens(M))
      module_set_mingens(res_M,vlist_dup(module_mingens(M),R));
    else
      module_set_mingens(res_M,module_gens(res_M)); }
  ring_unlink(module_ring(res_M));
  module_set_ring(res_M,ring_link(module_ring(M)));
  gb_dup(res_M,M);
  for (i=0;i<gb_NUM_STATS;i++) module_set_stat(res_M,i,module_stat(M,i));
  return res_M;
}

module module_generated_by_this_pv(pvector P,ring R)
{
  module res_J;
  res_J=module_init(pv_rank(P),R);
  module_set_gens(res_J,vlist_init());
  vlist_append_this_vector(module_gens(res_J), P, R );
  return res_J;
}

module module_generated_by_pv(pvector P,ring R)
{
  module res_J;
  res_J=module_init(pv_rank(P),R);
  module_set_gens(res_J,vlist_init());
  vlist_append_vector(module_gens(res_J), P, R );
  return res_J;
}

module this_vlist2module( vlist vl, int r, ring R)
{
  module res_M;
  shifts aux_ss;
  res_M=module_init(r,R);
  module_set_gens(res_M,vl);
  vlist_rank_and_shifts(vl,&r,&aux_ss);
  if (aux_ss != shifts_NULL) {
    module_set_shifts(res_M,shifts_link(aux_ss));
  }
  return res_M;
}

module vlist2module( vlist vl, int r, ring R)
{
  module res_M;
  res_M=module_init(r,R);
  module_set_gens(res_M,vlist_dup(vl,R));
  return res_M;
}

module module_sum(module M1, module M2, ring R)
{
  module res_M;
  vlist aux_vl1,aux_vl2;
  if (module_not_compatible(M1,M2)) return module_NULL;
  res_M=module_init(module_rank(M1),R);
  aux_vl1=vlist_dup(module_gens(M1),R);
  aux_vl2=vlist_dup(module_gens(M2),R);
  vlist_reinit_vectors(aux_vl1,R);
  vlist_reinit_vectors(aux_vl2,R);
  module_set_shifts(res_M,shifts_link(module_shifts(M1)));
  module_set_gens(res_M,vlist_union_these(aux_vl1,aux_vl2,R));
  return res_M;
}

void module_mult_this_by_pvcomp(module * M, pvcomp P, ring R)
{
  vlist_mult_this_by_pvcomp( & module_gens(*M),P,R);
  vlist_free( & module_gbasis(*M),R );
  pseries_free( module_pseries(*M) );
  module_unlink( module_syz(*M), R);
}

module module_mult_by_arpoly(module M, arpoly P, ring R)
{
  pvcomp aux_p;
  module res_M;
  aux_p=arpoly2pvcomp(P,R);
  res_M=module_mult_by_pvcomp(M,aux_p,R);
  pvcomp_free(aux_p,R);
  return res_M;
}

module module_mult_by_pvcomp(module M, pvcomp P, ring R)
{
  vlist aux_vl;
  module res_M;
  aux_vl=vlist_mult_by_pvcomp(module_gens(M),P,R);
  res_M=this_vlist2module(aux_vl,module_rank(M),R);
  module_set_shifts(res_M,shifts_link(module_shifts(M)));
  return res_M;
}

module module_mult_by_ideal(module M, module J, ring R)
{
  int n;
  module res_M;
  vlist aux_pl;
  evector aux_ev;
  vlist aux_vl,res_vl;
  vlist_atom aux_el;
  pvector aux_v;
  aux_pl=module_gens(J);
  n=vlist_len(aux_pl);
  aux_el=vlist_first(aux_pl);
  res_vl=vlist_init();
  while (n-- > 0) {
    aux_ev=vlist_atom_ev(aux_el);
    if (ev_is_split(aux_ev)) {
      aux_v=ev_whole_vector(aux_ev,R);
      aux_vl=vlist_mult_by_pvcomp( module_gens(M),pv_nth(aux_v,0),R);
      pv_free(aux_v,R);
    } else {
      aux_vl=vlist_mult_by_pvcomp( module_gens(M),vlist_atom_poly(aux_el),R);
    }
    res_vl=vlist_union_these(res_vl,aux_vl,R);
    aux_el=vlist_atom_next(aux_el); 
  }
  res_M=this_vlist2module(res_vl,module_rank(M),R);
  module_set_shifts(res_M,shifts_link(module_shifts(M)));
  return res_M;
}

module module_zero(int r, ring R)
{
  vlist aux_vl;
  aux_vl=vlist_init();
  vlist_append_this_vector(aux_vl,pv_zero(r,R),R);
  return this_vlist2module(aux_vl,r,R);
}

module module_freemod(int r, ring R)
{
  vlist aux_vl;
  int i;
  module M;
  aux_vl=vlist_init();
  for (i=0;i<r;i++)
    vlist_append_this_vector(aux_vl,pv_gen(r,i,R),R);
  M = this_vlist2module(aux_vl,r,R);
  module_set_flags(M, module_flags(M) | module_flag_is_freemod);
  return M;
}


coc_bool module_is_freemod(module M)
{
  return (module_flags(M) & module_flag_is_freemod) != 0;
}


cterm ideal2cterm(module M, ring R)
{
  int n;
  pvector aux_v;
  vlist_atom aux_atom;
  cterm aux_t;
  ctbuf aux_b;
  vlist aux_vl=module_gens(M);
  n=vlist_len(aux_vl);
  if (n==0) return cterm_mk_const(op_IDEAL);
  aux_b=ctbuf_empty();
  while (n-- > 0) {
    aux_atom=vlist_get_first(aux_vl);
    if (ev_is_split(vlist_atom_ev(aux_atom))) {
      aux_v=ev_whole_vector(vlist_atom_ev(aux_atom),R);
      aux_t=pvcomp2cterm(pv_nth(aux_v,0),R);
      pv_free(aux_v,R);
    } else {
      aux_t=pvcomp2cterm(vlist_atom_poly(aux_atom),R);
    }
    ctbuf_append(aux_b,aux_t);
    vlist_append(aux_vl,aux_atom); }
  return ctbuf_mk_cterm(aux_b,op_IDEAL);
}

cterm module2cterm(module M, ring R)
{
  int n;
  cterm params[1];
  vlist_atom aux_atom;
  cterm aux_t;
  ctbuf aux_b;
  vlist aux_vl;
  aux_vl=vlist_dup(module_gens_ck(M), R);
  n=vlist_len(aux_vl);
  if (n==0) {
    pvector aux_v;
    vlist_free(&aux_vl, R);
    aux_v=pv_init(module_rank(M));
    params[0]=cterm_noindent(pv2cterm2(aux_v,R));
    pv_free(aux_v,R);
    return cterm_mk_classic(op_MODULE,1,params); }
  aux_b=ctbuf_empty();
  if (module_shifts(M)!=shifts_NULL) {
    ctbuf_append(aux_b,shifts2cterm(module_shifts(M),R));
  }
  vlist_enlarge_elements(aux_vl, M->r, M->ss);
  while (n-- > 0) {
    aux_atom=vlist_get_first(aux_vl);
    aux_t=cterm_noindent(ev2cterm_normal(vlist_atom_ev(aux_atom),R));
    ctbuf_append(aux_b,aux_t);
    vlist_append(aux_vl,aux_atom); }
  vlist_free(&aux_vl, R);
  return ctbuf_mk_cterm(aux_b,op_MODULE);
}

void printf_module(module M, ring R)
{
  cterm t;
  t=module2cterm(M,R);
  put_cterm(t);
  cterm_free(t);
}

ctbuf module_description(module M, ring R)
{
  ctbuf res_b;
  cterm aux_t;
  res_b=ctbuf_empty();
  ctbuf_append(res_b,
    cterm_mk_table_entry("Gens",vlist2cterm(module_gens_ck(M),R)));

  ctbuf_append(res_b,
    cterm_mk_table_entry("MRC",int2cterm(module_rc(M))));

  if (module_shifts(M)!=shifts_NULL) {
    aux_t=shifts2cterm(module_shifts(M),R);
    ctbuf_append(res_b,
      cterm_mk_table_entry("Shifts",aux_t)); }

  if (module_gbasis(M)!=vlist_NULL) {
    aux_t=vlist2cterm(module_gbasis(M),R);
    ctbuf_append(res_b,
      cterm_mk_table_entry("GBasis",aux_t)); }

  if (module_mingens(M)!=vlist_NULL && vlist_len(module_mingens(M)) != 0) {
    if (module_mingens(M)==module_gens_ck(M)) {
      aux_t=string2cterm("IdenticalToGens");
    } else {
      aux_t=vlist2cterm(module_mingens(M),R);
    }
    ctbuf_append(res_b,
      cterm_mk_table_entry("MinGens",aux_t)); }

  if (M->syz_of_gens != module_NULL)
  {
    aux_t = module2cterm(M->syz_of_gens,R);
    ctbuf_append(res_b, cterm_mk_table_entry("SyzOfGens", aux_t));
  }

  if (module_syz(M)!=module_NULL) {
    aux_t=module2cterm(module_syz(M),R);
    ctbuf_append(res_b,
      cterm_mk_table_entry("Syz",aux_t)); }

  if (module_pseries(M)!=pseries_NULL) {
    aux_t=olist2cterm(pseries2olist(module_pseries(M)));
    ctbuf_append(res_b,
      cterm_mk_table_entry("PSeries",aux_t)); }

  if (flag_is_on(opt_panel_GROBNER,flag_GB_VERBOSE)) {

    if (gb_IVs(M)!=vlist_NULL && vlist_len(gb_IVs(M)) != 0) {
      aux_t=vlist2cterm(gb_IVs(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("IVs",aux_t)); }
  
    if (gb_CV(M)!=ev_NULL) {
      aux_t=ev2cterm_normal(gb_CV(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("CV",aux_t)); }
  
    if (gb_Rules(M)!=vlist_NULL) {
      aux_t=vlist2cterm(gb_Rules(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("Rules",aux_t)); }
  
    if (gb_IPs(M)!=vpairs_NULL) {
      aux_t=vpairs2cterm(gb_IPs(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("IPs",aux_t)); }

    if (module_discrepancy(M)!=0) {
      aux_t=int2cterm(module_discrepancy(M));
      ctbuf_append(res_b,
        cterm_mk_table_entry("Discrepancy",aux_t)); }

    ctbuf_append(res_b, cterm_mk_table_entry("KFL",gb_kf2cterm(M)));
  }

  if (module_DegTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_DegTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("DegTrunc",aux_t)); }

  if (module_SyzCompsNum(M)!=trunc_NONE) {
    aux_t=int2cterm(module_SyzCompsNum(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("SyzCompsNum",aux_t)); }

  if (module_ResTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_ResTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("ResTrunc",aux_t)); }

  if (module_RegTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_RegTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("RegTrunc",aux_t)); }

  return res_b;
}


ctbuf ideal_description(module M, ring R)
{
  ctbuf res_b;
  cterm aux_t;
  res_b=ctbuf_empty();
  ctbuf_append(res_b,
    cterm_mk_table_entry("Gens", this_olist2cterm(this_vlist2olist_ter(
      vlist_dup(module_gens(M),R),R))));

#ifdef DEBUG
  ctbuf_append(res_b,
    cterm_mk_table_entry("MRC",int2cterm(module_rc(M))));
#endif

  if (module_shifts(M)!=shifts_NULL) {
    aux_t=shifts2cterm(module_shifts(M),R);
    ctbuf_append(res_b,
      cterm_mk_table_entry("Shifts",aux_t)); }

  if (module_gbasis(M)!=vlist_NULL) {
    aux_t=this_olist2cterm(this_vlist2olist_ter(
      vlist_dup(module_gbasis(M),R),R));
    ctbuf_append(res_b,
      cterm_mk_table_entry("GBasis",aux_t)); }

  if (module_mingens(M)!=vlist_NULL && vlist_len(module_mingens(M)) != 0) {
    if (module_mingens(M)==module_gens_ck(M)) {
      aux_t=string2cterm("IdenticalToGens");
    } else {
      aux_t=this_olist2cterm(this_vlist2olist_ter(
        vlist_dup(module_mingens(M),R),R));
    }
    ctbuf_append(res_b,
      cterm_mk_table_entry("MinGens",aux_t)); }

  if (M->syz_of_gens != module_NULL)
  {
    aux_t = module2cterm(M->syz_of_gens,R);
    ctbuf_append(res_b, cterm_mk_table_entry("SyzOfGens", aux_t));
  }

  if (module_syz(M)!=module_NULL) {
    aux_t=module2cterm(module_syz(M),R);
    ctbuf_append(res_b,
      cterm_mk_table_entry("Syz",aux_t)); }

  if (module_pseries(M)!=pseries_NULL) {
    aux_t=olist2cterm(pseries2olist(module_pseries(M)));
    ctbuf_append(res_b,
      cterm_mk_table_entry("PSeries",aux_t)); }

  if (flag_is_on(opt_panel_GROBNER,flag_GB_VERBOSE)) {

    if (gb_IVs(M)!=vlist_NULL) {
      aux_t=this_olist2cterm(this_vlist2olist_ter(
        vlist_dup(gb_IVs(M),R),R));
      ctbuf_append(res_b,
        cterm_mk_table_entry("IVs",aux_t)); }
  
    if (gb_CV(M)!=ev_NULL) {
      aux_t=ev2cterm_normal(gb_CV(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("CV",aux_t)); }
  
    if (gb_Rules(M)!=vlist_NULL) {
      aux_t=this_olist2cterm(this_vlist2olist_ter(
        vlist_dup(gb_Rules(M),R),R));
      ctbuf_append(res_b,
        cterm_mk_table_entry("Rules",aux_t)); }
  
    if (gb_IPs(M)!=vpairs_NULL) {
      aux_t=vpairs2cterm(gb_IPs(M),R);
      ctbuf_append(res_b,
        cterm_mk_table_entry("IPs",aux_t)); }

    if (module_discrepancy(M)!=0) {
      aux_t=int2cterm(module_discrepancy(M));
      ctbuf_append(res_b,
        cterm_mk_table_entry("Discrepancy",aux_t)); }
  
    ctbuf_append(res_b, cterm_mk_table_entry("KFL",gb_kf2cterm(M)));
  }

  if (module_DegTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_DegTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("DegTrunc",aux_t)); }

  if (module_ResTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_ResTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("ResTrunc",aux_t)); }

  if (module_RegTrunc(M)!=trunc_NONE) {
    aux_t=int2cterm(module_RegTrunc(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("RegTrunc",aux_t)); }

  if (module_SyzCompsNum(M)!=trunc_NONE) {
    aux_t=int2cterm(module_SyzCompsNum(M));
    ctbuf_append(res_b,
      cterm_mk_table_entry("SyzCompsNum",aux_t)); }

  return res_b;
}
