/* $Id: modupoin.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include <gmp.h>
#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "eterms.h"
#include "evector.h"
#include "ivectors.h"
#include "module.h"
#include "monomial.h"
#include "poincare.h"
#include "modupoin.h"
#include "options.h"
#include "pvector.h"
#include "pseries.h"
#include "polys.h"
#include "ring.h"
#include "utils.h"
#include "terms.h"
#include "vlist.h"
#include "weights.h"

/*  auxiliary  */

TermList TrueDegVListNthCompToTList (vlist vl, term s, int nth, ring R)
{
  eterm t;
  term tt;
  iv_elem index;
  TermList auxTL;
  int NewLen =0, n;
  vlist_atom auxVL_el;
 
  n = vlist_len(vl);
  auxVL_el=vlist_first(vl);
  auxTL = NewTList (n, R->indetsNo);
 
  while ( n-- > 0 )
  {
    if (!vlist_atom_is_marked(auxVL_el))
    {
#ifdef PVCOMP_MONOPOLI
      tt = vlist_elem_nth_comp_lt(auxVL_el,nth);
#else
      tt = vlist_elem_nth_comp_lt_dup(auxVL_el,nth,R);
#endif
      if (tt != term_NULL)
      {
        if (s != term_NULL) tt = term_div(tt,s);
        t = TrueDeg_term2eterm(tt, R);
        if ( eterm_get_OccIndNo(t) == 1 ) /* TODO: check */
        {
          index = (Indets(t))[1];
          InsInSPList(index, t[index], SPList(auxTL));
          eterm_free(t);
        }
        else
        {
          MTLPutLast (MTList(auxTL), NewLen, t);
          t = NULL;
        }
#ifndef PVCOMP_MONOPOLI
	term_free(tt);
#endif
      }
    }
    auxVL_el = vlist_atom_next(auxVL_el) ;
  }
  TListReduceSize (auxTL, NewLen);
  return auxTL;
}

TermList VListNthCompToTList (vlist vl, term s, int nth, ring R)
{
  eterm t;
  term tt;
  iv_elem index;
  TermList auxTL;
  int NewLen =0, n;
  vlist_atom auxVL_el;

  n = vlist_len(vl);
  auxVL_el = vlist_first(vl);
  auxTL = NewTList (n, R->indetsNo);
  
  while ( n-- > 0 )
  {
    if (!vlist_atom_is_marked(auxVL_el)) 
    {
#ifdef PVCOMP_MONOPOLI
      tt = vlist_elem_nth_comp_lt(auxVL_el,nth);
#else
      tt = vlist_elem_nth_comp_lt_dup(auxVL_el,nth,R);
#endif
      if (tt != term_NULL) 
      {
        if (s != term_NULL)   term_div_and_assign(tt,s);
        t = term2eterm(tt);
        if ( eterm_get_OccIndNo(t) == 1 ) /* TODO: check */
        {
          index = (Indets(t))[1];
          InsInSPList (index, t[index], SPList(auxTL) );
          eterm_free (t);
        }
        else
        {
          MTLPutLast (MTList(auxTL), NewLen, t);
          t = NULL;
        }
#ifndef PVCOMP_MONOPOLI
	term_free(tt);
#endif
      }
    }
    auxVL_el = vlist_atom_next(auxVL_el) ;
  }
  TListReduceSize (auxTL, NewLen);
  return auxTL;
}

/* ---------------------------------------------- */

unipoly PoincareNumeratorOfVList(vlist theVList, shifts ss, int r, ring R)
{
  unipoly p;
  vlist aux_vl;
  int i;
  TermList theTList;

  StartPoincare(R->indetsNo);

  p = NewUnipoly(0);
  if (ss != shifts_NULL)  r = shifts_rank(ss);
  aux_vl = vlist_only_leading_terms(theVList, R);
  for ( i=1 ; i<=r ; i++ ) 
  {
    theTList = VListNthCompToTList (aux_vl, shifts_nth(ss,i), i-1, R);
    if ( MTListLen(theTList) != 0 || eterm_get_OccIndNo(SPList(theTList))!=0 )
      p = P1PlusXExpP2(p, shifts_deg_nth(ss,i), BigRecPoincare(theTList)); 
    else 
    {
      if ( UnipolyDegree(p)<shifts_deg_nth(ss,i) )
      {
        UPSetDeg(p, shifts_deg_nth(ss,i));
	if ( UPSize(p) < shifts_deg_nth(ss,i) )
	  p = UnipolyChangeSize(p, shifts_deg_nth(ss,i));
      }
      Add1ToCoeffOfXd(p, shifts_deg_nth(ss,i));
      EraseAndFreeTList(theTList);
    }
  }
  vlist_free(&aux_vl,R);
  
  return p;
}

arpoly MVPoincareNumeratorOfVList(vlist theVList, shifts ss, 
				  weilist wl, int r, ring R, ring P_R)
{
  arpoly res_p, aux_p;
  vlist aux_vl;
  int i;
  term mdeg;  
  TermList theTList, TLWeights;

  StartPoincare(R->indetsNo);

  res_p = NULL;
  if ( ss != shifts_NULL )  r = shifts_rank(ss);
  aux_vl = vlist_only_leading_terms(theVList, R);
  TLWeights = WeilistToTList(wl, R);
  for ( i=1 ; i<=r ; i++ )
  {
    theTList = TrueDegVListNthCompToTList (aux_vl, shifts_nth(ss,i), i-1, R);
    if ( MTListLen(theTList)!=0 || eterm_get_OccIndNo(SPList(theTList))!=0 )
    {
      aux_p = MVRecPoincare(theTList, TLWeights, P_R);
      if ( shifts_nth(ss,i)!=term_NULL )
      {
	mdeg = term_mdegree(shifts_nth(ss,i), R);
	aux_p = arpoly_this_x_term(aux_p, mdeg, P_R);
	term_free(mdeg);
      }
      res_p = arpoly_sum_these(res_p, aux_p, P_R);
    }
    else 
    {
      if ( shifts_nth(ss,i)!=term_NULL )
	mdeg = term_mdegree(shifts_nth(ss,i), R);      
      else
	mdeg = term_one(P_R);      
      res_p = arpoly_add_this_monomial(res_p, 
				       this_term2monomial(mdeg, P_R), 
				       P_R);
      EraseAndFreeTList(theTList);
    }
  }
  EraseAndFreeTList(TLWeights);
  vlist_free(&aux_vl,R);

  return res_p;
}

PSeries PoincareSeriesOfVList(vlist theVList, shifts ss, int r, ring R)
{
  PSeries PS;
  long t;
  
  PS = pseries_init();
#ifdef POINC_TIME
  t = clock();
#endif /* POINC_TIME */
  pseries_set_num(PS, PoincareNumeratorOfVList(theVList, ss, r, R));
#ifdef POINC_TIME
  printf("           PNum(thePList):  %4.2f s\n",(double)(clock()-t)/1000000);
#endif /* POINC_TIME */
  pseries_set_den_factors(PS, PoincareDenFactors(R)); 
  /* NOTE: it could be computed just once */

  return PS;
}

MVPSeries MVPoincareSeriesOfVList(vlist theVList, shifts ss, int r, 
				  weilist wl, ring R, ring P_R)
{
  MVPSeries PS;
 
  PS = MVPSeries_init(P_R);
#ifdef POINC_TIME
  t = clock();
#endif /* POINC_TIME */
  MVPSeries_set_num(PS, MVPoincareNumeratorOfVList(theVList,ss,wl,r,R,P_R));
#ifdef POINC_TIME
  printf("           PNum(thePList):  %4.2f s\n",(double)(clock()-t)/1000000);
#endif /* POINC_TIME */
  MVPSeries_set_den_factors(PS, MVPoincareDenFactors(wl,R));
  /* NOTE: it could be computed just once */
 
  return PS;
}

/* ---------------------------------------------- */
/* andranno da qualche altra parte da decidere */

PSeries module_PS_of_quotient(module M) 
{
  return PoincareSeriesOfVList(module_gbasis(M), module_shifts(M),
			       module_rank(M), module_ring(M));
}

void module_compute_PS_of_quotient_trunc(module M, int D)
{
  vlist aux_vl;
  ring R;

  R = module_ring(M);

  if (module_pseries(M)!=pseries_NULL) pseries_free(module_pseries(M));
  aux_vl = vlist_dup_trunc(module_gbasis(M), D, R);
  module_set_pseries(M,
      PoincareSeriesOfVList(aux_vl, module_shifts(M), module_rank(M), R));
  vlist_free(&aux_vl,R);
}




