/* $Id: monomial.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "coeffs.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "ring.h"
#include "terms.h"
#include "weights.h"

monomial monomial_dummy;

void monomial_adt_init(void)
{
  monomial_dummy=malloc_monomial();
}

monomial_list monomial_list_x_monomial(monomial_list p, monomial m, ring R)
{
  monomial_list aux_p,aux_p1;
  coeff aux_c,res_c;
  term aux_t,aux_t1,aux_t2;

  aux_c=monomial_coeff(m);
  monomial_set_next(monomial_dummy,NULL);
  aux_t2=monomial_term(m);
  aux_p=monomial_dummy;
  while (p != NULL) {
    coeff_set_mult(res_c,aux_c,monomial_coeff(p),R);

#ifndef IS_DOMAIN

    if (coeff_is_zero(res_c,R))
      R->ClearCoeff(&res_c);
    else {
      aux_t1=monomial_term(p);
      term_set_mult(aux_t,aux_t1,aux_t2);
      monomial_set_next(aux_p, monomial_mk(res_c, aux_t);
      aux_p=monomial_next(aux_p); }

#else

    aux_t1=monomial_term(p);
    term_set_mult(aux_t,aux_t1,aux_t2);
    monomial_set_mk(aux_p1,res_c,aux_t);
    monomial_set_next(aux_p, aux_p1);
    aux_p=aux_p1;

#endif

    p=monomial_next(p); }
  return monomial_next(monomial_dummy);
}

void monomial_mult_and_assign(monomial res_m, monomial m, ring R)
{
  coeff res_c;
  res_c=monomial_coeff(res_m);
  coeff_mult_this(res_c,monomial_coeff(m),R);
  monomial_set_coeff(res_m, res_c);
  term_mult_and_assign(monomial_term(res_m),monomial_term(m));
  monomial_set_next(res_m, NULL);
}

monomial monomial_list_dup(monomial p,ring R)
{
  monomial res_p, aux_p;
  if (p == NULL) return p;
  res_p=monomial_dup(p,R);
  aux_p=res_p;
  p=monomial_next(p);
  while (p != NULL) {
    monomial_set_next(aux_p,monomial_dup(p,R));
    p=monomial_next(p);
    aux_p=monomial_next(aux_p); }
  return res_p;
}

monomial monomial_opposite(monomial M, ring R)
{
  monomial res_m;
  res_m=monomial_dup(M,R);
  monomial_change_sign(res_m,R);
  return res_m;
}

monomial monomial_map(monomial M, ring R1, ring R2, ivector m)
{
  /* see description of poly_map */
  monomial res_m;
  res_m=malloc_monomial();
  monomial_set_coeff(res_m,coeff_dup(monomial_coeff(M),R2));
  monomial_set_term(res_m,
    term_map(monomial_term(M), ring_weights(R1), ring_weights(R2), m));
  monomial_set_next(res_m,NULL);
  return res_m;
}

monomial monomial_mk(coeff c, term t)
{
  monomial res_m;
  res_m=malloc_monomial();
  monomial_set_coeff(res_m,c);
  monomial_set_term(res_m,t);
  monomial_set_next(res_m,NULL);
  return res_m;
}

void monomial_free(monomial m,ring R)
{
  R->ClearCoeff(& monomial_coeff(m));
  term_free(monomial_term(m));
  free_monomial(m);
}

cterm monomial2cterm(monomial m,ring R) 
{
  cterm res_t;
  cterm params[2];
  if (m==NULL) res_t=cterm_NULL;
  else {
    params[0]=coeff2cterm(monomial_coeff(m),R);
    if (term_degree(monomial_term(m))==0)
      res_t=cterm_mk_classic(op_MONOMIAL,1,params);
    else {
      params[1]=term2cterm(monomial_term(m),R);
      res_t=cterm_mk_classic(op_MONOMIAL,2,params); }
  }
  return res_t;
}

monomial cterm2monomial(cterm t, ring R)
{
  /* t=mon(c:coeff,t:term) */
  /* t=mon(c:coeff) */
  cterm_list params;
  coeff aux_c;
  term aux_t;
  params=cterm_params(t);
  aux_c=cterm2coeff(params[0],R);
  if (cterm_arity(t)==2)
    aux_t=cterm2term(params[1],R);
  else
    aux_t=term_init(R->indetsNo);
  return monomial_mk(aux_c,aux_t);
}

monomial monomial_der(monomial m, unknown x, ring R)
{
  term aux_t;
  coeff aux_c;
  aux_t=term_der(monomial_term(m),x);
  if (aux_t==term_NULL) return NULL;
  aux_c=coeff_x_int(monomial_coeff(m), term_get_nth(aux_t,x+1)+1,R);
  if (coeff_is_zero(aux_c,R)) {
    term_free(aux_t);
    coeff_free(aux_c,R);
    return NULL; }
  return monomial_mk(aux_c,aux_t);
}

void printf_monomial(monomial m, ring R)
{
  cterm aux_t;
  aux_t=monomial2cterm(m,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

monomial monomial_square(monomial m, ring R)
{
  coeff aux_c;
  aux_c=coeff_mult(monomial_coeff(m),monomial_coeff(m),R);
  if (coeff_is_zero(aux_c,R)) { coeff_free(aux_c,R); return NULL; }
  return monomial_mk(aux_c,term_mult(monomial_term(m),monomial_term(m)));
}

monomial monomial_revert_this(monomial m)
{
  monomial curr,prev,next;

  prev=NULL;
  curr=m;

  while (curr!=NULL) {
    next=monomial_next(curr);
    monomial_set_next(curr,prev);
    prev=curr;
    curr=next; }

  return prev;
}
