/* $Id: monopoli.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */
 
#include "cocdefs.h"

#ifdef MONOPOLI

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "coeffs.h"
#include "ctbuf.h"
#include "cterm.h"
#include "ca.h"
#include "ck.h"
#include "ivectors.h"
#include "monomial.h"
#include "monopoli.h"
#include "olist.h"
#include "operator.h"
#include "object.h"
#include "options.h"
#include "polys.h"
#include "ring.h"
#include "terms.h"
#include "weights.h"

monopoli monopoli_remove_lm(monopoli p, ring R)
{
  monomial aux_m;
#ifdef DEBUG
  if (p == NULL) {
    INTERNAL_ERROR("monopoli_remove_lm(p,R): p is NULL");
    return p;
  }
#endif
  aux_m=p;
  p=monomial_next(p);
  monomial_free(aux_m,R);
  return p;
}


coc_bool monopoli_is_homog_deg(monomial m, int hd, ring R)
{
  if (m==NULL) return TRUE;
  if (monomial_deg(m) != hd) return FALSE;
  m=monomial_next(m);
  while (m != NULL) {
    if (hd != monomial_deg(m)) return FALSE;
    m=monomial_next(m); }
  return TRUE;
}

void monopoli_adt_init(void)
{
  /* NOTHING TO DO */
}

monopoli monopoli_sum(monopoli p1, monopoli p2, ring R)
{
  monopoli aux_p,res_p;
  int answer;

  res_p=monomial_dummy;
  aux_p=res_p;

  while ((p1 != NULL) && (p2 != NULL)) {
    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {
      coeff aux_c;
      case ANSWER_BOTH:
        aux_c=coeff_sum(monomial_coeff(p1),monomial_coeff(p2),R);
        if (coeff_is_zero(aux_c,R))
          R->ClearCoeff(&aux_c);
	else {
        monomial_set_next(aux_p,monomial_mk(aux_c,term_dup(monomial_term(p1))));
          aux_p=monomial_next(aux_p); }
        p1=monomial_next(p1);
        p2=monomial_next(p2);
	break;
      case ANSWER_FIRST:
        monomial_set_next(aux_p,monomial_dup(p1,R));
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
        break;
      case ANSWER_SECOND:
        monomial_set_next(aux_p,monomial_dup(p2,R));
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
        break;
    }
  }
  if (p1 == NULL) monomial_set_next(aux_p,monopoli_dup(p2,R));
  else monomial_set_next(aux_p,monopoli_dup(p1,R));
  return monomial_next(res_p);
}

monopoli unknown_pow2monopoli(unknown u,expType exponent,ring R)
{
  term aux_t;
  aux_t=term_init(R->indetsNo);
  term_put_nth(aux_t,u+1,exponent*iv_get_nth(R->weights,u+1));
  return monomial_mk(coeff_one(R),aux_t);
}

monopoli monopoli_map(monopoli p, ring R1, ring R2, ivector m)
{
  /* let p \in R1 */
  /* indets(R1) = x[1..k], indets(R2) = y[1..n] */
  /* m is the map, len(m) = n */
  /* if m[i] <> 0 then ey[i] := ex[m[i]] */ /* x[m[i]] -> y[i] */
  /* if m[i] = 0 then ey[i] := 0 */
  /* so if j \in 1..k but j \not\in m it means x[j] -> 1 */
  /* ASSUMPTIONS: CoeffRing(R1) = CoeffRing(R2) */
  /* m[i] \in 0..k \forall i \in 1..n */

  monopoli res_p;

  res_p = NULL;
  while (p != NULL) {
    res_p=monopoli_sum_these(res_p,monomial_map(p,R1,R2,m),R2);
    p = monomial_next(p);
  }
  return res_p;
}

monopoli monopoli_randomize(monopoli p, ring R)
{
  monopoli res_p,aux_p;
  coeff aux_c;
  monomial aux_m;
  if (p == NULL) return NULL;
  aux_p=monomial_dummy;
  monomial_set_next(aux_p,NULL);

  res_p=aux_p;
  while (p != NULL) {
    aux_c=coeff_random(R);
    if (coeff_is_zero(aux_c,R))
      R->ClearCoeff(&aux_c);
    else {
      monomial_set_mk(aux_m,aux_c,term_dup(monomial_term(p)));
      monomial_set_next(res_p,aux_m);
      res_p=monomial_next(res_p); }
    p = monomial_next(p); }
  return monomial_next(aux_p);
}

monopoli monopoli_mult( monopoli p1, monopoli p2, ring R ){
  monopoli auxPoly = NULL, res_p = NULL;
  while ( (p1 != NULL) ) {
    auxPoly=monopoli_dup( p2, R );
    auxPoly=monopoli_this_x_monomial( auxPoly, p1, R );    
    res_p=monopoli_sum_these( auxPoly, res_p, R );
    p1 = monomial_next(p1);
  }
  return res_p;
}

monopoli this_coeff2monopoli(coeff c, ring R)
{
  if (coeff_is_zero(c,R)) {
    R->ClearCoeff(&c);
    return NULL; }
  return monomial_mk(c,term_init(R->indetsNo));
}

monopoli monopoli_der(monopoli p, unknown x, ring R)
{
  monopoli res_p;
  monomial aux_m;
  if (p == NULL) return p;
  res_p=monomial_der(p,x,R);
  p=monomial_next(p);
  while (p != NULL) {
    aux_m=monomial_der(p,x,R);
    res_p=monopoli_sum_these(res_p,aux_m,R);
    p=monomial_next(p); }
  return res_p;
}

monopoli monopoli_sum_these(monopoli p1, monopoli p2, ring R)
{
  monopoli aux_p,res_p;
  int answer;

  res_p=monomial_dummy;
  aux_p=res_p;

  while ((p1 != NULL) && (p2 != NULL)) {
    monopoli p1_next,p2_next;
    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {
      case ANSWER_BOTH:
        p1_next=monomial_next(p1);
        p2_next=monomial_next(p2);
        R->AddCoeff(&monomial_coeff(p1),monomial_coeff(p1),
                                        monomial_coeff(p2),R);
        if (coeff_is_zero(monomial_coeff(p1),R))
          monomial_free(p1,R);
	else {
          monomial_set_next(aux_p,p1);
          aux_p=monomial_next(aux_p); }
	monomial_free(p2,R);
        p1=p1_next;
        p2=p2_next;
	break;
      case ANSWER_FIRST:
        monomial_set_next(aux_p,p1);
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
        break;
      case ANSWER_SECOND:
        monomial_set_next(aux_p,p2);
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
        break;
    }
  }
  if (p1 == NULL) monomial_set_next(aux_p,p2);
  else monomial_set_next(aux_p,p1);
  return monomial_next(res_p);
}

int monopoli_max_deg(monopoli m)
{
  int d=0;
  while (m != NULL) {
    d=MAX(d,monomial_deg(m));
    m=monomial_next(m); }
  return d;
}

/* ASSUMES that the polynomial m is non-zero */
coc_bool monopoli_ispurelt(monopoli m)
{
  return term_ispure(monomial_term(m));
}

coc_bool monopoli_eq(monopoli p1, monopoli p2, ring R)
{
  while ((p1 != NULL) && (p2 != NULL)) {
    if (!monomial_eq(p1,p2,R)) return FALSE;
    p1=monomial_next(p1);
    p2=monomial_next(p2); }
  return ((p1 == NULL) && (p2 == NULL));
}

monopoli monopoli_this_x_coeff(monopoli p, coeff c, ring R)
{
  monopoli res_p=p;
  while (p != NULL) {
    R->MulCoeff(&monomial_coeff(p),monomial_coeff(p),c,R);
    p=monomial_next(p); }
  return res_p;
}

monopoli monopoli_this_minus_term_x_monopoli_truncated (monopoli p1, term ME, 
  monopoli p2, iv_elem Trunc, ring R)
/* Trunc( p1-t*p2, Trunc ) */
{
  monopoli aux_p,res_p,p1_next,p2_next;
  int answer;
  iv_elem MEdeg=term_get_first (ME);
  
  res_p=monomial_dummy;
  aux_p=res_p;

  while ( (p1 != NULL) && (term_get_first(monomial_term(p1))>Trunc)) {
       p1_next=monomial_next(p1);
       monomial_free(p1,R);
       p1=p1_next; }

  while ((p2 != NULL) && 	 
	  ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
       p2_next=monomial_next(p2);
       monomial_free(p2,R);
       p2=p2_next; }

  if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

  while ((p1 != NULL) && (p2 != NULL)) {
    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {

      case ANSWER_BOTH:
        p1_next=monomial_next(p1);
        p2_next=monomial_next(p2);
        R->SubCoeff(&monomial_coeff(p1),monomial_coeff(p1),
                                        monomial_coeff(p2),R);
        if (coeff_is_zero(monomial_coeff(p1),R))
          monomial_free(p1,R);
	else {
          monomial_set_next(aux_p,p1);
          aux_p=monomial_next(aux_p); }
	monomial_free(p2,R);
        p1=p1_next;
        p2=p2_next;
       while ((p1 != NULL) && 
             (term_get_first(monomial_term(p1))>Trunc)) {
            p1_next=monomial_next(p1);
            monomial_free(p1,R);
            p1=p1_next; }

       while ((p2 != NULL) && 	 
             ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
            p2_next=monomial_next(p2);
            monomial_free(p2,R);
            p2=p2_next; }

        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

	break;

      case ANSWER_FIRST:
        monomial_set_next(aux_p,p1);
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
       while ((p1 != NULL) && 
             (term_get_first(monomial_term(p1))>Trunc)) {
            p1_next=monomial_next(p1);
            monomial_free(p1,R);
            p1=p1_next; }
        break;

      case ANSWER_SECOND:
        monomial_change_sign(p2,R);
        monomial_set_next(aux_p,p2);
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
        while ((p2 != NULL) && 	 
              ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
             p2_next=monomial_next(p2);
             monomial_free(p2,R);
             p2=p2_next; }
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
        break;
    }
  }
  if (p1 == NULL) {
    monomial_set_next(aux_p,p2);
    while (p2 != NULL) {
        monomial_change_sign(p2,R);
        term_mult_and_assign(monomial_term(p2),ME);
        p2=monomial_next(p2); }
  }
  else monomial_set_next(aux_p,p1);
  return monomial_next(res_p);
}

monopoli monopoli_this_plus_term_x_monopoli (monopoli p1, term ME, monopoli p2, 
                               ring R)

{
  monopoli aux_p,res_p;
  int answer;

  res_p=monomial_dummy;
  aux_p=res_p;

  if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

  while ((p1 != NULL) && (p2 != NULL)) {
    monopoli p1_next,p2_next;
    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {
      case ANSWER_BOTH:
        p1_next=monomial_next(p1);
        p2_next=monomial_next(p2);
        R->AddCoeff(&monomial_coeff(p1),monomial_coeff(p1),
                                        monomial_coeff(p2),R);
        if (coeff_is_zero(monomial_coeff(p1),R))
          monomial_free(p1,R);
	else {
          monomial_set_next(aux_p,p1);
          aux_p=monomial_next(aux_p); }
	monomial_free(p2,R);
        p1=p1_next;
        p2=p2_next;
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
	break;
      case ANSWER_FIRST:
        monomial_set_next(aux_p,p1);
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
        break;
      case ANSWER_SECOND:
        monomial_set_next(aux_p,p2);
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
        break;
    }
  }
  if (p1 == NULL) {
    monomial_set_next(aux_p,p2);
    while (p2 != NULL) {
        term_mult_and_assign(monomial_term(p2),ME);
        p2=monomial_next(p2); }
  }
  else monomial_set_next(aux_p,p1);
  return monomial_next(res_p);
}

monopoli monopoli_this_plus_term_x_monopoli_truncated (monopoli p1, term ME, 
  monopoli p2, iv_elem Trunc, ring R)

/* Trunc( p1+t*p2, Trunc ) */

{
  monopoli aux_p,res_p,p1_next,p2_next;
  int answer;
  iv_elem MEdeg=term_get_first (ME);

  res_p=monomial_dummy;
  aux_p=res_p;

  while ((p1 != NULL) && (term_get_first(monomial_term(p1))>Trunc)) {
       p1_next=monomial_next(p1);
       monomial_free(p1,R);
       p1=p1_next; }

  while ((p2 != NULL) && 	 
        ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
       p2_next=monomial_next(p2);
       monomial_free(p2,R);
       p2=p2_next; }

  if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

  while ((p1 != NULL) && (p2 != NULL)) {

    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {

      case ANSWER_BOTH:
        p1_next=monomial_next(p1);
        p2_next=monomial_next(p2);
        R->AddCoeff(&monomial_coeff(p1),monomial_coeff(p1),
                                        monomial_coeff(p2),R);
        if (coeff_is_zero(monomial_coeff(p1),R))
          monomial_free(p1,R);
	else {
          monomial_set_next(aux_p,p1);
          aux_p=monomial_next(aux_p); }
	monomial_free(p2,R);
        p1=p1_next;
        p2=p2_next;

       while ((p1 != NULL) &&         
             (term_get_first(monomial_term(p1))>Trunc)) {
            p1_next=monomial_next(p1);
            monomial_free(p1,R);
            p1=p1_next; }

       while ((p2 != NULL) && 	 
             ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
            p2_next=monomial_next(p2);
            monomial_free(p2,R);
            p2=p2_next; }

        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

	break;

      case ANSWER_FIRST:
        monomial_set_next(aux_p,p1);
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
       while ((p1 != NULL) && 
             (term_get_first(monomial_term(p1))>Trunc)) {
            p1_next=monomial_next(p1);
            monomial_free(p1,R);
            p1=p1_next; }
        break;

      case ANSWER_SECOND:
        monomial_set_next(aux_p,p2);
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
       while ((p2 != NULL) && 	 
             ((term_get_first(monomial_term(p2))+MEdeg)>Trunc)) {
             p2_next=monomial_next(p2);
             monomial_free(p2,R);
             p2=p2_next; }
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
        break;
    }
  }
  if (p1 == NULL) {
    monomial_set_next(aux_p,p2);
    while (p2 != NULL) {
        term_mult_and_assign(monomial_term(p2),ME);
        p2=monomial_next(p2); }
  }
  else monomial_set_next(aux_p,p1);
  return monomial_next(res_p);
}

void monopoli_free(monomial m, ring R)
{
  monomial next_m;
  while (m != NULL) {
    next_m=monomial_next(m);
    monomial_free(m,R);
    m = next_m; }
}

ivector monopoli_max_mdeg(monomial m, ring R)
{
  ivector aux_v,res_v;
  res_v=iv_init(weilist_len(ring_weilist(R)));
  while (m != NULL) {
    aux_v=monomial_mdeg(m,R);
    res_v=max_mdeg_lex_these(res_v,aux_v);
    m=monomial_next(m); }
  return res_v;
}

monopoli monopoli_add_this_monomial(monopoli p, monomial m, ring R)
{
  monopoli res_p, prev_p, next_p;
  int answer;
  term aux_t;

  if (p == NULL) {
    monomial_set_next(m,NULL);
    return m;
  }

  aux_t=monomial_term(m);

  answer = R->term_which_gt(monomial_term(p),aux_t,ring_ordering(R),R);

  if (answer == ANSWER_SECOND) {
    monomial_set_next(m,p);
    return m;
  }

  if (answer == ANSWER_BOTH) {
    R->AddCoeff(&monomial_coeff(p),monomial_coeff(p), monomial_coeff(m),R);
    next_p=monomial_next(p);
    monomial_free(m,R);
    if (coeff_is_zero(monomial_coeff(p),R)) {
      monomial_free(p,R);
      return next_p; }
    return p;
  }

  res_p=p; prev_p=p; p=monomial_next(p);

  while (p != NULL) {
    answer=R->term_which_gt(monomial_term(p),aux_t,ring_ordering(R),R);
    switch (answer) {
      case ANSWER_BOTH:
        next_p=monomial_next(p);
        R->AddCoeff(&monomial_coeff(p),monomial_coeff(p),monomial_coeff(m),R);
        monomial_free(m,R);
        if (coeff_is_zero(monomial_coeff(p),R)) {
          monomial_free(p,R);
          monomial_set_next(prev_p, next_p); }
        return res_p;
      case ANSWER_FIRST:
        prev_p=p;
        p=monomial_next(p);
        break;
      case ANSWER_SECOND:
        monomial_set_next(prev_p,m);
        monomial_set_next(m,p);
        return res_p;
        break;
    }
  };
  monomial_set_next(m,NULL);
  monomial_set_next(prev_p,m);
  return res_p;
}

int monopoli_size(monomial m, ring R)
{
  int c=0;
  while (m != NULL) {
    c+=monomial_size(m,R);
    m=monomial_next(m); }
  return c;
}

int monopoli_is_homog_return_deg(monomial m, ring R)
{
  int d;
  if (m==NULL) return -1;
  d=monomial_deg(m);
  m=monomial_next(m);
  while (m != NULL) {
    if (d != monomial_deg(m)) return -2; /* FALSE */
    m=monomial_next(m); }
  return d;
}

void monopoli_change_sign(monomial m, ring R)
{
  while (m != NULL) {
    monomial_change_sign(m,R);
    m=monomial_next(m); }
}

int monopoli_len(monomial m)
{
  int c=0;
  while (m != NULL) {
    c++;
    m=monomial_next(m); }
  return c;
}

monopoli monopoli_square(monopoli p, ring R)
{
  monopoli res_p;
  coeff aux_c;
  monomial aux_m,aux_m1;
  res_p=NULL;
  while (p != NULL) {
    aux_m1=monomial_square(p,R);
    if (aux_m1!=NULL) res_p=monopoli_sum_these(res_p,aux_m1,R);
    aux_c=coeff_sum(monomial_coeff(p),monomial_coeff(p),R);
    if (coeff_is_zero(aux_c,R)) {
      coeff_free(aux_c,R);
      p=monomial_next(p); }
    else {
      aux_m=monomial_mk(aux_c,term_dup(monomial_term(p)));
      p=monomial_next(p);
      res_p=monopoli_sum_these(res_p,monopoli_x_monomial(p,aux_m,R),R);
      monomial_free(aux_m,R); } }
  return res_p;
}

monopoli monopoli_this_x_term(monopoli p, term t, ring R)
{
  monopoli res_p=p;
  while (p != NULL) {
    term_mult_and_assign(monomial_term(p),t);
    p=monomial_next(p); }
  return res_p;
}

monopoli monopoli_this_minus_term_x_monopoli (monopoli p1, term ME,monopoli p2, 
                               ring R)
/* p1-ME*p2 */

{
  monopoli aux_p,res_p;
  int answer;

  res_p=monomial_dummy;
  aux_p=res_p;

  if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);

  while ((p1 != NULL) && (p2 != NULL)) {
    monopoli p1_next,p2_next;
    answer=R->term_which_gt(monomial_term(p1),monomial_term(p2),R->to,R);
    switch (answer) {
      case ANSWER_BOTH:
        p1_next=monomial_next(p1);
        p2_next=monomial_next(p2);
        R->SubCoeff(&monomial_coeff(p1),monomial_coeff(p1),
                                        monomial_coeff(p2),R);
        if (coeff_is_zero(monomial_coeff(p1),R))
          monomial_free(p1,R);
	else {
          monomial_set_next(aux_p,p1);
          aux_p=monomial_next(aux_p); }
	monomial_free(p2,R);
        p1=p1_next;
        p2=p2_next;
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
	break;
      case ANSWER_FIRST:
        monomial_set_next(aux_p,p1);
        aux_p=monomial_next(aux_p);
        p1=monomial_next(p1);
        break;
      case ANSWER_SECOND:
        monomial_change_sign(p2,R);
        monomial_set_next(aux_p,p2);
        aux_p=monomial_next(aux_p);
        p2=monomial_next(p2);
        if (p2 != NULL) term_mult_and_assign(monomial_term(p2),ME);
        break;
    }
  }
  if (p1 == NULL) {
    monomial_set_next(aux_p,p2);
    while (p2 != NULL) {
        term_mult_and_assign(monomial_term(p2),ME);
        p2=monomial_next(p2); }
  }
  else monomial_set_next(aux_p,p1);
  return monomial_next(res_p);
}

void monopoli_change_order(monopoli * p, ring R1, ring R2)
{ 
  *p=monopoli_sum_this_monomial_list(*p,R2);
}

monopoli monopoli_this_x_monomial(monopoli p, monomial m, ring R)
{
  monopoli res_p=p;
  coeff aux_c;
  term aux_t;
  aux_c=monomial_coeff(m);
  aux_t=monomial_term(m);
  while (p != NULL) {
    R->MulCoeff(&monomial_coeff(p),monomial_coeff(p),aux_c,R);
    term_mult_and_assign(monomial_term(p),aux_t);
    p=monomial_next(p); }
  return res_p;
}

monomial monopoli_last(monomial m)
{
  if (m==NULL) return NULL;
  while (monomial_next(m) != NULL) {
    m=monomial_next(m); }
  return m;
}

monopoli monopoli_sum_this_monomial_list_to_this_monopoli(monomial m, monopoli p, ring R)
{
  monomial next_m;
  while (m != NULL){
    next_m = monomial_next(m);
    p=monopoli_add_this_monomial( p, m, R );
    m=next_m; }
  return p;
}

ctbuf monopoli2ctbuf(monomial m, ring R) 
{
  ctbuf aux_b;
  cterm aux_t;

  aux_b = ctbuf_empty();
  while (m != NULL){
    aux_t=monomial2cterm(m,R);
    ctbuf_append(aux_b,aux_t);
    m = monomial_next(m); }
  return aux_b;
}

cterm monopoli2cterm(monomial m, ring R)
{
  return ctbuf_mk_cterm(monopoli2ctbuf(m,R),op_POLYNOMIAL);
}

olist monopoli2sppnum(monopoli p, ring R)
{
  olist res_ol,aux_ol;
  monomial aux_m;
  int n;
  res_ol = olist_init(monopoli_len(p));
  n=0;
  while (p != NULL) {
    aux_ol=olist_init(2);
    olist_set_nth(aux_ol, 1,
      object_clean(coeff2object(monomial_coeff(p),R)));
    olist_set_nth(aux_ol, 2, 
      olist2object(this_iv2olist(term_log(monomial_term(p),R))));
    olist_set_nth(res_ol, ++n, olist2object(aux_ol)); 
    term_free(monomial_term(p));
    aux_m=p;
    p=monomial_next(p);
    free_monomial(aux_m);
  }
  return res_ol;
}

/* --------------[ TODO_PUBLIC ]----------------- */

#if 0 /* TODO_PUBLIC */

monopoli monopoli_x_opp_term(monopoli p, term t, ring R)
{
  TODO("monopoli_x_opp_term");
  return NULL;
}

monopoli monopoli_this_x_opp_term(monopoli p, term t, ring R)
{
  TODO("monopoli_this_x_opp_term");
  return NULL;
}

/*
PolyToSP(P)    := Tagged([ [LC(M),Log(M)] | M In Monomials(P) ],'SpPoly');
*/


void printf_monopoli(monomial m, ring R)
{
  cterm aux_t;
  aux_t=monopoli2cterm(m,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

monomial monopoli_x_monomial(monomial ml, monomial m, ring R)
{
  monopoli aux_p,aux_p1;
  coeff aux_c,res_c;
  term aux_t,aux_t1,aux_t2;

  aux_c=monomial_coeff(m);
  monomial_set_next(monomial_dummy,NULL);
  aux_t2=monomial_term(m);
  aux_p=monomial_dummy;
  while (p != NULL) {
    coeff_set_mult(res_c,aux_c,monomial_coeff(p),R);

#ifndef IS_DOMAIN

    if (coeff_is_zero(res_c,R))
      R->ClearCoeff(&res_c);
    else {
      aux_t1=monomial_term(p);
      term_set_mult(aux_t,aux_t1,aux_t2);
      monomial_set_next(aux_p, monomial_mk(res_c, aux_t);
      aux_p=monomial_next(aux_p); }

#else

    aux_t1=monomial_term(p);
    term_set_mult(aux_t,aux_t1,aux_t2);
    monomial_set_mk(aux_p1,res_c,aux_t);
    monomial_set_next(aux_p, aux_p1);
    aux_p=aux_p1;

#endif

    p=monomial_next(p); }
  return monomial_next(monomial_dummy);
}


#endif /* TODO_PUBLIC */

#endif /* MONOPOLI */
