/* $Id: object.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ctbuf.h"
#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "cterm.h"
#include "cocfun.h"
#include "ctbuf.h"
#include "engine.h"
#include "evector.h"
#include "ivectors.h"
#include "monomial.h"
#include "module.h"
#include "object.h"
#include "operator.h"
#include "olist.h"
#include "omat.h"
#include "options.h"
#include "pipe.h"
#include "polys.h"
#include "pseries.h"
#include "pvector.h"
#include "ratfun.h"
#include "ring.h"
#include "record.h"
#include "terms.h"
#include "utils.h"
#include "vlist.h"

#ifdef OBJECT_DEBUG
#define OBJECT_DEBUG_MAX 1000
int objects_map[OBJECT_DEBUG_MAX+1];
#endif

object object_null,object_BREAK,object_ZERO,object_ONE,object_RETURN_NULL;
extern object math_error;

coc_bool objects_same_label(object o1, object o2)
{
  if (object_is_labelled(o1))
    if (object_is_labelled(o2))
      return strcmp(object_label(o1), object_label(o2)) == 0;
    else return FALSE;
  if (object_is_labelled(o2)) return FALSE;
  return TRUE;
}


object object_clean(object o)
{
  ring R;
  int c;
  R = object_ring(o);
  switch (object_type(o)) {
    case objtype_coeff: 

      c=ring_characteristic(R);
      switch (c) {
        case characteristic_Li: object_set_type(o,objtype_arbint); break;
        case characteristic_Lr: object_set_type(o,objtype_arbrat); break;
        case characteristic_Si: object_set_type(o,objtype_int); break;
        case characteristic_Sr: /* TODO: remove */
        case characteristic_Sd: object_set_type(o,objtype_double); break;
        default: if (c > 1) {
          object_set_type(o,objtype_zmod);
          o->v.zmod_value=zmod_mk((int)o->v.coeff_value,c);
        }
      }
      ring_unlink(R);
      object_set_ring(o,ring_NULL);
      break;
#ifdef ARPOLY_EEE
#if 0 /* TODO_EEE */
    case objtype_poly: poly_optimize_this(o->v.poly_value,R); break;
#endif
#endif
    case objtype_vlist: 
      o=object_cast(o,objtype_olist,R);
  }
  return o;
}

coc_bool objtype_lt(objtype t1,objtype t2)
{
  switch (t1) {

    case objtype_string: switch (t2) {
      case objtype_cdev: 
        return TRUE; } break;

    case objtype_cdev: switch (t2) {
      case objtype_string: 
        return TRUE; } break;

    case objtype_int: switch (t2) {
      case objtype_arbint: 
      case objtype_arbrat:
      case objtype_zmod:
      case objtype_coeff:
      case objtype_poly: 
      case objtype_ratfun: 
        return TRUE; } break;
   
    case objtype_arbint: switch (t2) {
      case objtype_arbrat: 
      case objtype_zmod:
      case objtype_poly: 
      case objtype_coeff:
      case objtype_ratfun: 
        return TRUE; } break;

    case objtype_arbrat: switch (t2) {
      case objtype_coeff:
      case objtype_poly: 
      case objtype_ratfun: 
        return TRUE; } break;

    case objtype_zmod: switch (t2) {
      case objtype_coeff:
      case objtype_poly:
      case objtype_ratfun:
        return TRUE; } break;

    case objtype_coeff: switch (t2) {
      case objtype_poly: 
        return TRUE; } break;

    case objtype_poly: switch (t2) {
      case objtype_ratfun: 
        return TRUE; } break;

    case objtype_ideal: switch (t2) {
      case objtype_module: 
        return TRUE; } break;

    default: return FALSE;
  }
  return FALSE;
}

void object_init_adt(void)
{

#ifdef OBJECT_DEBUG
  objects_map[0]=0;
#endif

  object_BREAK=cterm2object(cterm_mk_const(op_BREAK));
  object_RETURN_NULL=object_init(objtype_NULL,NULL,ring_NULL);
  object_null=object_init(objtype_NULL,NULL,ring_NULL);
  object_ZERO=int2object(0);
  object_ONE=int2object(1);
}

object object_simplify(object o)
{ 
  objtype t;
  ring R;
  t=object_type(o);
  R=object_ring(o);
  switch(t) {
  case objtype_ratfun:
    if ( ratfun_is_poly(o->v.ratfun_value,R) ) {
      object_set_type(o,objtype_poly);
      o->v.poly_value=this_ratfun2arpoly(o->v.ratfun_value,R); }
    break;
  case objtype_arbrat:
    if ( arbrat_is_arbint(o->v.arbrat_value) ) {
      object_set_type(o,objtype_arbint);
      o->v.arbint_value=this_arbrat2arbint(o->v.arbrat_value); }
    break; }
  return o;
}

object object_init(objtype t,char * id,ring R)
{
  object o;
  o=malloc_object();
  o->id=my_strdup(id);
  object_set_type(o,t);
  o->rc=1; /* self */
  o->v.NULL_value=NULL;
  switch(t)
  {
  case objtype_coeff: o->v.coeff_value=NULL;
    o->R=ring_link(R); break;
  case objtype_poly: o->v.poly_value=NULL; 
    o->R=ring_link(R); break;
  case objtype_vlist: o->v.vlist_value=vlist_NULL;
    o->R=ring_link(R); break;
  case objtype_ideal: o->v.module_value=module_NULL;
    o->R=ring_link(R); break;
  case objtype_module: o->v.module_value=module_NULL;
    o->R=ring_link(R); break;
  case objtype_ratfun: o->v.ratfun_value=ratfun_NULL; 
    o->R=ring_link(R); break;
  case objtype_pvector: o->v.pvector_value=pv_NULL; 
    o->R=ring_link(R); break;
  case objtype_pseries: o->v.pseries_value=NULL; break;
  case objtype_olist: o->v.olist_value=olist_NULL; o->R=ring_NULL; break;
  case objtype_omat: o->v.omat_value=omat_NULL; o->R=ring_NULL; break;
  case objtype_bool: o->v.bool_value=FALSE; o->R=ring_NULL; break;
  case objtype_cterm: o->v.cterm_value=cterm_NULL; o->R=ring_NULL; break;
  case objtype_int: o->v.int_value=0; o->R=ring_NULL; break;
  case objtype_string: o->v.string_value=NULL; o->R=ring_NULL; break;
  case objtype_ring: o->v.ring_value=NULL; o->R=ring_NULL; break;
  case objtype_panel: o->v.panel_value=NULL; o->R=ring_NULL; break;
  case objtype_err: o->v.err_value=NULL; o->R=ring_NULL; break;
  case objtype_NULL: o->v.NULL_value=NULL; o->R=ring_NULL; break;
  case objtype_objtype: o->v.objtype_value=objtype_NULL; o->R=ring_NULL; break;
  case objtype_arbint: o->v.arbint_value=arbint_NULL; o->R=ring_NULL; break;
  case objtype_arbrat: o->v.arbrat_value=arbrat_NULL; o->R=ring_NULL; break;
  case objtype_record: o->v.record_value=record_NULL; o->R=ring_NULL; break;
  case objtype_zmod: o->v.zmod_value=zmod_NULL; o->R=ring_NULL; break;
  case objtype_double: o->v.double_value=double_NULL; o->R=ring_NULL; break;
  case objtype_cdev: o->v.cdev_value=cdev_NULL; o->R=ring_NULL; break;
  case objtype_cocfun: o->v.cocfun_value=cocfun_NULL; o->R=ring_NULL; break;
  default: object_set_type(o,objtype_NULL);
    o->v.NULL_value=NULL; o->R=ring_NULL; 
    math_error=cocoa_internal_error("object_set_value");
  }

#ifdef OBJECT_DEBUG
  {
  o->num=objects_map[0]=MMOD(objects_map[0],OBJECT_DEBUG_MAX)+1;
  if (flag_is_on(opt_panel_DEBUG,flag_DEBUG_OBJ)) {
    char buffer[100];
    sprintf(buffer,"--> Created object %d (%s).\n",o->num,objtype2str(object_type(o)));
    ca_message_string(buffer); }
  objects_map[o->num]=(int)o; }
#endif

  return o;
}

void object_set_value(object o,valueType v)
{
  switch(object_type(o))
  {
  case objtype_NULL: o->v.NULL_value=NULL; break;
  case objtype_coeff: o->v.coeff_value=v.coeff_value; break;
  case objtype_ring: o->v.ring_value=v.ring_value; break;
  case objtype_panel: o->v.panel_value=v.panel_value; break;
  case objtype_poly: o->v.poly_value=v.poly_value; break;
  case objtype_bool: o->v.bool_value=v.bool_value; break;
  case objtype_cterm: o->v.cterm_value=v.cterm_value; break;
  case objtype_int: o->v.int_value=v.int_value; break;
  case objtype_pvector: o->v.pvector_value=v.pvector_value; break;
  case objtype_vlist: o->v.vlist_value=v.vlist_value; break;
  case objtype_string: o->v.string_value=v.string_value; break;
  case objtype_ideal:
  case objtype_module: o->v.module_value=v.module_value; break;
  case objtype_pseries: o->v.pseries_value=v.pseries_value; break;
  case objtype_objtype: o->v.objtype_value=v.objtype_value; break;
  case objtype_olist: o->v.olist_value=v.olist_value; break;
  case objtype_omat: o->v.omat_value=v.omat_value; break;
  case objtype_ratfun: o->v.ratfun_value=v.ratfun_value; break;
  case objtype_err: o->v.err_value=v.err_value; break;
  case objtype_arbint: o->v.arbint_value=v.arbint_value; break;
  case objtype_arbrat: o->v.arbrat_value=v.arbrat_value; break;
  case objtype_record: o->v.record_value=v.record_value; break;
  case objtype_zmod: o->v.zmod_value=v.zmod_value; break;
  case objtype_double: o->v.double_value=v.double_value; break;
  case objtype_cdev: o->v.cdev_value=v.cdev_value; break;
  case objtype_cocfun: o->v.cocfun_value=v.cocfun_value; break;
  default: math_error=cocoa_internal_error("object_set_value");
  }
}

void object_set_duplicate_value(object o,valueType v,ring R)
{
  switch(object_type(o))
  {
  case objtype_NULL: o->v.NULL_value=NULL; break;
  case objtype_coeff: o->v.coeff_value=coeff_dup(v.coeff_value,R); break;
  case objtype_ring: o->v.ring_value=ring_link(v.ring_value); break;
  case objtype_panel: o->v.panel_value=v.panel_value; break; /* TODO ? */
  case objtype_poly: o->v.poly_value=arpoly_dup(v.poly_value,R); break;
  case objtype_bool: o->v.bool_value=v.bool_value; break;
  case objtype_cterm: o->v.cterm_value=cterm_dup(v.cterm_value); break;
  case objtype_int: o->v.int_value=v.int_value; break;
  case objtype_pvector: o->v.pvector_value=pv_dup(v.pvector_value,R); break;
  case objtype_vlist: o->v.vlist_value=vlist_dup(v.vlist_value,R); break;
  case objtype_string: o->v.string_value=my_strdup(v.string_value); break;
  case objtype_ideal:
  case objtype_module: o->v.module_value=module_dup(v.module_value,R); break;
  case objtype_pseries: o->v.pseries_value=pseries_dup(v.pseries_value); break;
  case objtype_objtype: o->v.objtype_value=v.objtype_value; break;
  case objtype_olist: o->v.olist_value=olist_dup(v.olist_value); break;
  case objtype_omat: o->v.omat_value=omat_dup(v.omat_value); break;
  case objtype_ratfun: o->v.ratfun_value=ratfun_dup(v.ratfun_value,R); break;
  case objtype_err: o->v.err_value=my_strdup(v.err_value); break;
  case objtype_arbint: o->v.arbint_value=arbint_dup(v.arbint_value); break;
  case objtype_arbrat: o->v.arbrat_value=arbrat_dup(v.arbrat_value); break;
  case objtype_record: o->v.record_value=record_dup(v.record_value); break;
  case objtype_zmod: o->v.zmod_value=zmod_dup(v.zmod_value); break;
  case objtype_double: o->v.double_value=double_dup(v.double_value); break;
  case objtype_cdev: o->v.cdev_value=cdev_link(v.cdev_value); break;
  case objtype_cocfun: o->v.cocfun_value=cocfun_link(v.cocfun_value); break;
  default: math_error=cocoa_internal_error("object_set_duplicate_value");
  }
}

ring object_extract_ring(object o)
{
  if (object_ring(o) != ring_NULL) return object_ring(o);
  switch(object_type(o))
  {
  case objtype_olist: return olist_extract_ring(o->v.olist_value);
  case objtype_omat: return omat_extract_ring(o->v.omat_value);
  case objtype_record: return record_extract_ring(o->v.record_value);
  default: return ring_NULL;
  }
}

coc_bool object_extract_ring_ck(object o, ring * res_R)
{
  *res_R=object_ring(o);
  if (*res_R != ring_NULL) return TRUE;
  switch(object_type(o))
  {
  case objtype_olist: return olist_extract_ring_ck(o->v.olist_value, res_R);
  case objtype_omat: return omat_extract_ring_ck(o->v.omat_value, res_R);
  case objtype_record: return record_extract_ring_ck(o->v.record_value, res_R);
  default: return TRUE;
  }
}

coc_bool object_is_ring_homogeneous(object o) {
  ring dummy_R;
  return object_extract_ring_ck(o,&dummy_R);
}

void object_empty(object o)
{
  if (o->id!=NULL) {
    free_str1(o->id);
    o->id=NULL;
  }
  object_set_type(o,objtype_NULL);
  o->v.NULL_value=NULL;
}

object object_dup(object o)
{
  object res_o;
  if (o==object_NULL) return object_NULL;
  res_o=object_init(object_type(o),object_label(o),o->R);
  object_set_duplicate_value(res_o,o->v,o->R); 
  return res_o;
}

void object_free(object o)
{
  ring R;
  if (o==object_NULL) return;
  if (o->rc!=ZERO) {
    INTERNAL_ERROR("attempt to free a referenced object");
    return;
  }
  R = o->R;
  if (o->id!=NULL) {
    free_str1(o->id);
    o->id=NULL;
  }
  switch(object_type(o))
  {
    case objtype_NULL: break;
    case objtype_coeff:
      (*R->ClearCoeff)(&o->v.coeff_value);
      break;
    case objtype_ring: 
      ring_unlink(o->v.ring_value);
      o->v.ring_value=ring_NULL;
      break;
    case objtype_panel: 
      /* opt_panel_free(o->v.ring_value); */ /* TODO */
      o->v.ring_value=ring_NULL;
      break;
    case objtype_poly: 
      arpoly_free(o->v.poly_value,R); 
      o->v.poly_value=NULL;
      break;
    case objtype_bool: /* do nothing */ break;
    case objtype_cterm: cterm_free(o->v.cterm_value); break;
    case objtype_int: /* do nothing */ break;
    case objtype_pvector: 
      pv_free(o->v.pvector_value,R); break;
    case objtype_vlist: 
      vlist_free(&o->v.vlist_value,R); break;
    case objtype_string: 
      if (o->v.string_value!=NULL)
        free_str1(o->v.string_value); 
      break;
    case objtype_ideal: 
    case objtype_module: 
      module_unlink(o->v.module_value,R); 
      o->v.module_value=NULL;
      break;
    case objtype_pseries:
      pseries_free(o->v.pseries_value);
      o->v.pseries_value=NULL;
      break;
    case objtype_objtype: /* do nothing */ break;
    case objtype_olist:
      olist_free(o->v.olist_value);
      o->v.olist_value=NULL;
      break;
    case objtype_omat:
      omat_free(o->v.omat_value);
      o->v.omat_value=NULL;
      break;
    case objtype_ratfun:
      ratfun_free(o->v.ratfun_value,R);
      o->v.ratfun_value=NULL;
      break;
    case objtype_err: 
      if (o->v.err_value!=NULL)
        free_str(strlen(o->v.err_value),o->v.err_value); 
      break;
    case objtype_arbint:
      LiClear(&o->v.arbint_value);
      o->v.arbint_value=NULL;
      break;
    case objtype_arbrat:
      LrClear(&o->v.arbrat_value);
      o->v.arbrat_value=NULL;
      break;
    case objtype_record:
      record_free(o->v.record_value);
      o->v.record_value=NULL;
      break;
    case objtype_zmod:
      if (o->v.zmod_value != NULL) {
        free_zmod(o->v.zmod_value);
        o->v.zmod_value=NULL; }
      break;
    case objtype_double:
      free_double(o->v.double_value);
      o->v.double_value=NULL;
      break;
    case objtype_cdev: 
      cdev_unlink(o->v.cdev_value);
      o->v.cdev_value=cdev_NULL;
      break;
    case objtype_cocfun: 
      cocfun_unlink(o->v.cocfun_value);
      o->v.cocfun_value=cocfun_NULL;
      break;
    default: math_error=cocoa_internal_error("object_free");
    }

#ifdef OBJECT_DEBUG
  { if (flag_is_on(opt_panel_DEBUG,flag_DEBUG_OBJ)) {
      char buffer[100];
      sprintf(buffer,"--> Deleted object %d.\n",o->num);
      ca_message_string(buffer); }
    objects_map[o->num]=ZERO; }
#endif
  ring_unlink(R);
  free_object(o);
}

void printf_object(object o, cdev D)
{
  cterm t;
  t=object2cterm(o);
  ca_put(t,D);
  cterm_free(t);
}

void printf_objects(cdev D)
{
#ifdef OBJECT_DEBUG
  cterm t;
  t=objects2cterm();
  ca_put(t,D);
  cterm_free(t);
#else
  ca_message_string("Cannot print objects: set OBJECT_DEBUG flag and recompile");
#endif
}

cterm objects2cterm(void)
{
#ifdef OBJECT_DEBUG
  ctbuf aux_b;
  int i,n;
  aux_b=ctbuf_empty();
  n=objects_map[0];
  for (i=1;i<=n;i++)
    if (objects_map[i]!=ZERO) {
      cterm params[3];
      object aux_o;
      aux_o=(object)(objects_map[i]);
      params[0]=int2cterm(i);
      params[1]=objtype2cterm(object_type(aux_o),NULL);
      params[2]=object2cterm(aux_o);
      ctbuf_append(aux_b,cterm_mk_classic(op_OBJECT,3,params)); }
  return ctbuf_mk_cterm(aux_b,op_OBJECTS);
#else
  return cterm_NULL;
#endif
}

void object_describe(object o, cdev D)
{
  cterm aux_t;
  aux_t=object_description(o);
  ca_put(aux_t,D);
  cterm_free(aux_t);
}

cterm object_description(object o)
{
  ctbuf aux_b,aux_b1,aux_b2;
  cterm_list params[2];
  cterm aux_t, res_t;
  aux_b1=ctbuf_empty();
  aux_t=cterm_mk_table_entry("Type",objtype2cterm(object_type(o),
    object_label(o)));
  ctbuf_append(aux_b1, aux_t);
#ifdef OBJECT_DEBUG
  if (flag_is_on(opt_panel_DEBUG,flag_SHOW_OBJNUM)) {
    aux_t=cterm_mk_table_entry("Number",int2cterm(object_num(o)));
    ctbuf_append(aux_b1, aux_t);
  }
  aux_t=cterm_mk_table_entry("Ring",ring2cterm(object_ring(o)));
  ctbuf_append(aux_b1, aux_t);
  aux_t=cterm_mk_table_entry("RC",int2cterm(object_rc(o)));
  ctbuf_append(aux_b1, aux_t);
#endif

  res_t=cterm_NULL;
  switch(object_type(o))
  {
    case objtype_ideal:
      aux_b=ideal_description(o->v.module_value,o->R);
      aux_t=ctbuf_mk_cterm(aux_b,op_RECORD); 
      aux_b2=ctbuf_new(cterm_mk_table_entry("Value",aux_t));
      break;
    case objtype_module:
      aux_b=module_description(o->v.module_value,o->R);
      aux_t=ctbuf_mk_cterm(aux_b,op_RECORD); 
      aux_b2=ctbuf_new(cterm_mk_table_entry("Value",aux_t));
      break;
    case objtype_cocfun:
      res_t=cocfun_description(o->v.cocfun_value);
      break;
    default: {
      aux_t=cterm_mk_table_entry("Value",object2cterm(o));
      aux_b2=ctbuf_new(aux_t); }
  }
  if (res_t == cterm_NULL) {
    aux_b=ctbuf_concat_these(aux_b1,aux_b2);
    res_t=ctbuf_mk_cterm(aux_b,op_RECORD);
  }
  return res_t;
}


/* ...2object functions */

object int2object(int i)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_int,NULL,ring_NULL);
  aux_v.int_value=i;
  object_set_value(res_o,aux_v);
  return res_o;
}

object bool2object(coc_bool b)
{
  object res_o;
  valueType aux_v;
  if (b==bool_NULL) return object_NULL;
  res_o=object_init(objtype_bool,NULL,ring_NULL);
  aux_v.bool_value=b;
  object_set_value(res_o,aux_v);
  return res_o;
}

object zmod2object(zmod i)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_zmod,NULL,ring_NULL);
  aux_v.zmod_value=i;
  object_set_value(res_o,aux_v);
  return res_o;
}

object double2object(double * i)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_double,NULL,ring_NULL);
  aux_v.double_value=i;
  object_set_value(res_o,aux_v);
  return res_o;
}

object arbrat2object(arbrat i)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_arbrat,NULL,ring_NULL);
  aux_v.arbrat_value=i;
  object_set_value(res_o,aux_v);
  return res_o;
}

object arbint2object(arbint i)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_arbint,NULL,ring_NULL);
  aux_v.arbint_value=i;
  object_set_value(res_o,aux_v);
  return res_o;
}

object cterm2object(cterm t)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_cterm,NULL,ring_NULL);
  aux_v.cterm_value=t;
  object_set_value(res_o,aux_v);
  return res_o;
}

object cdev2object(cdev D)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_cdev,NULL,ring_NULL);
  aux_v.cdev_value=D;
  object_set_value(res_o,aux_v);
  return res_o;
}

object cocfun2object(cocfun f)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_cocfun, NULL, ring_NULL);
  aux_v.cocfun_value=f;
  object_set_value(res_o,aux_v);
  return res_o;
}

object ring2object(ring vR)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_ring,NULL,ring_NULL);
  aux_v.ring_value=vR;
  object_set_value(res_o,aux_v);
  return res_o;
}

object panel2object(opt_panel p)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_panel,NULL,ring_NULL);
  aux_v.panel_value=p;
  object_set_value(res_o,aux_v);
  return res_o;
}

object arpoly2object(arpoly p,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_poly,NULL,R);
  aux_v.poly_value=p;
  object_set_value(res_o,aux_v);
  return res_o;
}

object coeff2object(coeff c,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_coeff,NULL,R);
  aux_v.coeff_value=c;
  object_set_value(res_o,aux_v);
  return res_o;
}

object ratfun2object(ratfun F,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_ratfun,NULL,R);
  aux_v.ratfun_value=F;
  object_set_value(res_o,aux_v);
  return res_o;
}

object pv2object(pvector v,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_pvector,NULL,R);
  aux_v.pvector_value=v;
  object_set_value(res_o,aux_v);
  return res_o;
}

object vlist2object(vlist vl,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_vlist,NULL,R);
  if (vl==vlist_NULL) vl=vlist_init();
  aux_v.vlist_value=vl;
  object_set_value(res_o,aux_v);
  return res_o;
}

object string2object(char * v)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_string,NULL,ring_NULL);
  aux_v.string_value=v;
  object_set_value(res_o,aux_v);
  return res_o;
}

object error2object(char * v)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_err,NULL,ring_NULL);
  aux_v.err_value=v;
  object_set_value(res_o,aux_v);
  return res_o;
}

object module2object(module M,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_module,NULL,R);
  aux_v.module_value=M;
  object_set_value(res_o,aux_v);
  return res_o;
}

object ideal2object(module M,ring R)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_ideal,NULL,R);
  aux_v.module_value=M;
  object_set_value(res_o,aux_v);
  return res_o;
}

object record2object(record r)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_record,NULL,ring_NULL);
  aux_v.record_value=r;
  object_set_value(res_o,aux_v);
  return res_o;
}

object objtype2object(objtype ot, char * tag)
{
  object res_o;
  valueType aux_v;
  if (ot==objtype_int) ot=objtype_arbint; /* TODO: check */
  res_o=object_init(objtype_objtype,tag,ring_NULL);
  if (tag != NULL) {
    aux_v.objtype_value=objtype_tagged;
  } else {
    aux_v.objtype_value=ot;
  }
  object_set_value(res_o,aux_v);
  return res_o;
}

object olist2object(olist ol)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_olist,NULL,ring_NULL);
  aux_v.olist_value=ol;
  object_set_value(res_o,aux_v);
  return res_o;
}

object omat2object(omat om)
{
  object res_o;
  valueType aux_v;
  res_o=object_init(objtype_omat,NULL,ring_NULL);
  aux_v.omat_value=om;
  object_set_value(res_o,aux_v);
  return res_o;
}


/* object2... functions */

coc_bool object2bool(object o)
{
  coc_bool res_b;
  res_b=o->v.bool_value;
  object_unlink(o);
  return res_b;
}


#define term_pure_power(t,index,exp)  iv_almost_null(t,index,exp)

unknown object2indet(object o)
{ /* TODO: optimize (not urgent) */
  unknown res_x;
  ring R=o->R;
  res_x=unknown_NULL;
  if (object_type(o)==objtype_poly) res_x = arpoly2indet(o->v.poly_value,R);
  if (res_x==unknown_NULL) math_error=cocoa_error("EXPECTED_INDET");
  object_unlink(o);
  return res_x;
}

coeff object2coeff(object o)
{
  coeff res_c;
  if ((o==object_NULL) || (object_type(o)!=objtype_coeff))
    ca_message_string("INTERNAL ERROR: Function object2coeff");
  else {
    res_c=o->v.coeff_value;
    if (object_rc_is_not_one(o)) res_c=coeff_dup(res_c,o->R);
    else object_empty(o);
    object_unlink(o); }
  return res_c;
}

arpoly object2arpoly(object o,ring R)
{
  arpoly res_p;
  res_p=o->v.poly_value;
  if (object_rc_is_not_one(o)) res_p=arpoly_dup(res_p,R);
  else object_empty(o);
  object_unlink(o);
  return res_p;
}

zmod object2zmod(object o,ring R)
{
  zmod res_p;
  res_p=o->v.zmod_value;
  if (object_rc_is_not_one(o)) res_p=zmod_dup(res_p);
  else object_empty(o);
  object_unlink(o);
  return res_p;
}

double * object2double(object o,ring R)
{
  double * res_p;
  res_p=o->v.double_value;
  if (object_rc_is_not_one(o)) res_p=double_dup(res_p);
  else object_empty(o);
  object_unlink(o);
  return res_p;
}

ratfun object2ratfun(object o,ring R)
{
  ratfun res_F;
  res_F=o->v.ratfun_value;
  if (object_rc_is_not_one(o)) res_F=ratfun_dup(res_F,R);
  else object_empty(o);
  object_unlink(o);
  return res_F;
}

pvector object2pv(object o,ring R)
{
  pvector res_v;
  res_v=o->v.pvector_value;
  if (object_rc_is_not_one(o)) res_v=pv_dup(res_v,R);
  else object_empty(o);
  object_unlink(o);
  return res_v;
}

cdev object2cdev(object o)
{
  cdev res_d;
  res_d=o->v.cdev_value;
  if (object_rc_is_not_one(o)) res_d=cdev_link(res_d);
  else object_empty(o);
  object_unlink(o);
  return res_d;
}

cocfun object2cocfun(object o)
{
  cocfun res_f;
  res_f=o->v.cocfun_value;
  if (object_rc_is_not_one(o)) res_f=cocfun_link(res_f);
  else object_empty(o);
  object_unlink(o);
  return res_f;
}

vlist object2vlist(object o,ring R)
{
  vlist res_vl;
  res_vl=o->v.vlist_value;
  if (object_rc_is_not_one(o)) res_vl=vlist_dup(res_vl,R);
  else object_empty(o);
  object_unlink(o);
  return res_vl;
}

char * object2string(object o)
{
  char * res_s;
  if (object_type(o)!=objtype_string) { object_unlink(o); return my_strdup(""); }
  res_s=o->v.string_value;
  if (object_rc_is_not_one(o)) res_s=my_strdup(res_s);
  else object_empty(o);
  object_unlink(o);
  return res_s;
}

char * object2error(object o)
{
  char * res_s;
  if (object_type(o)!=objtype_err) return my_strdup("");
  res_s=o->v.err_value;
  if (object_rc_is_not_one(o)) res_s=my_strdup(res_s);
  else object_empty(o);
  object_unlink(o);
  return res_s;
}

module object2module(object o,ring R)
{
  module res_M;
  res_M=o->v.module_value;
  if (object_rc_is_not_one(o)) res_M=module_link(res_M);
  else object_empty(o);
  object_unlink(o);
  return res_M;
}

cterm object2cterm(object o)
{ 
  cterm t;
  ring R;

  if (o==object_NULL) return cterm_NULL;

  R = o->R;

  switch(object_type(o))
  {
  case objtype_NULL: t=cterm_mk_const(op_NULL_VALUE); break;
  case objtype_coeff:
    (*R->Coeff2Cterm)(&t,o->v.coeff_value,R);
    break;
  case objtype_ring: {
    t=ring2cterm(o->v.ring_value); }
    break;
  case objtype_panel: t=opt_panel2cterm(o->v.panel_value); break; 
  case objtype_poly: t=arpoly2cterm(o->v.poly_value,R); break;
  case objtype_bool: t=bool2cterm(o->v.bool_value); break;
  case objtype_cterm: t=cterm_dup(o->v.cterm_value); break;
  case objtype_int: t=int2cterm(o->v.int_value);break;
  case objtype_pvector: t=pv2cterm(o->v.pvector_value,R); break;
  case objtype_vlist: t=vlist2cterm(o->v.vlist_value,R); break;
  case objtype_string: t=string2cterm(o->v.string_value); break;
  case objtype_ideal: t=ideal2cterm(o->v.module_value,R); break;
  case objtype_module: t=module2cterm(o->v.module_value,R); break;
  case objtype_pseries: t=pseries2cterm(o->v.pseries_value); break;
  case objtype_objtype: 
    t=objtype2cterm(o->v.objtype_value,object_label(o)); break;
  case objtype_olist: t=olist2cterm(o->v.olist_value); break;
  case objtype_omat: t=omat2cterm(o->v.omat_value); break;
  case objtype_ratfun: t=ratfun2cterm(o->v.ratfun_value,R); break;
  case objtype_err: t=err2cterm(o->v.err_value); break;
  case objtype_arbint: Li2Cterm(&t,o->v.arbint_value,R); break;
  case objtype_arbrat: Lr2Cterm(&t,o->v.arbrat_value,R); break;
  case objtype_record: t=record2cterm(o->v.record_value); break;
  case objtype_zmod: t=zmod2cterm(o->v.zmod_value); break;
  case objtype_double: t=double2cterm(o->v.double_value); break;
  case objtype_cdev: t=cdev2cterm(o->v.cdev_value); break;
  case objtype_cocfun: t=cocfun2cterm(o->v.cocfun_value); break;
  default: t=cterm_mk_const(op_objtype_NULL);
  }
  if (current_ring != R && R != ring_NULL) {
    cterm params[2];
    params[0]=cterm_mk_NBS(ring_id(R));
    params[1]=t;
    t=cterm_mk_classic(op_COCOLON,2,params);
  }
  /* Print out a tagged object as Tagged(..,tag) EXCEPT if it is TAGGED(...) */
  if (cterm_op(t) != op_objtype_tagged && object_is_labelled(o))
  {
    cterm params[2];
    params[0] = t;
    params[1] = string2cterm(object_label(o));
    t = cterm_mk_classic(op_LABEL,2,params);
  }
  return t;
}

ring object2ring(object o)
{
  ring res_R=ring_NULL;
  res_R=o->v.ring_value;
  if (object_rc_is_not_one(o)) {
    res_R=ring_link(res_R);
  } else object_empty(o);
  object_unlink(o);
  return res_R;
}

opt_panel object2panel(object o)
{
  opt_panel res_p=opt_panel_NULL;
  if (o!=object_NULL)
    if (object_type(o)==objtype_panel) res_p=o->v.panel_value;
  /* NOTE: the object is inalterated */
  object_unlink(o);
  return res_p;
}

object object_coeff2arpoly(object o,ring R)
/* NOTE: R is the ring of polynomials */
{
  coeff aux_c;
  arpoly aux_p=NULL;
  aux_c=object2coeff(o);
  aux_p=this_coeff2arpoly(aux_c,R);
  return arpoly2object(aux_p,R);
}

olist object2olist(object o)
{
  olist res_ol;
  res_ol=o->v.olist_value;
  if (object_rc_is_not_one(o)) res_ol=olist_dup(res_ol);
  else object_empty(o);
  object_unlink(o);
  return res_ol;
}

record object2record(object o)
{
  record res_r;
  res_r=o->v.record_value;
  if (object_rc_is_not_one(o)) res_r=record_dup(res_r);
  else object_empty(o);
  object_unlink(o);
  return res_r;
}

omat object2omat(object o)
{
  omat res_m;
  res_m=o->v.omat_value;
  if (object_rc_is_not_one(o)) res_m=omat_dup(res_m);
  else object_empty(o);
  object_unlink(o);
  return res_m;
}

/* predicates */

coc_bool object_true(object o)
{
  return o->v.bool_value;
}

cterm objtype2cterm(objtype ot, char * tag)
{
  cterm t;
  switch(ot)
  {
  case objtype_NULL: t=cterm_mk_const(op_objtype_NULL); break;
  case objtype_coeff: t=cterm_mk_const(op_objtype_coeff); break;
  case objtype_ring: t=cterm_mk_const(op_objtype_ring); break;
  case objtype_poly: t=cterm_mk_const(op_objtype_poly); break;
  case objtype_bool: t=cterm_mk_const(op_objtype_bool); break;
  case objtype_cterm: t=cterm_mk_const(op_objtype_cterm); break;
  case objtype_int: t=cterm_mk_const(op_objtype_int);break;
  case objtype_pvector: t=cterm_mk_const(op_objtype_pvector); break;
  case objtype_string: t=cterm_mk_const(op_objtype_string); break;
  case objtype_ideal: t=cterm_mk_const(op_objtype_ideal); break;
  case objtype_module: t=cterm_mk_const(op_objtype_module); break;
  case objtype_pseries: t=cterm_mk_const(op_objtype_pseries); break;
  case objtype_objtype: t=cterm_mk_const(op_objtype_objtype); break;
  case objtype_olist: t=cterm_mk_const(op_objtype_olist); break;
  case objtype_omat: t=cterm_mk_const(op_objtype_omat); break;
  case objtype_ratfun: t=cterm_mk_const(op_objtype_ratfun); break;
  case objtype_panel: t=cterm_mk_const(op_objtype_panel); break;
  case objtype_err: t=cterm_mk_const(op_objtype_err); break;
  case objtype_arbint: t=cterm_mk_const(op_objtype_arbint); break;
  case objtype_arbrat: t=cterm_mk_const(op_objtype_arbrat); break;
  case objtype_record: t=cterm_mk_const(op_objtype_record); break;
  case objtype_zmod: t=cterm_mk_const(op_objtype_zp); break;
  case objtype_double: t=cterm_mk_const(op_objtype_double); break;
  case objtype_tagged: {
    if (tag==NULL) {
      math_error=cocoa_internal_error("objtype2cterm");
      t = cterm_NULL;
    } else {
      cterm params[1];
      params[0]=string2cterm(tag);
      t=cterm_mk_classic(op_objtype_tagged,1,params);
    }
    break;
  }
  case objtype_cdev: t=cterm_mk_const(op_objtype_cdev); break;
  case objtype_cocfun: t=cterm_mk_const(op_objtype_cocfun); break;
  default: t=cterm_mk_const(op_objtype_NULL);
  }
  return t;
}

char * objtype2str(objtype ot)
{
  char * res_s;
  switch(ot)
  {
  case objtype_NULL: res_s="NULL"; break;
  case objtype_coeff: res_s="COEFF"; break;
  case objtype_ring: res_s="RING"; break;
  case objtype_poly: res_s="POLY"; break;
  case objtype_bool: res_s="BOOL"; break;
  case objtype_cterm: res_s="CTERM"; break;
  case objtype_int: res_s="INT"; break; /* machine integer? */
  case objtype_pvector: res_s="VECTOR"; break;
  case objtype_string: res_s="STRING"; break;
  case objtype_ideal: res_s="IDEAL"; break;
  case objtype_module: res_s="MODULE"; break;
  case objtype_pseries: res_s="PSERIES"; break;
  case objtype_objtype: res_s="TYPE"; break;
  case objtype_olist: res_s="LIST"; break;
  case objtype_omat: res_s="MAT"; break;
  case objtype_ratfun: res_s="RATFUN"; break;
  case objtype_panel: res_s="PANEL"; break;
  case objtype_err: res_s="ERROR"; break;
  case objtype_arbint: res_s="INT"; break;
  case objtype_arbrat: res_s="RAT"; break;
  case objtype_record: res_s="RECORD"; break;
  case objtype_zmod: res_s="ZMOD"; break;
  case objtype_double: res_s="DOUBLE"; break;
  case objtype_tagged: res_s="TAGGED"; break;
  case objtype_cocfun: res_s="FUNCTION"; break;
  default: res_s="???";
  }
  return res_s;
}

/* conversion functions */

weilist convert_object2weilist(object o, ring R)
{
  weilist res_wl;
  switch(object_type(o))
  {
  case objtype_olist: 
  case objtype_omat: 
    res_wl=olist2weilist(o->v.olist_value, R);
    break;
  default: res_wl=weilist_NULL;
  }
  return res_wl;
}

/* CASTING */

/* IMPORTANT: object_cast* unlink the input object o */

object object_cast(object o, objtype t, ring res_R)
{
  object res_o;
  objtype ot=object_type(o);
  if (ot==t) return o;
  res_o=object_init(t,NULL,res_R);
  switch (t) {
    case objtype_ideal: 
      res_o->v.module_value=object_cast_ideal(o,res_R);
      if (math_error == object_NULL) {
        if (module_rank(res_o->v.module_value) <= 1)
          object_set_type(res_o,objtype_ideal);
        else {
          math_error=object_error_in_cast(ot,t);
          error_set_this_param(cterm_mk_NBS("MODULE to IDEAL: rank must be 1"));
        }
      }
      break;
    case objtype_module: 
      res_o->v.module_value=object_cast_module(o,res_R); break;
    case objtype_pvector: 
      res_o->v.pvector_value=object_cast_pvector(o,res_R); break;
    case objtype_poly: res_o->v.poly_value=object_cast_arpoly(o,res_R); break;
    case objtype_coeff: res_o->v.coeff_value=object_cast_coeff(o,res_R); break;
    case objtype_omat: res_o->v.omat_value=object_cast_omat(o); break;
    case objtype_olist: res_o->v.olist_value=object_cast_olist(o); break;
    case objtype_vlist: res_o->v.vlist_value=object_cast_vlist(o,res_R); break;
    case objtype_string: res_o->v.string_value=object_cast_string(o);
      break;
    case objtype_cdev: res_o->v.cdev_value=object_cast_cdev(o);
      break;
    case objtype_arbint: res_o->v.arbint_value=object_cast_arbint(o); break;
    case objtype_arbrat: res_o->v.arbrat_value=object_cast_arbrat(o); break;
    case objtype_ratfun: 
      res_o->v.ratfun_value=object_cast_ratfun(o,res_R); break;
    case objtype_int: res_o->v.int_value=object_cast_int(o); break;
    default: {
      math_error=object_error_in_cast(ot,t);
      object_unlink(o); }
  }
  if (math_error) {
    object_unlink(res_o);
    res_o=math_error;
    math_error=object_NULL;
  }
  return res_o;
}

char * object_cast_string(object o)
{
  char * res_S=NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_string: res_S=v.string_value; break;
    case objtype_cdev:   
      switch (cdev_type(v.cdev_value)) {
      case cdev_type_ostring:
        charbuf_cork(cdev_the_ostring(v.cdev_value));
        res_S=my_strdup(charbuf_buffer(cdev_the_ostring(v.cdev_value)));
	cdev_unlink(v.cdev_value);
	break;
      default: res_S=my_strdup(""); 
      }
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_string);
      object_unlink(aux_o); return res_S;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_S;
}

cdev object_cast_cdev(object o)
{
  cdev res_D=cdev_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_cdev: res_D=v.cdev_value; break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_cdev);
      object_unlink(aux_o); return res_D;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_D;
}

ratfun object_cast_ratfun(object o, ring R)
{
  ratfun res_F=ratfun_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);

  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_int:
      res_F=this_arpoly2ratfun(this_arbint2arpoly(
        int2arbint(v.int_value),R),R); 
      break;

    case objtype_arbint:
      res_F=this_arpoly2ratfun(this_arbint2arpoly(v.arbint_value,R),R); break;

    case objtype_zmod:
      if (ring_characteristic(R) != zmod_p(v.zmod_value)) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (characteristic <> %d)", 
          objtype2str(objtype_ratfun),zmod_p(v.zmod_value));
        error_set_this_param(cterm_mk_NBS(buffer)); }
      else res_F=this_arpoly2ratfun(int2arpoly(zmod_n(v.zmod_value),R),R);
      break;

    case objtype_poly:
      res_F=this_arpoly2ratfun(v.poly_value,R); break;

    case objtype_arbrat: {
      arbrat r=v.arbrat_value;
      arpoly aux_p1,aux_p2,aux_p,res_p;
      res_p=NULL;

      if (!ring_coeff_in_field(R)) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (coeffs are not in a field)", 
          objtype2str(objtype_ratfun));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      else {
        aux_p2=arbint2arpoly(arbrat_den(r),R);

        if (aux_p2 == NULL) {
          char buffer[100];
          math_error=cocoa_error("CANNOT_CAST");
          sprintf(buffer,"to %s (division by zero)", 
            objtype2str(objtype_ratfun));
          error_set_this_param(cterm_mk_NBS(buffer));
        }

        else {
          aux_p1=arbint2arpoly(arbrat_num(r),R);
          arpoly_div(aux_p1,aux_p2,&res_p,&aux_p,R);
          arpoly_free(aux_p1,R);
          arpoly_free(aux_p,R);
          res_F=this_arpoly2ratfun(res_p,R); }
        arpoly_free(aux_p2,R);
        free_verylongrat(r); } }
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_ratfun);
      object_unlink(aux_o); return res_F;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_F;
}

arbint object_cast_arbint(object o)
{

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;
  arbint res_I;
  
  res_I=arbint_NULL;

  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_arbint: res_I=v.arbint_value; break;
    case objtype_arbrat: 
      if (arbrat_is_arbint(v.arbrat_value))
        res_I=this_arbrat2arbint(v.arbrat_value); 
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s", objtype2str(objtype_arbint));
        error_set_this_param(cterm_mk_NBS(buffer));
        object_unlink(aux_o);
        return arbint_NULL;
      }
      break;
    case objtype_zmod:
      res_I=int2arbint(zmod_n(v.zmod_value));
      free_zmod(v.zmod_value);
      break;
    case objtype_coeff: res_I=this_coeff2arbint(v.coeff_value,R); break;
    case objtype_poly: 
      if (arpoly_deg(v.poly_value,R) > 0) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (non constant poly)", 
          objtype2str(objtype_arbint));
        error_set_this_param(cterm_mk_NBS(buffer));
        object_unlink(aux_o);
        return arbint_NULL;
      }
      else 
      {
        res_I=this_arpoly2arbint(v.poly_value,R);
        if (res_I == arbint_NULL)
          math_error=object_error_in_cast(t,objtype_arbint);
        break;
      }
    case objtype_int:
      res_I=int2arbint(v.int_value); break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_arbint);
      object_unlink(aux_o); return res_I;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */
  
  return res_I;
}


arbrat object_cast_arbrat(object o)
{

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;
  arbrat res_I;

  res_I=arbrat_NULL;
  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_arbrat: res_I=v.arbrat_value; break;
    case objtype_arbint: res_I=this_arbint2arbrat(v.arbint_value); break;
    case objtype_zmod: 
      res_I=int2arbrat(zmod_sym_n(v.zmod_value));
      free_zmod(v.zmod_value);
      break;
    case objtype_int:
      res_I=int2arbrat(v.int_value); break;
    case objtype_poly: 
      if (arpoly_deg(v.poly_value,R) > 0) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (non constant poly)", 
          objtype2str(objtype_arbint));
        error_set_this_param(cterm_mk_NBS(buffer));
        object_unlink(aux_o); return res_I;
      }
      else 
        res_I=this_arpoly2arbrat(v.poly_value,R); break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_arbrat);
      object_unlink(aux_o); return res_I;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */
  return res_I;
}

int object_cast_int(object o)
{
  int res_I=ZERO;

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;

  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_int: res_I=v.int_value; break;
    case objtype_coeff: 
      if (coeff_is_int(v.coeff_value,R))
        res_I=this_coeff2int(v.coeff_value,R);
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (integer too big)", 
          objtype2str(objtype_int));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      break;

    case objtype_poly: 
      if (arpoly_is_zero(v.poly_value,R)) break;
      if (arpoly_deg(v.poly_value,R) > 0) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (non constant poly)", 
          objtype2str(objtype_int));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      else if (coeff_is_int(arpoly_this_leading_coeff(v.poly_value,R),R))
        res_I=this_arpoly2int(v.poly_value,R); 
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (integer too big)", 
          objtype2str(objtype_int));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      break;
    case objtype_zmod:
      res_I=zmod_n(v.zmod_value);
      free_zmod(v.zmod_value);
      break;
    case objtype_arbrat:
      if (arbrat_is_int(v.arbrat_value))
        res_I=this_arbrat2int(v.arbrat_value); 
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s", 
          objtype2str(objtype_int));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      break;
    case objtype_arbint:
      if (arbint_is_int(v.arbint_value))
        res_I=this_arbint2int(v.arbint_value); 
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (integer too big)", 
          objtype2str(objtype_int));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_int);
      object_unlink(aux_o); return res_I;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_I;
}


coc_bool object_cast_bool(object o)
{
  coc_bool res_I=bool_NULL;

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;

  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_bool: res_I=v.bool_value; break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_bool);
      object_unlink(aux_o); return res_I;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_I;
}

module object_cast_module(object o, ring R)
{
  module res_M=module_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (t == objtype_err) { 
    math_error=o; 
    return res_M; /* <----[HERE]---- */
  }

  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_module:
    case objtype_ideal: res_M=v.module_value; break;
    case objtype_pvector: 
      res_M=module_generated_by_this_pv(v.pvector_value,R); 
      break;
    case objtype_arbint: {
      arpoly aux_p;
      aux_p=this_arbint2arpoly(v.arbint_value,R);
      res_M=module_generated_by_this_pv(pv_this_arpoly2pv(aux_p,0,1,R),R); }
      break;
    case objtype_poly: 
      res_M=
        module_generated_by_this_pv(pv_this_arpoly2pv(v.poly_value,0,1,R),R); 
      break;
    case objtype_olist:
      res_M=this_olist2module(v.olist_value,R); 
      break;
    case objtype_vlist:
      res_M=this_vlist2module(v.vlist_value,vlist_rank(v.vlist_value),R); 
      break;
    case objtype_omat:
      res_M=this_omat2module(v.omat_value,R); 
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_module);
      object_unlink(aux_o); return res_M;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_M;
}


module object_cast_ideal(object o, ring R)
{
  module res_M=module_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (t == objtype_err) { 
    math_error=o; 
    return res_M; /* <----[HERE]---- */
  }

  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_module: 
      res_M=v.module_value; 
      if (vlist_len(module_gens(res_M)) == 0)
        vlist_append_this_vector(module_gens(res_M),pv_zero(1,R),R);
      break;
    case objtype_ideal: res_M=v.module_value; break;
    case objtype_pvector: 
      res_M=module_generated_by_this_pv(v.pvector_value,R); 
      break;
    case objtype_arbint: {
      arpoly aux_p;
      aux_p=this_arbint2arpoly(v.arbint_value,R);
      res_M=module_generated_by_this_pv(pv_this_arpoly2pv(aux_p,0,1,R),R); }
      break;
    case objtype_poly: 
      res_M=
        module_generated_by_this_pv(pv_this_arpoly2pv(v.poly_value,0,1,R),R); 
      break;
    case objtype_olist:
      res_M=this_olist2ideal(v.olist_value,R); 
      break;
    case objtype_vlist:
      res_M=this_vlist2module(v.vlist_value,vlist_rank(v.vlist_value),R); 
      break;
    case objtype_omat:
      res_M=this_omat2module(v.omat_value,R); 
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_ideal);
      object_unlink(aux_o); return res_M;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_M;
}

pvector object_cast_pvector(object o, ring R)
{
  pvector res_v=pv_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_pvector: 
      res_v=v.pvector_value; break;
    case objtype_int: 
      res_v=pv_this_arpoly2pv(int2arpoly(v.int_value,R),0,1,R); break;
    case objtype_arbint: 
      res_v=pv_this_arpoly2pv(this_arbint2arpoly(v.arbint_value,R),0,1,R); 
      break;
    case objtype_poly: 
      res_v=pv_this_arpoly2pv(v.poly_value,0,1,R); break;
    case objtype_olist:
      res_v=this_olist2pv(v.olist_value,R); break;
    case objtype_omat:
      if (olist_len(v.omat_value) == 1) {
        object aux_o2;
        aux_o2=olist_nth(v.omat_value,1);
        res_v=this_olist2pv(aux_o2->v.olist_value,R);
        free_olist(1,v.olist_value); }
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (matrix has more than one row)", 
          objtype2str(objtype_pvector));
        error_set_this_param(cterm_mk_NBS(buffer)); }
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_pvector);
      object_unlink(aux_o); return res_v;
 }/* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_v;
}

omat object_cast_omat(object o)
{
  omat res_M=omat_NULL;

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;

  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_omat: 
      res_M=v.omat_value; break;
    case objtype_olist:
      if (olist_is_omat(v.olist_value))       /* check if it is a rect array */
        res_M=this_olist2omat(v.olist_value); /* if so, just tag it as a MAT */
      else {                                  /* if not, succede un casino...*/
        math_error=cocoa_error("MAT_RECT");
      }
      break;
    case objtype_pvector: {
      olist aux_ol;
      aux_ol=olist_init(1);
      olist_set_nth(aux_ol,1,olist2object(this_pv2olist(v.pvector_value,R)));
      res_M=this_olist2omat(aux_ol); }
      break;
    case objtype_vlist:
      res_M=this_vlist2omat(v.vlist_value,R); break;
    case objtype_ideal:
    case objtype_module:
      res_M=vlist2omat(module_gens_ck(v.module_value),R);
      module_unlink(v.module_value,R);
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_omat);
      object_unlink(aux_o); return res_M;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_M;
}


arpoly object_cast_arpoly(object o, ring R)
{
  arpoly res_p=NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_poly: 
      if (object_ring(aux_o) != R)
      { math_error = cocoa_error("RING_DEP");
        object_unlink(aux_o);
        return res_p;
      }
      res_p=v.poly_value; break;
    case objtype_arbint: res_p=this_arbint2arpoly(v.arbint_value,R); break;

    case objtype_arbrat: {
      arbrat r=v.arbrat_value;
      arpoly aux_p1,aux_p2,aux_p;
      res_p=NULL;

      if (!ring_coeff_in_field(R)) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (coeffs are not in a field)", 
          objtype2str(objtype_poly));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      else {
        aux_p2=arbint2arpoly(arbrat_den(r),R);

        if (aux_p2 == NULL) {
          char buffer[100];
          math_error=cocoa_error("CANNOT_CAST");
          sprintf(buffer,"to %s (division by zero)", 
            objtype2str(objtype_poly));
          error_set_this_param(cterm_mk_NBS(buffer));
        }

        else {
          aux_p1=arbint2arpoly(arbrat_num(r),R);
          arpoly_div(aux_p1,aux_p2,&res_p,&aux_p,R);
          arpoly_free(aux_p1,R);
          arpoly_free(aux_p,R); }
        arpoly_free(aux_p2,R);
	LrClear(&r);
	}
      }
      break;
    case objtype_coeff: res_p=this_coeff2arpoly(v.coeff_value,R); break;
    case objtype_int: res_p=int2arpoly(v.int_value,R); break;
    case objtype_zmod:
      if (ring_characteristic(R) != zmod_p(v.zmod_value)) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (characteristic <> %d)", 
          objtype2str(objtype_poly),zmod_p(v.zmod_value));
        error_set_this_param(cterm_mk_NBS(buffer)); }
      else res_p=int2arpoly(zmod_n(v.zmod_value),R);
	free_zmod(v.zmod_value);
      break;
    case objtype_pvector:
      if (pv_rank(v.pvector_value) != 1) {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (vector rank <> 1)", 
          objtype2str(objtype_poly));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      else {
        res_p=this_pvcomp2arpoly(pv_nth(v.pvector_value,0),R);
        pv_set_nth(v.pvector_value,0,NULL);
	pv_free(v.pvector_value,R);
      }
      break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_poly);
      object_unlink(aux_o); return res_p;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_p;
}

coeff object_cast_coeff(object o, ring R)
{
  coeff res_c=coeff_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_poly:
      if (arpoly_deg(v.poly_value,R)==0) {
        res_c=arpoly_leading_coeff(v.poly_value,R);
        arpoly_free(v.poly_value,R); }
      else {
        char buffer[100];
        math_error=cocoa_error("CANNOT_CAST");
        sprintf(buffer,"to %s (poly has degree > 1)", 
          objtype2str(objtype_coeff));
        error_set_this_param(cterm_mk_NBS(buffer));
      }
      break;
    case objtype_arbint: res_c=this_arbint2coeff(v.arbint_value,R); break;
    case objtype_arbrat: {
      arbrat r=v.arbrat_value;
      arpoly aux_p,res_p;
      arpoly_div(this_arbint2arpoly(arbrat_num(r),R),
        this_arbint2arpoly(arbrat_den(r),R),&res_p,&aux_p,R);
      res_c=arpoly_leading_coeff(res_p,R);
      arpoly_free(aux_p,R);
      arpoly_free(res_p,R); }
      break;
    case objtype_int: res_c=int2coeff(v.int_value,R); break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_coeff);
      object_unlink(aux_o); return res_c;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_c;
}

olist object_cast_olist(object o)
{
  olist res_ol=olist_NULL;

  /* STANDARD PROLOGUE ... */

  ring R;
  objtype t;
  object aux_o;
  valueType v;

  R=object_ring(o); t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_NULL: 
      res_ol=olist_init(0); break;
    case objtype_olist: 
      res_ol=v.olist_value; break;
    case objtype_vlist: 
      res_ol=this_vlist2olist_bis(v.vlist_value,R); break;
    case objtype_pvector:
      res_ol=this_pv2olist(v.pvector_value,R); break;
    case objtype_omat:
      res_ol=v.omat_value; break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_olist);
      object_unlink(aux_o); return res_ol;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_ol;
}

vlist object_cast_vlist(object o, ring R)
{
  vlist res_vl=vlist_NULL;

  /* STANDARD PROLOGUE ... */

  objtype t;
  object aux_o;
  valueType v;

  t=object_type(o);
  if (object_rc_is_not_one(o)) {
    aux_o=object_dup(o);
    object_unlink(o); }
  else {
    aux_o=o; }
  v=object_value(aux_o);

  switch (t) {

  /* ... STANDARD PROLOGUE */

    case objtype_vlist: 
      res_vl=v.vlist_value; break;
    case objtype_olist: 
      res_vl=this_olist2vlist_bis(v.olist_value,R); break;

  /* STANDARD EPILOGUE ... (modify HERE) */

    default: {
      math_error=object_error_in_cast(t,objtype_vlist);
      object_unlink(aux_o); return res_vl;
 } /* <----[HERE]---- */
  }
  object_empty(aux_o);
  object_unlink(aux_o);

  /* ... STANDARD EPILOGUE */

  return res_vl;
}

object object_error_types(char * err,objtype t1, char * sep,objtype t2)
{
  object res_o;
  char buffer[100];
  res_o=cocoa_error(err);
  sprintf(buffer,"%s%s%s",objtype2str(t1),sep,objtype2str(t2));
  error_set_this_param(cterm_mk_NBS(buffer));
  return res_o;
}


object object_error_types_term(char * err,objtype t1, 
  char * sep,objtype t2, char * termin)
{
  object res_o;
  char buffer[100];
  res_o=cocoa_error(err);
  sprintf(buffer,"%s%s%s %s",objtype2str(t1),sep,objtype2str(t2),termin);
  error_set_this_param(cterm_mk_NBS(buffer));
  return res_o;
}

