/* $Id: object.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef OBJECT_H
#define OBJECT_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#ifndef COCTYPES_H
# include "coctypes.h"
#endif

#define malloc_object() (object)mymalloc(sizeof(object_info),"object");
#define free_object(p) myfree(sizeof(object_info),p,"object")

#define pvcomp2object(p,R) arpoly2object(this_pvcomp2arpoly(p,R),R)

#define object_cast_pvcomp(o,R) this_arpoly2pvcomp(object_cast_arpoly(o,R),R)

#define object_rings_compatible(R,S) \
  rings_compatible(object_ring(R),object_ring(S))

#define objects_eq_comparable(o1,o2) \
  objtype_comparable(object_type(o1),object_type(o2))

#define objtype_le(t1,t2) \
  ((t1)==(t2) || objtype_lt(t1,t2))

#define objtype_comparable(t1,t2) \
  ((t1)==(t2) || objtype_lt(t1,t2) || objtype_lt(t2,t1))

#define object_incr_rc(o) ((o)->rc++)
#define object_decr_rc(o) ((o)->rc--)
#define object_rc_is_zero(o) ((o)->rc==ZERO)
#define object_rc_is_one(o) ((o)->rc==1)
#define object_rc_is_not_one(o) ((o)->rc!=1)

#define object_link(o) (object_incr_rc(o),o)
#define object_unlink(o) do \
  { if ((o)!=object_NULL) {\
      object_decr_rc(o); \
      if (object_rc_is_zero(o)) object_free(o);}} while(FALSE)

#define object_rc(o) ((o)->rc)
#define object_type(o) ((o)->t)

#define object_value(o) ((o)->v)
#define object_ring(o) ((o)->R)
#define object_set_type(o,T) (o)->t=(T)
#define object_is_error(o) ((o)->t==objtype_err)
#define object_num(o) ((o)->num)

#define object_set_ring(o,r) (o)->R=(r)
#define object_set_label(o,l) (o)->id=(l)
#define object_label(o) ((o)->id)
#define object_is_labelled(o) ((o)->id!=NULL)
coc_bool objects_same_label(object o1, object o2);

#define object_error_in_cast(t1,t2) \
  object_error_types("CANNOT_CAST",t1," to ",t2)

extern object object_BREAK,object_RETURN_NULL;

object object_clean(object o);

object object_error_types(char * err, objtype t1, char * sep,objtype t2);
object object_error_types_term(char * err,objtype t1, 
  char * sep,objtype t2, char * termin);

coc_bool objtype_lt(objtype t1,objtype t2);

object object_simplify(object o);

void object_init_adt(void);

object object_init(objtype t,char * id,ring R);
/* initialize an object with type t and identifier id and returns it */

void object_set_value(object o,valueType v);
/* update with v the valueType of the object o */

void object_empty(object o);
/* empty the object o */

object object_dup(object o);
/* create a copy of o and return it */

void object_free(object o);
/* free the object o, clearing the memory used for it */

void printf_object(object o, cdev D);

void printf_objects(cdev D);
cterm objects2cterm(void);

void object_describe(object o, cdev D);
cterm object_description(object o);

ring object_extract_ring(object o);
coc_bool object_extract_ring_ck(object o, ring * res_R);
coc_bool object_is_ring_homogeneous(object o);

#define object_is_ring_indep(o) (object_extract_ring(o) == ring_NULL)


/* ...2object functions */

/* ------------------------------------------------ *\

                   IMPORTANT

The ...2object functions move a value into an object.
They do NOT duplicate the value.
If you want to take a copy of the value you must
duplicate it before call the ...2object function.

\* ------------------------------------------------ */

object bool2object(coc_bool b);
object int2object(int i);
object zmod2object(zmod i);
object double2object(double * i);
object arbrat2object(arbrat i);
object arbint2object(arbint i);
object cterm2object(cterm t);
object cdev2object(cdev D);
object cocfun2object(cocfun f);
object ring2object(ring vR);
object panel2object(opt_panel p);
object coeff2object(coeff c,ring R);
object arpoly2object(arpoly p,ring R);
object ratfun2object(ratfun F,ring R);
object pv2object(pvector v,ring R);
object vlist2object(vlist vl,ring R);
object string2object(char * s);
object error2object(char * s);
object module2object(module M,ring R);
object ideal2object(module M,ring R);
object objtype2object(objtype ot, char * tag);
object olist2object(olist ol);
object omat2object(omat om);

object pseries2object(PSeries ps,ring R);
object record2object(record r);


/* object2... functions */

/* ------------------------------------------------ *\

                   IMPORTANT

The object2... functions extract a value from an object.

The functions always remove a link from the object.
If the object is referenced only one time (rc=1)
after this operation the object is cleared.
If you want to take a copy of the object you must
link it before call the object2... function
(see function object_link).

If the object is in the memory (o->l!=NULL) the
value of the object is duplicated.

\* ------------------------------------------------ */



coc_bool object2bool(object o);

#define object2int(o) object_cast_int(o)

int object2int(object o);
/* convert the object o to a integer */

unknown object2indet(object o);

coeff object2coeff(object o);
/* returns the coeff contained in the object o */

zmod object2zmod(object o,ring R);
double * object2double(object o,ring R);

ratfun object2ratfun(object o,ring R);

arpoly object2arpoly(object o,ring R);
/* returns the arpoly contained in the object o */

PSeries object2pseries(object o,ring R);

pvector object2pv(object o,ring R);
cdev object2cdev(object o);
cocfun object2cocfun(object o);
vlist object2vlist(object o,ring R);
char * object2string(object o);
char * object2error(object o);
module object2module(object o,ring R);
olist object2olist(object o);
omat object2omat(object o);
record object2record(object o);
char * object2err(object o);

/* 
  The function object2cterm is an exception: the object is only read
  and not cleared after the ...2... operation.
*/

cterm object2cterm(object o);
/* builds the cterm t associated with the object o and returns t */


/* 
  The function object2ring is an exception: the object is only read
  and not cleared after the ...2... operation.
  This is because the ring is a heavy object.
  The object is used only to have the address of the ring.
*/

ring object2ring(object o);
/* returns the ring contained in the object o */

/* object2panel is another exception */
opt_panel object2panel(object o);


/*
  The following function returns a new object of type poly
  starting from a coefficient.
  The object passed is clean.
*/

object object_coeff2arpoly(object o,ring R);
/* returns the object of type poly with value o of type coeff */
/* NOTE: the object is unlinked and a new object is returned */

/* predicates */

coc_bool object_true(object o);
/* return TRUE iff the object o contains TRUE */

cterm objtype2cterm(objtype ot, char * tag);
char * objtype2str(objtype ot);

/* conversion functions */

weilist convert_object2weilist(object o, ring R);

/* CASTING */

/* IMPORTANT: object_cast* unlink the input object o */

object object_cast(object o, objtype t, ring res_R);
ratfun object_cast_ratfun(object o, ring R);
int object_cast_int(object o);
arbint object_cast_arbint(object o);
arbrat object_cast_arbrat(object o);
coc_bool object_cast_bool(object o);
module object_cast_module(object o, ring R);
module object_cast_ideal(object o, ring R);
cdev object_cast_cdev(object o);
pvector object_cast_pvector(object o, ring R);
omat object_cast_omat(object o);
arpoly object_cast_arpoly(object o, ring R);
coeff object_cast_coeff(object o, ring R);
olist object_cast_olist(object o);
vlist object_cast_vlist(object o, ring R);
char * object_cast_string(object o);

#endif
