/* $Id: olist.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "coeffs.h"
#include "evector.h"
#include "indet.h"
#include "ivectors.h"
#include "module.h"
#include "math_lib.h"
#include "monomial.h"
#include "object.h"
#include "options.h"
#include "operator.h"
#include "olist.h"
#include "pipe.h"
#include "polys.h"
#include "pvector.h"
#include "ring.h"
#include "terms.h"
#include "utils.h"
#include "vlist.h"
#include "weights.h"

extern object math_error;
extern object cocoa_error(char * id);

/* internal */
void olist_cartesian_aux(int p, int nn, olist ol, olist res_ol);

olist olist_init(int n)
{
  int i;
  olist res_ol = olist_NULL;
  if (n+3 < INT_MAX/sizeof(object)) res_ol=malloc_olist(n);
  if (res_ol == olist_NULL) { math_error = cocoa_error("TOO_BIG"); return res_ol; }
  olist_set_len(res_ol,n);
  olist_set_size(res_ol,n);
  for (i=1;i<=n;i++) olist_set_nth(res_ol,i,object_NULL);
  return res_ol;
}

void olist_remove(olist ol, int j)
{
  int i,n;
  if (ol == olist_NULL) return;
  n = olist_len(ol);
  if (j > n) {
    math_error=cocoa_error("INDEX_TOO_BIG");
    return;
  } else if (j <= 0 ) {
    math_error=cocoa_error("INDEX_NEG");
    return;
  }
  object_unlink(olist_nth(ol, j));
  for (i=j+1; i<=n; i++) olist_set_nth(ol,i-1,olist_nth(ol, i));
  olist_set_len(ol,n-1);
}

olist olist_expand(olist l, int n)
{
  olist res_ol;
  res_ol=realloc_olist(l,n);
  olist_set_size(res_ol,olist_size(res_ol)+n);
  return res_ol;
}

olist olist_new(int n, object o)
{
  int i;
  olist res_ol;
  res_ol=malloc_olist(n);
  olist_set_len(res_ol,n);
  olist_set_size(res_ol,n);
  for (i=1;i<=n;i++) olist_set_nth(res_ol,i,object_link(o));
  return res_ol;
}

void olist_free(olist ol)
{
  int i,n;
  if (ol==olist_NULL) return;
  n=olist_len(ol);
  for (i=1;i<=n;i++) object_unlink(olist_nth(ol,i));
  free_olist(n,ol);
}

olist olist_dup(olist ol)
{
  int i,n;
  olist res_ol;
  n=olist_len(ol);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) olist_set_nth(res_ol,i,object_link(olist_nth(ol,i)));
  return res_ol;
}

olist olist2olist_of_olists(olist ol)
{
  int i,n;
  olist res_ol;
  n=olist_len(ol);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) 
    olist_set_nth(res_ol,i,
      object_cast(object_link(olist_nth(ol,i)),objtype_olist,ring_NULL));
  return res_ol;
}

coc_bool olist_is_omat(olist ol)
{
  int i,l,l1;
  object aux_o;
  l=-1;
  for (i=1;i<=olist_len(ol);i++)
  {
    if (object_type(olist_nth(ol,i)) != objtype_olist
        && object_type(olist_nth(ol,i)) != objtype_pvector) return FALSE;
    aux_o=olist_nth(ol,i);
    l1 = do_len(aux_o);
    if (l != -1 && l1 != l) return FALSE;
    l=l1;
  }
  if (l == -1) return FALSE;
  return TRUE;
}

/* olist depends upon no more than one ring */
/* if the result is TRUE, then res_R contains the ring */
coc_bool olist_extract_ring_ck(olist ol, ring * res_R) {
  int i;
  ring aux_R;
  *res_R=ring_NULL;
  for (i=1;i<=olist_len(ol);i++) {
    if (!object_extract_ring_ck(olist_nth(ol,i),res_R)) return FALSE;
    if (*res_R != ring_NULL) break;
  }
  for (;i<=olist_len(ol);i++) {
    if (!object_extract_ring_ck(olist_nth(ol,i),&aux_R)) return FALSE;
    if ((aux_R != ring_NULL) && (*res_R != aux_R)) return FALSE;
  }
  return TRUE;
}

ring olist_extract_ring(olist ol)
{
  int i;
  ring res_R;
  for (i=1;i<=olist_len(ol);i++) 
    if ((res_R=object_extract_ring(olist_nth(ol,i))) != ring_NULL) return res_R;
  return ring_NULL;
}

olist olist_change_sign(olist ol)
{
  int i,n;
  olist res_ol;
  n=olist_len(ol);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) olist_set_nth(res_ol,i,do_change_sign(olist_nth(ol,i)));
  return res_ol;
}

olist olist_extprod(olist ol,object o)
{
  int i,n;
  olist res_ol;
  n=olist_len(ol);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) 
    olist_set_nth(res_ol,i,do_mult(olist_nth(ol,i),o));
  return res_ol;
}

cterm this_olist2cterm(olist ol)
{
  cterm res_t;
  res_t=olist2cterm(ol);
  olist_free(ol);
  return res_t;
}

cterm olist2cterm(olist ol)
{
  int i,n;
  cterm res_t;
  cterm_list params;
  n=olist_len(ol);
  if (n == 0) return cterm_mk_const(op_LIST);
  params=malloc_cterm_list(n);
  for (i=1;i<=n;i++) params[i-1]=object2cterm(olist_nth(ol,i));
  res_t=cterm_mk_classic(op_LIST,n,params);
  free_cterm_list(n,params);
  return res_t;
}

int olist_do_size(olist ol)
{
  int i,n,sz;
  n=olist_len(ol);
  sz=0;
  for (i=1;i<=n;i++) sz+=do_size(olist_nth(ol,i));
  return sz;
}

olist olist_concat_these(olist ol1, olist ol2)
{ /* TODO: OPTIMIZE WITH olist_expand */
  int i, n, n1, n2;
  olist res_ol;
  n1=olist_len(ol1);
  n2=olist_len(ol2);
  n=n1+n2;
  res_ol=olist_init(n);
  for (i=1;i<=n1;i++)
    olist_set_nth(res_ol,i,olist_nth(ol1,i));
  for (i=1;i<=n2;i++)
    olist_set_nth(res_ol,i+n1,olist_nth(ol2,i));
  free_olist(olist_len(ol1),ol1);
  free_olist(olist_len(ol2),ol2);
  return res_ol;
}

olist olist_append_this(olist ol, object o)
{ 
  int i, n, sz;
  n=olist_len(ol);
  sz=olist_size(ol);
  if (n==sz) ol=olist_expand(ol,SIZE_OF_OLIST_BLOCK);
  olist_set_nth(ol,n+1,o);
  olist_set_len(ol,n+1);
  return ol;
}

coc_bool olist_isin(olist ol, object o)
{
  int i;
  for (i=1;i<=olist_len(ol);i++)
    if (do_is_eq(olist_nth(ol,i),o)) return TRUE;
  return FALSE;
}

olist olist_tail(olist ol)
{
  int n,i;
  olist res_ol;
  n=olist_len(ol)-1;
  res_ol=olist_init(n);
  for (i=1;i<=n;i++)
    olist_set_nth(res_ol,i,object_link(olist_nth(ol,i+1)));
  return res_ol;
}

olist olist_cartesian(olist ol)
{ /* Ex: cartesian([[1,2]><[3,4]]) = [[1, 3], [1, 4], [2, 3], [2, 4]] */
  int n,i,nn;
  olist res_ol,aux_ol;
  object aux_o;
  n=olist_len(ol);
  nn=1;
  for (i=1;i<=n;i++) {
    double new_nn;
    aux_o=olist_nth(ol,i);
    aux_ol=aux_o->v.olist_value;
    if (olist_len(aux_ol) > INT_MAX/nn) { math_error=cocoa_error("TOO_BIG"); break; }
    new_nn = nn * olist_len(aux_ol);
    nn = new_nn;
  }
  if (math_error != object_NULL) return olist_NULL;
  if (nn == 0) return olist_init(0);
  res_ol=olist_init(nn);
  if (math_error != object_NULL) return olist_NULL;
  for (i=1;i<=nn;i++)
    olist_set_nth(res_ol,i,olist2object(olist_init(n)));
  olist_cartesian_aux(1,nn,ol,res_ol);
  /* free_olist(n,ol); */ /* TODO: remove */
  return res_ol;
}

void olist_cartesian_aux(int p, int nn, olist ol, olist res_ol)
{
  int n, nn1, res_n, times,i,j,k,ii;
  object aux_o;
  olist aux_ol, aux_ol2;
  if (p>olist_len(ol)) return;
  aux_o=olist_nth(ol,p);
  aux_ol=aux_o->v.olist_value;
  n = olist_len(aux_ol);
  nn1 = nn/n;
  res_n = olist_len(res_ol);
  times = res_n / nn;
  for (i=1;i<=times;i++) {
    for (j=1;j<=n;j++) {
      for (k=1;k<=nn1;k++) {
        ii = (i-1)*nn+(j-1)*nn1+k;
	/* printf("    [i=%d,j=%d,k=%d,ii=%d,p=%d]\n",i,j,k,ii,p); */
        aux_o=olist_nth(res_ol,ii);
        aux_ol2=aux_o->v.olist_value;
	/* printf("[Res[%d,%d]:=L[%d,%d]]\n",ii,p,p,j); */
        olist_set_nth(aux_ol2,p,object_link(olist_nth(aux_ol,j)));
      }
    } 
  }
  olist_cartesian_aux(p+1,nn1,ol,res_ol); 
}

olist olist_sum(olist ol1, olist ol2)
{
  olist res_ol;
  object aux_o;
  int i,n;
  n=olist_len(ol1);
  if (olist_len(ol2)!=n) {
    math_error=cocoa_error("OLIST_SIZE");
    return olist_NULL; }
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) {
    aux_o=do_sum(olist_nth(ol1,i),olist_nth(ol2,i));
    if (object_is_error(aux_o)) {
      math_error=aux_o;
      olist_free(res_ol);
      return olist_NULL; }
    olist_set_nth(res_ol,i,aux_o); }
  return res_ol;
}

olist olist_sub(olist ol1, olist ol2)
{
  olist res_ol;
  object aux_o;
  int i,n;
  n=olist_len(ol1);
  if (olist_len(ol2)!=n) {
    math_error=cocoa_error("OLIST_SIZE");
    return olist_NULL; }
  res_ol=olist_init(n);
  for (i=1;i<=n;i++) {
    aux_o=do_subtract(olist_nth(ol1,i),olist_nth(ol2,i));
    if (object_is_error(aux_o)) {
      math_error=aux_o;
      olist_free(res_ol);
      return olist_NULL; }
    olist_set_nth(res_ol,i,aux_o); }
  return res_ol;
}

coc_bool olist_eq(olist ol1,olist ol2)
{
  int n, i;
  object aux_o1,aux_o2;
  if ((n=olist_len(ol1)) != olist_len(ol2)) return FALSE;
  for (i=1;i<=n;i++) {
    aux_o1=olist_nth(ol1,i);
    aux_o2=olist_nth(ol2,i);
    if (!objects_eq_comparable(aux_o1,aux_o2)) return FALSE;
    if (!do_is_eq(aux_o1,aux_o2)) return FALSE; }
  return TRUE;
}

/* CONVERSIONS FUNCTIONS */

pvector this_olist2pv(olist ol, ring R)
{
  int l,i;
  object aux_o;
  pvector res_v;
  pvcomp aux_p;

  l=olist_len(ol);
  if (l==0) {
    olist_free(ol);
    math_error=cocoa_error("EMPTY_VECTOR");
    return pv_NULL; }
  res_v=pv_init(l);
  for (i=1;i<=l;i++) {
    aux_p=object_cast_pvcomp(olist_nth(ol,i),R);
    olist_set_nth(ol,i,object_NULL);
    if (math_error!=object_NULL) {
      pv_free(res_v,R);
      olist_free(ol);
      return pv_NULL;
    }
    pv_set_nth(res_v,i-1,aux_p);
  }
  olist_free(ol);
  return res_v;
}

module this_olist2ideal(olist ol, ring R)
{
  vlist aux_vl;
  if (olist_len(ol) == 0) 
    ol=olist_append_this(ol,int2object(0));
  aux_vl=this_olist2vlist_bis(ol,R);
  if (math_error!=object_NULL) return module_NULL;
  return this_vlist2module(aux_vl,vlist_rank(aux_vl),R);
}

module this_olist2module(olist ol, ring R)
{
  vlist aux_vl;
  aux_vl=this_olist2vlist(ol,R);
  if (math_error!=object_NULL) return module_NULL;
  return this_vlist2module(aux_vl,vlist_rank(aux_vl),R);
}

vlist this_olist2vlist(olist ol,ring R)
{
  int l,i;
  object aux_o;
  vlist res_vl;
  olist aux_ol;
  pvector aux_v;

  l=olist_len(ol);
  res_vl=vlist_init();
  if (l==0) {
    olist_free(ol);
    return res_vl; }

  if (objtype_le(object_type(olist_nth(ol,1)), objtype_poly)) {
    aux_ol=olist_init(1);
    olist_set_nth(aux_ol,1,olist2object(ol));
    ol=aux_ol;
    l=1;
  }

  for (i=1;i<=l;i++) {
    objtype aux_ty;
    aux_o=olist_nth(ol,i);
    aux_ty=object_type(aux_o);
    if (aux_ty==objtype_olist || aux_ty==objtype_pvector) {
      aux_v=object_cast_pvector(aux_o,R);
    } else {
      math_error=cocoa_error("CANNOT_CAST");
      error_set_this_param(cterm_mk_NBS("to VLIST")); }


    /* aux_v=object_cast_pvector(olist_nth(ol,i)); */
    if (math_error!=object_NULL) {
      vlist_free(&res_vl,R);
      olist_free(ol);
      return vlist_NULL;
    }
    olist_set_nth(ol,i,object_NULL);
    vlist_append_this_vector(res_vl,aux_v,R);
  }
  olist_free(ol);
  return res_vl;
}

vlist this_olist2vlist_bis(olist ol,ring R)
{
  int l,i;
  object aux_o;
  vlist res_vl;
  pvector aux_v;

  l=olist_len(ol);
  res_vl=vlist_init();
  if (l==0) {
    olist_free(ol);
    return res_vl; }
  for (i=1;i<=l;i++) {
    aux_v=object_cast_pvector(olist_nth(ol,i),R);
     
    if (math_error!=object_NULL) {
      vlist_free(&res_vl,R);
      olist_free(ol);
      return vlist_NULL;
    }
    olist_set_nth(ol,i,object_NULL);
    vlist_append_this_vector(res_vl,aux_v,R);
  }
  olist_free(ol);
  return res_vl;
}

static weight olist2weight_raw(olist L, ring R)
{
  int i,aux_i,n;
  weight w;
  n=ring_indets_num(R);

  if (n!=olist_len(L)) {
    char buffer[100];
    math_error=cocoa_error( "BAD_PARAMS_NUM" );
    sprintf(buffer,": wanted a list of %d weights but found %d.",n,olist_len(L));
    error_set_this_param(cterm_mk_NBS(buffer));
    return weight_NULL; }

  w = iv_init(n);
  for (i=1;i<=n;i++) {
    aux_i=object_cast_int(object_link(olist_nth(L,i)));
    if (math_error!=object_NULL) {
      iv_free(w);
      return weight_NULL; }
    iv_set_nth(w,i,aux_i); }
  return w;
}

weilist olist2weilist(olist ol, ring R)
/* ol=[ [1,2,3], [2,4,0] ] for example */
{
  weight tmp;
  weilist res_wl;
  object aux_o;
  int i, n, good;
  if (olist_len(ol) < 1)
  {
    math_error = cocoa_error("BAD_WEIGHTS");
    return weilist_NULL;
  }
  aux_o = olist_nth(ol, 1);
  tmp = olist2weight(aux_o->v.olist_value, R);
  if (tmp == weight_NULL)
  {
    math_error = cocoa_error("BAD_WEIGHTS");
    return weilist_NULL;
  }
  res_wl=weilist_init(olist_len(ol));
  weilist_set_nth(res_wl, 1, tmp);
  n = olist_len(ol);
  good = 1;
  for (i=2; i <= n; i++)
  {
    aux_o=olist_nth(ol,i);
    tmp = olist2weight_raw(aux_o->v.olist_value, R);
    if (tmp == weight_NULL) good = 0;
    weilist_set_nth(res_wl, i, tmp);
  }
  if (good) return res_wl;
  weilist_free(res_wl);
  /* math_error will already have been set if we reach here */
  return weilist_NULL;
}

weight olist2weight(olist L, ring R)
{
  weight ans = olist2weight_raw(L, R);
  int n = ring_indets_num(R);
  int i;

  if (ans == weight_NULL) return ans;
  for (i=1; i <= n; i++)
    if (iv_get_nth(ans, i) <= 0)
    {
      math_error = cocoa_error("BAD_WEIGHTS");
      iv_free(ans);
      return weight_NULL;
    }
  return ans;
}

olist this_vlist2olist(vlist vl,ring R)
{ /* returns an olist of olists */
  vlist_atom aux_atom;
  olist aux_ol,res_ol;
  int i,r,n;
  shifts ss;
  n=vlist_len(vl);
  vlist_rank_and_shifts(vl,&r,&ss);
  vlist_enlarge_elements(vl,r,ss);
  res_ol=olist_init(n);
  i=1;
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_ol=this_ev2olist(vlist_atom_ev(aux_atom),R);
    olist_set_nth(res_ol,i++,olist2object(aux_ol));
    vlist_atom_set_ev(aux_atom,ev_NULL);
    vlist_atom_free(&aux_atom,R); }
  vlist_free(&vl,R);
  return res_ol;
}

olist vlist2olist(vlist vl,ring R)
{ /* returns an olist of olists */
  vlist_atom aux_atom;
  olist aux_ol,res_ol;
  int i,r,n;
  shifts ss;
  n=vlist_len(vl);
  vlist_rank_and_shifts(vl,&r,&ss);
  vlist_enlarge_elements(vl,r,ss);
  res_ol=olist_init(n);
  i=1;
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_ol=ev2olist(vlist_atom_ev(aux_atom),R);
    olist_set_nth(res_ol,i++,olist2object(aux_ol));
    vlist_append(vl,aux_atom); }
  return res_ol;
}

olist this_vlist2olist_bis(vlist vl,ring R)
{ /* returns an olist of rank-homogeneous vectors */
  vlist_atom aux_atom;
  olist res_ol;
  pvector aux_v;
  int i,r,n;
  shifts ss;
  n=vlist_len(vl);
  vlist_rank_and_shifts(vl,&r,&ss);
  vlist_enlarge_elements(vl,r,ss);
  res_ol=olist_init(n);
  i=1;
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_v=ev_whole_vector(vlist_atom_ev(aux_atom),R);
    olist_set_nth(res_ol,i++,pv2object(aux_v,R));
    ev_unlink(vlist_atom_ev(aux_atom),R);
    vlist_atom_set_ev(aux_atom,ev_NULL);
    vlist_atom_free(&aux_atom,R); }
  vlist_free(&vl,R);
  return res_ol;
}

olist this_vlist2olist_ter(vlist vl,ring R)
{ /* returns the olist of the first polys of each vector */
  vlist_atom aux_atom;
  pvcomp aux_p;
  olist res_ol;
  pvector aux_v;
  int i;
  int n=vlist_len(vl);
  res_ol=olist_init(n);
  i=1;
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_v=ev_whole_vector(vlist_atom_ev(aux_atom),R);
    aux_p=pv_nth(aux_v,0);
    pvcomp_Apply_Anti_Shift(&aux_p,aux_v,1,R);
    olist_set_nth(res_ol,i++, pvcomp2object(aux_p,R));
    pv_set_nth(aux_v,0,NULL);
    pv_free(aux_v,R);
    vlist_atom_free(&aux_atom,R); }
  vlist_free(&vl,R);
  return res_ol;
}

olist this_pv2olist(pvector v, ring R)
{
  olist res_ol;
  pvcomp aux_p;
  int i,r;
  r=pv_rank(v);
  res_ol=olist_init(r);
  for (i=0;i<r;i++) {
    aux_p=pv_nth(v,i);
    pvcomp_Apply_Anti_Shift(&aux_p,v,i+1,R);
    olist_set_nth(res_ol,i+1,pvcomp2object(aux_p,R));
    pv_set_nth(v,i,NULL); }
  pv_free(v,R);
  return res_ol;
}

olist shifts2olist(shifts ss, ring R)
{
  olist res_ol;
  int i,r;
  term aux_t;
  arpoly aux_p;
  r=shifts_rank(ss);
  res_ol=olist_init(r);
  for (i=1;i<=r;i++) {
    if (shifts_are_int(ss)) {
      olist_set_nth(res_ol,i,int2object(shift_get_int(shifts_nth(ss,i))));
    } else {
      aux_t=term_dup(shifts_nth(ss,i));
      aux_p=arpoly_this_term2arpoly(aux_t,R);
      olist_set_nth(res_ol,i,arpoly2object(aux_p,R));
    }
  }
  return res_ol;
}

olist this_ev2olist(evector ev, ring R)
{
  olist res_ol;
  pvcomp aux_p;
  int i,r;
  pvector v;
  r=ev_rank(ev);
  res_ol=olist_init(r);
  if (ev_is_split(ev)) {
    for (i=0;i<r;i++) {
      aux_p=pvcomp_sum_this_monomial_list_to_this_pvcomp(
        mv_nth(ev_irr(ev),i), pv_nth(ev_red(ev),i),R);
      pvcomp_Apply_Anti_Shift(&aux_p,ev_red(ev),i+1,R);
      olist_set_nth(res_ol,i+1,pvcomp2object(aux_p,R));
      pv_set_nth(ev_irr(ev),i,NULL);
      pv_set_nth(ev_red(ev),i,NULL); }
  } else {
    v=ev_vector(ev);
    for (i=0;i<r;i++) {
      aux_p=pv_nth(v,i);
      pvcomp_Apply_Anti_Shift(&aux_p,v,i+1,R);
      olist_set_nth(res_ol,i+1,pvcomp2object(aux_p,R));
      pv_set_nth(v,i,NULL); }
  }
  pv_free(v,R);
  return res_ol;
}

olist ev2olist(evector ev, ring R)
{
  olist res_ol;
  pvcomp aux_p;
  int i,r;
  pvector v;
  r=ev_rank(ev);
  res_ol=olist_init(r);
  if (ev_is_split(ev)) {
    for (i=0;i<r;i++) {
      aux_p=pvcomp_sum_this_monomial_list_to_this_pvcomp(
        monomial_list_dup(mv_nth(ev_irr(ev),i),R), 
        pvcomp_dup(pv_nth(ev_red(ev),i),R),R);
      pvcomp_Apply_Anti_Shift(&aux_p,ev_red(ev),i+1,R);
      olist_set_nth(res_ol,i+1,pvcomp2object(aux_p,R)); }
  } else {
    v=ev_vector(ev);
    for (i=0;i<r;i++) {
      aux_p=pvcomp_dup(pv_nth(v,i),R);
      pvcomp_Apply_Anti_Shift(&aux_p,v,i+1,R);
      olist_set_nth(res_ol,i+1,pvcomp2object(aux_p,R)); }
  }
  return res_ol;
}

shifts this_olist2shifts(olist ol, ring R)
{
  int i, aux_n, r;
  shifts res_ss;
  object aux_o;
  term aux_t;
  arpoly aux_p;
  r = olist_len(ol);
  res_ss=shifts_init(r);
  for (i=1;i<=olist_len(ol);i++) {
    aux_o = olist_nth(ol,i);
    if (objtype_le(object_type(aux_o), objtype_arbint)) {
      aux_n=object_cast_int(aux_o);
      if (math_error!=object_NULL) {
        shifts_free(res_ss);
        olist_free(ol);
        return shifts_NULL;
      }
      shifts_set_nth(res_ss,i,unknown_pow2term(0,-aux_n,R));
      shifts_set_rank(res_ss,-r);
    } else if (objtype_le(object_type(aux_o), objtype_poly)) {
      aux_p=object_cast_arpoly(aux_o,R);
      if (math_error!=object_NULL) {
        shifts_free(res_ss);
        olist_free(ol);
        return shifts_NULL;
      }
      aux_t=arpoly_leading_term(aux_p,R);
      arpoly_free(aux_p,R);
      shifts_set_nth(res_ss,i,aux_t);
    } else {
      shifts_free(res_ss);
      olist_free(ol);
      math_error=cocoa_error("BAD_PARAMS");
      return shifts_NULL;
    }
    olist_set_nth(ol,i,object_NULL);
  }
  olist_free(ol);
  return res_ss;
}

olist this_iv2olist(ivector v)
{
  olist res_ol;
  int i,n;
  n=iv_get_len(v);
  res_ol=olist_init(n);
  for (i=1;i<=n;i++)
    olist_set_nth(res_ol,i,int2object(iv_get_nth(v,i)));
  iv_free(v);
  return res_ol;
}

olist this_ctbuf2olist(ctbuf b)
{
  int arity,i;
  olist res_ol;
  object aux_o;
  arity=ctbuf_len(b);
  res_ol=olist_init(arity);
  for (i=1;i<=arity;i++) {
    aux_o=cterm2object(ctbuf_lpop(b));
    olist_set_nth(res_ol,i,aux_o); }
  ctbuf_free(b);
  return res_ol;
}

olist indet_ident2olist(char * ident,ring R)
{ /* returns the list of all the indeterminates with that name */
  /* if there are no indets with that name it returns olist_NULL */
  int i,n;
  indet_entry aux_e;
  olist res_ol;
  i = indet_info_search_entry_index(R->indets,ident);
  if (i == -1) return olist_NULL;
  aux_e=R->indets->indets[i];
  n=aux_e->size;
  res_ol=olist_init(n);
  for (i=1; i<=n; i++)
    olist_set_nth(res_ol,i,
      arpoly2object(arpoly_unknown2arpoly((unknown)(aux_e->offset+i-1),R),R));
  return res_ol;
}
