/* $Id: omat.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "engine.h"
#include "ivectors.h"
#include "jintf.h"
#include "module.h"
#include "math_lib.h"
#include "operator.h"
#include "olist.h"
#include "omat.h"
#include "ordering.h"
#include "ring.h"

extern object math_error;

typedef struct subomat_aux * subomat;

typedef struct subomat_aux 
{
  omat elems;
  int m,n; /* num of rows, num of columns (of the submatrix) */
  ivector rr; /* rows indices */
  ivector cc; /* columns indices */
} subomat_info;

#define subomat_elems(M) ((M)->elems)
#define subomat_row(M,i) iv_get_nth(M->rr,i)
#define subomat_col(M,j) iv_get_nth(M->cc,j)
#define subomat_ij(M,i,j) \
  omat_ij(subomat_elems(M),subomat_row(M,i),subomat_col(M,j))
#define subomat_num_rows(M) ((M)->m)
#define subomat_num_cols(M) ((M)->n)
#define subomat_rows(M) ((M)->rr)
#define subomat_cols(M) ((M)->cc)

#define subomat_set_elems(M,E) (M)->elems=(E)
#define subomat_set_row(M,i,r) iv_set_nth(M->rr,i,r)
#define subomat_set_col(M,j,c) iv_set_nth(M->cc,j,c)
#define subomat_set_num_rows(M,x) (M)->m=(x)
#define subomat_set_num_cols(M,x) (M)->n=(x)
#define subomat_set_rows(M,RR) (M)->rr=(RR)
#define subomat_set_cols(M,CC) (M)->cc=(CC)

#define malloc_subomat() (subomat)mymalloc(sizeof(subomat_info),"subomat")
#define free_subomat(p) myfree(sizeof(subomat_info),p,"subomat")

/* subomat */

omat subomat2omat(subomat S);
subomat omat2subomat(omat M);
subomat subomat_mk(omat M, int m, int n, ivector rr, ivector cc);
void subomat_free(subomat S);
subomat subomat_elim_row(subomat S, int row);
subomat subomat_elim_col(subomat S, int col);
subomat subomat_elim_rc(subomat S,int row,int col);

object subomat_det(subomat S);

omat omat_init(int m, int n)
{
  omat res_m;
  int i;
  res_m=(omat)olist_init(m);
  for (i=1;i<=m;i++)
    olist_set_nth(res_m,i,olist2object(olist_init(n)));
  return res_m;
}

omat omat_new(int m, int n, object o)
{
  omat res_m;
  int i;
  res_m=(omat)olist_init(m);
  for (i=1;i<=m;i++)
    olist_set_nth(res_m,i,olist2object(olist_new(n,o)));
  return res_m;
}

coc_bool omat_is_rational(omat M)
{ /* returns true iff all the elements are integers or rationals */
  int i,j,n,m;
  m=omat_nrow(M);
  n=omat_ncol(M);
  for (i=1;i<=m;i++)
    for (j=1;j<=n;j++) {
      if (!objtype_le(object_type(omat_ij(M,i,j)),objtype_arbrat)) return FALSE;
    }
  return TRUE;
}

coc_bool omat_is_zero(omat M)
{ /* returns true iff all the elements are integers or rationals */
  int i,j,n,m;
  m=omat_nrow(M);
  n=omat_ncol(M);
  for (i=1;i<=m;i++)
    for (j=1;j<=n;j++) {
      if (!object_is_zero(omat_ij(M,i,j))) return FALSE;
    }
  return TRUE;
}

omat ord2omat(ordering ord, ring R)
{
  omat res_M;
  int n;
  register int i,j;
  n=R->indetsNo;
  res_M=omat_init(n,n);
  for (i=1;i<=n;i++)
    for (j=1;j<=n;j++)
      omat_set_ij(res_M,i,j,int2object(ord[i-1][j-1]));
  return res_M;

}

void omat2ord(ordering * to, omat M)
{
  int n, i, j, aux_i, counter, index;

  n=omat_nrow(M); /* must be = omat_ncol(M) */
  for (i=0; i <= n-1; i++)
  {
    counter = 0;
    for (j=0; j <= n-1; j++)
    {
      aux_i = object_cast_int(object_link(omat_ij(M, i+1, j+1)));
      if (math_error != object_NULL) return;
      if (aux_i < SHRT_MIN || aux_i > SHRT_MAX)
      {
        math_error = cocoa_error("LARGE_ORD_MAT_ENTRY");
        return;
      }
      if (aux_i!=0)
      {
        counter++; 
        index=j;
        (*to)[i][j]=(toElem)aux_i;
      }
    }
    if (counter == 1) (*to)[i][n]=(toElem)index;
  }
  if (is_term_ordering(*to, n)) return;
  ca_warning_string("Non-term-ordering specified\n-- GBasis-related computations could fail to terminate");
  engine_made_output = TRUE;
  /*  In versions 3.7 to 4.0 non term orderings were errors... */
  /*  math_error = cocoa_error("NOT_TERM_ORDERING"); */
}

/* The function below is a copy of olist2cterm with cterm_noindent inserted */
cterm omat2cterm_aux(omat m)
{
  int i,n;
  cterm res_t;
  cterm_list params;
  olist ol = (olist)m;
  n=olist_len(ol);
  if (n == 0) return cterm_mk_const(op_LIST);
  params=malloc_cterm_list(n);
  for (i=1;i<=n;i++) params[i-1]=cterm_noindent(object2cterm(olist_nth(ol,i)));
  res_t=cterm_mk_classic(op_LIST,n,params);
  free_cterm_list(n,params);
  return res_t;
}

cterm omat2cterm(omat m)
{
  cterm params[1];
  params[0]=omat2cterm_aux(m);
  return cterm_mk_classic(op_MAT,1,params);
}

/* ------------------------------------ */

object omat_det(omat M)
{
  subomat aux_S;
  object det;

  if (! omat_is_square(M)) return object_NULL;
  det = try_Zdet(M);
  if (det != object_NULL) return det;
  object_unlink(math_error);
  math_error = object_NULL;
  aux_S = omat2subomat(M);
  det = subomat_det(aux_S);
  subomat_free(aux_S);
  return det;
}

omat omat_adjoint(omat M)
{
  omat res_M;
  object el;
  int i,j,n;
  subomat aux_S,S;
  if (! omat_is_square(M)) return omat_NULL;
  n=omat_nrow(M);
  res_M=omat_init(n,n);
  if (n==1) /* the general code is broken for n=1; special case anyway */
  {
    omat_set_ij(res_M, 1, 1, int2object(1));
    return res_M;
  }
  S=omat2subomat(M);
  for (i=1;i<=n;i++)
    for (j=1;j<=n;j++) {
      aux_S=subomat_elim_rc(S,j,i);
      el=subomat_det(aux_S);
      subomat_free(aux_S);
      if ((i+j) & 1) el=do_change_sign(el);
      omat_set_ij(res_M,i,j,el); }
  subomat_free(S);
  return res_M;
}

void printf_omat(omat M,ring R)
{
  cterm aux_t;
  aux_t=omat2cterm(M);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

subomat subomat_mk(omat M, int m, int n, ivector rr, ivector cc)
{
  subomat res_S;
  res_S=malloc_subomat();
  subomat_set_elems(res_S,M);
  subomat_set_num_rows(res_S,m);
  subomat_set_num_cols(res_S,n);
  subomat_set_rows(res_S,rr);
  subomat_set_cols(res_S,cc);
  return res_S;  
}

subomat omat2subomat(omat M)
{
  ivector rr,cc;
  register int i;
  int m,n;
  m=omat_nrow(M);
  n=omat_ncol(M);
  rr=iv_init(m);
  cc=iv_init(n);
  for (i=1;i<=m;i++) iv_set_nth(rr,i,i);
  for (i=1;i<=n;i++) iv_set_nth(cc,i,i);
  return subomat_mk(M,m,n,rr,cc);
}

void subomat_free(subomat S)
{
  iv_free(subomat_rows(S));
  iv_free(subomat_cols(S));
  free_subomat(S);
}

omat subomat2omat(subomat S)
{
  omat res_M;
  int m,n,i,j;
  m=subomat_num_rows(S);
  n=subomat_num_cols(S);
  res_M=omat_init(m,n);
  for (i=1;i<=m;i++)
    for (j=1;j<=n;j++)
      omat_set_ij(res_M,i,j,subomat_ij(S,i,j));
  return res_M;
}

subomat subomat_elim_row(subomat S, int row)
{
  ivector rr,cc;
  rr=iv_elim_nth(subomat_rows(S),row);
  cc=iv_dup(subomat_cols(S));
  return 
    subomat_mk(subomat_elems(S),subomat_num_rows(S)-1,
      subomat_num_cols(S),rr,cc);
}

subomat subomat_elim_col(subomat S, int col)
{
  ivector rr,cc;
  cc=iv_elim_nth(subomat_cols(S),col);
  rr=iv_dup(subomat_rows(S));
  return 
    subomat_mk(subomat_elems(S),subomat_num_rows(S),subomat_num_cols(S)-1,
      rr,cc);
}

subomat subomat_elim_rc(subomat S,int row,int col)
{
  subomat aux_S,res_S;
  aux_S=subomat_elim_col(S,col);
  res_S=subomat_elim_row(aux_S,row);
  subomat_free(aux_S);
  return res_S;
}

object subomat_det(subomat S)
{
  object aux_o,aux_o1,res_el;
  int n,j;
  subomat aux_S;
  n=subomat_num_cols(S);
  if (n==1) return object_link(subomat_ij(S,1,1));
  res_el=object_NULL;
  for (j=1;j<=n;j++) { /* for now work on last row; TODO search optimun */
    aux_S=subomat_elim_rc(S,n,j);
    if (object_is_zero(subomat_ij(S,n,j)) && res_el==object_NULL)
      res_el=object_link(subomat_ij(S,n,j));
    else if (!object_is_zero(subomat_ij(S,n,j))) {
      object det;
      det=subomat_det(aux_S);
      if (n+j & 1) {
        aux_o=do_change_sign(det); /* TODO: is inefficient */
        object_unlink(det);
        det=aux_o; }
      if (res_el!=object_NULL) {
        aux_o1=do_mult(det,subomat_ij(S,n,j));
        aux_o=do_sum(res_el,aux_o1);
        object_unlink(aux_o1);
        object_unlink(res_el);
        object_unlink(det);
        res_el=aux_o; }
      else {
        res_el=do_mult(det,subomat_ij(S,n,j));
        object_unlink(det); } }
    subomat_free(aux_S); }
  return res_el;
}

omat omat_mult(omat o1, omat o2)
{
  omat res_om;
  ring R;
  objtype t;
  object aux_o,aux_o1,aux_o2;
  int m,n,p,i,j,k;
  if (o1==omat_NULL || o2==omat_NULL) return omat_NULL;
  if (omat_ncol(o1) != omat_nrow(o2)) return omat_NULL;
  m=omat_nrow(o1);
  n=omat_ncol(o1);
  p=omat_ncol(o2);
  res_om=omat_init(m,p);
  aux_o=omat_ij(o1,1,1);
  t=object_type(aux_o);
  R=aux_o->R;
  for (i=1;i<=m;i++)
    for (j=1;j<=p;j++) {
      /* aux_o=object_zero(t,R); */
      aux_o=object_link(object_ZERO);
      for (k=1;k<=n;k++) {
        aux_o1=do_mult(omat_ij(o1,i,k),omat_ij(o2,k,j));
        if object_is_error(aux_o1) {
          object_unlink(aux_o);
          math_error=aux_o1;
          omat_free(res_om);
          return omat_NULL; }
        aux_o2=do_sum(aux_o,aux_o1);
        if object_is_error(aux_o2) {
          object_unlink(aux_o);
          object_unlink(aux_o1);
          math_error=aux_o2;
          omat_free(res_om);
          return omat_NULL; }
        object_unlink(aux_o);
        object_unlink(aux_o1);
        aux_o=aux_o2; }
      omat_set_ij(res_om,i,j,aux_o);}
  return res_om;
}

omat omat_id(int n)
{
  omat res_M;
  int i;
  res_M=omat_new(n,n,object_link(object_ZERO));
  for (i=1;i<=n;i++) {
    object_unlink(omat_ij(res_M,i,i));
    omat_set_ij(res_M,i,i,object_link(object_ONE)); }
  return res_M;
}

omat omat_inv(omat M)
{ /* must be a square matrix */
  object aux_o;
  omat res_M;
  if (omat_is_rational(M)) return Qmat_inverse(M);
  aux_o=omat_det(M);
  if (object_is_zero(aux_o)) {
    error_set_this_param(string2cterm("(det = 0)"));
    math_error=cocoa_error("CANNOT_INV");
    return omat_NULL;
  }
  do_inv(&aux_o);
  if (object_is_error(aux_o)) {
    math_error=aux_o;
    return omat_NULL;
  }
  res_M=omat_extprod(omat_adjoint(M),aux_o);
  object_unlink(aux_o);
  return res_M;
}

omat omat_pow(omat M, int n)
{
  omat aux_M,res_M;
  int i;
  if (! omat_is_square(M)) return omat_NULL;
  res_M = omat_id(omat_ncol(M));
  for (i=1;i<=n;i++) {
    aux_M = omat_mult(res_M,M);
    omat_free(res_M);
    if (math_error != object_NULL) return omat_NULL;
    res_M=aux_M;
  }
  return res_M;
}

olist omat_minors(omat M, int p /* order */ )
{
  int m,n,N,i,c;
  olist res_l;
  ivector rv,cv;
  subomat aux_S;

  m=omat_nrow(M);
  n=omat_ncol(M);

  if (p>MIN(n,m)) {
    math_error=cocoa_error("BAD_PARAMS");
    return olist_NULL;
  }

  N=int_binom(m,p)*int_binom(n,p);

  res_l=olist_init(N);

  rv=iv_init(p);
  cv=iv_init(p);

  for (i=1;i<=p;i++) iv_set_nth(rv,i,i);
  c=1;
  do {
    for (i=1;i<=p;i++) iv_set_nth(cv,i,i);
    do {
      aux_S=subomat_mk(M,p,p,rv,cv);
      olist_set_nth(res_l,c++,subomat_det(aux_S));
      free_subomat(aux_S); }
    while (iv_next_comb(cv,n)); }
  while (iv_next_comb(rv,m));
  iv_free(rv);
  iv_free(cv);
  return res_l;
}

omat this_olist2omat(olist ol)
{
  olist res_ol;
  res_ol=this_olist_of_olists2omat(olist2olist_of_olists(ol));
  olist_free(ol);
  return res_ol;
}

omat this_olist_of_olists2omat(olist ol)
{
  int nr,nc,i,j;
  olist aux_ol,res_ol,res_aux_ol;
  object aux_o, o_zero;
  o_zero=object_zero(objtype_arbint,ring_NULL);
  nr = olist_len(ol);
  nc = 0;
  for (i=1;i<=nr;i++) {
    aux_o=olist_nth(ol,i);
    aux_ol=aux_o->v.olist_value;
    nc = MAX(nc,olist_len(aux_ol));
  }
  res_ol=olist_init(nr);
  for (i=1;i<=nr;i++) {
    res_aux_ol=olist_new(nc,o_zero);
    aux_o=olist_nth(ol,i);
    aux_ol=aux_o->v.olist_value;
    for (j=1;j<=olist_len(aux_ol);j++) {
      object_unlink(olist_nth(res_aux_ol,j));
      olist_set_nth(res_aux_ol,j,object_link(olist_nth(aux_ol,j)));
    }
    aux_o=olist2object(res_aux_ol);
    olist_set_nth(res_ol,i,aux_o);
  }
  olist_free(ol);
  object_unlink(o_zero);
  return res_ol;
}
