/* $Id: operator.c,v 4.5 2004/07/27 10:57:44 cocoa Exp $ */

#include <stdlib.h>
#include "cocdefs.h"
#include "ca.h"
#include "operator.h"
#include "cterm.h"
#include "map.h"
#include "cocfun.h"

coop op_NULL;

package packages[MAX_NUM_PACKAGES];
int pkg_std_n, pkg_true_builtin_n, pkg_builtin_n, pkg_user_n;

map pkg_map;

op_info op_infos[MAX_NUM_OPERATORS];

int coop_pkg_std_op(char * op_id, int * pkg_n, coop * op)
{
  *op=op_map_coop(pkg_user_n,op_id);
  if (*op==op_NULL) {
    *op=op_map_coop(pkg_std_n,op_id);
    if (*op==op_NULL) {
      *op=op_map_coop(pkg_builtin_n,op_id);
      if (*op==op_NULL) *pkg_n=pkg_n_NULL;
      else *pkg_n=pkg_builtin_n; }
    else *pkg_n=pkg_std_n; }
  else *pkg_n=pkg_user_n;
  return 0;
}

void coop_init(void)
{
  static coc_bool already_called=FALSE;
  if (already_called) return;
  already_called=TRUE;
  pkg_map=map_new();
  pkg_true_builtin_n=pkg_builtin_n=pkg_init(DOLLAR_BUILTIN);
  pkg_std_n=pkg_init("$coclib");
  pkg_user_n=pkg_init("$user");
  pkg_set_nth(0,pkg_nth(pkg_builtin_n)); /* TRICK */
  pkg_builtin_n=0; /* TRICK */

  op_NULL = 0;
}

void pkg_clear(int pkg_n)
{
  /* TODO */
}

int pkg_index(pkg_id id)
{
  int pkg_n;
  pkg_n = map_get_name_id(pkg_map,id);
  if (pkg_n == pkg_true_builtin_n) return pkg_builtin_n;
  return pkg_n;
}

int pkg_init(pkg_id id)
{
  int i,n;
  package pkg;
  map m;
  if (pkg_is_def(id)) return pkg_index(id);
  pkg=malloc_pkg();
  m=map_new();
  pkg_set_op_map(pkg,m);
  if (m == map_NULL) {
    ca_internal_error_string("INTERNAL ERROR in function op_map_init: CANNOT CREATE op_map\n");
    abort(); }
  for (i=0;i<MAX_NUM_OPERATORS;i++) 
    pkg_set_op_info(pkg,i,malloc_op_info());
  /* TODO: do allocation only when it is necessary */

  n=map_add(&pkg_map,id);
  pkg_set_nth(n,pkg);
  return n;
}

coop op_map_add_if_not_exists(int pkg_n, nome n)
{
  coop res_op;
  if ((res_op=op_map_coop(pkg_n,n)) == op_NULL) {
    return op_map_add(pkg_n,n,NO_ARITY);
  }
  return res_op;
}

cocfun op_map_cocfun(char * p_id, char * fun_id)
{
  coop aux_op;
  coc_bool b;
  int pkg_n, n;
  cterm h;
  pkg_n = pkg_index(p_id);
  aux_op = op_map_coop(pkg_n,fun_id);
  if (aux_op == op_NULL) return cocfun_NULL;
  aux_op=op_index(aux_op);
  h = op_map_op_handler(pkg_n, aux_op);
  n = op_map_op_arity(pkg_n, aux_op);
  if (h == cterm_NULL) 
    return cocfun_mk(p_id, fun_id, n, cterm_NULL, cterm_NULL, FALSE);
  b=(cterm_op(cterm_param(h,2))==op_TRUE);
  return cocfun_mk(p_id, fun_id, 0, cterm_param(h,0), cterm_param(h,1),b);
}

coop op_map_add(int pkg_n, nome n,int arity)
{ 
  coop op;
  if ((op = map_add(&pkg_op_map(pkg_n),n)) == name_id_NULL) {
#ifdef DEBUG
    printf("internal error: coops map is FULL\n");
#endif
    return op_NULL; }
  else {
    op_map_set_op_arity(pkg_n,op,arity);
    op_map_set_op_handler(pkg_n,op,cterm_NULL);
    op_map_set_op_help(pkg_n,op,NULL);
    return op; }
}

coop op_map_add_handler(int pkg_n, nome n,int arity,cterm handler)
{ 
  coop op;
  if ((op = map_add(&pkg_op_map(pkg_n),n)) == name_id_NULL) {
#ifdef DEBUG
    cocoa_internal_error("coops map is FULL");
#endif
    return op_NULL; }
  else {
    op_map_set_op_arity(pkg_n,op,arity);
    op_map_set_op_handler(pkg_n,op,handler);
    op_map_set_op_help(pkg_n,op,NULL);
    return op; }
}

coop op_map_add_user_function(int pkg_n, nome n, int arity, 
  cterm handler, char * help)
{
  coop op;
  if ((op = map_add(&pkg_op_map(pkg_n),n)) == name_id_NULL) 
    return op_NULL;
  else {
    op_map_set_op_arity(pkg_n,op,arity);
    op_map_set_op_handler(pkg_n,op,handler);
    op_map_set_op_help(pkg_n,op,my_strdup(help));
    return op; }
}

coop op_map_mod_user_function(int pkg_n, coop op,int arity,
  cterm handler, char * help)
{
  op=op_index(op);
  cterm_free((cterm)op_map_op_handler(pkg_n,op));
  if (op_map_op_help(pkg_n,op) != NULL) free_str1(op_map_op_help(pkg_n,op));
  op_map_set_op_arity(pkg_n,op,arity);
  op_map_set_op_handler(pkg_n,op,handler);
  op_map_set_op_help(pkg_n,op,my_strdup(help));
  return op;
}

coop op_map_coop(int pkg_n, nome n)
{
  coop op;
  if (pkg_n == pkg_n_NULL) return op_NULL;
  if ((op = map_get_name_id(pkg_op_map(pkg_n),n)) == name_id_NULL)
    return op_NULL;
  else 
    return op_mk(pkg_n,op);
}
