/* $Id: options.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "cterm.h"
#include "ctbuf.h"
#include "options.h"
#include "operator.h"
#include "utils.h"

/* in the first entry of a panel we put the informations about the panel */

opt_panel opt_panel_new(char * descr)
{
  opt_panel p;
  p = malloc_panel();
  opt_panel_set_descr(p,descr);
  opt_panel_set_dim(p,0); /* number of options */
  opt_set_type(p,0,opt_type_PANEL);
  return p;
}

opt_id panels_env_get_oid(opt_panel pan, char * descr)
{
  opt_id oid;
  for (oid=1;oid<=opt_panel_dim(pan);oid++)
    if (strcmp(opt_descr(pan,oid),descr) == 0) return oid;
  return opt_id_NULL;
}

opt_id opt_new(opt_type t, opt_panel p, char * descr)
{
  opt_id o;
  o = opt_panel_dim(p)+1;
  if (o > opt_MAX_NUM) return opt_id_NULL;
  opt_panel_set_dim(p,o);
  opt_set_descr(p,o,descr);
  opt_set_type(p,o,t);
  opt_set_value(p,o,optv_NULL);
  if (t==opt_type_ENUM) optv_set_descrs(p,o,malloc_optv_descr());
  /*
        p [o ].vd=
         (sizeof(charstars)*10   <=100 ) ?
         (charstars) rum_malloc(sizeof(charstars)*10 ) :
         (charstars)malloc(sizeof(charstars)*10   )  ;
  */
  return o;
}

optv_id optv_new(opt_panel p, opt_id o, char * descr)
{
  int n;
  if (opt_get_type(p,o)<opt_type_ENUM) return optv_NULL;
  n=opt_get_type(p,o);
  if (n==opt_MAX_NUM_VALUES) return optv_NULL;
  opt_set_type(p,o,n+1); /* enumerative with one more element */ 
  optv_set_descr(p,o,n+1,descr);
  return opt_get_type(p,o);
}

void opt_set_value_check(opt_panel p, opt_id o, opt_value v)
{
  if (o>opt_panel_dim(p)) return;
  if (o<=0) return;
  if (opt_get_type(p,o)!=opt_type_NUM) {
    if (opt_get_type(p,o)==opt_type_FLAG) { if (v!=0) v=1; }
    else if (v>opt_get_type(p,o) || v<=0) return; }
  opt_set_value(p,o,v);
}

void printf_opt_panel(opt_panel p)
{
  cterm aux_t;
  aux_t=opt_panel2cterm(p);
  ca_put(aux_t, cdev_out);
  cterm_free(aux_t);
}

cterm opt_panel2cterm(opt_panel p)
{
  /* panel=panel(descr:NBS,o1:option,...,on:option) */
  int n;
  opt_id o;
  ctbuf aux_b;
  n=opt_panel_dim(p);
  aux_b=ctbuf_new(cterm_mk_NBS(opt_panel_descr(p)));
  for (o=1;o<=n;o++)
    ctbuf_append(aux_b,opt2cterm(p,o));
  return ctbuf_mk_cterm(aux_b,op_PANEL);
}

cterm opt2cterm(opt_panel p, opt_id o)
{
  /* option=option(t:opt_type,d:opt_descr,v:opt_value) */
  /* option=option(t:opt_type,d:opt_descr,v:opt_value,vd:optv_descr) */
  /* opt_descr=NBS */
  /* opt_value=number */
  ctbuf aux_b;
  aux_b=ctbuf_new(opt_type2cterm(opt_get_type(p,o)));
  ctbuf_append(aux_b,cterm_mk_NBS(opt_descr(p,o)));
  ctbuf_append(aux_b,int2cterm(opt_get_value(p,o)));
  if (opt_get_type(p,o)>opt_type_ENUM) /* enumerative not empty */
    ctbuf_append(aux_b,optv_descr2cterm(p,o));
  return ctbuf_mk_cterm(aux_b,op_OPTION);
}

cterm opt_type2cterm(opt_type t)
{
  /* opt_type=flag */
  /* opt_type=num */
  /* opt_type=enum(n:number) */
  cterm params[1];
  if (t==opt_type_FLAG) return cterm_mk_const(op_FLAG);
  if (t==opt_type_NUM) return cterm_mk_const(op_NUM);
  params[0]=int2cterm(t);
  return cterm_mk_classic(op_ENUM,1,params);
}

cterm optv_descr2cterm(opt_panel p, opt_id o)
{
  /* optv_descr=list(d1:NBS,...,dn:NBS) */
  ctbuf aux_b;
  int v,n;
  n=opt_get_type(p,o); /* ASSUMPTION: n>0 */
  aux_b=ctbuf_empty();
  for (v=1;v<=n;v++)
    ctbuf_append(aux_b,cterm_mk_NBS(optv_get_descr(p,o,v)));
  return ctbuf_mk_cterm(aux_b,op_LIST);
}


/* ---[ options OPERATORS ]----------------------------------------------- */

coop op_PANEL;
coop op_OPTION;
coop op_NUM;
coop op_ENUM;
coop op_FLAG;

void options_init(void)
{
  static coc_bool already_called=FALSE;
  if (already_called) return;
  already_called=TRUE;
  cterm_initialize();
  op_PANEL = op_map_add(pkg_builtin_n,"\040PANEL",NO_ARITY);
  op_OPTION = op_map_add(pkg_builtin_n,"\040OPTION",NO_ARITY);
  op_FLAG = op_map_add(pkg_builtin_n,"\040FLAG",2);
  op_NUM = op_map_add(pkg_builtin_n,"\040NUM",0);
  op_ENUM = op_map_add(pkg_builtin_n,"\040ENUM",1);
}
