/* $Id: ordering.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"
/* the following is from the factor package */
#include "Zmat.h"

#ifndef IVECTORS_H
# include "ivectors.h"
#endif

#ifndef ORDERING_H
# include "ordering.h"
#endif

#ifndef MATRIX_H
# include "matrix.h"
#endif

/* -----------[ cord ]-------------- */

cord cord_init(int t, ordering o, unknown x, unknown y)
{
  cord c;
  c = malloc_cord();
  cord_set_type(c,t);
  cord_set_x(c,x);
  cord_set_y(c,y);
  cord_set_ordering(c,o);
  return c;
} 

/* -----------[ ordering ]-------------- */

ordering ord_dup(ordering ord,ring R)
{
  register short int i,j;
  ordering res_ord;
  if (ord==ordering_NULL) return ordering_NULL;
  res_ord = malloc_ordering( R->indetsNo );
  for ( i = 0; i < R->indetsNo; i++ ) {
    res_ord[i] = malloc_ord_row( R->indetsNo );
    for ( j = 0; j <= R->indetsNo; j++ ) 
      res_ord[i][j] = ord[i][j]; }
  return res_ord;
}

void InitZeroOrder( ordering * myOrd , ring R ) {
  register short int i, j;
  *myOrd = malloc_ordering( R->indetsNo );
  for ( i = 0; i < R->indetsNo; i++ ) {
    (* myOrd)[i] = malloc_ord_row( R->indetsNo );
    for ( j = 0; j < R->indetsNo; j++ ) 
      (* myOrd)[i][j] = 0;
    (* myOrd)[i][R->indetsNo] = -1;
  }
}

void ClearOrder(ordering * myOrd, ring R ) {
  register short int i;
  if ((*myOrd)==ordering_NULL) return;
  for ( i = 0; i < R->indetsNo; i++ ) 
    free_ord_row( R->indetsNo, (* myOrd)[i] );
  free_ordering( R->indetsNo, *myOrd );
}

void SetZeroOrder( ordering myOrd, ring R ) {
  /* assume che lo spazio sia gia' allocato */
  register short int i, j;
  for ( i = 0; i < R->indetsNo; i++ ) {
    for ( j = 0; j < R->indetsNo; j++ ) 
      myOrd [i][j] = 0;
    myOrd [i][R->indetsNo] = -1;
  }
}

void SetLexOrder( ordering myOrd, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  for ( i = 0; i < R->indetsNo; i++ ) {
    myOrd [i][i] = 1;
    myOrd [i][R->indetsNo] = (toElem) i;
  }
}

void SetXelOrder( ordering myOrd, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  for ( i = 0; i < R->indetsNo; i++ ) {
    myOrd [i][R->indetsNo-i-1] = 1;
    myOrd [i][R->indetsNo] = (toElem) (R->indetsNo-i-1);
  }
}

void SetDegLexOrder( ordering  myOrd, weight weights, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  myOrd[0][R->indetsNo] = -1; /* il primo vettore ha tutte le componenti <> 0 */

  for ( i = 0; i < R->indetsNo; i++ ) 
    myOrd[0][i] = 1;
  for ( i = 1; i < R->indetsNo; i++ )  {
    myOrd[i][i-1] =  1;
    myOrd[i][R->indetsNo] = (toElem)(i-1);
  }
}

void SetDegRevLexOrder ( ordering myOrd, weight weights, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  myOrd[0][R->indetsNo] = -1; /* il primo vettore ha tutte le componenti <> 0 */

  for ( i = 0; i < R->indetsNo; i++ ) 
    myOrd[0][i] = 1;
  for ( i = 1; i < R->indetsNo; i++ )  {
    myOrd[i][R->indetsNo - i] =  -1;
    myOrd[i][R->indetsNo] = R->indetsNo - i;
  }
}

void SetElimOrder ( unknown firstVar, unknown lastVar, 
     ordering myOrd, weight weis, ring R ) {
  short int i,j, h, numVarToElim, numVarOK;
  SetZeroOrder( myOrd, R );
  if ( firstVar < 0 )
    firstVar = 0;
  if ( lastVar >= R->indetsNo )
    lastVar = R->indetsNo-1;
  numVarToElim = lastVar - firstVar + 1;
  numVarOK = R->indetsNo - numVarToElim;

  if ( numVarToElim == R->indetsNo ) {
    SetDegRevLexOrder( myOrd, weis, R ) ;
    return;
  }
  
  if ( numVarToElim == 1 ) 
    myOrd[0][R->indetsNo] = firstVar;
  else
    myOrd[0][R->indetsNo] = -1;
  for( i = firstVar; i <= lastVar; i++ )
    myOrd[0][i] = 1;
    
  for( i = 0; i < firstVar; i++ ) myOrd[1][i] = 1;
  for( i = lastVar+1; i < R->indetsNo; i++ ) myOrd[1][i] = 1;
  myOrd[1][R->indetsNo] = -1;
  
  j = R->indetsNo-1; i = 2;
  for( h = numVarOK; h > 1; h--, j--, i++ )
    if ( j > lastVar ) {
      myOrd[i][j] = -1;
      myOrd[i][R->indetsNo] = j;
    } else {
      myOrd[i] [j - numVarToElim] = -1;
      myOrd[i] [R->indetsNo] = j - numVarToElim;  
    }
  j = lastVar;
  for( h = 1; h < numVarToElim ; h++, j--, i++ ) {
    myOrd[i][j] = -1;
    myOrd[i][R->indetsNo] = j;  
  }
}

void SetAntiDegLexOrder ( ordering myOrd, weight weights, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  myOrd[0][R->indetsNo] = -1;  

  for ( i = 0; i < R->indetsNo; i++ ) 
    myOrd[0][i] = 1;
  for ( i = 1; i < R->indetsNo; i++ )  {
    myOrd[i][i-1] =  -1;
    myOrd[i][R->indetsNo] = i-1;
  }
}

void SetAntiDegRevLexOrder ( ordering myOrd, weight weights, ring R ) {
  register short int i;
  SetZeroOrder( myOrd, R );
  myOrd[0][R->indetsNo] = -1;  

  for ( i = 0; i < R->indetsNo; i++ ) 
    myOrd[0][i] = 1;
  for ( i = 1; i < R->indetsNo; i++ )  {
    myOrd[i][R->indetsNo - i] =  1;
    myOrd[i][R->indetsNo] = R->indetsNo - i;
  }
}

void PrintOrder( ordering myOrd, ring R ) {
  register short int i, j;
  for ( i = 0; i < R->indetsNo; i++ ) {
    for ( j = 0; j <= R->indetsNo; j++ ) 
      printf ("%d, ", myOrd[i][j]);
    printf( "\n" );
  }
}

int iv_which_gt_other(ivector v1, ivector v2, ordering ord)
{
  register short int i;
  register iv_elem temp;
  ivector aux_v;
  register int l=iv_get_len(v1);

  aux_v=iv_init(l);
  if ( iv_eq(v1,v2) )
    return ANSWER_BOTH;
  for ( i = 1; i <= l; i++ )
    iv_put_nth(aux_v,i,iv_get_nth(v1,i) - iv_get_nth(v2,i));
  i = 0;
  while ( i < l ) {
    temp = term_scalar( i, aux_v, ord);
    if ( temp > 0 )
      return ANSWER_FIRST;
    else if (temp < 0 )
      return ANSWER_SECOND;
    else 
      i++;
  }
  return ANSWER_BOTH;
}



matrix ord2matrix(ordering ord,ring R)
{
  matrix res_M;
  int n;
  register int i,j;
  n=R->indetsNo;
  res_M=matrix_init(n,n);
  for (i=0;i<n;i++)
    for (j=0;j<n;j++)
      matrix_set_elem(res_M,i,j,ord[i][j]);
  return res_M;
}

ordering matrix2ord(matrix M,ring R)
{
  register int i,j;
  int c;
  ordering res_ord=ordering_NULL;
  InitZeroOrder(&res_ord,R);
  for (i=0;i<R->indetsNo;i++) {
    c = 0;
    for (j=0;j<R->indetsNo;j++)
      if (matrix_elem(M,i,j) !=0 ) {
        res_ord[i][R->indetsNo]=j;
        c++;
        res_ord[i][j]=matrix_elem(M,i,j); }
    if (c != 1) res_ord[i][R->indetsNo]=-1; }
  return res_ord;
}

ordering ord_adjoint(ordering ord,ring R)
{
  matrix M,Adj;
  ordering res_ord;
  int n;
  n=R->indetsNo;
  M=ord2matrix(ord,R);
  Adj=matrix_adjoint(M,n);
  matrix_free(M,n,n);
  res_ord=matrix2ord(Adj,R);
  matrix_free(Adj,n,n);
  return res_ord;
}

int ord_det(ordering ord,ring R)
{
  matrix M;
  int det;
  int n;
  n=R->indetsNo;
  M=ord2matrix(ord,R);
  det=matrix_det(M,n);
  matrix_free(M,n,n);
  return det;
}

coc_bool ord_is_deg_compatible(ordering ord, ring R)
{
  register int i;
  if (ord==ordering_NULL) 
    return (R->whatTO == dlex_num || R->whatTO == drev_num);
  for (i=0;i<R->indetsNo;i++)
    if (ord[0][i]==0) return FALSE;
  return TRUE;
}


coc_bool is_term_ordering(ordering to, int n)
{
  int i, j;
  Zmat M;
  mpz_t D;

  for (j=0; j < n; j++)
  {
    for (i=0; i < n; i++)
    {
      if (to[i][j] > 0) break;
      if (to[i][j] < 0) return FALSE;
    }
    if (j == n) return FALSE; /* only for a whole column of zeroes */
  }

  M = Zmat_ctor(n, n);
  mpz_init(D);
  for (i=0; i < n; i++)
    for (j=0; j < n; j++)
      mpz_set_si(M->entry[i][j], (int)to[i][j]);
  Zdet(D, M);
  i = mpz_sgn(D);
  Zmat_dtor(M);
  mpz_clear(D);
  if (i == 0) return FALSE;
  return TRUE;
}
