/* pair.c : 0.1 : 15 Feb 94 : ADT pair and list of pairs */

#include "cocdefs.h"

# include "ca.h"
# include "ck.h"
# include "ctbuf.h"
# include "cterm.h"
# include "eterms.h"
# include "evector.h"
# include "ivectors.h"
# include "monomial.h"
# include "operator.h"
# include "options.h"
# include "polys.h"
# include "pair.h"
# include "terms.h"

/* labels: VPAIR */

weilist mdeg_pair_weights;
ordering mdeg_pair_ord;

/* ---------------[ VPAIR ]---------------- */

cterm vpair2cterm(vpair ps, ring R)
{
  cterm aux_params[3];
  evector V1,V2;
  V1=vpair_first(ps);
  V2=vpair_second(ps);
  if (ev_is_labelled(V1)) {
    aux_params[0]=int2cterm(ev_label(V1));
    aux_params[1]=int2cterm(ev_label(V2));
  } else {
    aux_params[0]=ev2cterm_normal(V1,R);
    aux_params[1]=ev2cterm_normal(V2,R); }
  aux_params[2]=term2cterm(vpair_lcm(ps),R);
  return cterm_noindent(cterm_mk_classic(op_LIST,3,aux_params));
}

cterm vpairs2cterm(vpair ps, ring R)
{
  ctbuf aux_b;
  aux_b=ctbuf_empty();
  while (ps != vpair_NULL) {
    ctbuf_append(aux_b,vpair2cterm(ps,R));
    ps=vpair_next(ps); }
  return ctbuf_mk_cterm(aux_b,op_LIST);
}

vpair vpairs_get_first( vpair * ps )
{
  vpair res_p;
  res_p=*ps;
  *ps=vpair_next(*ps);
  vpair_set_next(res_p,NULL);
  return res_p;
}

vpair NewVPair( ring R ) {
  vpair myPair = malloc_vpair();
  myPair -> next = NULL;
  myPair -> coprimes = FALSE;
  vpair_set_lcm(myPair, NULL);
  vpair_set_A(myPair, NULL);
  vpair_set_B(myPair, NULL);
  return myPair;
}

void FreeVPair( vpair * myPair,ring R ) {
  if ( * myPair != NULL ) {
    ev_unlink(vpair_first(*myPair),R);
    ev_unlink(vpair_second(*myPair),R);
    term_free( vpair_lcm(*myPair) );
#ifdef GB_NEWCRIT
    term_free( vpair_A(*myPair) );
    term_free( vpair_B(*myPair) );
#endif
    free_vpair(*myPair);
    *myPair = NULL;
  }
}

void FreeVPairs( vpair * myPair,ring R ) {
  if ( * myPair != NULL ) {
    ev_unlink(vpair_first(*myPair),R);
    ev_unlink(vpair_second(*myPair),R);
    term_free( vpair_lcm(*myPair) );
#ifdef GB_NEWCRIT
    term_free( vpair_A(*myPair) );
    term_free( vpair_B(*myPair) );
#endif
    FreeVPairs( & (*myPair) -> next,R );
    free_vpair(*myPair);
    *myPair = NULL;
  }
}

/* WARNING for gb-stuff see file gb : key-function Do_Make_Pair() */
vpair vpair_mk( evector E1, evector E2, opt_panel opts, ring R)
{
  register int sDeg;
  vpair res_p;
  res_p=NewVPair(R);
  vpair_set_first(res_p,ev_link(E1));
  vpair_set_second(res_p,ev_link(E2));
  vpair_set_lcm(res_p,term_lcm(vpair_first_lt(res_p),vpair_second_lt(res_p)));
#ifdef GB_NEWCRIT
  vpair_set_A(res_p,term_div(vpair_lcm(res_p),vpair_first_lt(res_p)));
  vpair_set_B(res_p,term_div(vpair_lcm(res_p),vpair_second_lt(res_p)));
#endif
  /* TODO: eliminare se possibile */
  vpair_set_coprimes(res_p,
    term_coprime(vpair_first_lt(res_p),vpair_second_lt(res_p)) );
  if ( flag_is_on(opts,flag_GB_SUGAR) ) {
    sDeg = term_degree(vpair_lcm(res_p));
    if ( ev_td(E1) > ev_td(E2) ) {
      sDeg += ev_td(E1);
      vpair_set_td(res_p,ev_td(E1)); }
    else {
      sDeg += ev_td(E2);
      vpair_set_td(res_p,ev_td(E2)); }
    vpair_set_wd(res_p,sDeg); }
  res_p->next = vpair_NULL;
  return res_p;
}

vpair vpair_dup(vpair P, ring R)
{
  vpair res_P;
  res_P=NewVPair(R);
  vpair_set_first(res_P,ev_link(vpair_first(P)));
  vpair_set_second(res_P,ev_link(vpair_second(P)));
#ifdef GB_NEWCRIT
  vpair_set_A(res_P,term_dup(vpair_A(P)));
  vpair_set_B(res_P,term_dup(vpair_B(P)));
#endif
  vpair_set_lcm(res_P,term_dup(vpair_lcm(P)));
  vpair_set_coprimes(res_P,vpair_are_coprimes(P));
  vpair_set_td(res_P,vpair_td(P));
  vpair_set_wd(res_P,vpair_wd(P));
  vpair_set_next(res_P,vpair_NULL);
  return res_P;
}

vpair vpairs_dup(vpair Ps, ring R)
{
  vpair res_Ps,aux_Ps;
  if (Ps == vpair_NULL) return vpair_NULL;
  res_Ps = aux_Ps = vpair_dup(Ps,R);
  Ps = vpair_next(Ps);
  while (Ps != vpair_NULL) {
    vpair_set_next(aux_Ps,vpair_dup(Ps,R));
    aux_Ps = vpair_next(aux_Ps);
    Ps = vpair_next(Ps);
  }
  return res_Ps;
}

int vpairs_len(vpair ps)
{
  int res_n;
  res_n=0;
  while (ps != vpair_NULL) {
    res_n++;
    ps=vpair_next(ps); }
  return res_n;
}
