/* $Id: pipe.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include <stdio.h>
#include <stdlib.h>

#ifndef PIPE_H
# include "pipe.h"
#endif

/* PRIMITIVES OF ADT coc_pipe */

coc_pipe pipe_new(void)
{ 
  coc_pipe p;
  p = malloc_pipe();
  p->lm=NULL;
  p->rm=NULL;
  p->len=0;
  return p;
}

coc_pipe pipe_free(coc_pipe p)
{ 
  pipe_atom a;
  if (p->len!=0) {
    a=p->lm;
    while (a!=p->rm) {
      a=a->r;
      free_pipe_atom(a->l); }
    free_pipe_atom(a); }
  free_pipe(p);
  return NULL;
}

pipe_atom new_pipe_atom(pipe_elem e,pipe_atom l,pipe_atom r);

pipe_atom new_pipe_atom(pipe_elem e,pipe_atom l,pipe_atom r)
{
  pipe_atom a;
  a = malloc_pipe_atom();
  a->e=e;
  a->l=l;
  a->r=r;
  return a;
}


coc_pipe pipe_lpush(coc_pipe p,pipe_elem e)
{ 
  pipe_atom a;
  a = new_pipe_atom(e,NULL,p->lm);
  p->lm=a;
  if (!p->len) p->rm=a;
  else a->r->l=a;
  p->len++;
  return p;
}

coc_pipe pipe_rpush(coc_pipe p,pipe_elem e)
{ 
  pipe_atom a;
  a = new_pipe_atom(e,p->rm,NULL);
  p->rm=a;
  if (!p->len) p->lm=a;
  else a->l->r=a;
  p->len++;
  return p;
}

pipe_elem pipe_lpop(coc_pipe p)
{ 
  pipe_elem e;
  pipe_atom a;
  if (!p->len) 
    return pipe_elem_NULL;
  else {
    a=p->lm;
    e=a->e;
    p->lm=p->lm->r;
    if (p->lm) p->lm->l=NULL;
    else p->rm=NULL; /* the pipe_elem was the last to pop */
    p->len--;
    free_pipe_atom(a);
    return e;
  }
}

pipe_elem pipe_rpop(coc_pipe p)
{
  pipe_elem e;
  pipe_atom a;
  if (!p->len) 
    return pipe_elem_NULL;
  else {
    a=p->rm;
    e=a->e;
    p->rm=p->rm->l;
    if (p->rm) p->rm->r=NULL;
    else p->lm=NULL; /* the pipe_elem was the last to pop */
    p->len--;
    free_pipe_atom(a);
    return e;
  }
}

coc_pipe pipe_concat_these(coc_pipe p1, coc_pipe p2)
{
  if (pipe_len(p1)==0) {
    free_pipe(p1);
    return p2; }
  if (pipe_len(p2)==0) {
    free_pipe(p2);
    return p1; }
  p1->rm->r=p2->lm;
  p2->lm->l=p1->rm;
  p1->rm=p2->rm;
  p1->len+=p2->len;
  free_pipe(p2);
  return p1;
}

/* PREDICATES */


/* I/O FUNCTIONS */

#if 0

void fprintf_pipe_pipe_elem(FILE * f,pipe_elem e)
{
  char buffer[30];
  sprintf(buffer,"%d",*e); /* IMPLEMENTATION DEPENDENT pipe_elem=*int */
  my_fputs(buffer,f);
  /* TODO: REMOVE DEPENDENCE */
}

void pipe_fprintf(FILE * f,coc_pipe p)
{
  pipe_atom a;
  a=p->lm;
  while (a) {
    fprintf_pipe_pipe_elem(f,a->e);
    fputs(" ",f);
    a=a->r; }
}

#endif
