/* $Id: poincare.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef ANNA
#include "cocdefs.h"
#include "coeffs.h"
#include "monomial.h"
#include "olist.h"
#include "object.h"
#include "polys.h"
#include "pseries.h"
#include "ring.h"
#include "terms.h"
#include "weights.h"
#endif

#ifdef ANNA
#include "AnnaIO.h"
#endif
 
#include "eterms.h"
#include "ivectors.h"
#include "poincare.h"
#include "unipoly.h"

/**********************************************************/

ints MFI_Occurrences, MFI_Indets;
int Ints120[121];
MixedTermList  MTL120[121];
TermList GlobalSplitterTList;

/**********************************************************/

#define malloc_MTList(len) \
  (MixedTermList)mymalloc( (len+1)*sizeof(eterm), "MTList" )
#define free_MTList(len,p)     myfree((len+1)*sizeof(eterm),p,"MTList")
#define realloc_MTList(p,oldsz,newsz) \
  (MixedTermList)myrealloc(p,(oldsz+1)*sizeof(eterm),\
  (newsz+1)*sizeof(eterm),"MTList" )

void poincare_init(int IndetsNo)
{
  static initialized = 0;
  if ( !initialized )
  {
    initialized = 1;

#ifdef COCOA_WITH_RUM
    /* TODO rum.h 22-9-98: if ( rums[rum].len==0 ) */
#ifdef ANNA
    rum_init_all();
    rum_init(12,2000); /* pointers */
    rum_init(eterm_size(IndetsNo), RUM_STD_SIZE);
#endif /* ANNA */
    rum_init(200*sizeof(eterm), 120);
    rum_init(sizeof(TermListStruct), RUM_STD_SIZE);
#endif /* COCOA_WITH_RUM */

#ifdef ANNA
    unipoly_rum_init(2*IndetsNo);
    PowerList = MakePowerList(MIN(IndetsNo, PoincareMaxPower));
#else
    unipoly_rum_init(PoincareMaxPower);
    PowerList = MakePowerList(PoincareMaxPower);
#endif /* ANNA */

    MFI_Occurrences = ints_init(IndetsNo);
    MFI_Indets = ints_init(IndetsNo);
    GlobalSplitterTList = NewTList(IndetsNo, IndetsNo);
  }
}

void StartPoincare(int IndetsNo)
{
  if ( IntsGetLen(MFI_Occurrences)!=IndetsNo )
  {
    ints_free(MFI_Occurrences);
    ints_free(MFI_Indets);
    SetMTListLen(GlobalSplitterTList, 0);
    EraseAndFreeTList(GlobalSplitterTList);
    
    MFI_Occurrences = ints_init(IndetsNo);
    MFI_Indets = ints_init(IndetsNo);
    GlobalSplitterTList = NewTList(IndetsNo, IndetsNo);
    /* TODO rum.h 22-9-98: if ( rums[rum].len==0 ) */
#ifdef COCOA_WITH_RUM
    rum_init(eterm_size(IndetsNo), RUM_STD_SIZE);
#endif
    unipoly_rum_reset(2*IndetsNo);
  }
}

#define malloc_TList() (TermList)mymalloc(sizeof(TermListStruct),"TList")
#define free_TList(p) {myfree(sizeof(TermListStruct),p,"TList");}
/**********************************************************/

#ifndef ANNA
/**********************  MV operations *************************/

int eterm_scalar(eterm t1, eterm t2)
{
  ints OccInd = Indets(t1);
  int i = IntsGetLen(OccInd), sum =0;

  for ( ; i>0 ; i-- )
    sum += t1[OccInd[i]] * t2[OccInd[i]];

  return sum;
}

#define SP_eterm_scalar(index,exp,t) (exp)*((t)[index])

term eterm_MultiDeg(eterm t, TermList Weights, ring MVP_R)
{
  int n =MTListLen(Weights);
  term aux_t;
  MixedTermList W =MTList(Weights);

  aux_t = term_init(n);
  for ( ; n>0 ; n-- )
    term_put_nth(aux_t, n, eterm_scalar(t, W[n]) );

  return aux_t;
}

term SPMultiDeg(int index, int exp, TermList Weights, ring MVP_R)
{
  int n =MTListLen(Weights);
  term aux_t;
  MixedTermList W =MTList(Weights);

  aux_t = term_init(n);
  for ( ; n>0 ; n-- )
    term_put_nth(aux_t, n, (W[n])[index]*exp );

  return aux_t;
}

/********  MV Mult ********/

arpoly P1MinusXMultiExpP2(arpoly p1, term ME, arpoly p2, ring R)
{
  p1 = arpoly_this_minus_term_x_arpoly(p1,ME,p2,R);
  term_free(ME);
/*  arpoly_free(p2, R);*/
  return p1;
}

arpoly P1PlusXMultiExpP2(arpoly p1, term ME, arpoly p2, ring R)
{
  p1 = arpoly_this_plus_term_x_arpoly(p1,ME,p2,R);
  term_free(ME);
/*  arpoly_free(p2, R);*/
  return p1;
}

arpoly MultByOneMinusXMultiExp(arpoly p, term ME, ring MVP_R)
{
  monomial myMon;
  arpoly res;

  myMon = monomial_mk( coeff_one(MVP_R), ME);
  monomial_change_sign(myMon,MVP_R);
  res = arpoly_plus_mxp(p, myMon, p, MVP_R);
  monomial_free(myMon,MVP_R);  

  return res;
}

arpoly MVP1TimesP2(arpoly p1, arpoly p2, ring MVP_R)
{
  arpoly res;

  res = arpoly_mult(p1, p2, MVP_R);
  arpoly_free(p1,MVP_R);
  arpoly_free(p2,MVP_R);

  return res;
}

/********  TruncMV Mult ********/

arpoly TruncP1MinusXMultiExpP2(arpoly p1, term ME, arpoly p2, 
			       int Trunc, ring R)
{
  p1 = arpoly_this_minus_term_x_arpoly_truncated(p1,ME,p2,Trunc,R);
  term_free(ME);
/*  arpoly_free(p2, R);*/
  return p1;
}

arpoly TruncP1PlusXMultiExpP2(arpoly p1, term ME, arpoly p2,
			      int Trunc, ring R)
{
  p1 = arpoly_this_plus_term_x_arpoly_truncated(p1,ME,p2,Trunc,R);
  term_free(ME);
/*  arpoly_free(p2, R);*/
  return p1;
}

arpoly TruncMultByOneMinusXMultiExp(arpoly p, term ME, 
				    int Trunc, ring MVP_R)
{
  monomial myMon;
  arpoly res;

  myMon = monomial_mk( coeff_one(MVP_R), ME);
  monomial_change_sign(myMon,MVP_R);
  res = arpoly_plus_mxp(p, myMon, p, MVP_R);
  monomial_free(myMon,MVP_R);  

  return res;
}

arpoly TruncMVP1TimesP2(arpoly p1, arpoly p2, int Trunc, ring MVP_R)
{
  arpoly res;

  res = arpoly_mult(p1, p2, MVP_R);
  arpoly_free(p1,MVP_R);
  arpoly_free(p2,MVP_R);

  return res;
}
#endif

/****************************************************************
 *************************  BASE CASES  *************************
 ****************************************************************/

unipoly UnipolySPPoincare(TermList theTList)
{
  eterm   SPL = SPList(theTList);
  ints OccInd = Indets(SPL);
  int i = IntsGetLen(OccInd);
  unipoly res;

  if ( i==eterm_degree(SPL) )    res = PowerExtDup(i, i);
  else
  {
    res = UnipolyOne(eterm_degree(SPL));
    for ( ; i>0 ; i-- )   MultByOneMinusXExp(res, SPL[OccInd[i]]);
  }
  EraseAndFreeTList(theTList);
  return res;
}

unipoly UnipolyLenOnePoincare(TermList theTList)
{
  eterm    SPL = SPList(theTList), myTerm = (MTList(theTList))[1];
  ints     OccInd = Indets(SPL);
  int   TDeg =eterm_degree(myTerm), i =IntsGetLen(OccInd), exp;
  unipoly p1, p2;

  if ( eterm_coprime(myTerm, SPL) )
  {
    if ( i==eterm_degree(SPL) )   p1 = PowerExtDup(i, i+TDeg);
    else
    {
      p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
      for ( ; i>0 ; i-- )  MultByOneMinusXExp(p1, SPL[OccInd[i]]);
    }
    MultByOneMinusXExp(p1, TDeg);
    EraseAndFreeTList(theTList);
    return p1;
  }
  else 
  {
    p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
    p2 = UnipolyOne(eterm_degree(SPL));
    for ( ; i>0 ; i-- )
    {
      exp = SPL[OccInd[i]];
      MultByOneMinusXExp(p1, exp);
      MultByOneMinusXExp(p2, exp-myTerm[OccInd[i]]);
    }   
    EraseAndFreeTList(theTList);
    return P1MinusXExpP2(p1, TDeg, p2);
  }
}

unipoly UnipolyOneTermAndSPPoincare(eterm theTerm, eterm SPL)
{
  ints  OccInd = Indets(SPL);
  int   TDeg = eterm_degree(theTerm), i = IntsGetLen(OccInd), exp;
  unipoly p1, p2 ;

  if (eterm_coprime(theTerm, SPL))
  {
    if ( i==eterm_degree(SPL) )   p1 = PowerExtDup(i, i+TDeg);
    else
    {
      p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
      for ( ; i>0 ; i-- )  MultByOneMinusXExp(p1, SPL[OccInd[i]]);
    }
    MultByOneMinusXExp(p1, eterm_degree(theTerm));
    eterm_free(theTerm);
    eterm_free(SPL);
    return p1;
  }

  p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
  p2 = UnipolyOne(eterm_degree(SPL));
  for ( ; i>0 ; i-- )
  {
    exp = SPL[OccInd[i]];
    MultByOneMinusXExp(p1, exp);
    MultByOneMinusXExp(p2, exp-theTerm[OccInd[i]]);
  }
  eterm_free(theTerm);
  eterm_free(SPL);
  return P1MinusXExpP2(p1, TDeg, p2);
}

#ifdef ANNA
unipoly TruncUnipolySPPoincare(TermList theTList, int Trunc)
{
  eterm   SPL = SPList(theTList);
  ints    OccInd = Indets(SPL);
  int i = IntsGetLen(OccInd);
  unipoly res;

  if ( i==eterm_degree(SPL) )    res = PowerExtDup(i, i);
  else
  {
    res = UnipolyOne(eterm_degree(SPL));
    for ( ; i>0 ; i-- )  TruncMultByOneMinusXExp(res, SPL[OccInd[i]], Trunc);
  }
  EraseAndFreeTList(theTList);
  return res;
}

unipoly TruncUnipolyLenOnePoincare(TermList theTList, int Trunc)
{
  eterm   SPL =SPList(theTList), myTerm=(MTList(theTList))[1];
  ints    OccInd = Indets(SPL);
  int   TDeg =eterm_degree(myTerm), i = IntsGetLen(OccInd), exp;
  unipoly p1, p2;

  if (eterm_coprime(myTerm, SPL))
  {
    if ( i==eterm_degree(SPL) )
      p1 = PowerExtDup(i, i+TDeg);
    else
    {
      p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
      for ( ; i>0 ; i-- ) TruncMultByOneMinusXExp(p1, SPL[OccInd[i]], Trunc);
    }
    TruncMultByOneMinusXExp(p1, eterm_degree(myTerm), Trunc);
    EraseAndFreeTList(theTList);
    return p1;
  }
  else 
  {
    p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
    p2 = UnipolyOne(eterm_degree(SPL));
    for ( ; i>0 ; i-- )
    {
      exp = SPL[OccInd[i]];
      TruncMultByOneMinusXExp(p1, exp, Trunc);
      TruncMultByOneMinusXExp(p2, exp-myTerm[OccInd[i]], Trunc);
    }
    EraseAndFreeTList(theTList);
    return TruncP1MinusXExpP2(p1, TDeg, p2, Trunc);
  }
}

unipoly TruncUnipolyOneTermAndSPPoincare(eterm theTerm, eterm SPL, int Trunc)
{
  ints OccInd = Indets(SPL);
  int i = IntsGetLen(OccInd),    TDeg =eterm_degree(theTerm), exp;
  unipoly p1, p2 ;

  if (eterm_coprime(theTerm, SPL))
  {
    if ( i==eterm_degree(SPL) )
      p1 = PowerExtDup(i, i+TDeg);
    else
    {
      p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
      for ( ; i>0 ; i-- ) TruncMultByOneMinusXExp(p1, SPL[OccInd[i]], Trunc);
    }
    TruncMultByOneMinusXExp(p1, eterm_degree(theTerm), Trunc);
    eterm_free(theTerm);
    eterm_free(SPL);
    return p1;
  }

  p1 = UnipolyOne(eterm_degree(SPL)+TDeg);
  p2 = UnipolyOne(eterm_degree(SPL));
  for ( ; i>0 ; i-- )
  {
    exp = SPL[OccInd[i]];
    TruncMultByOneMinusXExp(p1, exp, Trunc);
    TruncMultByOneMinusXExp(p2, exp-theTerm[OccInd[i]], Trunc);
  }
  eterm_free(theTerm);
  eterm_free(SPL);
  return TruncP1MinusXExpP2(p1, TDeg, p2, Trunc);
}
#endif

#ifndef ANNA
/******************  MV  *******************/
/******************  TruncMV  *******************/

arpoly TruncMVSPPoincare(TermList theTList,  TermList Weights,
			 int Trunc,  ring MVP_R)
{
  eterm   SPL = SPList(theTList);
  ints OccInd = Indets(SPL);
  int   n = IntsGetLen(OccInd), index;
  arpoly res =arpoly_one(MVP_R);

  for ( ; n>0 ; n-- )
  {
    index = OccInd[n];
    res = TruncMultByOneMinusXMultiExp(res, SPMultiDeg(index, SPL[index],
						       Weights, MVP_R),
				       Trunc, MVP_R);
  }
  EraseAndFreeTList(theTList);
  return res;
}


arpoly TruncMVLenOnePoincare(TermList theTList, TermList Weights, 
			     int Trunc, ring  MVP_R)
{
  eterm   SPL =SPList(theTList), myTerm=(MTList(theTList))[1];
  term    MExp;
  ints OccInd = Indets(SPL);
  int  n = IntsGetLen(OccInd), idx, exp;
  arpoly p1 =arpoly_one(MVP_R),p2;

  if (eterm_coprime(myTerm, SPL))
  {
    for ( ; n>0 ; n-- )
    {
      idx =  (Indets(SPL))[n];
      p1 = TruncMultByOneMinusXMultiExp(p1, SPMultiDeg(idx, (exp=SPL[idx]),
						     Weights, MVP_R),
				      Trunc, MVP_R);
    }
    p1 = TruncMultByOneMinusXMultiExp(p1, 
				      eterm_MultiDeg(myTerm, Weights, MVP_R), 
				      Trunc, 
				      MVP_R);
    EraseAndFreeTList(theTList);
    return p1;
  }
  else 
  {
    p2 = arpoly_one(MVP_R);
    for ( ; n>0 ; n-- )
    {	
      idx = (Indets(SPL))[n];
      p1 = TruncMultByOneMinusXMultiExp(p1, 
					SPMultiDeg(idx,  (exp=SPL[idx]),
						   Weights,  MVP_R),
					Trunc, 
					MVP_R);
      p2 = TruncMultByOneMinusXMultiExp(p2, 
					SPMultiDeg(idx,  exp-myTerm[idx],
						   Weights,  MVP_R),
					Trunc, 
					MVP_R);
    }
    MExp = eterm_MultiDeg(myTerm, Weights, MVP_R);
    EraseAndFreeTList(theTList);
    return TruncP1MinusXMultiExpP2(p1, MExp, p2, Trunc, MVP_R);
  }
}

arpoly TruncMVOneTermAndSPPoincare(eterm theTerm, eterm SPL, 
				   TermList Weights, int Trunc, ring MVP_R)
{
  ints OccInd = Indets(SPL);
  int n = IntsGetLen(OccInd), idx, exp;
  term  MExp;
  arpoly p1 =arpoly_one(MVP_R),p2;
  
  if (eterm_coprime(theTerm, SPL))
  {
    for ( ; n>0 ; n-- )
    {
      idx = (Indets(SPL))[n];
      p1 = TruncMultByOneMinusXMultiExp(p1, 
					SPMultiDeg(idx,  (exp =SPL[idx]),
						   Weights,  MVP_R),
					Trunc, 
					MVP_R);
    }
    p1 = TruncMultByOneMinusXMultiExp(p1, 
				      eterm_MultiDeg(theTerm,Weights,MVP_R), 
				      Trunc, 
				      MVP_R);
    eterm_free(theTerm);
    eterm_free(SPL);
    return p1;
  }
  else 
  {
    p2 = arpoly_one(MVP_R);
    for ( ; n>0 ; n-- )
    {
      idx = (Indets(SPL))[n];
      p1 = TruncMultByOneMinusXMultiExp(p1, 
					SPMultiDeg(idx, (exp =SPL[idx]),
						   Weights, MVP_R),
					Trunc, MVP_R);
      p2 = TruncMultByOneMinusXMultiExp(p2, 
					SPMultiDeg(idx, exp-theTerm[idx],
						   Weights, MVP_R),
					Trunc, MVP_R);
    }
    MExp =eterm_MultiDeg(theTerm, Weights, MVP_R);
    eterm_free(theTerm);
    eterm_free(SPL);
    return TruncP1MinusXMultiExpP2(p1, MExp, p2, Trunc, MVP_R);
  }
}

/******************  TruncMV end  ******************/

arpoly MVSPPoincare(TermList theTList, TermList Weights, ring MVP_R)
{
  eterm   SPL = SPList(theTList);
  ints    OccInd = Indets(SPL);
  int   n = IntsGetLen(OccInd), index;
  arpoly res =arpoly_one(MVP_R);

  for ( ; n>0 ; n-- )
  {
    index =  (Indets(SPL))[n];
    res = MultByOneMinusXMultiExp(res, 
				  SPMultiDeg(index, SPL[index], Weights,MVP_R),
				  MVP_R);
  }
  EraseAndFreeTList(theTList);
  return res;
}

arpoly MVLenOnePoincare(TermList theTList, TermList Weights, ring MVP_R)
{
  eterm   SPL =SPList(theTList), myTerm=(MTList(theTList))[1];
  term    MExp =eterm_MultiDeg(myTerm, Weights, MVP_R);
  ints    OccInd = Indets(SPL);
  int   n = IntsGetLen(OccInd), index, exp;
  arpoly p1 =arpoly_one(MVP_R),p2;

  if (eterm_coprime(myTerm, SPL))
  {
    for ( ; n>0 ; n-- )
    {
      index = (Indets(SPL))[n];
      p1 = MultByOneMinusXMultiExp(p1, 
				 SPMultiDeg(index, (exp =SPL[index]),
					    Weights, MVP_R),
				 MVP_R);
    }
    p1 = MultByOneMinusXMultiExp(p1, 
				 eterm_MultiDeg(myTerm, Weights, MVP_R), 
				 MVP_R);
    EraseAndFreeTList(theTList);
    return p1;
  }
  else 
  {
    p2 =arpoly_one(MVP_R);
    for ( ; n>0 ; n-- )
    {	
      index = (Indets(SPL))[n];
      p1 = MultByOneMinusXMultiExp(p1, 
				   SPMultiDeg(index,
					      (exp =SPL[index]),
					      Weights,
					      MVP_R),
				   MVP_R);
      p2 = MultByOneMinusXMultiExp(p2, 
				   SPMultiDeg(index,
					      exp-myTerm[index],
					      Weights,
					      MVP_R),
				   MVP_R);
    }
    EraseAndFreeTList(theTList);
    return P1MinusXMultiExpP2(p1, MExp, p2, MVP_R);
  }
}

arpoly MVOneTermAndSPPoincare (eterm theTerm, eterm SPL, 
			       TermList Weights, ring MVP_R)
{
  ints    OccInd = Indets(SPL);
  int   n = IntsGetLen(OccInd), index, exp;
  term  MExp =eterm_MultiDeg(theTerm, Weights, MVP_R);
  arpoly p1 =arpoly_one(MVP_R),p2;
  
  if (eterm_coprime(theTerm, SPL))
  {
    for ( ; n>0 ; n-- )
    {
      index = (Indets(SPL))[n];
      p1 = MultByOneMinusXMultiExp(p1, 
				   SPMultiDeg(index,  (exp =SPL[index]),
					      Weights,  MVP_R),
				   MVP_R);
    }
    p1 = MultByOneMinusXMultiExp(p1, 
				 eterm_MultiDeg(theTerm,Weights,MVP_R), 
				 MVP_R);
    eterm_free(theTerm);
    eterm_free(SPL);
    return p1;
  }
  else 
  {
    p2 = arpoly_one(MVP_R);
    for ( ; n>0 ; n-- )
    {
      index = (Indets(SPL))[n];
      p1 = MultByOneMinusXMultiExp(p1, 
				   SPMultiDeg(index,  (exp =SPL[index]),
					      Weights,  MVP_R),
				   MVP_R);
      p2 = MultByOneMinusXMultiExp(p2, 
				   SPMultiDeg(index,  exp-theTerm[index],
					      Weights,  MVP_R),
				   MVP_R);
    }
    eterm_free(theTerm);
    eterm_free(SPL);
    return P1MinusXMultiExpP2(p1, MExp, p2, MVP_R);
  }
}
/******************  MV end  *******************/
#endif  /* ANNA */

/* TermLists.c */

#define Random(len)( 1 )

TermList NewTList(int len, int indetsNo)
{
  TermList  aux = malloc_TList();

  if ( len <= 199 ) len = 199;
  SetMTListSize (aux, len);
  SetMTListLen  (aux, 0);
  MTList(aux) = malloc_MTList(len);
  SPList(aux) = eterm_init(indetsNo);
  
  return aux;
}

void EraseAndFreeTList(TermList theTList)
/* all the terms must be still allocated!! */
{
  int i =MTListLen(theTList);
  MixedTermList MTL =MTList(theTList);
  
  for ( ; i>=0 ; i-- )  eterm_free(MTL[i]);
  free_MTList(MTListSize(theTList), MTL);
  free_TList(theTList);
}

TermList TListMake(eterm theSPL, MixedTermList theMTL, int Size, int Len)
{
  TermList aux;

  aux = malloc_TList();
  MTList(aux) = theMTL;
  SPList(aux) = theSPL;
  SetMTListSize(aux, Size);
  TListReduceSize(aux, Len);

  return aux;
}

void TListChangeSize (TermList theTList, int NewSize)
{
  MTList(theTList) = realloc_MTList(MTList(theTList),
				    MTListLen(theTList), NewSize);
  SetMTListSize(theTList, NewSize);
}

void MTLAppend(MixedTermList MTL1, int *MTL1Len, 
	       MixedTermList MTL2, int MTL2Len)
/* allocated memory for MTL1 enough to contain MTL1+MTL2 */
{
  int i;

  for ( i=MTL2Len ; i>0 ; i-- )    MTLPutLast(MTL1, *MTL1Len, MTL2[i]);
}

void TListAppendMTL(TermList TL, MixedTermList MTL2, int MTL2Len)
/* allocated memory for MTL1 enough to contain MTL1+MTL2 */
{
  int i, MTL1Len =MTListLen(TL);
  MixedTermList MTL1=MTList(TL);
  
  for ( i=MTL2Len ; i>0 ; i-- )  MTLPutLast(MTL1, MTL1Len, MTL2[i]);
  SetMTListLen(TL, MTL1Len);
}

/********    Simple Power List    ********/

void InsInSPList(int Index, int Exp, eterm SPL)
{
  int CurrentExp =SPL[Index];
  
  if (CurrentExp == 0)
  {  
    eterm_put_non0_nth(SPL, Index, Exp);
    IntsPutLast(Indets(SPL), Index);
  }
  else if (CurrentExp > Exp)
    eterm_put_non0_nth(SPL, Index, Exp);
}

#define SPLDividesTerm(SPL,t)  sp_SmallMult(t,SPL) 

coc_bool TListSimplePowersDivide(TermList theTList, eterm t)
{
  eterm SPL =SPList(theTList);
  int   i;
  
  for ( i=eterm_get_indetsNo(SPL) ; i>0 ; i-- )
    if ( SPL[i]!=0 && SPL[i]<=t[i] )   return TRUE;
  return FALSE;
}

/************************    SPLIT    ************************/
 
void DelLinearSimplePowers(TermList theTList, int *LPSNo)
{
  eterm SPL = SPList(theTList);
  ints  OccInd = Indets(SPL);
  int   n = IntsGetLen(OccInd), sum = 0, index;
  
  for ( ; n>0 ; n-- )
    if ( SPL[(index=OccInd[n])] == 1 )
    {
      sum++;
      eterm_put0_nth(SPL, index);
      IntsMoveLastToNth(OccInd, n); 
    }
  *LPSNo = sum;  
}

eterm GetCoprimeSPList(TermList theTList)
{
  int    MTLLen =MTListLen(theTList), *OccInd;
  MixedTermList   MTL =MTList(theTList);
  int    i, n =MTLLen, index;
  eterm res, theSPL; 
  shortbitset CoprimeSupp;

  if ( IntsGetLen(Indets(SPList(theTList))) == 0 )   return NULL;
  CoprimeSupp = SqFr(SPList(theTList));  
  for ( ; n > 0 ; n--)
  {
    CoprimeSupp &=(~(SqFr(MTL[n]) ) );
    if ( CoprimeSupp == 0 )   return NULL;
  }
  theSPL = SPList(theTList);
  OccInd = Indets(theSPL);  
  res = eterm_init(eterm_get_indetsNo(theSPL));  
  for ( i=IntsGetLen(OccInd) ; i>0 ; i-- )
  {
    if ( bs_get_n(CoprimeSupp,OccInd[i]) )
    {
      index = OccInd[i];
      eterm_put_non0_nth(res, index, theSPL[index]);
      IntsPutLast(Indets(res), index);
      eterm_put0_nth(theSPL, index);      
      IntsMoveLastToNth(OccInd, i);
    }
  }  
  return res;
}

#ifndef ANNA
arpoly MVDelLinearSimplePowers(TermList theTList, 
				   TermList Weights, 
				   ring MVP_R)
{
  eterm   SPL = SPList(theTList);
  ints    OccInd = Indets(SPL);
  int     n = IntsGetLen(OccInd), index;
  arpoly  res = arpoly_one(MVP_R);
  
  for ( ; n>0 ; n-- )
    if ( SPL[(index=OccInd[n])] == 1 )
    {
      res = MultByOneMinusXMultiExp(res,
				    SPMultiDeg(index, 1, Weights, MVP_R),
				    MVP_R);
      eterm_put0_nth(SPL, index);
      IntsMoveLastToNth(OccInd, n);
    }
  return res;
}

arpoly MVDelCoprimeSimplePowers(TermList theTList, 
				    TermList Weights, 
				    ring MVP_R)
{
  eterm CoprimeSPL;
  int   i, index;
  arpoly res =arpoly_one(MVP_R);
  
  CoprimeSPL = GetCoprimeSPList(theTList);
  if ( CoprimeSPL != NULL )
  {
    for ( i=IntsGetLen(Indets(CoprimeSPL)) ; i>0 ; i-- )
    {
      index = (Indets(CoprimeSPL))[ i];
      res = MultByOneMinusXMultiExp(res,
				    SPMultiDeg(index, CoprimeSPL[index],
					       Weights, MVP_R),
				    MVP_R);
    }
    eterm_free(CoprimeSPL);
  }

  return res;
}
#endif  /* ANNA */

void MoveNotCoprimeSP(eterm FromSPList, eterm ToSPList, eterm theTerm)
{
  ints OccInd = Indets(FromSPList);
  int   i = IntsGetLen(OccInd), index;
  
  for ( ; i>0 ; i-- )
    if ( theTerm[(index = OccInd[i])] != 0 )
    {
      eterm_put_non0_nth(ToSPList, index, FromSPList[index]);
      IntsPutLast(Indets(ToSPList), index);
      eterm_put0_nth(FromSPList, index);
      IntsMoveLastToNth(OccInd, i);
    }
}

void MoveNotCoprime(TermList FromTList, TermList ToTList, eterm theTerm)
{
  int       FromMTLLen =MTListLen(FromTList),  ToMTLLen=0;
  MixedTermList   FromMTL =MTList(FromTList),  ToMTL =MTList(ToTList);
  int     n =FromMTLLen;
  
  for ( ; n > 0 ; n--)
    if ( !eterm_coprime(FromMTL[n], theTerm) ) 
    { 
      MTLPutLast(ToMTL, ToMTLLen, FromMTL[n]);
      MTLMoveLastToNth(FromMTL, FromMTLLen, n);
    }
  TListReduceSize(FromTList, FromMTLLen);
  TListReduceSize(ToTList, ToMTLLen);
  
  MoveNotCoprimeSP(SPList(FromTList), SPList(ToTList), theTerm);
}

TermList SplitIndets(TermList theTList)
{
  TermList      res =GlobalSplitterTList;
  int           resLen = 0;
  MixedTermList MTL = MTList(theTList),  resMTL =MTList(res);
  int   n = MTListLen(theTList),  m;
  eterm tn, tm;

  for ( ; n > 0 ; n--)
  {
    tn = MTL[n];
    for ( m=resLen ; m>0 ; m--)
    {
      tm = resMTL[m];
      if ( !eterm_coprime(tn, tm)  )
      {
        eterm_union_and_assign(tm, tn);
        if ( tn != MTL[n] ) eterm_free(tn);
        tn = tm;
        MTLMoveLastToNth(resMTL, resLen, m);
      }  
    }
    if ( tn != MTL[n] )    MTLPutLast(resMTL, resLen, tn);
    else MTLPutLast(resMTL, resLen, eterm_dup(tn));
  }
  if ( resLen == 1 )
  {
    eterm_free(resMTL[1]);
    return NULL;
  }
  SetMTListLen(res, resLen);
  
  return res;
}

/************************    INTERREDUCE    ************************/

void MTLOrderByDecrDegree(MixedTermList theMTList, int Len, int MaxDeg)
{
  eterm t;
  MixedTermList *MTLDeg=(MixedTermList*)mycalloc
   ((MaxDeg+1),sizeof(MixedTermList),"*MixedTermList");
  int *MTLLenDeg=(int*)mycalloc((MaxDeg+1), sizeof(int),"*int"), auxLen =0;
  int i, d;

  for ( d=MaxDeg ; d>=2 ; d--)    MTLDeg[d] = malloc_MTList(Len);

  for ( i=Len ; i>0 ; i--)
  {
    d = eterm_degree((t = theMTList[i]) ); 
    MTLPutLast( MTLDeg[d], MTLLenDeg[d], t);
  }
  for ( d=MaxDeg ; d>1 ; d--)
  {
    if (MTLLenDeg[d] != 0)
      for ( i=MTLLenDeg[d]; i>0; i--)
        MTLPutLast(theMTList, auxLen, (MTLDeg[d])[i]);
    free_MTList(Len, MTLDeg[d]);
  }
  myfree((MaxDeg+1)*sizeof(MixedTermList), MTLDeg, "*MixedTermList");  
  myfree((MaxDeg+1)*sizeof(int), MTLLenDeg, "*int");

}

coc_bool MTLDividesTerm(MixedTermList MTL, int MTLLen, eterm T)
{
  int i = MTLLen;

  for ( ; i>0 ; i-- ) if ( eterm_divides(MTL[i], T) )  return TRUE;

  return FALSE;
}

void InterreduceTList(TermList theTList)
{
  MixedTermList   MTL =MTList(theTList);
  int      MTLLen =MTListLen(theTList);
  int      n =MTLLen, j, MaxDeg =0;
  eterm    auxT;

  for ( ; n> 0 ; n--)
    if ( TListSimplePowersDivide(theTList,(auxT=MTL[n]) )  )
    {
      eterm_free(auxT);
      MTLMoveLastToNth(MTL, MTLLen, n);
    }
    else
      MaxDeg = MAX(MaxDeg, eterm_degree(auxT));
  if (MTLLen > 1)
  {
    MTLOrderByDecrDegree(MTL, MTLLen, MaxDeg);
    for ( n=MTLLen-1 ; n>0; n--)
    {
      auxT =  MTL[n];
      for ( j = MTLLen ; j>n ; j--)
        if ( eterm_divides(MTL[j], auxT)  )
        {
          eterm_free(auxT);
          MTLMoveLastToNth(MTL, MTLLen, n);
          break;
        }
    }
  }
  SetMTListLen(theTList, MTLLen);
}

/************************  PIVOT  ************************/

int MostFrequentIndet(TermList theTList)
{
  MixedTermList MTL =MTList(theTList);
  int IndNo = TListIndetsNo(theTList), exp, MFIndNo = 1, i, j;
  ints OccInd;
  
  for ( j=IndNo ; j>0 ; j-- )    MFI_Occurrences[j] = 0;
  for ( i=MTListLen(theTList) ; i>0 ; i--)
  {
    OccInd = Indets(MTL[i]);
    for ( j=IntsGetLen(OccInd) ; j>0 ; j-- )   MFI_Occurrences[OccInd[j]]++;
  }
  MFI_Indets[1] = (j=IndNo);
  i = MFI_Occurrences[j];
  for ( j-- ; j>0 ; j-- )
    if ((exp=MFI_Occurrences[j]) >= i)
    {
      if (exp > i)  { MFIndNo = 0;  i = exp; }
      MFI_Indets[++MFIndNo] = j;
    }  
 
  if ( i == 1 )       return 0;
  if ( MFIndNo == 1 ) return MFI_Indets[1];
  else                return MFI_Indets[MFIndNo/2];
}

MixedTermList GetTermsWithIndet(TermList theTList, int Indet, int *resLen)
{
  int i =MTListLen(theTList);
  eterm t;
  MixedTermList res =malloc_MTList(i), MTL =MTList(theTList);
  
  *resLen = 0;
  for ( ; i>0 ; i--)
    if ( (t =MTL[i])[Indet] )   MTLPutLast(res, *resLen, t);
  return res;
}

eterm VarPivotOf(TermList theTList)
{
  eterm res;
  int MFIndet;

  if ((MFIndet=MostFrequentIndet(theTList)) == 0)   return NULL;
 
  res = eterm_init(TListIndetsNo(theTList));
  eterm_put_non0_nth(res, MFIndet, 1);
  IntsPutLast(Indets(res), MFIndet);

  return res;
}

eterm GCD3PivotOf(TermList theTList)
{
  eterm res, t;
  int MTLLen, MFIndet, r;
  MixedTermList MTL;
  eterm t1, t2;

  if ((MFIndet=MostFrequentIndet(theTList)) == 0)   return NULL;
 
  MTL = GetTermsWithIndet(theTList, MFIndet, &MTLLen);

  t1 =  MTL[(r =Random(MTLLen)) ];
  MTLMoveLastToNth(MTL, MTLLen, r);
  t2 =  MTL[(r =Random(MTLLen)) ];
  MTLMoveLastToNth(MTL, MTLLen, r);
  res = eterm_gcd(t1, t2);
  if (MTLLen != 0)
  {
    res = eterm_gcd((t=res),  MTL[Random(MTLLen)]);
    eterm_free(t);
  }
  free_MTList(MTListLen(theTList),MTL);

  return res;
}

void BigPivotOf(TermList theTList, int *PivotIndex, int *PivotDeg)
{  
  int MTLLen, MFIndet, r, e1, e2;  
  MixedTermList MTL;

  MFIndet = MostFrequentIndet(theTList);
  *PivotIndex = MFIndet;
  if ( MFIndet == 0 ) return;
  
  MTL = MTList(theTList);
  MTLLen = MTListLen(theTList);
  r = 1; /* fix macro */
  while((e1 =(MTL[r++])[MFIndet]) == 0 );
  r = MTLLen;
  while((e2 =(MTL[r--])[MFIndet]) == 0 );
 
  *PivotDeg   = MIN(e1,e2);
}

/*
#define PivotOf(theTList,MTLLen)\
    ((MTLLen<3)?GCD3PivotOf(theTList):BigPivotOf(theTList))
*/

#define PivotOf(theTList,MTLLen) BigPivotOf(theTList)

eterm eterm_colon_SP(eterm theTerm, int index, int exp)
{
  int OldExp;
 
  if ((OldExp =theTerm[index]) != 0 )
  {
    if ( OldExp <= exp ) printf("OldExp <= exp");
    eterm_put_non0_nth(theTerm, index, OldExp-exp );
  }
  return theTerm;
}

void OccIndDel(eterm T, int index)
{
  ints    OccInd = Indets(T);
  int     n = IntsGetLen(OccInd);

  while ( OccInd[n] != index )  n--;
  IntsMoveLastToNth(OccInd, n);
}

void ReduceAndDivideBySimplePower( TermList theTList, TermList *DivTList,
                                   int PIndex, int PExp)
{
  int MTLLen = MTListLen(theTList), auxLen = MAX(MTListLen(theTList),199),
      DivMTLLen, MTLLenEe, index, TExp, i, e;
  ints DivMTLLenExp;
  eterm  T;
  MixedTermList   *DivMTLExp, MTL =MTList(theTList), DivMTL, MTLEe;
  eterm   DivSPL;
  
  if ( PExp>119 )
  {    
    DivMTLLenExp = (int*)mymalloc((PExp+2)*sizeof(int), "*int");
    DivMTLExp = (MixedTermList*)mymalloc((PExp+2)*sizeof(MixedTermList),
                                         "*MixedTermList");
  }
  else
  {  DivMTLLenExp = Ints120;     DivMTLExp = MTL120;   }
  for ( i=PExp+1 ; i>=0 ; i-- )
  {  DivMTLExp[i] = malloc_MTList(auxLen);    DivMTLLenExp[i] = 0;  }
  
  /* DivTList */
  DivSPL = eterm_colon_SP(eterm_dup(SPList(theTList)), PIndex, PExp);
  /* theTList */
  InsInSPList(PIndex, PExp, SPList(theTList));
 
  for ( i=MTLLen ; i > 0 ; i--)
  {
    TExp = (T=MTL[i])[PIndex];
    if ( TExp > PExp )
    {
      MTLMoveLastToNth(MTL, MTLLen, i);
      eterm_put_non0_nth(T, PIndex, TExp-PExp);
      MTLPutLast(DivMTLExp[PExp+1], DivMTLLenExp[PExp+1], T);
    }
    else if ( TExp == PExp )
    {
      MTLMoveLastToNth(MTL, MTLLen, i);
      if ( eterm_get_OccIndNo(T) == 2 )
      {
	if ((index = (Indets(T))[1]) == PIndex )  index = (Indets(T))[2];
        InsInSPList(index, T[index], DivSPL);
	eterm_free(T);
      }
      else
      {
	eterm_put0_nth(T, PIndex);
	OccIndDel(T, PIndex);
        MTLPutLast(DivMTLExp[TExp], DivMTLLenExp[TExp], T);
      }
    }
    else if ( TExp != 0 )
    {
      if ( eterm_get_OccIndNo(T) == 2 )
      {
	if ((index = (Indets(T))[1]) == PIndex )  index = (Indets(T))[2];
        InsInSPList(index, T[index], DivSPL);
      }
      else
      {
	eterm_put0_nth(T, PIndex);
	OccIndDel(T, PIndex);
        MTLPutLast(DivMTLExp[TExp], DivMTLLenExp[TExp], T);
      }
    }
    else /* TExp == 0 */ MTLPutLast(DivMTLExp[TExp], DivMTLLenExp[TExp], T);
  }
 
  /* DivTList(Interreduction) */
  DivMTL    = DivMTLExp[PExp];
  DivMTLLen = DivMTLLenExp[PExp];
  for ( e=PExp-1 ; e>0 ; e-- )
  {
    MTLEe    = DivMTLExp[e];
    MTLLenEe = DivMTLLenExp[e];
    for ( i=MTLLenEe ; i>0 ; i-- )
    {
      T = MTLEe[i];
      if ( SPLDividesTerm(DivSPL, T) )  MTLMoveLastToNth(MTLEe, MTLLenEe, i);
      else if ( MTLDividesTerm(DivMTL, DivMTLLen, T) )
        MTLMoveLastToNth(MTLEe, MTLLenEe, i);
      else
        MTLPutNth(MTLEe, i, eterm_dup(T));
      eterm_put_non0_nth(T, PIndex, e);
      IntsPutLast(Indets(T), PIndex);
    }
    if ( MTLLenEe!=0 )    MTLAppend(DivMTL, &DivMTLLen, MTLEe, MTLLenEe);
    free_MTList(auxLen, MTLEe);
  }
  MTLEe    = DivMTLExp[0];
  MTLLenEe = DivMTLLenExp[0];
  for ( i=MTLLenEe ; i>0 ; i-- )
  {
    T = MTLEe[i];
    if ( SPLDividesTerm(DivSPL, T) )   MTLMoveLastToNth(MTLEe, MTLLenEe, i);
    else if ( MTLDividesTerm(DivMTL, DivMTLLen, T) )
      MTLMoveLastToNth(MTLEe, MTLLenEe, i);
    else
      MTLPutNth(MTLEe, i, eterm_dup(T));
  }
  if ( MTLLenEe!=0 )    MTLAppend(DivMTL, &DivMTLLen, MTLEe, MTLLenEe);
  free_MTList(auxLen, MTLEe);
  
  MTLAppend(DivMTL, &DivMTLLen, DivMTLExp[PExp+1], DivMTLLenExp[PExp+1]);
  free_MTList(auxLen, DivMTLExp[PExp+1]);
  *DivTList = TListMake(DivSPL, DivMTL, auxLen, DivMTLLen);
 
  /* theTList */
  TListReduceSize(theTList, MTLLen);
  if ( PExp>119 )
  {
    myfree((PExp+2)*sizeof(int), DivMTLLenExp, "*int");
    myfree((PExp+2)*sizeof(MixedTermList), DivMTLExp, "*MixedTermList");
  }
}

void ReduceAndDivideByMixedTerm( TermList theTList, TermList *DivTList,
                                 eterm Pivot)
{
  int  MTLLen =MTListLen(theTList), OldMTLLen =MTLLen, 
       BMLen =0, CMTLLen =0, DivMTLLen =0;
  int index, TDeg;
  eterm     DivT,  t, auxSPL;
  int     i = MTLLen;
  MixedTermList   MTL =MTList(theTList),  BigMultMTL, CoprimeMTL, DivMTL;
  eterm   DivSPL;
 
  *DivTList = NewTList(MTLLen, TListIndetsNo(theTList));
  DivMTL = MTList(*DivTList);
  DivSPL = eterm_colon(eterm_dup(SPList(theTList)), Pivot);
  auxSPL =(SPList(*DivTList));  /* eterm_init() */

  BigMultMTL = malloc_MTList(MTLLen);
  CoprimeMTL = malloc_MTList(MTLLen);
 
  for ( ; i > 0 ; i-- )
  {
    t=MTL[i];
    TDeg = eterm_degree(t);

    if ( eterm_divides(Pivot,t) )
    {
      /* theTList */
      MTLMoveLastToNth(MTL, MTLLen, i);
      /* DivTList */
      if (sp_BigMult(t,Pivot))
	MTLPutLast(BigMultMTL, BMLen, eterm_colon(t,Pivot));
      else
      {
	DivT = eterm_colon(t, Pivot);
	if ( IntsGetLen(Indets(DivT)) == 1)
	{
	  index =  (Indets(DivT))[ 1];
	  InsInSPList(index, DivT[index], DivSPL);
	  InsInSPList(index, DivT[index] + Pivot[index], auxSPL);
	  eterm_free(DivT);
	}
	else
	  MTLPutLast(DivMTL, DivMTLLen, DivT);
      }
    }
    else if ( SPLDividesTerm(auxSPL, t) )
      ;
    else if (TDeg==eterm_degree((DivT=eterm_colon(eterm_dup(t),Pivot) )) )
      MTLPutLast(CoprimeMTL, CMTLLen, DivT);
    else if ( IntsGetLen(Indets(DivT)) == 1 )
    {
      index =  (Indets(DivT))[ 1];
      InsInSPList(index, DivT[index], DivSPL);
      InsInSPList(index, DivT[index] + Pivot[index], auxSPL);
      eterm_free(DivT);
    }
    else
      MTLPutLast(DivMTL, DivMTLLen, DivT);
  }
  /* theTList */
  MTLPutLast(MTL, MTLLen, Pivot);
  TListReduceSize(theTList, MTLLen);
  /* DivTList */
  eterm_colon(auxSPL, Pivot);
  SetMTListLen(*DivTList, DivMTLLen);
  InterreduceTList(*DivTList);
  DivMTLLen = MTListLen(*DivTList);

  if ( DivMTLLen != 0 )

    for ( i=CMTLLen ; i>0 ; i-- )
    {
      if ( SPLDividesTerm(DivSPL,(t=(CoprimeMTL)[i]) ) )
      {
	eterm_free(t);
	MTLMoveLastToNth(CoprimeMTL, CMTLLen, i);
      }
      else if ( MTLDividesTerm(DivMTL, DivMTLLen, t) )
      {
	eterm_free(t);
	MTLMoveLastToNth(CoprimeMTL, CMTLLen, i);
      }
    }
  else
    for ( i=CMTLLen ; i>0 ; i-- )
      if ( SPLDividesTerm(DivSPL,(t=(CoprimeMTL)[i]) ) )
      {
	eterm_free(t);
	MTLMoveLastToNth(CoprimeMTL, CMTLLen, i);
      }
  SPList(*DivTList) = DivSPL;
  if (CMTLLen!=0)  MTLAppend(DivMTL, &DivMTLLen, CoprimeMTL, CMTLLen);
  if (BMLen!=0)    MTLAppend(DivMTL, &DivMTLLen, BigMultMTL, BMLen);
  free_MTList(OldMTLLen, CoprimeMTL);
  free_MTList(OldMTLLen, BigMultMTL);
  eterm_free(auxSPL);
  TListReduceSize(*DivTList, DivMTLLen);
}

void ReduceAndDivideByPivot(TermList theTList, TermList *DivTList, eterm Pivot)
{
  if ( eterm_get_OccIndNo(Pivot) == 1)
  {
    int index = (Indets(Pivot))[1], exp = eterm_degree(Pivot);
    eterm_free(Pivot);
    ReduceAndDivideBySimplePower(theTList, DivTList, index, exp);
  }
  else
    ReduceAndDivideByMixedTerm(theTList, DivTList, Pivot);
}

/* TermLists.c: END */

/*************************************************************
 ************************    SPLIT    ************************
 *************************************************************/
 
unipoly TotalSplitPoincare(TermList theTList)
{
  unipoly   res = NULL;
  eterm  myTerm, SPL =SPList(theTList), mySPL;
  MixedTermList  MTL =MTList(theTList);
  int MTLLen =MTListLen(theTList), n =MTLLen, DegSum =0;
 
  for ( ; n>0 ; n--)
  {
    MoveNotCoprimeSP(SPL,
                     ( mySPL  =eterm_init(TListIndetsNo(theTList)) ),
                     ( myTerm =MTL[n] ) );
    if ( eterm_get_OccIndNo(mySPL)!=0 )
    {
      MTLMoveLastToNth(MTL, MTLLen, n);
      if ( res == NULL )   res = UnipolyOneTermAndSPPoincare(myTerm,mySPL);
      else  res = P1TimesP2(res, UnipolyOneTermAndSPPoincare(myTerm,mySPL));
    }
    else
    {
      DegSum += eterm_degree(myTerm);
      eterm_free(mySPL);
    }
  }
  n = IntsGetLen(Indets(SPL));
  if ((n==eterm_degree(SPL)) && res == NULL )
    res = PowerExtDup(n, n + DegSum);
  else
  {
    if ( res != NULL )
    {
      if ( UPSize(res) < UPDeg(res) + eterm_degree(SPL) + DegSum )
        res = UnipolyChangeSize(res, UPDeg(res) + eterm_degree(SPL) + DegSum);
    }
    else
      res = UnipolyOne(eterm_degree(SPL) + DegSum);
    for ( ; n>0 ; n-- )  MultByOneMinusXExp(res, SPL[(Indets(SPL))[n]]);
  }
  SetMTListLen(theTList,MTLLen);
  for ( n=MTLLen ; n>0 ; n--)   MultByOneMinusXExp(res, eterm_degree( MTL[n]));
  EraseAndFreeTList(theTList);
 
  return res;
}

#ifdef ANNA
unipoly TruncTotalSplitPoincare(TermList theTList, int d)
{
  unipoly   res;
  eterm  myTerm, SPL =SPList(theTList), mySPL;
  MixedTermList  MTL =MTList(theTList);
  int MTLLen =MTListLen(theTList), n =MTLLen, DegSum =0;
 
  /*
  printf("TruncTotalSplit %d:  ", d);
  */
  res = UnipolyOne(0);
  for ( ; n>0 ; n--)
  {
    MoveNotCoprimeSP(SPL,
                     ( mySPL =eterm_init(TListIndetsNo(theTList)) ),
                     ( myTerm =MTL[n] ) );
    if (IntsGetLen(Indets(mySPL))!=0)
    {
      MTLMoveLastToNth(MTL, MTLLen, n);
      res = P1TimesP2(res,
		    TruncUnipolyOneTermAndSPPoincare(myTerm,mySPL,d));
      /*
      printf("OneTermAndSP %d ", UPDeg(res));
      */
    }
    else
    {
      DegSum += eterm_degree(myTerm);
      eterm_free(mySPL);
    }
  }
  n = IntsGetLen(Indets(SPL));
  if ((n==eterm_degree(SPL)) &&(UPDeg(res)==0) )
  {
    FreeUnipoly(res);
    res = PowerExtDup(n, n+DegSum);
  }
  /*    CopyPower(n, res);*/
  else
  {
    {
      if ( UPSize(res) < UPDeg(res) + eterm_degree(SPL) + DegSum )
        res = UnipolyChangeSize(res, UPDeg(res) + eterm_degree(SPL) + DegSum);
    }
    for ( ; n>0 ; n-- )
    {
      /*      printf("p %d  ", SPL[(Indets(SPL))[n]]);*/
      TruncMultByOneMinusXExp(res, SPL[(Indets(SPL))[n]], d);
    }
  }
  SetMTListLen(theTList,MTLLen);
  for ( n=MTLLen ; n>0 ; n--)
  {
    /*    printf("m %d  ", eterm_degree( MTL[n]));*/
    TruncMultByOneMinusXExp(res, eterm_degree( MTL[n]), d);
  }
  EraseAndFreeTList(theTList);
  /*
  printf("\n");
  */
 
  return res;
}

unipoly TruncRecPoincare(TermList theTList, int n);
#endif /* ANNA - univariate trunc */

unipoly BigRecPoincare(TermList theTList);
unipoly RecPoincare(TermList theTList);
unipoly RecPoincareCoeffNth(TermList theTList, int n);

unipoly SplitPoincare(TermList theTList, TermList SplitterTList)
{
  unipoly   res;
  int n =MTListLen(SplitterTList), NewMTLLen =0, DegSum =0;
  eterm   SPL =SPList(theTList);
  TermList NewTL =NewTList(MTListLen(theTList),TListIndetsNo(theTList)),
           *auxTLs =(TermList*)mymalloc((n+1)*sizeof(TermList), "*TermList");
  MixedTermList   SMTL =MTList(SplitterTList), NewMTL =MTList(NewTL);
  
  res = NULL;
  for ( n =MTListLen(SplitterTList) ; n>0 ; n-- )
  {
    auxTLs[n] = NewTList(MTListLen(theTList),TListIndetsNo(theTList));
    MoveNotCoprime(theTList, auxTLs[n], SMTL[n]);
    eterm_free(SMTL[n]);
  }
  for ( n =MTListLen(SplitterTList) ; n>0 ; n-- )
    if ((MTListLen(auxTLs[n])!=1) ||(IntsGetLen(Indets(SPList(auxTLs[n])))!=0) )
      if ( res == NULL )
        res = BigRecPoincare(auxTLs[n]);
      else
        res = P1TimesP2(res, BigRecPoincare(auxTLs[n]));
    else
    {
      DegSum += eterm_degree((MTList(auxTLs[n]))[1]);
      MTLPutLast(NewMTL, NewMTLLen,(MTList(auxTLs[n]))[1]);
      SetMTListLen(auxTLs[n],0);
      EraseAndFreeTList(auxTLs[n]);
    }
  if ( UPSize(res) < UPDeg(res) + eterm_degree(SPL) + DegSum )
    res = UnipolyChangeSize(res, UPDeg(res) + eterm_degree(SPL) + DegSum);
  for ( n = IntsGetLen(Indets(SPL)) ; n>0 ; n-- )
    MultByOneMinusXExp(res, SPL[(Indets(SPL))[n]]);
  EraseAndFreeTList(theTList);
  for ( n=NewMTLLen ; n>0 ; n--)
    MultByOneMinusXExp(res, eterm_degree((NewMTL)[n]));
  SetMTListLen(NewTL,NewMTLLen);
  EraseAndFreeTList(NewTL);
 
  return res;
}

#ifndef ANNA
/*************************************************************************/
/******************  MV  *******************/

arpoly MVTotalSplitPoincare(TermList theTList, TermList Weights,ring MVP_R)
{
  MixedTermList  MTL =MTList(theTList);
  arpoly   res =arpoly_one(MVP_R), auxP;
  TermList   auxTL;
  eterm  t, myTerm, SPL =SPList(theTList), mySPL;
  int MTLLen =MTListLen(theTList), n =MTLLen, DegSum =0;

  for ( ; n>0 ; n--)
  {
    MoveNotCoprimeSP(SPL,
                     ( mySPL =eterm_init(TListIndetsNo(theTList)) ),
                     ( myTerm =MTL[n] ) );
    if ( IntsGetLen(Indets(mySPL)) != 0 )
    {
      MTLMoveLastToNth(MTL, MTLLen, n);
      res = MVP1TimesP2 
	(res,  MVOneTermAndSPPoincare(myTerm,mySPL,Weights,MVP_R),  MVP_R);
    }
    else
      eterm_free(mySPL);
  }
  n = IntsGetLen(Indets(SPL));
  for ( ; n>0 ; n-- )
    res = MultByOneMinusXMultiExp
     (res,  
      SPMultiDeg( (Indets(SPL))[n], SPL[(Indets(SPL))[n]], Weights,  MVP_R),
      MVP_R);
  SetMTListLen(theTList,MTLLen);
  for ( n=MTLLen ; n>0 ; n--)
    res = MultByOneMinusXMultiExp
     (res,  eterm_MultiDeg(MTL[n],Weights,MVP_R),  MVP_R);
  EraseAndFreeTList(theTList);
 
  return res;
}

arpoly MVRecPoincare(TermList theTList, TermList Weights, ring MVP_R);

arpoly MVSplitPoincare(TermList theTList, 
			   TermList SplitterTList, 
			   TermList Weights, 
			   ring MVP_R)
{
  arpoly   res, auxP;
  int n =MTListLen(SplitterTList), NewMTLLen =0, DegSum =0;
  eterm   SPL =SPList(theTList);
  TermList NewTL =NewTList(MTListLen(theTList),TListIndetsNo(theTList)),
           *auxTLs =(TermList*)mymalloc((n+1)*sizeof(TermList), "*TermList");
  MixedTermList   SMTL =MTList(SplitterTList), NewMTL =MTList(NewTL);
  
  res = NULL;
  for ( n =MTListLen(SplitterTList) ; n>0 ; n-- )
  {
    auxTLs[n] = NewTList(MTListLen(theTList),TListIndetsNo(theTList));
    MoveNotCoprime(theTList, auxTLs[n], SMTL[n]);
    eterm_free(SMTL[n]);
  }
  for ( n =MTListLen(SplitterTList) ; n>0 ; n-- )
    if ((MTListLen(auxTLs[n])!=1) ||(IntsGetLen(Indets(SPList(auxTLs[n])))!=0) )
      if ( res == NULL )
        res = MVRecPoincare(auxTLs[n], Weights, MVP_R);
      else
        res = MVP1TimesP2(res, MVRecPoincare(auxTLs[n], Weights, MVP_R), MVP_R);
    else
    {
      MTLPutLast(NewMTL, NewMTLLen,(MTList(auxTLs[n]))[1]);
      SetMTListLen(auxTLs[n],0);
      EraseAndFreeTList(auxTLs[n]);
    }
  for ( n = IntsGetLen(Indets(SPL)) ; n>0 ; n-- )
    res = MultByOneMinusXMultiExp
     (res,  
      SPMultiDeg( (Indets(SPL))[n], SPL[(Indets(SPL))[n]], Weights,  MVP_R),
      MVP_R);
  EraseAndFreeTList(theTList);
  for ( n=NewMTLLen ; n>0 ; n--)
    res = MultByOneMinusXMultiExp
     (res,  eterm_MultiDeg((NewMTL)[n],Weights,MVP_R),  MVP_R);
  SetMTListLen(NewTL,NewMTLLen);
  EraseAndFreeTList(NewTL);
 
  return res;
}

/******************  TruncMV  *******************/

arpoly TruncMVTotalSplitPoincare(TermList theTList, 
				     TermList Weights, 
				     int Trunc,
				     ring MVP_R)
{
  arpoly   res, auxP;
  TermList   auxTL;
  eterm  t, myTerm;
  MixedTermList   MTL =MTList(theTList);
  eterm   SPL =SPList(theTList), mySPL;
  int i =MTListLen(theTList);
    
  MoveNotCoprimeSP(SPL,
		   (mySPL =eterm_init(TListIndetsNo(theTList))),
		   (myTerm =MTL[i--]) );
  res = TruncMVOneTermAndSPPoincare(myTerm, mySPL, Weights, Trunc, MVP_R);
  while( i > 1 )
  {
    MoveNotCoprimeSP(SPL, 
		     (mySPL =eterm_init(TListIndetsNo(theTList))), 
		     (myTerm =MTL[i--]) );
    res = TruncMVP1TimesP2 
     (res,  TruncMVOneTermAndSPPoincare(myTerm,mySPL,Weights,Trunc,MVP_R), 
       Trunc,  MVP_R);
  }
  SetMTListLen(theTList,1);
  return TruncMVP1TimesP2 
   (res, TruncMVLenOnePoincare(theTList,Weights,Trunc,MVP_R), Trunc,  MVP_R);
}

arpoly TruncMVRecPoincare(TermList theTList,  TermList Weights, 
			      int Trunc,   ring MVP_R);

arpoly TruncMVSplitPoincare(TermList theTList,  TermList SplitterTList, 
				TermList Weights,  int Trunc,  ring MVP_R)
{
  arpoly   res, auxP;
  TermList   auxTL;
  MixedTermList   SMTL =MTList(SplitterTList);
  int n =MTListLen(SplitterTList);
  
  auxTL = NewTList(MTListLen(theTList),TListIndetsNo(theTList));
  MoveNotCoprime(theTList, auxTL,(SMTL)[n--] );
  res = TruncMVRecPoincare(auxTL, Weights, Trunc, MVP_R);
  
  for ( ; n > 1 ; n--)
  {
    auxTL = NewTList(MTListLen(theTList),TListIndetsNo(theTList));
    MoveNotCoprime(theTList, auxTL,(SMTL)[n] );
    res = TruncMVP1TimesP2(res, 
			    TruncMVRecPoincare(auxTL, Weights, Trunc, MVP_R), 
			    Trunc, 
			    MVP_R);
  }
  EraseAndFreeTList(SplitterTList);
  return TruncMVP1TimesP2(res, 
			   TruncMVRecPoincare(theTList, Weights, Trunc, MVP_R), 
			   Trunc, 
			   MVP_R);
}

/******************  TruncMV end  *******************/
/******************  MV end  *******************/
#endif  /* ANNA */

unipoly BigRecPoincare(TermList theTList)
{
  int MTLLen = MTListLen(theTList), PivotDeg, PivotIndex;
  TermList DivTList, SplitterTL;

  if (MTLLen == 0) return    UnipolySPPoincare(theTList);
  if (MTLLen == 1) return    UnipolyLenOnePoincare(theTList);
  BigPivotOf(theTList, &PivotIndex, &PivotDeg);
  if ( PivotIndex == 0 )  return    TotalSplitPoincare(theTList);
  if ((MTLLen < TListIndetsNo(theTList)) && (MTLLen > 4) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
      return SplitPoincare(theTList, SplitterTL);
  ReduceAndDivideBySimplePower(theTList, &DivTList, PivotIndex, PivotDeg);
  
  return P1PlusXExpP2(BigRecPoincare(theTList),  PivotDeg, 
		      BigRecPoincare(DivTList));
}

unipoly RecPoincare(TermList theTList)
{
  int MTLLen = MTListLen(theTList), PivotDeg;
  eterm Pivot;
  TermList DivTList, SplitterTL;
    
  if (MTLLen == 0) return    UnipolySPPoincare(theTList);
  if (MTLLen == 1) return    UnipolyLenOnePoincare(theTList);
  if ((Pivot =GCD3PivotOf(theTList)) == NULL )  
    return    TotalSplitPoincare(theTList);
  if ((MTLLen < TListIndetsNo(theTList)) && (MTLLen > 4) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
    {
      eterm_free(Pivot);
      return     SplitPoincare(theTList, SplitterTL);
    }
  PivotDeg = eterm_degree(Pivot);
  ReduceAndDivideByPivot(theTList, &DivTList, Pivot);
  
  return P1PlusXExpP2(RecPoincare(theTList),  PivotDeg, 
		      RecPoincare(DivTList));
/*
  return P1TimesP2(unipoly_dup(PowerList[LPSNo]),
		   P1PlusXExpP2(RecPoincare(theTList), PivotDeg, 
				RecPoincare(DivTList)),
		   P_R);
*/
}

#ifdef ANNA /* ANNA - univariate trunc */
void TLTrunc(TermList theTList, int Trunc)
{
  int n, index, MTLLen =MTListLen(theTList);
  MixedTermList MTL =MTList(theTList);
  eterm   SPL =SPList(theTList);
  ints OccInd = Indets(SPL);
  
  for ( n=IntsGetLen(OccInd) ; n>0 ; n-- )
    if ( SPL[(index = OccInd[n])] > Trunc )
    {
      /*
      printf("P");
      */
      eterm_put0_nth(SPL, index);
      IntsMoveLastToNth(OccInd, n);
    }

  for ( n=MTLLen ; n>0 ; n-- )
    if ( eterm_degree(MTL[n]) > Trunc )
    {
      /*
      printf(">");
      */
      eterm_free(MTL[n]);
      MTLMoveLastToNth(MTL, MTLLen, n);
    }
  TListReduceSize(theTList, MTLLen);
}

unipoly TruncRecPoincare(TermList theTList, int n)
{
  int MTLLen = MTListLen(theTList), PivotDeg, PivotIndex;
  TermList DivTList /*, SplitterTL*/;
    
  if (MTLLen == 0) return    TruncUnipolySPPoincare(theTList, n);
  if (MTLLen == 1) return    TruncUnipolyLenOnePoincare(theTList, n);
  BigPivotOf(theTList, &PivotIndex, &PivotDeg);
  if ( PivotIndex == 0 )  return    TruncTotalSplitPoincare(theTList, n);
  /*
  if ((MTLLen < TListIndetsNo(theTList)) && (MTLLen > 4) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
      return SplitPoincare(theTList, SplitterTL);
      */
  ReduceAndDivideBySimplePower(theTList, &DivTList, PivotIndex, PivotDeg);
  TLTrunc(DivTList, n);

  return TruncP1PlusXExpP2(TruncRecPoincare(theTList, n),  PivotDeg, 
		      TruncRecPoincare(DivTList, n-PivotDeg), n);
}

#if 0
unipoly RecPoincareCoeffNth(TermList theTList, int n)
{
  int MTLLen = MTListLen(theTList), PivotDeg;
  eterm Pivot;
  TermList DivTList, SplitterTL;
    
  if (MTLLen == 0) return    UnipolySPPoincare(theTList);
  if (MTLLen == 1) return    UnipolyLenOnePoincare(theTList);
  if ((Pivot =BigPivotOf(theTList)) == NULL )  
    return    TotalSplitPoincare(theTList);
  if ((MTListLen(theTList) <= TListIndetsNo(theTList)/2) &&
     (MTListLen(theTList) >= 8) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
    {
      eterm_free(Pivot);
      return     SplitPoincare(theTList, SplitterTL);
    }
  PivotDeg = eterm_degree(Pivot);
  ReduceAndDivideByPivot(theTList, &DivTList, Pivot);
  
  return P1PlusXExpP2(BigRecPoincare(theTList),  PivotDeg, 
		      BigRecPoincare(DivTList));
}
#endif
#endif /* ANNA - univariate trunc */

#ifndef ANNA 

/******************  MV  *******************/

arpoly TruncMVRecPoincare (TermList theTList, 
			       TermList TLWeights, 
			       int Trunc, 
			       ring MVP_R)
{
  int MTLLen = MTListLen(theTList);
  term PivotMultiDeg;
  int PivotWDeg;
  eterm Pivot;
  TermList DivTList, SplitterTL;
  arpoly CoprimeSPPoly;
    
  if (Trunc <= 0) 
  {
    /* printf("Trunc = %d <= 0\n", Trunc); */
    EraseAndFreeTList(theTList);
    if ( Trunc==0 ) 
      return arpoly_one(MVP_R);
    else
      return NULL; 
  }
  if (MTLLen == 0)   
    return    TruncMVSPPoincare(theTList, TLWeights, Trunc, MVP_R);
  if (MTLLen == 1)   
    return    TruncMVLenOnePoincare(theTList, TLWeights, Trunc, MVP_R);
  if ((Pivot =GCD3PivotOf(theTList)) == NULL )  
    return    TruncMVTotalSplitPoincare(theTList, TLWeights, Trunc, MVP_R);
  if ((MTListLen(theTList) <= TListIndetsNo(theTList)/2) &&
     (MTListLen(theTList) >= 8) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
      {
	eterm_free(Pivot);
	return TruncMVSplitPoincare(theTList,SplitterTL,TLWeights,Trunc,MVP_R);
      }
  PivotMultiDeg = eterm_MultiDeg(Pivot, TLWeights, MVP_R);
  PivotWDeg = term_get_nth(PivotMultiDeg, 1);
  CoprimeSPPoly = MVDelCoprimeSimplePowers(theTList, TLWeights, MVP_R);
  ReduceAndDivideByPivot(theTList, &DivTList, Pivot );

  return TruncMVP1TimesP2
   (CoprimeSPPoly,
     TruncP1PlusXMultiExpP2(TruncMVRecPoincare(theTList, TLWeights,
					       Trunc, MVP_R), 
			    PivotMultiDeg, 
			    TruncMVRecPoincare(DivTList, TLWeights,
					       Trunc-PivotWDeg, MVP_R),
			    Trunc, MVP_R),
     Trunc, MVP_R);
}

arpoly MVRecPoincare(TermList theTList, TermList Weights, ring MVP_R)
{
  int MTLLen = MTListLen(theTList);
  term PivotDeg;
  eterm Pivot;
  TermList DivTList, SplitterTL;
  arpoly CoprimeSPPoly;  
  
  if (MTLLen == 0)   return    MVSPPoincare(theTList, Weights, MVP_R);
  if (MTLLen == 1)   return    MVLenOnePoincare(theTList, Weights, MVP_R);
  if ((Pivot =GCD3PivotOf(theTList)) == NULL )  
    return    MVTotalSplitPoincare(theTList, Weights, MVP_R);
  if ((MTListLen(theTList) <= TListIndetsNo(theTList)/2) &&
     (MTListLen(theTList) >= 8) )
    if ((SplitterTL=SplitIndets(theTList))!= NULL )
    {
      eterm_free(Pivot);
      return  MVSplitPoincare(theTList, SplitterTL, Weights, MVP_R);
    }
  PivotDeg = eterm_MultiDeg(Pivot, Weights, MVP_R);
  CoprimeSPPoly = MVDelCoprimeSimplePowers(theTList, Weights, MVP_R);
  ReduceAndDivideByPivot(theTList, &DivTList, Pivot );
  return MVP1TimesP2
   (CoprimeSPPoly,  
     P1PlusXMultiExpP2(MVRecPoincare(theTList, Weights, MVP_R),  PivotDeg,
		       MVRecPoincare(DivTList, Weights, MVP_R),  MVP_R),
     MVP_R);
}

/******************  MV end  *******************/

TermList WeilistToTList(weilist wl, ring R)
{
  eterm t;
  TermList aux;
  int n,i;
  int NewLen=0;

  n = weilist_len(wl);
  aux = NewTList(n, R->indetsNo);
  
  for ( i=1 ; i<=n ; i++ )
  {
    t = term2eterm(weight2term(weilist_nth(wl,i)));
    MTLPutLast(MTList(aux), NewLen, t);
    t = NULL;
  }
  SetMTListLen(aux, NewLen);
  
  return aux;
}

olist PoincareDenFactors(ring R)
{
  olist ol,aux_ol;
  weight w =ring_weights(R);
  int i, n;

  ol = olist_init(n=R->indetsNo);
  for ( i=1 ; i<=n ; i++ )
  {
    aux_ol = olist_init(1);
    olist_set_nth(aux_ol,1,int2object(weight_nth(w,i)));
    olist_set_nth(ol,i,olist2object(aux_ol));
  }
  return ol;
}

olist MVPoincareDenFactors(weilist Weights, ring R)
{
  olist ol,aux_ol;
  int i, j;
  int IndNo, WNo, n;
  TermList TLWeights;
  MixedTermList W;

  TLWeights = WeilistToTList(Weights, R);
  W = MTList(TLWeights);
  WNo = MTListLen(TLWeights);
  ol=olist_init(IndNo=R->indetsNo);
  for (i=1 ; i<=IndNo ; i++)
  {
    aux_ol=olist_init(WNo);
    for ( n=WNo ; n>0 ; n-- )
      olist_set_nth(aux_ol, n, int2object(SP_eterm_scalar(i, 1, W[n])));
    olist_set_nth(ol,i,olist2object(aux_ol)); 
  }
  EraseAndFreeTList(TLWeights);
  return ol;
}
#else  /* ANNA */

/*
UPCoeff PoincareNumCoeffNth(TermList theTList, int n)
{
  return RecPoincareCoeffNth(theTList, n);
}
*/

unipoly TLTruncPoincareNumerator(TermList theTList, int n)
{
  TLTrunc(theTList, n);
  return TruncRecPoincare(theTList, n);
}

#endif  /* ANNA */

unipoly TLPoincareNumerator(TermList theTList)
{
  StartPoincare(TListIndetsNo(theTList));
  
  return BigRecPoincare(theTList);
}




