/* poincare.h */

/*-----------------------------------------------------------------------------

	Part of project CoCoA

-----------------------------------------------------------------------------*/

#ifndef POINCARE_H
#define POINCARE_H

#ifndef ANNA
#include "auxtypes.h"
#include "coctypes.h"
#else  /* ANNA */
#include "AnnaUtils.h"
#endif   /* ANNA */

/**********************************************************/
typedef eterm      * MixedTermList;
typedef struct TermListStruct * TermList;
 
typedef struct TermListStruct
{
  MixedTermList   MixedTerms;
  int MixedTermListSize;
  int MixedTermListLen;
} TermListStruct; 

/********************    list of simple powers    ********************/
/*
    SPList(theTList) is represented with an eterm ( MTL[0] )
    which is the product of the simple powers.
*/

void InsInSPList(int Index, int Exp, eterm SPL);

/********************    list of mixed terms    ********************/

#define MTLPutNth(MTList,N,t)    (MTList)[N] = t
#define MTLPutLast(MTList,len,t)        (MTList)[++len] = t
#define MTLMoveLastToNth(MTList,len,N)  (MTList)[N] = (MTList)[len--]

/********************    list of terms    ********************/

#define SPList(theTList)      ((theTList)->MixedTerms)[0]
#define MTList(theTList)      (theTList)->MixedTerms
#define MTListSize(theTList)   (theTList)->MixedTermListSize
#define MTListLen(theTList)    (theTList)->MixedTermListLen
#define TListIndetsNo(theTList) eterm_get_indetsNo(SPList(theTList))

#define SetMTListSize(theTList,size)  (theTList)->MixedTermListSize = size
#define SetMTListLen(theTList,len)    (theTList)->MixedTermListLen = len

#define TListReduceSize(TL,newsz) \
        {  SetMTListLen(TL,newsz);\
	   if ( MTListSize(TL)!=199 && MTListSize(TL) > newsz*16)\
	     TListChangeSize(TL, newsz);}

TermList NewTList(int len,int indetsNo);
void EraseAndFreeTList(TermList theTList);
void TListChangeSize  (TermList theTList, int NewLen);

void InterreduceTList (TermList theTList);

void poincare_init(int IndetsNo);
void StartPoincare(int IndetsNo);

/**********************************************************************/

#ifdef ANNA
UPCoeff PoincareNumCoeffNth(TermList theTList, int n);
unipoly TLTruncPoincareNumerator(TermList theTList, int n);
unipoly TLIncrPoincareNumerator
        (TermList OldTList, unipoly OldPNum, TermList NewTList);
#endif  /* ANNA */

unipoly TLPoincareNumerator(TermList theTList);

#ifndef ANNA 
olist PoincareDenFactors(ring R);
olist MVPoincareDenFactors(weilist Weights, ring R);

unipoly RecPoincare(TermList theTList);
unipoly BigRecPoincare(TermList theTList);
arpoly MVRecPoincare(TermList theTList, TermList Weights, ring MVP_R);

TermList VListToTLists (vlist vl, ring R);
TermList WeilistToTList (weilist wl, ring R);
#endif  /* ANNA */

#endif /* POINCARE_H */
