/* $Id: polys.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "coeffs.h"
#include "eterms.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "options.h"
#include "polys.h"
#include "ring.h"
#include "terms.h"
#include "weights.h"

void polynomials_init()
{
  monomial_adt_init();
#ifdef MONOPOLI
  monopoli_adt_init();
#endif
#ifdef TRIPOLI
  tripoli_adt_init();
#endif
#ifdef EEE
  eee_adt_init();
#endif
}

unknown arpoly2indet(arpoly p, ring R)
{
  expType expo;
  unknown res_x,aux_x;
  term aux_t;
  res_x=unknown_NULL;
  if ((p != NULL) && arpoly_len(p,R) == 1) {
    aux_t=arpoly_leading_term(p,R);
    if (term_pure_power(aux_t,&aux_x,&expo)
      && term_get_true_nth(aux_t,aux_x,R)==1) res_x=aux_x-1;
    term_free(aux_t); 
  }
  return res_x;
}

arpoly cterm2arpoly(cterm t, ring R)
{
  /* t=poly(m1:mon,...,mn:mon) */
  int arity=cterm_arity(t);
  monomial aux_m;
  register int i;
  cterm_list params;
  arpoly res_p;
  params=cterm_params(t);
  res_p=NULL;
  for (i=0;i<arity;i++) {
    aux_m=cterm2monomial(params[i],R);
    res_p=arpoly_sum_these(res_p,arpoly_this_monomial2arpoly(aux_m,R),R); }
  return res_p;
}

int this_arpoly2int(arpoly p, ring R)
{
  if (p == NULL) return ZERO;
  return this_coeff2int(arpoly_this_leading_coeff(p,R),R);
}

arbrat this_arpoly2arbrat(arpoly p, ring R)
{
  arbrat res_n;
  if (p == NULL) return int2arbrat(0);
  res_n=this_coeff2arbrat(arpoly_leading_coeff(p,R),R);
  arpoly_free(p,R);
  return res_n;
}

arbint this_arpoly2arbint(arpoly p, ring R)
{
  arbint res_n;
  if (p == NULL) return int2arbint(0);
  res_n=this_coeff2arbint(arpoly_leading_coeff(p,R),R);
  arpoly_free(p,R);
  return res_n;
}

arpoly arpoly_sub_these(arpoly p1, arpoly p2, ring R)
{
  arpoly_change_sign(p2,R);
  return arpoly_sum_these(p1,p2,R);
}

pvcomp pvcomp_sub_these(pvcomp p1, pvcomp p2, ring R)
{
  pvcomp_change_sign(p2,R);
  return pvcomp_sum_these(p1,p2,R);
}

arpoly arpoly_full_aux(int deg, term aux_t, unknown u, ring R)
/* return the polynomial with support s.t. degree of the subterm from 
   unknwown u is deg and the subterm from first unknown to u-1 is the
   same of aux_t */
{
  arpoly aux_p;
  weight w;
  int i, d;
  if (deg < 0) return NULL;
  if (deg==0) {
    for (i=u+1;i<=R->indetsNo;i++) term_put_nth(aux_t,i,0);
    return arpoly_this_term2arpoly(term_dup(aux_t),R); }
  if (u==R->indetsNo) return NULL;
  w=ring_weights(R);
  aux_p=NULL;
  for (d=0;d<=deg;d++) {
    term_put_nth(aux_t,u+1,d*weight_nth(w,u+1));
    aux_p=arpoly_sum_these(aux_p,
      arpoly_full_aux(deg-d*weight_nth(w,u+1),aux_t,(unknown)(u+1),R),R); }
  return aux_p;
}

arpoly arpoly_full_support_of_degree(int deg, ring R)
{
  term aux_t;
  arpoly res_p;
  aux_t=term_init(R->indetsNo);
  res_p=arpoly_full_aux(deg,aux_t,0,R);
  term_free(aux_t);
  return res_p;
}

arpoly arpoly_power(arpoly p, expType exp, ring R)
{
  arpoly aux_p,aux_p2,aux_p3,res_p;
  if (exp < 1) return arpoly_one(R);
  if (exp == 1) return arpoly_dup(p,R);
  res_p=arpoly_one(R);
  aux_p=arpoly_dup(p,R);
  while (exp > 1) {
    if (exp & 1) {
      aux_p3=arpoly_mult(res_p,aux_p,R);
      arpoly_free(res_p,R);
      res_p=aux_p3; }
    exp /= 2;
    aux_p2=arpoly_square(aux_p,R);
    arpoly_free(aux_p,R);
    aux_p=aux_p2; }
  aux_p3=arpoly_mult(res_p,aux_p,R);
  arpoly_free(res_p,R);
  arpoly_free(aux_p,R);
  res_p=aux_p3;
  return res_p;
}

coc_bool arpoly_div(arpoly f, arpoly g, arpoly * q, arpoly * r, ring R)
{
  coeff aux_cg, aux_c;
  term aux_tg,aux_tr,aux_t;
  monomial aux_m;
  arpoly aux_p;

  if (g == NULL) return FALSE;
  if (f == NULL) {
    *q=NULL;
    *r=NULL;
    return TRUE; }

  if (arpoly_is_nonzero_const(g,R)) {
    aux_c=coeff_inverse(arpoly_this_leading_coeff(g,R),R);
    *q = arpoly_x_coeff(f,aux_c,R);
    *r = NULL;
    coeff_free(aux_c,R);
    return TRUE; }

  aux_cg=coeff_inverse(arpoly_this_leading_coeff(g,R),R);
  aux_tg=arpoly_leading_term(g,R);
  *r=arpoly_dup(f,R);
  *q=NULL;
  aux_tr=arpoly_leading_term(*r,R);

  while (term_divides(aux_tg,aux_tr)) {
    
    aux_t=term_div(aux_tr,aux_tg);
    coeff_set_mult(aux_c,aux_cg,arpoly_this_leading_coeff(*r,R),R);
    monomial_set_mk(aux_m,aux_c,aux_t);
    aux_p=arpoly_x_monomial(g,aux_m,R);
    *r=arpoly_sub_these(*r,aux_p,R);
    *q=arpoly_sum_these(*q,arpoly_this_monomial2arpoly(aux_m,R),R);

    term_free(aux_tr);
    aux_tr=term_NULL;

    if ((*r) != NULL) 
      aux_tr=arpoly_leading_term(*r,R);
    else break;
  }
  if (aux_tr != term_NULL) term_free(aux_tr); 
  coeff_free(aux_cg,R); 
  term_free(aux_tg);

  return TRUE;
}

arpoly arpoly_homog(arpoly p, unknown x, ring R)
{
  int d,d1;
  monomial aux_ml,aux_ml_next;
  arpoly res_p;
  d = arpoly_deg(p,R);
  aux_ml=arpoly2monopoli(p,R);
  res_p=NULL;
  while (aux_ml != NULL) {
    d1 = d-monomial_deg(aux_ml);
    aux_ml_next=monomial_next(aux_ml);
    if (d1 != 0) monomial_this_x_unknown_pow(aux_ml,x,d1);
    res_p=arpoly_sum_these(res_p,arpoly_this_monomial2arpoly(aux_ml,R),R);
    aux_ml=aux_ml_next;
  }
  return res_p;
}

pvcomp pvcomp_homog(pvcomp p, unknown x, ring R)
{
  int d,d1;
  monomial aux_ml,aux_ml_next;
  pvcomp res_p;
  d = pvcomp_deg(p,R);
  aux_ml=pvcomp2monopoli(p,R);
  res_p=NULL;
  while (aux_ml != NULL) {
    d1 = d-monomial_deg(aux_ml);
    aux_ml_next=monomial_next(aux_ml);
    if (d1 != 0) monomial_this_x_unknown_pow(aux_ml,x,d1);
    res_p=pvcomp_sum_these(res_p,this_monomial2pvcomp(aux_ml,R),R);
    aux_ml=aux_ml_next;
  }
  return res_p;
}

coc_bool arpoly_is_zero(arpoly p, ring R)
{
  return p==NULL;
}

coc_bool arpoly_is_one(arpoly p, ring R)
{
  coeff aux_c;
  if (!arpoly_is_nonzero_const(p,R)) return FALSE;
  aux_c=arpoly_this_leading_coeff(p,R);
  return coeff_is_one(aux_c,R);
}

arpoly arpoly_der_n(arpoly p, unknown x, int n, ring R)
{
  arpoly res_p,aux_p;
  int i;
  res_p=arpoly_der(p,x,R);
  for (i=2;i<=n && (res_p != NULL);i++) {
    aux_p=arpoly_der(res_p,x,R);
    arpoly_free(res_p,R);
    res_p=aux_p; }
  return res_p;
}

arpoly arpoly_der_nn(arpoly p, int * der_list, ring R)
{
  int m,n,i;
  unknown x;
  arpoly aux_p,res_p;
  m=der_list[0];
  if (m==0) return arpoly_dup(p,R);
  x=der_list[1];
  n=der_list[2];
  res_p=arpoly_der_n(p,x,n,R);
  for (i=2;i<=m && (res_p != NULL);i++) {
    x=der_list[i*2-1];
    n=der_list[i*2];
    aux_p=arpoly_der_n(res_p,x,n,R);
    arpoly_free(res_p,R);
    res_p=aux_p; }
  return res_p;
}

void printf_arpoly(arpoly p,ring R)
{
  cterm aux_t;
  aux_t=arpoly2cterm(p,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

void printf_pvcomp(pvcomp p,ring R)
{
  cterm aux_t;
  aux_t=pvcomp2cterm(p,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

/* --------------[ TODO_PUBLIC ]----------------- */

#if 0 /* TODO_PUBLIC */

polynomial this_arbrat2poly(arbrat r, ring R)
{
  polynomial res_p,aux_p;
  poly_div(this_arbint2poly(arbrat_num(r),R),
    this_arbint2poly(arbrat_den(r),R),&res_p,&aux_p,R);
  poly_free(aux_p,R);
  return res_p;
}

void poly_make_monic(polynomial * p, ring R)
{
  coeff aux_c;
  aux_c=poly_leading_coeff(*p,R);
  R->InvertCoeff(&aux_c,aux_c,R);
  *p=poly_this_x_coeff(*p,aux_c,R);
  R->ClearCoeff(&aux_c);
}

#endif /* TODO_PUBLIC */
