/* $Id: pp.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */
/* [ HANDLERS ] */

#include "cocdefs.h"

#include <stdio.h>
#include <string.h>

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "map.h"
#include "options.h"
#include "operator.h"
#include "pp.h"
#include "utils.h"

typedef int pp_handler(cdev d, cterm t);

pp_handler * pp_handlers[MAX_NUM_OPERATORS];

int pp_indent=0;

#define pp_set_indent(indent) \
  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_INDENT)) { pp_indent=indent; }
#define pp_incr_indent(n) \
  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_INDENT)) { pp_indent+=n; }
#define pp_decr_indent(n) \
  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_INDENT)) { pp_indent-=n; }

int pp_verbatim(cdev d, cterm t);

int pp_space_if_notindent(cdev d)
{
  if (!flag_is_on(opt_panel_ENGINE,flag_ENGINE_INDENT)) cdev_putc(' ',d);
  return pp_NOERR;
}

void pp_init(void)
{
  int i;
  for (i=0;i<MAX_NUM_OPERATORS;i++) pp_handlers[i]=NULL;
  pp_handlers[op_ALL]=&pp_all;
  pp_handlers[op_AND]=&pp_and;
  pp_handlers[op_ASSIGN]=&pp_assignment;
  pp_handlers[op_AT]=&pp_at;
  pp_handlers[op_BIND]=&pp_bind;
  pp_handlers[op_COLON]=&pp_colon;
  pp_handlers[op_COCOLON]=&pp_cocolon;
  pp_handlers[op_COMMANDS]=&pp_commands;
  pp_handlers[op_COMPONENT]=&pp_comp;
  pp_handlers[op_COND]=&pp_cond;
  pp_handlers[op_CHANGE_SIGN]=&pp_change_sign;
  pp_handlers[op_DASHES]=&pp_dashes;
  pp_handlers[op_DOTS]=&pp_dots;
  pp_handlers[op_DEFINE_AND_USE_RING]=&pp_define_and_use_ring;
  pp_handlers[op_DEFINE_RING]=&pp_define_ring;
  pp_handlers[op_ERROR]=&pp_error;
  pp_handlers[op_IS_EQ]=&pp_is_eq;
  pp_handlers[op_IS_GE]=&pp_is_ge;
  pp_handlers[op_IS_GT]=&pp_is_gt;
  pp_handlers[op_IS_LE]=&pp_is_le;
  pp_handlers[op_IS_LT]=&pp_is_lt;
  pp_handlers[op_EEE]=&pp_eee;
  pp_handlers[op_EEE_LBL]=&pp_eee_lbl;
  pp_handlers[op_EVAL]=&pp_eval;
  pp_handlers[op_FOR]=&pp_for;
  pp_handlers[op_FOREACH]=&pp_foreach;
  pp_handlers[op_FRAC]=&pp_frac;
  pp_handlers[op_FUN_APPL]=&pp_fun_appl;
  pp_handlers[op_INDET_INFO]=&pp_indet_info;
  pp_handlers[op_INDET_DECL]=&pp_indet_decl;
  pp_handlers[op_IF]=&pp_if;
  pp_handlers[op_IDEAL]=&pp_ideal;
  pp_handlers[op_INDET]=&pp_indet;
  pp_handlers[op_LIST]=&pp_list;
  pp_handlers[op_LONG_INT]=&pp_coeff_ppd_NONE;
  pp_handlers[op_LONG_RAT]=&pp_coeff_ppd_NONE;
  pp_handlers[op_MAT]=&pp_mat;
  pp_handlers[op_MODULE]=&pp_module;
  pp_handlers[op_MULT]=&pp_mult;
  pp_handlers[op_MEM_STAT]=&pp_mem_stat;
  pp_handlers[op_MONOMIAL]=&pp_monomial_call;
  pp_handlers[op_NOINDENT]=&pp_noindent;
  pp_handlers[op_NBS]=&pp_NBS;
  pp_handlers[op_NEWLINE]=&pp_newline;
  pp_handlers[op_OR]=&pp_or;
  pp_handlers[op_PARENTHESIS]=&pp_par;
  pp_handlers[op_POLYNOMIAL]=&pp_polynomial;
  pp_handlers[op_PANEL]=&pp_panel;
  pp_handlers[op_POLY_RING]=&pp_poly_ring;
  pp_handlers[op_POWER]=&pp_pow1;
  pp_handlers[op_PROGRAM]=&pp_define;
  pp_handlers[op_PROGRAM2]=&pp_define2;
  pp_handlers[op_RING_ZP]=&pp_ring_zp;
  pp_handlers[op_RATFUN]=&pp_ratfun;
  pp_handlers[op_REPEAT]=&pp_repeat;
  pp_handlers[op_RECORD]=&pp_record;
  pp_handlers[op_RECORDSEL]=&pp_record_sel;
  pp_handlers[op_RANGE]=&pp_range;
  pp_handlers[op_SHORT_INT]=&pp_coeff_ppd_NONE;
  pp_handlers[op_SOURCE]=&pp_source;
  pp_handlers[op_STRING]=&pp_string;
  pp_handlers[op_SUM]=&pp_sum;
  pp_handlers[op_SUBTRACT]=&pp_subtract;
  pp_handlers[op_SHORT_RAT]=&pp_coeff_ppd_NONE;
  pp_handlers[op_SHORT_DOUBLE]=&pp_coeff_ppd_NONE;
  pp_handlers[op_TERROR]=&pp_terror;
  pp_handlers[op_TERM]=&pp_term;
  pp_handlers[op_USING]=&pp_using;
  pp_handlers[op_WHILE]=&pp_while;
  pp_handlers[op_VAR]=&pp_var;
  pp_handlers[op_ZP]=&pp_coeff_ppd_NONE;
  pp_handlers[op_ZPVALUE]=&pp_zpvalue;
}

int pp_cterm_noindent(cdev d, cterm t)
{
  int n;
  if (flag_is_on(opt_panel_ENGINE, flag_ENGINE_INDENT)) {
    flag_unset(opt_panel_ENGINE, flag_ENGINE_INDENT);
    n=pp_cterm(d,t);
    flag_set(opt_panel_ENGINE, flag_ENGINE_INDENT);
    return n; }
  else
    return pp_cterm(d,t);
}

int pp_cterm_main(cdev d, cterm t, int indent)
{
  if (t == cterm_NULL) return 1; /* TODO: scegliere errore */
  pp_set_indent(indent);
  if (cterm_op(t)==op_STRING) 
    cdev_puts(cterm_string(cterm_param(t,0)),d);
  else 
    pp_cterm(d,t);
  cdev_flush(d);
  return 0;
}

int pp_nl_indent(cdev d)
{
  if (flag_is_on(opt_panel_ENGINE,flag_ENGINE_INDENT)) {
    cdev_puts("\n",d);
    pp_cdev_puts_spaces(d,pp_indent); }
}

int pp_cdev_puts_spaces(cdev d, int n)
{
  int i;
  static char s[200];
  for (i=0;i<n;i++) s[i]=' ';
  s[n]='\0';
  cdev_puts(s,d);
}

int pp_cterm(cdev d, cterm t)
{
  coop op;
  int pkg_n;
  if (t==cterm_NULL) { cdev_puts("",d); return 0; }

  op=cterm_op(t);
  pkg_n=op_package(op);
  op=op_index(op);

  if (pkg_n == pkg_builtin_n) {
    pp_handler * h;
    h = pp_handlers[op];
    if (h == NULL) pp_verbatim(d,t);
    else (*h)(d,t); }

  else pp_verbatim(d,t);
  return 0;
}

int pp_verbatim(cdev d, cterm t)
{
  int i,pkg_n;
  coop op;
  op=cterm_op(t);
  pkg_n=op_package(op);
  op=op_index(op);
  if (pkg_n!=pkg_builtin_n && pkg_n != pkg_std_n && pkg_n != pkg_user_n) {
    cdev_puts((char *)pkg_name(pkg_n),d);
    cdev_puts(".",d);
  }
  cdev_puts((char *)op_map_op_name(pkg_n, op),d);
  if (cterm_arity(t)>0) {
    cdev_puts("(",d);
    pp_cterm_noindent(d,cterm_param(t,0));
    for (i=1;i<cterm_arity(t);i++) {
      cdev_puts(", ",d); 
      pp_cterm_noindent(d,cterm_param(t,i)); }
    cdev_puts(")",d); }
}

/* ---[ HANDLERS ]--------------------------------------------- */

int pp_eee(cdev d, cterm t)
{
  char * prefix;
  int arity=cterm_arity(t);
  if (arity == 0) 
    cdev_puts("0",d);
  else
    pp_eee_prefix(d, "", t);
}

int pp_eee_mon_noprefix(cdev d, char * prefix, char c, cterm t);

int pp_eee_prefix(cdev d, char * prefix, cterm t)
{
  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  register int i;
  if (cterm_op(t) != op_EEE) {
    cdev_puts("> ",d);
    pp_cterm(d,t);
    return 0; }
  if (arity==1) {
    pp_eee_mon_noprefix(d, prefix, ' ', params[0]);
    return 0; }
  pp_eee_mon_noprefix(d, prefix, '|', params[0]);
  for (i=1;i<arity-1;i++) {
    cdev_puts("\n",d);
    cdev_puts(prefix,d);
    cdev_puts("|\n",d);
    pp_eee_mon(d, prefix, '|', params[i]); }
  cdev_puts("\n",d);
  cdev_puts(prefix,d);
  cdev_puts("|\n",d);
  pp_eee_mon(d, prefix, ' ', params[arity-1]);
}

int pp_eee_lbl(cdev d, cterm t)
{
#ifdef DEBUG
  int i;
  if (opt_get_value(opt_panel_DEBUG,opt_DEBUG_EEE_PP) ==
    optv_DEBUG_EEE_PP_GRAPH2)
      for (i=1;i<=cterm_arity(t);i++) cdev_putc('O',d); 
  else
#endif
  pp_list(d,t);
}

int pp_eee_mon_noprefix(cdev d, char * prefix, char c, cterm t)
{
  int i,n;
  char * s;
  cterm_list params=cterm_params(t);
  cdev_puts("|--",d);
  pp_cterm(d,params[0]);
  cdev_puts("--",d);
  n=cterm_arity(params[0])+5;
  s=malloc_str(n);
  s[0]=c;
  for (i=1;i<n;i++) s[i]=' ';
  s[n]='\0';
  prefix=my_strsum(prefix,s);
  pp_eee_prefix(d,prefix,params[1]);
  free_str1(s);
  free_str1(prefix);
}

int pp_eee_mon(cdev d, char * prefix, char c, cterm t)
{
  int i,n;
  char * s;
  cterm_list params=cterm_params(t);
  cdev_puts(prefix,d);
  cdev_puts("|--",d);
  pp_cterm(d,params[0]);
  cdev_puts("--",d);
  n=cterm_arity(params[0])+5;
  s=malloc_str(n);
  s[0]=c;
  for (i=1;i<n;i++) s[i]=' ';
  s[n]='\0';
  prefix=my_strsum(prefix,s);
  pp_eee_prefix(d,prefix,params[1]);
  free_str1(prefix);
  free_str1(s);
}

int pp_record_sel(cdev d, cterm t) { pp_bop(d,t,"."); } 
int pp_colon(cdev d, cterm t) { pp_bop(d,t," : "); }
int pp_bind(cdev d, cterm t) { pp_bop(d,t," = "); }
int pp_mult(cdev d, cterm t) { pp_bop(d,t," * "); }
int pp_sum(cdev d, cterm t) { pp_bop(d,t," + "); }
int pp_zpvalue(cdev d, cterm t) { pp_bop(d,t," % "); }
int pp_or(cdev d, cterm t) { pp_bop(d,t," OR "); }
int pp_and(cdev d, cterm t) { pp_bop(d,t," AND "); }
int pp_at(cdev d, cterm t) { pp_bop(d,t,"@"); }
int pp_cocolon(cdev d, cterm t) { pp_bop(d,t," :: "); }
int pp_subtract(cdev d, cterm t) { pp_bop(d,t," - "); }
int pp_frac(cdev d, cterm t) { pp_bop(d,t," / "); }
int pp_pow1(cdev d, cterm t) { pp_bop(d,t,"^"); }
int pp_range(cdev d, cterm t) { pp_bop(d,t,".."); }

int pp_define_ring(cdev d, cterm t) {
  int i;
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(" ::= ",d);
  pp_cterm(d,cterm_param(t,1));
  t=cterm_param(t,2);
  for (i=0;i<cterm_arity(t);i++) {
    cdev_puts(",",d);
    pp_cterm(d,cterm_param(t,i));
  }
}

int pp_is_eq(cdev d, cterm t) { pp_bop(d,t," = "); }
int pp_is_le(cdev d, cterm t) { pp_bop(d,t," <= "); }
int pp_is_lt(cdev d, cterm t) { pp_bop(d,t," < "); }
int pp_is_ge(cdev d, cterm t) { pp_bop(d,t," >= "); }
int pp_is_gt(cdev d, cterm t) { pp_bop(d,t," > "); }
int pp_change_sign(cdev d, cterm t) { pp_uop(d,t,"-"); }

int pp_cdev_puts_1(char * s, cdev d)
{
  if (!flag_is_on(opt_panel_ENGINE, flag_ENGINE_INDENT)) {
    cdev_puts(" ",d);
  }
  cdev_puts(s,d);
}

int pp_ratfun(cdev d, cterm t) {
  pp_polynomial_par(d,TRUE,cterm_param(t,0));
  cdev_puts("/",d);
  pp_polynomial_par(d,FALSE,cterm_param(t,1));
  return 0;
}

int pp_if(cdev d, cterm t)
{
  cterm_list params;
  int arity,i;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("If ",d);
  pp_cterm(d,params[0]);
  cdev_puts(" Then ",d);
  pp_cterm_indent(d,params[1],2);

  for (i=3; i<arity; i+=2) {
    pp_cdev_puts_1("ElsIf ",d);
    pp_cterm(d,params[i-1]);
    cdev_puts(" Then ",d);
    pp_cterm_indent(d,params[i],2);
  }

  if (arity==i) {
    pp_cdev_puts_1("Else ",d);
    pp_cterm_indent(d,params[arity-1],2); }
  pp_cdev_puts_1("EndIf",d);
}

int pp_cond(cdev d, cterm t)
{
  cterm_list params;
  int arity,i;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("Cond ",d);
  pp_cterm(d,params[0]);
  cdev_puts(" Then ",d);
  pp_cterm_indent(d,params[1],2);

  for (i=3; i<arity; i+=2) {
    pp_cdev_puts_1("ElsIf ",d);
    pp_cterm(d,params[i-1]);
    cdev_puts(" Then ",d);
    pp_cterm_indent(d,params[i],2);
  }

  if (arity==i) {
    pp_cdev_puts_1("Else ",d);
    pp_cterm_indent(d,params[arity-1],2); }
  pp_cdev_puts_1("EndCond",d);
}

int pp_for(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("For ",d);
  pp_cterm(d,params[0]);
  cdev_puts(" := ",d);
  pp_cterm(d,params[2]);
  cdev_puts(" To ",d);
  pp_cterm(d,params[3]);
  if (arity==5) {
    cdev_puts(" Step ",d);
    pp_cterm(d,params[4]);
  }
  cdev_puts(" Do ",d);
  pp_cterm_indent(d,params[1],2);
  pp_cdev_puts_1("EndFor", d);
}

int pp_foreach(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("Foreach ",d);
  pp_cterm(d,params[0]);
  cdev_puts(" In ",d);
  pp_cterm_noindent(d,params[1]);
  cdev_puts(" Do ",d);
  pp_cterm_indent(d,params[arity-1],2);
  pp_cdev_puts_1("EndForeach", d);
}

int pp_catch(cdev d, cterm t)
{
  cdev_puts("Catch ",d);
  pp_cterm_indent(d,cterm_param(t,0),2);
  if (cterm_arity(t) == 2) {
    pp_cdev_puts_1("In ",d);
    pp_cterm(d,cterm_param(t,1)); }
  pp_cdev_puts_1("EndCatch", d);
}

int pp_while(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("While ",d);
  pp_cterm(d,params[0]);
  pp_cdev_puts_1(" Do ",d);
  pp_cterm_indent(d,params[arity-1],2);
  pp_cdev_puts_1("EndWhile", d);
}

int pp_using(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("Using ",d);
  pp_cterm(d,params[0]);
  pp_cdev_puts_1(" Do ",d);
  pp_cterm_indent(d,params[arity-1],2);
  pp_cdev_puts_1("EndUsing", d);
}

int pp_repeat(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);
  cdev_puts("Repeat ",d);
  pp_cterm_indent(d,params[0],2);
  cdev_puts("Until ",d);
  pp_cterm(d,params[arity-1]);
}

int pp_commands(cdev d, cterm t)
{
  int i;
  if (cterm_arity(t)==0) return 0;
  pp_cterm(d,cterm_param(t,0));
  for (i=1;i<cterm_arity(t);i++) {
    cdev_puts("; ",d);
    if (cterm_op(cterm_param(t,i)) != op_SKIP) {
      pp_nl_indent(d);
      pp_cterm(d,cterm_param(t,i)); } }
  return 0;
}

int pp_params1(cdev d, int n)
{
  int i;
  if (n == NO_ARITY) {
    cdev_puts("(...)",d);
  } else if (n == 0) { 
    cdev_puts("()",d); 
  } else {
    char buffer[5];
    sprintf(buffer,"(X%d",1);
    cdev_puts(buffer,d);
    for (i=2; i<=n; i++) {
      sprintf(buffer,",X%d",i);
      cdev_puts(buffer,d);
    }
    cdev_puts(")",d);
  }
  return 0;
}

int pp_params(cdev d, cterm t)
{
  if (cterm_op(t) == op_DOTS) cdev_puts("(...)",d);
  else {
    cdev_puts("(",d);
    pp_cterms_sep(d,t,",");
    cdev_puts(")",d);
  }
  return 0;
}

int pp_define_aux(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);

  if (cterm_op(params[3]) == op_TRUE) { /* short form F(X) := E */
    pp_cterm(d, params[0]);
    pp_params(d,params[1]);
    pp_incr_indent(2);
    cdev_puts(" := ",d);
    pp_cterm(d,cterm_param(params[2],0)); /* params[2] = Return(Expr) */
    pp_decr_indent(2);
  } else {
    cdev_puts("Define ",d);
    pp_cterm(d, params[0]);
    if (cterm_op(params[2]) == op_BUILTIN) {
      pp_params1(d,cterm2int(params[1]));
      cdev_puts(" /* built in */ ",d);
    } else {
      pp_params(d,params[1]);
      if (cterm_op(params[arity-1]) != op_NULL_VALUE) {
	cdev_puts("Help ",d);
	pp_cterm(d,params[arity-1]);
      }
      pp_cterm_indent(d,params[2],2);
    }
    cdev_puts("EndDefine;",d);
  }
}

int pp_define(cdev d, cterm t)
{
  int n;
  if ((cterm_op(cterm_param(t,3)) == op_FALSE ) /* not F(X) := E */
      && !flag_is_on(opt_panel_ENGINE, flag_ENGINE_INDENT)) {
    flag_set(opt_panel_ENGINE, flag_ENGINE_INDENT);
    n=pp_define_aux(d,t);
    flag_unset(opt_panel_ENGINE, flag_ENGINE_INDENT);
    return n; }
  else
    return pp_define_aux(d,t);
}

int pp_define2(cdev d, cterm t)
{
  cterm_list params;
  int arity;
  arity=cterm_arity(t);
  params=cterm_params(t);

  pp_cterm(d, params[0]);
  if (cterm_op(params[2]) == op_BUILTIN) {
    pp_params1(d,cterm2int(params[1]));
  } else {
    pp_params(d,params[1]);
  }
}

int pp_source(cdev d, cterm t) {
  cdev_puts("<< ",d);
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(";\n",d);
  return 0;
}

int pp_mat(cdev d, cterm t)
{
  /* Mat(LIST(LIST)) */
  cterm aux_t;
  cterm_list params;
  int arity;
  register int i;
  if (cterm_arity(t) == 0) {
    cdev_puts("Mat()",d);
    return 0;
  }
  params=cterm_params(t);
  aux_t=params[0];
  params=cterm_params(aux_t);
  arity=cterm_arity(aux_t);

  cdev_puts("Mat([",d);
  if (arity>0) {
    cdev_puts("\n",d);
    pp_cdev_puts_spaces(d,pp_indent+2);
    pp_cterm(d,params[0]);
    for (i=1;i<arity;i++) {
      cdev_puts(",\n",d);
      pp_cdev_puts_spaces(d,pp_indent+2);
      pp_cterm(d,params[i]); }
    cdev_puts("\n",d);
    pp_cdev_puts_spaces(d,pp_indent); }
  cdev_puts("])",d);
  return 0;
}

int pp_ideal(cdev d, cterm t)
{
  int i;
    cdev_puts("Ideal(",d);
    pp_incr_indent(2);
    if (cterm_arity(t)>0) {
      pp_nl_indent(d);
      pp_cterm(d,cterm_param(t,0));
      for (i=1;i<cterm_arity(t);i++) {
        cdev_putc(',',d);
        pp_space_if_notindent(d);
        pp_nl_indent(d);
        pp_cterm(d,cterm_param(t,i)); } }
    pp_decr_indent(2);
    cdev_puts(")",d);
}

int pp_module(cdev d, cterm t)
{
  int i;
    cdev_puts("Module(",d);
    pp_incr_indent(2);
    if (cterm_arity(t)>0) {
      pp_nl_indent(d);
      pp_cterm(d,cterm_param(t,0));
      for (i=1;i<cterm_arity(t);i++) {
        cdev_putc(',',d);
        pp_space_if_notindent(d);
        pp_nl_indent(d);
        pp_cterm(d,cterm_param(t,i)); } }
    pp_decr_indent(2);
    cdev_puts(")",d);
}


int pp_comp(cdev d, cterm t)
{
  pp_cterm(d,cterm_param(t,0));
  cdev_puts("[",d);
  pp_cterm(d,cterm_param(t,1));
  cdev_puts("]",d);
}

int pp_noindent(cdev d, cterm t)
{
  return pp_cterm_noindent(d,cterm_param(t,0));
}

int pp_mem_stat(cdev d, cterm t)
{
  char buffer[100];
  sprintf(buffer,"--> Memory Used: %s bytes: (Max.: %s bytes).\n",
    cterm_string(cterm_param(t,1)),cterm_string(cterm_param(t,0)));
  cdev_puts(buffer,d);
}

int pp_all(cdev d, cterm t)
{
  cterm_list params;
  params=cterm_params(t);
  cdev_puts("[",d);
  pp_cterm(d,params[3]);
  cdev_puts("|",d);
  pp_cterm(d,params[0]);
  cdev_puts(" In ", d);
  pp_cterm(d,params[1]);
  if (cterm_op(params[2]) != op_TRUE) {
    cdev_puts(" And ", d);
    pp_cterm(d, params[2]); }
  cdev_puts("]",d);
}

int pp_panel(cdev d, cterm t)
{
  char * descr;
  int n,l,i;
  cterm_list params,aux_params;
  params=cterm_params(t);
  cdev_puts("\n",d);
  n=cterm_arity(t)-1;
  l=0;
  for (i=1;i<=n;i++) {
    aux_params=cterm_params(params[i]);
    l=MAX(l,(int)strlen(cterm_string(aux_params[1]))); }
  for (i=1;i<=n;i++) pp_option(d, params[i], i, l);
}

int pp_ring_zp(cdev d, cterm t)
{
  cdev_puts("Z/(",d);
  if (cterm_arity(t) != 0) pp_cterm(d,cterm_param(t,0));
  cdev_puts(")",d);
}

int pp_eval(cdev d, cterm t)
{ /* t=eval(e:expr,r:ring_name) */
  if (cterm_arity(t)==2) {
    pp_cterm(d,cterm_param(t,1));
    cdev_puts(" :: ",d); }
  pp_cterm(d,cterm_param(t,0));
}

int pp_list(cdev d, cterm t)
{
  register int i;
  if (cterm_arity(t)==0) {cdev_puts("[ ]",d); return pp_NOERR; }
  cdev_puts("[",d);
  pp_incr_indent(2);
  if (cterm_arity(t)>0) {
    pp_nl_indent(d);
    pp_cterm(d,cterm_param(t,0));
    for (i=1;i<cterm_arity(t);i++) {
      cdev_putc(',',d);
      pp_space_if_notindent(d);
      pp_nl_indent(d);
      pp_cterm(d,cterm_param(t,i)); } }
  pp_decr_indent(2);
  cdev_puts("]",d);
  return 0;
}

int pp_term(cdev d, cterm t)
{
  register int i;
  if (cterm_op(t)!=op_TERM) return 0;
  for (i=0;i<cterm_arity(t);i++) pp_pow(d,cterm_param(t,i));
  return 0;
}

int pp_par(cdev d, cterm t)
{
  cdev_puts("(",d);
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(")",d);
  return 0;
}

int pp_define_and_use_ring(cdev d, cterm t)
{
  cdev_puts("Use ",d);
  return pp_define_ring(d,t);
}

int pp_record(cdev d, cterm t)
{ /* t=record(s1:bind,...,sn:bind) */
  register int i;
  if (cterm_arity(t)==0) {cdev_puts("Record[ ]",d); return pp_NOERR; }
  cdev_puts("Record[",d);
  pp_incr_indent(2);
  if (cterm_arity(t)>0) {
    pp_nl_indent(d);
    pp_cterm(d,cterm_param(t,0));
    for (i=1;i<cterm_arity(t);i++) {
      cdev_putc(',',d);
      pp_space_if_notindent(d);
      pp_nl_indent(d);
      pp_cterm(d,cterm_param(t,i)); } }
  pp_decr_indent(2);
  cdev_puts("]",d);
  return 0;
}

int pp_var(cdev d, cterm t)
{
  char *s, *p;
  t=cterm_param(t,0);
  if (cterm_op(t) == op_NBS) {
    char buffer[100];
    s=cterm_string(t);
    if (p=strchr(s,'#')) { /* = and not == */
      strncpy(buffer,s,p-s);
      buffer[p-s]='\0';
      cdev_puts(buffer,d); }
    else 
      cdev_puts(s,d);
  } else {
    cdev_puts("Var(",d);
    pp_cterm(d,t);
    cdev_puts(")",d);
  }
}

int pp_fun_appl(cdev d, cterm t)
{
  int i;
  pp_cterm(d,cterm_param(t,0));
  if (cterm_arity(t)==3) {
    cdev_puts(".",d);
    pp_cterm(d,cterm_param(t,1));
    t=cterm_param(t,2); }
  else t=cterm_param(t,1);
  cdev_puts("(",d);
  if (cterm_arity(t)>0) {
    pp_cterm_noindent(d,cterm_param(t,0));
    for (i=1;i<cterm_arity(t);i++) {
      cdev_puts(",",d);
      pp_cterm_noindent(d,cterm_param(t,i)); } }
  cdev_puts(")",d);
}

int pp_indet(cdev d, cterm t)
{ /* t=indet(id:lower_ident,i1:expr,...,in:expr) */
  register int i;
  cdev_puts(cterm_string(cterm_param(t,0)),d);
  if (cterm_arity(t)>1) {
    cdev_puts("[",d);
    pp_cterm(d,cterm_param(t,1));
    for (i=2;i<cterm_arity(t);i++) {
      cdev_puts(",",d);
      pp_cterm(d,cterm_param(t,i)); }
    cdev_puts("]",d); }
}

coc_bool pp_is_pure_power(cterm t) {
/* COCOA USED TO CRASH ON: Use Z/(6)[x];Poly(2)^(-1);              */
/* NORMALLY RATFUNS HAVE MONIC DENOMINATOR; NOT POSS. IN THIS CASE */
/* IT NO LONGER CRASHES BUT PRINTS BADLY: 1/(2x^2) prints as 1/2x^2*/
/* THIS CAN OCCUR ONLY IF THE DENOM CANNOT BE MADE MONIC AND IS A PURE POWER */
/* (2x^2)^(-1) CAN GENERATE THIS: SEE mi_power, and do_inv         */  
  /* assumed t = POLY(...) */
  if (cterm_arity(t) > 1) return FALSE;                /* sum of >=2 terms */
  if (cterm_arity(cterm_param(t,0)) == 1) return TRUE; /* just a coeff     */
  return cterm_arity(cterm_param(cterm_param(t,0),1))<=1;
}

int pp_polynomial_par(cdev d, coc_bool is_num, cterm t)
{
  if ((is_num && cterm_arity(t) <= 1) || pp_is_pure_power(t))
    return pp_polynomial(d,t);
  else {
    cdev_puts("(",d);
    pp_polynomial(d,t);
    cdev_puts(")",d);
    return pp_NOERR;
  } 
}

int pp_monomial_call(cdev d, cterm t)
{
  return pp_monomial(d, t, ppd_NONE);
}

int pp_polynomial(cdev d, cterm t)
{
  register int i;
  if (cterm_arity(t)==0) cdev_puts("0",d);
  else {
    pp_monomial(d, cterm_param(t,0), ppd_NONE);
    for (i=1; i<cterm_arity(t); i++) {
      cdev_putc(' ',d);
      pp_monomial(d, cterm_param(t,i), ppd_WITH_PLUS);
    }
  }
}

int pp_number1(cdev d, cterm t, pp_directive ppd)
{
  pp_number(d,cterm_string(cterm_param(t,0)),ppd);
}

int pp_rat(cdev d, cterm t, pp_directive ppd)
{
  pp_number(d,cterm_string(cterm_param(t,0)),ppd);
  if (!pp_is_one_abs_number(cterm_string(cterm_param(t,1)))) {
    cdev_puts("/",d);
    pp_number(d,cterm_string(cterm_param(t,1)),ppd); }
}

int pp_zp(cdev d, cterm t, pp_directive ppd)
{
  int p,n;
  char buffer[30];
  p = atoi(cterm_string(cterm_param(t,0)));
  n = atoi(cterm_string(cterm_param(t,1)));
  if (n > p/2) {
    if (ppd_is_WITH_PLUS(ppd)) {
      sprintf(buffer,"- %d",p-n);
    } else {
      sprintf(buffer,"-%d",p-n);
    }
  } else sprintf(buffer,"%d",n);
  cdev_puts(buffer,d);
}

int pp_coeff_ppd_NONE(cdev d, cterm t)
{
  return pp_coeff(d, t, ppd_NONE);
}

int pp_NBS(cdev d, cterm t)
{
  return cdev_puts(cterm_string(t),d);
}

int pp_indet_info(cdev d, cterm t)
{
  int i;
  if (cterm_arity(t)==0) return pp_NOERR;
  pp_cterm(d,cterm_param(t,0));
  for (i=1;i<cterm_arity(t);i++) {
    cdev_puts(",",d);
    pp_cterm(d,cterm_param(t,i)); }
  return 0;
}

int pp_indet_decl(cdev d, cterm t)
{
  int i;
  cdev_puts(cterm_string(cterm_param(t,0)),d);
  if (cterm_arity(t)>1) {
    cdev_puts("[",d); 
    pp_cterm(d,cterm_param(t,1));
    for (i=2;i<cterm_arity(t);i++) {
      cdev_puts(",",d);
      pp_cterm(d,cterm_param(t,i)); }
    cdev_puts("]",d); }
}

int pp_poly_ring(cdev d, cterm t)
{
  cterm_list params;
  int i;
  pp_cterm(d,cterm_param(t,1));
  cdev_puts("[",d);
  if (cterm_op(cterm_param(t,0))==op_LIST) {
    params=cterm_params(cterm_param(t,0));
    for (i=0;i<cterm_arity(cterm_param(t,0));i++) pp_cterm(d,params[i]); }
  else pp_cterm(d,cterm_param(t,0));
  cdev_puts("]",d);
  if (cterm_arity(t)==3) {
    cdev_puts(",",d);
    pp_cterm(d,cterm_param(t,2)); }
}

int pp_asis_param0(cdev d, cterm t)
{
  return cdev_puts(cterm_string(cterm_param(t,0)),d);
}

int pp_string(cdev d, cterm t)
{
  cdev_puts("\"",d);
  cdev_puts(cterm_string(cterm_param(t,0)),d);
  return cdev_puts("\"",d);
}

int pp_newline(cdev d, cterm t)
{
  return cdev_puts("\n",d);
}

int pp_dashes(cdev d, cterm t)
{
  return cdev_puts("-------------------------------\n",d);
}

int pp_dots(cdev d, cterm t)
{
  return cdev_puts("...",d);
}

int pp_assignment(cdev d, cterm t)
{
  cterm_list params=cterm_params(t);
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(" := ",d);
  pp_cterm(d,cterm_param(t,1));
}

int pp_error(cdev d, cterm t)
{
  char * aux_s;
  aux_s = cterm_string(cterm_param(t,0));
  /*  ATTENTION:  The string below comes from the file cocmisc.coc! */
  if (strcmp(aux_s, "Quitting...") == 0) return 0;
  cdev_puts("ERROR: ",d);
  cdev_puts(aux_s, d);
  if (error_param != cterm_NULL) {
    cdev_puts(" ", d);
    pp_cterm_main(d, error_param, strlen(aux_s)); }
  if (!flag_is_on(opt_panel_ENGINE,flag_ENGINE_NO_ERROR_CONTEXT) && error_context != cterm_NULL && error_context != cterm_unknown_context) {
    cdev_puts("\nCONTEXT: ",d);
    pp_cterm_main(d, error_context, 13); }
}

int pp_terror(cdev d, cterm t)
{ /* TERROR("error", linenumber, device) */
  char buffer[200];

  sprintf(buffer,"ERROR: %s in line %s of device %s", 
    cterm_string(cterm_param(t,0)), 
    cterm_string(cterm_param(t,1)), 
    cterm_string(cterm_param(t,2)));
  cdev_puts(buffer, d);
}


/* ---[ HANDLERS auxiliary functions ]------------------------- */

int pp_cterm_indent(cdev d, cterm t, int n)
{
  pp_incr_indent(n);
  pp_nl_indent(d);
  pp_cterm(d,t);
  pp_decr_indent(2);
  pp_nl_indent(d);
}

int pp_cterms_sep(cdev d, cterm t, char * sep)
{
  cterm_list params;
  int arity,i;
  arity=cterm_arity(t);
  params=cterm_params(t);
  if (arity==0) return pp_NOERR;
  pp_cterm(d, params[0]);
  for (i=1;i<arity;i++) {
    cdev_puts(sep,d);
    pp_cterm(d, params[i]); }
  return pp_NOERR;
}

int pp_coeff(cdev d, cterm t, pp_directive ppd)
{
  coop op=cterm_op(t);
  if (ppd_is_WITH_PLUS(ppd) && pp_is_positive_coeff(t)) cdev_puts("+ ",d);
  if (ppd_is_IF_NOT_ONE(ppd) && pp_is_one_abs_coeff(t) 
      && (!pp_is_positive_coeff(t))) {
    if (ppd_is_WITH_PLUS(ppd)) {
      cdev_puts("- ",d);
    } else {
      cdev_putc('-',d);
    }
  }
  else if (!(ppd_is_IF_NOT_ONE(ppd) && pp_is_one_abs_coeff(t))) {
    if (op==op_LONG_INT) return pp_number1(d,t,ppd);
    else if (op==op_SHORT_INT) return pp_number1(d,t,ppd);
    else if (op==op_LONG_RAT) return pp_rat(d,t,ppd);
    else if (op==op_SHORT_RAT) return pp_rat(d,t,ppd);
    else if (op==op_SHORT_DOUBLE) return pp_number1(d,t,ppd);
    else if (op==op_ZP) return pp_zp(d,t,ppd); }
  return pp_NOERR;
}

int pp_uop(cdev d, cterm t, char * sep)
{
  cdev_puts(sep,d);
  pp_cterm(d,cterm_param(t,0));
}

int pp_bop(cdev d, cterm t, char * sep)
{
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(sep,d);
  pp_cterm(d,cterm_param(t,1));
}

int pp_bop_plus(cdev d, cterm t, char * before, char * sep, char * after)
{
  cdev_puts(before,d);
  pp_cterm(d,cterm_param(t,0));
  cdev_puts(sep,d);
  pp_cterm(d,cterm_param(t,1));
  cdev_puts(after,d);
}

int pp_monomial(cdev d, cterm t, pp_directive ppd)
{
  int arity=cterm_arity(t);
  cterm_list params=cterm_params(t);
  if (cterm_op(t)==op_MONOMIAL) {
    if (arity==2) {
      pp_coeff(d,params[0],ppd_IF_NOT_ONE+ppd);
      pp_term(d,params[1]); }
    else pp_coeff(d,params[0],ppd);
  } 
}

int pp_number(cdev d, char * s, pp_directive ppd)
{
  if (s[0]=='-') {
    cdev_putc('-',d); 
    s++;
    if (ppd_is_WITH_PLUS(ppd)) {
      cdev_putc(' ',d);
    }
  }
  cdev_puts(s,d);
}

coc_bool pp_is_positive_coeff(cterm t)
{  
  coop op=cterm_op(t);
  cterm_list params=cterm_params(t);
  if (op==op_LONG_INT || op==op_SHORT_INT || op==op_LONG_RAT || 
      op==op_SHORT_RAT || op==op_SHORT_DOUBLE) 
    return pp_is_positive_number(cterm_string(params[0]));
  else if (op==op_ZP) 
    return pp_is_positive_zp(cterm_string(params[0]),cterm_string(params[1]));
}

coc_bool pp_is_one_abs_coeff(cterm t)
{  
  coop op=cterm_op(t);
  cterm_list params=cterm_params(t);
  if (op==op_LONG_INT || op==op_SHORT_INT || op==op_SHORT_DOUBLE) 
    return pp_is_one_abs_number(cterm_string(params[0]));
  else if (op==op_LONG_RAT || op==op_SHORT_RAT) 
    return pp_is_one_abs_number(cterm_string(params[0])) &&
           pp_is_one_abs_number(cterm_string(params[1]));
  else if (op==op_ZP) 
    return pp_is_one_abs_zp(cterm_string(params[0]),cterm_string(params[1]));
}

coc_bool pp_is_one_abs_zp(char * p,char *s)
{
  return (atoi(s)==atoi(p)-1)||(atoi(s)==1);
}

coc_bool pp_is_positive_zp(char * p, char * s)
{
  return atoi(s)<=(atoi(p)/2);
}

coc_bool pp_is_positive_number(char * s)
{
  return s[0]!='-';
}

coc_bool pp_is_one_abs_number(char * s)
{
  return ((strcmp(s,"1")==0)||(strcmp(s,"-1")==0));
}

int pp_option_aux(cdev d, char * descr, int idx, int value)
{
  if (idx==value) cdev_puts("(o) ",d); else cdev_puts("( ) ",d);
  cdev_puts(descr,d);
  cdev_puts("\n",d);
}

int pp_option(cdev d, cterm t, int o, int maxldescr)
{
  int i,ivalue,n,j;
  coop op;
  char * descr,*value;
  cterm_list params;
  params=cterm_params(t);
  descr=cterm_string(params[1]);
  cdev_puts(descr,d);
  for (i=1;i<=maxldescr-(int)strlen(descr);i++) cdev_puts(".",d);
  cdev_puts("... : ",d);
  op=cterm_op(params[0]);
  value=cterm_string(params[2]);
  ivalue=atoi(value);
  if (op==op_NUM) { cdev_puts(value,d); cdev_puts("\n",d); }
  else if (op==op_FLAG) {
    if (ivalue==TRUE) cdev_puts("TRUE",d); else cdev_puts("FALSE",d);
    cdev_puts("\n",d); }
  else {
    n=cterm_arity(params[3]);
    params=cterm_params(params[3]);
    pp_option_aux(d,cterm_string(params[0]),1,ivalue);
    for (i=1;i<n;i++) {
      for (j=1;j<=maxldescr+3;j++) cdev_puts(" ",d);
      cdev_puts(" : ",d);
      pp_option_aux(d,cterm_string(params[i]),i+1,ivalue); }
  }
} 
