/* pseries.c : 0.2 : 1 Feb 94 : ADT poincare series */

/*------------------------------------------------------------------------------

	part of project cocoa

	revision history

------------------------------------------------------------------------------*/
 
#include "cocdefs.h"

#include <stdio.h>
#include <stdlib.h>

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "operator.h"
#include "olist.h"
#include "object.h"
#include "options.h"
#include "monomial.h"
#include "polys.h"
#include "pseries.h"
#include "ring.h"
#include "unipoly.h"
#include "utils.h"

PSeries pseries_init()
{
  PSeries res_PS;

  res_PS = malloc_pseries();
  res_PS->PolyNum = NULL;

  return res_PS;
}

void pseries_free (PSeries PS)
{
  if (PS!=NULL) 
  {
    FreeUnipoly(PS->PolyNum);
    olist_free(pseries_den_factors(PS));
    free_pseries(PS); 
  }
}

PSeries pseries_dup (PSeries PS)
{
  PSeries res_PS=NULL;

  if (PS!=NULL) 
  {
    res_PS = pseries_init();
    res_PS->PolyNum = unipoly_dup(PS->PolyNum);
    pseries_set_den_factors(res_PS,olist_dup(pseries_den_factors(PS)));
  }
  return res_PS;
}

cterm pseries2cterm (PSeries PS)
{
  return olist2cterm (pseries2olist(PS));
}

olist pseries2olist (PSeries PS)
{
  olist ol;
  object aux_o;

  ol = olist_init(2);
  aux_o = olist2object(unipoly2sppnum(pseries_num(PS)));
  object_set_label(aux_o, my_strdup("SpPoly"));
  olist_set_nth(ol, 1, aux_o);
  aux_o = olist2object(olist_dup(pseries_den_factors(PS)));
  olist_set_nth(ol,2,aux_o);
  return ol;
}

PSeries olist2pseries (olist ol)
{
  olist aux_ol;
  PSeries res_PS = pseries_init();

  aux_ol = object2olist(object_link(olist_nth(ol,1)));
  pseries_set_num(res_PS, sppnum2unipoly(aux_ol));
  olist_free(aux_ol); /* MARK-12 */
  pseries_set_den_factors(res_PS,object2olist(object_link(olist_nth(ol,2))));

  return res_PS;
}

olist MVPSeries2olist (MVPSeries PS)
{
  olist ol;
  object aux_o;

  ol = olist_init(2);
  aux_o = olist2object(arpoly2sppnum(MVPSeries_num(PS), MVPSeries_ring(PS)));
  object_set_label(aux_o, my_strdup("SpPoly"));
  olist_set_nth(ol, 1, aux_o);
  aux_o = olist2object(MVPSeries_den_factors(PS));
  olist_set_nth(ol,2,aux_o);
  free_MVPSeries(PS);
  return ol;
}

MVPSeries MVPSeries_init(ring R)
{
  MVPSeries res_PS;

  res_PS = malloc_MVPSeries();
  res_PS->PolyNum = NULL;
  res_PS->MVPRing = R;
  MVPSeries_set_den_factors(res_PS,NULL);

  return res_PS;
}

void MVPSeries_free (MVPSeries PS)
{
  if (PS!=NULL) 
  {
    arpoly_free(PS->PolyNum, PS->MVPRing);
    olist_free(MVPSeries_den_factors(PS));
    free_MVPSeries(PS); 
  }
}

MVPSeries MVPSeries_dup (MVPSeries PS)
{
  MVPSeries res_PS = NULL;

  if (PS!=NULL) 
  {
    res_PS = MVPSeries_init(MVPSeries_ring(PS));
    res_PS->PolyNum = arpoly_dup(PS->PolyNum, MVPSeries_ring(PS));
    MVPSeries_set_den_factors(res_PS, MVPSeries_den_factors(PS));
    MVPSeries_set_ring(res_PS, MVPSeries_ring(PS));
  }
  return res_PS;
}
