/* $Id: pvector.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ vector of polys

[ shifts ] [ pvector ]
---------------------------------------------------------------------------- */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "cterm.h"
#include "ctbuf.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "options.h"
#include "polys.h"
#include "pvector.h"
#include "terms.h"
#include "ring.h"
#include "weights.h"

#ifdef DEBUG_SHIFTS_RC

shifts shifts_link(shifts ss)
{
  return ((ss) == shifts_NULL ? NULL : (shifts_incr_rc(ss),ss));
}

void shifts_unlink(shifts ss)
{
  if ((ss)!=shifts_NULL) {
    shifts_decr_rc(ss);
    if (shifts_rc_is_zero(ss)) shifts_free(ss);
  }
}

#endif


/* ---[ shifts ]--------------------------------------- */

shifts shifts_init(int r)
{
  shifts res_ss;
  res_ss=malloc_shifts(r);
  shifts_set_rank(res_ss,r);
  shifts_set_rc(res_ss,1);
  shifts_set_ss(res_ss,malloc_term_list(r));
  return res_ss;
}

cterm shift2cterm(shifts ss, int n, ring R)
{
  shift s;
  if (shifts_are_int(ss)) return int2cterm(shift_get_int(shifts_nth(ss,n)));
  else {
    s=shifts_nth(ss,n);
    if (s==term_NULL || term_degree(s)==0) return int2cterm(0);
    return term2cterm(s,R);
  }
}

cterm shifts2cterm(shifts ss, ring R)
{
  register int i;
  cterm params[1];
  ctbuf aux_b;
  if (ss==shifts_NULL) return cterm_mk_const(op_LIST);
  aux_b=ctbuf_empty();
  for (i=1;i<=shifts_rank(ss);i++)
    ctbuf_append(aux_b,shift2cterm(ss,i,R));
  params[0]=ctbuf_mk_cterm(aux_b,op_LIST);
  return cterm_noindent(cterm_mk_classic(op_SHIFTS,1,params));
}

shifts shifts_dup(shifts ss)
{
  shifts res_ss;
  register int i;
  if (ss==shifts_NULL) return shifts_NULL;
  res_ss=shifts_init(shifts_rank(ss));
  for (i=1;i<=shifts_rank(ss);i++)
    shifts_set_nth(res_ss,i,term_dup(shifts_nth(ss,i)));
  return res_ss;
}

coc_bool shifts_eq(shifts ss1, shifts ss2)
{
  register int i;
  if (ss1 == shifts_NULL && ss2 == shifts_NULL) return TRUE;
  if (shifts_rank(ss1) != shifts_rank(ss2)) return FALSE;
  for (i=1;i<=shifts_rank(ss1);i++)
    if (!term_eq(shifts_nth(ss1,i),shifts_nth(ss2,i))) return FALSE;
  return TRUE;
}

void shifts_free(shifts ss)
{
  int i;
  if (ss == shifts_NULL) return;
  for (i=1;i<=shifts_rank(ss);i++)
    term_free(shifts_nth(ss,i));
  free_term_list(shifts_rank(ss),shifts_ss(ss));
  free_shifts(shifts_rank(ss),ss);
}

void shifts_expand(shifts ss)
{ /* the last element is not assigned */
  shift * aux_ss;
  int i,r;
  shifts_incr_rank(ss);
  r = shifts_rank(ss);
  aux_ss=shifts_ss(ss);
  shifts_set_ss(ss,malloc_term_list(r));
  for (i=1;i<=r-1;i++)
    shifts_set_nth(ss,i,aux_ss[i-1]);
  shifts_set_nth(ss,r,term_NULL);
  free_term_list(r-1,aux_ss);
}

void shifts_append(shifts * ss, term t)
{
  if (*ss == shifts_NULL) *ss = shifts_init(1);
  else shifts_expand(*ss);
  shifts_set_nth(*ss,shifts_rank(*ss),term_dup(t));
}

/* ---[ pvector ]--------------------------------------- */

coc_bool pv_is_homog(pvector v, ring R)
{
  int i,d;
  for (i=0;i<pv_rank(v);i++) if (pv_nth(v,i) != NULL) break;
  if (i == pv_rank(v)) return TRUE;
  d=pvcomp_is_homog_return_deg(pv_nth(v,i),R);
  if (d == -2) return FALSE;
  for (i++;i<pv_rank(v);i++)
    if (!pvcomp_is_homog_deg(pv_nth(v,i),d,R)) return FALSE;
  return TRUE;
}

void pv_Apply_Shifts(pvector v, shifts ss, ring R)
{
  int i;
  if (v == pv_NULL) return;
  if (ss==shifts_NULL) return;
  pv_Apply_Anti_Shifts(v,R); /* ensures no previous applied shifts */
  for (i=0;i<pv_rank(v);i++) 
    pv_set_nth(v,i,pvcomp_this_x_term(pv_nth(v,i),shifts_nth(ss,i+1),R));
  pv_set_shifts(v,shifts_link(ss));
}

void pv_Apply_Anti_Shifts(pvector v, ring R)
{
  int i;
  term aux_t;
  shifts ss;
  ss=pv_shifts(v);
  if (ss==shifts_NULL) return;
  for (i=0;i<pv_rank(v);i++) {
    aux_t=shifts_nth(ss,i+1);
    term_inverse(aux_t);
    pv_set_nth(v,i,pvcomp_this_x_term(pv_nth(v,i),aux_t,R));
    term_inverse(aux_t); }
}

void pvcomp_Apply_Anti_Shift(pvcomp * p, pvector v, int i, ring R)
{
  term aux_t;
  shifts ss;
  ss=pv_shifts(v);
  if (ss==shifts_NULL) return;
  aux_t=shifts_nth(ss,i);
  term_inverse(aux_t);
  *p=pvcomp_this_x_term(*p,aux_t,R);
  term_inverse(aux_t);
}

int pv_size(pvector v, ring R)
{
  int res_i;
  register int i;
  res_i=0;
  for (i=0;i<pv_rank(v);i++) res_i+=pvcomp_size(pv_nth(v,i),R);
  return res_i;
}

pvector pv_init(int r)
{
  pvector res_v;
  register int i;
  res_v=malloc_pv();
  pv_set_rank(res_v,r);
  pv_set_comp(res_v,malloc_pv_comp(r));
  pv_set_shifts(res_v,shifts_NULL);
  for (i=0;i<r;i++) pv_set_nth(res_v,i,NULL);
  return res_v;
}


void pv_free(pvector v,ring R)
{
  register int i;
  if (v != pv_NULL) {
    for (i=0;i<pv_rank(v);i++) pvcomp_free(pv_nth(v,i),R);
    free_pv_comp(pv_rank(v),pv_comp(v));
    shifts_unlink(pv_shifts(v));
    free_pv(v); }
}

coc_bool pv_is_zero(pvector v)
{
  register int i;
  for (i=0;i<pv_rank(v);i++)
    if (pv_nth(v,i) != NULL) return FALSE;
  return TRUE;
}

pvector pv_sum(pvector v1, pvector v2, ring R)
{
  pvector res_v;
  register int i;
  if (pv_rank(v2) > pv_rank(v1)) { res_v=v1; v1=v2; v2=res_v; }
  res_v=pv_init(pv_rank(v1));
  for (i=0;i<pv_rank(v1);i++)
    pv_set_nth(res_v,i,pvcomp_sum(pv_nth(v1,i),pv_nth(v2,i),R));
  pv_set_shifts(res_v,shifts_link(pv_shifts(v1)));
  return res_v;
}

pvector pv_subtract(pvector v1, pvector v2, ring R)
{
  pvector res_v;
  register int i;
  if (pv_rank(v2) > pv_rank(v1)) { res_v=v1; v1=v2; v2=res_v; }
  res_v=pv_init(pv_rank(v1));
  for (i=0;i<pv_rank(v1);i++)
    pv_set_nth(res_v,i,pvcomp_sub(pv_nth(v1,i),pv_nth(v2,i),R));
  pv_set_shifts(res_v,shifts_link(pv_shifts(v1)));
  return res_v;
}

void pv_enlarge(pvector v, int r, shifts ss)
{ /* we assume shifts(v,i) = ss[i] forall i < rank(v) */
  int i;
  if (r <= pv_rank(v)) return;
  if (ss != shifts_NULL) {
    shifts_unlink(pv_shifts(v));
    pv_set_shifts(v,shifts_link(ss));
  }
  pv_set_comp(v,realloc_pv_comp(pv_comp(v),pv_rank(v),r));
  for (i=pv_rank(v);i<r;i++) pv_set_nth(v,i,NULL);
  pv_set_rank(v,r);
}

pvector pv_sum_these(pvector v1, pvector v2, ring R)
{
  register int i;
  pvector aux_v;
  if (pv_rank(v2) > pv_rank(v1)) { aux_v=v1; v1=v2; v2=aux_v; }
  for (i=0;i<pv_rank(v2);i++)
    pv_set_nth(v1,i,pvcomp_sum_these(pv_nth(v1,i),pv_nth(v2,i),R));
  shifts_unlink(pv_shifts(v2));
  free_pv_comp(pv_rank(v2),pv_comp(v2));
  free_pv(v2);
  return v1;
}

void pv_this_x_coeff(pvector v, coeff c, ring R)
{
  register int i;
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(v,i,pvcomp_this_x_coeff(pv_nth(v,i),c,R));
}

void pv_change_sign(pvector v, ring R)
{
  register int i;
  pvcomp aux_p;
  for (i=0;i<pv_rank(v);i++) {
    aux_p=pv_nth(v,i);
    pvcomp_change_sign(aux_p,R);
  }
}

pvector pv_mult(pvcomp p, pvector v, ring R)
{
  pvector res_v;
  register int i;
  res_v=pv_init(pv_rank(v));
  pv_set_shifts(res_v,shifts_link(pv_shifts(v)));
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(res_v,i,pvcomp_mult(p,pv_nth(v,i),R));
  return res_v;
}

pvector pv_dup(pvector v, ring R)
{
  pvector res_v;
  register int i;
  if (v==pv_NULL) return pv_NULL;
  res_v=pv_init(pv_rank(v));
  pv_set_shifts(res_v,shifts_link(pv_shifts(v)));
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(res_v,i,pvcomp_dup(pv_nth(v,i),R));
  return res_v;
}

pvector pv_fill(pvcomp p, int r, ring R)
{
  pvector res_v;
  register int i;
  res_v=pv_init(r);
  for (i=0;i<r;i++)
    pv_set_nth(res_v,i,pvcomp_dup(p,R));
  return res_v;
}

cterm pv2cterm(pvector v, ring R)
{
  cterm_list params;
  cterm res_t;
  shifts ss;
  register int i;
  params=malloc_cterm_list(pv_rank(v));
  if (pv_shifts(v) != shifts_NULL) {
    pv_Apply_Anti_Shifts(v,R);
    for (i=0;i<pv_rank(v);i++)
      params[i]=pvcomp2cterm(pv_nth(v,i),R);
    ss = pv_shifts(v);
    pv_set_shifts(v,shifts_NULL);
    pv_Apply_Shifts(v,ss,R);
    shifts_unlink(ss); /* TAPPUL! (TODO) */
  } else
    for (i=0;i<pv_rank(v);i++)
      params[i]=pvcomp2cterm(pv_nth(v,i),R);
  res_t=cterm_mk_classic(op_VECTOR,pv_rank(v),params);
  free_cterm_list(pv_rank(v),params);
  return res_t;
}

cterm pv2cterm2(pvector v, ring R)
{
  cterm_list params;
  cterm res_t;
  shifts ss;
  register int i;
  params=malloc_cterm_list(pv_rank(v));
  if (pv_shifts(v) != shifts_NULL) {
    pv_Apply_Anti_Shifts(v,R);    
    for (i=0;i<pv_rank(v);i++)
      params[i]=pvcomp2cterm(pv_nth(v,i),R);
    ss = pv_shifts(v);
    pv_set_shifts(v,shifts_NULL);
    pv_Apply_Shifts(v,ss,R);
    shifts_unlink(ss); /* TAPPUL! (TODO) */
  }
  else
    for (i=0;i<pv_rank(v);i++)
      params[i]=pvcomp2cterm(pv_nth(v,i),R);
  res_t=cterm_mk_classic(op_LIST,pv_rank(v),params);
  free_cterm_list(pv_rank(v),params);
  return res_t;
}

void printf_pv(pvector v, ring R)
{
  cterm aux_t;
  aux_t=pv2cterm(v,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}



void pv_leading_monomial(pvector v, monomial * m, int * pos, ring R)

#ifdef PVCOMP_TRIPOLI_OR_EEE

{
  pv_leading_pos(v,pos,R);
  *m=pvcomp_leading_monomial(pv_nth(v,*pos),R);
}

#else

{
  register int i;
  monomial m1;
  int answer,r;
  *pos=0;
  i = 0;
  r = pv_rank(v);
  while (i<r && (pv_nth(v,i)== NULL)) i++;
  if (i==r) *m = NULL;
  else {
    *m = pvcomp_this_leading_monomial(pv_nth(v,i),R);
    *pos=i;
    if (R->mord==TO_POS)
      for (i=*pos+1;i<r;i++)
        if (pv_nth(v,i) != NULL) {
          m1=pvcomp_this_leading_monomial(pv_nth(v,i),R); /* m1 <> NULL */
          answer=monomial_which_gt(*m,m1,R);
          if (answer == ANSWER_SECOND) {
            *m=m1;
            *pos=i; } }
    *m=monomial_dup(*m,R); }
}

#endif

coeff pv_leading_coeff(pvector v, ring R)
{
  coeff res_c;
  int aux_n;
  pv_leading_pos(v, &aux_n, R);
  res_c=pvcomp_leading_coeff(pv_nth(v, aux_n),R);
  return res_c;
}

#ifdef PVCOMP_MONOPOLI

void pv_extract_lm(pvector v, monomial * m, int * pos, ring R)
{
  register int i;
  monomial m1;
  int answer,r;
  *pos=0;
  i = 0;
  r = pv_rank(v);
  while (i<r && (pv_nth(v,i) == NULL)) i++;
  if (i==r) *m = NULL;
  else {
    *m = pvcomp_this_leading_monomial(pv_nth(v,i),R);
    *pos=i;
    if (R->mord==TO_POS)
      for (i=*pos+1;i<r;i++)
        if (pv_nth(v,i) != NULL) {
          m1=pvcomp_this_leading_monomial(pv_nth(v,i),R);
          answer=monomial_which_gt(*m,m1,R);
          if (answer == ANSWER_SECOND) {
            *m=m1;
            *pos=i; } } }
  if (*m != NULL) {
    pv_set_nth(v,*pos,monomial_next(*m));
    monomial_set_next(*m,NULL);
  }
}

#elif defined(PVCOMP_TRIPOLI_OR_EEE)

void pv_extract_lm(pvector v, monomial * m, int * pos, ring R)
{
  pvcomp aux_p;
  pv_leading_pos(v,pos,R);
  aux_p=pv_nth(v,*pos);
  if (aux_p == NULL) {
    *m=NULL;
  } else {
    *m=pvcomp_extract_lm(&aux_p,R);
    pv_set_nth(v,*pos,aux_p);
  }
}

#endif

void pv_leading_pos(pvector v, int * pos, ring R)

#ifdef PVCOMP_TRIPOLI_OR_EEE

{
  register int i;
  int answer,r;
  *pos=0;
  i = 0;
  r = pv_rank(v);
  while ((pv_nth(v,i) == NULL) && i<r) i++;
  if (i!=r) {
    *pos = i;
    if (R->mord==TO_POS) {
      for (i=*pos+1;i<r;i++)
        if (pv_nth(v,i) != NULL) {
          answer=pvcomp_lm_which_gt(pv_nth(v,*pos),pv_nth(v,i),R);
          if (answer == ANSWER_SECOND) *pos=i; } } }
}

#else

{
  register int i;
  monomial m,m1;
  int answer,r;
  *pos=0;
  i = 0;
  r = pv_rank(v);
  while (i < r && (pv_nth(v,i) == NULL)) i++;
  if (i!=r) {
    m = pvcomp_this_leading_monomial(pv_nth(v,i),R);
    *pos=i;
    if (R->mord==TO_POS)
      for (i=*pos+1;i<r;i++)
        if (pv_nth(v,i) != NULL) {
          m1=pvcomp_this_leading_monomial(pv_nth(v,i),R);
          answer=monomial_which_gt(m,m1,R);
          if (answer == ANSWER_SECOND) {
            m=m1;
            *pos=i; } } }
}

#endif

coc_bool pv_eq(pvector v1, pvector v2, ring R)
{
  register int i;

  if (pv_rank(v1) != pv_rank(v2)) return FALSE;
  for (i=0;i<pv_rank(v1);i++)
    if (!pvcomp_eq(pv_nth(v1,i),pv_nth(v2,i),R)) return FALSE;
  return TRUE;
}

void pv_make_monic(pvector v, ring R)
{
  int pos;
  coeff aux_c;
  register int i;
  monomial m;

  if (v==pv_NULL) return;

  pv_leading_monomial(v,&m,&pos,R);
  if (m == NULL) return;
  aux_c=monomial_coeff(m);
  R->InvertCoeff(&aux_c,aux_c,R);
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(v,i,pvcomp_this_x_coeff(pv_nth(v,i),aux_c,R));
  monomial_free(m,R);
}

pvector pv_pvcomp2pv(pvcomp p, int pos, int r, ring R)
{
  register int i;
  pvector res_v;
  res_v=pv_init(r);
  for (i=0;i<pos;i++) pv_set_nth(res_v,i,NULL);
  for (i>pos;i<r;i++) pv_set_nth(res_v,i,NULL);
  pv_set_nth(res_v,pos,pvcomp_dup(p,R));
  return res_v;
}

pvector this_pv_plus_this_monomial(pvector v, monomial m, int pos, ring R)
{
  pv_set_nth(v,pos,pvcomp_sum_these(pv_nth(v,pos),
    this_monomial2pvcomp(m,R),R));
  return v;
}

pvector pv_this_pvcomp2pv(pvcomp p, int pos, int r, ring R)
{
  register int i;
  pvector res_v;
  res_v=pv_init(r);
  for (i=0;i<pos;i++) pv_set_nth(res_v,i,NULL);
  for (i>pos;i<r;i++) pv_set_nth(res_v,i,NULL);
  pv_set_nth(res_v,pos,p);
  return res_v;
}

pvector pv_only_leading_term(pvector v, ring R)

#ifdef PVCOMP_TRIPOLI_OR_EEE

{
  int pos;
  pvcomp aux_p;
  pv_leading_pos(v,&pos,R);
  aux_p=pvcomp_only_leading_term(pv_nth(v,pos),R);
  return pv_this_pvcomp2pv(aux_p,pos,pv_rank(v),R);
}

#else

{
  pvector res_v;
  int pos,r;
  monomial m;
  r=pv_rank(v);
  pv_leading_monomial(v,&m,&pos,R); /* TODO OPTIMIZE: pv_leading_term */
  if (m == NULL) return pv_gen(r,0,R);
  res_v=this_term2pv(monomial_term(m),pos,r,R);
  pv_set_shifts(res_v,shifts_link(pv_shifts(v)));
  R->ClearCoeff(&monomial_coeff(m));
  free_monomial(m);
  return res_v;
}

#endif

pvector pv_zero(int r, ring R)
{
  register int i;
  pvector res_v;
  res_v=pv_init(r);
  for (i=0;i<r;i++) pv_set_nth(res_v,i,NULL);
  return res_v;
}

pvector pv_gen(int r, int pos, ring R)
{
  register int i;
  pvector res_v;
  res_v=pv_init(r);
  if (pos<r) pv_set_nth(res_v,pos,pvcomp_one(R));
  return res_v;
}

pvector pv_x_monomial(pvector v, monomial m, ring R)
{
  register int i;
  int r=pv_rank(v);
  pvector res_v;
  res_v=pv_init(r);
  pv_set_shifts(res_v,shifts_link(pv_shifts(v)));
  for (i=0;i<r;i++) pv_set_nth(res_v,i,pvcomp_x_monomial(pv_nth(v,i),m,R));
  return res_v;
}

pvector pv_this_x_monomial(pvector v, monomial m, ring R)
{
  register int i;
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(v,i,pvcomp_this_x_monomial(pv_nth(v,i),m,R));
  return v;
}

pvector pv_this_x_term(pvector v, term t, ring R)
{
  register int i;
  for (i=0;i<pv_rank(v);i++)
    pv_set_nth(v,i,pvcomp_this_x_term(pv_nth(v,i),t,R));
  return v;
}

int pv_deg(pvector v, ring R)
{
  int res_i;
  register int i;
  res_i=pvcomp_deg(pv_nth(v,0),R);
  for (i=1;i<pv_rank(v);i++) res_i=MAX(res_i,pvcomp_deg(pv_nth(v,i),R));
  return res_i;
}

ivector pv_mdeg(pvector v, ring R)
{
  ivector res_v;
  register int i;
  res_v=pvcomp_mdeg(pv_nth(v,0),R);
  for (i=1;i<pv_rank(v);i++) 
    res_v=max_mdeg_lex_these(res_v,pvcomp_mdeg(pv_nth(v,i),R));
  return res_v;
}

void pv_change_order(pvector v, ring from_R, ring to_R)
{
  register int i;
  pvcomp aux_p;
  for (i=0;i<pv_rank(v);i++) {
    aux_p=pv_nth(v,i);
    pvcomp_change_order(&aux_p,from_R,to_R);
    pv_set_nth(v,i,aux_p); }
}

void pv_zero2pv_NULL(pvector *v, ring R)
{
  if (pv_is_zero(*v)) {
    pv_free(*v,R);
    *v=pv_NULL; }
}

pvector pv_homog(pvector v, unknown x, ring R)
{
  register int i;
  int r=pv_rank(v);
  pvector res_v;
  res_v=pv_init(r);
  pv_set_shifts(res_v,shifts_link(pv_shifts(v)));
  for (i=0;i<r;i++) pv_set_nth(res_v,i,pvcomp_homog(pv_nth(v,i),x,R));
  return res_v;
}
