/* $Id: ratfun.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "coeffs.h"
#include "cterm.h"
#include "ctbuf.h"
#include "hilop.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "polys.h"
#include "ratfun.h"
#include "ring.h"
#include "terms.h"

ratfun ratfun_init( void )
{
  ratfun res_F;
  res_F = malloc_ratfun();
  ratfun_set_num(res_F,NULL);
  ratfun_set_den(res_F,NULL);
  return res_F;
}

void ratfun_free( ratfun F, ring R )
{
   if (F==ratfun_NULL) return;
   arpoly_free( ratfun_num(F),R );
   arpoly_free( ratfun_den(F),R );
   free_ratfun(F);
}

ratfun ratfun_dup( ratfun F, ring R )
{
  ratfun res_F;
  if (F==ratfun_NULL) return ratfun_NULL;
  res_F=ratfun_init();
  ratfun_set_num(res_F,arpoly_dup(ratfun_num(F),R));
  ratfun_set_den(res_F,arpoly_dup(ratfun_den(F),R));
  return res_F;
}

void ratfun_simplify( ratfun F, ring R )
{
  coeff aux_c;
  arpoly aux_p,aux_p1,aux_p2;
  if (ratfun_num(F) == NULL) {
    arpoly_free(ratfun_den(F), R);
    ratfun_set_den(F,arpoly_one(R));
  } else {
    aux_p = arpoly_gcd(ratfun_num(F), ratfun_den(F), R);
    if (arpoly_deg(aux_p,R) > 0) {
      arpoly_div(ratfun_num(F), aux_p, &aux_p1, &aux_p2, R);
      arpoly_free(ratfun_num(F), R);
      ratfun_set_num(F, aux_p1);
      arpoly_free(aux_p2, R);
      arpoly_div(ratfun_den(F), aux_p, &aux_p1, &aux_p2, R);
      arpoly_free(ratfun_den(F), R);
      ratfun_set_den(F, aux_p1);
      arpoly_free(aux_p2, R);
    }
    arpoly_free(aux_p,R);
    /* can be still simplified - see CoCoa 1.5 */
    /* oppure [OK] dividere numeratore e denominatore per il Lc(denominatore) */
  aux_c=coeff_inverse(arpoly_this_leading_coeff(ratfun_den(F),R),R);
  ratfun_set_den(F,arpoly_this_x_coeff(ratfun_den(F),aux_c,R));
  ratfun_set_num(F,arpoly_this_x_coeff(ratfun_num(F),aux_c,R));
  coeff_free(aux_c,R);
  }
}

ratfun ratfun_mk(arpoly N, arpoly D, ring R)
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F,N);
  ratfun_set_den(res_F,D);
  ratfun_simplify(res_F, R);
  return res_F;
}

ratfun ratfun_cons(arpoly N, arpoly D, ring R)
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F,N);
  ratfun_set_den(res_F,D);
  return res_F;
}

ratfun ratfun_create(arpoly N, arpoly D, ring R)
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F,arpoly_dup(N,R));
  ratfun_set_den(res_F,arpoly_dup(D,R));
  ratfun_simplify(res_F, R);
  return res_F;
}

ratfun ratfun_neg( ratfun F, ring R )
{
  ratfun res_F;
  res_F=ratfun_dup(F, R);
  ratfun_change_sign(res_F,R);
  return res_F;
}

ratfun ratfun_sum( ratfun F, ratfun G, ring R )
{
  arpoly aux_p,aux_p1,aux_q,aux_r;
  ratfun res_F;

  aux_p = arpoly_lcm(ratfun_den(F), ratfun_den(G), R );
  arpoly_div(aux_p, ratfun_den(F), &aux_q, &aux_r, R);
  aux_p1 = arpoly_mult( ratfun_num(F), aux_q, R);
  arpoly_free(aux_q,R);
  arpoly_free(aux_r,R);
  
  arpoly_div(aux_p, ratfun_den(G), &aux_q, &aux_r, R);
  res_F=ratfun_mk(arpoly_sum_these(aux_p1,arpoly_mult(ratfun_num(G),aux_q,R), R),
    aux_p,R);
  arpoly_free(aux_q,R);
  arpoly_free(aux_r,R);
  return res_F;
}

ratfun ratfun_sub( ratfun F, ratfun G, ring R )
{
  arpoly aux_p,aux_p1,aux_q,aux_r;
  
  aux_p = arpoly_lcm(ratfun_den(F), ratfun_den(G), R );
  arpoly_div(aux_p, ratfun_den(F), &aux_q, &aux_r, R);
  aux_p1 = arpoly_mult( ratfun_num(F), aux_q, R);
  arpoly_free(aux_q,R);
  arpoly_free(aux_r,R);
  
  arpoly_div(aux_p, ratfun_den(G), &aux_q, &aux_r, R);
  arpoly_free(aux_r,R);
  aux_r = arpoly_mult(ratfun_num(G), aux_q,R);
  arpoly_free(aux_q,R);
  arpoly_change_sign(aux_r,R);
  return ratfun_mk(arpoly_sum_these(aux_p1,aux_r, R),aux_p,R );
}

ratfun ratfun_mult( ratfun F, ratfun G, ring R )
{
  ratfun res_F, aux_F1,aux_F2;
  aux_F1=ratfun_create(ratfun_num(F), ratfun_den(G),R);
  aux_F2=ratfun_create(ratfun_num(G), ratfun_den(F),R);
  res_F=ratfun_init();
  ratfun_set_num(res_F, arpoly_mult(ratfun_num(aux_F1), ratfun_num(aux_F2), R));
  ratfun_set_den(res_F, arpoly_mult(ratfun_den(aux_F1), ratfun_den(aux_F2), R));
  ratfun_free(aux_F1,R);
  ratfun_free(aux_F2,R);
  return res_F;
}

ratfun ratfun_inv_this( ratfun F, ring R )
{
  arpoly aux_p;
  aux_p=ratfun_num(F);
  ratfun_set_num(F, ratfun_den(F));
  ratfun_set_den(F, aux_p);
  ratfun_simplify(F, R);
  return F;
}

ratfun ratfun_inv( ratfun F, ring R )
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F, arpoly_dup(ratfun_den(F),R));
  ratfun_set_den(res_F, arpoly_dup(ratfun_num(F),R));
  ratfun_simplify(F, R);
  return res_F;
}

ratfun ratfun_div( ratfun F, ratfun G, ring R )
{
  ratfun res_F, aux_F1,aux_F2;
  aux_F1=ratfun_create(ratfun_num(F), ratfun_num(G),R);
  aux_F2=ratfun_create(ratfun_den(G), ratfun_den(F),R);
  res_F=ratfun_init();
  ratfun_set_num(res_F, arpoly_mult(ratfun_num(aux_F1), ratfun_num(aux_F2), R));
  ratfun_set_den(res_F, arpoly_mult(ratfun_den(aux_F1), ratfun_den(aux_F2), R));
  ratfun_free(aux_F1,R);
  ratfun_free(aux_F2,R);
  return res_F;
}

ratfun ratfun_der( ratfun F, unknown x, ring R )
{
  ratfun res_F;
  arpoly aux_P, aux_P1, auxP_2;
  
  res_F = ratfun_init();
  aux_P1 = arpoly_der(ratfun_num(F), x, R);
  aux_P = arpoly_mult(aux_P1, ratfun_den(F), R);
  arpoly_free(aux_P1, R);
  aux_P1 = arpoly_der(ratfun_den(F), x, R);
  arpoly_change_sign(aux_P1, R);
  ratfun_set_num(res_F,
    arpoly_sum_these(aux_P,arpoly_mult(aux_P1,ratfun_num(F),R),R));
  arpoly_free(aux_P1, R);
  
  ratfun_set_den(res_F, arpoly_square(ratfun_den(F),R));
  ratfun_simplify( res_F, R );
  return res_F;
}

ratfun ratfun_der_n( ratfun F, unknown x, int n, ring R )
{
  ratfun res_F,aux_F;
  int i;
  res_F = ratfun_der(F,x,R);
  for (i=2; i<=n && !ratfun_is_zero(res_F); i++) {
    aux_F = ratfun_der(res_F,x,R);
    ratfun_free(res_F,R);
    res_F = aux_F; }
  return res_F;
}

ratfun ratfun_der_nn(ratfun F, int * der_list, ring R)
{
  int m,n,i;
  unknown x;
  ratfun aux_F,res_F;
  m=der_list[0];
  if (m==0) return ratfun_dup(F,R);
  x=der_list[1];
  n=der_list[2];
  res_F=ratfun_der_n(F,x,n,R);
  for (i=2;i<=m && ! ratfun_is_zero(res_F);i++) {
    x=der_list[i*2-1];
    n=der_list[i*2];
    aux_F=ratfun_der_n(res_F,x,n,R);
    ratfun_free(res_F,R);
    res_F=aux_F; }
  return res_F;
}

ratfun ratfun_power(ratfun p, expType exp, ring R)
{
  ratfun aux_p,aux_p2,aux_p3,res_p;
  if (exp < 0) {
    aux_p=ratfun_power(p,-exp,R);
    return ratfun_inv_this(aux_p,R); }
  if (exp == 0) return ratfun_one(R);
  if (exp == 1) return ratfun_dup(p,R);
  res_p=ratfun_one(R);
  aux_p=ratfun_dup(p,R);
  while (exp > 1) {
    if (exp & 1) {
      aux_p3=ratfun_mult(res_p,aux_p,R);
      ratfun_free(res_p,R);
      res_p=aux_p3; }
    exp /= 2;
    aux_p2=ratfun_square(aux_p,R);
    ratfun_free(aux_p,R);
    aux_p=aux_p2; }
  aux_p3=ratfun_mult(res_p,aux_p,R);
  ratfun_free(res_p,R);
  res_p=aux_p3;
  return res_p;
}

ratfun arpoly2ratfun( arpoly p, ring R )
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F, arpoly_dup(p,R));
  ratfun_set_den(res_F, arpoly_one(R));
  return res_F;
}

ratfun this_arpoly2ratfun( arpoly p, ring R )
{
  ratfun res_F;
  res_F=ratfun_init();
  ratfun_set_num(res_F, p);
  ratfun_set_den(res_F, arpoly_one(R));
  return res_F;
}

cterm ratfun2cterm( ratfun F, ring R )
{
  cterm params[2];
  if (arpoly_is_one(ratfun_den(F),R)) {
    return arpoly2cterm(ratfun_num(F),R);
  }
  params[0]=arpoly2cterm(ratfun_num(F),R);
  params[1]=arpoly2cterm(ratfun_den(F),R);
  return cterm_mk_classic(op_RATFUN,2,params);
}

void printf_ratfun( ratfun F, ring R )
{
  cterm aux_t;
  aux_t=ratfun2cterm(F,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

ratfun cterm2ratfun( cterm t, ring R )
{
  arpoly num,den;
  ratfun res_F;
  cterm_list params;
  params=cterm_params(t);
  num=cterm2arpoly(params[0],R);
  den=cterm2arpoly(params[1],R);
  res_F=ratfun_mk(num,den,R);
  return res_F;
}

arpoly this_ratfun2arpoly(ratfun F, ring R)
{
  arpoly res_p;
  res_p=ratfun_num(F);
  ratfun_set_num(F,NULL);
  ratfun_free(F,R);
  return res_p;
}
