/* ratfun.h : 0.1 : July 10, 1994 : ADT for rational functions in CoCoA */

/*------------------------------------------------------------------------------

	part of project cocoa

	revision history

------------------------------------------------------------------------------*/

#ifndef RATFUN_H
#define RATFUN_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#ifndef COCTYPES_H
# include "coctypes.h"
#endif

#define malloc_ratfun() (ratfun)mymalloc(sizeof(ratfun_info),"ratfun")
#define free_ratfun(p) myfree(sizeof(ratfun_info),p,"ratfun")

#define ratfun_num(F) ((F)->num)
#define ratfun_den(F) ((F)->den)

#define ratfun_set_num(F,N) (F)->num=N
#define ratfun_set_den(F,D) (F)->den=D

#define ratfun_change_sign(F,R) arpoly_change_sign(ratfun_num(F),R)
#define ratfun_is_zero(F) (ratfun_num(F) == NULL)
#define ratfun_is_poly(F,R) arpoly_is_one(ratfun_den(F),R)
#define ratfun_eq(F,G,R) (arpoly_eq(ratfun_num(F),ratfun_num(G),R) && \
  arpoly_eq(ratfun_den(F),ratfun_den(G),R))

#define ratfun_one(R) ratfun_mk(arpoly_one(R),arpoly_one(R),R)
#define ratfun_zero(R) ratfun_mk(NULL,arpoly_one(R),R)

#define ratfun_square(F,R) ratfun_mult(F,F,R)

ratfun ratfun_init( void );
void ratfun_free( ratfun F, ring R );
ratfun ratfun_dup( ratfun F, ring R );


ratfun ratfun_mk(arpoly N, arpoly D, ring R);
ratfun ratfun_cons(arpoly N, arpoly D, ring R);
ratfun ratfun_create(arpoly N, arpoly D, ring R);
ratfun ratfun_neg( ratfun F, ring R );
ratfun ratfun_sum( ratfun F, ratfun G, ring R );
ratfun ratfun_sub( ratfun F, ratfun G, ring R );
ratfun ratfun_mult( ratfun F, ratfun G, ring R );
ratfun ratfun_inv_this( ratfun F, ring R );
ratfun ratfun_inv( ratfun F, ring R );
/* assume F invertible */
ratfun ratfun_div( ratfun F, ratfun G, ring R );
/* assume G not zero */
ratfun ratfun_der( ratfun F, unknown x, ring R );
ratfun ratfun_der_n( ratfun F, unknown x, int n, ring R );
ratfun ratfun_der_nn(ratfun F, int * der_list, ring R);

ratfun ratfun_power(ratfun p, expType exp, ring R);

ratfun arpoly2ratfun( arpoly p, ring R );
ratfun this_arpoly2ratfun( arpoly p, ring R );
arpoly this_ratfun2arpoly(ratfun F, ring R);

void ratfun_simplify( ratfun F, ring R );

cterm ratfun2cterm( ratfun F, ring R );
void printf_ratfun( ratfun F, ring R );
ratfun cterm2ratfun( cterm t, ring R );

#endif


