/* $Id: ring.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ck.h"
#include "coeffs.h"
#include "cterm.h"
#include "indet.h"
#include "ivectors.h"
#include "module.h"
#include "memoryco.h"
#include "ordering.h"
#include "operator.h"
#include "polys.h"
#include "ring.h"
#include "rum.h"
#include "scalars.h"
#include "terms.h"
#include "weights.h"

cterm ring2cterm(ring R);
cterm characteristic2cterm(int c);

#ifdef DEBUG_RING_RC

ring ring_link(ring R)
{
  return (R != ring_NULL ? (ring_incr_rc(R),R) : ring_NULL);
}

void ring_unlink(ring R)
{
  if ((R)!=ring_NULL) ring_unlink_do(R);
}

#endif


void ring_set_ordering(ring R,ordering to,short whatTO,
  unknown first_unknown, unknown last_unknown)
{
  /* TODO: to is aliased, ot is created. Study a better choice */
  R->to=to; /* old to must be deleted */
  R->whatTO=whatTO;
  R->first_u=first_unknown;
  R->last_u=last_unknown;
  switch (whatTO) {
  case lex_num : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_lex;
    R->term2scalars=term2scalars_lex;
    R->ot=R->to;
    R->det=1;
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_lex;
    R->ivec_term2scalars=ivec_term2scalars_lex;
#endif
#endif
    R->term_which_gt=term_which_gt_lex; 
    break;

  case xel_num : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_xel;
    R->term2scalars=term2scalars_xel;
    R->ot=R->to;
    R->det=((R->indetsNo/2 & 1) ? -1 : 1);
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_xel;
    R->ivec_term2scalars=ivec_term2scalars_xel;
#endif
#endif
    R->term_which_gt=term_which_gt_xel; 
    break;

/* TODO: fare scalars2term_deglex e term2scalars_deglex */

  case dlex_num : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_gen;
    R->term2scalars=term2scalars_gen;
    R->ot=ord_adjoint(R->to,R);
    R->det=ord_det(R->ot,R);
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_gen;
    R->ivec_term2scalars=ivec_term2scalars_gen;
#endif
#endif
    R->term_which_gt=term_which_gt_deglex;
    break;

  case drev_num : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_degrevlex;
    R->term2scalars=term2scalars_degrevlex;
    R->ot=R->to;
    R->det=1;
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_degrevlex;
    R->ivec_term2scalars=ivec_term2scalars_degrevlex;
#endif
#endif
    R->term_which_gt=term_which_gt_degrevlex;
    break;

  case elim_num : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_gen;
    R->term2scalars=term2scalars_gen;
    R->ot=ord_adjoint(R->to,R);
    R->det=ord_det(R->to,R);
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_gen;
    R->ivec_term2scalars=ivec_term2scalars_gen;
#endif
#endif
    R->term_which_gt=term_which_gt_elim;
    break;

  default : 
#ifdef TRIPOLI_OR_EEE
    R->scalars2term=scalars2term_gen;
    R->term2scalars=term2scalars_gen;
    R->ot=ord_adjoint(R->to,R);
    R->det=ord_det(R->to,R);
#ifdef EEE
    R->ivec_scalars2term=ivec_scalars2term_gen;
    R->ivec_term2scalars=ivec_term2scalars_gen;
#endif
#endif
    R->term_which_gt=term_which_gt_other;
    break;
  }
}

ring ring_init(char * ring_name,ringtype rt,int characteristic,ring coeffs,
         indet_info indets,ordering to,short whatTO,short mord)
{
  ring R;
  register int i;
  int rum;

  R=malloc_ring();
  ring_set_locked(R,FALSE);
  ring_set_weilist(R,NULL);
  ring_set_rc(R,1);
  R->ring_name=my_strdup(ring_name);
  R->rtype=rt;
  R->quid=module_NULL;
  ring_set_local_env(R,env_init());
  R->characteristic=characteristic;
  R->halfP=characteristic/2;
  if (R->rtype==ringtype_coeff) R->coeffs=R;
  else R->coeffs=coeffs;
  R->indetsNo=count_indets(indets);
  R->indets=indets;

  R->to=to;
  R->whatTO=whatTO;
  R->mord=mord;

  R->weights=iv_NULL;

  /* R->level=R->indetsNo; */
  ring_set_level(R,R->indetsNo/2);

#ifdef COCOA_WITH_RUM

  if (R->indetsNo != 0) {

#ifdef MONOPOLI

    if (R->indetsNo <= 100) {
      rum=sizeof(expType)*(R->indetsNo+2);
      /* rum_empty(rum); */
      rum_init(rum,RUM_STD_SIZE); }

#endif /* MONOPOLI */

  }

#endif /* COCOA_WITH_RUM */

#ifdef TRIPOLI_OR_EEE

  R->ot=ordering_NULL;
  R->scalars2term = NULL;
  R->term2scalars = NULL;

#ifdef EEE
  R->ivec_scalars2term = NULL;
  R->ivec_term2scalars = NULL;
  R->eee_dum=malloc_eee_mon_array(R->indetsNo+1);
  for (i=0;i<=R->indetsNo;i++)
    eee_set_dummy(R,i,malloc_eee_mon());
#endif

#ifdef TRIPOLI
  R->tripoli_dum=trnode_array_init(R->indetsNo+1,R);
#endif

#endif

  R->term_which_gt = NULL;

  for (i=0;i<NUM_AUX_IV;i++) R->aux_iv[i]=iv_init(R->indetsNo);

  if ( characteristic > 1 ) { /* integers module p */
    
    R->InitCoeff   =  & ZpInit ;
    R->ClearCoeff   =  & ZpClear ;
    R->CopyCoeff   =  & ZpCopy;
    R->MakeCoeff    =  & ZpMakeCoeff ;
    R->Str2Coeff  =  & ZpStr2Coeff;
    R->Cterm2Coeff = & ZpCterm2Coeff;
    
    R->InvertCoeff =  & ZpInvert ;
    R->NegCoeff    =  & ZpNeg;
    R->ChangeSignToCoeff   = & ZpChangeSign;
    R->RandomizeCoeff    = & ZpRandomize;
    
    R->AddCoeff    =  & ZpAdd ;
    R->SubCoeff     =  & ZpSub ;
    R->MulCoeff    =  & ZpMul ;
    R->DivCoeff    =  & ZpDiv ;
    R->ModCoeff    =  & ZpMod ;
    
    R->AreEqualsCoeff   = & AreZpEquals ;
    R->IsPositiveCoeff  = & IsZpPositive ;
    R->IsZeroCoeff   = & IsZpZero ;
    R->IsOneCoeff     = & IsZpOne ;
    R->IsMinusOneCoeff = & IsZpMinusOne ;
    R->Coeff2Cterm = & Zp2Cterm;
    R->Li2Coeff = & Li2Zp;

  } else if ( characteristic == characteristic_Si ) {
    R->InitCoeff     =  & SiInit ;
    R->ClearCoeff     =  & SiClear ;
    R->CopyCoeff     =  & SiCopy;
    R->MakeCoeff      =  & SiMakeCoeff ;
    R->Str2Coeff    =   & SiStr2Coeff;
    R->Cterm2Coeff = & SiCterm2Coeff;
    
    R->InvertCoeff    =  & SiInvert ;
    R->NegCoeff      =   & SiNeg;
    R->ChangeSignToCoeff   = & SiChangeSign;
    R->RandomizeCoeff    = & SiRandomize;
    
    R->AddCoeff    = & SiAdd ;
    R->SubCoeff     = & SiSub ;
    R->MulCoeff    = & SiMul ;
    R->DivCoeff    = & SiDiv ;
    R->ModCoeff    = & SiMod ;
        
    R->AreEqualsCoeff   = & AreSiEquals ;
    R->IsPositiveCoeff  = & IsSiPositive ;
    R->IsZeroCoeff   = & IsSiZero ;
    R->IsOneCoeff     = & IsSiOne ;
    R->IsMinusOneCoeff = & IsSiMinusOne ;
    R->Coeff2Cterm = & Si2Cterm;
    R->Li2Coeff = & Li2Si;

  } else if ( characteristic ==  characteristic_Sr ) {
    R->InitCoeff     =  & SrInit ;
    R->ClearCoeff     =  & SrClear ;
    R->CopyCoeff     =  & SrCopy;
    R->MakeCoeff      =  & SrMakeCoeff ;
    R->Str2Coeff    =    & SrStr2Coeff;
    R->Cterm2Coeff = & SrCterm2Coeff;
    
    R->InvertCoeff    =  & SrInvert ;
    R->NegCoeff      =   & SrNeg;
    R->ChangeSignToCoeff   = & SrChangeSign;
    R->RandomizeCoeff    = & SrRandomize;
    
    R->AddCoeff    = & SrAdd ;
    R->SubCoeff     = & SrSub ;
    R->MulCoeff    = & SrMul ;
    R->DivCoeff    = & SrDiv ;
    R->ModCoeff    = & SrMod ;
        
    R->AreEqualsCoeff   = & AreSrEquals ;
    R->IsPositiveCoeff  = & IsSrPositive ;
    R->IsZeroCoeff   = & IsSrZero ;
    R->IsOneCoeff     = & IsSrOne ;
    R->IsMinusOneCoeff = & IsSrMinusOne ;
    R->Coeff2Cterm = & Sr2Cterm;
    R->Li2Coeff = & Li2Sr;

  } else if ( characteristic == characteristic_Li ) { 
    R->InitCoeff     =  & LiInit ;
    R->ClearCoeff     =  & LiClear ;
    R->CopyCoeff     =  & LiCopy;
    R->MakeCoeff      =  & LiMakeCoeff ;
    R->Str2Coeff    =   & LiStr2Coeff;
    R->Cterm2Coeff = & LiCterm2Coeff;
    
    R->InvertCoeff    =  & LiInvert ;
    R->NegCoeff      =   & LiNeg;
    R->ChangeSignToCoeff   = & LiChangeSign;
    R->RandomizeCoeff    = & LiRandomize;
    
    R->AddCoeff    = & LiAdd ;
    R->SubCoeff     = & LiSub ;
    R->MulCoeff    = & LiMul ;
    R->DivCoeff    = & LiDiv ;
    R->ModCoeff    = & LiMod ;
        
    R->AreEqualsCoeff   = & AreLiEquals ;
    R->IsPositiveCoeff  = & IsLiPositive ;
    R->IsZeroCoeff   = & IsLiZero ;
    R->IsOneCoeff     = & IsLiOne ;
    R->IsMinusOneCoeff = & IsLiMinusOne ;
    R->Coeff2Cterm = & Li2Cterm;
    R->Li2Coeff = & Li2Li;

  } else if ( characteristic ==  characteristic_Lr ) {
    R->InitCoeff     =  & LrInit ;
    R->ClearCoeff     =  & LrClear ;
    R->CopyCoeff     =  & LrCopy;
    R->MakeCoeff      =  & LrMakeCoeff ;
    R->Str2Coeff    =   & LrStr2Coeff;
    R->Cterm2Coeff = & LrCterm2Coeff;
    
    R->InvertCoeff    =  & LrInvert ;
    R->NegCoeff      =   & LrNeg;
    R->ChangeSignToCoeff   = & LrChangeSign;
    R->RandomizeCoeff    = & LrRandomize;
    
    R->AddCoeff    = & LrAdd ;
    R->SubCoeff    = & LrSub ;
    R->MulCoeff    = & LrMul ;
    R->DivCoeff    = & LrDiv ;
    R->ModCoeff    = & LrMod ;
        
    R->AreEqualsCoeff   = & AreLrEquals ;
    R->IsPositiveCoeff  = & IsLrPositive ;
    R->IsZeroCoeff   = & IsLrZero ;
    R->IsOneCoeff     = & IsLrOne ;
    R->IsMinusOneCoeff = & IsLrMinusOne ;
    R->Coeff2Cterm = & Lr2Cterm;
    R->Li2Coeff = & Li2Lr;

  } else if ( characteristic ==  characteristic_Sd ) {
    R->InitCoeff     =  & SdInit ;
    R->ClearCoeff     =  & SdClear ;
    R->CopyCoeff     =  & SdCopy;
    R->MakeCoeff      =  & SdMakeCoeff ;
    R->Str2Coeff    =   & SdStr2Coeff;
    R->Cterm2Coeff = & SdCterm2Coeff;
    
    R->InvertCoeff    =  & SdInvert ;
    R->NegCoeff      =   & SdNeg;
    R->ChangeSignToCoeff   = & SdChangeSign;
    R->RandomizeCoeff    = & SdRandomize;
    
    R->AddCoeff    = & SdAdd ;
    R->SubCoeff     = & SdSub ;
    R->MulCoeff    = & SdMul ;
    R->DivCoeff    = & SdDiv ;
    R->ModCoeff    = & SdMod ;
        
    R->AreEqualsCoeff   = & AreSdEquals ;
    R->IsPositiveCoeff  = & IsSdPositive ;
    R->IsZeroCoeff   = & IsSdZero ;
    R->IsOneCoeff     = & IsSdOne ;
    R->IsMinusOneCoeff = & IsSdMinusOne ;
    R->Coeff2Cterm = & Sd2Cterm;
    R->Li2Coeff = & Li2Sd;

  }
  return R;
}

coc_bool ring_coeff_in_field(ring R)
{
  if (abs(R->characteristic)==1) return FALSE;
  if (R->characteristic > 1) return IsPrime(R->characteristic);
  return TRUE;
}


ordering ring_build_ordering(ring R)
{
  ordering to;
  InitZeroOrder(&to,R);
  switch (R->whatTO) {
  case lex_num : 
    SetLexOrder(to,R);
    break;

  case xel_num : 
    SetXelOrder(to,R);
    break;

  case dlex_num : 
    SetDegLexOrder(to,R->weights,R);
    break;

  case drev_num : 
    SetDegRevLexOrder(to,R->weights,R);
    break;

  case elim_num : /* must be already defined */
  default : 
    cocoa_internal_error(
      "unexpected undefined ordering in ring_build_ordering");
    break;
  }
  return to;
}


void ring_free_ordering(ring R)
{
  ClearOrder(&R->to,R);
#ifdef TRIPOLI_OR_EEE
  if (R->indetsNo != 0 && R->to != R->ot) ClearOrder(&R->ot,R);
#endif
}

void ring_free(ring R)
{
  register int i;
  if (R == ring_NULL) return;

  if (R->ring_name != NULL) free_str(strlen(R->ring_name),R->ring_name);
  indet_info_free(R->indets);

  ring_free_ordering(R);
  if (ring_local_env(R) != env_NULL) {
    env_free(ring_local_env(R));
  }
  weilist_free(ring_weilist(R));

#ifdef TRIPOLI
  if (R->indetsNo!=0) trnode_array_free(R->indetsNo+1,R->tripoli_dum);
#endif

#ifdef EEE
  for (i=0;i<=R->indetsNo;i++) free_eee_mon(eee_dummy(R,i));
  free_eee_mon_array(R->indetsNo+1,R->eee_dum);
#endif

  for (i=0;i<NUM_AUX_IV;i++) iv_free(R->aux_iv[i]);
  weight_free(R->weights);
  free_ring(R);
}

void ring_clear(ring R)
{
  env_free(ring_local_env(R));
  ring_set_local_env(R,env_init());
}

cterm characteristic2cterm(int c)
{
  cterm params[1],res_t;
  if (c>1) {
    params[0]=int2cterm(c);
    res_t=cterm_mk_classic(op_RING_ZP,1,params); }
  else switch(c) {
    case characteristic_Li:res_t=cterm_mk_NBS("Z"); break;
    case characteristic_Lr:res_t=cterm_mk_NBS("Q");break;
    case characteristic_Si:res_t=cterm_mk_NBS("ShZ");break;
    case characteristic_Sr:res_t=cterm_mk_NBS("ShQ");break;
    case characteristic_Sd:res_t=cterm_mk_NBS("ShD");break;
    default:res_t=cterm_mk_NBS("?");
  }
  return res_t;
}

cterm ring_poly2cterm(ring R)
{
  ring aux_R;
  cterm params[2],res_t;
  params[0]=indet_info2cterm(R->indets);
  params[1]=characteristic2cterm(R->characteristic);
  res_t=cterm_mk_classic(op_POLY_RING,2,params);
  return res_t;
}

cterm ring_env2cterm(ring R)
{
  cterm params[2];
  params[0]=ring_id2cterm(R);
  params[1]=ring2cterm(R);
  return cterm_mk_classic(op_IS_EQ,2,params);
}

cterm ring2cterm(ring R)
{
  cterm params[2],res_t;
  if (R == ring_NULL) return cterm_NULL;
  switch(R->rtype) {
    case ringtype_coeff:
      res_t=characteristic2cterm(R->characteristic);
      break;
    case ringtype_poly:
      res_t=ring_poly2cterm(R);
      break;
    case ringtype_quotient:
      params[0]=ring_poly2cterm(R);
      params[1]=module2cterm(R->quid,R);
      res_t=cterm_mk_classic(op_FRAC,2,params);
      break;
    default: res_t=cterm_NULL;
  }
  return res_t;
}
