/* $Id: rum.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#ifndef RUM_H
# include "rum.h"
#endif

typedef struct rum_aux {
  void ** slots;
  int top;
  int len;
} rum_stack;

long int rum_stat_counter;

rum_stack rums[RUM_N_STACKS+1];

short rum_status[RUM_MAX_ALLOCATION+1]; 
/* very large number: if the program attempt to allocate more then
   RUM_MAX_ALLOCATION bytes at a time, you cannot use RUM */

#ifdef RUM_DEBUG
typedef struct rum_stat_aux {
  coc_bool do_stat;
  int found_full;
  int found_empty;
  int level_max;
  int level_min;
  int level_init;
  long int mallocs;
  long int frees;
} rum_stat_aux;

rum_stat_aux rum_stat[RUM_N_STACKS+1];
#endif

#define rum_NORMAL 0
#define rum_EMPTY 1
#define rum_FULL 2
#define rum_FULL_AND_EMPTY 3

void rum_empty(int rum);

/* TODO: examinate the possibility of using shifts (?) */

#define rum_is_empty(rum) (rum_status[rum] & rum_EMPTY)
#define rum_is_full(rum) (rum_status[rum] & rum_FULL)

void rum_init_all(void){
  register int i;
  static coc_bool already_called=FALSE;
  if (already_called) return;
  already_called=TRUE;

#ifdef RUM_DEBUG
  rum_stat_counter=0;
#endif

  for (i=0;i<RUM_N_STACKS;i++) {
    rums[i].len=0;
    rums[i].top=-1;

#ifdef RUM_DEBUG
#ifdef RUM_DEBUG_SHOW_UNUSED
    rum_stat[i].do_stat=TRUE;
#else
    rum_stat[i].do_stat=FALSE;
#endif
#endif

  }
  for (i=0;i<=RUM_MAX_ALLOCATION;i++)
    rum_status[i]=rum_FULL_AND_EMPTY;

}

void rum_init(int rum, int size) {

  if (rum<RUM_N_STACKS) {
    int i;
    if (rums[rum].len!=0) {
      size = MIN(RUM_MAX_SIZE,rums[rum].len+size);
      if (size==rums[rum].len) {
#ifdef DEBUG_RUM
        printf("--[ DEBUG: rum_init %3d, SAME size : %6d ]--\n",rum,size);
#endif
      } else {
        /* rum_empty(rum); */
        rums[rum].slots=(void **)realloc(rums[rum].slots,size*sizeof(void *));
        rums[rum].len=size; 
        /* rums[rum].top=-1; */
#ifdef DEBUG_RUM
        printf("--[ DEBUG: rum_init %3d, NEW size is %6d ]--\n",rum,size);
#endif
      } }
    else {
      size = MIN(RUM_MAX_SIZE,size);
      rums[rum].slots=(void **)malloc(size*sizeof(void *));
      rums[rum].len=size; 
      rums[rum].top=-1;
      rum_status[rum]=rum_EMPTY;
      i=MIN(RUM_INIT_LOAD,size);
      while (--i>0) rum_free(rum,(void *)malloc(rum));
#ifdef DEBUG_RUM
      printf("--[ DEBUG: rum_init %3d, the size is %6d ]--\n",rum,size);
#endif
#ifdef RUM_DEBUG
      rum_stat[rum].do_stat=TRUE;
      rum_stat[rum].found_full=0;
      rum_stat[rum].found_empty=0;
      rum_stat[rum].level_max=rums[rum].top;
      rum_stat[rum].level_min=rums[rum].top;
      rum_stat[rum].level_init=rums[rum].top;
      rum_stat[rum].mallocs=rums[rum].top;
      rum_stat[rum].frees=rums[rum].top;
#endif
    }

  }
}

void * rum_malloc(int size)
{

  register rum_stack * rum = &rums[size];

#ifdef RUM_DEBUG
  if (size<RUM_N_STACKS) {
    rum_stat[size].mallocs++;
    rum_stat[size].level_min=MIN(rum_stat[size].level_min,rums[size].top-1); }
#endif

  if (rum_is_empty(size)) {
#ifdef RUM_DEBUG
    rum_stat[size].found_empty++;
#endif
    return (void *)malloc(size); }

#ifdef RUM_DEBUG
  if (rum_stat_counter++==RUM_STAT_FREQ) {
    rum_stat_counter=0;
    rum_show_statistics(0); }
#endif

  if (!rum->top) {
    register int i;
    i=MIN(RUM_RELOAD,rums[size].len);
    while (--i>0) rum_free(size,(void *)malloc(size));

#ifdef RUM_DEBUG
  rum_stat[size].found_empty++;
#endif

    /* rum_status[size]=rum_EMPTY; */
    /* printf("Room # %d was empty.\n",size);*/ }
  else
    rum_status[size]=rum_NORMAL;
  return rum->slots[rum->top--];
}


void rum_free(int size, void * p)
{

  register rum_stack * rum = &rums[size];

#ifdef DEBUG_CHECK_NIL_FREE
  if (p==0) {
    printf("rum_free: attempt to free a NIL pointer\n"); abort();
  }
#endif

#ifdef RUM_DEBUG
  if (size<RUM_N_STACKS) {
    rum_stat[size].frees++;
    rum_stat[size].level_max=MAX(rum_stat[size].level_max,rums[size].top+1); }
#endif

  if (rum_is_full(size)) {
#ifdef RUM_DEBUG
    rum_stat[size].found_full++;
#endif
    free(p); }
  else {

#ifdef RUM_DEBUG
  if (rum_stat_counter++==RUM_STAT_FREQ) {
    rum_stat_counter=0;
    rum_show_statistics(0); }
#endif

    if (rum->top==rum->len-2) {
      rum_status[size]=rum_FULL;
#ifdef RUM_DEBUG
      rum_stat[size].found_full++;
#endif
      /* printf("Room # %d is full.\n",size);*/ }
    else
      rum_status[size]=rum_NORMAL;
    rum->slots[++rum->top]=p;
  }
}


/* auxiliary functions */

void rum_empty(int rum)
{
  register int i;
  if (rum_status[rum] != rum_FULL_AND_EMPTY) {
    for (i=0;i<=rums[rum].top;i++) free(rums[rum].slots[i]);
    free(rums[rum].slots);
    rum_status[rum]=rum_FULL_AND_EMPTY;
    rums[rum].len=0;
    rums[rum].top=-1; }
}

void rum_show_statistics( int n )
{
#ifdef RUM_DEBUG
  register int i;
  if (n==0) printf("------ RUM STATISTICS ------\n");
  else printf("------ RUM STATISTICS (traffic >= %d) ------\n",n);
  for (i=0;i<RUM_N_STACKS;i++)
    if ( rum_stat[i].do_stat && (rum_stat[i].mallocs+rum_stat[i].frees)>=n) {
      printf("-- RUM %3d (%d) : ",i,rums[i].len);
      printf("%dM, ",rum_stat[i].mallocs);
      printf("%dF, ",rum_stat[i].frees);
      printf("Empty/Full %d/%d, ",rum_stat[i].found_empty,
        rum_stat[i].found_full);
      printf("L: ");
      /* printf("Initial = %d ",rum_stat[i].level_init); */
      printf("%d < %d < %d\n",rum_stat[i].level_min, rums[i].top, 
        rum_stat[i].level_max); }
  printf("\n");
#else
  printf("------ RUM STATISTICS NOT AVAILABLE ------\n");
#endif
}
