/* subst.c : 0.1 : 19 Oct 94 : ADT substitutions */

/*------------------------------------------------------------------------------

	part of project cocoa

------------------------------------------------------------------------------*/

#include "cocdefs.h"

# include "coeffs.h"
# include "ivectors.h"
# include "monomial.h"
# include "polys.h"
# include "pvector.h"
# include "ring.h"
# include "subst.h"
# include "terms.h"
# include "weights.h"

subst subst_init(ring R)
{
  subst res_s;
  res_s=malloc_subst();
  res_s->filter=iv_init(R->indetsNo);
  res_s->sub=malloc_subst_comp(R->indetsNo);
  return res_s;
}

void subst_free(subst s,ring R)
{
  unknown u;
  for (u=0; u<ring_indets_num(R); u++) 
    if (subst_is_defined(s,u))
      arpoly_free(subst_get(s,u),R);
  iv_free(s->filter);
  free_subst_comp(ring_indets_num(R),s->sub);
  free_subst(s);
}

arpoly subst_in_term(term t, subst s, ring R)
{
  unknown u;
  expType exp;
  arpoly res_p,aux_p,aux_p1;
  res_p=arpoly_one(R);
  for (u=0;u<R->indetsNo;u++) {
    exp=term_get_true_nth(t,u+1,R);
    if (exp!=0) {
      if (subst_is_defined(s,u))
        aux_p=arpoly_power(subst_get(s,u),exp,R);
      else
        aux_p=arpoly_unknown_pow2arpoly(u,exp,R);
      aux_p1=arpoly_mult(aux_p,res_p,R);
      arpoly_free(res_p,R);
      arpoly_free(aux_p,R);
      res_p=aux_p1; } }
  return res_p;
}

arpoly subst_in_monomial(monomial m, subst s, ring R)
{
  return arpoly_this_x_coeff(
    subst_in_term(monomial_term(m),s,R),
    monomial_coeff(m),R);
}

arpoly subst_in_arpoly(arpoly p, subst s, ring R)
{
  monomial aux_ml,aux_ml1;
  arpoly res_p,aux_p;
  if ((p == NULL) || arpoly_is_nonzero_const(p,R)) return arpoly_dup(p,R);
  aux_ml=arpoly2monopoli(p,R);
  aux_ml1=aux_ml;

  res_p=NULL;
  while (aux_ml != NULL) {
    aux_p=subst_in_monomial(aux_ml,s,R);
    res_p=arpoly_sum_these(res_p,aux_p,R);
    aux_ml=monomial_next(aux_ml); }
  monopoli_free(aux_ml1,R);
  return res_p;
}
