/* $Id: terms.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"
#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "indet.h"
#include "ivectors.h"
#include "mathcoc.h"
#include "operator.h"
#include "terms.h"
#include "ring.h"

/* Check if a term is pure, i.e. is a power of a single indeterminate */
/* The term 1 is regarded as being pure.                              */
coc_bool term_ispure(term t)
{
  int nexps = 0;
  int i;
  for (i=1;i<=iv_get_len(t);i++)
    if (term_get_nth(t,i) > 0 && nexps++ > 0) return FALSE;
  return TRUE;

}

term term_true(term t, ring R)
{
  int i;
  term res_t;
  res_t=term_init(iv_get_len(t));
  for (i=1;i<=iv_get_len(t);i++)
    term_put_nth(res_t,i,iv_get_nth(t,i)/iv_get_nth(ring_weights(R),i));
  return res_t;
}

ivector term_log(term t, ring R)
{
  register int i;
  ivector res_v;
  res_v=iv_init(iv_get_len(t));
  for (i=1;i<=iv_get_len(t);i++)
    iv_set_nth(res_v,i,term_get_nth(t,i)/iv_get_nth(R->weights,i));
  return res_v;
}

/* return ANSWER_BOTH iff t1==t2 */
/* return ANSWER_FIRST iff t1>t2 */
/* return ANSWER_SECOND iff t2>t1 */
/* return ANSWER_NC otherwise (never used if total ordering) */

int term_which_gt_lex(term t1, term t2, ordering ord, ring R)
{
  register int i = 1;
 
  while ( iv_get_nth(t1,i)==iv_get_nth(t2,i) )
    if ( i++==iv_get_len(t1) )  return ANSWER_BOTH;
  if (iv_get_nth(t1,i) > iv_get_nth(t2,i)) return ANSWER_FIRST;
  return ANSWER_SECOND;
}

int term_which_gt_xel(term t1, term t2, ordering ord, ring R)
{
  register int i;
  for (i=iv_get_len(t1);i>=1;i--) {
    if (iv_get_nth(t1,i) > iv_get_nth(t2,i)) return ANSWER_FIRST;
    if (iv_get_nth(t1,i) < iv_get_nth(t2,i)) return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

int term_which_gt_deglex(term t1, term t2, ordering ord, ring R)
{
  register int i;
  if (term_degree(t1)>term_degree(t2)) return ANSWER_FIRST;
  if (term_degree(t1)<term_degree(t2)) return ANSWER_SECOND;
  for (i=1;i<iv_get_len(t1);i++) {
    if (iv_get_nth(t1,i) > iv_get_nth(t2,i)) return ANSWER_FIRST;
    if (iv_get_nth(t1,i) < iv_get_nth(t2,i)) return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

int term_which_gt_degrevlex(term t1, term t2, ordering ord, ring R)
{
  register int i;
  if (term_degree(t1)!=term_degree(t2)) 
    if (term_degree(t1)<term_degree(t2)) return ANSWER_SECOND;
      else return ANSWER_FIRST;
  for (i=iv_get_len(t1);i>1;i--) {
    if (iv_get_nth(t1,i) != iv_get_nth(t2,i))
      if (iv_get_nth(t1,i) < iv_get_nth(t2,i)) return ANSWER_FIRST;
        else return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

int term_which_gt_degrevlex_opt(term t1, term t2)
{
  register int i;
  for (i=iv_get_len(t1);i>1;i--) {
    if (iv_get_nth(t1,i) != iv_get_nth(t2,i))
      if (iv_get_nth(t1,i) < iv_get_nth(t2,i)) return ANSWER_FIRST;
        else return ANSWER_SECOND; }
  return ANSWER_BOTH;
}

int term_which_gt_elim(term t1, term t2, ordering ord, ring R)
{
  register int i,j;
  int n,s1,s2; 
  unknown u,u2;
  u2=R->last_u+1;
  for (u=R->first_u+1,s1=s2=0;u<=u2;u++) {
    s1+=iv_get_nth(t1,u);
    s2+=iv_get_nth(t2,u);
  }
  if (s1>s2) return ANSWER_FIRST;
  if (s2>s1) return ANSWER_SECOND;
  if (term_degree(t1)>term_degree(t2)) return ANSWER_FIRST;
  if (term_degree(t1)<term_degree(t2)) return ANSWER_SECOND;
  n=iv_get_len(t1);
  for (i=2;i<n;i++) {
    j=(int)(ord[i][n]+1);
    if (iv_get_nth(t1,j) > iv_get_nth(t2,j)) return ANSWER_SECOND;
    if (iv_get_nth(t1,j) < iv_get_nth(t2,j)) return ANSWER_FIRST; }
  return ANSWER_BOTH;
}

int term_which_gt_other(term t1, term t2, ordering ord, ring R)
{
  register short int i;
  register iv_elem temp;
  register int l=iv_get_len(t1);
  if ( term_eq(t1,t2) )
    return ANSWER_BOTH;
  for ( i = 1; i <= l; i++ )
    iv_put_nth(R->aux_iv[0],i,term_get_nth(t1,i) - term_get_nth(t2,i));
  i = 0;
  while ( i < l ) {
    temp = term_scalar( i, R->aux_iv[0], ord);
    if ( temp > 0 )
      return ANSWER_FIRST;
    else if (temp < 0 )
      return ANSWER_SECOND;
    else 
      i++;
  }
  return ANSWER_BOTH;
}

iv_elem term_scalar(int i, term t, ordering ord)
{
  register toElem * el;
  register iv_elem result, j;
  register int l=iv_get_len(t);
  el = ord[i];
  if ( el[ l ] <  0 ) {
    for ( result = 0, j = 0; j < l ; j++ )
      result += el[j] * term_get_nth(t,j+1);
    return result;
  } else {
    j = el[l];
    return el[ j ] * term_get_nth(t,j+1);
  }
}

ivec_elem ivec_term_scalar(int i, ivec t, ordering ord)
{
  register toElem * el;
  register ivec_elem result, j;
  register int l=ivec_len(t);
  el = ord[i];
  if ( el[ l ] <  0 ) {
    for ( result = 0, j = 0; j < l ; j++ )
      result += el[j] * ivec_nth(t,j+1);
    return result;
  } else {
    j = el[l];
    return el[ j ] * ivec_nth(t,j+1);
  }
}

term cterm2term(cterm t, ring R)
{  
  /* R poly_ring */
  /* t=term(p1:pow,...,pn:pow) */
  cterm_list params=cterm_params(t);
  cterm_list subparams;
  int arity=cterm_arity(t);
  unknown aux_u;
  short int exponent;
  term res_t;
  register int i;
  res_t=term_init(R->indetsNo);
  for (i=0;i<arity;i++) {
    subparams=cterm_params(params[i]);
    aux_u=mi_cterm2unknown(subparams[0],R);
    exponent=cterm2int(subparams[1]);
    term_put_nth(res_t,aux_u+1,exponent); }
  return res_t;
}

cterm term2cterm(term t,ring R) 
{
  cterm aux_t;
  register int i;
  ctbuf b;
  cterm params[2];
  b=ctbuf_empty();
  if (term_degree(t)!=0) {
    for (i=0;i<R->indetsNo;i++)
      if (term_get_nth(t,i+1)!=0) { 
        params[0]=indet_info_get_cterm(R->indets,(unknown)i);
        params[1]=int2cterm(term_get_nth(t,i+1)/iv_get_nth(R->weights,i+1));
        aux_t=cterm_mk_classic(op_POWER,2,params);
        ctbuf_append(b,aux_t); }
    }
  return ctbuf_mk_cterm(b,op_TERM);
}

unknown term_first_nonzero_indet(term t, ring R)
{
  /* returns the minimun indet with exponent > 0 */
  /* returns R->indetsNo+1 if all exponents are 0 */
  register int i;
  if (term_degree(t)==0) return R->indetsNo+1;
  for (i=1;i<=R->indetsNo;i++)
    if (term_get_nth(t,i)!=0) return i;
  return R->indetsNo+1;
}

void printf_term(term t, ring R)
{
  cterm aux_t;
  aux_t=term2cterm(t,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

term term_der(term t, unknown x)
{
  int n;
  term res_t;
  n=term_get_nth(t,x+1);
  if (n==0) return term_NULL;
  res_t=term_dup(t);
  term_put_nth(res_t,x+1,n-1);
  return res_t;
}

term unknown_pow2term(unknown u, expType n, ring R)
{
  term res_t;
  res_t=term_init(ring_indets_num(R));
  term_put_nth(res_t,u+1,n*iv_get_nth(R->weights,u+1));
  return res_t;
}
