/* $Id: terms.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef TERMS_H
#define TERMS_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#ifndef COCTYPES_H
# include "coctypes.h"
#endif


#define term_is_gt(t1,t2,R) (R->term_which_gt(t1,t2,R->to,R) == ANSWER_FIRST)

#define term_is_one(t) (iv_get_sum(t)==0)

#define term_this_x_unknown_pow(t,x,n) iv_incr_nth(t,x+1,n)

#define term_inverse(t) iv_change_sign(t)

#define term_get_first(t) term_get_nth(t,1)
#define term_get_nth(t,n) iv_get_nth(t,n)
#define term_put_nth(t,n,e) iv_put_nth(t,n,e)
#define term_set_nth(t,n,e) iv_set_nth(t,n,e)

#define term_get_true_nth(t,n,R) (iv_get_nth(t,n)/iv_get_nth(R->weights,n))

#define term_one(R) term_init(R->indetsNo)

#define term_init(n) iv_init(n)
/* initialize a term */

#define term_free(t) iv_free(t)

#define term_assign(dest,src) iv_assign(dest,src)
/* dest := src */

#define term_fill(t,e) iv_fill(t,e)

#define term_create_and_fill(n,e) iv_create_and_fill(n,e)

#define term_dup(t) iv_dup(t)
/* returns a duplicate of t */

#define term_lcm(t1,t2) iv_maxs(t1,t2)
/* returns lcm(t1,t2) */

#define term_deglcm(t1,t2) iv_sum_maxs(t1,t2)

#define term_gcd(t1,t2) iv_mins(t1,t2)
/* returns gcd(t1,t2) */

#define term_degree(t) (t==term_NULL ? 0 : iv_get_sum(t))
/* return the natural degree of t (sum of exponents) */

coc_bool term_ispure(term t);

#define term_set_degree(t,D) iv_set_total(t,D)

#define term_divides(t1,t2) iv_all_le(t1,t2)
/* TRUE iff t1|t2 */

#define term_which_divides(t1,t2) iv_which_all_le(t1,t2)
/* return ANSWER_BOTH iff t1==t2 */
/* return ANSWER_FIRST iff t1|t2 */
/* return ANSWER_SECOND iff t2|t1 */
/* return ANSWER_NC otherwise */

#define term_eq(t1,t2) (iv_get_sum(t1)==iv_get_sum(t2) && iv_eq(t1,t2))
/* TRUE iff t1==t2 */

#define term_coprime(t1,t2) iv_coprime(t1,t2)
/* TRUE iff no common indeterminates in t1 and t2 */

#define term_set_mult(t,t1,t2) iv_set_sum(t,t1,t2)

#define term_mult(t1,t2) iv_sum(t1,t2)
/* return t1*t2 */

#define term_mult_and_assign(t1,t2) iv_sum_and_assign(t1,t2)
/* return t1*=t2 */

#define term_div_and_assign(t1,t2) iv_subtract_and_assign(t1,t2)
/* return t1*=t2 */

#define term_div(t1,t2) iv_subtract(t1,t2)

#define term_square(t) iv_sum(t,t)
/* return t^2 */

#define term_square_and_assign(t) iv_sum_and_assign(t,t)
/* return t^2 in place of t */

#define term_take_lcm(t,t1,t2) iv_fill_maxs(t,t1,t2)
/* t = lcm(t1,t2) */

#define term_take_div(t,t1,t2) iv_fill_subtract(t,t1,t2)
/* t = t1/t2 */

#define term_map(t,w1,w2,m) iv_map(t,w1,w2,m)

/* The following 4 functions compare two terms and: */

/* return ANSWER_BOTH iff t1==t2 */
/* return ANSWER_FIRST iff t1>t2 */
/* return ANSWER_SECOND iff t2>t1 */
/* return ANSWER_NC otherwise (never used if total ordering) */

ivector term_log(term t, ring R);

int term_which_gt_lex(term t1, term t2, ordering ord, ring R);
int term_which_gt_xel(term t1, term t2, ordering ord, ring R);

int term_which_gt_deglex(term t1, term t2, ordering ord, ring R);

int term_which_gt_degrevlex(term t1, term t2, ordering ord, ring R);

int term_which_gt_elim(term t1, term t2, ordering ord, ring R);

int term_which_gt_other(term t1, term t2, ordering ord, ring R);

term term_true(term t, ring R);

term cterm2term(cterm t, ring R);
/* return the term represented by cterm t */

cterm term2cterm(term t, ring R);
/* return the cterm corresponding to the term t */

iv_elem term_scalar(int i, term t, ordering ord);
/* return the scalar product of t and the i-th row of ord */

ivec_elem ivec_term_scalar(int i, ivec t, ordering ord);
/* return the scalar product of t and the i-th row of ord */

void printf_term(term t, ring R);

term term_der(term t, unknown x);

term unknown_pow2term(unknown u, expType n, ring R);

unknown term_first_nonzero_indet(term t,ring R);


/* BIGATTI: START */

#define term_auto_increase_nth(t,n)   iv_auto_increase_nth(t,n)

#define term_colon(t1,t2)  iv_bounded_subtract(t1,t2)

#define term_pure_power(t,index,exp)  iv_almost_null(t,index,exp)

#define term_get_indetsNo(t) iv_get_len(t)

/* BIGATTI: END */


#endif  /* TERMS_H */
