/* $Id: tripoli.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

int tripoli_c_dummy;

#ifdef TRIPOLI

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "coeffs.h"
#include "ctbuf.h"
#include "ivectors.h"
#include "monomial.h"
#include "mlist.h"
#include "ordering.h"
#include "operator.h"
#include "options.h"
#include "pipe.h"
#include "polys.h"
#include "scalars.h"
#include "terms.h"
#include "tripoli.h"
#include "weights.h"

int tripoli_eq_l_counter;
int tripoli_eq_counter;

mon mon_dummy,mon_dummy1;

/* --------------------------------- *\
            POLYNOMIAL
\* --------------------------------- */

/* ENTRY FUNCTIONS */

tripoli tripoli_remove_lm(tripoli p, ring R)
{
  monomial aux_m;
#ifdef DEBUG                                                           
  if (p == NULL) {                       
    INTERNAL_ERROR("tripoli_remove_lm(p,R): p is NULL");
    return p;                            
  }
#endif                                  
  OPTIMIZE("tripoli_remove_lm");                  
  aux_m=tripoli_leading_monomial(p,R);
  monomial_set_next(aux_m,NULL);                                                
  return
   tripoli_sum_this_monomial_list_to_this_tripoli(
     monomial_opposite(aux_m,R),p,R);
}

tripoli tripoli_sum_this_monomial_list_to_this_tripoli(monomial m,
  tripoli p, ring R)
{
  monomial next_m;
  while (m != NULL){
    next_m = monomial_next(m);
    p=tripoli_add_this_monomial( p, m, R );
    m=next_m; }
  return p;
}

tripoli tripoli_der(tripoli p, unknown x, ring R)
{
  monomial aux_ml,aux_ml1,aux_m;
  tripoli res_p,aux_p;
  aux_ml=tripoli2monopoli(p,R);
  aux_ml1=aux_ml;
  if (aux_ml == NULL) return p;
  aux_m=monomial_der(aux_ml,x,R);
  if (aux_m!=NULL)
    res_p=this_monomial2tripoli(aux_m,R);
  else
    res_p=NULL;
  aux_ml=monomial_next(aux_ml);
  while (aux_ml != NULL) {
    aux_m=monomial_der(aux_ml,x,R);
    if (aux_m!=NULL) {
      aux_p=this_monomial2tripoli(aux_m,R);
      res_p=tripoli_sum_these(res_p,aux_p,R); }
    aux_ml=monomial_next(aux_ml); }
  monopoli_free(aux_ml1,R);
  return res_p;
}

tripoli tripoli_sum_this_monopoli(monopoli m, ring R)
{
  tripoli res_p=NULL;
  monomial next_m;
  while (m != NULL){
    next_m = monomial_next(m);
    tripoli_add_this_monomial_unique(m,&res_p,R);
    m=next_m; }
  return res_p;
}

tripoli tripoli_sum_monopoli(monopoli m, ring R)
{
  tripoli res_p=NULL;
  monomial next_m;
  OPTIMIZE("tripoli_sum_monopoli");
  while (m != NULL) {
    next_m = monomial_next(m);
    tripoli_add_this_monomial_unique(monomial_dup(m,R),&res_p,R); /* OPT here */
    m=next_m; }
  return res_p;
}

tripoli tripoli_sum_this_monomial(tripoli p, monomial m, ring R)
{
  TODO("tripoli_sum_this_monomial");
  return p;
}

int tripoli_size(tripoli p, ring R)
{
  if (p == NULL) return sizeof(tripoli);
  return tripoli_size_l(p,0,R)+sizeof(tripoli);
}

term tripoli_this_leading_term(tripoli p, ring R)
{
  TODO("tripoli_this_leading_term (note that is impossible)");
  return term_NULL;
}

olist tripoli2sppnum(tripoli p, ring R)
{
  monopoli aux_m;
  olist res_ol;
  OPTIMIZE("tripoli2sppnum");
  aux_m=tripoli2monopoli(p,R);
  res_ol=monopoli2sppnum(aux_m,R);
  /* monopoli_free(aux_m,R); */
  return res_ol;
}

int tripoli_lm_which_gt(tripoli p1, tripoli p2, ring R)
{
  tripoli aux_p1,aux_p2;
  trnode aux_tp1,aux_tp2;
  register int i;
  aux_p1=p1;
  aux_p2=p2;
  for (i=1;i<=R->level;i++) {
    aux_tp1=tripoli_trnode(aux_p1);
    aux_tp2=tripoli_trnode(aux_p2);
    if (trnode_scalar(aux_tp1)<trnode_scalar(aux_tp2)) return ANSWER_SECOND;
    else if (trnode_scalar(aux_tp1)>trnode_scalar(aux_tp2))
      return ANSWER_FIRST;
    else {
      aux_p1=trnode_homo_comp(aux_tp1);
      aux_p2=trnode_homo_comp(aux_tp2); } }
  if (R->level==R->indetsNo) return ANSWER_FIRST;
  return scalars_which_gt_lex(
    mon_scalars(tripoli_monomials(aux_p1)),
    mon_scalars(tripoli_monomials(aux_p2)));
}

/* TODO: optimize the following (scan the poly, not get scalars) */
coc_bool tripoli_is_nonzero_const(tripoli p, ring R)
{
  scalars aux_ss;
  register int i;
  if (p == NULL) return FALSE;
  aux_ss=tripoli_leading_scalars(p,R);
  for (i=1;i<=R->indetsNo;i++)
    if (scalars_nth(aux_ss,i)) {scalars_free(aux_ss); return FALSE; }
  scalars_free(aux_ss);
  return TRUE;
}

tripoli scalars2tripoli(scalars ss,ring R)
{
  tripoli aux_p;
  register int i;
  int n,j;
  scalars aux_ss;
  mon aux_m;

  n=R->indetsNo;
  j=R->level;
  aux_ss=scalars_init(n-j);
  for (i=j+1;i<=n;i++) 
    scalars_set_nth(aux_ss,i-j,scalars_nth(ss,i));
  mon_set_mk(aux_m,coeff_one(R),aux_ss);
  tripoli_set_monomials(aux_p,aux_m);

  for (i=j;i>0;i--) {
    scalar s;
    s = scalars_nth(ss,i);
    tripoli_set_trnode(aux_p,trnode_mk(s,aux_p)); }
  return aux_p;
}

tripoli unknown_pow2tripoli(unknown u,expType n,ring R)
{
  term aux_t;
  scalars aux_ss;
  tripoli res_p;
  aux_t=term_init(R->indetsNo);
  term_put_nth(aux_t,u+1,n*weight_nth(R->weights,u+1));
  aux_ss=term2scalars(aux_t,R);
  res_p=scalars2tripoli(aux_ss,R);
  term_free(aux_t);
  scalars_free(aux_ss);
  return res_p;
}

tripoli tripoli_coeff2tripoli(coeff c,ring R)
{
  coeff aux_c;
  aux_c=coeff_dup(c,R);
  return this_coeff2tripoli(aux_c,R);
}

tripoli this_coeff2tripoli(coeff c,ring R)
{
  tripoli aux_p;
  scalars aux_ss;
  register int i;
  if (coeff_is_zero(c,R)) {
    R->ClearCoeff(&c);
    return NULL; }
  aux_ss=scalars_init(R->indetsNo - R->level);
  tripoli_set_monomials(aux_p,mon_mk(c,aux_ss));
  for (i=1;i<=R->level;i++)
    tripoli_set_trnode(aux_p,trnode_mk(0,aux_p));
  return aux_p;
}

term tripoli_leading_term(tripoli p, ring R)
{
  term res_t;
  scalars aux_ss;
  aux_ss=tripoli_leading_scalars(p,R);
  res_t=scalars2term(aux_ss,R);
  scalars_free(aux_ss);
  return res_t;
}

tripoli tripoli_only_leading_term(tripoli p,ring R)
{ 
  scalars aux_ss;
  tripoli res_p;
  mon aux_m;
  aux_ss=tripoli_leading_scalars(p,R);
  mon_set_mk(aux_m,coeff_one(R),aux_ss);
  res_p=mon2tripoli(aux_m,R);
  mon_free(aux_m,R);
  return res_p;
}

coc_bool tripoli_eq(tripoli p1, tripoli p2, ring R)
{
  coc_bool res_b;
  tripoli_eq_counter++;
  if (p1 == NULL) return (p2 == NULL);
  
  /* else: poly != 0 */

  return tripoli_eq_l(p1,p2,0,R);
}

tripoli tripoli_this_x_opp_term(tripoli p, term t, ring R)
{
  /* return -t*p and destroy p */
  tripoli res_p;
  scalars aux_ss,ss;
  register int i;

  if (p == NULL) return NULL;

  if (term_is_one(t)) {tripoli_change_sign(p,R);return p;}

  aux_ss=term2scalars(t,R);
  ss=scalars_init(R->indetsNo-R->level);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i-R->level,scalars_nth(aux_ss,i));
  res_p=tripoli_this_x_opp_scalars_l(p,ss,aux_ss,0,R);
  scalars_free(aux_ss);
  return res_p;
}

tripoli tripoli_this_x_term(tripoli p, term t, ring R)
{
  /* return -t*p and destroy p */
  tripoli res_p;
  scalars aux_ss,ss;
  register int i;

  if (p == NULL) return NULL;

  if (term_is_one(t)) return p;

  aux_ss=term2scalars(t,R);
  ss=scalars_init(R->indetsNo-R->level);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i-R->level,scalars_nth(aux_ss,i));
  res_p=tripoli_this_x_scalars_l(p,ss,aux_ss,0,R);
  scalars_free(aux_ss);
  return res_p;
}

tripoli tripoli_this_x_monomial(tripoli p, monomial m, ring R)
{
  mon aux_m;
  tripoli res_p;
  scalars aux_ss,ss;
  register int i;

  if (p == NULL) return NULL;

  if (monomial_is_zero(m,R)) {
    tripoli_free(p,R);
    return NULL; }
  
  aux_ss=term2scalars(monomial_term(m),R);
  ss=scalars_init(R->indetsNo-R->level);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i-R->level,scalars_nth(aux_ss,i));
  mon_set_mk(aux_m,monomial_coeff(m),ss);
  res_p=tripoli_this_x_mon_l(p,aux_m,aux_ss,0,R);
  scalars_free(mon_scalars(aux_m));
  scalars_free(aux_ss);
  free_mon(aux_m);
  return res_p;
}

tripoli tripoli_x_opp_term(tripoli p, term t, ring R)
{
  /* return -t*p */
  tripoli res_p;
  scalars aux_ss,ss;
  register int i;

  if (p == NULL) return NULL;

  if (term_is_one(t)) { tripoli_change_sign(tripoli_dup(p,R),R);return p; }
  
  aux_ss=term2scalars(t,R);
  ss=scalars_init(R->indetsNo-R->level);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i-R->level,scalars_nth(aux_ss,i));
  res_p=tripoli_x_opp_scalars_l(p,ss,aux_ss,0,R);
  scalars_free(aux_ss);
  scalars_free(ss);
  return res_p;
}

tripoli tripoli_x_monomial(tripoli p, monomial m, ring R)
{
  tripoli res_p;
  scalars ss,aux_ss;
  register int i;
  mon aux_m;

  if (p == NULL) return NULL;

  if (monomial_is_zero(m,R)) return NULL;
  
  aux_ss=term2scalars(monomial_term(m),R);
  ss=scalars_init(R->indetsNo-R->level);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i-R->level,scalars_nth(aux_ss,i));
  mon_set_mk(aux_m,monomial_coeff(m),ss);
  res_p=tripoli_x_mon_l(p,aux_m,aux_ss,0,R);
  scalars_free(aux_ss);
  scalars_free(mon_scalars(aux_m));
  free_mon(aux_m);
  return res_p;
}

monomial tripoli_extract_lm(tripoli *p, ring R)
{
  monomial res_m;
  res_m=tripoli_leading_monomial(*p,R);
  OPTIMIZE("tripoli_extract_lm");                  
  *p=tripoli_remove_lm(*p,R);
  return res_m;
}

monomial tripoli_leading_monomial(tripoli p, ring R)
{
  monomial res_m;
  mon aux_m;
  aux_m=tripoli_leading_mon(p,R);
  if (aux_m == NULL) return NULL;
  monomial_set_mk(res_m,mon_coeff(aux_m),scalars2term(mon_scalars(aux_m),R));
  scalars_free(mon_scalars(aux_m));
  free_mon(aux_m);
  return res_m;
}

coeff tripoli_this_leading_coeff(tripoli p, ring R)
{
  trnode aux_tp;
  int n; /* indicates the current level */
  mon aux_m;

  if (p == NULL) return coeff_zero(R);

  for (n=1;n<=R->level;n++) {
    aux_tp=tripoli_trnode(p);
    p=trnode_homo_comp(aux_tp); }

  aux_m=tripoli_monomials(p);
  return mon_coeff(aux_m);
}

int tripoli_deg(tripoli p, ring R)
{
  if (p == NULL) return 0;
  else {
    int deg_i;
    scalars ss;
    ss=scalars_init(R->indetsNo);
    deg_i=tripoli_deg_l(p,ss,0,R);
    scalars_free(ss);
    return deg_i; }
}

void tripoli_change_sign(tripoli p, ring R)
{
  if (p!=NULL) tripoli_change_sign_l(p,0,R);
}

int tripoli_len(tripoli p, ring R)
{
  if (p == NULL) return 1;
  else return tripoli_len_l(p,0,R);
}

cterm tripoli2cterm(tripoli p,ring R)
{
  if (p == NULL) return cterm_mk_const(op_POLYNOMIAL);
  else {
    ctbuf aux_b;
    cterm res_t;
    scalars aux_ss;
    aux_ss=scalars_init(R->indetsNo);
    aux_b=tripoli2ctbuf_l(p,aux_ss,0,R);
    res_t=ctbuf_mk_cterm(aux_b,op_POLYNOMIAL);
    scalars_free(aux_ss);
    return res_t; }
}


/* AUXILIARY FUNCTIONS */

tripoli tripoli_dup_l(tripoli p,int level,ring R)
{
  tripoli res_p;
  if (level==R->level) 
    tripoli_set_monomials(res_p,monomials_dup(tripoli_monomials(p),R));
  else
    tripoli_set_trnode(res_p,trnode_dup_l(tripoli_trnode(p),level,R));
  return res_p;
}

tripoli tripoli_mult_l(tripoli p1, tripoli p2, int level, ring R)
{
  tripoli res_p;
  trnode aux_tp,aux_tp2;
  if (p1 == NULL) return NULL;
  else if (p2 == NULL) return NULL;
  else if (level==R->level) return tripoli_monopoli_mult(p1,p2,R);

  /* else (level < R->level) */
  res_p=NULL;
  aux_tp=tripoli_trnode(p1); /* TODO: study a strategy to select which p */
  aux_tp2=tripoli_trnode(p2); /* TODO: study a strategy to select which p */
  do {
    tripoli aux_p;
    aux_p=tripoli_mult_homo_tripoli(aux_tp,aux_tp2,level,R);
    tripoli_addthis2this_l(&res_p,aux_p,level,R);
    aux_tp=trnode_next_comp(aux_tp);
  } 
  while (aux_tp != NULL);
  return res_p;
}


tripoli tripoli_merge_l(tripoli p1,tripoli p2,int level,ring R)
{
  tripoli res_p;
  trnode aux_tp, aux_tp1, aux_tp2, res_tp;

  if (p1 == NULL) return p2;
  else if (p2 == NULL) return p1;
  else if (level==R->level) {
    tripoli_set_monomials(res_p,
      monomials_merge_normal(tripoli_monomials(p1),tripoli_monomials(p2),R));
    return res_p; }

  /* else (level < R->level) */

  res_tp=tripoli_dummy(R,level);
  aux_tp=res_tp;
  aux_tp1=tripoli_trnode(p1);
  aux_tp2=tripoli_trnode(p2);
  do {
    trnode aux_tp1_next,aux_tp2_next;
    scalar s1,s2;
    s1=trnode_scalar(aux_tp1);
    s2=trnode_scalar(aux_tp2);
    if (s1==s2) {
      tripoli aux_p;
      aux_p=tripoli_merge_l(trnode_homo_comp(aux_tp1),
                             trnode_homo_comp(aux_tp2),level+1,R);
      aux_tp1_next=trnode_next_comp(aux_tp1);
      aux_tp2_next=trnode_next_comp(aux_tp2);
      if (aux_p != NULL) {
        trnode_set_homo_comp(aux_tp1,aux_p);
        trnode_set_next_comp(aux_tp,aux_tp1);
        aux_tp=trnode_next_comp(aux_tp); }
      else free_trnode(aux_tp1);
      free_trnode(aux_tp2);
      aux_tp1=aux_tp1_next;
      aux_tp2=aux_tp2_next; }
    else if (s1>s2) {
      trnode_set_next_comp(aux_tp,aux_tp1);
      aux_tp=trnode_next_comp(aux_tp);             
      aux_tp1=trnode_next_comp(aux_tp1); }
    else { /* s2>s1 */
      trnode_set_next_comp(aux_tp,aux_tp2);
      aux_tp=trnode_next_comp(aux_tp);
      aux_tp2=trnode_next_comp(aux_tp2); } }
  while((aux_tp1 != NULL) && (aux_tp2 != NULL));
  if (aux_tp1 == NULL)      
    trnode_set_next_comp(aux_tp,aux_tp2);
  else trnode_set_next_comp(aux_tp,aux_tp1);
  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));
  return res_p;
}

tripoli tripoli_monopoli_sum(tripoli p1, tripoli p2, ring R)
{
  tripoli res_p;
  tripoli_set_monomials(res_p,monomials_sum(tripoli_monomials(p1),
                                         tripoli_monomials(p2),R));
  return res_p;
}

tripoli tripoli_monopoli_mult(tripoli p1, tripoli p2, ring R)
{
  tripoli res_p;
  tripoli_set_monomials(res_p,monomials_mult(tripoli_monomials(p1),
                                          tripoli_monomials(p2),R));
  return res_p;
}

void tripoli_monopoli_add2this(tripoli * this,tripoli p, ring R)
{
  monomials_add2these(tripoli_monomials_star(*this),tripoli_monomials(p),R);
}

tripoli tripoli_sum_l(tripoli p1,tripoli p2,int level,ring R)
{
  scalar s1,s2;
  tripoli res_p;
  trnode aux_tp, aux_tp1, aux_tp2, res_tp;
  if (p1 == NULL) return trnode_dup_l(p2,level,R);
  else if (p2 == NULL) return tripoli_dup_l(p1,level,R);
  else if (level==R->level) return tripoli_monopoli_sum(p1,p2,R);
  /* level < R->level */
  res_tp=tripoli_dummy(R,level);
  aux_tp=res_tp;
  aux_tp1=tripoli_trnode(p1);
  aux_tp2=tripoli_trnode(p2);
  do {
    s1=trnode_scalar(aux_tp1);
    s2=trnode_scalar(aux_tp2);
    if (s1==s2) {
      tripoli aux_p;
      aux_p=tripoli_sum_l(trnode_homo_comp(aux_tp1),
                           trnode_homo_comp(aux_tp2),level+1,R);
      if (aux_p != NULL) {
        trnode_set_next_comp(aux_tp,trnode_mk(s1,aux_p));
        aux_tp=trnode_next_comp(aux_tp); }
      aux_tp1=trnode_next_comp(aux_tp1);
      aux_tp2=trnode_next_comp(aux_tp2); }
    else if (s1>s2) {
      tripoli aux_p;
      aux_p=tripoli_dup_l(trnode_homo_comp(aux_tp1),level+1,R);
      trnode_set_next_comp(aux_tp,trnode_mk(s1,aux_p));
      aux_tp=trnode_next_comp(aux_tp);
      aux_tp1=trnode_next_comp(aux_tp1); }
    else { /* s2>s1 */
      tripoli aux_p;
      aux_p=tripoli_dup_l(trnode_homo_comp(aux_tp2),level+1,R);
      trnode_set_next_comp(aux_tp,trnode_mk(s2,aux_p));
      aux_tp=trnode_next_comp(aux_tp);
      aux_tp2=trnode_next_comp(aux_tp2); } }
  while((aux_tp1 != NULL) && (aux_tp2 != NULL));
  if (aux_tp1 == NULL)      
    trnode_set_next_comp(aux_tp,trnode_dup_l(aux_tp2,level,R));
  else trnode_set_next_comp(aux_tp,trnode_dup_l(aux_tp1,level,R));
  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));
  return res_p;
}

tripoli tripoli_mult_homo_tripoli(trnode homo,trnode tp,int level,ring R)
{
  /* --------------------------------------------------------------- *\

  NOTE: homo could be a list of components, but we consider only the
        first one 

  \* --------------------------------------------------------------- */
  tripoli res_p;
  trnode aux_tp,aux_tp1,res_tp;
  res_tp=tripoli_dummy(R,level);
  aux_tp1=res_tp;
  aux_tp=tp;
  do {
    scalar aux_s;
    tripoli aux_p;
    aux_s=trnode_scalar(homo)+trnode_scalar(aux_tp);
    aux_p=tripoli_mult_l(trnode_homo_comp(homo),
                          trnode_homo_comp(aux_tp),level+1,R);
    trnode_set_next_comp(aux_tp1,trnode_mk(aux_s,aux_p));
    aux_tp1=trnode_next_comp(aux_tp1);
    aux_tp=trnode_next_comp(aux_tp);
  }
  while (aux_tp != NULL);
  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));
  return res_p;
}

cterm tripoli2cterm_l(tripoli p,int level,ring R)
{
  if (p == NULL) return cterm_mk_const(op_TRIPOLI);
  if (level==R->level) return tripoli_monopoli2cterm(p,R);
  return trnode2cterm_l(tripoli_trnode(p),level,R);
}

void special_printf_tripoli(tripoli p, ring R)
{
  special_printf_trnode_l(p,0,R);
}

void special_printf_trnode_l(tripoli p,int level,ring R)
{
  cterm aux_t;
  aux_t=tripoli2cterm_l(p,level,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

void tripoli_free_l(tripoli p,int level,ring R)
{
  if (p != NULL)
    if (level==R->level) monomials_free(tripoli_monomials(p),R);
    else trnode_free_l(tripoli_trnode(p),level,R);
}

mon tripoli_leading_mon(tripoli p, ring R)
{
  scalars res_ss,aux_ss;
  trnode aux_tp;
  int n; /* indicates the current level */
  mon aux_m; /*,res_m; */

  if (p == NULL) return NULL;
  res_ss=scalars_init(R->indetsNo);

  for (n=1;n<=R->level;n++) {
    aux_tp=tripoli_trnode(p);
    scalars_set_nth(res_ss,n,trnode_scalar(aux_tp));
    p=trnode_homo_comp(aux_tp); }

  aux_m=tripoli_monomials(p);
  aux_ss=mon_scalars(aux_m);
  for (;n <= R->indetsNo;n++)
    scalars_set_nth(res_ss,n,scalars_nth(aux_ss,n-R->level)); 
  return mon_mk(coeff_dup(mon_coeff(aux_m),R), res_ss);
}

scalars tripoli_leading_scalars(tripoli p,ring R)
{
  scalars res_ss,aux_ss;
  trnode aux_tp;
  int n; /* indicates the current level */
  mon aux_m;

  res_ss=scalars_init(R->indetsNo);
  if (p == NULL) return res_ss;

  for (n=1;n<=R->level;n++) {
    aux_tp=tripoli_trnode(p);
    scalars_set_nth(res_ss,n,trnode_scalar(aux_tp));
    p=trnode_homo_comp(aux_tp); }

  aux_m=tripoli_monomials(p);
  aux_ss=mon_scalars(aux_m);
  for (;n <= R->indetsNo;n++)
    scalars_set_nth(res_ss,n,scalars_nth(aux_ss,n-R->level));
  return res_ss;
}

tripoli tripoli_only_leading_mon(tripoli p,ring R)
{ 
  mon aux_m;
  tripoli res_p;
  aux_m=tripoli_leading_mon(p,R);
  if (aux_m == NULL) return this_monomial2tripoli(monomial_zero(R),R);
  res_p=mon2tripoli(aux_m,R);
  mon_free(aux_m,R);
  return res_p;
}

tripoli mon2tripoli(mon m,ring R)
{
  tripoli aux_p;
  register int i;
  scalars aux_ss;
  aux_p=tripoli_monopoli_mon2tripoli(m,R);
  aux_ss=mon_scalars(m);
  for (i=R->level;i>=1;i--)
    tripoli_set_trnode(aux_p,trnode_mk(scalars_nth(aux_ss,i),aux_p));
  return aux_p;
}

tripoli tripoli_monopoli_mon2tripoli(mon m, ring R)
{
  tripoli res_p;
  scalars aux_ss,ss;
  int i,j=R->level;
  ss=mon_scalars(m);
  aux_ss=scalars_init(R->indetsNo-j);
  for (i=j+1;i<=R->indetsNo;i++) 
    scalars_set_nth(aux_ss,i-j,scalars_nth(ss,i));
  tripoli_set_monomials(res_p,mon_mk(coeff_dup(mon_coeff(m),R),aux_ss));
  return res_p;
}

coc_bool tripoli_eq_l(tripoli p1,tripoli p2,int level,ring R)
{
  trnode aux_tp1, aux_tp2;

  tripoli_eq_l_counter++;

  if (level==R->level) 
    return monomials_eq(tripoli_monomials(p1),tripoli_monomials(p2),R);
  
  /* level < R->level */

  /* then compare homo-components */

  aux_tp1=tripoli_trnode(p1);
  aux_tp2=tripoli_trnode(p2);
  while ((aux_tp1 != NULL) && (aux_tp2 != NULL)) {
    if (trnode_scalar(aux_tp1) != trnode_scalar(aux_tp2)) return FALSE;
    if (!tripoli_eq_l(trnode_homo_comp(aux_tp1),
                      trnode_homo_comp(aux_tp2),level+1,R))
      return FALSE;
    aux_tp1=trnode_next_comp(aux_tp1);
    aux_tp2=trnode_next_comp(aux_tp2); }

  if ((aux_tp1 != NULL) || (aux_tp2 != NULL)) return FALSE;

  return TRUE;
}

tripoli tripoli_this_x_opp_scalars_l(tripoli p,scalars ss,
  scalars aux_ss,int level,ring R)
{
  scalar aux_s;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp;

  if (level==R->level) {
    tripoli_set_monomials(res_p,monomials_these_x_opp_scalars(tripoli_monomials(p),
      ss,R));
    return res_p; }
  
  /* level < R->level */
  
  /* first work on homo-components */

  aux_tp=tripoli_trnode(p);
  while (aux_tp != NULL) {
    aux_p=tripoli_this_x_opp_scalars_l(trnode_homo_comp(aux_tp),
      ss,aux_ss,level+1,R);
    trnode_set_homo_comp(aux_tp, aux_p);
    aux_tp=trnode_next_comp(aux_tp); }

  /* finally adjust scalars of homo-components (i.e. structure of poly) */

  aux_s=scalars_nth(aux_ss,level+1);
  aux_tp=res_tp;
  while (aux_tp != NULL) {
    trnode_set_scalar(aux_tp,trnode_scalar(aux_tp)+aux_s);
    aux_tp=trnode_next_comp(aux_tp); }

  tripoli_set_trnode(res_p,res_tp);

  return res_p;
}


tripoli tripoli_this_x_scalars_l(tripoli p,scalars ss,
  scalars aux_ss,int level,ring R)
{
  scalar aux_s;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp;

  if (level==R->level) {
    tripoli_set_monomials(res_p,monomials_these_x_scalars(tripoli_monomials(p),
      ss,R));
    return res_p; }
  
  /* level < R->level */
  
  /* first work on homo-components */

  aux_tp=tripoli_trnode(p);
  while (aux_tp != NULL) {
    aux_p=tripoli_this_x_scalars_l(trnode_homo_comp(aux_tp),
      ss,aux_ss,level+1,R);
    trnode_set_homo_comp(aux_tp, aux_p);
    aux_tp=trnode_next_comp(aux_tp); }

  /* finally adjust scalars of homo-components (i.e. structure of poly) */

  aux_s=scalars_nth(aux_ss,level+1);
  aux_tp=res_tp;
  while (aux_tp != NULL) {
    trnode_set_scalar(aux_tp,trnode_scalar(aux_tp)+aux_s);
    aux_tp=trnode_next_comp(aux_tp); }

  tripoli_set_trnode(res_p,res_tp);

  return res_p;
}



tripoli tripoli_this_x_mon_l(tripoli p,mon m,scalars ss,int level,ring R)
{
  scalar aux_s;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp;

  if (level==R->level) {
    tripoli_set_monomials(res_p,monomials_these_x_mon(tripoli_monomials(p),m,R));
    return res_p; }
  
  /* level < R->level */
  
  /* first work on homo-components */

  aux_tp=tripoli_trnode(p);
  while (aux_tp != NULL) {
    aux_p=tripoli_this_x_mon_l(trnode_homo_comp(aux_tp),m,ss,level+1,R);
    trnode_set_homo_comp(aux_tp, aux_p);
    aux_tp=trnode_next_comp(aux_tp); }

  /* then remove zero components */

  res_tp=trnode_rm_zero_comp(tripoli_trnode(p),level,R);

  if (res_tp == NULL) return NULL;

  /* finally adjust scalars of homo-components (i.e. structure of poly) */

  aux_s=scalars_nth(ss,level+1);
  aux_tp=res_tp;
  while (aux_tp != NULL) {
    trnode_set_scalar(aux_tp,trnode_scalar(aux_tp)+aux_s);
    aux_tp=trnode_next_comp(aux_tp); }

  tripoli_set_trnode(res_p,res_tp);

  return res_p;
}

tripoli tripoli_x_opp_scalars_l(tripoli p,scalars ss,scalars aux_ss,int level,
  ring R)
{
  scalar aux_s,aux_s_tp;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp,tp,aux_tp1;

  if (level==R->level) {
    tripoli_set_monomials(res_p,monomials_x_opp_scalars(tripoli_monomials(p),ss,R));
    return res_p; }
  
  /* level < R->level */
  
  tp=tripoli_trnode(p);
  aux_s=scalars_nth(aux_ss,level+1);

  res_tp=tripoli_dummy(R,level);

  aux_tp=res_tp;
  while (tp != NULL) {
    aux_p=tripoli_x_opp_scalars_l(trnode_homo_comp(tp),ss,aux_ss,level+1,R);
    if (aux_p != NULL) {
      tripoli_set_mk(aux_tp1,trnode_scalar(tp)+aux_s,aux_p);
      trnode_set_next_comp(aux_tp, aux_tp1);
      aux_tp=trnode_next_comp(aux_tp); }
    tp=trnode_next_comp(tp); }

  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));

  return res_p;
}

#if 0 /* TODO: decide which is better */
tripoli tripoli_x_mon_l_NEW(tripoli p,int level)
{
  scalar aux_s,aux_s_tp;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp,tp,aux_tp1;

  if (level++==tripoli_x_mon_l_max_l) {
    tripoli_set_monomials(res_p,monomials_x_mon(tripoli_monomials(p)));
    return res_p; }
  
  /* level < R->level */
  
  tp=tripoli_trnode(p);
  aux_s=scalars_nth(tripoli_x_mon_l_scalars,level);

  res_tp=tripoli_dummy(tripoli_x_mon_l_ring,level);

  aux_tp=res_tp;
  while (tp != NULL) {
    aux_p=tripoli_x_mon_l(trnode_homo_comp(tp),level);
    if (aux_p != NULL) {
      tripoli_set_mk(aux_tp1,trnode_scalar(tp)+aux_s,aux_p);
      trnode_set_next_comp(aux_tp, aux_tp1);
      aux_tp=trnode_next_comp(aux_tp); }
    tp=trnode_next_comp(tp); }

  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));

  return res_p;
}
#endif

tripoli tripoli_x_mon_l(tripoli p,mon m,scalars ss,int level,ring R)
{
  scalar aux_s;
  tripoli res_p,aux_p;
  trnode aux_tp,res_tp,tp,aux_tp1;

  if (level==R->level) {
    tripoli_set_monomials(res_p,monomials_x_mon(tripoli_monomials(p),m,R));
    return res_p; }
  
  /* level < R->level */
  
  tp=tripoli_trnode(p);

  if (trnode_next_comp(tp) == NULL) {
    tripoli_set_trnode(res_p,trnode_mk(
      trnode_scalar(tp)+scalars_nth(ss,level+1),
      tripoli_x_mon_l(trnode_homo_comp(tp),m,ss,level+1,R)));
    return res_p;
  }

  aux_s=scalars_nth(ss,level+1);

  res_tp=tripoli_dummy(R,level);

  aux_tp=res_tp;
  while (tp != NULL) {

    aux_p=tripoli_x_mon_l(trnode_homo_comp(tp),m,ss,level+1,R);

#ifndef IS_DOMAIN

    if (aux_p != NULL) {
      tripoli_set_mk(aux_tp1,trnode_scalar(tp)+aux_s,aux_p);
      trnode_set_next_comp(aux_tp, aux_tp1);
      aux_tp=trnode_next_comp(aux_tp); }

#else

    tripoli_set_mk(aux_tp1,trnode_scalar(tp)+aux_s,aux_p);
    trnode_set_next_comp(aux_tp, aux_tp1);
    aux_tp=trnode_next_comp(aux_tp);

#endif

    tp=trnode_next_comp(tp); }

  tripoli_set_trnode(res_p,trnode_next_comp(res_tp));

  return res_p;
}

int tripoli_deg_l(tripoli p, scalars ss, int level, ring R)
{
  int deg_i;
  trnode tp;
  if (level==R->level) return monomials_deg(tripoli_monomials(p),ss,R);
  
  /* level < R->level */

  deg_i=0;
  tp = tripoli_trnode(p);
  while( (tp != NULL) ) {
    scalars_set_nth(ss,level+1,trnode_scalar(tp));
    deg_i=MAX(deg_i,tripoli_deg_l(trnode_homo_comp(tp),ss,level+1,R));
    tp = trnode_next_comp(tp); }
  return deg_i;
}

void tripoli_change_sign_l(tripoli p, int level, ring R)
{
  if (level==R->level) monomials_change_sign(tripoli_monomials(p),R);
  
  /* level < R->level */

  else {
    trnode aux_tp;
    aux_tp=tripoli_trnode(p);
    while(aux_tp != NULL) {
      tripoli_change_sign_l(trnode_homo_comp(aux_tp),level+1,R);
      aux_tp=trnode_next_comp(aux_tp); } }
}

int tripoli_len_l(tripoli p, int level, ring R)
{
  trnode aux_tp;
  int res_sum;
  
  if (level==R->level) return monomials_len(tripoli_monomials(p));

  aux_tp=tripoli_trnode(p);
  res_sum=0;
  while (aux_tp != NULL) {
    res_sum+=tripoli_len_l(trnode_homo_comp(aux_tp),level+1,R);
    aux_tp=trnode_next_comp(aux_tp); }
  return res_sum;
}

int tripoli_size_l(tripoli p, int level, ring R)
{
  trnode aux_tp;
  int res_sum;
  
  if (level==R->level) return monomials_size(tripoli_monomials(p),R);

  aux_tp=tripoli_trnode(p);
  res_sum=0;
  while (aux_tp != NULL) {
    res_sum+=tripoli_size_l(trnode_homo_comp(aux_tp),level+1,R)+
      sizeof(struct trnode_aux);
    aux_tp=trnode_next_comp(aux_tp); }
  return res_sum;
}


ctbuf tripoli2ctbuf_l(tripoli p,scalars ss,int level,ring R)
{
  trnode tp;
  ctbuf res_b,aux_b;

  if (level==R->level) return monomials2ctbuf(tripoli_monomials(p),ss,R);
  
  /* level < R->level */

  res_b=ctbuf_empty();
  tp=tripoli_trnode(p);
  while(tp != NULL) {
    scalars_set_nth(ss,level+1,trnode_scalar(tp));
    aux_b=tripoli2ctbuf_l(trnode_homo_comp(tp),ss,level+1,R);
    res_b=ctbuf_concat_these(res_b,aux_b);
    tp = trnode_next_comp(tp); }
  return res_b;
}

tripoli P1MinusXMultiExpP2 (tripoli p1, term t, tripoli p2, 
                               ring R)
{
  tripoli aux_p;
  aux_p=tripoli_this_x_opp_term(p2,t,R);
  return tripoli_sum_these(p1,aux_p,R);
}

tripoli P1PlusXMultiExpP2 (tripoli p1, term t, tripoli p2, 
                               ring R)
{
  tripoli aux_p;
  aux_p=tripoli_this_x_term(p2,t,R);
  return tripoli_sum_these(p1,aux_p,R);
}


/* ------[ TRIPOLI ]------------------------*/

void tripoli_adt_init(void)
{
  mon_dummy=malloc_mon();
  mon_dummy1=malloc_mon();
}

trnode trnode_dup_l(trnode tp,int level,ring R) /*X*/
{
  trnode res_tp;
  if (tp == NULL) return NULL;
  tripoli_set_mk(res_tp,trnode_scalar(tp),
    tripoli_dup_l(trnode_homo_comp(tp),level+1,R));
  trnode_set_next_comp(res_tp, trnode_dup_l(trnode_next_comp(tp),level,R));
  return res_tp;
}

trnode trnode_mk(scalar s,tripoli p) 
{
  trnode res_tp;
  res_tp=malloc_trnode();
  trnode_set_scalar(res_tp,s);
  trnode_set_homo_comp(res_tp,p);
  trnode_set_next_comp(res_tp,NULL);
  return res_tp;
}

cterm trnode2cterm_l(trnode tp,int level,ring R)
{ 
  trnode aux_tp;
  ctbuf aux_b;
  aux_tp=tp;
  aux_b=ctbuf_empty();
  while (aux_tp != NULL) {
    cterm params[2];
    params[0]=int2cterm(trnode_scalar(aux_tp));
    params[1]=tripoli2cterm_l(trnode_homo_comp(aux_tp),level+1,R);
    ctbuf_append(aux_b,cterm_mk_classic(op_HOMO,2,params));
    aux_tp=trnode_next_comp(aux_tp); }
  return ctbuf_mk_cterm(aux_b,op_TRIPOLI);
}

void printf_trnode_l(trnode tp,int level,ring R)
{
  cterm aux_t;
  aux_t=trnode2cterm_l(tp,level,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

void trnode_free_l(trnode tp,int level,ring R)
{
  if (tp != NULL) {
    tripoli_free_l(trnode_homo_comp(tp),level+1,R);
    trnode_free_l(trnode_next_comp(tp),level,R);
    free_trnode(tp); }
}

trnode trnode_rm_zero_comp(trnode tp,int level,ring R)
{
  trnode res_tp,aux_tp,next_tp;
  res_tp=tripoli_dummy(R,level);
  trnode_set_next_comp(res_tp,tp);
  aux_tp=res_tp;
  while (tp != NULL) {
    next_tp=trnode_next_comp(tp);
    if (trnode_homo_comp(tp) == NULL) {
      trnode_set_next_comp(aux_tp,next_tp);
      free_trnode(tp); }
    else
      aux_tp=trnode_next_comp(aux_tp);
    tp=next_tp; }
  return trnode_next_comp(res_tp);
}

void tripoli_plus_mxp_l(tripoli * zis,mon m,tripoli p,scalars ss,
  int level,ring R)
{ /* zis += m*p */ 
  /* NOTE: p is preserved */

  if (level==R->level) {
    mon aux_m;
    aux_m=tripoli_monomials(*zis);
    monomials_plus_mxmm(&aux_m,m,tripoli_monomials(p),R);
    tripoli_set_monomials(*zis,aux_m); }
  
  else /* level < R->level */ {

  scalar s,s1;
  trnode res_tp, aux_tp, aux_tp_next, aux_tp1, tripoli_dum;
  tripoli_dum=tripoli_dummy(R,level);
  trnode_set_next_comp(tripoli_dum,tripoli_trnode(*zis));
  res_tp=tripoli_dum;
  aux_tp=res_tp;
  aux_tp1=tripoli_trnode(p);
  do {
    aux_tp_next=trnode_next_comp(aux_tp);
    s=trnode_scalar(aux_tp_next);
    s1=trnode_scalar(aux_tp1)+scalars_nth(ss,level+1);
    if (s==s1) {
      tripoli aux_p;
      tripoli_plus_mxp_l(&trnode_homo_comp(aux_tp_next),
                          m,trnode_homo_comp(aux_tp1),ss,level+1,R);
      if (trnode_homo_comp(aux_tp_next)==NULL) {
        trnode_set_next_comp(aux_tp,trnode_next_comp(aux_tp_next));
        free_trnode(aux_tp_next); }
      else aux_tp=trnode_next_comp(aux_tp);
      aux_tp1=trnode_next_comp(aux_tp1); }
    else if (s1>s) {
      trnode aux_tp_new;
      tripoli_set_mk(aux_tp_new,s1,tripoli_x_mon_l(
        trnode_homo_comp(aux_tp1),m,ss,level+1,R));
      trnode_set_next_comp(aux_tp_new,trnode_next_comp(aux_tp));
      trnode_set_next_comp(aux_tp,aux_tp_new);
      aux_tp1=trnode_next_comp(aux_tp1); 
      aux_tp=trnode_next_comp(aux_tp); }
    else aux_tp=trnode_next_comp(aux_tp); }
  while((trnode_next_comp(aux_tp) != NULL)
        && (aux_tp1 != NULL));
  if (trnode_next_comp(aux_tp) == NULL)
    while ((aux_tp1 != NULL)) {
      scalar aux_s;
      tripoli aux_p;
      aux_s=trnode_scalar(aux_tp1)+scalars_nth(ss,level+1);
      aux_p=tripoli_x_mon_l(trnode_homo_comp(aux_tp1),m,ss,level+1,R);
      trnode_set_next_comp(aux_tp,trnode_mk(aux_s,aux_p));
      aux_tp=trnode_next_comp(aux_tp);
      aux_tp1=trnode_next_comp(aux_tp1); }
  tripoli_set_trnode(*zis,trnode_next_comp(res_tp)); }
}

/* --------------------------------- *\
           TRNODE_ARRAY
\* --------------------------------- */

trnode * trnode_array_init(int n, ring R)
{
  register int i;
  trnode * res_tp_array;
  res_tp_array=malloc_trnode_array(n);
  for (i=0;i<n;i++)
    tripoli_set_mk(res_tp_array[i],(scalar)0,NULL);
  return res_tp_array;
}

trnode * trnode_array_dup(int n, trnode * tp_array,ring R)
{
  register int i;
  trnode * res_tp_array;
  tripoli aux_p;
  res_tp_array=trnode_array_init(n,R);
  for (i=0;i<n;i++) {
    tripoli_set_trnode(aux_p,tp_array[i]);
    aux_p=tripoli_dup(aux_p,R);
    tripoli_set_mk(res_tp_array[i],(scalar)0,aux_p); }
  return res_tp_array;
}

void trnode_array_free(int n, trnode * tp_array)
{
  register int i;
  for (i=0;i<n;i++)
    free_trnode(tp_array[i]);
  free_trnode_array(n,tp_array);
}

/* --------------------------------- *\
            MONOMIALS
\* --------------------------------- */

mon monomials_dup(mon m,ring R)
{
  mon res_m, aux_m;
  if (m == NULL) return m;
  res_m=mon_dup(m,R);
  aux_m=res_m;
  m=mon_next(m);
  while (m != NULL) {
    mon_set_next(aux_m,mon_dup(m,R));
    m=mon_next(m);
    aux_m=mon_next(aux_m); }
  mon_set_next(aux_m,NULL);
  return res_m;
}

#if 1
mon monomials_merge(mon m1, mon m2, ring R, mon * last_mon)
{
  /* suppose scalars(first(m1)) > scalars(first(m2)) = s */
  /* last_mon is the last mon of m1 s.t. scalars(last_mon)>s */

  /* suppose m1 NOT EMPTY */

  mon aux_m,res_m,aux_m1,aux_m2;
  scalars aux_ss;
  res_m=mon_dummy;

  aux_m1=m1;
  aux_m2=m2;

  aux_ss=mon_scalars(m2);
  *last_mon=NULL;
  while ((aux_m1 != NULL) &&
         scalars_which_gt_lex(aux_ss,mon_scalars(aux_m1))==ANSWER_SECOND) {
    /* printf("."); */
    *last_mon=aux_m1;
    aux_m1=mon_next(aux_m1); }

  if ((*last_mon) == NULL) aux_m=res_m;
  else {
    mon_set_next(res_m,m1);
    aux_m=*last_mon; }

  while ((aux_m1 != NULL) && (aux_m2 != NULL)) {
    mon aux_m1_next,aux_m2_next;
    switch (scalars_which_gt_lex(mon_scalars(aux_m1),mon_scalars(aux_m2))) {
      case ANSWER_BOTH:
        aux_m1_next=mon_next(aux_m1);
        aux_m2_next=mon_next(aux_m2);
        R->AddCoeff(&mon_coeff(aux_m1),mon_coeff(aux_m1),mon_coeff(aux_m2),R);
        if (coeff_is_zero(mon_coeff(aux_m1),R)) {
          mon_set_next(aux_m1,NULL);
          monomials_free(aux_m1,R); }
	else {
          mon_set_next(aux_m,aux_m1);
          aux_m=mon_next(aux_m); }
        mon_set_next(aux_m2,NULL);
	monomials_free(aux_m2,R);
        aux_m1=aux_m1_next;
        aux_m2=aux_m2_next;
	break;
      case ANSWER_FIRST:
        mon_set_next(aux_m,aux_m1);
        aux_m=mon_next(aux_m);
        aux_m1=mon_next(aux_m1);
        break;
      case ANSWER_SECOND:
        mon_set_next(aux_m,aux_m2);
        aux_m=mon_next(aux_m);
        aux_m2=mon_next(aux_m2);
        break;
    }
  }
  if (aux_m1 == NULL) mon_set_next(aux_m,aux_m2);
  else mon_set_next(aux_m,aux_m1);
  return mon_next(res_m);
}
#endif

#if 1
mon monomials_merge_normal(mon m1, mon m2, ring R)
{
  /* suppose scalars(first(m1)) > scalars(first(m2)) = s */
  /* last_mon is the last mon of m1 s.t. scalars(last_mon)>s */

  /* suppose m1 NOT EMPTY */

  mon aux_m,res_m;
  scalars aux_ss;
  aux_m=res_m=mon_dummy;
  mon_set_next(res_m,m1);

  aux_ss=mon_scalars(m2);
  while ((m1 != NULL) &&
         scalars_which_gt_lex(aux_ss,mon_scalars(m1))==ANSWER_SECOND) {
    aux_m=m1;
    m1=mon_next(m1); }

  while ((m1 != NULL) && (m2 != NULL)) {
    mon m1_next,m2_next;
    switch (scalars_which_gt_lex(mon_scalars(m1),mon_scalars(m2))) {
      mon aux_mm;
      case ANSWER_BOTH:
        m1_next=mon_next(m1);
        m2_next=mon_next(m2);
        R->AddCoeff(&mon_coeff(m1),mon_coeff(m1),mon_coeff(m2),R);
        if (coeff_is_zero(mon_coeff(m1),R)) {
          mon_set_next(m1,NULL);
          monomials_free(m1,R); }
	else {
          mon_set_next(aux_m,m1);
          aux_m=mon_next(aux_m); }
        mon_set_next(m2,NULL);
	monomials_free(m2,R);
        m1=m1_next;
        m2=m2_next;
	break;
      case ANSWER_FIRST:
        mon_set_next(aux_m,m1);
        aux_m=mon_next(aux_m);
        m1=mon_next(m1);
        break;
      case ANSWER_SECOND:
        mon_set_next(aux_m,m2);
        aux_m=mon_next(aux_m);
        m2=mon_next(m2);
        break;
    }
  }
  if (m1 == NULL) mon_set_next(aux_m,m2);
  else mon_set_next(aux_m,m1);
  return mon_next(res_m);
}
#endif

mon monomials_mult(mon m1, mon m2, ring R)
{ /* OPTIMIZE: consider case m2 is monomial */
  mon aux_m1,res_m,first_mon,new_first_mon;
  res_m=NULL;
  aux_m1=m1; /* TODO: study a strategy to select which m */
  res_m=monomials_mult_mon_monomials(aux_m1,m2,R);
  aux_m1=mon_next(aux_m1);
  first_mon=res_m;
  while ((aux_m1 != NULL)) {
    mon aux_m;
    aux_m=monomials_mult_mon_monomials(aux_m1,m2,R);
    monomials_addthis2this(&first_mon,aux_m,R,&new_first_mon);
    first_mon=new_first_mon;
    aux_m1=mon_next(aux_m1);
  } 
  return res_m; 
}

mon monomials_mult_mon_monomials(mon m,mon mm,ring R)
{  
  /* --------------------------------------------------------------- *\

  NOTE: m could be a list of monomials, but we consider only the
        first one 

  \* --------------------------------------------------------------- */
  mon res_m,aux_m,aux_mm;

  res_m=mon_dummy;
  aux_m=res_m;
  aux_mm=mm;
  do {
    scalars aux_ss;
    coeff aux_c;
    coeff_set_mult(aux_c,mon_coeff(m),mon_coeff(aux_mm),R);
    scalars_set_sum(&aux_ss,mon_scalars(m),mon_scalars(aux_mm));
    mon_set_next(aux_m,mon_mk(aux_c,aux_ss));
    aux_mm=mon_next(aux_mm);
    aux_m=mon_next(aux_m);
  }
  while (aux_mm != NULL);
  return mon_next(res_m);  
}

void monomials_add2these(mon * ziis, mon m, ring R)
{ /* m is preserved */
  scalar s;
  mon res_m, aux_m, aux_m_next, aux_m1;

  mon_set_next(mon_dummy,*ziis);
  res_m=mon_dummy;
  aux_m=res_m;
  aux_m1=m;
  do {
    int answer;
    aux_m_next=mon_next(aux_m);
    answer=scalars_which_gt_lex(mon_scalars(aux_m_next),mon_scalars(aux_m1));
    switch (answer) {
      mon aux_mm,aux_m_new;
      case ANSWER_BOTH:
        R->AddCoeff(&mon_coeff(aux_m_next),
                     mon_coeff(aux_m_next),mon_coeff(aux_m1),R);
        if (coeff_is_zero(mon_coeff(aux_m_next),R)) {
          mon_set_next(aux_m,mon_next(aux_m_next));
          mon_set_next(aux_m_next,NULL);
          monomials_free(aux_m_next,R); }
        aux_m1=mon_next(aux_m1);
	break;
      case ANSWER_SECOND:
        mon_set_mk(aux_m_new,mon_coeff(aux_m1),
          scalars_dup(mon_scalars(aux_m1)));
        mon_set_next(aux_m_new,mon_next(aux_m_next));
        mon_set_next(aux_m_next,aux_m_new);
        aux_m1=mon_next(aux_m1);
        break;
      case ANSWER_FIRST:
        /* do nothing */
        break;
    }
    aux_m=mon_next(aux_m); }
  while((mon_next(aux_m) != NULL) && (aux_m1 != NULL));
  if (mon_next(aux_m) == NULL) 
    mon_set_next(aux_m,monomials_dup(aux_m1,R));
  *ziis=mon_next(res_m);
}

mon monomials_sum(mon m1, mon m2, ring R)
{
  mon aux_m,res_m,aux_m1,aux_m2;
  int answer;
  res_m=mon_dummy;
  aux_m=res_m;
  aux_m1=m1;
  aux_m2=m2;
  while ((aux_m1 != NULL) && (aux_m2 != NULL)) {
    coeff aux_c;
    answer=scalars_which_gt_lex(mon_scalars(aux_m1),mon_scalars(aux_m2));
    switch (answer) {
      mon aux_mm;
      case ANSWER_BOTH:
        aux_c=coeff_sum(mon_coeff(aux_m1),mon_coeff(aux_m2),R);
        if (!coeff_is_zero(aux_c,R)) {
          mon_set_mk(aux_mm,aux_c,scalars_dup(mon_scalars(aux_m1)));
          mon_set_next(aux_m,aux_mm);
          aux_m=mon_next(aux_m); }
        aux_m1=mon_next(aux_m1);
        aux_m2=mon_next(aux_m2);
	break;
      case ANSWER_FIRST:
        mon_set_mk(aux_mm,coeff_dup(mon_coeff(aux_m1),R),
                           scalars_dup(mon_scalars(aux_m1)));
        mon_set_next(aux_m,aux_mm);
        aux_m=mon_next(aux_m);
        aux_m1=mon_next(aux_m1);
        break;
      case ANSWER_SECOND:
        mon_set_mk(aux_mm,coeff_dup(mon_coeff(aux_m2),R),
                           scalars_dup(mon_scalars(aux_m2)));
        mon_set_next(aux_m,aux_mm);
        aux_m=mon_next(aux_m);
        aux_m2=mon_next(aux_m2);
        break;
    }
  }
  if (aux_m1 == NULL) mon_set_next(aux_m,monomials_dup(aux_m2,R));
  else mon_set_next(aux_m,monomials_dup(aux_m1,R));
  return mon_next(res_m);
}

cterm special_monomials2cterm(mon mm,ring R)
{
  mon aux_mm;
  ctbuf aux_b;
  aux_mm=mm;
  aux_b=ctbuf_empty();
  while (aux_mm != NULL) {
    ctbuf_append(aux_b,special_monomial2cterm(aux_mm,R));
    aux_mm=mon_next(aux_mm); }
  return ctbuf_mk_cterm(aux_b,op_CLASSIC);
}

void monomials_free(mon m,ring R)
{
  if (m != NULL) {
    monomials_free(mon_next(m),R);
    mon_free(m,R); }
}

coc_bool monomials_eq(mon m1, mon m2, ring R)
{
  while ((m1 != NULL) && (m2 != NULL)) {
    if (! mon_eq(m1,m2,R)) return FALSE;
    m1=mon_next(m1);
    m2=mon_next(m2); }
  return ((m1 == NULL) && (m2 == NULL));
}

mon monomials_these_x_opp_scalars(mon mm, scalars ss, ring R)
{
  mon aux_m,res_m,next_m;
  scalars aux_ss;

  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  while (mm != NULL) {
    next_m=mon_next(mm);
    R->ChangeSignToCoeff(& mon_coeff(mm),R);
    scalars_add_to_this(mon_scalars(mm),ss);
    aux_m=mon_next(aux_m);
    mm=next_m; }
  return mon_next(res_m);
}

mon monomials_these_x_scalars(mon mm, scalars ss, ring R)
{
  mon aux_m,res_m,next_m;
  scalars aux_ss;

  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  while (mm != NULL) {
    next_m=mon_next(mm);
    scalars_add_to_this(mon_scalars(mm),ss);
    aux_m=mon_next(aux_m);
    mm=next_m; }
  return mon_next(res_m);
}

mon monomials_these_x_mon(mon mm, mon m, ring R)
{
  mon aux_m,res_m,next_m;
  coeff aux_c,aux_c_mm,res_c;
  scalars aux_ss;
  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  aux_c=mon_coeff(m);
  aux_ss=mon_scalars(m);
  while (mm != NULL) {
    next_m=mon_next(mm);
    aux_c_mm=mon_coeff(mm);
    coeff_set_mult(res_c,aux_c,aux_c_mm,R);
    if (coeff_is_zero(res_c,R)) {
      R->ClearCoeff(& res_c);
      mon_free(mm,R);
      mon_set_next(aux_m,next_m); }
    else {
      R->ClearCoeff(& aux_c_mm);
      mon_set_coeff(mm,res_c);
      scalars_add_to_this(mon_scalars(mm),aux_ss);
      aux_m=mon_next(aux_m); }
    mm=next_m; }
  return mon_next(res_m);
}

mon monomials_x_opp_scalars(mon mm, scalars ss, ring R)
{
  mon aux_m,res_m,next_m,aux_m1;

  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  while (mm != NULL) {
    next_m=mon_next(mm);
    mon_set_mk(aux_m1,coeff_opposite(mon_coeff(mm),R),
      scalars_sum(mon_scalars(mm),ss));
    mon_set_next(aux_m,aux_m1);
    aux_m=mon_next(aux_m);
    mm=next_m; }
  return mon_next(res_m); 
}

/* TODO: optimize (coeff=1, term=1) */

mon monomials_x_mon(mon mm, mon m, ring R)
{
  mon aux_m,aux_m1;
  coeff aux_c,res_c;
  scalars aux_ss,aux_ss1,aux_ss2;

  mon_set_next(mon_dummy,mm);
  aux_m=mon_dummy;
  aux_ss2=mon_scalars(m);
  aux_c=mon_coeff(m);
  while (mm != NULL) {
    coeff_set_mult(res_c,aux_c,mon_coeff(mm),R);

#ifndef IS_DOMAIN

    if (coeff_is_zero(res_c,R))
      R->ClearCoeff(& res_c);
    else {
      aux_ss1=mon_scalars(mm);
      scalars_set_sum(aux_ss,aux_ss1,aux_ss2);
      mon_set_mk(aux_m1,res_c,aux_ss);
      mon_set_next(aux_m,aux_m1);
      aux_m=aux_m1; }

#else

    aux_ss1=mon_scalars(mm);
    scalars_set_sum(&aux_ss,aux_ss1,aux_ss2);
    mon_set_mk(aux_m1,res_c,aux_ss);
    mon_set_next(aux_m,aux_m1);
    aux_m=mon_next(aux_m);

#endif

    mm=mon_next(mm); }
  return mon_next(mon_dummy); 
}

mon monomials_x_mon_NEWW(mon mm, mon m, ring R)
{
  mon aux_m,res_m,next_m,aux_m1;
  coeff aux_c,aux_c_mm,res_c;
  scalars aux_ss;

  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  aux_ss=mon_scalars(m);
  aux_c=mon_coeff(m);

  if (coeff_is_one(aux_c,R))

  while (mm != NULL) {
    next_m=mon_next(mm);
    mon_set_mk(aux_m1,coeff_dup(mon_coeff(mm),R),
      scalars_sum(mon_scalars(mm),aux_ss));
    mon_set_next(aux_m,aux_m1);
    aux_m=mon_next(aux_m);
    mm=next_m; }

  else

  while (mm != NULL) {
    next_m=mon_next(mm);
    aux_c_mm=mon_coeff(mm);
    coeff_set_mult(res_c,aux_c,aux_c_mm,R);
    if (coeff_is_zero(res_c,R))
      R->ClearCoeff(& res_c);
    else {
      mon_set_mk(aux_m1,res_c,scalars_sum(mon_scalars(mm),aux_ss));
      mon_set_next(aux_m,aux_m1);
      aux_m=mon_next(aux_m); }
    mm=next_m; }

  return mon_next(res_m); 
}


#if 0
mon monomials_x_mon_NEW(mon mm)
{
  mon aux_m,res_m,next_m,aux_m1;
  coeff aux_c,aux_c_mm,res_c;
  scalars aux_ss;

  res_m=mon_dummy;
  mon_set_next(res_m,mm);
  aux_m=res_m;
  aux_c=mon_coeff(tripoli_x_mon_l_mon);
  aux_ss=mon_scalars(tripoli_x_mon_l_mon);
  while (mm != NULL) {
    next_m=mon_next(mm);
    aux_c_mm=mon_coeff(mm);
    coeff_set_mult(res_c,aux_c,aux_c_mm,tripoli_x_mon_l_ring);
    if (coeff_is_zero(res_c,tripoli_x_mon_l_ring))
      tripoli_x_mon_l_ring->ClearCoeff(& res_c);
    else {
      mon_set_mk(aux_m1,res_c,scalars_sum(mon_scalars(mm),aux_ss));
      mon_set_next(aux_m,aux_m1);
      aux_m=mon_next(aux_m); }
    mm=next_m; }
  return mon_next(res_m); 
}
#endif

int monomials_deg(mon m,scalars ss,ring R)
{
  int deg_i;
  deg_i=0;
  while (m != NULL) {
    deg_i=MAX(deg_i,mon_deg(m,ss,R));
    m = mon_next(m); }
  return deg_i;
}

void monomials_change_sign(mon m,ring R)
{
  while (m != NULL) {
    mon_change_sign(m,R);
    m=mon_next(m); }
}

int monomials_len(mon m)
{
  int res_sum;
  res_sum=0;
  while (m != NULL) {
    res_sum++;
    m=mon_next(m); }
  return res_sum;
}

int monomials_size(mon m, ring R)
{
  int res_sum;
  res_sum=0;
  while (m != NULL) {
    res_sum+=sizeof(struct mon_aux) +    
      (scalars_len(mon_scalars(m))+2)*sizeof(iv_elem)+
      coeff_size(mon_coeff(m),R)-sizeof(coeff);
    m=mon_next(m); }
  return res_sum;
}

ctbuf monomials2ctbuf(mon m,scalars ss,ring R)
{
  ctbuf res_b;
  res_b=ctbuf_empty();
  while (m != NULL) {
    ctbuf_append(res_b,mon2cterm(m,ss,R));
    m=mon_next(m); }
  return res_b;
}

void monomials_plus_mxmm(mon * ziis,mon m,mon mm,ring R)
{ /* m is preserved */
  scalar s;
  mon res_m, aux_m, aux_m_next, aux_m1;

  mon_set_next(mon_dummy1,*ziis);
  res_m=mon_dummy1; /* cannot use mon_dummy because I call monomials_x_mon */
  aux_m=res_m;
  aux_m1=mm;
  do {
    int answer;
    scalars aux_ss;
    aux_m_next=mon_next(aux_m);
    scalars_set_sum(&aux_ss,mon_scalars(aux_m1),mon_scalars(m));
    answer=scalars_which_gt_lex(mon_scalars(aux_m_next),aux_ss);
    switch (answer) {
      mon aux_mm,aux_m_new;
      case ANSWER_BOTH:
        R->AddCoeff(&mon_coeff(aux_m_next),
                     mon_coeff(aux_m_next),coeff_mult(mon_coeff(m),
                                                      mon_coeff(aux_m1),R),R);
        if (coeff_is_zero(mon_coeff(aux_m_next),R)) {
          mon_set_next(aux_m,mon_next(aux_m_next));
          mon_free(aux_m_next,R); }
        else aux_m=mon_next(aux_m);
        scalars_free(aux_ss);     
        aux_m1=mon_next(aux_m1);
	break;
      case ANSWER_SECOND:
        mon_set_mk(aux_m_new,
          coeff_mult(mon_coeff(aux_m1),mon_coeff(m),R),aux_ss);
        mon_set_next(aux_m_new,aux_m_next);
        mon_set_next(aux_m,aux_m_new);
        aux_m1=mon_next(aux_m1);
        aux_m=mon_next(aux_m);
        break;
      case ANSWER_FIRST:
        /* do nothing */
        scalars_free(aux_ss);
        aux_m=mon_next(aux_m);
        break;
    } }
  while((mon_next(aux_m) != NULL) && (aux_m1 != NULL));

  if (mon_next(aux_m) == NULL)
    mon_set_next(aux_m,monomials_x_mon(aux_m1,m,R)); 

  *ziis=mon_next(res_m);
}

cterm special_monomial2cterm(mon m,ring R)
{
  scalars aux_ss;
  register int i;
  ctbuf aux_b;
  cterm params[2];
  aux_ss=mon_scalars(m);
  aux_b=ctbuf_empty();
  for (i=1;i<=scalars_len(aux_ss);i++) 
    ctbuf_append(aux_b,int2cterm(scalars_nth(aux_ss,i)));
  params[0]=coeff2cterm(mon_coeff(m),R);
  params[1]=ctbuf_mk_cterm(aux_b,op_SCALARS);
  return cterm_mk_classic(op_MON,2,params);
}

/* --------------------------------- *\
               MON
\* --------------------------------- */

mon mon_mk(coeff c,scalars ss)
{
  mon res_m=malloc_mon();
  mon_set_coeff(res_m,c);
  mon_set_scalars(res_m,ss);
  mon_set_next(res_m,NULL);
  return res_m;
}

void printf_mon(mon m,ring R)
{
  cterm aux_t;
  aux_t=special_monomials2cterm(m,R);
  put_cterm(aux_t);
  cterm_free(aux_t);
}

void mon_free(mon m,ring R)
{
  R->ClearCoeff(& mon_coeff(m));
  scalars_free(mon_scalars(m));
  free_mon(m);
}

coc_bool mon_eq(mon m1, mon m2, ring R)
{
  return (R->AreEqualsCoeff(mon_coeff(m1),mon_coeff(m2)) &&
          scalars_eq(mon_scalars(m1),mon_scalars(m2)));
}

int mon_deg(mon m,scalars ss,ring R)
{
  scalars aux_ss;
  term aux_t;
  int deg_i,i;
  aux_ss=mon_scalars(m);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i,scalars_nth(aux_ss,i-R->level));
  aux_t=scalars2term(ss,R);
  deg_i=term_degree(aux_t);
  term_free(aux_t);
  return deg_i;
}

cterm mon2cterm(mon m,scalars ss,ring R)
{
  cterm res_t;
  cterm params[2];
  term aux_t;
  scalars aux_ss;
  int i;
  params[0]=coeff2cterm(mon_coeff(m),R);
  aux_ss=mon_scalars(m);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i,scalars_nth(aux_ss,i-R->level));
  aux_t=scalars2term(ss,R);
  if (term_degree(aux_t)==0)
    res_t=cterm_mk_classic(op_MONOMIAL,1,params);
  else {
    params[1]=term2cterm(aux_t,R);
    res_t=cterm_mk_classic(op_MONOMIAL,2,params); }
  term_free(aux_t);
  return res_t;
}

tripoli tripoli_this_x_coeff(tripoli p, coeff c, ring R)
{
  /* TODO: IMPROVE */
  tripoli res_p,aux_p;
  aux_p=tripoli_coeff2tripoli(c,R);
  res_p=tripoli_mult(aux_p,p,R);
  tripoli_free(aux_p,R);
  tripoli_free(p,R);
  return res_p;
}

tripoli tripoli_square(tripoli p, ring R)
{
  /* TODO: improve */
  return tripoli_mult(p,p,R);
}

tripoli this_term2tripoli(term t, ring R)
{
  monomial aux_m;
  monomial_set_mk(aux_m,coeff_one(R),t);
  return this_monomial2tripoli(aux_m,R);
}

tripoli this_monomial2tripoli(monomial m, ring R)
{
  /* TODO: ->macro */
  return tripoli_this_x_monomial(tripoli_one(R),m,R);
}

void tripoli_change_order(tripoli * p, ring R1, ring R2) 
{
  monomial aux_m;
  if ((*p) != NULL) {
    tripoli res_p;
    aux_m=tripoli2monopoli(*p,R1);
    res_p=NULL;
    while (aux_m!=NULL) {
      tripoli_add_this_monomial_unique(aux_m,&res_p,R2);
      aux_m=monomial_next(aux_m); }
    *p=res_p; }
}


ivector tripoli_max_mdeg(tripoli p, ring R)
{
  monopoli aux_m;
  ivector res_iv;
  OPTIMIZE("tripoli_max_mdeg");
  aux_m=tripoli2monopoli(p,R);
  res_iv=monopoli_max_mdeg(aux_m,R);
  monopoli_free(aux_m,R);
  return res_iv;
}

coc_bool tripoli_is_homog(tripoli p, ring R)
{
  monopoli aux_m;
  coc_bool res_b;
  OPTIMIZE("tripoli_is_homog");
  aux_m=tripoli2monopoli(p,R);
  res_b=monopoli_is_homog(aux_m,R);
  monopoli_free(aux_m,R);
  return res_b;
}

int tripoli_is_homog_return_deg(tripoli p, ring R)
{
  monopoli aux_m;
  int res_n;
  OPTIMIZE("tripoli_is_homog_return_deg");
  aux_m=tripoli2monopoli(p,R);
  res_n=monopoli_is_homog_return_deg(aux_m,R);
  monopoli_free(aux_m,R);
  return res_n;
}

coc_bool tripoli_is_homog_deg(tripoli p, int hd, ring R)
{
  monopoli aux_m;
  coc_bool res_b;
  OPTIMIZE("tripoli_is_homog");
  aux_m=tripoli2monopoli(p,R);
  res_b=monopoli_is_homog_deg(aux_m,hd,R);
  monopoli_free(aux_m,R);
  return res_b;
}

tripoli tripoli_map(tripoli p, ring R, ring S, ivector m)
{
  monopoli aux_p,aux_p2;
  tripoli res_p;
  OPTIMIZE("tripoli_map");
  aux_p=tripoli2monopoli(p,R);
  aux_p2=monopoli_map(aux_p,R,S,m);
  res_p=this_monopoli2tripoli(aux_p2,S);
  monopoli_free(aux_p,R);
  return res_p;
}

monopoli this_tripoli2monopoli(tripoli p,ring R) 
{
  monopoli res_m;
  res_m=tripoli2monopoli(p,R);
  tripoli_free(p,R);
  return res_m;
}

monopoli tripoli2monopoli(tripoli p, ring R)
{
  mlist aux_ML;
  scalars aux_ss;
  monomial res_m;
  if (p == NULL) return NULL;
  aux_ss=scalars_init(R->indetsNo);
  aux_ML=tripoli2mlist_l(p,aux_ss,0,R);
  res_m=this_mlist2monopoli(aux_ML,R);
  scalars_free(aux_ss);
  return res_m;
}

mlist monomials2mlist(mon m,scalars ss,ring R)
{
  mlist res_ML;
  res_ML=mlist_init();
  while (m != NULL) {
    mlist_append(res_ML,mon2longmon(m,ss,R));
    m=mon_next(m); }
  return res_ML;
}

mon mon2longmon(mon m,scalars ss,ring R)
{
  scalars aux_ss;
  int i;
  aux_ss=mon_scalars(m);
  for (i=R->level+1;i<=R->indetsNo;i++)
    scalars_set_nth(ss,i,scalars_nth(aux_ss,i-R->level));
  return mon_mk(coeff_dup(mon_coeff(m),R),scalars_dup(ss));
}

mlist tripoli2mlist_l(tripoli p, scalars ss, int level, ring R)
{
  mlist res_ML,aux_ML;
  trnode tp;

  if (level==R->level) return monomials2mlist(tripoli_monomials(p),ss,R);

  tp=tripoli_trnode(p);
  res_ML=mlist_init();
  while (tp != NULL) {
    scalars_set_nth(ss,level+1,trnode_scalar(tp));
    aux_ML=tripoli2mlist_l(trnode_homo_comp(tp),ss,level+1,R);
    res_ML=mlist_concat_these(res_ML,aux_ML);
    tp = trnode_next_comp(tp); }
  return res_ML;
}

monopoli this_mlist2monopoli(mlist ML,ring R)
{
  /* NOTE: ML CANNOT be empty */
  int n;
  mon aux_m;
  coeff aux_c;
  term aux_t;
  monomial res_m,res_m_save;
  aux_m=mlist_get_first(ML);
  aux_c=mon_coeff(aux_m);
  aux_t=scalars2term(mon_scalars(aux_m),R);
  monomial_set_mk(res_m,aux_c,aux_t);
  res_m_save=res_m;
  n=mlist_len(ML);
  while(n-- > 0) {
    aux_m=mlist_get_first(ML);
    aux_c=mon_coeff(aux_m);
    aux_t=scalars2term(mon_scalars(aux_m),R);
    monomial_set_next(res_m,monomial_mk(aux_c,aux_t));
    scalars_free(mon_scalars(aux_m));
    free_mon(aux_m);
    res_m=monomial_next(res_m); }
  mlist_free(ML);
  return res_m_save;
}

void tripoli_add_this_monomial_unique(monomial m, tripoli *p, ring R)
{
  /* TODO: Optimize */ /* unique => we know that this monomial is not here */
  tripoli aux_p;
  aux_p=this_monomial2tripoli(m,R);
  *p=tripoli_sum_these(aux_p,*p,R);
}

tripoli tripoli_add_this_monomial(tripoli p, monomial m, ring R)
{
  /* TODO: Optimize */
  tripoli aux_p;
  aux_p=this_monomial2tripoli(m,R);
  return tripoli_sum_these(aux_p,p,R);
}

tripoli tripoli_randomize(tripoli p, ring R)
{
  monopoli aux_m,aux_m2;
  tripoli res_p;
  aux_m=tripoli2monopoli(p,R);
  aux_m2=monopoli_randomize(aux_m,R);
  res_p=monopoli2tripoli(aux_m2,R);
  monopoli_free(aux_m,R);
  monopoli_free(aux_m2,R);
  return res_p;
}

tripoli tripoli_this_plus_term_x_tripoli_truncated(tripoli p, term t, 
  tripoli p2, iv_elem Trunc, ring R)
{
  TODO("tripoli_this_plus_term_x_tripoli_truncated");
  return NULL;
}

tripoli tripoli_this_minus_term_x_tripoli_truncated(tripoli p, term t, 
  tripoli p2, iv_elem Trunc, ring R)
{
  TODO("tripoli_this_minus_term_x_tripoli_truncated");
  return NULL;
}

tripoli tripoli_this_plus_term_x_tripoli(tripoli p, term t, 
  tripoli p2, ring R)
{
  TODO("tripoli_this_plus_term_x_tripoli");
  return NULL;
}

tripoli tripoli_this_minus_term_x_tripoli(tripoli p, term t, 
  tripoli p2, ring R)
{
  TODO("tripoli_this_minus_term_x_tripoli");
  return NULL;
}

#endif /* TRIPOLI */
