/* $Id: utils.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

/* [ charbuf ] */

#include "cocdefs.h"
#include "machine.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "options.h"
#include "utils.h"

#ifdef COCOA_WITH_RUM

#ifndef RUM_H
# include "rum.h"
#endif

#endif

coc_bool utils_MEMORY_FLAG=0;
coc_bool utils_MEMSTAT_FULL=0;

/*  ==========  memory management  ==========  */

#ifdef DEBUG
size_t memused = 0;
size_t maxmemused = 0;
size_t MPmemused = 0;
size_t MPmaxmemused = 0;

FILE * debuglog=stdout;

#endif


void *myrealloc(void *oldptr, size_t old_size, size_t new_size,char * m)
{
  void *ret;

  ret = realloc (oldptr, new_size);
#ifdef DEBUG_MALLOC
  printf("FREED %lx of size=%d (realloc)\n", (long)oldptr, old_size);
  printf("ALLOC %lx of size=%d (realloc)\n", (long)ret, new_size);
#endif
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d->%d:R]\n",m,new_size-old_size,oldptr,ret);
    else
      fprintf(debuglog,"[%s:%d:R]\n",m,new_size-old_size);
    fflush(debuglog); }
#endif

  if (ret == 0)
    {
      /* perror ("cannot allocate in libmp"); */
      abort ();
    }
#ifdef DEBUG
  memused += (new_size - old_size);
  if (memused > maxmemused )
    maxmemused = memused;
#endif


  return ret;
}


#ifdef DEBUG

void debug_free(size_t size, void *p,char * m){
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d:F]\n",m,size,p);
    else
      fprintf(debuglog,"[%s:%d:F]\n",m,size);
    fflush(debuglog); }
#ifdef DEBUG_MALLOC
  printf("FREED %lx of size=%d\n", (long)p, size);
#endif
  memused -= size;
#ifdef MEMORY_CLEAR_WHEN_FREE
  memset(p,'\0',size);
#endif
#ifdef COCOA_WITH_RUM
  if (size<=RUM_N_STACKS) rum_free(size,p); else free(p);
#else
  free (p);
#endif
}


void *debug_malloc(size_t size,char * m) {
  void * p;

#ifdef COCOA_WITH_RUM
  p = (void *)((size<=RUM_N_STACKS) ? rum_malloc(size) : malloc(size));
#else
  p = (void *)malloc(size);
#endif
#ifdef DEBUG_MALLOC
  printf("ALLOC %lx of size=%d\n", (long)p, size);
#endif
  
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d:A]\n",m,size,p);
    else
      fprintf(debuglog,"[%s:%d:A]\n",m,size);
    fflush(debuglog); }
#endif

  if (p == NULL) {
    abort();
  }
  memused += size;
  if (memused > maxmemused )
    maxmemused = memused;
  return p;
}

/* ----no RUM_N_STACKS check--- */

void debug_free1(size_t size, void *p,char * m){
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d:F]\n",m,size,p);
    else
      fprintf(debuglog,"[%s:%d:F]\n",m,size);
    fflush(debuglog); }

  memused -= size;
#ifdef COCOA_WITH_RUM
  rum_free(size,p);
#else
  free (p);
#endif
}


void *debug_malloc1(size_t size,char * m) {
  void * p;

#ifdef COCOA_WITH_RUM
  p = (void *)rum_malloc(size);
#else
  p = (void *)malloc(size);
#endif

  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d:A]\n",m,size,p);
    else
      fprintf(debuglog,"[%s:%d:A]\n",m,size);
    fflush(debuglog); }

  if (p == NULL) {
    abort();
  }
  memused += size;
  if (memused > maxmemused )
    maxmemused = memused;
  return p;
}


#endif /* DEBUG */

/* ------------------------------------------------------------------------ */

void *mycalloc(int n,size_t size, char * m) {
  void *p = calloc(n,size);
  if (p == NULL) {
    abort();
  }
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[%s:%d:%d:C]\n",m,n*size,p);
    else
      fprintf(debuglog,"[%s:%d:C]\n",m,n*size);
    fflush(debuglog); }
#endif

#ifdef DEBUG
  memused += n*size;
  if (memused > maxmemused )
    maxmemused = memused;
#endif


  return p;
}

void MPfree(void *p, size_t size)
{
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[MP:%d:%d:F]\n",size,p);
    else
      fprintf(debuglog,"[MP:%d:F]\n",size);
    fflush(debuglog); }
#endif

#ifdef DEBUG
  MPmemused -= size;
#endif
  free (p);
}

void *MPmalloc(size_t size) {
  void *p = malloc(size);
  if (p == NULL) {
    abort();
  }
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[MP:%d:%d:A]\n",size,p);
    else
      fprintf(debuglog,"[MP:%d:A]\n",size);
    fflush(debuglog); }
#endif

#ifdef DEBUG
  MPmemused += size;
  if (MPmemused > MPmaxmemused )
    MPmaxmemused = MPmemused;
#endif
  return p;
}

void *MPrealloc(void *oldptr, size_t old_size, size_t new_size)
{
  void *ret;

/* trick that doew not work on x86
  ret = malloc(new_size);
  memcpy(ret,oldptr,old_size);
  free(oldptr);
*/
  
  ret = realloc (oldptr, new_size);
  
  if (ret == 0)
    {
      /* perror ("cannot allocate in libmp"); */
      abort ();
    }
#ifdef DEBUG
  if (utils_MEMORY_FLAG) {
    if (utils_MEMSTAT_FULL)
      fprintf(debuglog,"[MP:%d:%d->%d:R]\n",new_size-old_size,oldptr,ret);
    else
      fprintf(debuglog,"[MP:%d:R]\n",new_size-old_size);
    fflush(debuglog); }
#endif

#ifdef DEBUG
  MPmemused += (new_size - old_size);
  if (MPmemused > MPmaxmemused )
    MPmaxmemused = MPmemused;
#endif
  return ret;
}


void initmem(void) {
  mp_set_memory_functions(&MPmalloc, &MPrealloc, &MPfree);
}

void printmem(char * message) {

if (message!=NULL) printf("%% %s\n",message);
 printf("%%-------------------------------------------------------------------------------\n");
#ifdef DEBUG
  printf("%% mem      = %10lu", memused);
  printf("  maxmem   = %10lu\n", maxmemused);
  printf("%% MPmem    = %10lu", MPmemused);
  printf("  MPmaxmem = %10lu\n", MPmaxmemused);
  maxmemused = memused;
  MPmaxmemused = MPmemused;
#endif
                printf("%%-------------------------------------------------------------------------------\n");

}

char * my_strdup(char*s)
{
  char * dup=NULL;
  if (s!=NULL) {
    dup = malloc_str(strlen(s));
    strcpy(dup,s); }
  return dup;
}

char * my_strdup_nospaces(char*s)
{
  char * dup=NULL;
  int i,l;
  if (s!=NULL) {
    l = strlen(s);
    for (i=l-1;i>=0 && s[i]==' ';i--);
    dup = malloc_str(i+1);
    strncpy(dup,s,i+1);
    dup[i+1]='\0'; }
  return dup;
}

char * my_upper(char*s)
{
  char * dup=NULL;
  if (s!=NULL) {
    register int i;
    int l;
    l = strlen(s);
    dup = malloc_str(l);
    for (i=0;i<l;i++)
      dup[i]=UPPER(s[i]);
    dup[l]='\0';}
  return dup;
}

char * my_strsum(char * s1, char * s2)
{
  char * res_s;
  res_s=malloc_str(strlen(s1)+strlen(s2));
  res_s[0]='\0';
  strcat(res_s,s1);
  strcat(res_s,s2);
  return res_s;
}

char * path_tail(char * path) {
  int n;
  for (n=strlen(path);n>=0;n--) {
    if (path[n]=='/') break;
  }
  return path+n+1;
}

/* This signature comes from ca.h */
char* ca_cocoa_packages();

char * pkg_path(char * pkg_id)
{
  char * filepath;
  char * respath;
  char * prefix;
  prefix = ca_cocoa_packages();
  filepath=malloc_str(strlen(prefix)+strlen(pkg_id)+7);
  sprintf(filepath,"%s/%s.cpkg\0",prefix,++pkg_id);
  respath=build_path(filepath);
  free_str1(filepath);
  return respath;
}

int my_strcasecmp(const char *s1, const char *s2)
{
  for ( ; tolower(*s1) == tolower(*s2) ; s1++, s2++)
          if (*s1 =='\0') return 0;
        return (tolower(*s1) - tolower(*s2) );
}

int str_split(char * s, char * sep, char ** s1, char ** s2)
{ /* given s = X sep Y returns s1=X and s2 = Y */
  /* if NO sep then s1=NULL s2=s */
  *s1=my_strtok(s,sep);
  *s2=my_strtok(NULL,"");
  if (*s2 == NULL) {*s2 = *s1; *s1=NULL;}
  return 0;
}

char * my_strtok(char * s1, const char * s2)
{
  return strtok(s1,s2);
}

void test_strtok(void)
{
  char * s;
  char * t;
  s="prova/tokens/stringa";
  t=my_strtok(s,"/");
  do {
    printf("%s\n",t);
    t=my_strtok(NULL,"/");
  } while (t);
  printf("s=%s\n",s);
}

/* operations on integers */

int int_binom(int n, int k)
{
  int i,res;
  k=MIN(k,n-k);
  res=1;
  for (i=1;i<=k;i++)
    res=res*(n-i+1)/i;
  return res;
}

/* -----------[ charbuf ]-------------- */

charbuf charbuf_init(int size, int blocksize)
{
  charbuf B;
  B = malloc_charbuf();
  charbuf_set_buffer(B, malloc_str(size));
  charbuf_set_len(B,0);
  charbuf_set_size(B,size);
  charbuf_set_blocksize(B,blocksize);
  charbuf_cork(B);
  return B;
}

void charbuf_free(charbuf B)
{
  free_str(charbuf_size(B),charbuf_buffer(B));
  free_charbuf(B);
}

void charbuf_enlarge(charbuf B)
{
  B->buffer=realloc_str(B->buffer, B->size, B->size+B->blocksize);
  B->size+=B->blocksize;
}

void charbuf_putc(charbuf B, char c)
{
  if (charbuf_len(B) == charbuf_size(B)) charbuf_enlarge(B);
  B->buffer[B->len++]=c;
  B->buffer[B->len]='\0';
}

void charbuf_puts(charbuf B, char * s)
{
  while (*s) { charbuf_putc(B,*s++); }
}

void charbuf_cork(charbuf B)
{
  B->buffer[B->len]='\0';
}

void charbuf_test()
{
  charbuf B;
  char c;
  B=charbuf_init(10,4);
  for (c=' '; c<= '\176' ; c++) {
    charbuf_putc(B,c);
    charbuf_cork(B);
    printf("B=%s, L=%d, S=%d\n", B->buffer, B->len, B->size);
  }
  charbuf_puts(B,"THEEND");
  printf("B=%s, L=%d, S=%d\n", B->buffer, B->len, B->size);
  charbuf_free(B);
}
