/* $Id: utils.h,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#ifndef UTILS_H
#define UTILS_H

#include <stdio.h>

# include "coctypes.h"
# include "auxtypes.h"
# include "rum.h"

#ifdef DEBUG
extern FILE * debuglog;
#endif

extern coc_bool utils_MEMORY_FLAG;

/*  ============  general macros  ============  */

#ifdef ABS
#undef ABS
#endif

#define ABS(A)       ( (A) < 0 ? (- (A)) : (A) )
#define MAX(A, B)    ( (A) > (B) ? (A) : (B) )
#define MIN(A, B)    ( (A) > (B) ? (B) : (A) )
#define MMOD(A, B)   ( (A) % (B) >= 0 ? (A) % (B) : (A) % (B) + (B) ) 
#define UPPER(C)     ( islower(C) ? (C)-'a'+'A' : (C) )

#define is_special_var(id) (\
  (strcmp(id,"PKG") == 0) || \
  (strcmp(id,"DEV") == 0) || \
  (strcmp(id,"ENV") == 0) || \
  (strcmp(id,"ERR") == 0))|| \
  (strcmp(id,"ComputationStack") == 0)

#ifdef DEBUG

#define TODO(S) do {\
  if (flag_is_on(opt_panel_DEBUG,flag_SHOW_TODO)) { \
    char buffer[100];\
    sprintf(buffer,"--[ to be IMPLEMENTED: %s ]--\n",S);\
    ca_message_string(buffer); } } while(FALSE)

#define OPTIMIZE(S) do {\
  if (flag_is_on(opt_panel_DEBUG,flag_SHOW_OPTIMIZE)) { \
    char buffer[100];\
    sprintf(buffer,"--[ to be OPTIMIZED: %s ]--\n",S);\
    ca_message_string(buffer); } } while(FALSE)

#else

#define TODO(S)
#define OPTIMIZE(S)

#endif

/* ---------------------------------------------------------------------- *\

	Function Free MACROS

\* ---------------------------------------------------------------------- */

#ifdef DEBUG

#define myfree(S,P,D) debug_free(S,P,D)

#else

#ifdef COCOA_WITH_RUM
#define myfree(S,P,D) (((S)<=RUM_N_STACKS) ? rum_free(S,P) : free(P))
#else
#define myfree(S,P,D) free(P)
#endif

#endif

/* ----no RUM_N_STACKS check--- */

#ifdef DEBUG

#define myfree1(S,P,D) debug_free1(S,P,D)

#else

#ifdef COCOA_WITH_RUM
#define myfree1(S,P,D) rum_free(S,P)
#else
#define myfree1(S,P,D) free(P)
#endif

#endif

/* ---------------------------------------------------------------------- *\

	Function Alloc MACROS

\* ---------------------------------------------------------------------- */


#ifdef DEBUG

#define mymalloc(S,D) debug_malloc(S,D)

#else

#ifdef COCOA_WITH_RUM
#define mymalloc(S,D) (((S)<=RUM_N_STACKS) ? rum_malloc(S) : malloc(S))
#else
#define mymalloc(S,D) malloc(S)
#endif

#endif

/* ----no RUM_N_STACKS check--- */

#ifdef DEBUG

#define mymalloc1(S,D) debug_malloc1(S,D)

#else

#ifdef COCOA_WITH_RUM
#define mymalloc1(S,D) rum_malloc(S)
#else
#define mymalloc1(S,D) malloc(S)
#endif

#endif

/*  ==========  memory management  ==========  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* #include <limits.h> */ /* TODO: remove this line */

#ifdef DEBUG
extern size_t memused;
extern size_t maxmemused;
extern size_t MPmemused;
extern size_t MPmaxmemused;
#endif

char * my_strdup_nospaces(char*s);
char * my_strdup(char * s);
char * my_upper(char*s);
char * my_strsum(char * s1, char * s2);
/* return s1+s2, allocating the space for the result */
int my_strcasecmp(const char *s1, const char *s2);

int str_split(char * s, char * sep, char ** s1, char ** s2);
char * my_strtok(char * s1, const char * s2);
void test_strtok(void);

void initmem(void);
void *debug_malloc(size_t size,char *message);
void *debug_malloc1(size_t size,char *message);
void *mycalloc(int n,size_t size,char *message);
void debug_free(size_t size, void *p, char *message);
void debug_free1(size_t size, void *p, char *message);
void *myrealloc(void *oldptr, size_t old_size, size_t new_size, char * message);
void printmem(char *message);
void my_fputs(char *s, FILE * f);

/*  ==========  string management  ==========  */

#define malloc_str(size) (char *)mymalloc((size+1)*sizeof(char),"string")
#define free_str(size,p) myfree((size+1)*sizeof(char),p,"string")
#define free_str1(p) do { \
  if (p != NULL) myfree((strlen(p)+1)*sizeof(char),p,"string");} while(0)

#define realloc_str(m, old_n, new_n) \
  (char *)myrealloc(m, (old_n+1), (new_n+1),"string")

char * path_tail(char * path);
char * pkg_path(char * pkg_id);

/*  ==========  integers management  ==========  */

int int_binom(int n, int k);
int  IntGCD(register int m, register int n);
int  IntLCM (int x, int y);
coc_bool IsPrime( long int N );

/*  ==========  errors management  ==========  */

#define INTERNAL_ERROR(s) ca_internal_error_string(s)


/* -----------[ charbuf ]-------------- */

#define malloc_charbuf() \
  (charbuf)mymalloc(sizeof(charbuf_info),"charbuf")
#define free_charbuf(B) myfree(sizeof(charbuf_info),B,"charbuf")

#define charbuf_buffer(B) ((B)->buffer)
#define charbuf_len(B) ((B)->len)
#define charbuf_size(B) ((B)->size)
#define charbuf_blocksize(B) ((B)->blocksize)

#define charbuf_set_buffer(B, b) (B)->buffer=b
#define charbuf_set_len(B, n) (B)->len=n
#define charbuf_set_size(B, n) (B)->size=n
#define charbuf_set_blocksize(B, n) (B)->blocksize=n

charbuf charbuf_init(int size, int blocksize);
void charbuf_free(charbuf B);
void charbuf_enlarge(charbuf B);
void charbuf_putc(charbuf B, char c);
void charbuf_puts(charbuf B, char * s);
void charbuf_cork(charbuf B);
void charbuf_test();


#endif
