/* $Id: vlist.c,v 4.4.1.1 2004/06/18 14:24:25 cocoa Exp $ */

#include "cocdefs.h"

#include "ca.h"
#include "ck.h"
#include "cterm.h"
#include "ctbuf.h"
#include "coeffs.h"
#include "evector.h"
#include "ivectors.h"
#include "monomial.h"
#include "operator.h"
#include "options.h"
#include "polys.h"
#include "pvcomp.h"
#include "pvector.h"
#include "ring.h"
#include "vlist.h"
#include "terms.h"

#ifdef PVCOMP_MONOPOLI

term vlist_elem_nth_comp_lt(vlist_atom E, int N) 
{
  monopoli aux_p;
  aux_p=pv_nth(vlist_atom_vector(E),N);
  if (aux_p == NULL) return term_NULL;
  else return monomial_term(aux_p);
}

#endif

/* return a duplicate of the lt of the nth component of vlist_atom */
term vlist_elem_nth_comp_lt_dup(vlist_atom E, int N, ring R) 
{
  pvcomp aux_p;
  aux_p=pv_nth(vlist_atom_vector(E),N);
  if (aux_p == NULL) return term_NULL;
  return pvcomp_leading_term(aux_p,R);
}

vlist vlist_init(void)
{
  vlist res_vl;
  res_vl=malloc_vlist();
  vlist_set_len(res_vl,0);
  vlist_set_first(res_vl,vlist_atom_NULL);
  vlist_set_last(res_vl,vlist_atom_NULL);
  return res_vl;
}

void vlist_insert_at_head(vlist vl, vlist_atom a)
{
  vlist_atom_set_next(a,vlist_first(vl));
  vlist_incr_len(vl);
  vlist_set_first(vl,a);
  if (vlist_len(vl)==1 /* was empty */) vlist_set_last(vl,a);
}

void vlist_Insert_AntiDeg(vlist vl, evector v)
{
  /* insert v before the first element w with Deg(LT(w))>=Deg(LT(v)) */
  int n,D;
  vlist_atom aux_atom,new_atom,aux_atom_prec;
  new_atom=vlist_mk_atom(v);
  n=vlist_len(vl);
  if (n==0) { vlist_insert_at_head(vl,new_atom); return; }
  aux_atom=vlist_first(vl);
  aux_atom_prec=vlist_atom_NULL;
  D = ev_deg_lt(v);
  while (n-- > 0) {
    if (ev_deg_lt(vlist_atom_ev(aux_atom))>=D) {
      if (aux_atom_prec == vlist_atom_NULL) 
        vlist_insert_at_head(vl,new_atom);
      else {
        vlist_atom_set_next(new_atom, vlist_atom_next(aux_atom_prec));
        vlist_atom_set_next(aux_atom_prec, new_atom);
        vlist_incr_len(vl); }
      return;
    }
    aux_atom_prec=aux_atom;
    aux_atom=vlist_atom_next(aux_atom); }
  vlist_append(vl,new_atom);
}

vlist_atom vlist_mk_atom(evector ev)
{
  vlist_atom res_atom;
  res_atom=malloc_vlist_atom();
  vlist_atom_set_ev(res_atom,ev_link(ev));
  vlist_atom_unmark(res_atom);
  vlist_atom_set_next(res_atom,vlist_atom_NULL);
  return res_atom;
}

void vlist_mark_nth(vlist vl, int n)
{
  vlist_atom aux_atom;
  int l=vlist_len(vl);
  aux_atom=vlist_first(vl);
  if (n>l) return; /* ERROR */
  while (n-- > 1) aux_atom=vlist_atom_next(aux_atom);
  vlist_atom_mark(aux_atom);
}

vlist_atom vlist_get_first(vlist vl)
/* NOTE: for efficiency is assumed that vl is not empty */
{
  vlist_atom res_atom;
  res_atom=vlist_first(vl);
  vlist_decr_len(vl);
  vlist_set_first(vl,vlist_atom_next(res_atom));
  vlist_atom_set_next(res_atom,vlist_atom_NULL);
  return res_atom;
}

evector vlist_get_first_ev(vlist vl)
/* NOTE: for efficiency is assumed that vl is not empty */
{
  vlist_atom aux_atom;
  evector res_ev;
  aux_atom=vlist_first(vl);
  vlist_decr_len(vl);
  vlist_set_first(vl,vlist_atom_next(aux_atom));
  res_ev=vlist_atom_ev(aux_atom);
  free_vlist_atom(aux_atom);
  return res_ev;
}

vlist vlist_change_order(vlist vl, ring R1, ring R2)
{
  vlist res_vl;
  pvector aux_v;
  vlist_atom el;
  int n;

  res_vl = vlist_init();
  n = vlist_len(vl);
  el = vlist_first(vl);

  while (n-- > 0) {
    aux_v = vlist_atom_whole_vector(el,R1);
    pv_change_order(aux_v,R1,R2);
    vlist_append_this_vector(res_vl, aux_v, R2);
    el = vlist_atom_next(el); }

  return res_vl;
}

void vlist_to_rules(vlist vl, ring R)
{
  vlist_atom el;
  int n;
 
  n = vlist_len(vl);
  el = vlist_first(vl);
  while (n-- > 0) {
    ev_to_rule(vlist_atom_ev(el),R);
    el = vlist_atom_next(el); }
}

void vlist_reinit_vectors(vlist vl, ring R)
{
  vlist_atom el;
  int n;
 
  n = vlist_len(vl);
  el = vlist_first(vl);
  while (n-- > 0) {
    ev_reinit(vlist_atom_ev(el),R);
    el = vlist_atom_next(el); }
}

vlist vlist_tail(vlist vl, ring R)
{
  vlist res_vl;
  vlist_atom aux_atom;
  int n;
  res_vl=vlist_init();
  if (vlist_len(vl)<=1) return res_vl;
  vlist_append(vl,vlist_get_first(vl));
  n=vlist_len(vl)-1;
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    vlist_append(res_vl,vlist_atom_dup(aux_atom,R));
    vlist_append(vl,aux_atom); }
  return res_vl;
}

void vlist_append(vlist vl, vlist_atom a)
{
  if (vlist_atom_next(a) != vlist_atom_NULL) {
    my_fputs("INTERNAL ERROR: vlist_append: wrong vlist_atom\n",stdout);
  }
  if (vlist_is_empty(vl))
    vlist_set_first(vl,a);
  else
    vlist_atom_set_next(vlist_last(vl),a);
  vlist_incr_len(vl);
  vlist_set_last(vl,a);
}

vlist vlist_concat_these(vlist vl1, vlist vl2)
{
  if (vlist_is_empty(vl1)) return vl2;
  if (vlist_is_empty(vl2)) return vl1;
  vlist_atom_set_next(vlist_last(vl1),vlist_first(vl2));
  vlist_set_len(vl1,vlist_len(vl1)+vlist_len(vl2));
  vlist_set_last(vl1,vlist_last(vl2));
  free_vlist(vl2); /* just the header */
  return vl1;
}

void vlist_free(vlist * vl, ring R)
{
  vlist_atom aux_el;
  int n;
  if (*vl==vlist_NULL) return;
  n=vlist_len(*vl);
  while (n-- > 0) {
    aux_el=vlist_get_first(*vl);
    vlist_atom_free(&aux_el,R); }
  free_vlist(*vl);
  *vl=vlist_NULL;
}

void vlist_atom_free(vlist_atom * atom, ring R)
{
  if (vlist_atom_ev(*atom) != ev_NULL) 
    ev_unlink(vlist_atom_ev(*atom),R);
  free_vlist_atom(*atom);
}

vlist vlist_get_rels(vlist vl, ring R)
{
  vlist res_vl;
  int n;
  pvector pv;
  vlist_atom aux_el;
  if (vl==vlist_NULL) return vlist_NULL;
  res_vl=vlist_init();
  n=vlist_len(vl);
  while (n-- > 0) {
    aux_el=vlist_get_first(vl);
    pv=ev_rel(vlist_atom_ev(aux_el));
    if (pv != pv_NULL)
      vlist_append_ev(res_vl,ev_init(pv_dup(pv,R),R));
    vlist_append(vl,aux_el); 
  }
  if (vlist_len(res_vl) == 0) {
    vlist_free(&res_vl,R);
    return vlist_NULL;
  }
  return res_vl;
}

vlist vlist_dup(vlist vl, ring R)
{
  vlist res_vl;
  int n;
  vlist_atom aux_el;
  if (vl==vlist_NULL) return vlist_NULL;
  res_vl=vlist_init();
  n=vlist_len(vl);
  while (n-- > 0) {
    aux_el=vlist_get_first(vl);
    vlist_append(res_vl,vlist_atom_dup(aux_el,R));
    vlist_append(vl,aux_el); }
  return res_vl;
}

vlist vlist_dup_trunc(vlist vl, int D, ring R)
{
  vlist res_vl;
  int n;
  vlist_atom aux_el;
  if (vl==vlist_NULL) return vlist_NULL;
  res_vl=vlist_init();
  n=vlist_len(vl);
  while (n-- > 0) {
    aux_el=vlist_get_first(vl);
    if (term_degree(ev_lt(vlist_atom_ev(aux_el))) <= D)
      vlist_append(res_vl,vlist_atom_dup(aux_el,R));
    vlist_append(vl,aux_el); }
  return res_vl;
}

vlist_atom vlist_atom_dup(vlist_atom atom, ring R)
{
  vlist_atom res_atom;
  res_atom=malloc_vlist_atom();
  vlist_atom_set_ev(res_atom,ev_dup(vlist_atom_ev(atom),R));
  vlist_atom_set_mark(res_atom,vlist_atom_is_marked(atom));
  vlist_atom_set_next(res_atom,vlist_atom_NULL);
  return res_atom;
}

void vlist_append_this_vector(vlist vl,pvector v,ring R)
{
  evector aux_ev;
  aux_ev=ev_init(v,R);
  vlist_append(vl,vlist_mk_atom(aux_ev));
  ev_unlink(aux_ev,R);
}

void vlist_append_vector(vlist vl,pvector v,ring R)
{
  evector aux_ev;
  aux_ev=ev_init(pv_dup(v,R),R);
  vlist_append(vl,vlist_mk_atom(aux_ev));
  ev_unlink(aux_ev,R);
}

coc_bool vlist_isin(vlist vl, pvector v, ring R)
{
  vlist_atom aux_atom;
  int n=vlist_len(vl);
  aux_atom=vlist_first(vl);
  while (n-- > 0) {
    if (pv_eq(vlist_atom_vector(aux_atom),v,R)) return TRUE;
    aux_atom=vlist_atom_next(aux_atom); }
  return FALSE;
}

vlist vlist_union_these(vlist vl1, vlist vl2, ring R)
{
  vlist_atom aux_atom;
  int n=vlist_len(vl2);
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl2);
    if (!vlist_isin(vl1,vlist_atom_vector(aux_atom),R)) 
      vlist_append(vl1,aux_atom);
    else vlist_atom_free(&aux_atom,R); }
  free_vlist(vl2);
  return vl1;
}

void vlist_mult_this_by_pvcomp(vlist * vl, pvcomp p, ring R)
{
  vlist_atom aux_atom;
  evector aux_ev;
  pvector aux_v, aux_v1;
  int n=vlist_len(*vl);
  while(n-- > 0) {
    aux_atom=vlist_get_first(*vl);
    aux_ev=vlist_atom_ev(aux_atom);
    if (ev_is_split(aux_ev)) {
      aux_v1=ev_whole_vector(aux_ev,R);
      aux_v=pv_mult(p,aux_v1,R);
      pv_free(aux_v1,R);
    } else {
      aux_v=pv_mult(p,ev_vector(aux_ev),R);
    }
    vlist_atom_free(&aux_atom,R);
    aux_ev=ev_init(aux_v,R);
    vlist_append(*vl,vlist_mk_atom(aux_ev));
    ev_unlink(aux_ev,R);
  }
}

vlist vlist_mult_by_pvcomp(vlist vl, pvcomp p, ring R)
{
  vlist res_vl;
  res_vl=vlist_dup(vl,R);
  vlist_mult_this_by_pvcomp(&res_vl,p,R);
  return res_vl;
}

vlist vlist_only_leading_terms(vlist vl, ring R)
{
  vlist res_vl;
  vlist_atom aux_atom;
  evector aux_ev;
  pvector aux_v;
  int n=vlist_len(vl);
  res_vl=vlist_init();
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_ev=vlist_atom_ev(aux_atom);
    aux_v=term2pv(ev_lt(aux_ev),ev_lmpos(aux_ev),ev_rank(aux_ev),R);
    aux_ev=ev_init(aux_v,R);
    vlist_append(res_vl,vlist_mk_atom(aux_ev));
    ev_unlink(aux_ev,R);
    vlist_append(vl,aux_atom); }
  return res_vl;
}

shifts vlist_leading_terms_to_shifts(vlist vl,ring R)
{
  vlist_atom aux_atom;
  evector aux_ev;
  pvector aux_v;
  term aux_t;
  shifts res_ss;
  int i,n;
  n=vlist_len(vl);
  res_ss=shifts_init(n);
  for (i=1;i<=n;i++) {
    aux_atom=vlist_get_first(vl);
    aux_ev=vlist_atom_ev(aux_atom);
    if (ev_lm(aux_ev) == NULL) aux_t=term_init(ring_indets_num(R));
    else aux_t=term_dup(ev_lt(aux_ev));
    shifts_set_nth(res_ss,i,aux_t);
    vlist_append(vl,aux_atom); }
  return res_ss;
}

cterm vlist2cterm(vlist vl, ring R)
{
  cterm aux_t;
  ctbuf aux_b;
  vlist_atom aux_el;
  int n=vlist_len(vl);
  aux_b=ctbuf_empty();
  while (n-- > 0) {
    aux_el=vlist_get_first(vl);
    aux_t=ev2cterm_normal(vlist_atom_ev(aux_el),R);
    ctbuf_append(aux_b,aux_t);
    vlist_append(vl,aux_el); }
  return ctbuf_mk_cterm(aux_b,op_LIST);
}

cterm vlist2cterm_plus(vlist vl, ring R)
{
  cterm aux_t;
  ctbuf aux_b;
  cterm aux_param[2];
  vlist_atom aux_atom;
  int n=vlist_len(vl);
  aux_b=ctbuf_empty();
  while (n-- > 0) {
    aux_atom=vlist_get_first(vl);
    aux_param[0]=monomial2cterm(ev_lm(vlist_atom_ev(aux_atom)),R);
    aux_param[1]=pv2cterm(vlist_atom_vector(aux_atom),R);
    aux_t=cterm_mk_classic(op_PAIR,2,aux_param);
    ctbuf_append(aux_b,aux_t);
    vlist_append(vl,aux_atom); }
  return ctbuf_mk_cterm(aux_b,op_LIST);
}

void printf_vlist(vlist vl,ring R) 
{
  cterm t;
  t=vlist2cterm(vl,R);
  put_cterm(t);
  cterm_free(t);
}  

void printf_vlist_plus(vlist vl,ring R) 
{
  cterm t;
  t=vlist2cterm_plus(vl,R);
  put_cterm(t);
  cterm_free(t);
}  

void vlist_make_monic(vlist vl, ring R)
{
  vlist_atom aux_el;
  int n=vlist_len(vl);
  aux_el=vlist_first(vl);
  while (n-- > 0) {
    vlist_atom_make_monic(aux_el,R);
    aux_el=vlist_atom_next(aux_el); }
}

void vlist_enlarge_elements(vlist vl, int r, shifts ss)
{
  vlist_atom aux_el;
  pvector aux_v;
  evector aux_ev;
  int n=vlist_len(vl);
  aux_el=vlist_first(vl);
  while (n-- > 0) {
    aux_ev=vlist_atom_ev(aux_el);
    if (ev_red(aux_ev) != pv_NULL) pv_enlarge(ev_red(aux_ev),r,ss);
    if (ev_tail(aux_ev) != pv_NULL) pv_enlarge(ev_tail(aux_ev),r,ss);
    aux_el=vlist_atom_next(aux_el); }
}

int vlist_rank(vlist vl)
{
  int res_n;
  vlist_atom aux_el;
  evector aux_ev;
  int n=vlist_len(vl);
  aux_el=vlist_first(vl);
  res_n=1;
  while (n-- > 0) {
    aux_ev=vlist_atom_ev(aux_el);
    res_n=MAX(res_n,ev_rank(aux_ev));
    aux_el=vlist_atom_next(aux_el); }
  return res_n;
}

void vlist_rank_and_shifts(vlist vl,int * r,shifts * ss)
{
  int n, aux_n;
  vlist_atom aux_el;
  evector aux_ev;
  n=vlist_len(vl);
  aux_el=vlist_first(vl);
  *ss=shifts_NULL;
  *r=0;
  while (n-- > 0) {
    aux_ev=vlist_atom_ev(aux_el);
    aux_n=ev_rank(aux_ev);
    if (aux_n > *r) {
      *r=aux_n;
      *ss=ev_shifts(aux_ev);
    }
    aux_el=vlist_atom_next(aux_el); }
}

void vlist_apply_shifts(vlist vl, shifts ss, ring R)
{
  vlist_atom aux_atom;
  int n;
  if (vl==vlist_NULL) return;
  n=vlist_len(vl);
  aux_atom=vlist_first(vl);
  while (n-- > 0) {
    ev_apply_shifts(vlist_atom_ev(aux_atom),ss,R);
    aux_atom=vlist_atom_next(aux_atom); }
}

void vlist_force_rc1(vlist vl, ring R)
{
  vlist_atom aux_atom;
  int n;
  if (vl==vlist_NULL) return;
  n=vlist_len(vl);
  aux_atom=vlist_first(vl);
  while (n-- > 0) {
    ev_force_rc1(&vlist_atom_ev(aux_atom),R);
    aux_atom=vlist_atom_next(aux_atom); }
}

vlist vlist_homog(vlist pl, unknown x, ring R)
{
  if (pl==vlist_NULL) return vlist_NULL;
  else {
    vlist res_pl;
    int n;
    vlist_atom aux_el;
    res_pl=vlist_init();
    n=vlist_len(pl);
    while (n-- > 0) {
      aux_el=vlist_get_first(pl);
      vlist_append(res_pl,vlist_atom_homog(aux_el,x,R));
      vlist_append(pl,aux_el); }
    return res_pl; }
}

vlist_atom vlist_atom_homog(vlist_atom el, unknown x, ring R)
{
  vlist_atom res_el;
  res_el=malloc_vlist_atom();
  vlist_atom_set_ev(res_el,ev_homog(vlist_atom_ev(el),x,R));
  vlist_atom_set_mark(res_el,vlist_atom_is_marked(el));
  return res_el;
}
