/* vlist.h : 0.1 : 12 Jan 96 : ADT list of extended vectors */

#ifndef VLIST_H
#define VLIST_H

#ifndef AUXTYPES_H
# include "auxtypes.h"
#endif

#ifndef COCTYPES_H
# include "coctypes.h"
#endif

/* HERE ev is DIRECTLY the atom of the list: these 2 macros
   are for modularity */

#define vlist_atom_ev(A) ((A)->ev)
#define vlist_atom_vector(A) ev_vector(vlist_atom_ev(A))
#define vlist_atom_whole_vector(A,R) ev_whole_vector(vlist_atom_ev(A),R)
#define vlist_atom_poly(A) pv_nth(vlist_atom_vector(A),0)

#define vlist_atom_next(E) ((E)->next)
#define vlist_atom_is_marked(E) ((E)->marked)
#define vlist_atom_is_not_marked(E) (!(E)->marked)
#define vlist_atom_mark(E) (E)->marked=TRUE
#define vlist_atom_unmark(E) (E)->marked=FALSE
#define vlist_atom_set_mark(E,M) (E)->marked=M
#define vlist_atom_set_next(E,N) (E)->next=N
#define vlist_atom_set_ev(A,EV) (A)->ev=EV

#define vlist_first(PL) ((PL)->first)
#define vlist_last(PL) ((PL)->last)
#define vlist_len(PL) ((PL)->len)
#define vlist_len_ck(PL) ((PL)==vlist_NULL ? 0 : (PL)->len)

#define vlist_set_first(PL,E) (PL)->first=E
#define vlist_set_last(PL,E) (PL)->last=E
#define vlist_set_len(PL,L) (PL)->len=L


#define vlist_NULL NULL
#define vlist_atom_NULL NULL

#define vlist_first_ev(VL) vlist_atom_ev(vlist_first(VL))
#define vlist_first_vector(VL) ev_vector(vlist_first_ev(VL))
#define vlist_first_poly(VL) pv_nth(vlist_first_vector(VL),0)

#define vlist_incr_len(PL) (PL)->len++
#define vlist_decr_len(PL) (PL)->len--

#define vlist_is_empty(PL) (PL==vlist_NULL || vlist_len(PL)==0)

#define malloc_vlist() \
  (vlist)mymalloc(sizeof(vlist_header),"vlist")
#define free_vlist(p) myfree(sizeof(vlist_header),p,"vlist")

#define malloc_vlist_atom() \
  (vlist_atom)mymalloc(sizeof(vlist_atom_info),"vlist_atom")
#define free_vlist_atom(p) myfree(sizeof(vlist_atom_info),p,"vlist_atom")


#define vlist_atom_make_monic(A,R) ev_make_monic(vlist_atom_ev(A),R)


#ifdef PVCOMP_MONOPOLI
term vlist_elem_nth_comp_lt(vlist_atom E, int N);
#endif

term vlist_elem_nth_comp_lt_dup(vlist_atom E, int N, ring R);
vlist vlist_init(void);

void vlist_insert_at_head(vlist vl, vlist_atom a);
void vlist_Insert_AntiDeg(vlist vl, evector v);

vlist_atom vlist_mk_atom(evector v);

void vlist_mark_nth(vlist vl, int n);

vlist vlist_change_order(vlist vl, ring R1, ring R2);

vlist_atom vlist_get_first(vlist PL);
/* NOTE: for efficiency is assumed that pl is not empty */

evector vlist_get_first_ev(vlist PL);
/* NOTE: for efficiency is assumed that pl is not empty */

void vlist_to_rules(vlist vl, ring R);
void vlist_reinit_vectors(vlist vl, ring R);

vlist vlist_tail(vlist vl, ring R);
/* return a copy of vl without the first element */

int vlist_rank(vlist vl);
void vlist_rank_and_shifts(vlist vl,int * r,shifts * ss);

void vlist_append(vlist PL, vlist_atom E);
vlist vlist_concat_these(vlist PL1, vlist PL2);

#define vlist_append_ev(VL, EV) vlist_append(VL,vlist_mk_atom(EV))

#define vlist_concat( PL1, PL2, R) \
  vlist_concat_these(vlist_dup(PL1,R),vlist_dup(PL2,R))

void vlist_free(vlist * pl, ring R);
void vlist_atom_free(vlist_atom * atom, ring R);


vlist vlist_get_rels(vlist vl, ring R);
vlist vlist_dup(vlist pl, ring R);
vlist vlist_dup_trunc(vlist pl, int D, ring R);
vlist_atom vlist_atom_dup(vlist_atom atom, ring R);
void vlist_append_this_vector(vlist pl,pvector p,ring R);
void vlist_append_vector(vlist pl,pvector p,ring R);
coc_bool vlist_isin(vlist pl, pvector p, ring R);
vlist vlist_union_these(vlist pl1, vlist pl2, ring R);
void vlist_mult_this_by_pvcomp(vlist * pl, pvcomp p, ring R);
vlist vlist_mult_by_pvcomp(vlist pl, pvcomp p, ring R);
vlist vlist_only_leading_terms(vlist pl, ring R);
shifts vlist_leading_terms_to_shifts(vlist vl, ring R);
cterm vlist2cterm(vlist pl, ring R);
cterm vlist2cterm_plus(vlist vl, ring R);
void printf_vlist(vlist pl,ring R);
void printf_vlist_plus(vlist vl,ring R);

void vlist_enlarge_elements(vlist vl, int r, shifts ss);

void vlist_make_monic(vlist pl, ring R);

void vlist_force_rc1(vlist vl, ring R);
void vlist_apply_shifts(vlist vl, shifts ss, ring R);
vlist vlist_homog(vlist pl, unknown x, ring R);
vlist_atom vlist_atom_homog(vlist_atom el, unknown x, ring R);

#endif
