#include "DMPZfactors.h"
#include "jalloc.h"


DMPZfactors DMPZfactors_ctor()
{
  DMPZfactors this;

  this = (DMPZfactors)MALLOC(sizeof(struct DMPZfactors_struct));
  this->list = NULL;
  this->multiplicity = 1;
  mpz_init_set_ui(this->content, 1);
  return this;
}


void DMPZfactors_dtor(DMPZfactors this)
{
  mpz_clear(this->content);
  DMPZlist_dtor(this->list);
  FREE(this);
}

void DMPZfactors_content(DMPZfactors this, mpz_t n)
{
  mpz_mul(this->content, this->content, n);
}


void DMPZfactors_negate(DMPZfactors this)
{
  if (this->multiplicity & 1) mpz_neg(this->content, this->content);
}


void DMPZfactors_add(DMPZfactors this, const DMPZ f)
{
  DMPZ fcopy;
  
  fcopy = DMPZcopy(f);
  /* force LC (wrt to deglex) > 0; order deglex imposed by DMPZfactor(..) */
  fcopy = DMPZsort(fcopy, DMPZorder_deglex);
  if (mpz_sgn(fcopy->coeff) < 0) DMPZnegate(fcopy);
  this->list = DMPZlist_append(this->list,
                               DMPZlist_ctor(fcopy, this->multiplicity));
}

void DMPZfactors_multiplicity(DMPZfactors this, int m)
{
  this->multiplicity = m;
}

int DMPZfactors_nfactors(const DMPZfactors this)
{
  DMPZlist iter;
  int nfactors;

  nfactors = 0;
  for (iter=this->list; iter; iter = iter->next)
    nfactors += iter->deg;
  return nfactors;
}
