#ifndef DMPZfactors_H
#define DMPZfactors_H

#include "DMPZ.h"
#include "DMPZlist.h"

/* To make sure that the product of the factors in the list multiplied    */
/* by the content is equal to the original polynomial (with correct sign) */
/* the sign of the leading coefficient (wrt to deglex) of the original    */
/* polynomial must be included in the integer content; all polynomial     */
/* factors are forced to have positive leading coefficient (wrt deglex).  */

struct DMPZfactors_struct
{
  mpz_t content;    /* signed integer content */
  DMPZlist list;    /* list of factors, each tagged with its multiplicity */
  int multiplicity; /* used by DMPZfactors_add to know multiplicity of factor */
};

typedef struct DMPZfactors_struct *DMPZfactors;


DMPZfactors DMPZfactors_ctor();
void DMPZfactors_dtor(DMPZfactors A);
void DMPZfactors_content(DMPZfactors A, mpz_t n);
void DMPZfactors_negate(DMPZfactors A);
void DMPZfactors_add(DMPZfactors A, const DMPZ f);
void DMPZfactors_multiplicity(DMPZfactors A, int m);
int DMPZfactors_nfactors(const DMPZfactors A);


#endif
