#include "jerror.h"
#include "DUPFF.h"
#include "DUPFFmod.h"

void DUPFFmul3mod(DUPFF ans, const DUPFF x, const DUPFF y, const DUPFF m)
{
  if (ans == m) { JERROR(JERROR_ALIASING); return; }
  if (ans->maxdeg < DUPFFdeg(x) + DUPFFdeg(y))
  {
    JERROR(JERROR_DEG_TOO_LOW);
    ans->deg = -1;
    return;
  }
  DUPFFmul3(ans, x, y);
  DUPFFrem2(ans, m);
}


void DUPFFexpt3mod(DUPFF ans, const DUPFF base, const int power, const DUPFF m)
{
  if (power < 1) { JERROR(JERROR_EXPT_MOD); ans->deg = -1; return; }
  if (ans->maxdeg < 2*DUPFFdeg(base)-2)
  {
    JERROR(JERROR_DEG_TOO_LOW);
    ans->deg = -1;
    return;
  }
  if (power == 1)
  {
    DUPFFcopy2(ans, base);
    return;
  }
  DUPFFexpt3mod(ans, base, power/2, m);
  DUPFFmul3mod(ans, ans, ans, m);
  if ((power&1) == 0) return;
  DUPFFmul3mod(ans, ans, base, m);
}
