
/* Standard "read once" trick */
#ifndef DUPZ_H
#define DUPZ_H
/***************************************************************************/
/* File really starts here                                                 */

/* DUPZs are dense univariate polynomials with big integer coefficients.   */


/* we have to include stdio.h before gmp.h otherwise i/o functions vanish */
#include <stdio.h>
#include <gmp.h>

struct DUPZstruct
{
  int maxdeg;
  int deg;
  mpz_t *coeffs;
};

typedef struct DUPZstruct *DUPZ;

DUPZ DUPZnew(int maxdeg);
DUPZ int_to_DUPZ(int i);
void DUPZinit(DUPZ x, int maxdeg);
void DUPZfree(DUPZ x);
void DUPZassign(DUPZ lhs, const DUPZ rhs);
void DUPZswap(DUPZ x, DUPZ y);
DUPZ DUPZcopy(const DUPZ f);
void DUPZcopy2(DUPZ dest, const DUPZ src);
int DUPZequal(const DUPZ x, const DUPZ y);

int DUPZdeg(const DUPZ f);
#define DUPZlc(f) (f->coeffs[f->deg])
void DUPZshift(DUPZ f, int xpower);
void DUPZreverse(DUPZ f);

void DUPZneg1(DUPZ f);
void DUPZadd3(DUPZ sum, const DUPZ x, const DUPZ y);
void DUPZsub3(DUPZ diff, const DUPZ x, const DUPZ y);
DUPZ DUPZadd(const DUPZ x, const DUPZ y);
DUPZ DUPZsub(const DUPZ x, const DUPZ y);
DUPZ DUPZmul(const DUPZ x, const DUPZ y);
void DUPZmul3(DUPZ ans, const DUPZ x, const DUPZ y);
void DUPZsquare(DUPZ f);
void DUPZmul2z(DUPZ f, const mpz_t c);
void DUPZmod2z(DUPZ f, const mpz_t c);

void DUPZshift_add(DUPZ f, const DUPZ g, const int deg, const mpz_t coeff);
void DUPZshift_sub(DUPZ f, const DUPZ g, const int deg, const mpz_t coeff);

/* DUPZdiv2z does "floor" division, and ignores non-zero remainders */
void DUPZdiv2z(DUPZ f, mpz_t n);
/* All the division functions below return "prematurely" if an inexact division is found */
DUPZ DUPZdiv(const DUPZ num, const DUPZ den);
DUPZ DUPZrem(const DUPZ num, const DUPZ den);
void DUPZrem2(DUPZ x, const DUPZ m);
void DUPZdiv4(DUPZ quot, DUPZ rem, const DUPZ num, const DUPZ den);

void DUPZmod2(DUPZ f, DUPZ g, mpz_t Q);
void DUPZmdiv2(DUPZ f, DUPZ g, mpz_t Q);
void DUPZmmul3z(DUPZ product, DUPZ f, mpz_t c, mpz_t Q);
void DUPZlinear_shift(DUPZ f, mpz_t a);

/* To avoid trouble with Macintoshes which cannot printf... */
#ifdef FACTOR_DEBUG
#include "DUPZprint.h"
#endif

#endif

