#include "DUPZcra.h"
#include "jerror.h"
#include "mpz_cra_ui.h"

/**********************************************************************/
/* NOTE the CRA construction is symmetric (-modulus/2,..., modulus/2) */
/**********************************************************************/

int DUPZcra(DUPZ f1, const mpz_t m1, const DUPFF f2, FFelem m2)
{
  int i, changed, df1, df2;
  mpz_t tmp; /* used as workspace in mpz_cra_ui_raw */
  FFelem m1modm2;

  df1 = DUPZdeg(f1);
  df2 = DUPFFdeg(f2);
  if (df2 > f1->maxdeg) { JERROR(JERROR_DEG_TOO_LOW); return 1; }

  mpz_init(tmp);
  m1modm2 = mpz_fdiv_ui(m1, (unsigned long)m2);
  changed = 0;
  for (i=1+df1; i <= df2; i++) mpz_set_ui(f1->coeffs[i], 0);
  for (i=0; i <= df2; i++)
    changed |= mpz_cra_ui_raw(f1->coeffs[i], m1, f2->coeffs[i], m2, tmp, m1modm2);
  for (i=1+df2; i <= df1; i++)
    changed |= mpz_cra_ui_raw(f1->coeffs[i], m1, 0, m2, tmp, m1modm2);

  if (df2 > df1) f1->deg = df2;
  mpz_clear(tmp);
  return changed;
}
