#ifndef DUPZcra_h
#define DUPZcra_h

#include <gmp.h>
#include "DUPZ.h"
#include "DUPFF.h"

/* This Chinese Remaindering function computes a symmetric modular value. */
/* i.e. the value is in the range -(m1*m2)/2 to +(m1*m2)/2.               */

/* f1 is a DUPZ with coefficients in the range -m1/2 to m1/2,             */
/* m1 is a positive (odd?) number (the modulus for f1),                   */
/* f2 is a DUPFF over the current finite field with m2 elements,          */
/* m2 is the characteristic of the current finite field (odd small prime) */
/* It is necessary that m1 and m2 be coprime.                             */
/* The argument f1 is modified to have coefficients in the range -m1*m2/2 */
/* to +m1*m2/2, so that it equals f1 modulo m1, and equals f2 modulo m2.  */
/* Result is 0 if the coeffs of f1 were unchanged, otherwise none-zero.   */

int DUPZcra(DUPZ f1, const mpz_t m1, const DUPFF f2, FFelem m2);

/* BUGS: the argument m2 appears to be quite superfluous.                 */

#endif
