#include <stddef.h>
#include "DUPZfactor_info.h"
#include "logi.h"
#include "jalloc.h"
#include "DUPZfactor_liftq.h"

DUPZfactor_info DUPZfactor_info_ctor(DUPZ f, DUPZfactors output)
{
  int i, df;
  DUPZfactor_info ans;

  ans = (DUPZfactor_info)MALLOC(sizeof(struct DUPZfactor_info_struct));
  ans->f = f; /* deliberately do not copy */
  ans->irreds = output;
  ans->nprimes = 0;
  ans->max_nprimes = 25;
  ans->FFq = (FF*)MALLOC(ans->max_nprimes*sizeof(struct FFstruct));
  ans->qfactors = (DUPFFlist*)MALLOC(ans->max_nprimes*sizeof(DUPFFlist));
  df = DUPZdeg(f);
  ans->fds = (int*)MALLOC((1+df)*sizeof(int));
  for (i=0; i <= df; i++) ans->fds[i] = 1;

  mpz_init(ans->Q);
  mpz_init(ans->recip_lcf);
  /* Set these fields to NULL to mean "unused" */
  ans->bounds = NULL;
  ans->lifter = NULL;

  return ans;
}


void DUPZfactor_info_dtor(DUPZfactor_info this)
{
  int i;

  mpz_clear(this->recip_lcf);
  mpz_clear(this->Q);
  DUPZfree(this->f);
  for (i=0; i < this->nprimes; i++)
  {
    FFdtor(this->FFq[i]);
    DUPFFlist_dtor(this->qfactors[i]);
  }
  FREE(this->FFq);
  FREE(this->qfactors);
  if (this->fds) FREE(this->fds);
  /* The two fields below may be "unused", if so don't call their dtors */
  if (this->bounds) DUPZfactor_bound_dtor(this->bounds);
  if (this->lifter) DUPZfactor_lift_dtor(this->lifter);
  FREE(this);
}

/***************************************************************************/
/* Set the target_height based on information in bounds and the prime.     */

void DUPZfactor_info_set_target_height(DUPZfactor_info this)
{
  this->target_height = 1 + (int)((logi(2) + this->bounds->leading_coeff + this->bounds->lift_bound)/logi(this->p));
}
