#include <stddef.h>
#include "DUPZfactors.h"
#include "jalloc.h"
#include "jerror.h"


DUPZfactors DUPZfactors_ctor()
{
  DUPZfactors this;

  this = (DUPZfactors)MALLOC(sizeof(struct DUPZfactors_struct));
  this->list = NULL;
  this->multiplicity = 1;
  this->reversed = 0;
  mpz_init_set_ui(this->content, 1);
  return this;
}


void DUPZfactors_dtor(DUPZfactors this)
{
  mpz_clear(this->content);
  DUPZlist_dtor(this->list);
  FREE(this);
}

void DUPZfactors_content(DUPZfactors this, mpz_t n)
{
  mpz_mul(this->content, this->content, n);
}


void DUPZfactors_negate(DUPZfactors this)
{
  if (this->multiplicity & 1) mpz_neg(this->content, this->content);
}


void DUPZfactors_add(DUPZfactors this, const DUPZ f)
{
  DUPZ fcopy;
  
#ifdef FACTOR_DEBUG
  printf("%d ", DUPZdeg(f));
#endif
  fcopy = DUPZcopy(f);
  if (this->reversed) DUPZreverse(fcopy);
  /* The next lines just make sure that the lc of each factor is >0       */
  /* Also negate the "content" if lc < 0 and the multiplicity is odd.     */
  if (mpz_sgn(DUPZlc(fcopy)) < 0)
  {
    DUPZneg1(fcopy);
    if (this->multiplicity&1) mpz_neg(this->content, this->content);
  }
  this->list = DUPZlist_append(this->list,
                               DUPZlist_ctor(fcopy, this->multiplicity));
}

void DUPZfactors_multiplicity(DUPZfactors this, int m)
{
  this->multiplicity = m;
}

void DUPZfactors_reversed(DUPZfactors this, int yes_no)
{
  this->reversed = yes_no;
}

int DUPZfactors_nfactors(const DUPZfactors this)
{
  DUPZlist iter;
  int nfactors;

  nfactors = 0;
  for (iter=this->list; iter; iter = iter->next)
    nfactors += iter->deg;
  return nfactors;
}
