#include "FindPrimRoot.h"
#include "primes.h"

/* No number less than 2^32 can have more than 10 distinct prime factors.*/
static FFelem prime_factor[10];
static FFelem prime_power[10];

static void Factorize32bits(FFelem n)
/* Factors are placed in prime_factor[], and exponents in prime_power[]. */
/* The end of the factors is marked by prime_factor[k] = 1.              */
{
  FFelem p;
  int nfactors = 0;

  for (p = 2;; p = nextprime(p))
  {
    if (p*p > n) break;
    if (n%p != 0) continue;

    prime_factor[nfactors] = p;
    prime_power[nfactors] = 1;
    n /= p;
    while (n%p == 0)
    {
      n /= p;
      prime_power[nfactors]++;
    }
    nfactors++;
  }
  prime_factor[nfactors] = n;
  if (n==1) return;
  prime_power[nfactors] = 1;
  prime_factor[nfactors+1] = 1;
}


static FFelem BinaryPowerModP(FFelem n, FFelem e, FFelem p) /* p<65536 */
{
  if (e==0) return 1;
  if (e==1) return n;
  if ((e&1)==0) return BinaryPowerModP((n*n)%p, e/2, p);
  return (BinaryPowerModP((n*n)%p, e/2, p) * n)%p;
}


static int FindPrimRootTest(FFelem r, FFelem p)
{
  int i, q;

  i = 0;
  for (q=prime_factor[i]; q != 1; q=prime_factor[++i])
    if (BinaryPowerModP(r, p/q, p) == 1) return 0;
  return 1;
}

FFelem FindPrimRoot(FFelem p)
{
  int q;

  if (p == 2) return 1;
  if (p > MAX_PRIME || !isprime(p)) return 0; /* 0 indicates failure */
  Factorize32bits(p-1);
  for (q=2;;q++)
    if (FindPrimRootTest(q, p)) return q;
}
