#ifndef GBM_h
#define GBM_h

#include <gmp.h>
#include "jaa.h"
#include "FFmat.h"
#include "Qmat.h"

/***************************************************************************/

struct GeneralizedBuchbergerMoeller_struct
{
  int nideals;   /* number of (0-diml) ideals we are intersecting          */
const  int *quotdim;  /* vector space dimension of the quotient by each ideal   */
/*const*/  FFelem **one;  /* vector repr of 1 modulo each input ideal               */
/*const*/  FFmat **Mmult; /* for each ideal a collection of multiplication matrices*/
  
  int nvars;     /* number of variables = dimension of the space           */
  int npp;       /* number of power products currently in use              */
  int npp_max;   /* maximum number of power products we can manage         */
                 /* this value can be increased by calling BM_grow         */
  int maxdeg;    /* highest degree PP used in projective case; 0 for affine*/
  int GBsize;    /* number of elements in the Groebner basis               */
  int *GB;       /* indices of the rows of M which corr to GB elems        */
  int *sep;      /* indices of the rows of M which corr to separators      */
  int **pp;      /* the power product which gave rise to this row of M     */
  int ncols;     /* number of columns in the matrix M (below)              */
  int nseps;     /* the number of "separators" found so far                */
  FFelem **image;/* the image of the corr PP modulo each input ideal       */
  FFelem **M;    /* matrix of reductions of the PPs modulo given ideals,   */
                 /* subsequently row-reduced.                              */
  int (*cmp)(const void*,const void*); /* power product order              */
};
typedef struct GeneralizedBuchbergerMoeller_struct *GBM;


/***************************************************************************/
/* This is the public interface.                                           */

GBM GBM_affine_mod_p(int nvars, int nideals, const int *quotdim, /*const*/ FFelem **one, /*const*/ FFmat **Mmult, int (*cmp)(const void*, const void*));
void GBM_dtor(GBM this);
void GBM_affine(GBM *trace_out, Qmat *GBtails_out, int nvars, int nideals, const int *quotdim, mpq_t **one, Qmat **Mmult, int (*cmp)(const void*, const void*));

/*
void BM_affine(BMGB *char0_ptr, BM *modp_ptr, int nvars, int npoints, mpz_t **points, int (*cmp)(const void*, const void*));
void BM_projective(BMGB *char0_ptr, BM *modp_ptr, int nvars, int npoints, mpz_t **points, int (*cmp)(const void*, const void*));
void BM_dtor(BM this);
void BMGB_dtor(BMGB this);
BM BM_affine_mod_p(int nvars, int npoints, FFelem **points, int (*cmp)(const void*, const void*));
BM BM_projective_mod_p(int nvars, int npoints, FFelem **points, int (*cmp)(const void*, const void*));
*/

#if 0
void BM_print_GBasis(BMGB char0, BM modp);
void BM_print_separators(BMGB char0, BM modp);
#endif

#endif
