#include "add_logs.h"

#include <math.h>

/***************************************************************************/
/* Compute log(1+x) without losing too much precision if x is small.       */
/* Some systems already have this function (perhaps called log1p),         */
/* nonetheless we define it here as it is not present on all systems.      */
/* Implementation is simple rather than especially fast/accurate/...       */

static double log1plus(double x)
{
  if (x > 0.01) return log(1.0+x);
  return x*(1-x*(1/2.0-x*(1/3.0-x*(1/4.0-x/5.0))));
}

/***************************************************************************/
/* Assuming log1=log(x), and log2=log(y), compute log(x+y).                */
/* Replacing log1plus(x) by log(1+x) causes noticeable loss of precision   */
/* only if the larger of log1 and log2 is very small.  Since add_logs is   */
/* not called very often, extreme speed is not essential.                  */

double add_logs(double log1, double log2)
{
  if (log1 < log2 - 40) return log2;
  if (log2 < log1 - 40) return log1;
  if (log1 < log2) return log2 + log1plus(exp(log1-log2));
  return log1 + log1plus(exp(log2-log1));
}
