/* I am not allowed to use slashes in include strings so the misguided */
/* can use this code on non-Unix machines.  Use -I when compiling.     */

/* There are currently 4 calls to arbint2... functions.  The MPW       */
/* compiler on the Macintosh complains unless I cast the args to void* */

#include <stdlib.h>
#include "jintf.h"

/* this is from directory factor (but ought to be in GMP) */
#include "mpz_lcm.h"
#include "mpz_alias.h"
#include "primes.h"
#include "BM.h"
#include "GBM.h"
#include "FF.h"
#include "PPstream.h"
#include "WGD.h"

/* These are from directory multivariate */
#include "DMPZgcd.h"
#include "DMPZfactor.h"
#include "DMPZfactor_modp.h"
#include "DMPZevalQ.h"
/* These are from directory linalg */
#include "Qsolve.h"
#include "Zkernel.h"
#include "Zdet.h"
#include "FFdet.h"

/***************************************************************************/
/* CoCoA includes */

#include "ca.h"
#include "ck.h"
#include "coctypes.h"
#include "coeffs.h"
#include "cterm.h"
#include "engine.h"
#include "ivectors.h"
#include "mathcoc.h"
#include "monomial.h"
#include "object.h"
#include "olist.h"
#include "omat.h"
#include "options.h"
#include "polys.h"
#include "ring.h"
#include "terms.h"
#include "utils.h"

/***************************************************************************/
#ifdef ARPOLY_MONOPOLI

/* Compute in D common denominator of coeffs of f if f has rational coeffs. */
/* Otherwise D is set to 1.                                                 */
static void arpoly_common_denom(mpz_t D, arpoly f, ring R)
{
  monomial iter;
  
  mpz_set_ui(D, 1);
  if (ring_characteristic(R) != characteristic_Lr) return;
  for (iter = f; iter != NULL; iter = monomial_next(iter))
    mpz_lcm(D, D, mpq_denref((verylongrat)monomial_coeff(iter)));
}

static DMPZ arpoly2DMPZ(arpoly poly, ring R)
{
  monomial iter;
  int var, *exps;
  int charR = ring_characteristic(R);
  arbint z;
  DMPZ f;

  NVARS = R->indetsNo;
  f = NULL;
  for (iter = poly; iter != NULL; iter = monomial_next(iter))
  {
    exps = (int*)malloc(NVARS*sizeof(int));
    for (var = 0; var < NVARS; var++)
      exps[var] = term_get_true_nth(monomial_term(iter), var+1, R);
    if (charR > 1)
      z = int2arbint((int)monomial_coeff(iter));
    else z = this_coeff2arbint(coeff_dup(monomial_coeff(iter), R), R);
    f = DMPZprepend(z, exps, f);
    mpz_clear(z); free_verylongint(z);
  }
  return DMPZreverse(f); /* restore original order of the terms */
}

/* Converts D*poly to a DMPZ -- D should be the result of arpoly_common_denom */
static DMPZ arpoly2DMPZ_common_denom(arpoly poly, ring R, mpz_t D)
{
  monomial iter;
  int var, *exps;
  mpz_t coeff;
  DMPZ f;

  mpz_init(coeff);
  NVARS = R->indetsNo;
  f = NULL;
  for (iter = poly; iter != NULL; iter = monomial_next(iter))
  {
    exps = (int*)malloc(NVARS*sizeof(int));
    for (var = 0; var < NVARS; var++)
      exps[var] = term_get_true_nth(monomial_term(iter), var+1, R);
    mpz_divexact(coeff, D, mpq_denref((verylongrat)monomial_coeff(iter)));
    mpz_mul(coeff, coeff, mpq_numref((verylongrat)monomial_coeff(iter)));
    f = DMPZprepend(coeff, exps, f);
  }
  mpz_clear(coeff);
  return DMPZreverse(f); /* restore original order of the terms */
}

#elif defined(ARPOLY_TRIPOLI_OR_EEE)

static DMPZ arpoly2DMPZ(arpoly poly, ring R)
{
  TODO("arpoly2DMPZ");
  return NULL;
}

#endif

static arpoly DMPZ2arpoly(DMPZ poly, ring R)
{
  DMPZ iter;
  int var;
  term aux_t;
  coeff aux_c;
  monomial aux_m;
  arpoly res_p;
  
  res_p = NULL;
  for (iter = poly; iter; iter = iter->next)
  {
    if (mpz_sgn(iter->coeff) == 0) continue;
    aux_c = arbint2coeff((void*)iter->coeff, R);
    aux_t = term_init(R->indetsNo);
    for (var = 0; var < NVARS; var++)
      term_put_nth(aux_t, (var+1), iter->exps[var] * iv_get_nth(R->weights, var+1));
    aux_m = monomial_mk(aux_c, aux_t);
    res_p = arpoly_sum_these(res_p, arpoly_this_monomial2arpoly(aux_m, R), R);
  }
  return res_p;
}

static olist this_DMPZfactors2olist(DMPZfactors facs, ring R)
{
  DMPZlist iter;
  olist res_ol, aux_ol;
  int i, n, content;
  n = DMPZlist_length(facs->list);
  content = 0;
  if (mpz_cmp_ui(facs->content, 1) != 0) content = 1;
  res_ol = olist_init(n+content);
  for (i=1, iter=facs->list; iter; i++, iter = iter->next)
  {
    aux_ol = olist_init(2);
    olist_set_nth(aux_ol, 1,
                  arpoly2object(DMPZ2arpoly(iter->poly, R), R));
    olist_set_nth(aux_ol, 2, int2object(iter->deg));
    olist_set_nth(res_ol, i, olist2object(aux_ol));
  }
  if (content)
  {
    aux_ol = olist_init(2);
    olist_set_nth(aux_ol, 1, arpoly2object(monomial_mk(arbint2coeff((void*)facs->content, R), term_init(R->indetsNo)), R));
    olist_set_nth(aux_ol, 2, int2object(1));
    olist_set_nth(res_ol, n+content, olist2object(aux_ol));
  }
                           
  DMPZfactors_dtor(facs);
  return res_ol;
}


object cocoa_DMPZfactor(arpoly poly, ring R)
{
  DMPZ f;
  olist ans;
  int p = ring_characteristic(R);

  if (p > 1 && !isprime(p)) return cocoa_error("BAD_CHAR");
  f = arpoly2DMPZ(poly, R);
  if (p < 2)
    ans = this_DMPZfactors2olist(DMPZfactor(f), R);
  else
    ans = this_DMPZfactors2olist(DMPZfactor_modp(f, p), R);
  DMPZdtor(f);
  return olist2object(ans);
}


arpoly arpoly_gcd_JAA(arpoly fin, arpoly gin, ring R)
{
  DMPZ f, g, gcd;
/*  int p = ring_characteristic(R); NOT YET USED */
  arpoly ans;
  mpz_t D;

  /* Can't do multivariate GCD in char>0 yet.   */
  /* if (p > 1) return cocoa_error("BAD_CHAR"); */
  mpz_init(D);
  arpoly_common_denom(D, fin, R);
  f = arpoly2DMPZ_common_denom(fin, R, D);
  arpoly_common_denom(D, gin, R);
  g = arpoly2DMPZ_common_denom(gin, R, D);
  mpz_clear(D);
  gcd = DMPZgcd(f, g);
  ans = DMPZ2arpoly(gcd, R);
  DMPZdtor(gcd);
  DMPZdtor(g);
  DMPZdtor(f);
  return ans;
}

/***************************************************************************/

object jaa_eval(arpoly f, olist point, ring R)
{
  int i, nvars;
  DMPZ poly;
  arbrat tmp;
  mpz_t D;
  mpq_t *value, *p;
  
  mpz_init_set_ui(D, 1);
  if (ring_characteristic(R) != characteristic_Lr)
    poly = arpoly2DMPZ(f, R);
  else
  {
    arpoly_common_denom(D, f, R);
    poly = arpoly2DMPZ_common_denom(f, R, D);
  }
  nvars = ring_indets_num(R);
  p = (mpq_t*)malloc(nvars*sizeof(mpq_t));
  for (i=0; i < nvars; i++)
  {
    tmp = object_cast_arbrat(object_link(olist_nth(point, i+1)));
    mpq_init(p[i]);mpq_set(p[i], tmp);
    mpq_clear(tmp); free_verylongrat(tmp); /* don't ask me why... */
  }
  NVARS = nvars;
  value = (mpq_t*)malloc(sizeof(mpq_t));
  mpq_init(*value);
  DMPZevalQ(*value, poly, p);
  DMPZdtor(poly);
  for (i=0; i < nvars; i++) mpq_clear(p[i]);
  free(p);
  mpz_mul(mpq_denref(*value), mpq_denref(*value), D);
  mpq_canonicalize(*value);
  mpz_clear(D);
  return arbrat2object(*value); /* *value is owned by caller */
}


/***************************************************************************/

/* Return value is the modulus, or zero if an error occurred. */
static FFelem omat_is_FFmat(omat Min)
{
  int nrows = omat_nrow(Min);
  int ncols = omat_ncol(Min);
  int i, j;
  FFelem p, q;
  p = q = 0;

  for (i=0; i < nrows; i++)
    for (j=0; j < ncols; j++)
    {
      if (object_type(omat_ij(Min, i+1, j+1)) != objtype_zmod) return 0;
      q = zmod_p(omat_ij(Min, i+1, j+1)->v.zmod_value);
      if (p == 0) p = q;
      if (p != q) return 0; /* ERROR: mixed moduli! */
    }

  return p;
}


static void omat_to_FFmat(FFmat *Mout, omat Min)
{
  int i, j, nrows, ncols;
  
  nrows = omat_nrow(Min);
  ncols = omat_ncol(Min);
  *Mout = FFmat_ctor(nrows, ncols);

  for (i=0; i < nrows; i++)
    for (j=0; j < ncols; j++)
      (*Mout)->entry[i][j] = object_cast_int(object_link(omat_ij(Min, i+1, j+1)));
}


static omat FFmat_to_omat(FFelem p, FFmat M)
{
  const int nrows = M->nrows;
  const int ncols = M->ncols;
  omat ans = omat_init(nrows, ncols);
  int i, j;

  for (i=0; i < nrows; ++i)
    for (j=0; j < ncols; ++j)
      omat_set_ij(ans, i+1, j+1, zmod2object(zmod_mk(M->entry[i][j], p)));
  return ans;
}


static FFmat FFmat_transpose(FFmat M)
{
  const int nrows = M->nrows;
  const int ncols = M->ncols;
  FFmat ans = FFmat_ctor(ncols, nrows);
  int i, j;

  for (i=0; i < nrows; ++i)
    for (j=0; j < ncols; ++j)
      ans->entry[j][i] = M->entry[i][j];
  return ans;
}


static FFmat FFkernel2(FFmat M)
{
  const int nrows = M->nrows;
  const int ncols = M->ncols;
  FFmat ans;
  FFelem *Mi, *Mj, *c, *x;
  int i, j, k, s, *d;
  FFelem q, Mik;
  int KerDim;
  unsigned int count;
  FFelem shift = CurrentFF.shift;
  FFelem limit = CurrentFF.k;
  FFelem p = CurrentFF.prime;


  c = (FFelem*)malloc(nrows*sizeof(FFelem));
  d = (int*)malloc(ncols*sizeof(int));
  for (i=0; i < nrows; i++) c[i] = 0;

  KerDim = 0;
  count = 0;
  for (k=0; k < ncols; k++)
  {
    for (j=0; j < nrows; j++) M->entry[j][k] %= p;
    for (j=0; j < nrows; j++)
      if (c[j] == 0 && M->entry[j][k] != 0) break;
    if (j == nrows) { KerDim++; d[k] = -1; continue; }
    Mj = M->entry[j];
    q = p - FFdiv(1, Mj[k]);
    for (i=k+1; i < ncols; i++) Mj[i] = FFmul(Mj[i]%p, q);
    for (i=0; i < nrows; i++)
    {
      if (i == j) continue;
      Mi = M->entry[i];
      Mik = Mi[k];
      if (Mik == 0) continue;
      Mi[k] = 0;
      /* for (s=k+1; s < ncols; s++) Mi[s] += Mik*Mj[s]; */
      {
        FFelem *Mis = &Mi[k+1],
               *Mjs = &Mj[k+1],
               *last = &Mi[ncols-1];
        while (Mis <= last) *(Mis++) += Mik * *(Mjs++);
      }
    }
    c[j] = k+1; /* this is only a boolean */
    d[k] = j;
    if (++count < limit) continue;
    count = 0;
    for (i=0; i < nrows; i++)
    {
      Mi = M->entry[i];
      for (j=k+1; j < ncols; j++)
        if (Mi[j] >= shift) Mi[j] -= shift;
    }
  }

  ans = FFmat_ctor(KerDim, ncols);
  if (KerDim == 0) return ans;
  s = 0;
  x = ans->entry[s];
  for (k=0; k < ncols; k++)
  {
    if (d[k] != -1) continue;
    for (i=0; i < ncols; i++)
    {
      if (d[i] == -1) x[i] = (i == k);
      else x[i] = M->entry[d[i]][k];
    }
    s++;
    x = ans->entry[s];
  }
  free(c);
  free(d);

  return ans;
}


olist lin_kernel(const omat M)
{
  olist answer, vec;
  mpz_t *copy;
  Zmat matrix, basis;
  mpz_t D, gcd, tmp2;
  mpq_t tmpq;
  arbrat tmp;
  int kerdim, nrows, ncols, i, j;
  FFelem p = 0;

  if ((p = omat_is_FFmat(M)))
  {
    FFmat Mp, Kp;

    omat_to_FFmat(&Mp, M);
    {
      FF Fp = FFctor(p);
      FFselect(Fp);
      Kp = FFkernel2(Mp);
      FFdtor(Fp);
    }

    if (Kp == 0) { FFmat_dtor(Mp); return olist_init(0); }
    ncols = Kp->ncols;
    answer = olist_init(Kp->nrows);
    for (i=0; i < Kp->nrows; i++)
    {
      vec = olist_init(ncols);
      for (j=0; j < ncols; j++)
	olist_set_nth(vec, j+1, zmod2object(zmod_mk(Kp->entry[i][j], p)));
      olist_set_nth(answer, i+1, olist2object(vec));
    }
    FFmat_dtor(Kp);
    FFmat_dtor(Mp);
    return answer;
  }
  /* We shall automatically transpose the input matrix!! */
  nrows = omat_ncol(M);
  ncols = omat_nrow(M);

  mpz_init_set_ui(D, 1);
  mpz_init(tmp2);
  mpz_init(gcd);
  mpq_init(tmpq);
  for (i=0; i < nrows; i++)
  {
    for (j=0; j < ncols; j++)
    {
      tmp = object_cast_arbrat(object_link(omat_ij(M, j+1, i+1)));
      if (math_error != object_NULL)
      {
        mpq_clear(tmpq);mpz_clear(gcd);mpz_clear(tmp2);mpz_clear(D);
        return olist_init(0);
      }
      mpq_set(tmpq, tmp);
      mpq_clear(tmp); free_verylongrat(tmp);
      mpz_gcd(gcd, D, mpq_denref(tmpq));
      mpz_divexact(tmp2, mpq_denref(tmpq), gcd);
      mpz_mul(D, D, tmp2);
    }
  }

  matrix = Zmat_ctor(nrows, ncols);
  for (i=0; i < nrows; i++)
  {
     for (j=0; j < ncols; j++)
     {
       tmp = object_cast_arbrat(object_link(omat_ij(M, j+1, i+1)));
       mpq_set(tmpq, tmp);
       mpq_clear(tmp); free_verylongrat(tmp);
       mpz_set(matrix->entry[i][j], mpq_numref(tmpq));
       mpz_divexact(tmp2, D, mpq_denref(tmpq));
       mpz_mul(matrix->entry[i][j], matrix->entry[i][j], tmp2);
     }
  }
  kerdim = Zmat_left_kernel_Zbasis(&basis, matrix);
  answer = olist_init(kerdim);
  for (i=0; i < kerdim; i++)
  {
    vec = olist_init(nrows);
    for (j=0; j < nrows; j++)
    {
      copy = (mpz_t*)MALLOC(sizeof(mpz_t));
      mpz_init(*copy);
      mpz_set(*copy, basis->entry[i][j]);
      olist_set_nth(vec, j+1, arbint2object((void*)*copy));
    }
    olist_set_nth(answer, i+1, olist2object(vec));
  }

  Zmat_dtor(matrix);
  Zmat_dtor(basis);
  mpz_clear(tmp2);
  mpq_clear(tmpq);
  mpz_clear(D);
  mpz_clear(gcd);

  return answer;
}


olist lin_solve(const omat M, const olist L)
{
  olist answer = olist_NULL;
  int i, j, nrows, ncols, rank, error, r;
  mpq_t *copy;
  Qmat matrix, soln, RHS;
  arbrat tmp;
  
  nrows = omat_nrow(M);
  ncols = omat_ncol(M);
  r = olist_len(L);                  /* number of RHS vectors */
  if (r == 0) return olist_init(0);  /* no RHS vectors */
  error = 0;
  for (i=1; i <= r; i++)
  {
    olist tmp = object_cast_olist(object_link(olist_nth(L, i)));
    if (math_error != object_NULL || olist_len(tmp) != nrows) error = 1;
    olist_free(tmp);
    if (error) break;
  }
  if (error) { math_error=cocoa_error("BAD_ARGS"); return olist_init(0); }

  matrix = Qmat_ctor(nrows, ncols);
  RHS = Qmat_ctor(nrows, r);
  soln = Qmat_ctor(ncols, r);
  rank = -1; /* Used also as a flag to detect bad input */
  for (i=0; i < nrows; i++)
  {
    for (j=0; j < ncols; j++)
    {
      tmp = object_cast_arbrat(object_link(omat_ij(M, i+1, j+1)));
      if (tmp == NULL) { math_error=cocoa_error("BAD_ARGS"); goto tidy_up; }
      mpq_set(matrix->entry[i][j], tmp);
      mpq_clear(tmp); free_verylongrat(tmp);
    }
  }
  for (j=0; j < r; j++)
  {
    olist v = object_cast_olist(object_link(olist_nth(L, j+1)));
    for (i=0; i < nrows; i++)
    {
      tmp = object_cast_arbrat(object_link(olist_nth(v, i+1)));
      if (tmp == NULL) { olist_free(v); math_error=cocoa_error("BAD_ARGS"); goto tidy_up; }
      mpq_set(RHS->entry[i][j], tmp);
      mpq_clear(tmp); free_verylongrat(tmp);
    }
    olist_free(v);
  }
  rank = Qsolve(soln, matrix, RHS);

  /* Now make a copy of the solution to be owned by answer */
  answer = olist_init(r);
  for (j=0; j < r; j++)
  {
    olist ansv;
    if (mpz_sgn(mpq_denref(soln->entry[0][j])) == 0) ansv = olist_init(0);
    else
    {
      ansv = olist_init(ncols);
      for (i=0; i < ncols; i++)
      {
        copy = (mpq_t*)MALLOC(sizeof(mpq_t));
        mpq_init(*copy);
        mpq_set(*copy, soln->entry[i][j]);
        olist_set_nth(ansv, i+1, arbrat2object(*copy));
      }
    }
    olist_set_nth(answer, j+1, olist2object(ansv));
  }
tidy_up:
  Qmat_dtor(matrix);
  Qmat_dtor(RHS);
  Qmat_dtor(soln);

  if (rank < 0) return olist_init(0);
  return answer;
}


omat Qmat_inverse(const omat M)
{
  omat answer=omat_NULL;
  int i, j, n, rank;
  Qmat matrix, inverse, identity;
  mpq_t *copy;
  arbrat tmp;
  
  n = omat_nrow(M);

  /* Convert matrix to our format... */
  matrix = Qmat_ctor(n, n);
  inverse = Qmat_ctor(n, n);
  identity = Qmat_ctor(n, n);
  for (i=0; i < n; i++)
  {
    for (j=0; j < n; j++)
    {
      tmp = object_cast_arbrat(object_link(omat_ij(M, i+1, j+1)));
      mpq_set(matrix->entry[i][j], tmp);
      mpq_clear(tmp); free_verylongrat(tmp);
    }
    mpq_set_ui(identity->entry[i][i], 1, 1);
  }

  rank = Qsolve(inverse, matrix, identity);
  if (rank == n) /* if matrix was invertible... */
  {
    /* ... make a copy of the solution to be owned by answer */
    answer = omat_init(n, n);
    for (i=0; i < n; i++)
    {
      for (j=0; j < n; j++)
      {
        copy = (mpq_t*)MALLOC(sizeof(mpq_t));
        mpq_init(*copy);
        mpq_set(*copy, inverse->entry[i][j]);
        omat_set_ij(answer, i+1, j+1, arbrat2object(*copy));
      }
    }
  }
  
  /* Free memory */
  Qmat_dtor(matrix);
  Qmat_dtor(inverse);
  Qmat_dtor(identity);

  if (rank != n) {
    error_set_param(string2cterm("(det = 0)"));
    math_error=cocoa_error("CANNOT_INV");
  }

  return answer;
}


/* ASSUMES input matrix is square */
/* Here we compute det if matrix is of integers or of integers modulo N. */
/* If N is not prime, we just compute integer det and then reduce mod N; */
/* it is not easy to do this better as N need not be square-free.        */
object try_Zdet(omat M)
{
  FFelem p = 0, q;
  int i, j, n;
  mpz_t *D;
  Zmat matrix;
  arbint tmp;
  
  n = omat_nrow(M);
  matrix = Zmat_ctor(n, n);
  for (i=0; i < n; i++)
  {
    for (j=0; j < n; j++)
    {
      if (object_type(omat_ij(M, i+1, j+1)) == objtype_zmod)
      {
        q = zmod_p(omat_ij(M, i+1, j+1)->v.zmod_value);
        if (p == 0) p = q;
        if (p != q) { Zmat_dtor(matrix); return object_NULL; } /* mixed moduli! */
      }
      tmp = object_cast_arbint(object_link(omat_ij(M, i+1, j+1)));
      if (tmp == NULL || math_error != object_NULL) { Zmat_dtor(matrix); return object_NULL; }
      mpz_set(matrix->entry[i][j], tmp);
      mpz_clear(tmp); free_verylongint(tmp);
    }
  }
  if (isprime(p))
  {
    FFelem D;
    FF Fp = FFctor(p);
    FFmat matrix_p;
    FFselect(Fp);
    matrix_p = FFmat_ctor(n, n);
    Zmat_to_FFmat(matrix_p, matrix);
    D = FFdet(matrix_p);
    FFmat_dtor(matrix_p);
    Zmat_dtor(matrix);
    FFdtor(Fp);
    return zmod2object(zmod_mk_norm(D, p));
  }
  /* Either an integer det or the modulus is not prime */
  D = (mpz_t*)MALLOC(sizeof(mpz_t));
  mpz_init(*D);
  Zdet(*D, matrix);
  Zmat_dtor(matrix);
  if (p == 0) return arbint2object((void*)*D);
  q = mpz_fdiv_ui(*D, p);
  return zmod2object(zmod_mk_norm(q, p));
}


/***************************************************************************/


static int deg(int nvars, int *pp)
{
  int d, i;
  
  d = 0;
  for (i=0; i < nvars; i++) d += pp[i];
  return d;
}

static int pp_cmp_nvars;
static ring pp_cmp_ring;
/* ASSUMES the two args are never equal */
static int pp_cmp(const void *arg1, const void *arg2)
{
  int var, ans;
  int **pp1 = (int**)arg1;
  int **pp2 = (int**)arg2;
  ring R = pp_cmp_ring;
  int nvars = pp_cmp_nvars;
  term PP1, PP2;

  PP1 = term_init(R->indetsNo);
  for (var = 0; var < nvars; var++)
    term_put_nth(PP1, (var+1), pp1[0][var] * iv_get_nth(R->weights, var+1));
  PP2 = term_init(R->indetsNo);
  for (var = 0; var < nvars; var++)
    term_put_nth(PP2, (var+1), pp2[0][var] * iv_get_nth(R->weights, var+1));
  ans = term_is_gt(PP1, PP2, R);
  term_free(PP1); term_free(PP2);
  if (ans) return 1;
  return -1;
}



object BM_interface(object L, ring R)
{
  int i, j, tmp, nvars, npoints, affine, modular;
  FFelem **points_p = NULL; /* to keep compiler quiet */
  mpz_t **points = NULL;    /* to keep compiler quiet */
  arbint entry;
  olist args, l, point;
  int p = 0;                /* to keep compiler quiet */
  FF Fp = NULL;             /* to keep compiler quiet */

  modular = (ring_characteristic(R) > 1);
  if (modular)
  {
    p = ring_characteristic(R);
    Fp = FFctor(p);
    if (Fp == NULL) return cocoa_error("BAD_CHAR");
    FFselect(Fp);
  }
  args = object2olist(object_link(L));
  entry = object_cast_arbint(object_link(olist_nth(args, 1)));
  affine = mpz_sgn(*(mpz_t*)(entry));
  mpz_clear(entry); free_verylongint(entry);
  l = object2olist(object_link(olist_nth(args, 2)));
  olist_free(args);
  npoints = olist_len(l);
  if (npoints > 0) nvars = olist_len(olist_nth(l, 1)->v.olist_value);
  else nvars = 0;

  if (modular)
  {
    points_p = (FFelem**)malloc(npoints*sizeof(FFelem*));
    for (i=0; i < npoints; i++)
    {
      points_p[i] = (FFelem*)malloc(nvars*sizeof(FFelem));
      point = object2olist(object_link(olist_nth(l, i+1)));
      for (j=0; j < nvars; j++)
      {
        entry = object_cast_arbint(object_link(olist_nth(point, j+1)));
        points_p[i][j] = mpz_fdiv_ui(entry, p);
        /* I bet this leaks memory -- Antonio can fix it; it's his stupid interface */
        mpz_clear(entry); free_verylongint(entry);
      }
      olist_free(point);
    }
  }
  else /* non-modular case */
  {
    points = (mpz_t**)malloc(npoints*sizeof(mpz_t*));
    for (i=0; i < npoints; i++)
    {
      points[i] = (mpz_t*)malloc(nvars*sizeof(mpz_t));
      for (j=0; j < nvars; j++) mpz_init(points[i][j]);
      point = object2olist(object_link(olist_nth(l, i+1)));
      for (j=0; j < nvars; j++)
      {
        entry = object_cast_arbint(object_link(olist_nth(point, j+1)));
        mpz_set(points[i][j], entry);
        /* I bet this leaks memory -- Antonio can fix it; it's his stupid interface */
        mpz_clear(entry); free_verylongint(entry);
      }
      olist_free(point);
    }
  }
  olist_free(l);
  if (affine)
  {
    int GBsize, var;
    arpoly GBelem, sep;
    olist ans, GB, seps;
    BMGB char0;
    BM modp;
    coeff c;
    monomial summand;
    term PP;
            
    pp_cmp_ring = R;
    pp_cmp_nvars = nvars;
    if (modular) modp = BM_affine_mod_p(nvars, npoints, points_p, pp_cmp);
    else BM_affine(&char0, &modp, nvars, npoints, points, pp_cmp);
    pp_cmp_ring = NULL;
    pp_cmp_nvars = 0;

    if (modp == NULL) return olist2object(olist_init(0));/* to indicate "error" */
    GBsize = modp->GBsize;
    GB = olist_init(GBsize);
    for (i=0; i < GBsize; i++)
    {
      c = int2coeff(1, R);
      PP = term_init(R->indetsNo);
      for (var = 0; var < nvars; var++)
        term_put_nth(PP, (var+1), (modp->pp[modp->GB[i]][var] * iv_get_nth(R->weights, var+1)));
      summand = monomial_mk(c, PP);
      GBelem = arpoly_this_monomial2arpoly(summand, R);
      for (j=0; j < npoints; j++)
      {
        if (modular)
        {
          tmp = modp->M[modp->GB[i]][j+npoints];
          if (tmp == 0) continue;
          c = int2coeff(tmp, R);
        }
        else
        {
          if (mpq_sgn(char0->GB[i][j]) == 0) continue;
          c = arbrat2coeff(char0->GB[i][j], R);
        }
        PP = term_init(R->indetsNo);
        for (var = 0; var < nvars; var++)
          term_put_nth(PP, (var+1), modp->pp[modp->sep[j]][var] * iv_get_nth(R->weights, var+1));
        summand = monomial_mk(c, PP);
        GBelem = arpoly_sum_these(GBelem, arpoly_this_monomial2arpoly(summand, R), R);
      }
      olist_set_nth(GB, i+1, arpoly2object(GBelem, R));
    }

    seps = olist_init(npoints);
    for (i=0; i < npoints; i++)
    {
      sep = NULL;
      for (j=npoints-1; j >= 0; j--)
      {
        if (modular)
        {
          tmp = modp->M[modp->sep[i]][j+npoints];
          if (tmp == 0) continue;
          c = int2coeff(tmp, R);
        }
        else
        {
          if (mpq_sgn(char0->sep[i][j]) == 0) continue;
          c = arbrat2coeff(char0->sep[i][j], R);
        }
        PP = term_init(R->indetsNo);
        for (var = 0; var < nvars; var++)
          term_put_nth(PP, (var+1), modp->pp[modp->sep[j]][var] * iv_get_nth(R->weights, var+1));
        summand = monomial_mk(c, PP);
        sep = arpoly_sum_these(sep, arpoly_this_monomial2arpoly(summand, R), R);
      }
      olist_set_nth(seps, i+1, arpoly2object(sep, R));
    }

    BM_dtor(modp);
    if (modular) { FFdtor(Fp); for (i=0; i < npoints; i++) FREE(points_p[i]); FREE(points_p); }
    else { BMGB_dtor(char0); for (i=0; i < npoints; i++) { for (j=0; j < nvars; j++) mpz_clear(points[i][j]); FREE(points[i]);} FREE(points); }
    ans = olist_init(2);
    olist_set_nth(ans, 1, olist2object(GB));
    olist_set_nth(ans, 2, olist2object(seps));
    return olist2object(ans);
  }

  /* else projective case... */
  else
  {
    int GBsize, var, n, top, k, d;
    arpoly GBelem, sep;
    olist ans, GB, seps;
    BMGB char0;
    BM modp;
    coeff c;
    monomial summand;
    term PP;

    pp_cmp_ring = R;
    pp_cmp_nvars = nvars;
    if (modular) modp = BM_projective_mod_p(nvars, npoints, points_p, pp_cmp);
    else BM_projective(&char0, &modp, nvars, npoints, points, pp_cmp);
    pp_cmp_ring = NULL;
    pp_cmp_nvars = 0;
    if (modp == NULL) return olist2object(olist_init(0));/* to indicate "error" */
    GBsize = modp->GBsize;
    GB = olist_init(GBsize);
    for (i=0; i < GBsize; i++)
    {
      c = int2coeff(1, R);
      PP = term_init(R->indetsNo);
      for (var = 0; var < nvars; var++)
        term_put_nth(PP, (var+1), (modp->pp[modp->GB[i]][var] * iv_get_nth(R->weights, var+1)));
      summand = monomial_mk(c, PP);
      GBelem = arpoly_this_monomial2arpoly(summand, R);
      d = deg(nvars, modp->pp[modp->GB[i]]);
      for (top=modp->npp-1;top>0;top--)if(deg(nvars,modp->pp[top])==d)break;

      for (j=0; j < npoints; j++)
      {
        if (modular)
        {
          tmp = modp->M[modp->GB[i]][j+npoints];
          if (tmp == 0) continue;
          c = int2coeff(tmp, R);
        }
        else
        {
          if (mpq_sgn(char0->GB[i][j]) == 0) continue;
          c = arbrat2coeff(char0->GB[i][j], R);
        }
        for (n=top; n >= 0; n--) if (modp->issep[n] == j) break;
        PP = term_init(R->indetsNo);
        for (var = 0; var < nvars; var++)
          term_put_nth(PP, (var+1), modp->pp[n][var] * iv_get_nth(R->weights, var+1));
        summand = monomial_mk(c, PP);
        GBelem = arpoly_sum_these(GBelem, arpoly_this_monomial2arpoly(summand, R), R);
      }
      olist_set_nth(GB, i+1, arpoly2object(GBelem, R));
    }

    seps = olist_init(npoints);
    for (i=0; i < npoints; i++)
    {
      k = modp->projective_sep[i];
      d=deg(nvars,modp->pp[k]);for (top=modp->npp-1;top>=0;top--)if(deg(nvars,modp->pp[top])==d)break;
      sep = NULL;
      for (j=npoints-1; j >= 0; j--)
      {
        if (modular)
        {
          tmp = modp->M[k][j+npoints];
          if (tmp == 0) continue;
          c = int2coeff(tmp, R);
        }
        else
        {
          if (mpq_sgn(char0->sep[i][j]) == 0) continue;
          c = arbrat2coeff(char0->sep[i][j], R);
        }
        for (n=top; n >= 0; n--) if (modp->issep[n] == j) break;
        PP = term_init(R->indetsNo);
        for (var = 0; var < nvars; var++)
          term_put_nth(PP, (var+1), modp->pp[n][var] * iv_get_nth(R->weights, var+1));
        summand = monomial_mk(c, PP);
        sep = arpoly_sum_these(sep, arpoly_this_monomial2arpoly(summand, R), R);
      }
      olist_set_nth(seps, i+1, arpoly2object(sep, R));
    }

    BM_dtor(modp);
    if (modular) { FFdtor(Fp); for (i=0; i < npoints; i++) FREE(points_p[i]); FREE(points_p); }
    else { BMGB_dtor(char0); for (i=0; i < npoints; i++) { for (j=0; j < nvars; j++) mpz_clear(points[i][j]); FREE(points[i]);} FREE(points); }
    ans = olist_init(2);
    olist_set_nth(ans, 1, olist2object(GB));
    olist_set_nth(ans, 2, olist2object(seps));
    return olist2object(ans);
  }
}


/***************************************************************************/

static void omat_to_Qmat(Qmat *Mout, omat Min)
{
  int i, j, nrows, ncols;
  arbrat tmp;
  
  nrows = omat_nrow(Min);
  ncols = omat_ncol(Min);
  *Mout = Qmat_ctor(nrows, ncols);

  for (i=0; i < nrows; i++)
  {
    for (j=0; j < ncols; j++)
    {
      tmp = object_cast_arbrat(object_link(omat_ij(Min, i+1, j+1)));
      mpq_set((*Mout)->entry[i][j], tmp);
      mpq_clear(tmp); free_verylongrat(tmp);
    }
  }
}




/***************************************************************************/
/* GBM expects as input a list of lists of matrices (one for each ideal).  */
/* It assumes that 1 is always represented as the vector (1, 0, 0, ..., 0).*/

object GBM_interface(object args, ring R)
{
  int i, j, nvars, nideals, x, GBsize;
  int *quotdim, **sep;
  mpq_t **one;
  Qmat **Mmult, GBtails;
  olist GB, L, L1, input_matrices;
  omat omatrix;
  GBM trace;

  term PP;
  monomial summand;
  coeff c;
  arpoly GBelem;
  
  nvars = ring_indets_num(R);
  L1 = object2olist(object_link(args));
  L = object2olist(object_link(olist_nth(L1, 1)));
  nideals = olist_len(L);
  quotdim = (int*)malloc(nideals*sizeof(int));
  one = (mpq_t**)malloc(nideals*sizeof(mpq_t*));
  Mmult = (Qmat**)malloc(nideals*sizeof(Qmat*));
  for (i=0; i < nideals; i++)
  {
    input_matrices = object2olist(object_link(olist_nth(L, i+1)));
    Mmult[i] = (Qmat*)malloc(nvars*sizeof(Qmat));
    for (x=0; x < nvars; x++)
    {
      omatrix = object2omat(object_link(olist_nth(input_matrices, x+1)));
      omat_to_Qmat(&Mmult[i][x], omatrix);
      omat_free(omatrix);
    }
    olist_free(input_matrices);
    quotdim[i] = Mmult[i][0]->nrows;
    one[i] = (mpq_t*)malloc(quotdim[i]*sizeof(mpq_t));
    for (j=0; j < quotdim[i]; j++)
      mpq_init(one[i][j]);
    mpq_set_ui(one[i][0], 1, 1);
  }
  olist_free(L1);
  olist_free(L);
  pp_cmp_ring = R;
  pp_cmp_nvars = nvars;
  GBM_affine(&trace, &GBtails, nvars, nideals, quotdim, one, Mmult, pp_cmp);
  pp_cmp_ring = NULL;
  pp_cmp_nvars = 0;

  for (i=0; i < nideals; i++)
  {
    for (x=0; x < nvars; x++)
      Qmat_dtor(Mmult[i][x]);
    free(Mmult[i]);
    for (j=0; j < quotdim[i]; j++)
      mpq_clear(one[i][j]);
    free(one[i]);
  }
  free(Mmult);
  free(quotdim);
  free(one);

  sep = (int**)malloc(trace->nseps*sizeof(int*));
  j = 0;
  for (i=0; i < trace->nseps; i++)
  {
    while (trace->sep[j] < 0) j++;
    sep[i] = trace->pp[trace->sep[j]];
    j++;
  }
  
  GBsize = trace->GBsize;
  GB = olist_init(GBsize);
  for (i=0; i < GBsize; i++)
  {
    c = int2coeff(1, R);
    PP = term_init(R->indetsNo);
    for (x = 0; x < nvars; x++)
      term_put_nth(PP, (x+1), (trace->pp[trace->GB[i]][x] * iv_get_nth(R->weights, x+1)));
    summand = monomial_mk(c, PP);
    GBelem = arpoly_this_monomial2arpoly(summand, R);
    for (j=0; j < trace->nseps; j++)
    {
      if (mpq_sgn(GBtails->entry[j][i]) == 0) continue;
      c = arbrat2coeff(GBtails->entry[j][i], R);

      PP = term_init(R->indetsNo);
      for (x = 0; x < nvars; x++)
        term_put_nth(PP, (x+1), sep[j][x] * iv_get_nth(R->weights, x+1));
      summand = monomial_mk(c, PP);
      GBelem = arpoly_sum_these(GBelem, arpoly_this_monomial2arpoly(summand, R), R);
    }
    olist_set_nth(GB, i+1, arpoly2object(GBelem, R));
  }
  free(sep);
  Qmat_dtor(GBtails);
  GBM_dtor(trace);
  return olist2object(GB);
}

/* Same as GBM_interface but assumes coeff ring is a (small) finite field */
object GBM_interface_FF(object args, ring R)
{
  int i, j, nvars, nideals, Mshift, x, GBsize;
  int *quotdim;
  FF Fp;
  FFelem **one;
  FFmat **Mmult;
  olist GB, L, L1, input_matrices;
  omat omatrix;
  GBM trace;

  term PP;
  monomial summand;
  coeff c;
  arpoly GBelem;
  
  int p = ring_characteristic(R);
  if (!isprime(p)) { math_error = cocoa_error("BAD_CHAR"); return object_NULL; }
  Fp = FFctor(p);
  FFselect(Fp);
  nvars = ring_indets_num(R);
  L1 = object2olist(object_link(args));
  L = object2olist(object_link(olist_nth(L1, 1)));
  nideals = olist_len(L);
  quotdim = (int*)malloc(nideals*sizeof(int));
  one = (FFelem**)malloc(nideals*sizeof(FFelem*));
  Mmult = (FFmat**)malloc(nideals*sizeof(FFmat*));
  for (i=0; i < nideals; i++)
  {
    input_matrices = object2olist(object_link(olist_nth(L, i+1)));
    Mmult[i] = (FFmat*)malloc(nvars*sizeof(FFmat));
    for (x=0; x < nvars; x++)
    {
      omatrix = object2omat(object_link(olist_nth(input_matrices, x+1)));
      omat_to_FFmat(&Mmult[i][x], omatrix);
    }
    quotdim[i] = Mmult[i][0]->nrows;
    one[i] = (FFelem*)malloc(quotdim[i]*sizeof(FFelem));
    for (j=0; j < quotdim[i]; j++) one[i][j] = 0;
    one[i][0] = 1;
  }
  olist_free(L1);
  olist_free(L);
  pp_cmp_ring = R;
  pp_cmp_nvars = nvars;
  trace = GBM_affine_mod_p(nvars, nideals, quotdim, one, Mmult, pp_cmp);
  pp_cmp_ring = NULL;
  pp_cmp_nvars = 0;
  FFdtor(Fp); /* We do not need Fp hereafter. */

  Mshift = 0;
  for (i=0; i < nideals; ++i) Mshift += quotdim[i];
  GBsize = trace->GBsize;
  GB = olist_init(GBsize);
  for (i=0; i < GBsize; i++)
  {
    c = int2coeff(1, R);
    PP = term_init(R->indetsNo);
    for (x = 0; x < nvars; x++)
      term_put_nth(PP, (x+1), (trace->pp[trace->GB[i]][x] * iv_get_nth(R->weights, x+1)));
    summand = monomial_mk(c, PP);
    GBelem = arpoly_this_monomial2arpoly(summand, R);
    for (j=0; j < Mshift; j++)
    {
      if (trace->M[trace->GB[i]][j+Mshift] == 0) continue;
      c = int2coeff(trace->M[trace->GB[i]][j+Mshift], R);

      PP = term_init(R->indetsNo);
      for (x = 0; x < nvars; x++)
        term_put_nth(PP, (x+1), trace->pp[trace->sep[j]][x] * iv_get_nth(R->weights, x+1));
      summand = monomial_mk(c, PP);
      GBelem = arpoly_sum_these(GBelem, arpoly_this_monomial2arpoly(summand, R), R);
    }
    olist_set_nth(GB, i+1, arpoly2object(GBelem, R));
  }
  GBM_dtor(trace);
  return olist2object(GB);
}

/***************************************************************************/
/* W A R N I N G ! !                                                       */
/* I have used the "globals" pp_cmp_nvars and pp_cmp_ring whose values     */
/* MUST be left unaltered between calls to these three functions.          */
/* Failure to observe this restriction may lead to GRAVE PROBLEMS.         */

object do_NewPPStream(ring R)
{
  PPstream PPs;
  mpz_t *addr;

  pp_cmp_nvars = ring_indets_num(R);
  pp_cmp_ring = R;
  PPs = PPstream_ctor(ring_indets_num(R), pp_cmp); /* freed only when NextPP returns NULL, see below */
  addr = (mpz_t*)malloc(sizeof(mpz_t));
  mpz_init_set_ui(*addr, (unsigned long)PPs);
  return arbint2object(addr);
}


object do_NextPP(object o, ring R_unused)
{
  PPstream PPs;
  int var, *t;
  int nvars = pp_cmp_nvars;
  ring R = pp_cmp_ring;
  coeff c;
  term pp;
  monomial m;

  PPs = (PPstream)mpz_get_ui(*(mpz_t*)(o->v.arbint_value));
  t = PPstream_next(PPs);
  if (t == NULL) { PPstream_dtor(PPs); return arpoly2object(NULL, R); }
  c = int2coeff(1, R);
  pp = term_init(nvars);
  for (var = 0; var < nvars; var++)
    term_put_nth(pp, (var+1), t[var] * iv_get_nth(R->weights, var+1));
  FREE(t);
  m = monomial_mk(c, pp);
  return arpoly2object(arpoly_sum_these(NULL, arpoly_this_monomial2arpoly(m, R), R), R);
}


object do_AvoidLastPP(object o, ring R_unused)
{
  PPstream PPs;

  PPs = (PPstream)mpz_get_ui(*(mpz_t*)(o->v.arbint_value));
  PPstream_avoid_last(PPs);
  return object_NULL;
}


/***************************************************************************/

/* NextPrime returns 0 if arg is negative or too big */
object do_NextPrime(arbint N)
{
  FFelem n;
  mpz_t *Nptr = (mpz_t*)N;

  if (mpz_sgn(*Nptr) < 0) return int2object(0);
  if (mpz_cmp_ui(*Nptr, MAX_PRIME) >= 0) return int2object(0);
  n = mpz_get_ui(*Nptr);
  n = nextprime(n);
  return arbint2object(int2arbint(n)); /* cannot use int2object directly -- another of Antonio's little surprises... che gioia! */
}

/* NextPrime returns 0 if arg is negative or too big */
object do_NextPrime2(int N)
{
  FFelem n;

  if (N < 0) return int2object(0);
  if (N >= MAX_PRIME) return int2object(0);
  n = nextprime(N);
  return arbint2object(int2arbint(n)); /* cannot use int2object directly -- another of Antonio's little surprises... che gioia! */
}


object do_Mod2Rat(arbint res, arbint mod, arbint Q)
{
  int wrong;
  mpz_t *pres = (mpz_t*)res;
  mpz_t *pmod = (mpz_t*)mod;
  mpz_t *pQ = (mpz_t*)Q;
  mpz_t den;
  mpz_t num;
  object ans;

  mpz_init(den);
  mpz_init(num);
  wrong = modular_to_rational(num, den, *pQ, *pres, *pmod);
  if (wrong) { mpz_set_ui(num, 0); mpz_set_ui(den, 1); }
  ans = arbrat2object(arbrat_mk(&num, &den));
  mpz_clear(den);
  mpz_clear(num);
  return ans;
}
