#include "mpz_lcm.h"

#if __GNU_MP_VERSION==2

/***************************************************************************/
/* dest := LCM(dest, other); aliasing NOT permitted!                       */
static void mpz_lcm2(mpz_t dest, mpz_t other)
{
  mpz_t gcd;

  mpz_init(gcd);
  mpz_gcd(gcd, dest, other);
  mpz_divexact(dest, dest, gcd);
  mpz_clear(gcd);
  mpz_mul(dest, dest, other);
  if (mpz_sgn(dest) < 0) mpz_neg(dest, dest);
}

/***************************************************************************/
/* Compute LCM of arg1 and arg2 putting result into dest.                  */
/* Aliasing between args is permitted.                                     */

void mpz_lcm(mpz_t dest, mpz_t arg1, mpz_t arg2)
{
  if (dest == arg1)
  {
    if (dest == arg2) return;
    else { mpz_lcm2(dest, arg2); return; }
  }
  if (dest == arg2) { mpz_lcm2(dest, arg1); return; }
  mpz_set(dest, arg2);
  mpz_lcm2(dest, arg1);
}

#endif
