#include "mpz_lift_recip.h"

/* Lift a reciprocal value using Hensel lifting                          */
/* Assume that inputs satisfy for some integer M:                        */
/*   recip*value = 1 mod M  and                                          */
/*   M | modulus  and                                                    */
/*   modulus | M^2                                                       */
/* Upon return recip will be altered so that recip*value = 1 mod modulus */

void mpz_lift_recip(mpz_t recip, mpz_t value, mpz_t modulus)
{
  mpz_t tmp;

  mpz_init(tmp);
  mpz_fdiv_r(tmp, value, modulus);
  mpz_mul(tmp, tmp, recip);
  mpz_fdiv_r(tmp, tmp, modulus);
  mpz_neg(tmp, tmp);
  mpz_add_ui(tmp, tmp, 2);
  mpz_mul(tmp, tmp, recip);
  mpz_fdiv_r(recip, tmp, modulus);
  mpz_clear(tmp);
}
