#include <stdio.h>

#include "primes.h"
#include "FindPrimRoot.h"
#include "FF.h"
#include "DUPFF.h"
#include "DUPFFprint.h"


void singletest(FFelem p, int d);

int main(int argc, char **argv)
{
  singletest(19, 5);
  singletest(19, 18);
  singletest(97, 5);
  singletest(97, 25);
  singletest(3571, 5);
  singletest(3571, 15);
  singletest(3571, 25);
  singletest(7, 19);
  singletest(3, 37);
  return(0);
}

void singletest(FFelem p, int d)
{
  FF K;
  DUPFF f, g, h, k;
  int i;

  printf("Starting test for p=%d and d=%d\n", p, d);
  K = FFctor(p);
  FFselect(K);
  f = DUPFFnew(d);
  g = DUPFFnew(d);
  for (i=0; i<=d; i++) f->coeffs[i] = (d+i)%p;
  g->coeffs[0] = p/2;
  for (i=1; i<=d; i++) g->coeffs[i] = (g->coeffs[i-1]+f->coeffs[i])%p;
  f->deg = d; if (f->coeffs[d] == 0) f->coeffs[d] = p/3+1;
  g->deg = d; if (g->coeffs[d] == 0) g->coeffs[d] = 2*p/3-1;


  printf("f is ");
  DUPFFprint(f);
  printf("g is ");
  DUPFFprint(g);

  h = DUPFFnew(2*d);
  DUPFFmul3(h, g, f);
  if (DUPFFdeg(h) != DUPFFdeg(f)+DUPFFdeg(g))
    printf("degree loss in multiply\n");
  printf("h=f*g\n");

  if (!DUPFFequal(DUPFFdiv(h,g), f))/*memory leak*/
    printf("h/g != f\n");

  if (!DUPFFequal(DUPFFdiv(h,f), g))/*memory leak*/
    printf("h/f != g\n");
  if (DUPFFrem(h,g)->deg >= 0)/*memory leak*/
    printf("h mod g is not zero\n");
  if (DUPFFrem(h,f)->deg >= 0)/*memory leak*/
    printf("h mod f is not zero\n");

  k = DUPFFadd(h, f);
  if (!DUPFFequal(DUPFFrem(k,g),
                  DUPFFrem(f,g)))/*double memory leak*/
    printf("h+f != f mod g\n");
  DUPFFfree(k);

  k = DUPFFadd(h, g);
  if (!DUPFFequal(DUPFFrem(k,f),
                  DUPFFrem(g,f)))/*double memory leak*/
    printf("h+g != g mod f\n");

  DUPFFfree(k);
  DUPFFfree(f);
  DUPFFfree(g);
  DUPFFfree(h);
  FFdtor(K);
  printf("test for p=%d and d=%d complete\n\n\n\n\n\n", p, d);
}
