/* Testing whether 3-arg multiplication works when the output is the same
   as the one of the inputs (or even both inputs). Similarly for div4. */

#include <stdio.h>

#include "primes.h"
#include "FindPrimRoot.h"
#include "FF.h"
#include "DUPFF.h"
#include "DUPFFprint.h"


void singletest(FFelem p, int d);

int main(int argc, char **argv)
{
  singletest(19, 5);
  singletest(19, 18);
  singletest(97, 5);
  singletest(97, 25);
  singletest(3571, 5);
  singletest(3571, 15);
  singletest(3571, 25);
  singletest(7, 19);
  singletest(3, 37);
  return(0);
}

void singletest(FFelem p, int d)
{
  FF K;
  DUPFF f, g, h, k, zero, junk;
  int i;

  printf("Starting test for p=%d and d=%d\n", p, d);
  K = FFctor(p);
  FFselect(K);
  f = DUPFFnew(2*d);
  for (i=0; i<=d; i++) f->coeffs[i] = (4*i*i+7*i+d)%p;
  f->deg = d; if (f->coeffs[d] == 0) f->coeffs[d] = p/3+1;
  g = DUPFFnew(2*d);
  DUPFFcopy2(g, f);
  k = DUPFFnew(2*d);
  junk = DUPFFnew(2*d);

  printf("f is ");
  DUPFFprint(f);
  printf("g:=f\n");

  h = DUPFFnew(2*d);
  DUPFFmul3(h, g, f);
  DUPFFmul3(k, f, f);
  if (!DUPFFequal(h, k))
    printf("f*f != g*f\n");
  DUPFFcopy2(junk, f);
  DUPFFsquare(junk);
  if (!DUPFFequal(junk, k))
    printf("********square(f) != f*f********\n");

  zero = DUPFFnew(-1);
  DUPFFdiv4(k, zero, h, f);
  if (!DUPFFequal(k, f))
    printf("(f*f)/f != f\n");

  DUPFFcopy2(k, f);
  DUPFFmul3(g, g, f);
  DUPFFmul3(k, f, k);
  DUPFFmul3(f, f, f);
  if (!DUPFFequal(h, f))
    printf("h != f\n");
  if (!DUPFFequal(g, f))
    printf("g != f\n");
  if (!DUPFFequal(k, f))
    printf("k != f\n");
  if (!DUPFFequal(h, f))
    printf("h != f\n");

  /* now test division */
  /* create some (not very) random polyomials */
  for (i=0; i<=d; i++) f->coeffs[i] = (4*i*i+7*i+d)%p;
  f->deg = d; if (f->coeffs[d] == 0) f->coeffs[d] = p/3+1;
  for (i=0; i<=d; i++) g->coeffs[i] = (i*i*i+3*i+8)%p;
  g->deg = d; if (g->coeffs[d] == 0) g->coeffs[d] = p-2;
  DUPFFdiv4(h, k, f, g);
  DUPFFmul3(h, h, g);
  DUPFFadd3(h, h, k);
  DUPFFsub3(h, h, f);
  if (DUPFFdeg(h) >= 0)
    printf("********div4(h, k, f, g) FAILED********\n");
  DUPFFdiv4(h, k, f, g);
  DUPFFcopy2(h, f);
  DUPFFdiv4(junk, h, h, g);
  if (!DUPFFequal(k, h))
    printf("********div4(junk, h, h, g) FAILED********\n");
  DUPFFdiv4(h, k, f, g);
  DUPFFcopy2(k, f);
  DUPFFdiv4(k, junk, k, g);
  if (!DUPFFequal(k, h))
    printf("********div4(k, junk, k, g) FAILED********\n");


  DUPFFfree(zero);
  DUPFFfree(junk);
  DUPFFfree(k);
  DUPFFfree(f);
  DUPFFfree(g);
  DUPFFfree(h);
  FFdtor(K);
  printf("test for p=%d and d=%d complete\n\n\n\n\n\n", p, d);
}
