#include <stdio.h>

#include "primes.h"
#include "FindPrimRoot.h"
#include "FF.h"
#include "DUPFF.h"
#include "DUPFFprint.h"


/***************************************************************************/
/*  Somewhat haphazard test file for gcds                                  */
/***************************************************************************/


void singletest(FFelem p, int d);

int main(int argc, char **argv)
{
  singletest(19, 5);
  singletest(19, 18);
  singletest(97, 5);
  singletest(97, 25);
  singletest(3571, 5);
  singletest(3571, 15);
  singletest(3571, 25);
  singletest(7, 19);
  singletest(3, 37);
  return(0);
}


void check(const DUPFF f, const DUPFF g, const DUPFF h)
{
  DUPFF k, l, q, r, s;
  int d;

  d = DUPFFdeg(f); /* f, g, and h all have same degree */
  q = DUPFFnew(d);
  r = DUPFFnew(d);
  s = DUPFFnew(d);

  /* the code below leaks memory like crazy */
  k = DUPFFgcd(f, g);
  printf("deg(gcd)=%d\n", DUPFFdeg(k));
  DUPFFdiv4(q, r, f, k);
  if (r->deg >= 0) printf("f mod gcd(f, g) != 0\n");
  DUPFFdiv4(s, r, g, k);
  if (r->deg >= 0) printf("g mod gcd(f, g) != 0\n");
  if (DUPFFdeg(DUPFFgcd(q, s)) != 0)
    printf("gcd(f/gcd(f, g), g/gcd(f, g)) != 1\n");
  l = DUPFFgcd(DUPFFmul(f, h), DUPFFmul(g, h));
  DUPFFdiv4(q, r, l, k);
  if (r->deg >= 0) printf("gcd(f, g) | gcd(f*h, g*h) fails\n");
  DUPFFdiv4(s, r, q, h);
  if (r->deg >= 0) printf("h | gcd(f*h, g*h)/gcd(f, g) fails\n");

  DUPFFcopy2(r, f);
  DUPFFcopy2(s, g);
  DUPFFgcd2(r, s);
  if (DUPFFdeg(k) != DUPFFdeg(r) || DUPFFdeg(DUPFFrem(k, r)) >= 0)
    printf("gcd(f, g) != gcd2(f, g)\n");
}


void singletest(FFelem p, int d)
{
  FF K;
  DUPFF f, g, h;
  int i;

  printf("Starting test for p=%d and d=%d\n", p, d);
  K = FFctor(p);
  FFselect(K);
  f = DUPFFnew(d);
  g = DUPFFnew(d);
  h = DUPFFnew(d);
  for (i=0; i<=d; i++) f->coeffs[i] = (4*i*i+7*i+3)%p;
  for (i=0; i<=d; i++) g->coeffs[i] = (i*i+17*i+9)%p;
  for (i=0; i<=d; i++) h->coeffs[i] = (i*i*i+3*i*i+7*i+4)%p;
  f->deg = d; if (f->coeffs[d] == 0) f->coeffs[d] = p/3+1;
  g->deg = d; if (g->coeffs[d] == 0) g->coeffs[d] = 2*p/3-1;
  h->deg = d; if (h->coeffs[d] == 0) h->coeffs[d] = 3*p/5+1;

  printf("f is ");
  DUPFFprint(f);
  printf("g is ");
  DUPFFprint(g);
  printf("h is ");
  DUPFFprint(h);

  printf("checking fgh...\n");
  check(f, g, h);
  printf("checking ghf...\n");
  check(g, h, f);
  printf("checking hfg...\n");
  check(h, f, g);
  FFdtor(K);
  printf("test for p=%d and d=%d complete\n\n\n\n\n\n", p, d);
}
