#include <stdio.h>

#include "DUPFF.h"
#include "DUPFFprint.h"
#include "DUPFFddf.h"
#include "DUPFFsqfrd.h"
#include "jerror.h"
#include "FF.h"
#include "DUPFFfactor.h"


void singletest(FFelem p, int d);

int main()
{
  singletest(19, 5);
  singletest(19, 17);
  singletest(19, 18);
  singletest(97, 5);
  singletest(97, 25);
  singletest(7, 19);
  singletest(3, 37);
  singletest(3, 38);

  if (LAST_ERROR != 0)
    printf("********ERROR %d OCCURRED********\n", LAST_ERROR);

  return(0);
}


void check(const DUPFF f)
{
  DUPFF tmp;
  int j;
  DUPFFlist facs, iter, facs2;

  facs = DUPFFfactor(f);
  printf("Number of irreducible factors is %d\n", DUPFFlist_length(facs));

  tmp = DUPFFnew(2*DUPFFdeg(f));
  tmp->deg = 0;
  tmp->coeffs[0] = 1;
  /* Multiply out the factorization, and check it is right */
  for (iter=facs; iter; iter = iter->next)
  {
    for (j=0; j < iter->deg; j++) DUPFFmul3(tmp, tmp, iter->poly);
    facs2 = DUPFFfactor(iter->poly);
    if (DUPFFlist_length(facs2) != 1 || facs2->deg != 1)
      printf("********IRREDUCIBLE ISN'T********\n");
    DUPFFlist_dtor(facs2);
  }
  if (DUPFFdeg(tmp) != DUPFFdeg(f))
    printf("********DEG ERROR IN FACTORIZATION********\n");
  DUPFFrem2(tmp, f);
  if (DUPFFdeg(tmp) >= 0)
    printf("********FACTORIZATION WRONG********\n");


  /* clean up allocated memory */
  DUPFFfree(tmp);
  DUPFFlist_dtor(facs);

  printf("Check complete.\n\n\n");
}


void singletest(FFelem p, int d)
{
  FF K;
  DUPFF h;
  int i;

  printf("Starting test for p=%d and d=%d\n", p, d);
  K = FFctor(p);
  FFselect(K);

  h = DUPFFnew(d);
  for (i=0; i<=d; i++) h->coeffs[i] = (i*i*i+3*i*i+7*i+4)%p;
  h->deg = d; if (h->coeffs[d] == 0) h->coeffs[d] = 3*p/5+1;

  printf("h is ");
  DUPFFprint(h);

  printf("checking irreducible factorization for h...\n");
  check(h);

  DUPFFfree(h);
  FFdtor(K);
  
  printf("test for p=%d and d=%d complete\n\n\n\n\n\n", p, d);
}
