/* simple test for polynomials with GMP integer coefficients */

#include <stdio.h>

#include "jerror.h"
#include "DUPZ.h"


void singletest(int d);

int main(int argc, char **argv)
{
  singletest(1);
  singletest(2);
  singletest(4);
  singletest(8);
  singletest(16);
  singletest(32);
  singletest(64);

  if (LAST_ERROR != 0)
    printf("********ERROR %d OCCURRED********\n", LAST_ERROR);

  return(0);
}


void singletest(int d)
{
  int i;
  DUPZ f, g, h, tmp;

  printf("doing test for degree %d\n", d);

  f = DUPZnew(2*d);
  g = DUPZnew(2*d);
  h = DUPZnew(2*d);

  f->deg = d;
  for (i=0; i <= d; i++) mpz_set_si(f->coeffs[i], 1234567*i + 987654321);

  DUPZmul3(g, f, f);
  if (DUPZdeg(g) != 2*DUPZdeg(f))
    printf("deg(f^2) != 2*deg(f)\n");
  DUPZcopy2(h, f);
  if (!DUPZequal(h, f))
    printf("h != f\n");
  DUPZmul3(h, h, f);
  if (!DUPZequal(g, h))
    printf("f*f != h*f\n");
  /*  DUPZmul3(f, f, f);  NOT YET IMPLEMENTED */

  tmp = DUPZdiv(h, f);
  if (!DUPZequal(tmp, f))
    printf("(f^2)/f != f\n");
  DUPZfree(tmp);

  tmp = DUPZdiv(h, h);
  if (DUPZdeg(tmp) != 0)
    printf("h/h does not have degree 0\n");
  DUPZfree(tmp);

  tmp = DUPZadd(h, f);
  DUPZdiv4(h, tmp, tmp, f);
  DUPZrem2(h, f);
  if (DUPZdeg(h) != 0)
    printf("deg(remainder((f^2+f)/f, f)) != 0\n");
  
  DUPZfree(tmp);
  DUPZfree(h);
  DUPZfree(g);
  DUPZfree(f);
}
