#include <stdio.h>
#include <stdlib.h>

#include "jerror.h"
#include "DUPZ.h"
#include "DUPZfactors.h"
#include "DUPZfactor.h"

int main()
{
  char comment[999];
  int d, i, last_char, error_count, nfactors;
  DUPZ f, zero;
  DUPZfactors facs;
  DUPZlist iter;

  error_count = 0;
  zero = DUPZnew(-1); /* used in a call to DUPZdiv4 below */
  while(1)
  {
    scanf("%[^\n]", comment);
    last_char = scanf("%d", &d);
    if (last_char == EOF) break;
    printf("%s\nDegree is %d\n", comment, d);
    f = DUPZnew(d);
    for (i=d; i >= 0; i--)
    {
      mpz_inp_str(f->coeffs[i], stdin, 10);
    }
    scanf("\nnfactors=%d", &nfactors); /* the number of irreducible factors */
    scanf("\n"); /* Need to do this so we read the next comment properly */

    for (i=d; (i >= 0) && (mpz_sgn(f->coeffs[i]) == 0); i--);
    f->deg = i;

    facs = DUPZfactor(f);
    
    d = DUPZfactors_nfactors(facs);
    if (d != nfactors)
    {
      printf("***** WRONG NUMBER OF FACTORS (%d instead of %d)\n", d, nfactors);
      error_count++;
    }
    if (d == 1) printf("The polynomial is irreducible\n");
    else printf("There are %d factors.\n", d);

    for (iter=facs->list; iter; iter = iter->next)
      for (i=iter->deg; i > 0; i--)
        DUPZdiv4(f, zero, f, iter->poly);
    if (DUPZdeg(f) != 0)
    {
      printf("***** DEGREE ERROR IN FACTORIZATION!\n");
      error_count++;
    }
    if (mpz_cmp(f->coeffs[0], facs->content) != 0)
    {
      printf("***** COEFFICIENT ERROR IN FACTORIZATION!\n");
      error_count++;
    }
    DUPZfactors_dtor(facs);
    DUPZfree(f);
    if (LAST_ERROR != 0)
      printf("********ERROR %d OCCURRED********\n", LAST_ERROR);
    printf("\n");
  }

  DUPZfree(zero);
  if (error_count != 0) printf("There were %d erroneous factorizations\n", error_count);

  return(0);
}
